@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -89,9 +90,9 @@ Z_to_raw(GEN x, GEN ff)
   case t_FF_FpXQ:
     return scalarpol(x, varn(T));
   case t_FF_F2xq:
-    return Z_to_F2x(x,varn(T));
+    return Z_to_F2x(x, T[1]);
   default:
-    return Z_to_Flx(x,pp,T[1]);
+    return Z_to_Flx(x, pp, T[1]);
   }
 }
 
@@ -109,7 +110,7 @@ Rg_to_raw(GEN x, GEN ff)
     return gel(x,2);
   }
   pari_err_TYPE("Rg_to_raw",x);
-  return NULL; /* LCOV_EXCL_LINE */
+  return NULL;/* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -132,12 +133,9 @@ FFX_to_raw(GEN x, GEN ff)
 }
 
 static GEN
-FFC_to_raw(GEN x, GEN ff)
-{ pari_APPLY_same(Rg_to_raw(gel(x, i), ff)) }
-
+FFC_to_raw(GEN x, GEN ff) { pari_APPLY_same(Rg_to_raw(gel(x, i), ff)) }
 static GEN
-FFM_to_raw(GEN x, GEN ff)
-{ pari_APPLY_same(FFC_to_raw(gel(x, i), ff)) }
+FFM_to_raw(GEN x, GEN ff) { pari_APPLY_same(FFC_to_raw(gel(x, i), ff)) }
 
 /* in place */
 static GEN
@@ -204,15 +202,11 @@ Fq_to_FF(GEN x, GEN ff)
 
 static int
 FF_samechar(GEN x, GEN y)
-{
-  return x[1] == y[1] && equalii(gel(x,4),gel(y,4));
-}
+{ return x[1] == y[1] && equalii(gel(x,4),gel(y,4)); }
 
 int
 FF_samefield(GEN x, GEN y)
-{
-  return FF_samechar(x, y) && gidentical(gel(x,3),gel(y,3));
-}
+{ return FF_samechar(x, y) && gidentical(gel(x,3),gel(y,3)); }
 
 int
 FF_equal(GEN x, GEN y)
@@ -239,11 +233,7 @@ FF_equal1(GEN x)
 
 static int
 Fp_cmp_1(GEN x, GEN p)
-{
-  pari_sp av = avma;
-  int b = equalii(x, addis(p,-1));
-  avma = av; return b;
-}
+{ pari_sp av = avma; return gc_bool(av, equalii(x, addis(p,-1))); }
 
 int
 FF_equalm1(GEN x)
@@ -364,6 +354,19 @@ FF_mod(GEN x)
   }
 }
 
+long
+FF_var(GEN x)
+{
+  switch(x[1])
+  {
+  case t_FF_FpXQ:
+    return varn(gel(x,3));
+  case t_FF_F2xq:
+  default:
+    return gel(x,3)[1]>>VARNSHIFT;
+  }
+}
+
 long
 FF_f(GEN x)
 {
@@ -606,31 +609,30 @@ FF_Frobenius(GEN x, long e)
   ulong pp;
   GEN r, T, p, z=_initFF(x,&T,&p,&pp);
   ulong n = umodsu(e, FF_f(x));
+  pari_sp av = avma;
   if (n==0) return gcopy(x);
   switch(x[1])
   {
   case t_FF_FpXQ:
-    r=FpXQ_pow(gel(x,2),p,T,p);
+    r=FpX_Frobenius(T,p);
     if (n>1) r=FpXQ_autpow(r,n,T,p);
     break;
   case t_FF_F2xq:
-    r=F2xq_sqr(gel(x,2),T);
+    r=F2x_Frobenius(T);
     if (n>1) r=F2xq_autpow(r,n,T);
     break;
   default:
-    r=Flxq_powu(gel(x,2),pp,T,pp);
+    r=Flx_Frobenius(T,pp);
     if (n>1) r=Flxq_autpow(r,n,T,pp);
   }
+  r = gerepileupto(av, r);
   return _mkFF(x,z,r);
 }
 
-GEN
-FFX_preimage(GEN x, GEN F, GEN y)
+static GEN
+FFX_preimage_i(GEN x, GEN y, GEN F, GEN T, GEN p, long pp)
 {
-  GEN r, T, p, z;
-  ulong pp;
-  if (FF_equal0(x)) return FF_zero(y);
-  z=_initFF(y,&T,&p,&pp);
+  GEN r;
   F = FFX_to_raw(F, y);
   switch(y[1])
   {
@@ -643,11 +645,34 @@ FFX_preimage(GEN x, GEN F, GEN y)
   default:
     r = FlxqX_rem(Flx_to_FlxX(gel(x,2),T[1]), F, T, pp);
   }
+  return r;
+}
+
+GEN
+FFX_preimage(GEN x, GEN F, GEN y)
+{
+  GEN r, T, p, z;
+  ulong pp;
+  if (FF_equal0(x)) return FF_zero(y);
+  z = _initFF(y,&T,&p,&pp);
+  r = FFX_preimage_i(x, y, F, T, p, pp);
   if (degpol(r) > 0) return NULL;
   r = (y[1] == t_FF_FpXQ)? Fq_to_FpXQ(gel(r,2),T, p): gel(r,2);
   return _mkFF(y,z,r);
 }
 
+GEN
+FFX_preimagerel(GEN x, GEN F, GEN y)
+{
+  pari_sp av = avma;
+  GEN r, T, p;
+  ulong pp;
+  if (FF_equal0(x)) return FF_zero(y);
+  _getFF(y,&T,&p,&pp);
+  r = FFX_preimage_i(x, y, F, T, p, pp);
+  return gerepilecopy(av, raw_to_FFX(r, y));
+}
+
 GEN
 FF_mul(GEN x, GEN y)
 {
@@ -814,7 +839,7 @@ Z_FF_div(GEN n, GEN x)
     break;
   case t_FF_F2xq:
     r = F2xq_inv(A,T); /*Check for division by 0*/
-    if(!mpodd(n)) { avma = av; r = zero_Flx(A[1]); }
+    if(!mpodd(n)) { set_avma(av); r = zero_Flx(A[1]); }
     break;
   default:
     r = gerepileupto(av, Flx_Fl_mul(Flxq_inv(A,T,pp),umodiu(n,pp),pp));
@@ -914,7 +939,7 @@ FF_ispower(GEN x, GEN K, GEN *pt)
     r = Flxq_sqrtn(gel(x,2),K,T,pp,NULL);
     break;
   }
-  if (!r) { avma = av; return 0; }
+  if (!r) return gc_long(av,0);
   if (pt) { *pt = z; (void)_mkFF(x,z,r); }
   return 1;
 }
@@ -1217,7 +1242,7 @@ FqV_to_FpXQV(GEN x, GEN T)
       gel(y,i) = scalarpol(gel(x,i), v);
       s = 1;
     }
-  if (!s) { avma = av; return x; }
+  if (!s) { set_avma(av); return x; }
   return y;
 }
 
@@ -1395,9 +1420,7 @@ FF_elltwist(GEN E)
 }
 
 static long
-F3x_equalm1(GEN x)
-{ return degpol(x)==0 && x[2] == 2; }
-
+F3x_equalm1(GEN x) { return degpol(x)==0 && x[2] == 2; }
 GEN
 FF_ellrandom(GEN E)
 {
@@ -1607,9 +1630,9 @@ FFX_roots(GEN Pf, GEN ff)
 }
 
 static GEN
-raw_to_FFXC(GEN x, GEN ff)
-{ pari_APPLY_type(t_COL, raw_to_FFX(gel(x,i), ff)); }
-
+raw_to_FFXC(GEN x, GEN ff) { pari_APPLY_type(t_COL, raw_to_FFX(gel(x,i), ff)); }
+static GEN
+raw_to_FFXM(GEN x, GEN ff) { pari_APPLY_same(raw_to_FFXC(gel(x,i), ff)); }
 static GEN
 raw_to_FFX_fact(GEN F, GEN ff)
 {
@@ -1636,6 +1659,30 @@ FFX_zero(GEN ff, long v)
   return r;
 }
 
+GEN
+FFX_add(GEN Pf, GEN Qf, GEN ff)
+{
+  pari_sp av = avma;
+  GEN r,T,p;
+  ulong pp;
+  GEN P = FFX_to_raw(Pf, ff);
+  GEN Q = FFX_to_raw(Qf, ff);
+  _getFF(ff,&T,&p,&pp);
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    r = FpXX_add(P, Q, p);
+    break;
+  case t_FF_F2xq:
+    r = F2xX_add(P, Q);
+    break;
+  default:
+    r = FlxX_add(P, Q, pp);
+  }
+  if (!lgpol(r)) { set_avma(av); return FFX_zero(ff, varn(Pf)); }
+  return gerepilecopy(av, raw_to_FFX(r, ff));
+}
+
 static GEN
 FFX_wrap2(GEN Pf, GEN Qf, GEN ff, GEN FpXQX(GEN, GEN, GEN, GEN),
           GEN F2xqX(GEN, GEN, GEN), GEN FlxqX(GEN, GEN, GEN, ulong))
@@ -1657,7 +1704,7 @@ FFX_wrap2(GEN Pf, GEN Qf, GEN ff, GEN FpXQX(GEN, GEN, GEN, GEN),
   default:
     r = FlxqX(P, Q, T, pp);
   }
-  if (!lgpol(r)) { avma = av; return FFX_zero(ff, varn(Pf)); }
+  if (!lgpol(r)) { set_avma(av); return FFX_zero(ff, varn(Pf)); }
   return gerepilecopy(av, raw_to_FFX(r, ff));
 }
 
@@ -1665,6 +1712,10 @@ GEN
 FFX_mul(GEN Pf, GEN Qf, GEN ff)
 { return FFX_wrap2(Pf, Qf, ff, FpXQX_mul, F2xqX_mul, FlxqX_mul); }
 
+GEN
+FFX_gcd(GEN Pf, GEN Qf, GEN ff)
+{ return FFX_wrap2(Pf, Qf, ff, FpXQX_gcd, F2xqX_gcd, FlxqX_gcd); }
+
 GEN
 FFX_sqr(GEN Pf, GEN ff)
 {
@@ -1684,7 +1735,7 @@ FFX_sqr(GEN Pf, GEN ff)
   default:
     r = FlxqX_sqr(P, T, pp);
   }
-  if (!lgpol(r)) { avma = av; return FFX_zero(ff, varn(Pf)); }
+  if (!lgpol(r)) { set_avma(av); return FFX_zero(ff, varn(Pf)); }
   return gerepilecopy(av, raw_to_FFX(r, ff));
 }
 
@@ -1692,6 +1743,108 @@ GEN
 FFX_rem(GEN Pf, GEN Qf, GEN ff)
 { return FFX_wrap2(Pf, Qf, ff, FpXQX_rem, F2xqX_rem, FlxqX_rem); }
 
+GEN
+FFX_resultant(GEN Pf, GEN Qf, GEN ff)
+{
+  pari_sp av = avma;
+  GEN r,T,p;
+  ulong pp;
+  GEN P = FFX_to_raw(Pf, ff);
+  GEN Q = FFX_to_raw(Qf, ff);
+  GEN z = _initFF(ff,&T,&p,&pp);
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    r = FpXQX_resultant(P, Q, T, p);
+    break;
+  case t_FF_F2xq:
+    r = F2xqX_resultant(P, Q, T);
+    break;
+  default:
+    r = FlxqX_resultant(P, Q, T, pp);
+  }
+  return gerepileupto(av, _mkFF(ff,z,r));
+}
+
+GEN
+FFX_disc(GEN Pf, GEN ff)
+{
+  pari_sp av = avma;
+  GEN r,T,p;
+  ulong pp;
+  GEN P = FFX_to_raw(Pf, ff);
+  GEN z = _initFF(ff,&T,&p,&pp);
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    r = FpXQX_disc(P, T, p);
+    break;
+  case t_FF_F2xq:
+    r = F2xqX_disc(P, T);
+    break;
+  default:
+    r = FlxqX_disc(P, T, pp);
+  }
+  return gerepileupto(av, _mkFF(ff,z,r));
+}
+
+static GEN
+gc_gcdext(pari_sp av, GEN r, GEN *u, GEN *v)
+{
+  if (!u && !v) return gerepilecopy(av, r);
+  if (u  &&  v) gerepileall(av, 3, &r, u, v);
+  else          gerepileall(av, 2, &r, u ? u: v);
+  return r;
+}
+
+GEN
+FFX_extgcd(GEN Pf, GEN Qf, GEN ff, GEN *pt_Uf, GEN *pt_Vf)
+{
+  pari_sp av = avma;
+  GEN r,T,p;
+  ulong pp;
+  GEN P = FFX_to_raw(Pf, ff);
+  GEN Q = FFX_to_raw(Qf, ff);
+  _getFF(ff,&T,&p,&pp);
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    r = FpXQX_extgcd(P, Q, T, p, pt_Uf, pt_Vf);
+    break;
+  case t_FF_F2xq:
+    r = F2xqX_extgcd(P, Q, T, pt_Uf, pt_Vf);
+    break;
+  default:
+    r = FlxqX_extgcd(P, Q, T, pp, pt_Uf, pt_Vf);
+  }
+  if (pt_Uf) *pt_Uf = raw_to_FFX(*pt_Uf, ff);
+  if (pt_Vf) *pt_Vf = raw_to_FFX(*pt_Vf, ff);
+  return gc_gcdext(av, raw_to_FFX(r, ff), pt_Uf, pt_Vf);
+}
+
+GEN
+FFX_halfgcd(GEN Pf, GEN Qf, GEN ff)
+{
+  pari_sp av = avma;
+  GEN r,T,p;
+  ulong pp;
+  GEN P = FFX_to_raw(Pf, ff);
+  GEN Q = FFX_to_raw(Qf, ff);
+  _getFF(ff,&T,&p,&pp);
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    r = FpXQX_halfgcd(P, Q, T, p);
+    break;
+  case t_FF_F2xq:
+    r = F2xqX_halfgcd(P, Q, T);
+    break;
+  default:
+    r = FlxqX_halfgcd(P, Q, T, pp);
+  }
+  return gerepilecopy(av, raw_to_FFXM(r, ff));
+}
+
 GEN
 FFXQ_sqr(GEN Pf, GEN Qf, GEN ff)
 { return FFX_wrap2(Pf, Qf, ff, FpXQXQ_sqr, F2xqXQ_sqr, FlxqXQ_sqr); }
@@ -1721,7 +1874,30 @@ FFXQ_mul(GEN Pf, GEN Qf, GEN Sf, GEN ff)
   default:
     r = FlxqXQ_mul(P, Q, S, T, pp);
   }
-  if (!lgpol(r)) { avma = av; return FFX_zero(ff, varn(Pf)); }
+  if (!lgpol(r)) { set_avma(av); return FFX_zero(ff, varn(Pf)); }
+  return gerepilecopy(av, raw_to_FFX(r, ff));
+}
+
+GEN
+FFXQ_minpoly(GEN Pf, GEN Qf, GEN ff)
+{
+  pari_sp av = avma;
+  GEN r,T,p;
+  ulong pp;
+  GEN P = FFX_to_raw(Pf, ff);
+  GEN Q = FFX_to_raw(Qf, ff);
+  _getFF(ff,&T,&p,&pp);
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    r = FpXQXQ_minpoly(P, Q, T, p);
+    break;
+  case t_FF_F2xq:
+    r = FlxX_to_F2xX(FlxqXQ_minpoly(F2xX_to_FlxX(P), F2xX_to_FlxX(Q), F2x_to_Flx(T), 2UL));
+    break;
+  default:
+    r = FlxqXQ_minpoly(P, Q, T, pp);
+  }
   return gerepilecopy(av, raw_to_FFX(r, ff));
 }
 
@@ -1746,10 +1922,10 @@ FFX_ispower(GEN Pf, long k, GEN ff, GEN *pt_r)
   default:
     s = FlxqX_ispower(P, k, T, pp, pt_r);
   }
-  if (s==0) { avma = av; return 0; }
+  if (s==0) return gc_long(av,0);
   if (pt_r)
     *pt_r = gerepilecopy(av, raw_to_FFX(*pt_r, ff));
-  else avma = av;
+  else set_avma(av);
   return 1;
 }
 
@@ -1886,7 +2062,7 @@ ffgen(GEN T, long v)
       }
     default:
       pari_err_TYPE("ffgen",T);
-      return NULL;
+      return NULL;/* LCOV_EXCL_LINE */
   }
   if (v < 0) v = varn(T);
   if (lgefint(p)==3)
@@ -2105,7 +2281,7 @@ FFM_wrap(GEN M, GEN ff, GEN (*Fq)(GEN,GEN,GEN),
   case t_FF_F2xq: M = F2xq(M,T); break;
   default: M = Flxq(M,T,pp); break;
   }
-  if (!M) { avma = av; return NULL; }
+  if (!M) return gc_NULL(av);
   return gerepilecopy(av, raw_to_FFM(M, ff));
 }
 
@@ -2130,7 +2306,7 @@ FFM_FFM_wrap(GEN M, GEN N, GEN ff,
   case t_FF_F2xq: M = F2xq(M, N, T); break;
   default: M = Flxq(M, N, T, pp); break;
   }
-  if (!M) { avma = av; return NULL; }
+  if (!M) return gc_NULL(av);
   return gerepilecopy(av, raw_to_FFM(M, ff));
 }
 
@@ -2154,7 +2330,7 @@ FFM_FFC_wrap(GEN M, GEN C, GEN ff,
   case t_FF_F2xq: C = F2xq(M, C, T); break;
   default: C = Flxq(M, C, T, pp); break;
   }
-  if (!C) { avma = av; return NULL; }
+  if (!C) return gc_NULL(av);
   return gerepilecopy(av, raw_to_FFC(C, ff));
 }
 
@@ -2184,7 +2360,7 @@ FFM_deplin(GEN M, GEN ff)
   case t_FF_F2xq: C = F2xqM_deplin(M, T); break;
   default: C = FlxqM_deplin(M, T, pp); break;
   }
-  if (!C) { avma = av; return NULL; }
+  if (!C) return gc_NULL(av);
   return gerepilecopy(av, raw_to_FFC(C, ff));
 }
 
@@ -2217,7 +2393,7 @@ FFM_rank(GEN M, GEN ff)
   case t_FF_F2xq: r = F2xqM_rank(M,T); break;
   default: r = FlxqM_rank(M,T,pp); break;
   }
-  avma = av; return r;
+  return gc_long(av,r);
 }
 GEN
 FFM_det(GEN M, GEN ff)
