@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -26,10 +27,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
  * a projective point (x : y : z) on E by [z*x , z^2*y , z].  This is
  * probably not the fastest representation available for the given
  * problem, but they're easy to implement and up to 60% faster than
- * the school-book method used in Fle_mulu(). */
+ * the school-book method. */
 
 /* Cost: 1M + 8S + 1*a + 10add + 1*8 + 2*2 + 1*3.
- * Source: http://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian.html#doubling-dbl-2007-bl */
+ * http://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian.html#doubling-dbl-2007-bl */
 INLINE void
 Flj_dbl_indir_pre(GEN P, GEN Q, ulong a4, ulong p, ulong pi)
 {
@@ -70,7 +71,7 @@ Flj_dbl_pre(GEN P, ulong a4, ulong p, ulong pi)
 }
 
 /* Cost: 11M + 5S + 9add + 4*2.
- * Source: http://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian.html#addition-add-2007-bl */
+ * http://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian.html#addition-add-2007-bl */
 INLINE void
 Flj_add_indir_pre(GEN P, GEN Q, GEN R, ulong a4, ulong p, ulong pi)
 {
@@ -92,8 +93,8 @@ Flj_add_indir_pre(GEN P, GEN Q, GEN R, ulong a4, ulong p, ulong pi)
   r = Fl_double(Fl_sub(S2, S1, p), p);
 
   if (H == 0) {
-    if (r == 0) Flj_dbl_indir_pre(P, R, a4, p, pi); /*Points are equal, double*/
-    else { R[1] = R[2] = 1; R[3] = 0; } /* Points are opposite, return zero */
+    if (r == 0) Flj_dbl_indir_pre(P, R, a4, p, pi); /* P = Q */
+    else { R[1] = R[2] = 1; R[3] = 0; } /* P = -Q */
     return;
   }
   I = Fl_sqr_pre(Fl_double(H, p), p, pi);
@@ -124,22 +125,19 @@ Flj_neg(GEN Q, ulong p)
 { return mkvecsmall3(Q[1], Fl_neg(Q[2], p), Q[3]); }
 
 typedef struct {
-  ulong n; /* The number being represented */
   ulong pbits, nbits;  /* Positive bits and negative bits */
-  ulong lnzb; /* Leading non-zero bit */
+  ulong lnzb; /* Leading nonzero bit */
 } naf_t;
 
 /* Return the signed binary representation (i.e. the Non-Adjacent Form
- * in base 2) of 0 <= a < 2^63 */
+ * in base 2) of a; a = x.pbits - x.nbits (+ 2^BILif < 0; this
+ * exceptional case can happen if a > 2^(BIL-1)) */
 static void
 naf_repr(naf_t *x, ulong a)
 {
+  ulong pbits = 0, nbits = 0, c0 = 0, c1, a0;
   long t, i;
-  ulong pbits, nbits;
-  ulong c0 = 0, c1, a0;
 
-  x->n = a;
-  pbits = nbits = 0;
   for (i = 0; a; a >>= 1, ++i) {
     a0 = a & 1;
     c1 = (c0 + a0 + ((a & 2) >> 1)) >> 1;
@@ -152,16 +150,11 @@ naf_repr(naf_t *x, ulong a)
   }
   c1 = c0 >> 1;
   t = c0 - (c1 << 1);
-  /* Note that we don't need to check whether t < 0, since a >= 0 implies
-   * that this most significant signed bit must be non-negative. */
+  /* since a >= 0, we have t >= 0; if i = BIL, pbits (virtually) overflows;
+   * that leading overflowed bit is implied and not recorded in pbits */
   if (t > 0 && i != BITS_IN_LONG) pbits |= (1UL << i);
-
   x->pbits = pbits;
   x->nbits = nbits;
-  /* Note that expu returns the least nonzero bit in the argument,
-   * like the bit-scan-rev instruction on Intel architectures. */
-  /* Using pbits here is justified by the fact that a >= 0, so the
-   * most significant bit must be positive. */
   x->lnzb = t? i-2: i-3;
 }
 
@@ -169,8 +162,8 @@ naf_repr(naf_t *x, ulong a)
 static GEN
 Flj_mulu_pre_naf(GEN P, ulong n, ulong a4, ulong p, ulong pi, const naf_t *x)
 {
-  GEN R, Pinv;
-  ulong pbits, nbits, lnzb;
+  GEN R, Pi;
+  ulong pbits, nbits;
   ulong m;
 
   if (n == 0) return mkvecsmall3(1, 1, 0);
@@ -181,28 +174,53 @@ Flj_mulu_pre_naf(GEN P, ulong n, ulong a4, ulong p, ulong pi, const naf_t *x)
 
   pbits = x->pbits;
   nbits = x->nbits;
-  lnzb = x->lnzb;
-
-  Pinv = Flj_neg(P, p);
-  m = (1UL << lnzb);
+  Pi = nbits? Flj_neg(P, p): NULL;
+  m = (1UL << x->lnzb);
   for ( ; m; m >>= 1) {
     Flj_dbl_pre_inplace(R, a4, p, pi);
     if (m & pbits)
       Flj_add_pre_inplace(R, P, a4, p, pi);
     else if (m & nbits)
-      Flj_add_pre_inplace(R, Pinv, a4, p, pi);
+      Flj_add_pre_inplace(R, Pi, a4, p, pi);
   }
-  avma = (pari_sp)R; return R;
+  return gc_const((pari_sp)R, R);
 }
 
 GEN
 Flj_mulu_pre(GEN P, ulong n, ulong a4, ulong p, ulong pi)
 {
-  naf_t x;
-  naf_repr(&x, n);
+  naf_t x; naf_repr(&x, n);
   return Flj_mulu_pre_naf(P, n, a4, p, pi, &x);
 }
 
+struct _Flj { ulong a4, p, pi; };
+
+static GEN
+_Flj_add(void *E, GEN P, GEN Q)
+{
+  struct _Flj *ell=(struct _Flj *) E;
+  return Flj_add_pre(P, Q, ell->a4, ell->p, ell->pi);
+}
+
+static GEN
+_Flj_mul(void *E, GEN P, GEN n)
+{
+  struct _Flj *ell = (struct _Flj *) E;
+  long s = signe(n);
+  GEN Q;
+  if (s==0) return mkvecsmall3(1, 1, 0);
+  Q = Flj_mulu_pre(P, itou(n), ell->a4, ell->p, ell->pi);
+  return s>0 ? Q : Flj_neg(Q, ell->p);
+}
+
+GEN
+FljV_factorback_pre(GEN P, GEN L, ulong a4, ulong p, ulong pi)
+{
+  struct _Flj E;
+  E.a4 = a4; E.p = p; E.pi = pi;
+  return gen_factorback(P, L, (void*)&E, &_Flj_add, &_Flj_mul);
+}
+
 ulong
 Flj_order_ufact(GEN P, ulong n, GEN F, ulong a4, ulong p, ulong pi)
 {
@@ -225,11 +243,11 @@ Flj_order_ufact(GEN P, ulong n, GEN F, ulong a4, ulong p, ulong pi)
     b = Flj_mulu_pre(P, n / q, a4, p, pi);
 
     naf_repr(&x, pp);
-    for (j = 0; j < ei && b[3] != 0; ++j)
+    for (j = 0; j < ei && b[3]; ++j)
       b = Flj_mulu_pre_naf(b, pp, a4, p, pi, &x);
-    if (b[3] != 0) return 0;
+    if (b[3]) return 0;
     for (k = 0; k < j; ++k) res *= pp;
-    avma = av;
+    set_avma(av);
   }
   return res;
 }
@@ -277,41 +295,60 @@ random_Flj_pre(ulong a4, ulong a6, ulong p, ulong pi)
   return mkvecsmall3(x, y, 1);
 }
 
+GEN
+Flj_changepointinv_pre(GEN P, GEN ch, ulong p, ulong pi)
+{
+  ulong c, u, r, s, t, u2, u3;
+  ulong x  = uel(P,1), y = uel(P,2), z = uel(P,3);
+  GEN w;
+  if (z == 0) return Flv_copy(P);
+  u = ch[1]; r = ch[2];
+  s = ch[3]; t = ch[4];
+  u2 = Fl_sqr_pre(u, p, pi); u3 = Fl_mul_pre(u, u2, p, pi);
+  c = Fl_mul_pre(u2, x, p, pi);
+  w = cgetg(4, t_VECSMALL);
+  uel(w,1) = Fl_add(c, Fl_mul_pre(r, Fl_sqr_pre(z, p, pi), p, pi), p);
+  uel(w,2) = Fl_add(Fl_mul_pre(u3 ,y, p, pi),
+                    Fl_mul_pre(z, Fl_add(Fl_mul_pre(s,c,p,pi),
+                                         Fl_mul_pre(z,t,p,pi), p), p, pi), p);
+  uel(w,3) = z;
+  return w;
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                              Fle                                  **/
 /**                                                                   **/
 /***********************************************************************/
 GEN
-Fle_changepoint(GEN x, GEN ch, ulong p)
+Fle_changepoint(GEN P, GEN ch, ulong p)
 {
-  ulong p1,u,r,s,t,v,v2,v3;
+  ulong c, u, r, s, t, v, v2, v3;
   GEN z;
-  if (ell_is_inf(x)) return x;
+  if (ell_is_inf(P)) return ellinf();
   u = ch[1]; r = ch[2];
   s = ch[3]; t = ch[4];
   v = Fl_inv(u, p); v2 = Fl_sqr(v,p); v3 = Fl_mul(v,v2,p);
-  p1 = Fl_sub(x[1],r,p);
+  c = Fl_sub(uel(P,1),r,p);
   z = cgetg(3,t_VECSMALL);
-  z[1] = Fl_mul(v2, p1, p);
-  z[2] = Fl_mul(v3, Fl_sub(x[2], Fl_add(Fl_mul(s,p1, p),t, p),p),p);
+  z[1] = Fl_mul(v2, c, p);
+  z[2] = Fl_mul(v3, Fl_sub(uel(P,2), Fl_add(Fl_mul(s,c, p),t, p),p),p);
   return z;
 }
 
 GEN
-Fle_changepointinv(GEN x, GEN ch, ulong p)
+Fle_changepointinv(GEN P, GEN ch, ulong p)
 {
-  ulong u, r, s, t, X, Y, u2, u3, u2X;
+  ulong c, u, r, s, t, u2, u3;
   GEN z;
-  if (ell_is_inf(x)) return x;
-  X = x[1]; Y = x[2];
+  if (ell_is_inf(P)) return ellinf();
   u = ch[1]; r = ch[2];
   s = ch[3]; t = ch[4];
   u2 = Fl_sqr(u, p); u3 = Fl_mul(u,u2,p);
-  u2X = Fl_mul(u2,X, p);
+  c = Fl_mul(u2,uel(P,1), p);
   z = cgetg(3, t_VECSMALL);
-  z[1] = Fl_add(u2X,r,p);
-  z[2] = Fl_add(Fl_mul(u3,Y,p), Fl_add(Fl_mul(s,u2X,p), t, p), p);
+  z[1] = Fl_add(c,r,p);
+  z[2] = Fl_add(Fl_mul(u3,uel(P,2),p), Fl_add(Fl_mul(s,c,p), t, p), p);
   return z;
 }
 static GEN
@@ -421,7 +458,7 @@ Fle_mul(GEN P, GEN n, ulong a4, ulong p)
   return _Fle_mul(&E, P, n);
 }
 
-/* Finds a random non-singular point on E */
+/* Finds a random nonsingular point on E */
 GEN
 random_Fle_pre(ulong a4, ulong a6, ulong p, ulong pi)
 {
@@ -518,14 +555,6 @@ Fl_elldisc(ulong a4, ulong a6, ulong p)
   return Fl_elldisc_pre(a4, a6, p, get_Fl_red(p));
 }
 
-static ulong
-nonsquare_Fl(ulong p)
-{
-  ulong a;
-  do a = random_Fl(p); while (krouu(a, p) >= 0);
-  return a;
-}
-
 void
 Fl_elltwist_disc(ulong a4, ulong a6, ulong D, ulong p, ulong *pa4, ulong *pa6)
 {
@@ -633,7 +662,7 @@ static void
 FleV_mulu_pre_naf_inplace(GEN P, ulong n, GEN a4, ulong p, ulong pi, const naf_t *x)
 {
   pari_sp av = avma;
-  ulong pbits, nbits, lnzb, m;
+  ulong pbits, nbits, m;
   GEN R;
   if (n == 1) return;
 
@@ -643,9 +672,7 @@ FleV_mulu_pre_naf_inplace(GEN P, ulong n, GEN a4, ulong p, ulong pi, const naf_t
 
   pbits = x->pbits;
   nbits = x->nbits;
-  lnzb = x->lnzb;
-
-  m = (1UL << lnzb);
+  m = (1UL << x->lnzb);
   for ( ; m; m >>= 1) {
     FleV_dbl_pre_inplace(R, a4, p, pi);
     if (m & pbits)
@@ -653,13 +680,12 @@ FleV_mulu_pre_naf_inplace(GEN P, ulong n, GEN a4, ulong p, ulong pi, const naf_t
     else if (m & nbits)
       FleV_sub_pre_inplace(R, P, a4, p, pi);
   }
-  avma = av;
+  set_avma(av);
 }
 
 void
 FleV_mulu_pre_inplace(GEN P, ulong n, GEN a4, ulong p, ulong pi)
 {
-  naf_t x;
-  naf_repr(&x, n);
+  naf_t x; naf_repr(&x, n);
   FleV_mulu_pre_naf_inplace(P, n, a4, p, pi, &x);
 }
