@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -133,26 +134,19 @@ FpJ_to_FpE(GEN P, GEN p)
   }
 }
 
-struct _FpE
-{
-  GEN a4,a6;
-  GEN p;
-};
-
+struct _FpE { GEN p,a4,a6; };
 static GEN
 _FpJ_dbl(void *E, GEN P)
 {
   struct _FpE *ell = (struct _FpE *) E;
   return FpJ_dbl(P, ell->a4, ell->p);
 }
-
 static GEN
 _FpJ_add(void *E, GEN P, GEN Q)
 {
   struct _FpE *ell=(struct _FpE *) E;
   return FpJ_add(P, Q, ell->a4, ell->p);
 }
-
 static GEN
 _FpJ_mul(void *E, GEN P, GEN n)
 {
@@ -162,7 +156,7 @@ _FpJ_mul(void *E, GEN P, GEN n)
   if (!s || ell_is_inf(P)) return ellinf();
   if (s<0) P = FpJ_neg(P, e->p);
   if (is_pm1(n)) return s>0? gcopy(P): P;
-  return gerepilecopy(av, gen_pow(P, n, e, &_FpJ_dbl, &_FpJ_add));
+  return gerepilecopy(av, gen_pow_i(P, n, e, &_FpJ_dbl, &_FpJ_add));
 }
 
 GEN
@@ -200,34 +194,43 @@ FpE_to_mod(GEN x, GEN p)
 }
 
 GEN
-FpE_changepoint(GEN x, GEN ch, GEN p)
+FpE_changepoint(GEN P, GEN ch, GEN p)
 {
   pari_sp av = avma;
-  GEN p1,z,u,r,s,t,v,v2,v3;
-  if (ell_is_inf(x)) return x;
-  u = gel(ch,1); r = gel(ch,2);
-  s = gel(ch,3); t = gel(ch,4);
+  GEN c, z, u, r, s, t, v, v2, v3;
+  if (ell_is_inf(P)) return P;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = p[2];
+    z = Fle_changepoint(ZV_to_Flv(P, pp), ZV_to_Flv(ch, pp), pp);
+    return gerepileupto(av, Flv_to_ZV(z));
+  }
+  u = gel(ch,1); r = gel(ch,2); s = gel(ch,3); t = gel(ch,4);
   v = Fp_inv(u, p); v2 = Fp_sqr(v,p); v3 = Fp_mul(v,v2,p);
-  p1 = Fp_sub(gel(x,1),r,p);
+  c = Fp_sub(gel(P,1),r,p);
   z = cgetg(3,t_VEC);
-  gel(z,1) = Fp_mul(v2, p1, p);
-  gel(z,2) = Fp_mul(v3, Fp_sub(gel(x,2), Fp_add(Fp_mul(s,p1, p),t, p),p),p);
+  gel(z,1) = Fp_mul(v2, c, p);
+  gel(z,2) = Fp_mul(v3, Fp_sub(gel(P,2), Fp_add(Fp_mul(s,c, p),t, p),p),p);
   return gerepileupto(av, z);
 }
 
 GEN
-FpE_changepointinv(GEN x, GEN ch, GEN p)
+FpE_changepointinv(GEN P, GEN ch, GEN p)
 {
-  GEN u, r, s, t, X, Y, u2, u3, u2X, z;
-  if (ell_is_inf(x)) return x;
-  X = gel(x,1); Y = gel(x,2);
-  u = gel(ch,1); r = gel(ch,2);
-  s = gel(ch,3); t = gel(ch,4);
+  GEN u, r, s, t, u2, u3, c, z;
+  if (ell_is_inf(P)) return P;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = p[2];
+    z = Fle_changepointinv(ZV_to_Flv(P, pp), ZV_to_Flv(ch, pp), pp);
+    return Flv_to_ZV(z);
+  }
+  u = gel(ch,1); r = gel(ch,2); s = gel(ch,3); t = gel(ch,4);
   u2 = Fp_sqr(u, p); u3 = Fp_mul(u,u2,p);
-  u2X = Fp_mul(u2,X, p);
+  c = Fp_mul(u2, gel(P,1), p);
   z = cgetg(3, t_VEC);
-  gel(z,1) = Fp_add(u2X,r,p);
-  gel(z,2) = Fp_add(Fp_mul(u3,Y,p), Fp_add(Fp_mul(s,u2X,p), t, p), p);
+  gel(z,1) = Fp_add(c,r,p);
+  gel(z,2) = Fp_add(Fp_mul(u3,gel(P,2),p), Fp_add(Fp_mul(s,c,p), t, p), p);
   return z;
 }
 
@@ -238,7 +241,7 @@ nonsquare_Fp(GEN p)
   GEN a;
   do
   {
-    avma = av;
+    set_avma(av);
     a = randomi(p);
   } while (kronecker(a, p) >= 0);
   return a;
@@ -351,7 +354,7 @@ _FpE_mul(void *E, GEN P, GEN n)
   if (s<0) P = FpE_neg(P, e->p);
   if (is_pm1(n)) return s>0? gcopy(P): P;
   if (equalis(n,2)) return _FpE_dbl(E, P);
-  Q = gen_pow(FpE_to_FpJ(P), n, e, &_FpJ_dbl, &_FpJ_add);
+  Q = gen_pow_i(FpE_to_FpJ(P), n, e, &_FpJ_dbl, &_FpJ_add);
   return gerepileupto(av, FpJ_to_FpE(Q, e->p));
 }
 
@@ -363,7 +366,7 @@ FpE_mul(GEN P, GEN n, GEN a4, GEN p)
   return _FpE_mul(&E, P, n);
 }
 
-/* Finds a random non-singular point on E */
+/* Finds a random nonsingular point on E */
 
 GEN
 random_FpE(GEN a4, GEN a6, GEN p)
@@ -372,7 +375,7 @@ random_FpE(GEN a4, GEN a6, GEN p)
   GEN x, x2, y, rhs;
   do
   {
-    avma= ltop;
+    set_avma(ltop);
     x   = randomi(p); /*  x^3+a4*x+a6 = x*(x^2+a4)+a6  */
     x2  = Fp_sqr(x, p);
     rhs = Fp_add(Fp_mul(x, Fp_add(x2, a4, p), p), a6, p);
@@ -537,31 +540,21 @@ FpE_chord_update(GEN R, GEN P, GEN Q, GEN a4, GEN p, GEN *pt_R)
   }
 }
 
-/* Returns the Miller function f_{m, Q} evaluated at the point P using
-   the standard Miller algorithm.
- */
-
-struct _FpE_miller
-{
-  GEN p, a4, P;
-};
-
+struct _FpE_miller { GEN p, a4, P; };
 static GEN
 FpE_Miller_dbl(void* E, GEN d)
 {
   struct _FpE_miller *m = (struct _FpE_miller *)E;
   GEN p = m->p, a4 = m->a4, P = m->P;
   GEN v, line;
-  GEN num = Fp_sqr(gel(d,1), p);
-  GEN denom = Fp_sqr(gel(d,2), p);
+  GEN N = Fp_sqr(gel(d,1), p);
+  GEN D = Fp_sqr(gel(d,2), p);
   GEN point = gel(d,3);
   line = FpE_tangent_update(point, P, a4, p, &point);
-  num  = Fp_mul(num, line, p);
+  N  = Fp_mul(N, line, p);
   v = FpE_vert(point, P, a4, p);
-  denom = Fp_mul(denom, v, p);
-  return mkvec3(num, denom, point);
+  D = Fp_mul(D, v, p); return mkvec3(N, D, point);
 }
-
 static GEN
 FpE_Miller_add(void* E, GEN va, GEN vb)
 {
@@ -570,48 +563,47 @@ FpE_Miller_add(void* E, GEN va, GEN vb)
   GEN v, line, point;
   GEN na = gel(va,1), da = gel(va,2), pa = gel(va,3);
   GEN nb = gel(vb,1), db = gel(vb,2), pb = gel(vb,3);
-  GEN num   = Fp_mul(na, nb, p);
-  GEN denom = Fp_mul(da, db, p);
+  GEN N = Fp_mul(na, nb, p);
+  GEN D = Fp_mul(da, db, p);
   line = FpE_chord_update(pa, pb, P, a4, p, &point);
-  num  = Fp_mul(num, line, p);
+  N = Fp_mul(N, line, p);
   v = FpE_vert(point, P, a4, p);
-  denom = Fp_mul(denom, v, p);
-  return mkvec3(num, denom, point);
+  D = Fp_mul(D, v, p); return mkvec3(N, D, point);
 }
 
+/* Returns the Miller function f_{m, Q} evaluated at the point P using
+ * the standard Miller algorithm. */
 static GEN
 FpE_Miller(GEN Q, GEN P, GEN m, GEN a4, GEN p)
 {
-  pari_sp ltop = avma;
+  pari_sp av = avma;
   struct _FpE_miller d;
-  GEN v, num, denom;
+  GEN v, N, D;
 
   d.a4 = a4; d.p = p; d.P = P;
-  v = gen_pow(mkvec3(gen_1,gen_1,Q), m, (void*)&d, FpE_Miller_dbl, FpE_Miller_add);
-  num = gel(v,1); denom = gel(v,2);
-  return gerepileupto(ltop, Fp_div(num, denom, p));
+  v = gen_pow_i(mkvec3(gen_1,gen_1,Q), m, (void*)&d,
+                FpE_Miller_dbl, FpE_Miller_add);
+  N = gel(v,1); D = gel(v,2);
+  return gerepileuptoint(av, Fp_div(N, D, p));
 }
 
 GEN
 FpE_weilpairing(GEN P, GEN Q, GEN m, GEN a4, GEN p)
 {
   pari_sp ltop = avma;
-  GEN num, denom, result;
-  if (ell_is_inf(P) || ell_is_inf(Q) || ZV_equal(P,Q))
-    return gen_1;
-  num    = FpE_Miller(P, Q, m, a4, p);
-  denom  = FpE_Miller(Q, P, m, a4, p);
-  result = Fp_div(num, denom, p);
-  if (mpodd(m))
-    result  = Fp_neg(result, p);
-  return gerepileupto(ltop, result);
+  GEN N, D, w;
+  if (ell_is_inf(P) || ell_is_inf(Q) || ZV_equal(P,Q)) return gen_1;
+  N = FpE_Miller(P, Q, m, a4, p);
+  D = FpE_Miller(Q, P, m, a4, p);
+  w = Fp_div(N, D, p);
+  if (mpodd(m)) w  = Fp_neg(w, p);
+  return gerepileuptoint(ltop, w);
 }
 
 GEN
 FpE_tatepairing(GEN P, GEN Q, GEN m, GEN a4, GEN p)
 {
-  if (ell_is_inf(P) || ell_is_inf(Q))
-    return gen_1;
+  if (ell_is_inf(P) || ell_is_inf(Q)) return gen_1;
   return FpE_Miller(P, Q, m, a4, p);
 }
 
@@ -700,7 +692,7 @@ path_extends_to_floor(GEN j_prev, GEN j, GEN T, GEN p, GEN Phi2, ulong max_len)
     j_next = FqX_quad_root(Phi2_j, T, p);
     if (!j_next)
     { /* j is on the floor */
-      avma = ltop;
+      set_avma(ltop);
       return 1;
     }
 
@@ -713,7 +705,7 @@ path_extends_to_floor(GEN j_prev, GEN j, GEN T, GEN p, GEN Phi2, ulong max_len)
    * point to the last element in the path. */
   Phi2_j = FqX_div_by_X_x(FqXY_evalx(Phi2, j, T, p), j_prev, T, p, NULL);
   mult = FqX_nbroots(Phi2_j, T, p);
-  avma = ltop;
+  set_avma(ltop);
   return mult == 0;
 }
 
@@ -757,7 +749,7 @@ Fp_elljissupersingular(GEN j, GEN p)
   {
     GEN S = init_Fq(p, 2, fetch_var());
     int res = jissupersingular(j, S, p);
-    (void)delete_var(); avma = ltop; return res;
+    (void)delete_var(); return gc_bool(ltop, res);
   }
 }
 
@@ -832,7 +824,7 @@ get_table_size(GEN pordmin, GEN B)
   GEN t = ceilr( sqrtr( divri(itor(pordmin, DEFAULTPREC), B) ) );
   if (is_bigint(t))
     pari_err_OVERFLOW("ellap [large prime: install the 'seadata' package]");
-  avma = av;
+  set_avma(av);
   return itos(t) >> 1;
 }
 
@@ -984,7 +976,7 @@ Fp_ellcard_Shanks(GEN c4, GEN c6, GEN p)
         tx[i] = mod2BIL(gel(P,1));
         ty[i] = mod2BIL(gel(P,2));
       }
-      avma = av2;
+      set_avma(av2);
     }
     P = FpE_add(gel(pts,j-1),mfh,a4,p); /* = (s-1).F */
     if (ell_is_inf(P)) { h = mului(s-1,B); goto FOUND; }
@@ -1005,7 +997,7 @@ Fp_ellcard_Shanks(GEN c4, GEN c6, GEN p)
     for (i=1; i<=s; i++) { ty[i] = ti[i]; ti[i] = p1[i]; }
     /* ty is sorted. ti = permutation sorting tx */
     if (DEBUGLEVEL >= 6) timer_printf(&T, "[Fp_ellcard_Shanks] sorting");
-    avma = av2;
+    set_avma(av2);
 
     gaffect(fg, gel(pts,1));
     for (j=2; j<=nb; j++) /* pts[j] = j.fg = (s*j).F */
@@ -1015,7 +1007,7 @@ Fp_ellcard_Shanks(GEN c4, GEN c6, GEN p)
       gaffect(P, gel(pts,j));
     }
     /* replace fg by nb.fg since we do nb points at a time */
-    avma = av2;
+    set_avma(av2);
     fg = gcopy(gel(pts,nb)); /* copy: we modify (temporarily) pts[nb] below */
     av2 = avma;
 
@@ -1025,7 +1017,7 @@ Fp_ellcard_Shanks(GEN c4, GEN c6, GEN p)
       long m, l = 1, r = s+1;
       long k, k2, j2;
 
-      avma = av2;
+      set_avma(av2);
       k = mod2BIL(gel(ftest,1));
       while (l < r)
       {
@@ -1144,7 +1136,7 @@ Fl_ellcard_Shanks(ulong c4, ulong c6, ulong p)
 
   if (!c6) {
     GEN ap = ap_j1728(utoi(c4), utoipos(p));
-    avma = av; return p+1 - itos(ap);
+    return gc_long(av, p+1 - itos(ap));
   }
 
   pordmin = (ulong)(1 + 4*sqrt((double)p));
@@ -1220,9 +1212,9 @@ FOUND:
       }
       B = itou(C);
     }
-    A = (p2p - A) % B; avma = av;
+    A = (p2p - A) % B; set_avma(av);
   }
-  avma = av; return KRO==1? h: p2p-h;
+  return gc_long(av, KRO==1? h: p2p-h);
 }
 
 /** ellap from CM (original code contributed by Mark Watkins) **/
@@ -1325,8 +1317,14 @@ Fp_ellj_nodiv(GEN a4, GEN a6, GEN p)
 GEN
 Fp_ellj(GEN a4, GEN a6, GEN p)
 {
-  pari_sp av=avma;
-  GEN z = Fp_ellj_nodiv(a4, a6, p);
+  pari_sp av = avma;
+  GEN z;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = p[2];
+    return utoi(Fl_ellj(umodiu(a4,pp), umodiu(a6,pp), pp));
+  }
+  z = Fp_ellj_nodiv(a4, a6, p);
   return gerepileuptoint(av,Fp_div(gel(z,1),gel(z,2),p));
 }
 
@@ -1341,7 +1339,7 @@ Fp_ellcard_CM(GEN a4, GEN a6, GEN p)
   {
     GEN j = Fp_ellj_nodiv(a4, a6, p);
     long CM = Fp_ellj_get_CM(gel(j,1), gel(j,2), p);
-    if (!CM) { avma = av; return NULL; }
+    if (!CM) return gc_NULL(av);
     a = ec_ap_cm(CM,a4,a6,p);
   }
   return gerepileuptoint(av, subii(addiu(p,1),a));
@@ -1372,7 +1370,7 @@ Fl_elltrace(ulong a4, ulong a6, ulong p)
   lp = expu(p);
   if (lp <= minss(56, BITS_IN_LONG-2)) return p+1-Fl_ellcard_Shanks(a4, a6, p);
   av = avma; a = subui(p+1, Fp_ellcard(utoi(a4), utoi(a6), utoipos(p)));
-  avma = av; return itos(a);
+  return gc_long(av, itos(a));
 }
 long
 Fl_elltrace_CM(long CM, ulong a4, ulong a6, ulong p)
@@ -1382,7 +1380,7 @@ Fl_elltrace_CM(long CM, ulong a4, ulong a6, ulong p)
   if (!CM) return Fl_elltrace(a4,a6,p);
   if (p < (1<<11)) return Fl_elltrace_naive(a4, a6, p);
   av = avma; a = ec_ap_cm(CM, utoi(a4), utoi(a6), utoipos(p));
-  avma = av; return itos(a);
+  return gc_long(av, itos(a));
 }
 
 static GEN
@@ -1489,7 +1487,7 @@ nonsquare_FpXQ(GEN T, GEN p)
   }
   do
   {
-    avma = av;
+    set_avma(av);
     a = random_FpX(n, v, p);
   } while (FpXQ_issquare(a, T, p));
   return a;
@@ -1579,26 +1577,19 @@ FpXQE_sub(GEN P, GEN Q, GEN a4, GEN T, GEN p)
   return gerepileupto(av, FpXQE_add_slope(P, FpXQE_neg_i(Q, p), a4, T, p, &slope));
 }
 
-struct _FpXQE
-{
-  GEN a4,a6;
-  GEN T,p;
-};
-
+struct _FpXQE { GEN a4,a6,T,p; };
 static GEN
 _FpXQE_dbl(void *E, GEN P)
 {
   struct _FpXQE *ell = (struct _FpXQE *) E;
   return FpXQE_dbl(P, ell->a4, ell->T, ell->p);
 }
-
 static GEN
 _FpXQE_add(void *E, GEN P, GEN Q)
 {
   struct _FpXQE *ell=(struct _FpXQE *) E;
   return FpXQE_add(P, Q, ell->a4, ell->T, ell->p);
 }
-
 static GEN
 _FpXQE_mul(void *E, GEN P, GEN n)
 {
@@ -1608,7 +1599,7 @@ _FpXQE_mul(void *E, GEN P, GEN n)
   if (!s || ell_is_inf(P)) return ellinf();
   if (s<0) P = FpXQE_neg(P, e->T, e->p);
   if (is_pm1(n)) return s>0? gcopy(P): P;
-  return gerepileupto(av, gen_pow(P, n, e, &_FpXQE_dbl, &_FpXQE_add));
+  return gerepilecopy(av, gen_pow_i(P, n, e, &_FpXQE_dbl, &_FpXQE_add));
 }
 
 GEN
@@ -1619,7 +1610,7 @@ FpXQE_mul(GEN P, GEN n, GEN a4, GEN T, GEN p)
   return _FpXQE_mul(&E, P, n);
 }
 
-/* Finds a random non-singular point on E */
+/* Finds a random nonsingular point on E */
 
 GEN
 random_FpXQE(GEN a4, GEN a6, GEN T, GEN p)
@@ -1629,7 +1620,7 @@ random_FpXQE(GEN a4, GEN a6, GEN T, GEN p)
   long v = get_FpX_var(T), d = get_FpX_degree(T);
   do
   {
-    avma= ltop;
+    set_avma(ltop);
     x   = random_FpX(d,v,p); /*  x^3+a4*x+a6 = x*(x^2+a4)+a6  */
     x2  = FpXQ_sqr(x, T, p);
     rhs = FpX_add(FpXQ_mul(x, FpX_add(x2, a4, p), T, p), a6, p);
@@ -1676,7 +1667,6 @@ FpXQE_log(GEN a, GEN b, GEN o, GEN a4, GEN T, GEN p)
   return gerepileuptoint(av, gen_PH_log(a, b, o, (void*)&e, &FpXQE_group));
 }
 
-
 /***********************************************************************/
 /**                                                                   **/
 /**                            Pairings                               **/
@@ -1777,16 +1767,7 @@ FpXQE_chord_update(GEN R, GEN P, GEN Q, GEN a4, GEN T, GEN p, GEN *pt_R)
   }
 }
 
-/* Returns the Miller function f_{m, Q} evaluated at the point P using
-   the standard Miller algorithm.
- */
-
-struct _FpXQE_miller
-{
-  GEN p;
-  GEN T, a4, P;
-};
-
+struct _FpXQE_miller { GEN p, T, a4, P; };
 static GEN
 FpXQE_Miller_dbl(void* E, GEN d)
 {
@@ -1794,14 +1775,13 @@ FpXQE_Miller_dbl(void* E, GEN d)
   GEN p  = m->p;
   GEN T = m->T, a4 = m->a4, P = m->P;
   GEN v, line;
-  GEN num = FpXQ_sqr(gel(d,1), T, p);
-  GEN denom = FpXQ_sqr(gel(d,2), T, p);
+  GEN N = FpXQ_sqr(gel(d,1), T, p);
+  GEN D = FpXQ_sqr(gel(d,2), T, p);
   GEN point = gel(d,3);
   line = FpXQE_tangent_update(point, P, a4, T, p, &point);
-  num  = FpXQ_mul(num, line, T, p);
+  N = FpXQ_mul(N, line, T, p);
   v = FpXQE_vert(point, P, a4, T, p);
-  denom = FpXQ_mul(denom, v, T, p);
-  return mkvec3(num, denom, point);
+  D = FpXQ_mul(D, v, T, p); return mkvec3(N, D, point);
 }
 
 static GEN
@@ -1813,49 +1793,49 @@ FpXQE_Miller_add(void* E, GEN va, GEN vb)
   GEN v, line, point;
   GEN na = gel(va,1), da = gel(va,2), pa = gel(va,3);
   GEN nb = gel(vb,1), db = gel(vb,2), pb = gel(vb,3);
-  GEN num   = FpXQ_mul(na, nb, T, p);
-  GEN denom = FpXQ_mul(da, db, T, p);
+  GEN N = FpXQ_mul(na, nb, T, p);
+  GEN D = FpXQ_mul(da, db, T, p);
   line = FpXQE_chord_update(pa, pb, P, a4, T, p, &point);
-  num  = FpXQ_mul(num, line, T, p);
+  N = FpXQ_mul(N, line, T, p);
   v = FpXQE_vert(point, P, a4, T, p);
-  denom = FpXQ_mul(denom, v, T, p);
-  return mkvec3(num, denom, point);
+  D = FpXQ_mul(D, v, T, p); return mkvec3(N, D, point);
 }
 
+/* Returns the Miller function f_{m, Q} evaluated at the point P using
+ * the standard Miller algorithm. */
 static GEN
 FpXQE_Miller(GEN Q, GEN P, GEN m, GEN a4, GEN T, GEN p)
 {
-  pari_sp ltop = avma;
+  pari_sp av = avma;
   struct _FpXQE_miller d;
-  GEN v, num, denom, g1;
+  GEN v, N, D, g1;
 
   d.a4 = a4; d.T = T; d.p = p; d.P = P;
   g1 = pol_1(get_FpX_var(T));
-  v = gen_pow(mkvec3(g1,g1,Q), m, (void*)&d, FpXQE_Miller_dbl, FpXQE_Miller_add);
-  num = gel(v,1); denom = gel(v,2);
-  return gerepileupto(ltop, FpXQ_div(num, denom, T, p));
+  v = gen_pow_i(mkvec3(g1,g1,Q), m, (void*)&d,
+                FpXQE_Miller_dbl, FpXQE_Miller_add);
+  N = gel(v,1); D = gel(v,2);
+  return gerepileupto(av, FpXQ_div(N, D, T, p));
 }
 
 GEN
 FpXQE_weilpairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, GEN p)
 {
-  pari_sp ltop = avma;
-  GEN num, denom, result;
+  pari_sp av = avma;
+  GEN N, D, w;
   if (ell_is_inf(P) || ell_is_inf(Q) || ZXV_equal(P,Q))
     return pol_1(get_FpX_var(T));
-  num    = FpXQE_Miller(P, Q, m, a4, T, p);
-  denom  = FpXQE_Miller(Q, P, m, a4, T, p);
-  result = FpXQ_div(num, denom, T, p);
-  if (mpodd(m))
-    result  = FpX_neg(result, p);
-  return gerepileupto(ltop, result);
+  N = FpXQE_Miller(P, Q, m, a4, T, p);
+  D = FpXQE_Miller(Q, P, m, a4, T, p);
+  w = FpXQ_div(N, D, T, p);
+  if (mpodd(m)) w = FpX_neg(w, p);
+  return gerepileupto(av, w);
 }
 
 GEN
 FpXQE_tatepairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, GEN p)
 {
-  if (ell_is_inf(P) || ell_is_inf(Q))
-    return pol_1(get_FpX_var(T));
+  if (ell_is_inf(P) || ell_is_inf(Q)) return pol_1(get_FpX_var(T));
   return FpXQE_Miller(P, Q, m, a4, T, p);
 }
 
@@ -1891,7 +1871,6 @@ FpXQ_elljissupersingular(GEN j, GEN T, GEN p)
    * the j-invariants are in FF_{p^{2 - e}}. */
   ulong d = get_FpX_degree(T);
   GEN S;
-  int res;
 
   if (degpol(j) <= 0) return Fp_elljissupersingular(constant_coeff(j), p);
   if (abscmpiu(p, 5) <= 0) return 0; /* j != 0*/
@@ -1905,17 +1884,15 @@ FpXQ_elljissupersingular(GEN j, GEN T, GEN p)
     GEN j_pow_p = FpXQ_pow(j, p, T, p);
     GEN j_sum = FpX_add(j, j_pow_p, p), j_prod;
     long var = varn(T);
-    if (degpol(j_sum) > 0) { avma = ltop; return 0; /* j not in Fp^2 */ }
+    if (degpol(j_sum) > 0) return gc_bool(ltop,0); /* j not in Fp^2 */
     j_prod = FpXQ_mul(j, j_pow_p, T, p);
-    if (degpol(j_prod) > 0 ) { avma = ltop; return 0; /* j not in Fp^2 */ }
+    if (degpol(j_prod) > 0 ) return gc_bool(ltop,0); /* j not in Fp^2 */
     j_sum = constant_coeff(j_sum); j_prod = constant_coeff(j_prod);
     S = mkpoln(3, gen_1, Fp_neg(j_sum, p), j_prod);
     setvarn(S, var);
     j = pol_x(var);
   }
-  res = jissupersingular(j, S, p);
-  avma = ltop;
-  return res;
+  return gc_bool(ltop, jissupersingular(j,S,p));
 }
 
 /***********************************************************************/
@@ -2064,5 +2041,3 @@ FpXQ_ellgens(GEN a4, GEN a6, GEN ch, GEN D, GEN m, GEN T, GEN p)
   }
   return gerepilecopy(av, P);
 }
-
-
