@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -184,6 +185,8 @@ Flm_Fl_add(GEN x, ulong y, ulong p)
   }
   return z;
 }
+GEN
+Flm_Fl_sub(GEN x, ulong y, ulong p) { return Flm_Fl_add(x, Fl_neg(y,p), p); }
 
 GEN
 Flm_add(GEN x, GEN y, ulong p)
@@ -236,7 +239,10 @@ void
 Flv_Fl_mul_part_inplace(GEN x, ulong y, ulong p, long l)
 {
   long i;
-  for (i=1;i<=l;i++) x[i] = Fl_mul(x[i], y, p);
+  if (HIGHWORD(y | p))
+    for(i=1; i<=l; i++) x[i] = Fl_mul(x[i], y, p);
+  else
+    for(i=1; i<=l; i++) x[i] = (x[i] * y) % p;
 }
 void
 Flv_Fl_mul_inplace(GEN x, ulong y, ulong p)
@@ -258,26 +264,55 @@ Flm_Fl_mul_inplace(GEN y, ulong x, ulong p)
 }
 
 /* return x * y */
-GEN
-Flm_Fl_mul(GEN y, ulong x, ulong p)
+static GEN
+Flm_Fl_mul_pre_i(GEN y, ulong x, ulong p, ulong pi)
 {
   long i, j, m, l = lg(y);
   GEN z = cgetg(l, t_MAT);
   if (l == 1) return z;
   m = lgcols(y);
-  if (HIGHWORD(x | p))
-    for(j=1; j<l; j++) {
-      GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
-      for(i=1; i<m; i++) c[i] = Fl_mul(ucoeff(y,i,j), x, p);
-    }
-  else
-    for(j=1; j<l; j++) {
-      GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
-      for(i=1; i<m; i++) c[i] = (ucoeff(y,i,j) * x) % p;
-    }
+  for(j=1; j<l; j++) {
+    GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
+    for(i=1; i<m; i++) uel(c,i) = Fl_mul_pre(ucoeff(y,i,j), x, p, pi);
+  }
   return z;
 }
 
+/* return x * y */
+static GEN
+Flm_Fl_mul_OK(GEN y, ulong x, ulong p)
+{
+  long i, j, m, l = lg(y);
+  GEN z = cgetg(l, t_MAT);
+  if (l == 1) return z;
+  m = lgcols(y);
+  for(j=1; j<l; j++) {
+    GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
+    for(i=1; i<m; i++) uel(c,i) = (ucoeff(y,i,j) * x) % p;
+  }
+  return z;
+}
+
+/* return x * y */
+GEN
+Flm_Fl_mul_pre(GEN y, ulong x, ulong p, ulong pi)
+{
+  if (HIGHWORD(p))
+    return Flm_Fl_mul_pre_i(y, x, p, pi);
+  else
+    return Flm_Fl_mul_OK(y, x, p);
+}
+
+/* return x * y */
+GEN
+Flm_Fl_mul(GEN y, ulong x, ulong p)
+{
+  if (HIGHWORD(p))
+    return Flm_Fl_mul_pre_i(y, x, p, get_Fl_red(p));
+  else
+    return Flm_Fl_mul_OK(y, x, p);
+}
+
 GEN
 Flv_neg(GEN x, ulong p)
 { pari_APPLY_ulong(Fl_neg(uel(x,i), p)) }
@@ -433,22 +468,6 @@ Flm_Flc_mul_i(GEN x, GEN y, long lx, long l, ulong p, ulong pi)
   return z;
 }
 
-INLINE GEN
-F2m_F2c_mul_i(GEN x, GEN y, long lx, long l)
-{
-  long j;
-  GEN z = NULL;
-
-  for (j=1; j<lx; j++)
-  {
-    if (!F2v_coeff(y,j)) continue;
-    if (!z) z = vecsmall_copy(gel(x,j));
-    else F2v_add_inplace(z,gel(x,j));
-  }
-  if (!z) z = zero_F2v(l);
-  return z;
-}
-
 GEN
 FpM_mul(GEN x, GEN y, GEN p)
 {
@@ -669,7 +688,7 @@ Flm_mul_sw(GEN A, GEN B, long m, long n, long p, ulong l, ulong li)
     gerepileall(av, 3, &M3, &M7, &V2);  /* destroy V1, M2 */
   add_slices_ip(m2, p1, V2, 0, m2, 0, p1, M7, 0, m2, 0, p1, C, m1, 0, l);
   add_slices_ip(m2, p2, V2, 0, m2, 0, p2, M3, 0, m2, 0, p2, C, m1, p1, l);
-  avma = av; return C;
+  set_avma(av); return C;
 }
 
 /* Strassen-Winograd used for dim >= ZM_sw_bound */
@@ -688,6 +707,15 @@ Flm_mul_i(GEN x, GEN y, long l, long lx, long ly, ulong p, ulong pi)
     return Flm_mul_sw(x, y, l - 1, lx - 1, ly - 1, p, pi);
 }
 
+GEN
+Flm_mul_pre(GEN x, GEN y, ulong p, ulong pi)
+{
+  long lx=lg(x), ly=lg(y);
+  if (ly==1) return cgetg(1,t_MAT);
+  if (lx==1) return zero_Flm(0, ly-1);
+  return Flm_mul_i(x, y, lgcols(x), lx, ly, p, pi);
+}
+
 GEN
 Flm_mul(GEN x, GEN y, ulong p)
 {
@@ -697,23 +725,6 @@ Flm_mul(GEN x, GEN y, ulong p)
   return Flm_mul_i(x, y, lgcols(x), lx, ly, p, get_Fl_red(p));
 }
 
-GEN
-F2m_mul(GEN x, GEN y)
-{
-  long i,j,l,lx=lg(x), ly=lg(y);
-  GEN z;
-  if (ly==1) return cgetg(1,t_MAT);
-  z = cgetg(ly,t_MAT);
-  if (lx==1)
-  {
-    for (i=1; i<ly; i++) gel(z,i) = mkvecsmall(0);
-    return z;
-  }
-  l = coeff(x,1,1);
-  for (j=1; j<ly; j++) gel(z,j) = F2m_F2c_mul_i(x, gel(y,j), lx, l);
-  return z;
-}
-
 struct _Flm
 {
   ulong p;
@@ -732,36 +743,21 @@ _Flm_one(void *E)
 GEN
 Flm_powu(GEN x, ulong n, ulong p)
 {
-  pari_sp av = avma;
   struct _Flm d;
   if (!n) return matid(lg(x)-1);
   d.p = p;
-  return gerepileupto(av, gen_powu(x, n, (void*)&d, &_Flm_sqr, &_Flm_mul));
+  return gen_powu(x, n, (void*)&d, &_Flm_sqr, &_Flm_mul);
 }
 GEN
 Flm_powers(GEN x, ulong n, ulong p)
 {
-  pari_sp av = avma;
   struct _Flm d;
   d.p = p;
   d.n = lg(x)-1;
-  return gerepileupto(av, gen_powers(x, n, 1, (void*)&d,
-                          &_Flm_sqr, &_Flm_mul, &_Flm_one));
+  return gen_powers(x, n, 1, (void*)&d,
+                    &_Flm_sqr, &_Flm_mul, &_Flm_one);
 }
 
-static GEN
-_F2m_mul(void *data, GEN x, GEN y)
-{ (void) data; return F2m_mul(x,y); }
-static GEN
-_F2m_sqr(void *data, GEN x)
-{ (void) data; return F2m_mul(x,x); }
-GEN
-F2m_powu(GEN x, ulong n)
-{
-  pari_sp av = avma;
-  if (!n) return matid(lg(x)-1);
-  return gerepileupto(av, gen_powu(x, n,NULL, &_F2m_sqr, &_F2m_mul));
-}
 static GEN
 _FpM_mul(void *p , GEN x, GEN y)
 { return FpM_mul(x,y,(GEN)p); }
@@ -771,7 +767,6 @@ _FpM_sqr(void *p, GEN x)
 GEN
 FpM_powu(GEN x, ulong n, GEN p)
 {
-  pari_sp av = avma;
   if (!n) return matid(lg(x)-1);
   if (lgefint(p) == 3)
   {
@@ -784,7 +779,7 @@ FpM_powu(GEN x, ulong n, GEN p)
       z = Flm_to_ZM(Flm_powu(ZM_to_Flm(x, pp), n, pp));
     return gerepileupto(av, z);
   }
-  return gerepileupto(av, gen_powu(x, n, (void*)p, &_FpM_sqr, &_FpM_mul));
+  return gen_powu(x, n, (void*)p, &_FpM_sqr, &_FpM_mul);
 }
 
 /*Multiple a column vector by a line vector to make a matrix*/
@@ -907,25 +902,6 @@ Flx_dotproduct(GEN x, GEN y, ulong p)
     return Flv_dotproductspec_i(x+2, y+2, p, get_Fl_red(p), lx);
 }
 
-ulong
-F2v_dotproduct(GEN x, GEN y)
-{
-  long i, lx = lg(x);
-  ulong c;
-  if (lx <= 2) return 0;
-  c = uel(x,2) & uel(y,2);
-  for (i=3; i<lx; i++) c ^= uel(x,i) & uel(y,i);
-#ifdef LONG_IS_64BIT
-  c ^= c >> 32;
-#endif
-  c ^= c >> 16;
-  c ^= c >>  8;
-  c ^= c >>  4;
-  c ^= c >>  2;
-  c ^= c >>  1;
-  return c & 1;
-}
-
 GEN
 FpM_FpC_mul(GEN x, GEN y, GEN p)
 {
@@ -976,14 +952,6 @@ Flm_Flc_mul_pre_Flx(GEN x, GEN y, ulong p, ulong pi, long sv)
   return Flx_renormalize(z, l + 1);
 }
 
-GEN
-F2m_F2c_mul(GEN x, GEN y)
-{
-  long l, lx = lg(x);
-  if (lx==1) return cgetg(1,t_VECSMALL);
-  l = coeff(x,1,1);
-  return F2m_F2c_mul_i(x, y, lx, l);
-}
 /* RgV_to_RgX(FpM_FpC_mul(x,y,p), v), p != NULL, memory clean */
 GEN
 FpM_FpC_mul_FpX(GEN x, GEN y, GEN p, long v)
@@ -1004,9 +972,9 @@ FpM_FpC_mul_FpX(GEN x, GEN y, GEN p, long v)
       gel(z,i+1) = gerepileuptoint(av, p1);
       break;
     }
-    avma = av;
+    set_avma(av);
   }
-  if (!i) { avma = (pari_sp)(z + l+1); return pol_0(v); }
+  if (!i) { set_avma((pari_sp)(z + l+1)); return pol_0(v); }
   z[0] = evaltyp(t_POL) | evallg(i+2);
   z[1] = evalsigne(1) | evalvarn(v);
   for (; i; i--)
@@ -1058,16 +1026,6 @@ gen_matid(long n, void *E, const struct bb_field *S)
   return y;
 }
 
-GEN
-matid_F2m(long n)
-{
-  GEN y = cgetg(n+1,t_MAT);
-  long i;
-  if (n < 0) pari_err_DOMAIN("matid_F2m", "dimension","<",gen_0,stoi(n));
-  for (i=1; i<=n; i++) { gel(y,i) = zero_F2v(n); F2v_set(gel(y,i),i); }
-  return y;
-}
-
 GEN
 matid_F2xqM(long n, GEN T)
 {
@@ -1131,6 +1089,21 @@ Fp_to_mod(GEN z, GEN p)
   retmkintmod(modii(z, p), icopy(p));
 }
 
+GEN
+FpX_to_mod_raw(GEN z, GEN p)
+{
+  long i,l = lg(z);
+  GEN x;
+  if (l == 2)
+  {
+    x = cgetg(3,t_POL); x[1] = z[1];
+    gel(x,2) = mkintmod(gen_0,p); return x;
+  }
+  x = cgetg(l,t_POL);
+  for (i=2; i<l; i++) gel(x,i) = to_intmod(gel(z,i), p);
+  x[1] = z[1]; return normalizepol_lg(x,l);
+}
+
 /* z in Z[X], return z * Mod(1,p), normalized*/
 GEN
 FpX_to_mod(GEN z, GEN p)
@@ -1148,6 +1121,32 @@ FpX_to_mod(GEN z, GEN p)
   x[1] = z[1]; return normalizepol_lg(x,l);
 }
 
+GEN
+FpXC_to_mod(GEN z, GEN p)
+{
+  long i,l = lg(z);
+  GEN x = cgetg(l, t_COL);
+  p = icopy(p);
+  for (i=1; i<l; i++)
+    gel(x,i) = FpX_to_mod_raw(gel(z,i), p);
+  return x;
+}
+
+static GEN
+FpXC_to_mod_raw(GEN x, GEN p)
+{ pari_APPLY_type(t_COL, FpX_to_mod_raw(gel(x,i), p)) }
+
+GEN
+FpXM_to_mod(GEN z, GEN p)
+{
+  long i,l = lg(z);
+  GEN x = cgetg(l, t_MAT);
+  p = icopy(p);
+  for (i=1; i<l; i++)
+    gel(x,i) = FpXC_to_mod_raw(gel(z,i), p);
+  return x;
+}
+
 /* z in Z^n, return z * Mod(1,p), normalized*/
 GEN
 FpV_to_mod(GEN z, GEN p)
@@ -1238,19 +1237,36 @@ FpXQC_to_mod(GEN z, GEN T, GEN p)
   long i,l = lg(z);
   GEN x = cgetg(l, t_COL);
   if (l == 1) return x;
-  T = FpX_to_mod(T, p);
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
   for (i=1; i<l; i++)
-    gel(x,i) = mkpolmod(FpX_to_mod(gel(z,i), p), T);
+    gel(x,i) = mkpolmod(FpX_to_mod_raw(gel(z,i), p), T);
   return x;
 }
 
 static GEN
 Fq_to_mod_raw(GEN x, GEN T, GEN p)
 {
-  GEN z = typ(x)==t_INT ? Fp_to_mod(x, p): FpX_to_mod(x, p);
+  GEN z = typ(x)==t_INT ? mkintmod(modii(x, p), p): FpX_to_mod_raw(x, p);
   return mkpolmod(z, T);
 }
 
+/* z in Z^n, return z * Mod(1,p), normalized*/
+GEN
+FqC_to_mod(GEN z, GEN T, GEN p)
+{
+  GEN x;
+  long i,l = lg(z);
+  if (!T) return FpC_to_mod(z, p);
+  x = cgetg(l, t_COL);
+  if (l == 1) return x;
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
+  for (i=1; i<l; i++)
+    gel(x,i) = Fq_to_mod_raw(gel(z, i), T, p);
+  return x;
+}
+
 /* z in Z^n, return z * Mod(1,p), normalized*/
 static GEN
 FqC_to_mod_raw(GEN x, GEN T, GEN p)
@@ -1265,7 +1281,8 @@ FqM_to_mod(GEN z, GEN T, GEN p)
   if (!T) return FpM_to_mod(z, p);
   x = cgetg(l, t_MAT);
   if (l == 1) return x;
-  T = FpX_to_mod(T, p);
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
   for (i=1; i<l; i++)
     gel(x,i) = FqC_to_mod_raw(gel(z, i), T, p);
   return x;
@@ -1336,7 +1353,7 @@ gen_FpM_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p)
       }
     }
     b = FpX_renormalize(b, m+2);
-    if (lgpol(b)==0) {avma = btop; continue; }
+    if (lgpol(b)==0) {set_avma(btop); continue; }
     M = FpX_halfgcd(b, pol_xn(m, 0), p);
     Q = FpX_neg(FpX_normalize(gcoeff(M, 2, 1),p),p);
     W = B; lQ =lg(Q);
@@ -1365,7 +1382,7 @@ gen_FpM_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p)
         return gerepilecopy(ltop, shallowtrans(V));
       gerepileall(av, 2, &V, &W);
     }
-    avma = btop;
+    set_avma(btop);
   }
   return NULL;
 }
@@ -1495,7 +1512,7 @@ wrap_relker(void*E, GEN x) { return ZV_zMs_mul(x, (GEN)E); }
 
 /* Solve f(X) = B (mod p^e); blackbox version of ZlM_gauss */
 GEN
-gen_ZpM_Dixon(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
+gen_ZpM_Dixon_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
 {
   struct wrapper_modp_s W;
   pari_sp av = avma;
@@ -1513,7 +1530,7 @@ gen_ZpM_Dixon(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
     B = ZC_Z_divexact(ZC_sub(B, f(E, xi)), p);
     if (gc_needed(av,2))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"gen_ZpM_Dixon. i=%ld",i);
+      if(DEBUGMEM>1) pari_warn(warnmem,"gen_ZpM_Dixon_Wiedemann. i=%ld",i);
       gerepileall(av,3, &pi,&B,&xb);
     }
     xi = gen_FpM_Wiedemann((void*)&W, wrap_relcomb_modp, FpC_red(B, p), p);
@@ -1544,7 +1561,7 @@ ZpMs_ZpCs_solve(GEN M, GEN A, long nbrow, GEN p, long e)
   pari_timer ti;
   if (DEBUGLEVEL) timer_start(&ti);
   RgMs_structelim(M, nbrow, gel(A, 1), &pcol, &prow);
-  if (!pcol) { avma = av; return NULL; }
+  if (!pcol) return gc_NULL(av);
   if (DEBUGLEVEL)
     timer_printf(&ti,"structured elimination (%ld -> %ld)",nbi,lg(pcol)-1);
   n = lg(pcol)-1;
@@ -1553,9 +1570,9 @@ ZpMs_ZpCs_solve(GEN M, GEN A, long nbrow, GEN p, long e)
     gel(Mp, i) = vecprow(gel(M,pcol[i]), prow);
   Ap = zCs_to_ZC(vecprow(A, prow), n);
   if (DEBUGLEVEL) timer_start(&ti);
-  Rp = gen_ZpM_Dixon((void*)Mp,wrap_relcomb, Ap, p, e);
+  Rp = gen_ZpM_Dixon_Wiedemann((void*)Mp,wrap_relcomb, Ap, p, e);
   if (DEBUGLEVEL) timer_printf(&ti,"linear algebra");
-  if (!Rp) { avma = av; return NULL; }
+  if (!Rp) return gc_NULL(av);
   lR = lg(Rp)-1;
   if (typ(Rp) == t_COL)
   {
@@ -1612,8 +1629,7 @@ FpMs_structelim_back(GEN M, GEN V, GEN prow, GEN p)
         GEN C = gel(M,i), F = gel(C,1), E = gel(C,2);
         long c=0, cj=0, lF = lg(F);
         for(j=1; j<lF; j++)
-          if (!gel(R,F[j]))
-          { c++; cj=j; }
+          if (!gel(R,F[j])) { c++; cj=j; }
         if (c>=2) continue;
         if (c==1)
         {
@@ -1621,7 +1637,9 @@ FpMs_structelim_back(GEN M, GEN V, GEN prow, GEN p)
           GEN s = gen_0;
           for(j=1; j<lF; j++)
             if (j!=cj) s = Fp_add(s, mulis(gel(R,F[j]), E[j]), p);
-          gel(R,F[cj]) = gerepileupto(av, Fp_div(Fp_neg(s, p), stoi(E[cj]), p));
+          /* s /= -E[cj] mod p */
+          s = E[cj] < 0? Fp_divu(s, -E[cj], p): Fp_divu(Fp_neg(s,p), E[cj], p);
+          gel(R,F[cj]) = gerepileuptoint(av, s);
           f++;
         }
         W[i]=0;
@@ -1644,7 +1662,7 @@ FpMs_leftkernel_elt_col(GEN M, long nbcol, long nbrow, GEN p)
   struct wrapper_modp_s W;
   if (DEBUGLEVEL) timer_start(&ti);
   RgMs_structelim_col(M, nbcol, nbrow, cgetg(1,t_VECSMALL), &pcol, &prow);
-  if (!pcol) { avma = av; return NULL; }
+  if (!pcol) return gc_NULL(av);
   if (DEBUGLEVEL)
     timer_printf(&ti,"structured elimination (%ld -> %ld)",nbcol,lg(pcol)-1);
   n = lg(pcol)-1;
@@ -1657,7 +1675,7 @@ FpMs_leftkernel_elt_col(GEN M, long nbcol, long nbrow, GEN p)
   av2 = avma;
   for(;;)
   {
-    avma = av2;
+    set_avma(av2);
     B = random_FpV(n, p);
     MB = FpV_FpMs_mul(B, Mp, p);
     if (DEBUGLEVEL) timer_start(&ti);
