@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -31,7 +32,7 @@ get_FpX_red(GEN T, GEN *B)
 
 /* FpX are polynomials over Z/pZ represented as t_POL with
  * t_INT coefficients.
- * 1) Coefficients should belong to {0,...,p-1}, though non-reduced
+ * 1) Coefficients should belong to {0,...,p-1}, though nonreduced
  * coefficients should work but be slower.
  *
  * 2) p is not assumed to be prime, but it is assumed that impossible divisions
@@ -63,7 +64,7 @@ Z_to_FpX(GEN a, GEN p, long v)
   pari_sp av = avma;
   GEN z = cgetg(3, t_POL);
   GEN x = modii(a, p);
-  if (!signe(x)) { avma =av; return pol_0(v); }
+  if (!signe(x)) { set_avma(av); return pol_0(v); }
   z[1] = evalsigne(1) | evalvarn(v);
   gel(z,2) = x; return z;
 }
@@ -126,7 +127,7 @@ FpX_add(GEN x,GEN y,GEN p)
   for (i=2; i<ly; i++) gel(z,i) = Fp_add(gel(x,i),gel(y,i), p);
   for (   ; i<lx; i++) gel(z,i) = modii(gel(x,i), p);
   z = ZX_renormalize(z, lx);
-  if (!lgpol(z)) { avma = (pari_sp)(z + lx); return pol_0(varn(x)); }
+  if (!lgpol(z)) { set_avma((pari_sp)(z + lx)); return pol_0(varn(x)); }
   return z;
 }
 
@@ -235,7 +236,7 @@ FpX_subspec(GEN x,GEN y,GEN p, long nx, long ny)
     for (   ; i<ny; i++) gel(z,i) = Fp_neg(gel(y,i), p);
   }
   z = FpX_renormalize(z-2, lz);
-  if (!lgpol(z)) { avma = (pari_sp)(z + lz); return pol_0(0); }
+  if (!lgpol(z)) { set_avma((pari_sp)(z + lz)); return pol_0(0); }
   return z;
 }
 
@@ -255,14 +256,14 @@ Fp_FpX_sub(GEN x, GEN y, GEN p)
   if (ly <= 3) {
     z = cgetg(3, t_POL);
     x = (ly == 3)? Fp_sub(x, gel(y,2), p): modii(x, p);
-    if (!signe(x)) { avma = (pari_sp)(z + 3); return pol_0(varn(y)); }
+    if (!signe(x)) { set_avma((pari_sp)(z + 3)); return pol_0(varn(y)); }
     z[1] = evalsigne(1)|y[1]; gel(z,2) = x; return z;
   }
   z = cgetg(ly,t_POL);
   gel(z,2) = Fp_sub(x, gel(y,2), p);
   for (i = 3; i < ly; i++) gel(z,i) = Fp_neg(gel(y,i), p);
   z = ZX_renormalize(z, ly);
-  if (!lgpol(z)) { avma = (pari_sp)(z + ly); return pol_0(varn(x)); }
+  if (!lgpol(z)) { set_avma((pari_sp)(z + ly)); return pol_0(varn(x)); }
   z[1] = y[1]; return z;
 }
 
@@ -276,7 +277,7 @@ FpX_convol(GEN x, GEN y, GEN p)
   for (i=2; i<ly; i++) gel(z,i) = Fp_mul(gel(x,i),gel(y,i), p);
   for (   ; i<lx; i++) gel(z,i) = modii(gel(x,i), p);
   z = ZX_renormalize(z, lx);
-  if (!lgpol(z)) { avma = (pari_sp)(z + lx); return pol_0(varn(x)); }
+  if (!lgpol(z)) { set_avma((pari_sp)(z + lx)); return pol_0(varn(x)); }
   return z;
 }
 
@@ -318,6 +319,12 @@ FpX_mulu(GEN y, ulong x,GEN p)
   return z;
 }
 
+GEN
+FpX_divu(GEN y, ulong x, GEN p)
+{
+  return FpX_Fp_div(y, utoi(umodui(x, p)), p);
+}
+
 GEN
 FpX_Fp_mulspec(GEN y,GEN x,GEN p,long ly)
 {
@@ -336,6 +343,12 @@ FpX_Fp_mul(GEN y,GEN x,GEN p)
   setvarn(z, varn(y)); return z;
 }
 
+GEN
+FpX_Fp_div(GEN y, GEN x, GEN p)
+{
+  return FpX_Fp_mul(y, Fp_inv(x, p), p);
+}
+
 GEN
 FpX_Fp_mul_to_monic(GEN y,GEN x,GEN p)
 {
@@ -398,7 +411,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
     if (pr)
     {
       av0 = avma; x = FpX_red(x, p);
-      if (pr == ONLY_DIVIDES) { avma=av0; return signe(x)? NULL: pol_0(vx); }
+      if (pr == ONLY_DIVIDES) { set_avma(av0); return signe(x)? NULL: pol_0(vx); }
       if (pr == ONLY_REM) return x;
       *pr = x;
     }
@@ -414,14 +427,14 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
     }
     av0 = avma;
     if (equali1(lead)) return FpX_red(x, p);
-    else return gerepileupto(av0, FpX_Fp_mul(x, Fp_inv(lead,p), p));
+    else return gerepileupto(av0, FpX_Fp_div(x, lead, p));
   }
   av0 = avma; dz = dx-dy;
   if (lgefint(p) == 3)
   { /* assume ab != 0 mod p */
     ulong pp = to_Flx(&x, &y, p);
     z = Flx_divrem(x, y, pp, pr);
-    avma = av0; /* HACK: assume pr last on stack, then z */
+    set_avma(av0); /* HACK: assume pr last on stack, then z */
     if (!z) return NULL;
     z = leafcopy(z);
     if (pr && pr != ONLY_DIVIDES && pr != ONLY_REM)
@@ -432,7 +445,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
     return Flx_to_ZX_inplace(z);
   }
   lead = equali1(lead)? NULL: gclone(Fp_inv(lead,p));
-  avma = av0;
+  set_avma(av0);
   z=cgetg(dz+3,t_POL); z[1] = x[1];
   x += 2; y += 2; z += 2;
   for (dy1=dy-1; dy1>=0 && !signe(gel(y, dy1)); dy1--);
@@ -447,7 +460,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
     if (lead) p1 = mulii(p1,lead);
     gel(z,i-dy) = gerepileuptoint(av,modii(p1, p));
   }
-  if (!pr) { if (lead) gunclone(lead); return z-2; }
+  if (!pr) { guncloneNULL(lead); return z-2; }
 
   rem = (GEN)avma; av = (pari_sp)new_chunk(dx+3);
   for (sx=0; ; i--)
@@ -457,13 +470,13 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     p1 = modii(p1,p); if (signe(p1)) { sx = 1; break; }
     if (!i) break;
-    avma=av;
+    set_avma(av);
   }
   if (pr == ONLY_DIVIDES)
   {
-    if (lead) gunclone(lead);
-    if (sx) { avma=av0; return NULL; }
-    avma = (pari_sp)rem; return z-2;
+    guncloneNULL(lead);
+    if (sx) return gc_NULL(av0);
+    return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
   rem[0] = evaltyp(t_POL) | evallg(lr);
@@ -478,7 +491,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
     gel(rem,i) = gerepileuptoint(av, modii(p1,p));
   }
   rem -= 2;
-  if (lead) gunclone(lead);
+  guncloneNULL(lead);
   if (!sx) (void)FpX_renormalize(rem, lr);
   if (pr == ONLY_REM) return gerepileupto(av0,rem);
   *pr = rem; return z-2;
@@ -487,12 +500,17 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
 GEN
 FpX_div_by_X_x(GEN a, GEN x, GEN p, GEN *r)
 {
-  long l = lg(a)-1, i;
-  GEN z = cgetg(l, t_POL);
-  z[1] = evalsigne(1) | evalvarn(0);
+  long l = lg(a), i;
+  GEN z;
+  if (l <= 3)
+  {
+    if (r) *r = l == 2? gen_0: icopy(gel(a,2));
+    return pol_0(0);
+  }
+  l--; z = cgetg(l, t_POL); z[1] = evalsigne(1) | evalvarn(0);
   gel(z, l-1) = gel(a,l);
-  for (i=l-2; i>1; i--) /* z[i] = a[i+1] + x*z[i+1] */
-    gel(z, i) = Fp_addmul(gel(a,i+1), x, gel(z,i+1), p);
+  for (i = l-2; i > 1; i--) /* z[i] = a[i+1] + x*z[i+1] */
+    gel(z,i) = Fp_addmul(gel(a,i+1), x, gel(z,i+1), p);
   if (r) *r = Fp_addmul(gel(a,2), x, gel(z,2), p);
   return z;
 }
@@ -626,6 +644,9 @@ matid2_FpXM(long v)
             mkcol2(pol_0(v),pol_1(v)));
 }
 
+static GEN
+FpX_shift(GEN a, long n) { return RgX_shift_shallow(a, n); }
+
 static GEN
 FpX_halfgcd_split(GEN x, GEN y, GEN p)
 {
@@ -634,12 +655,12 @@ FpX_halfgcd_split(GEN x, GEN y, GEN p)
   GEN y1, r, q;
   long l = lgpol(x), n = l>>1, k;
   if (lgpol(y)<=n) return matid2_FpXM(varn(x));
-  R = FpX_halfgcd(RgX_shift_shallow(x,-n),RgX_shift_shallow(y,-n),p);
+  R = FpX_halfgcd(FpX_shift(x,-n), FpX_shift(y,-n), p);
   V = FpXM_FpX_mul2(R,x,y,p); y1 = gel(V,2);
   if (lgpol(y1)<=n) return gerepilecopy(av, R);
   q = FpX_divrem(gel(V,1), y1, p, &r);
   k = 2*n-degpol(y1);
-  S = FpX_halfgcd(RgX_shift_shallow(y1,-k), RgX_shift_shallow(r,-k),p);
+  S = FpX_halfgcd(FpX_shift(y1,-k), FpX_shift(r,-k), p);
   return gerepileupto(av, FpXM_mul2(S,FpX_FpXM_qmul(q,R,p),p));
 }
 
@@ -695,7 +716,7 @@ FpX_gcd_basecase(GEN a, GEN b, GEN p)
     }
     av = avma; c = FpX_rem(a,b,p); a=b; b=c;
   }
-  avma = av; return a;
+  return gc_const(av, a);
 }
 
 GEN
@@ -708,7 +729,7 @@ FpX_gcd(GEN x, GEN y, GEN p)
     (void)new_chunk((lg(x) + lg(y)) << 2); /* scratch space */
     pp = to_Flx(&x, &y, p);
     x = Flx_gcd(x, y, pp);
-    avma = av; return Flx_to_ZX(x);
+    set_avma(av); return Flx_to_ZX(x);
   }
   x = FpX_red(x, p);
   y = FpX_red(y, p);
@@ -739,17 +760,17 @@ FpX_gcd_check(GEN x, GEN y, GEN p)
   b = FpX_red(y, p);
   while (signe(b))
   {
-    GEN g = gcdii(p, leading_coeff(b));
-    if (!equali1(g)) return gerepileuptoint(av,g);
-    c = FpX_rem(a,b,p); a = b; b = c;
+    GEN g;
+    if (!invmod(leading_coeff(b), p, &g)) return gerepileuptoint(av,g);
+    b = FpX_Fp_mul_to_monic(b, g, p);
+    c = FpX_rem(a, b, p); a = b; b = c;
     if (gc_needed(av,1))
     {
-      if (DEBUGMEM>1)
-        pari_warn(warnmem,"FpX_gcd_check (d = %ld)",degpol(b));
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_gcd_check (d = %ld)",degpol(b));
       gerepileall(av,2,&a,&b);
     }
   }
-  avma = av; return NULL;
+  return gc_NULL(av);
 }
 
 static GEN
@@ -835,7 +856,7 @@ FpX_rescale(GEN P, GEN h, GEN p)
 {
   long i, l = lg(P);
   GEN Q = cgetg(l,t_POL), hi = h;
-  Q[l-1] = P[l-1];
+  gel(Q,l-1) = gel(P,l-1);
   for (i=l-2; i>=2; i--)
   {
     gel(Q,i) = Fp_mul(gel(P,i), hi, p);
@@ -848,6 +869,32 @@ FpX_rescale(GEN P, GEN h, GEN p)
 GEN
 FpX_deriv(GEN x, GEN p) { return FpX_red(ZX_deriv(x), p); }
 
+/* Compute intformal(x^n*S)/x^(n+1) */
+static GEN
+FpX_integXn(GEN x, long n, GEN p)
+{
+  long i, lx = lg(x);
+  GEN y;
+  if (lx == 2) return ZX_copy(x);
+  y = cgetg(lx, t_POL); y[1] = x[1];
+  for (i=2; i<lx; i++)
+  {
+    GEN xi = gel(x,i);
+    if (!signe(xi))
+      gel(y,i) = gen_0;
+    else
+    {
+      ulong j = n+i-1;
+      ulong d = ugcd(j, umodiu(xi, j));
+      if (d==1)
+        gel(y,i) = Fp_divu(xi, j, p);
+      else
+        gel(y,i) = Fp_divu(diviuexact(xi, d), j/d, p);
+    }
+  }
+  return ZX_renormalize(y, lx);;
+}
+
 GEN
 FpX_integ(GEN x, GEN p)
 {
@@ -857,7 +904,7 @@ FpX_integ(GEN x, GEN p)
   y = cgetg(lx+1, t_POL); y[1] = x[1];
   gel(y,2) = gen_0;
   for (i=3; i<=lx; i++)
-    gel(y,i) = Fp_div(gel(x,i-1), utoipos(i-2), p);
+    gel(y,i) = signe(gel(x,i-1))? Fp_divu(gel(x,i-1), i-2, p): gen_0;
   return ZX_renormalize(y, lx+1);;
 }
 
@@ -870,7 +917,7 @@ FpX_Newton(GEN P, long n, GEN p)
 {
   pari_sp av = avma;
   GEN dP = FpX_deriv(P, p);
-  GEN Q = FpXn_recip(FpX_div(RgX_shift(dP,n), P, p), n);
+  GEN Q = FpXn_recip(FpX_div(FpX_shift(dP,n), P, p), n);
   return gerepilecopy(av, Q);
 }
 
@@ -878,24 +925,37 @@ GEN
 FpX_fromNewton(GEN P, GEN p)
 {
   pari_sp av = avma;
-  long n = itos(modii(constant_coeff(P), p))+1;
-  GEN z = FpX_neg(FpX_integ(RgX_shift(P,-1),p),p);
-  GEN Q = FpXn_recip(FpXn_exp(z, n, p), n);
-  return gerepilecopy(av, Q);
+  if (lgefint(p)==3)
+  {
+    ulong pp = p[2];
+    GEN Q = Flx_fromNewton(ZX_to_Flx(P, pp), pp);
+    return gerepileupto(av, Flx_to_ZX(Q));
+  } else
+  {
+    long n = itos(modii(constant_coeff(P), p))+1;
+    GEN z = FpX_neg(FpX_shift(P,-1),p);
+    GEN Q = FpXn_recip(FpXn_expint(z, n, p), n);
+    return gerepilecopy(av, Q);
+  }
 }
 
 GEN
 FpX_invLaplace(GEN x, GEN p)
 {
   pari_sp av = avma;
-  long i, e = 0, l = lg(x);
-  GEN t = gen_1, y = cgetg(l,t_POL);
+  long i, d = degpol(x);
+  GEN t, y;
+  if (d <= 1) return gcopy(x);
+  t = Fp_inv(factorial_Fp(d, p), p);
+  y = cgetg(d+3, t_POL);
   y[1] = x[1];
-  for (i=2; i<l; i++)
+  for (i=d; i>=2; i--)
   {
-    gel(y,i) = Fp_div(gel(x,i), t, p);
-    e++; t = Fp_mulu(t,e,p);
+    gel(y,i+2) = Fp_mul(gel(x,i+2), t, p);
+    t = Fp_mulu(t, i, p);
   }
+  gel(y,3) = gel(x,3);
+  gel(y,2) = gel(x,2);
   return gerepilecopy(av, y);
 }
 
@@ -903,13 +963,18 @@ GEN
 FpX_Laplace(GEN x, GEN p)
 {
   pari_sp av = avma;
-  long i, e = 0, l = lg(x);
-  GEN t = gen_1, y = cgetg(l,t_POL);
+  long i, d = degpol(x);
+  GEN t = gen_1;
+  GEN y;
+  if (d <= 1) return gcopy(x);
+  y = cgetg(d+3, t_POL);
   y[1] = x[1];
-  for (i=2; i<l; i++)
+  gel(y,2) = gel(x,2);
+  gel(y,3) = gel(x,3);
+  for (i=2; i<=d; i++)
   {
-    gel(y,i) = Fp_mul(gel(x,i), t, p);
-    e++; t = Fp_mulu(t,e,p);
+    t = Fp_mulu(t, i, p);
+    gel(y,i+2) = Fp_mul(gel(x,i+2), t, p);
   }
   return gerepilecopy(av, y);
 }
@@ -919,7 +984,7 @@ FpX_is_squarefree(GEN f, GEN p)
 {
   pari_sp av = avma;
   GEN z = FpX_gcd(f,FpX_deriv(f,p),p);
-  avma = av;
+  set_avma(av);
   return degpol(z)==0;
 }
 
@@ -972,11 +1037,10 @@ FpX_eval(GEN x,GEN y,GEN p)
       }
     r = (i==j)? y: Fp_powu(y,i-j+1,p);
     p1 = Fp_addmul(gel(x,j), p1, r, p);
-    if ((i & 7) == 0) { affii(p1, res); p1 = res; avma = av; }
+    if ((i & 7) == 0) { affii(p1, res); p1 = res; set_avma(av); }
   }
  fppoleval:
-  modiiz(p1,p,res);
-  avma = av; return res;
+  modiiz(p1,p,res); return gc_const(av, res);
 }
 
 /* Tz=Tx*Ty where Tx and Ty coprime
@@ -1012,7 +1076,7 @@ FpX_resultant(GEN a, GEN b, GEN p)
     pari_sp av = avma;
     ulong pp = to_Flx(&a, &b, p);
     long r = Flx_resultant(a, b, pp);
-    avma = av;
+    set_avma(av);
     return utoi(r);
   }
 
@@ -1030,7 +1094,7 @@ FpX_resultant(GEN a, GEN b, GEN p)
     lb = gel(b,db+2);
     c = FpX_rem(a,b, p);
     a = b; b = c; dc = degpol(c);
-    if (dc < 0) { avma = av; return gen_0; }
+    if (dc < 0) return gc_const(av, gen_0);
 
     if (both_odd(da,db)) res = subii(p, res);
     if (!equali1(lb)) res = Fp_mul(res, Fp_powu(lb, da - dc, p), p);
@@ -1267,23 +1331,23 @@ FpX_divrem_Barrettspec(GEN x, long l, GEN mg, GEN T, GEN p, GEN *pr)
 }
 
 static GEN
-FpX_divrem_Barrett_noGC(GEN x, GEN mg, GEN T, GEN p, GEN *pr)
+FpX_divrem_Barrett(GEN x, GEN mg, GEN T, GEN p, GEN *pr)
 {
   GEN q = NULL, r = FpX_red(x, p);
-  long l = lgpol(r), lt = degpol(T), lm = 2*lt-1;
+  long l = lgpol(r), lt = degpol(T), lm = 2*lt-1, v = varn(T);
   long i;
   if (l <= lt)
   {
     if (pr == ONLY_REM) return r;
-    if (pr == ONLY_DIVIDES) return signe(x)? NULL: pol_0(varn(x));
+    if (pr == ONLY_DIVIDES) return signe(r)? NULL: pol_0(v);
     if (pr) *pr = r;
-    return pol_0(varn(T));
+    return pol_0(v);
   }
   if (lt <= 1)
     return FpX_divrem_basecase(r,T,p,pr);
   if (pr != ONLY_REM && l>lm)
   {
-    q = cgetg(l-lt+2, t_POL);
+    q = cgetg(l-lt+2, t_POL); q[1] = T[1];
     for (i=0;i<l-lt;i++) gel(q+2,i) = gen_0;
   }
   while (l>lm)
@@ -1298,42 +1362,40 @@ FpX_divrem_Barrett_noGC(GEN x, GEN mg, GEN T, GEN p, GEN *pr)
     for(i=0; i<lz; i++) gel(r+2+l-lm,i) = gel(zr,2+i);
     l = l-lm+lz;
   }
-  if (pr != ONLY_REM)
-  {
-    if (l > lt)
-    {
-      GEN zq = FpX_divrem_Barrettspec(r+2,l,mg,T,p,&r);
-      if (!q) q = zq;
-      else
-      {
-        long lq = lgpol(zq);
-        for(i=0; i<lq; i++) gel(q+2,i) = gel(zq,2+i);
-      }
-    }
-    else
-      r = FpX_renormalize(r, l+2);
-  }
-  else
+  if (pr == ONLY_REM)
   {
     if (l > lt)
       r = FpX_divrem_Barrettspec(r+2, l, mg, T, p, ONLY_REM);
     else
       r = FpX_renormalize(r, l+2);
-    r[1] = x[1]; return FpX_renormalize(r, lg(r));
+    setvarn(r, v); return r;
   }
-  if (pr) { r[1] = x[1]; r = FpX_renormalize(r, lg(r)); }
-  q[1] = x[1]; q = FpX_renormalize(q, lg(q));
+  if (l > lt)
+  {
+    GEN zq = FpX_divrem_Barrettspec(r+2,l,mg,T,p, pr? &r: NULL);
+    if (!q) q = zq;
+    else
+    {
+      long lq = lgpol(zq);
+      for(i=0; i<lq; i++) gel(q+2,i) = gel(zq,2+i);
+    }
+  }
+  else if (pr)
+    r = FpX_renormalize(r, l+2);
+  setvarn(q, v); q = FpX_renormalize(q, lg(q));
   if (pr == ONLY_DIVIDES) return signe(r)? NULL: q;
-  if (pr) *pr = r;
+  if (pr) { setvarn(r, v); *pr = r; }
   return q;
 }
 
 GEN
 FpX_divrem(GEN x, GEN T, GEN p, GEN *pr)
 {
-  GEN B, y = get_FpX_red(T, &B);
-  long dy = degpol(y), dx = degpol(x), d = dx-dy;
-  if (pr==ONLY_REM) return FpX_rem(x, y, p);
+  GEN B, y;
+  long dy, dx, d;
+  if (pr==ONLY_REM) return FpX_rem(x, T, p);
+  y = get_FpX_red(T, &B);
+  dy = degpol(y); dx = degpol(x); d = dx-dy;
   if (!B && d+3 < FpX_DIVREM_BARRETT_LIMIT)
     return FpX_divrem_basecase(x,y,p,pr);
   else if (lgefint(p)==3)
@@ -1352,8 +1414,8 @@ FpX_divrem(GEN x, GEN T, GEN p, GEN *pr)
   {
     pari_sp av=avma;
     GEN mg = B? B: FpX_invBarrett(y, p);
-    GEN q1 = FpX_divrem_Barrett_noGC(x,mg,y,p,pr);
-    if (!q1) {avma=av; return NULL;}
+    GEN q1 = FpX_divrem_Barrett(x,mg,y,p,pr);
+    if (!q1) return gc_NULL(av);
     if (!pr || pr==ONLY_DIVIDES) return gerepilecopy(av, q1);
     gerepileall(av,2,&q1,pr);
     return q1;
@@ -1377,24 +1439,15 @@ FpX_rem(GEN x, GEN T, GEN p)
   {
     pari_sp av = avma;
     GEN mg = B? B: FpX_invBarrett(y, p);
-    return gerepileupto(av, FpX_divrem_Barrett_noGC(x, mg, y, p, ONLY_REM));
+    return gerepileupto(av, FpX_divrem_Barrett(x, mg, y, p, ONLY_REM));
   }
 }
 
 static GEN
-FpV_producttree(GEN xa, GEN s, GEN p, long vs)
+FpXV_producttree_dbl(GEN t, long n, GEN p)
 {
-  long n = lg(xa)-1;
-  long m = n==1 ? 1: expu(n-1)+1;
-  long i, j, k, ls = lg(s);
+  long i, j, k, m = n==1 ? 1: expu(n-1)+1;
   GEN T = cgetg(m+1, t_VEC);
-  GEN t = cgetg(ls, t_VEC);
-  for (j=1, k=1; j<ls; k+=s[j++])
-    gel(t, j) = s[j] == 1 ?
-             deg1pol(gen_1, Fp_neg(gel(xa,k), p), vs):
-             deg2pol_shallow(gen_1,
-               Fp_neg(Fp_add(gel(xa,k), gel(xa,k+1), p), p),
-               Fp_mul(gel(xa,k), gel(xa,k+1), p), vs);
   gel(T,1) = t;
   for (i=2; i<=m; i++)
   {
@@ -1409,9 +1462,23 @@ FpV_producttree(GEN xa, GEN s, GEN p, long vs)
 }
 
 static GEN
-FpX_FpV_multieval_tree(GEN P, GEN xa, GEN T, GEN p)
+FpV_producttree(GEN xa, GEN s, GEN p, long vs)
+{
+  long n = lg(xa)-1;
+  long j, k, ls = lg(s);
+  GEN t = cgetg(ls, t_VEC);
+  for (j=1, k=1; j<ls; k+=s[j++])
+    gel(t, j) = s[j] == 1 ?
+             deg1pol(gen_1, Fp_neg(gel(xa,k), p), vs):
+             deg2pol_shallow(gen_1,
+               Fp_neg(Fp_add(gel(xa,k), gel(xa,k+1), p), p),
+               Fp_mul(gel(xa,k), gel(xa,k+1), p), vs);
+  return FpXV_producttree_dbl(t, n, p);
+}
+
+static GEN
+FpX_FpXV_multirem_dbl_tree(GEN P, GEN T, GEN p)
 {
-  pari_sp av = avma;
   long i,j,k;
   long m = lg(T)-1;
   GEN t;
@@ -1430,19 +1497,26 @@ FpX_FpV_multieval_tree(GEN P, GEN xa, GEN T, GEN p)
     }
     gel(Tp, i) = t;
   }
+  return Tp;
+}
+
+static GEN
+FpX_FpV_multieval_tree(GEN P, GEN xa, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long j,k;
+  GEN Tp = FpX_FpXV_multirem_dbl_tree(P, T, p);
+  GEN R = cgetg(lg(xa), t_VEC);
+  GEN u = gel(T, 1);
+  GEN v = gel(Tp, 1);
+  long n = lg(u)-1;
+  for (j=1, k=1; j<=n; j++)
   {
-    GEN R = cgetg(lg(xa), t_VEC);
-    GEN u = gel(T, i+1);
-    GEN v = gel(Tp, i+1);
-    long n = lg(u)-1;
-    for (j=1, k=1; j<=n; j++)
-    {
-      long c, d = degpol(gel(u,j));
-      for (c=1; c<=d; c++, k++)
-        gel(R,k) = FpX_eval(gel(v, j), gel(xa,k), p);
-    }
-    return gerepileupto(av, R);
+    long c, d = degpol(gel(u,j));
+    for (c=1; c<=d; c++, k++)
+      gel(R,k) = FpX_eval(gel(v, j), gel(xa,k), p);
   }
+  return gerepileupto(av, R);
 }
 
 static GEN
@@ -1492,11 +1566,19 @@ GEN
 FpV_polint(GEN xa, GEN ya, GEN p, long vs)
 {
   pari_sp av = avma;
-  GEN s = producttree_scheme(lg(xa)-1);
-  GEN T = FpV_producttree(xa, s, p, vs);
-  long m = lg(T)-1;
-  GEN P = FpX_deriv(gmael(T, m, 1), p);
-  GEN R = FpV_inv(FpX_FpV_multieval_tree(P, xa, T, p), p);
+  GEN s, T, P, R;
+  long m;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = p[2];
+    P = Flv_polint(ZV_to_Flv(xa, pp), ZV_to_Flv(ya, pp), pp, evalvarn(vs));
+    return gerepileupto(av, Flx_to_ZX(P));
+  }
+  s = producttree_scheme(lg(xa)-1);
+  T = FpV_producttree(xa, s, p, vs);
+  m = lg(T)-1;
+  P = FpX_deriv(gmael(T, m, 1), p);
+  R = FpV_inv(FpX_FpV_multieval_tree(P, xa, T, p), p);
   return gerepileupto(av, FpVV_polint_tree(T, R, s, xa, ya, p, vs));
 }
 
@@ -1536,6 +1618,132 @@ FpV_invVandermonde(GEN L, GEN den, GEN p)
   return gerepilecopy(av, M);
 }
 
+static GEN
+FpXV_producttree(GEN xa, GEN s, GEN p)
+{
+  long n = lg(xa)-1;
+  long j, k, ls = lg(s);
+  GEN t = cgetg(ls, t_VEC);
+  for (j=1, k=1; j<ls; k+=s[j++])
+    gel(t, j) = s[j] == 1 ?
+             gel(xa,k): FpX_mul(gel(xa,k),gel(xa,k+1),p);
+  return FpXV_producttree_dbl(t, n, p);
+}
+
+static GEN
+FpX_FpXV_multirem_tree(GEN P, GEN xa, GEN T, GEN s, GEN p)
+{
+  pari_sp av = avma;
+  long j, k, ls = lg(s);
+  GEN Tp = FpX_FpXV_multirem_dbl_tree(P, T, p);
+  GEN R = cgetg(lg(xa), t_VEC);
+  GEN v = gel(Tp, 1);
+  for (j=1, k=1; j<ls; k+=s[j++])
+  {
+    gel(R,k) = FpX_rem(gel(v, j), gel(xa,k), p);
+    if (s[j] == 2)
+      gel(R,k+1) = FpX_rem(gel(v, j), gel(xa,k+1), p);
+  }
+  return gerepileupto(av, R);
+}
+
+GEN
+FpX_FpXV_multirem(GEN P, GEN xa, GEN p)
+{
+  pari_sp av = avma;
+  GEN s = producttree_scheme(lg(xa)-1);
+  GEN T = FpXV_producttree(xa, s, p);
+  return gerepileupto(av, FpX_FpXV_multirem_tree(P, xa, T, s, p));
+}
+
+/* T = ZV_producttree(P), R = ZV_chinesetree(P,T) */
+static GEN
+FpXV_chinese_tree(GEN A, GEN P, GEN T, GEN R, GEN s, GEN p)
+{
+  long m = lg(T)-1, ls = lg(s);
+  long i,j,k;
+  GEN Tp = cgetg(m+1, t_VEC);
+  GEN M = gel(T, 1);
+  GEN t = cgetg(lg(M), t_VEC);
+  for (j=1, k=1; j<ls; k+=s[j++])
+    if (s[j] == 2)
+    {
+      pari_sp av = avma;
+      GEN a = FpX_mul(gel(A,k), gel(R,k), p), b = FpX_mul(gel(A,k+1), gel(R,k+1), p);
+      GEN tj = FpX_rem(FpX_add(FpX_mul(gel(P,k), b, p),
+            FpX_mul(gel(P,k+1), a, p), p), gel(M,j), p);
+      gel(t, j) = gerepileupto(av, tj);
+    }
+    else
+      gel(t, j) = FpX_rem(FpX_mul(gel(A,k), gel(R,k), p), gel(M, j), p);
+  gel(Tp, 1) = t;
+  for (i=2; i<=m; i++)
+  {
+    GEN u = gel(T, i-1), M = gel(T, i);
+    GEN t = cgetg(lg(M), t_VEC);
+    GEN v = gel(Tp, i-1);
+    long n = lg(v)-1;
+    for (j=1, k=1; k<n; j++, k+=2)
+    {
+      pari_sp av = avma;
+      gel(t, j) = gerepileupto(av, FpX_rem(FpX_add(FpX_mul(gel(u, k), gel(v, k+1), p),
+              FpX_mul(gel(u, k+1), gel(v, k), p), p), gel(M, j), p));
+    }
+    if (k==n) gel(t, j) = gel(v, k);
+    gel(Tp, i) = t;
+  }
+  return gmael(Tp,m,1);
+}
+
+static GEN
+FpXV_sqr(GEN x, GEN p)
+{ pari_APPLY_type(t_VEC, FpX_sqr(gel(x,i), p)) }
+
+static GEN
+FpXT_sqr(GEN x, GEN p)
+{
+  if (typ(x) == t_POL)
+    return FpX_sqr(x, p);
+  pari_APPLY_type(t_VEC, FpXT_sqr(gel(x,i), p))
+}
+
+static GEN
+FpXV_invdivexact(GEN x, GEN y, GEN p)
+{ pari_APPLY_type(t_VEC, FpXQ_inv(FpX_div(gel(x,i), gel(y,i),p), gel(y,i),p)) }
+
+static GEN
+FpXV_chinesetree(GEN P, GEN T, GEN s, GEN p)
+{
+  GEN T2 = FpXT_sqr(T, p), P2 = FpXV_sqr(P, p);
+  GEN mod = gmael(T,lg(T)-1,1);
+  return FpXV_invdivexact(FpX_FpXV_multirem_tree(mod, P2, T2, s, p), P, p);
+}
+
+static GEN
+gc_chinese(pari_sp av, GEN T, GEN a, GEN *pt_mod)
+{
+  if (!pt_mod)
+    return gerepileupto(av, a);
+  else
+  {
+    GEN mod = gmael(T, lg(T)-1, 1);
+    gerepileall(av, 2, &a, &mod);
+    *pt_mod = mod;
+    return a;
+  }
+}
+
+GEN
+FpXV_chinese(GEN A, GEN P, GEN p, GEN *pt_mod)
+{
+  pari_sp av = avma;
+  GEN s = producttree_scheme(lg(P)-1);
+  GEN T = FpXV_producttree(P, s, p);
+  GEN R = FpXV_chinesetree(P, T, s, p);
+  GEN a = FpXV_chinese_tree(A, P, T, R, s, p);
+  return gc_chinese(av, T, a, pt_mod);
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                              FpXQ                                 **/
@@ -1688,8 +1896,8 @@ FpXQ_pow(GEN x, GEN n, GEN T, GEN p)
   }
   if (s < 0) x = FpXQ_inv(x,T,p);
   D.p = p; D.T = FpX_get_red(T,p);
-  y = gen_pow(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
-  return gerepileupto(av, y);
+  y = gen_pow_i(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
+  return gerepilecopy(av, y);
 }
 
 GEN /*Assume n is very small*/
@@ -1708,8 +1916,8 @@ FpXQ_powu(GEN x, ulong n, GEN T, GEN p)
     return Flx_to_ZX_inplace(gerepileuptoleaf(av, y));
   }
   D.T = FpX_get_red(T, p); D.p = p;
-  y = gen_powu(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
-  return gerepileupto(av, y);
+  y = gen_powu_i(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
+  return gerepilecopy(av, y);
 }
 
 /* generates the list of powers of x of degree 0,1,2,...,l*/
@@ -1774,23 +1982,19 @@ FpX_FpXQ_eval(GEN Q, GEN x, GEN T, GEN p)
 }
 
 GEN
-FpXC_FpXQV_eval(GEN P, GEN x, GEN T, GEN p)
-{
-  long i, l = lg(P);
-  GEN res = cgetg(l, t_COL);
-  for (i=1; i<l; i++)
-    gel(res,i) = FpX_FpXQV_eval(gel(P,i), x, T, p);
-  return res;
-}
+FpXC_FpXQV_eval(GEN x, GEN v, GEN T, GEN p)
+{ pari_APPLY_type(t_COL, FpX_FpXQV_eval(gel(x,i), v, T, p)) }
 
 GEN
-FpXM_FpXQV_eval(GEN Q, GEN x, GEN T, GEN p)
+FpXM_FpXQV_eval(GEN x, GEN v, GEN T, GEN p)
+{ pari_APPLY_same(FpXC_FpXQV_eval(gel(x,i), v, T, p)) }
+
+GEN
+FpXC_FpXQ_eval(GEN x, GEN F, GEN T, GEN p)
 {
-  long i, l = lg(Q);
-  GEN y = cgetg(l, t_MAT);
-  for (i=1; i<l; i++)
-    gel(y,i) = FpXC_FpXQV_eval(gel(Q,i), x, T, p);
-  return y;
+  long d = brent_kung_optpow(RgXV_maxdegree(x), lg(x)-1, 1);
+  GEN Fp = FpXQ_powers(F, d, T, p);
+  return FpXC_FpXQV_eval(x, Fp, T, p);
 }
 
 GEN
@@ -1819,20 +2023,25 @@ FpXQ_autpow_sqr(void *E, GEN x)
 }
 
 static GEN
-FpXQ_autpow_mul(void *E, GEN x, GEN y)
+FpXQ_autpow_msqr(void *E, GEN x)
 {
   struct _FpXQ *D = (struct _FpXQ*)E;
-  return FpX_FpXQ_eval(x, y, D->T, D->p);
+  return FpX_FpXQV_eval(FpXQ_autpow_sqr(E, x), D->aut, D->T, D->p);
 }
 
 GEN
 FpXQ_autpow(GEN x, ulong n, GEN T, GEN p)
 {
+  pari_sp av = avma;
   struct _FpXQ D;
+  long d;
   if (n==0) return FpX_rem(pol_x(varn(x)), T, p);
   if (n==1) return FpX_rem(x, T, p);
   D.T = FpX_get_red(T, p); D.p = p;
-  return gen_powu(x,n,(void*)&D,FpXQ_autpow_sqr,FpXQ_autpow_mul);
+  d = brent_kung_optpow(degpol(T), hammingl(n)-1, 1);
+  D.aut = FpXQ_powers(x, d, T, p);
+  x = gen_powu_fold(x,n,(void*)&D,FpXQ_autpow_sqr,FpXQ_autpow_msqr);
+  return gerepilecopy(av, x);
 }
 
 static GEN
@@ -1857,9 +2066,11 @@ FpXQ_auttrace_sqr(void *E, GEN x)
 GEN
 FpXQ_auttrace(GEN x, ulong n, GEN T, GEN p)
 {
+  pari_sp av = avma;
   struct _FpXQ D;
   D.T = FpX_get_red(T, p); D.p = p;
-  return gen_powu(x,n,(void*)&D,FpXQ_auttrace_sqr,FpXQ_auttrace_mul);
+  x = gen_powu_i(x,n,(void*)&D,FpXQ_auttrace_sqr,FpXQ_auttrace_mul);
+  return gerepilecopy(av, x);
 }
 
 static GEN
@@ -1883,9 +2094,11 @@ FpXQ_autsum_sqr(void *E, GEN x)
 GEN
 FpXQ_autsum(GEN x, ulong n, GEN T, GEN p)
 {
+  pari_sp av = avma;
   struct _FpXQ D;
   D.T = FpX_get_red(T, p); D.p = p;
-  return gen_powu(x,n,(void*)&D,FpXQ_autsum_sqr,FpXQ_autsum_mul);
+  x = gen_powu_i(x,n,(void*)&D,FpXQ_autsum_sqr,FpXQ_autsum_mul);
+  return gerepilecopy(av, x);
 }
 
 static GEN
@@ -1910,9 +2123,11 @@ FpXQM_autsum_sqr(void *E, GEN x)
 GEN
 FpXQM_autsum(GEN x, ulong n, GEN T, GEN p)
 {
+  pari_sp av = avma;
   struct _FpXQ D;
   D.T = FpX_get_red(T, p); D.p = p;
-  return gen_powu(x, n, (void*)&D, FpXQM_autsum_sqr, FpXQM_autsum_mul);
+  x = gen_powu_i(x, n, (void*)&D, FpXQM_autsum_sqr, FpXQM_autsum_mul);
+  return gerepilecopy(av, x);
 }
 
 static long
@@ -1983,12 +2198,10 @@ int
 FpXQ_issquare(GEN x, GEN T, GEN p)
 {
   pari_sp av;
-  long res;
   if (lg(x) == 2 || absequalui(2, p)) return 1;
   if (lg(x) == 3) return Fq_issquare(gel(x,2), T, p);
-  /* Ng = g^((q-1)/(p-1)) */
-  av = avma; res = kronecker(FpXQ_norm(x,T,p), p) == 1;
-  avma = av; return res;
+  av = avma; /* Ng = g^((q-1)/(p-1)) */
+  return gc_bool(av, kronecker(FpXQ_norm(x,T,p), p) == 1);
 }
 int
 Fp_issquare(GEN x, GEN p)
@@ -2017,7 +2230,7 @@ Fq_ispower(GEN x, GEN K, GEN T, GEN p)
   Q = subiu(powiu(p,d), 1);
   Q = diviiexact(Q, gcdii(Q, K));
   d = gequal1(Fq_pow(x, Q, T,p));
-  avma = av; return d;
+  return gc_long(av, d);
 }
 
 /* discrete log in FpXQ for a in Fp^*, g in FpXQ^* of order ord */
@@ -2068,7 +2281,7 @@ _FpXQ_rand(void *data)
   GEN z;
   do
   {
-    avma=av;
+    set_avma(av);
     z=random_FpX(get_FpX_degree(D->T),get_FpX_var(D->T),D->p);
   } while (!signe(z));
   return z;
@@ -2223,7 +2436,7 @@ FpXQ_trace(GEN x, GEN TB, GEN p)
   GEN dT = FpX_deriv(T,p);
   long n = degpol(dT);
   GEN z = FpXQ_mul(x, dT, TB, p);
-  if (degpol(z)<n) { avma = av; return gen_0; }
+  if (degpol(z)<n) return gc_const(av, gen_0);
   return gerepileuptoint(av, Fp_div(gel(z,2+n), gel(T,3+n),p));
 }
 
@@ -2260,7 +2473,7 @@ FpXQ_transmul_init(GEN tau, GEN T, GEN p)
     bht = FpXn_mul(bt, h, n-1, p);
   else
   {
-    GEN bh = FpX_div(RgX_shift_shallow(tau, n-1), T, p);
+    GEN bh = FpX_div(FpX_shift(tau, n-1), T, p);
     bht = FpX_recipspec(bh+2, lgpol(bh), n-1);
     setvarn(bht, vT);
   }
@@ -2274,11 +2487,11 @@ FpXQ_transmul(GEN tau, GEN a, long n, GEN p)
   GEN t1, t2, t3, vec;
   GEN bt = gel(tau, 1), bht = gel(tau, 2), ft = gel(tau, 3);
   if (signe(a)==0) return pol_0(varn(a));
-  t2 = RgX_shift_shallow(FpX_mul(bt, a, p),1-n);
+  t2 = FpX_shift(FpX_mul(bt, a, p),1-n);
   if (signe(bht)==0) return gerepilecopy(ltop, t2);
-  t1 = RgX_shift_shallow(FpX_mul(ft, a, p),-n);
+  t1 = FpX_shift(FpX_mul(ft, a, p),-n);
   t3 = FpXn_mul(t1, bht, n-1, p);
-  vec = FpX_sub(t2, RgX_shift_shallow(t3, 1), p);
+  vec = FpX_sub(t2, FpX_shift(t3, 1), p);
   return gerepileupto(ltop, vec);
 }
 
@@ -2361,7 +2574,7 @@ gener_FpXQ_i(GEN T, GEN p, GEN p_1, GEN Lp, GEN Lq)
   long vT = varn(T), f = degpol(T), l = lg(Lq);
   GEN F = FpX_Frobenius(T, p);
   int p_is_2 = is_pm1(p_1);
-  for (av = avma;; avma = av)
+  for (av = avma;; set_avma(av))
   {
     GEN t, g = random_FpX(f, vT, p);
     long i;
@@ -2486,40 +2699,11 @@ FpXn_sqr(GEN a, long n, GEN p)
   return FpX_red(ZXn_sqr(a, n), p);
 }
 
-GEN
-FpXn_exp(GEN h, long e, GEN p)
-{
-  pari_sp av = avma, av2;
-  long v = varn(h), n=1;
-  GEN f = pol_1(v), g = pol_1(v);
-  ulong mask = quadratic_prec_mask(e);
-  av2 = avma;
-  if (signe(h)==0 || degpol(h)<1 || !gequal0(gel(h,2)))
-    pari_err_DOMAIN("FpXn_exp","valuation", "<", gen_1, h);
-  for (;mask>1;)
-  {
-    GEN q, w;
-    long n2 = n;
-    n<<=1; if (mask & 1) n--;
-    mask >>= 1;
-    g = FpX_sub(FpX_mulu(g,2,p), FpXn_mul(f, FpXn_sqr(g, n2, p), n2, p), p);
-    q = FpX_deriv(FpXn_red(h,n2), p);
-    w = FpX_add(q, FpXn_mul(g, FpX_sub(FpX_deriv(f, p), FpXn_mul(f,q,n-1, p), p),n-1, p), p);
-    f = FpX_add(f, FpXn_mul(f, FpX_sub(FpXn_red(h, n), FpX_integ(w,p), p), n, p), p);
-    if (gc_needed(av2,2))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"FpXn_exp, e = %ld", n);
-      gerepileall(av2, 2, &f, &g);
-    }
-  }
-  return gerepileupto(av, f);
-}
-
 /* (f*g) \/ x^n */
 static GEN
 FpX_mulhigh_i(GEN f, GEN g, long n, GEN p)
 {
-  return RgX_shift_shallow(FpX_mul(f,g, p),-n);
+  return FpX_shift(FpX_mul(f,g, p),-n);
 }
 
 static GEN
@@ -2561,7 +2745,7 @@ FpXn_inv(GEN f, long e, GEN p)
     mask >>= 1;
     fr = FpXn_red(f, n);
     u = FpXn_mul(W, FpXn_mulhigh(fr, W, n2, n, p), n-n2, p);
-    W = FpX_sub(W, RgX_shift_shallow(u, n2), p);
+    W = FpX_sub(W, FpX_shift(u, n2), p);
     if (gc_needed(av2,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"FpXn_inv, e = %ld", n);
@@ -2570,3 +2754,41 @@ FpXn_inv(GEN f, long e, GEN p)
   }
   return gerepileupto(av, W);
 }
+
+GEN
+FpXn_expint(GEN h, long e, GEN p)
+{
+  pari_sp av = avma, av2;
+  long v = varn(h), n=1;
+  GEN f = pol_1(v), g = pol_1(v);
+  ulong mask = quadratic_prec_mask(e);
+  av2 = avma;
+  for (;mask>1;)
+  {
+    GEN u, w;
+    long n2 = n;
+    n<<=1; if (mask & 1) n--;
+    mask >>= 1;
+    u = FpXn_mul(g, FpX_mulhigh_i(f, FpXn_red(h, n2-1), n2-1, p), n-n2, p);
+    u = FpX_add(u, FpX_shift(FpXn_red(h, n-1), 1-n2), p);
+    w = FpXn_mul(f, FpX_integXn(u, n2-1, p), n-n2, p);
+    f = FpX_add(f, FpX_shift(w, n2), p);
+    if (mask<=1) break;
+    u = FpXn_mul(g, FpXn_mulhigh(f, g, n2, n, p), n-n2, p);
+    g = FpX_sub(g, FpX_shift(u, n2), p);
+    if (gc_needed(av2,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXn_exp, e = %ld", n);
+      gerepileall(av2, 2, &f, &g);
+    }
+  }
+  return gerepileupto(av, f);
+}
+
+GEN
+FpXn_exp(GEN h, long e, GEN p)
+{
+  if (signe(h)==0 || degpol(h)<1 || !gequal0(gel(h,2)))
+    pari_err_DOMAIN("FpXn_exp","valuation", "<", gen_1, h);
+  return FpXn_expint(FpX_deriv(h, p), e, p);
+}
