@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -14,7 +15,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
-
 /*******************************************************************/
 /**                                                               **/
 /**           Isomorphisms between finite fields                  **/
@@ -670,7 +670,7 @@ Flx_nbfactff(GEN P, GEN T, ulong p)
   pari_sp av = avma;
   GEN F = gel(Flx_degfact(P, p), 1);
   long s = simpleff_to_nbfact(F, get_Flx_degree(T));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 /* dummy implementation */
@@ -710,7 +710,7 @@ FpX_nbfactff(GEN P, GEN T, GEN p)
   pari_sp av = avma;
   GEN F = gel(FpX_degfact(P, p), 1);
   long s = simpleff_to_nbfact(F, get_FpX_degree(T));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 GEN
@@ -873,7 +873,7 @@ FpXQX_nbroots(GEN f, GEN T, GEN p)
 {
   pari_sp av = avma;
   GEN z = FpXQX_split_part(f, T, p);
-  avma = av; return degpol(z);
+  return gc_long(av, degpol(z));
 }
 
 long
@@ -885,7 +885,7 @@ FlxqX_nbroots(GEN f, GEN T, ulong p)
 {
   pari_sp av = avma;
   GEN z = FlxqX_split_part(f, T, p);
-  avma = av; return degpol(z);
+  return gc_long(av, degpol(z));
 }
 
 static GEN
@@ -956,9 +956,9 @@ FlxqX_split_Berlekamp(GEN *t, GEN xp, GEN T, ulong p)
         pari_sp av = avma;
         GEN S = FlxqX_get_red(a, T, p);
         b = FlxqX_rem(pol, S, T,p);
-        if (degpol(b)<=0) { avma=av; continue; }
+        if (degpol(b)<=0) { set_avma(av); continue; }
         b = FlxqXQ_halfFrobenius_i(b, xp, FlxqX_rem(Xp, S, T, p), S, T, p);
-        if (degpol(b)<=0) { avma=av; continue; }
+        if (degpol(b)<=0) { set_avma(av); continue; }
         gel(b,2) = Flxq_sub(gel(b,2), gen_1,T,p);
         b = FlxqX_gcd(a,b, T,p); lb = degpol(b);
         if (lb && lb < la)
@@ -967,14 +967,13 @@ FlxqX_split_Berlekamp(GEN *t, GEN xp, GEN T, ulong p)
           t[L] = FlxqX_div(a,b,T,p);
           t[i]= b; L++;
         }
-        else avma = av;
+        else set_avma(av);
       }
     }
   }
   return d;
 }
 
-
 static long
 FpXQX_split_Berlekamp(GEN *t, GEN T, GEN p)
 {
@@ -1010,9 +1009,9 @@ FpXQX_split_Berlekamp(GEN *t, GEN T, GEN p)
         pari_sp av = avma;
         GEN S = FpXQX_get_red(a, T, p);
         b = FqX_rem(pol, S, T,p);
-        if (degpol(b)<=0) { avma=av; continue; }
+        if (degpol(b)<=0) { set_avma(av); continue; }
         b = FpXQXQ_halfFrobenius_i(b, xp, FpXQX_rem(Xp, S, T, p), S, T, p);
-        if (degpol(b)<=0) { avma=av; continue; }
+        if (degpol(b)<=0) { set_avma(av); continue; }
         gel(b,2) = Fq_sub(gel(b,2), gen_1,T,p);
         b = FqX_gcd(a,b, T,p); lb = degpol(b);
         if (lb && lb < la)
@@ -1021,7 +1020,7 @@ FpXQX_split_Berlekamp(GEN *t, GEN T, GEN p)
           t[L] = FqX_div(a,b,T,p);
           t[i]= b; L++;
         }
-        else avma = av;
+        else set_avma(av);
       }
     }
   }
@@ -1161,15 +1160,16 @@ F2xqX_ispower(GEN f, long k, GEN T, GEN *pt_r)
 {
   pari_sp av = avma;
   GEN lc, F;
-  long i, l, n = degpol(f), v = varn(f);
+  long i, l, n = degpol(f);
   if (n % k) return 0;
   lc = F2xq_sqrtn(leading_coeff(f), stoi(k), T, NULL);
-  if (!lc) { av = avma; return 0; }
+  if (!lc) return gc_long(av,0);
   F = F2xqX_factor_squarefree(f, T); l = lg(F)-1;
   for(i=1; i<=l; i++)
-    if (i%k && degpol(gel(F,i))) { avma = av; return 0; }
+    if (i%k && degpol(gel(F,i))) return gc_long(av,0);
   if (pt_r)
   {
+    long v = varn(f);
     GEN r = scalarpol(lc, v), s = pol1_F2xX(v, T[1]);
     for(i=l; i>=1; i--)
     {
@@ -1178,7 +1178,7 @@ F2xqX_ispower(GEN f, long k, GEN T, GEN *pt_r)
       r = F2xqX_mul(r, s, T);
     }
     *pt_r = gerepileupto(av, r);
-  } else av = avma;
+  } else set_avma(av);
   return 1;
 }
 
@@ -1206,7 +1206,7 @@ F2xqX_roots_edf(GEN Sp, GEN xp, GEN Xp, GEN T, GEN V, long idx)
     GEN R = gel(F2xqXQ_auttrace(mkvec3(xp, Xp, a), dT, S, T), 3);
     f = F2xqX_gcd(R, Sp, T);
     if (degpol(f) > 0 && degpol(f) < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   f = gerepileupto(btop, F2xqX_normalize(f, T));
   ff = F2xqX_div(Sp, f, T);
@@ -1345,10 +1345,10 @@ FlxqX_ispower(GEN f, ulong k, GEN T, ulong p, GEN *pt_r)
   long i, l, n = degpol(f), v = varn(f);
   if (n % k) return 0;
   lc = Flxq_sqrtn(leading_coeff(f), stoi(k), T, p, NULL);
-  if (!lc) { av = avma; return 0; }
+  if (!lc) return gc_long(av,0);
   F = FlxqX_factor_squarefree_i(f, NULL, T, p); l = lg(F)-1;
   for(i=1; i<=l; i++)
-    if (i%k && degpol(gel(F,i))) { avma = av; return 0; }
+    if (i%k && degpol(gel(F,i))) return gc_long(av,0);
   if (pt_r)
   {
     GEN r = scalarpol(lc, v), s = pol1_FlxX(v, T[1]);
@@ -1359,7 +1359,7 @@ FlxqX_ispower(GEN f, ulong k, GEN T, ulong p, GEN *pt_r)
       r = FlxqX_mul(r, s, T, p);
     }
     *pt_r = gerepileupto(av, r);
-  } else av = avma;
+  } else set_avma(av);
   return 1;
 }
 
@@ -1379,7 +1379,7 @@ FlxqX_roots_split(GEN Sp, GEN xp, GEN Xp, GEN S, GEN T, ulong p)
     if (DEBUGLEVEL >= 7) timer_printf(&ti, "FlxqXQ_halfFrobenius");
     f = FlxqX_gcd(FlxX_Flx_sub(R, pol1_Flx(vT), p), Sp, T, p);
     if (degpol(f) > 0 && degpol(f) < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   return gerepileupto(btop, FlxqX_normalize(f, T, p));
 }
@@ -1511,17 +1511,16 @@ FpXQX_ispower(GEN f, ulong k, GEN T, GEN p, GEN *pt_r)
   {
     ulong pp = p[2];
     GEN fp = ZXX_to_FlxX(f, pp, varn(T));
-    if (!FlxqX_ispower(fp, k, ZX_to_Flx(T, pp), pp, pt_r))
-    { avma = av; return 0; }
+    if (!FlxqX_ispower(fp, k, ZX_to_Flx(T,pp), pp, pt_r)) return gc_long(av,0);
     if (pt_r) *pt_r = gerepileupto(av, FlxX_to_ZXX(*pt_r));
-    else avma = av;
+    else set_avma(av);
     return 1;
   }
   lc = FpXQ_sqrtn(leading_coeff(f), stoi(k), T, p, NULL);
-  if (!lc) { av = avma; return 0; }
+  if (!lc) return gc_long(av,0);
   F = FpXQX_factor_Yun(f, T, p); l = lg(F)-1;
   for(i=1; i <= l; i++)
-    if (i%k && degpol(gel(F,i))) { avma = av; return 0; }
+    if (i%k && degpol(gel(F,i))) return gc_long(av,0);
   if (pt_r)
   {
     GEN r = scalarpol(lc, v), s = pol_1(v);
@@ -1532,7 +1531,7 @@ FpXQX_ispower(GEN f, ulong k, GEN T, GEN p, GEN *pt_r)
       r = FpXQX_mul(r, s, T, p);
     }
     *pt_r = gerepileupto(av, r);
-  } else av = avma;
+  } else set_avma(av);
   return 1;
 }
 
@@ -1556,7 +1555,7 @@ FpXQX_roots_split(GEN Sp, GEN xp, GEN Xp, GEN S, GEN T, GEN p)
     if (DEBUGLEVEL >= 7) timer_printf(&ti, "FpXQXQ_halfFrobenius");
     f = FpXQX_gcd(FqX_Fq_sub(R, pol_1(vT), T, p), Sp, T, p);
     if (degpol(f) > 0 && degpol(f) < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   return gerepileupto(btop, FpXQX_normalize(f, T, p));
 }
@@ -1882,7 +1881,7 @@ F2xqX_edf_simple(GEN Sp, GEN xp, GEN Xp, GEN Sq, long d, GEN T, GEN V, long idx)
       g = F2xX_add(w0, F2xqXQ_sqr(g,S,T));
     f = F2xqX_gcd(g, Sp, T);
     if (degpol(f) > 0 && degpol(f) < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   f = F2xqX_normalize(f, T);
   ff = F2xqX_div(Sp, f , T);
@@ -2055,7 +2054,7 @@ FlxqX_ddf_degree(GEN S, GEN XP, GEN T, ulong p)
   {
     b = xq ? FlxqX_FlxqXQV_eval(b, xq, S, T, p)
            : FlxqXQ_pow(b, q, S, T, p);
-    if (gequal(b,X)) { avma = av; return i-1; }
+    if (gequal(b,X)) return gc_long(av,i-1);
     hash_insert_long(&h, b, i-1);
   }
   if (DEBUGLEVEL>=7) timer_printf(&ti,"FlxqX_ddf_degree: baby");
@@ -2065,9 +2064,9 @@ FlxqX_ddf_degree(GEN S, GEN XP, GEN T, ulong p)
   for(i = 2; i <= m+1; i++)
   {
     g = FlxqX_FlxqXQV_eval(g, xq, S, T, p);
-    if (hash_haskey_long(&h, g, &j)) { avma=av; return l*i-j; }
+    if (hash_haskey_long(&h, g, &j)) return gc_long(av, l*i-j);
   }
-  avma = av; return n;
+  return gc_long(av,n);
 }
 
 static GEN
@@ -2255,10 +2254,10 @@ FlxqX_edf_simple(GEN Sp, GEN xp, GEN Xp, GEN Xq, long d, GEN T, ulong p, GEN V,
       GEN R = FlxqXQ_halfFrobenius_i(FlxX_Flx_add(t, r, p), xp, Xp, S, T, p);
       f = FlxqX_gcd(FlxX_Flx_sub(R, pol1_Flx(vT), p), Sp, T, p);
       if (degpol(f) > 0 && degpol(f) < n) break;
-      avma = btop2;
+      set_avma(btop2);
     }
     if (degpol(f) > 0 && degpol(f) < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   f = FlxqX_normalize(f, T, p);
   ff = FlxqX_div(Sp, f , T, p);
@@ -2343,7 +2342,7 @@ FlxqX_nbfact_Frobenius(GEN S, GEN Xq, GEN T, ulong p)
     s = Flx_nbfactff(FlxX_to_Flx(u), T, p);
   else
     s = ddf_to_nbfact(FlxqX_ddf_Shoup(S, Xq, T, p));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 long
@@ -2356,7 +2355,7 @@ FlxqX_nbfact(GEN S, GEN T, ulong p)
     s = Flx_nbfactff(FlxX_to_Flx(u), T, p);
   else
     s = ddf_to_nbfact(FlxqX_ddf_Shoup(S, FlxqX_Frobenius(S, T, p), T, p));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 GEN
@@ -2407,7 +2406,7 @@ FpXQX_ddf_degree(GEN S, GEN XP, GEN T, GEN p)
   {
     b = xq ? FpXQX_FpXQXQV_eval(b, xq, S, T, p)
            : FpXQXQ_pow(b, q, S, T, p);
-    if (gequal(b,X)) { avma = av; return i-1; }
+    if (gequal(b,X)) return gc_long(av,i-1);
     hash_insert_long(&h, simplify_shallow(b), i-1);
   }
   if (DEBUGLEVEL>=7) timer_printf(&ti,"FpXQX_ddf_degree: baby");
@@ -2417,9 +2416,9 @@ FpXQX_ddf_degree(GEN S, GEN XP, GEN T, GEN p)
   for(i = 2; i <= m+1; i++)
   {
     g = FpXQX_FpXQXQV_eval(g, xq, S, T, p);
-    if (hash_haskey_long(&h, simplify_shallow(g), &j)) { avma=av; return l*i-j; }
+    if (hash_haskey_long(&h, simplify_shallow(g), &j)) return gc_long(av,l*i-j);
   }
-  avma = av; return n;
+  return gc_long(av,n);
 }
 
 static GEN
@@ -2638,10 +2637,10 @@ FpXQX_edf_simple(GEN Sp, GEN xp, GEN Xp, GEN Xq, long d, GEN T, GEN p, GEN V, lo
       GEN R = FpXQXQ_halfFrobenius_i(FqX_Fq_add(t, r, T, p), xp, Xp, S, T, p);
       f = FpXQX_gcd(FqX_Fq_add(R, gen_m1, T, p), Sp, T, p);
       if (degpol(f) > 0 && degpol(f) < n) break;
-      avma = btop2;
+      set_avma(btop2);
     }
     if (degpol(f) > 0 && degpol(f) < n) break;
-    avma = btop;
+    set_avma(btop);
   }
   f = FpXQX_normalize(f, T, p);
   ff = FpXQX_div(Sp, f , T, p);
@@ -2730,7 +2729,7 @@ FpXQX_nbfact_Frobenius(GEN S, GEN Xq, GEN T, GEN p)
     s = FpX_nbfactff(simplify_shallow(u), T, p);
   else
     s = ddf_to_nbfact(FpXQX_ddf_Shoup(S, Xq, T, p));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 long
@@ -2749,7 +2748,7 @@ FpXQX_nbfact(GEN S, GEN T, GEN p)
     s = FpX_nbfactff(simplify_shallow(u), T, p);
   else
     s = ddf_to_nbfact(FpXQX_ddf_Shoup(S, FpXQX_Frobenius(S, T, p), T, p));
-  avma = av; return s;
+  return gc_long(av,s);
 }
 long
 FqX_nbfact(GEN u, GEN T, GEN p)
@@ -2804,7 +2803,7 @@ FlxqX_is_squarefree(GEN P, GEN T, ulong p)
 {
   pari_sp av = avma;
   GEN z = FlxqX_gcd(P, FlxX_deriv(P, p), T, p);
-  avma = av; return degpol(z)==0;
+  return gc_long(av, degpol(z)==0);
 }
 
 long
@@ -2812,7 +2811,7 @@ FqX_is_squarefree(GEN P, GEN T, GEN p)
 {
   pari_sp av = avma;
   GEN z = FqX_gcd(P, FqX_deriv(P, T, p), T, p);
-  avma = av; return degpol(z)==0;
+  return gc_long(av, degpol(z)==0);
 }
 
 /* as RgX_to_FpXQ(FqX_to_FFX), leaving alone t_FFELT */
@@ -2837,7 +2836,7 @@ static GEN
 factmod_init(GEN f, GEN *pD, GEN *pT, GEN *pp)
 {
   const char *s = "factormod";
-  GEN T = NULL, p = NULL, D = *pD;
+  GEN T, p, D = *pD;
   if (typ(f) != t_POL) pari_err_TYPE(s,f);
   if (!D)
   {
@@ -2847,31 +2846,32 @@ factmod_init(GEN f, GEN *pD, GEN *pT, GEN *pp)
     if (t != t_INTMOD && t != code(t_POLMOD,t_INTMOD)) pari_err_TYPE(s,f);
     return RgX_to_FqX(f, *pT, *pp);
   }
-  switch(typ(D))
-  {
-    case t_INT: p = D; break;
-    case t_FFELT: { *pD = NULL; *pT = D; return RgX_to_FFX(f,D); }
-    case t_VEC:
-      if (lg(D) == 3)
-      {
-        p = gel(D,1);
-        T = gel(D,2);
-        if (typ(p) == t_INT) break;
-        if (typ(T) == t_INT) { swap(T,p); break; }
-      }
-    default: pari_err_TYPE(s,D);
-  }
-  if (signe(p) <= 0) pari_err_PRIME(s,p);
-  if (T)
-  {
-    if (typ(T) != t_POL) pari_err_TYPE(s,p);
-    T = RgX_to_FpX(T,p);
-    if (varncmp(varn(T), varn(f)) <= 0) pari_err_PRIORITY(s, T, "<=", varn(f));
-  }
+  if (typ(D) == t_FFELT) { *pD = NULL; *pT = D; return RgX_to_FFX(f,D); }
+  if (!ff_parse_Tp(D, &T, &p, 1)) pari_err_TYPE(s,D);
+  if (T && varncmp(varn(T), varn(f)) <= 0)
+    pari_err_PRIORITY(s, T, "<=", varn(f));
   *pT = T; *pp = p; return RgX_to_FqX(f, T, p);
 }
 #undef code
 
+int
+ff_parse_Tp(GEN Tp, GEN *T, GEN *p, long red)
+{
+  long t = typ(Tp);
+  *T = *p = NULL;
+  if (t == t_INT) { *p = Tp; return cmpiu(*p, 1) > 0; }
+  if (t != t_VEC || lg(Tp) != 3) return 0;
+  *T = gel(Tp,1);
+  *p = gel(Tp,2);
+  if (typ(*p) != t_INT)
+  {
+    if (typ(*T) != t_INT) return 0;
+    swap(*T, *p); /* support both [T,p] and [p,T] */
+  }
+  if (red) *T = RgX_to_FpX(*T, *p);
+  return cmpiu(*p, 1) > 0 && typ(*T) == t_POL && RgX_is_ZX(*T);
+}
+
 static GEN
 to_Fq(GEN x, GEN T, GEN p)
 {
@@ -2918,13 +2918,6 @@ to_FqC(GEN P, GEN T, GEN p)
   return P;
 }
 
-static GEN
-FpXC_to_mod(GEN x, GEN p)
-{ pari_APPLY_type(t_COL, FpX_to_mod(gel(x,i),p)) }
-static GEN
-FqXC_to_mod(GEN x, GEN T, GEN p)
-{ pari_APPLY_type(t_COL, FqX_to_mod(gel(x,i), T, p)) }
-
 GEN
 factmod(GEN f, GEN D)
 {
@@ -2949,7 +2942,7 @@ simplefactmod(GEN f, GEN D)
   pari_sp av = avma;
   GEN T, p;
   f = factmod_init(f, &D, &T,&p);
-  if (lg(f) <= 3) { avma = av; return trivial_fact(); }
+  if (lg(f) <= 3) { set_avma(av); return trivial_fact(); }
   f = D? FqX_degfact(f, T, p): FFX_degfact(f, T);
   return gerepileupto(av, Flm_to_ZM(f));
 }
@@ -2974,7 +2967,7 @@ factormodSQF(GEN f, GEN D)
   pari_sp av = avma;
   GEN F, T, p;
   f = factmod_init(f, &D, &T,&p);
-  if (lg(f) <= 3) { avma = av; return trivial_fact(); }
+  if (lg(f) <= 3) { set_avma(av); return trivial_fact(); }
   if (!D)
     F = sqf_to_fact(FFX_factor_squarefree(f, T));
   else
@@ -2990,7 +2983,7 @@ factormodDDF(GEN f, GEN D)
   pari_sp av = avma;
   GEN F, T, p;
   f = factmod_init(f, &D, &T,&p);
-  if (lg(f) <= 3) { avma = av; return trivial_fact(); }
+  if (lg(f) <= 3) { set_avma(av); return trivial_fact(); }
   if (!D) return FFX_ddf(f, T);
   F = FqX_ddf(f,T,p);
   gel(F,1) = FqXC_to_mod(gel(F,1), T,p);
