@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -108,7 +109,7 @@ ZpX_HenselLift(GEN V, GEN W, long j, GEN f, GEN pd, GEN p0, GEN p1, int noinv)
   t = FpX_red(t, pd);
   t = ZX_Z_mul(t,p0);
   s = ZX_Z_mul(s,p0);
-  avma = av;
+  set_avma(av);
   a2 = ZX_add(a,s);
   b2 = ZX_add(b,t);
 
@@ -129,7 +130,7 @@ ZpX_HenselLift(GEN V, GEN W, long j, GEN f, GEN pd, GEN p0, GEN p1, int noinv)
   t = FpX_red(t, pd);
   t = ZX_Z_mul(t,p0);
   s = ZX_Z_mul(s,p0);
-  avma = av;
+  set_avma(av);
   gel(W,j)   = ZX_add(u,t);
   gel(W,j+1) = ZX_add(v,s);
 }
@@ -155,7 +156,7 @@ ZpXQ_HenselLift(GEN V, GEN W, long j, GEN f, GEN Td, GEN T1, GEN pd, GEN p0, GEN
   t = FpXQX_red(t, Td, pd);
   t = RgX_Rg_mul(t,p0);
   s = RgX_Rg_mul(s,p0);
-  avma = av;
+  set_avma(av);
 
   a2 = RgX_add(a,s);
   b2 = RgX_add(b,t);
@@ -178,7 +179,7 @@ ZpXQ_HenselLift(GEN V, GEN W, long j, GEN f, GEN Td, GEN T1, GEN pd, GEN p0, GEN
   t = FpXQX_red(t, Td, pd);
   t = RgX_Rg_mul(t,p0);
   s = RgX_Rg_mul(s,p0);
-  avma = av;
+  set_avma(av);
   gel(W,j)   = RgX_add(u,t);
   gel(W,j+1) = RgX_add(v,s);
 }
@@ -341,7 +342,6 @@ GEN
 ZqX_liftroot(GEN f, GEN a, GEN T, GEN p, long e)
 { return T ? ZpXQX_liftroot(f, a,T , p, e): ZpX_liftroot(f, a, p, e); }
 
-
 /* U = NULL treated as 1 */
 static void
 BezoutPropagate(GEN link, GEN v, GEN w, GEN pe, GEN U, GEN f, long j)
@@ -392,28 +392,19 @@ bezout_lift_fact(GEN pol, GEN Q, GEN p, long e)
 /* Front-end for ZpX_liftfact:
    lift the factorization of pol mod p given by L to p^N (if possible) */
 GEN
-polhensellift(GEN pol, GEN L, GEN p, long N)
+polhensellift(GEN pol, GEN L, GEN Tp, long N)
 {
-  GEN T = NULL;
-  long i, l, t;
+  GEN T, p;
+  long i, l;
   pari_sp av = avma;
   void (*chk)(GEN, const char*);
 
   if (typ(pol) != t_POL) pari_err_TYPE("polhensellift",pol);
   RgX_check_ZXX(pol, "polhensellift");
   if (!is_vec_t(typ(L)) || lg(L) < 3) pari_err_TYPE("polhensellift",L);
-  t = typ(p);
-  if (t == t_VEC) /* [p, T] */
-  {
-    T = gel(p,2);
-    if (typ(T) != t_POL) pari_err_TYPE("polhensellift",pol);
-    RgX_check_ZX(T, "polhensellift");
-    p = gel(p,1); t = typ(p);
-  }
-  chk = T? RgX_check_ZXX: RgX_check_ZX;
-  if (t != t_INT) pari_err_TYPE("polhensellift",p);
   if (N < 1) pari_err_DOMAIN("polhensellift", "precision", "<", gen_1,stoi(N));
-
+  if (!ff_parse_Tp(Tp, &T, &p, 0)) pari_err_TYPE("polhensellift",Tp);
+  chk = T? RgX_check_ZXX: RgX_check_ZX;
   l = lg(L); L = leafcopy(L);
   for (i = 1; i < l; i++)
   {
@@ -428,7 +419,7 @@ static GEN
 FqV_roots_from_deg1(GEN x, GEN T, GEN p)
 {
   long i,l = lg(x);
-  GEN r = cgetg(l,t_VEC);
+  GEN r = cgetg(l,t_COL);
   for (i=1; i<l; i++) { GEN P = gel(x,i); gel(r,i) = Fq_neg(gel(P,2), T, p); }
   return r;
 }
@@ -575,6 +566,18 @@ ZpXQX_liftroot_vald(GEN f, GEN a, long v, GEN T, GEN p, long e)
 GEN
 ZpXQX_liftroot(GEN f, GEN a, GEN T, GEN p, long e) { return ZpXQX_liftroot_vald(f,a,0,T,p,e); }
 
+GEN
+ZpXQX_liftroots(GEN f, GEN S, GEN T, GEN p, long e)
+{
+  long i, n = lg(S)-1, d = degpol(f);
+  GEN r;
+  if (n == d) return ZpXQX_liftroots_full(f, S, T, powiu(p, e), p, e);
+  r = cgetg(n+1, typ(S));
+  for (i=1; i <= n; i++)
+    gel(r,i) = ZpXQX_liftroot(f, gel(S,i), T, p, e);
+  return r;
+}
+
 /* Same as ZpX_liftroot for the polynomial X^n-b*/
 GEN
 Zp_sqrtnlift(GEN b, GEN n, GEN a, GEN p, long e)
@@ -620,7 +623,6 @@ Zp_sqrtnlift(GEN b, GEN n, GEN a, GEN p, long e)
   return gerepileuptoint(ltop,a);
 }
 
-
 /* Same as ZpX_liftroot for the polynomial X^2-b */
 GEN
 Zp_sqrtlift(GEN b, GEN a, GEN p, long e)
@@ -695,7 +697,7 @@ ZpXQ_log(GEN a, GEN T, GEN p, long N)
       g = powuu(pp,2*i*k-w);
     }
     else g = powiu(p,2*i*k);
-    gel(pol,i+2) = Fp_div(g, utoi(z),pe);
+    gel(pol,i+2) = Fp_divu(g, z,pe);
   }
   if( DEBUGLEVEL>=3) timer_printf(&ti,"pol(%ld)",l);
   s = FpX_FpXQ_eval(pol, FpXQ_sqr(b, T, pe), T,  pe);
@@ -711,6 +713,29 @@ ZpXQ_log(GEN a, GEN T, GEN p, long N)
 /***********************************************************************/
 /* q = p^N */
 
+GEN
+gen_ZpM_Dixon(GEN F, GEN V, GEN q, GEN p, long N, void *E,
+                            GEN lin(void *E, GEN F, GEN d, GEN q),
+                            GEN invl(void *E, GEN d))
+{
+  pari_sp av = avma;
+  long N2, M;
+  GEN VN2, V2, VM, bil;
+  GEN q2, qM;
+  V = FpM_red(V, q);
+  if (N == 1) return invl(E, V);
+  N2 = (N + 1)>>1; M = N - N2;
+  F = FpM_red(F, q);
+  qM = powiu(p, M);
+  q2 = M == N2? qM: mulii(qM, p);
+  /* q2 = p^N2, qM = p^M, q = q2 * qM */
+  VN2 = gen_ZpM_Dixon(F, V, q2, p, N2, E, lin, invl);
+  bil = lin(E, F, VN2, q);
+  V2 = ZM_Z_divexact(ZM_sub(V, bil), q2);
+  VM = gen_ZpM_Dixon(F, V2, qM, p, M, E, lin, invl);
+  return gerepileupto(av, FpM_red(ZM_add(VN2, ZM_Z_mul(VM, q2)), q));
+}
+
 GEN
 gen_ZpX_Dixon(GEN F, GEN V, GEN q, GEN p, long N, void *E,
                             GEN lin(void *E, GEN F, GEN d, GEN q),
@@ -734,6 +759,66 @@ gen_ZpX_Dixon(GEN F, GEN V, GEN q, GEN p, long N, void *E,
   return gerepileupto(av, FpX_red(ZX_add(VN2, ZX_Z_mul(VM, q2)), q));
 }
 
+GEN
+gen_ZpM_Newton(GEN x, GEN p, long n, void *E,
+                      GEN eval(void *E, GEN f, GEN q),
+                      GEN invd(void *E, GEN V, GEN v, GEN q, long M))
+{
+  pari_sp ltop = avma, av;
+  long N = 1, N2, M;
+  long mask;
+  GEN q = p;
+  if (n == 1) return gcopy(x);
+  mask = quadratic_prec_mask(n);
+  av = avma;
+  while (mask > 1)
+  {
+    GEN qM, q2, v, V;
+    N2 = N; N <<= 1;
+    q2 = q;
+    if (mask&1UL) { /* can never happen when q2 = p */
+      N--; M = N2-1;
+      qM = diviiexact(q2,p); /* > 1 */
+      q = mulii(qM,q2);
+    } else {
+      M = N2;
+      qM = q2;
+      q = sqri(q2);
+    }
+    /* q2 = p^N2, qM = p^M, q = p^N = q2 * qM */
+    mask >>= 1;
+    v = eval(E, x, q);
+    V = ZM_Z_divexact(gel(v,1), q2);
+    x = FpM_sub(x, ZM_Z_mul(invd(E, V, v, qM, M), q2), q);
+    if (gc_needed(av, 1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"gen_ZpX_Newton");
+      gerepileall(av, 2, &x, &q);
+    }
+  }
+  return gerepileupto(ltop, x);
+}
+
+static GEN
+_ZpM_invd(void *E, GEN V, GEN v, GEN q, long M/*unused*/)
+{
+  (void)E; (void)M;
+  return FpM_mul(V, gel(v,2), q);
+}
+
+static GEN
+_ZpM_eval(void *E, GEN x, GEN q)
+{
+  GEN f = RgM_Rg_sub_shallow(FpM_mul(x, FpM_red((GEN) E, q), q), gen_1);
+  return mkvec2(f, x);
+}
+
+GEN
+ZpM_invlift(GEN M, GEN C, GEN p, long n)
+{
+  return gen_ZpM_Newton(C, p, n, (void *)M, _ZpM_eval, _ZpM_invd);
+}
+
 GEN
 gen_ZpX_Newton(GEN x, GEN p, long n, void *E,
                       GEN eval(void *E, GEN f, GEN q),
@@ -836,7 +921,7 @@ ZpXQX_divrem(GEN x, GEN Sp, GEN T, GEN q, GEN p, long e, GEN *pr)
   bi = ZpXQ_inv(b, T, p, e);
   S2 = FqX_Fq_mul_to_monic(S, bi, T, q);
   Q = FpXQX_divrem(x, S2, T, q, pr);
-  if (pr==ONLY_DIVIDES && !Q) { avma = av; return NULL; }
+  if (pr==ONLY_DIVIDES && !Q) { set_avma(av); return NULL; }
   if (pr==ONLY_REM || pr==ONLY_DIVIDES) return gerepileupto(av, Q);
   Q = FpXQX_FpXQ_mul(Q, bi, T, q);
   gerepileall(av, 2, &Q, pr);
@@ -916,7 +1001,7 @@ ZpXQ_sqrt(GEN a, GEN T, GEN p, long e)
 
 GEN
 ZpX_ZpXQ_liftroot_ea(GEN P, GEN S, GEN T, GEN p, long n, void *E,
-                     int early(void *E, GEN x, GEN q))
+                     GEN early(void *E, GEN x, GEN q))
 {
   pari_sp ltop = avma, av;
   long N, r;
@@ -944,8 +1029,13 @@ ZpX_ZpXQ_liftroot_ea(GEN P, GEN S, GEN T, GEN p, long n, void *E,
     Sq = FpX_sub(S, ZX_Z_mul(H, q2), q);
     if (DEBUGLEVEL > 3)
       timer_printf(&ti,"ZpX_ZpXQ_liftroot: reaching prec %ld",N);
-    if (mask==1 || (early && early(E, Sq, q)))
+    if (mask==1)
       return gerepileupto(ltop, Sq);
+    if (early)
+    {
+      GEN Se = early(E, Sq, q);
+      if (Se) return gerepileupto(ltop, Se);
+    }
     qq = sqri(q); N <<= 1;
     if (mask&1UL) { qq = diviiexact(qq, p); N--; }
     mask >>= 1;
@@ -986,3 +1076,148 @@ ZpXQM_prodFrobenius(GEN M, GEN T, GEN p, long e)
   GEN z = FpXQM_autsum(mkvec2(xp, M), get_FpX_degree(T), T, powiu(p,e));
   return gerepilecopy(av, gel(z,2));
 }
+
+/* Canonical lift of polynomial */
+
+static GEN _can_invl(void *E, GEN V) {(void) E; return V; }
+
+static GEN _can_lin(void *E, GEN F, GEN V, GEN q)
+{
+  GEN v = RgX_splitting(V, 3);
+  (void) E;
+  return FpX_sub(V,ZXV_dotproduct(v, F), q);
+}
+
+static GEN
+_can_iter(void *E, GEN f, GEN q)
+{
+  GEN h = RgX_splitting(f,3);
+  GEN h1s = ZX_sqr(gel(h,1)), h2s = ZX_sqr(gel(h,2)), h3s = ZX_sqr(gel(h,3));
+  GEN h12 = ZX_mul(gel(h,1), gel(h,2));
+  GEN h13 = ZX_mul(gel(h,1), gel(h,3));
+  GEN h23 = ZX_mul(gel(h,2), gel(h,3));
+  GEN h1c = ZX_mul(gel(h,1), h1s);
+  GEN h3c = ZX_mul(gel(h,3), h3s);
+  GEN th = ZX_mul(ZX_sub(h2s,ZX_mulu(h13,3)),gel(h,2));
+  GEN y = FpX_sub(f,ZX_add(RgX_shift_shallow(h3c,2),ZX_add(RgX_shift_shallow(th,1),h1c)),q);
+  (void) E;
+  return mkvecn(7,y,h1s,h2s,h3s,h12,h13,h23);
+}
+
+static GEN
+_can_invd(void *E, GEN V, GEN v, GEN qM, long M)
+{
+  GEN h1s=gel(v,2), h2s=gel(v,3), h3s=gel(v,4);
+  GEN h12=gel(v,5), h13=gel(v,6), h23=gel(v,7);
+  GEN F = mkvec3(ZX_sub(h1s,RgX_shift_shallow(h23,1)),RgX_shift_shallow(ZX_sub(h2s,h13),1),
+                 ZX_sub(RgX_shift_shallow(h3s,2),RgX_shift_shallow(h12,1)));
+  (void)E;
+  return gen_ZpX_Dixon(ZXV_Z_mul(F, utoi(3)), V, qM, utoi(3), M, NULL,
+                                                 _can_lin, _can_invl);
+}
+
+static GEN
+F3x_frobeniuslift(GEN P, long n)
+{ return gen_ZpX_Newton(Flx_to_ZX(P),utoi(3), n, NULL, _can_iter, _can_invd); }
+
+static GEN _can5_invl(void *E, GEN V) {(void) E; return V; }
+
+static GEN _can5_lin(void *E, GEN F, GEN V, GEN q)
+{
+  ulong p = *(ulong*)E;
+  GEN v = RgX_splitting(V, p);
+  return FpX_sub(V,ZXV_dotproduct(v, F), q);
+}
+
+/* P(X,t) -> P(X*t^n,t) mod (t^p-1) */
+static GEN
+_shift(GEN P, long n, ulong p, long v)
+{
+  long i, l=lg(P);
+  GEN r = cgetg(l,t_POL); r[1] = P[1];
+  for(i=2;i<l;i++)
+  {
+    long s = n*(i-2)%p;
+    GEN ci = gel(P,i);
+    if (typ(ci)==t_INT)
+      gel(r,i) = monomial(ci, s, v);
+    else
+      gel(r,i) = RgX_rotate_shallow(ci, s, p);
+  }
+  return FpXX_renormalize(r, l);
+}
+
+struct _can_mul
+{
+  GEN T, q;
+  ulong p;
+};
+
+static GEN
+_can5_mul(void *E, GEN A, GEN B)
+{
+  struct _can_mul *d = (struct _can_mul *)E;
+  GEN a = gel(A,1), b = gel(B,1);
+  long n = itos(gel(A,2));
+  GEN bn = _shift(b, n, d->p, get_FpX_var(d->T));
+  GEN c = FpXQX_mul(a, bn, d->T, d->q);
+  return mkvec2(c, addii(gel(A,2), gel(B,2)));
+}
+
+static GEN
+_can5_sqr(void *E, GEN A)
+{
+  return _can5_mul(E,A,A);
+}
+
+static GEN
+_can5_iter(void *E, GEN f, GEN q)
+{
+  pari_sp av = avma;
+  struct _can_mul D;
+  ulong p = *(ulong*)E;
+  long i, vT = fetch_var();
+  GEN N, P, d, V, fs;
+  D.q = q; D.T = ZX_Z_sub(pol_xn(p,vT),gen_1);
+  D.p = p;
+  fs = mkvec2(_shift(f, 1, p, vT), gen_1);
+  N = gel(gen_powu_i(fs,p-1,(void*)&D,_can5_sqr,_can5_mul),1);
+  N = ZXX_evalx0(FpXQX_red(N,polcyclo(p,vT),q));
+  P = FpX_mul(N,f,q);
+  P = RgX_deflate(P, p);
+  d = RgX_splitting(N, p);
+  V = cgetg(p+1,t_VEC);
+  gel(V,1) = ZX_mulu(gel(d,1), p);
+  for(i=2; i<= (long)p; i++)
+    gel(V,i) = ZX_mulu(RgX_shift_shallow(gel(d,p+2-i), 1), p);
+  (void)delete_var(); return gerepilecopy(av, mkvec2(ZX_sub(f,P),V));
+}
+
+static GEN
+_can5_invd(void *E, GEN H, GEN v, GEN qM, long M)
+{
+  ulong p = *(long*)E;
+  return gen_ZpX_Dixon(gel(v,2), H, qM, utoi(p), M, E, _can5_lin, _can5_invl);
+}
+
+GEN
+Flx_Teichmuller(GEN P, ulong p, long n)
+{
+  return p==3 ? F3x_frobeniuslift(P,n):
+         gen_ZpX_Newton(Flx_to_ZX(P),utoi(p), n, &p, _can5_iter, _can5_invd);
+}
+
+GEN
+polteichmuller(GEN P, ulong p, long n)
+{
+  pari_sp av = avma;
+  GEN q = NULL;
+  if (typ(P)!=t_POL || !RgX_is_FpX(P,&q))
+    pari_err_TYPE("polteichmuller",P);
+  if (q && !equaliu(q,p))
+    pari_err_MODULUS("polteichmuller",q,utoi(p));
+  if (n <= 0)
+    pari_err_DOMAIN("polteichmuller", "precision", "<=",gen_0,stoi(n));
+  return gerepileupto(av, p==2 ? F2x_Teichmuller(RgX_to_F2x(P), n)
+                               : Flx_Teichmuller(RgX_to_Flx(P, p), p, n));
+}
