@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -41,7 +42,7 @@ check_quaddisc_imag(GEN x, long *r, const char *f)
 }
 
 /* X^2 + b X + c is the canonical quadratic t_POL of discriminant D.
- * Dodd is non-zero iff D is odd */
+ * Dodd is nonzero iff D is odd */
 static void
 quadpoly_bc(GEN D, long Dodd, GEN *b, GEN *c)
 {
@@ -119,7 +120,7 @@ Qfb0(GEN x, GEN y, GEN z, GEN d, long prec)
   if (typ(z)!=t_INT) pari_err_TYPE("Qfb",z);
   D = qfb_disc3(x,y,z);
   check_quaddisc(D, &s, &r, "Qfb");
-  avma = av;
+  set_avma(av);
   if (s < 0) return qfi(x, y, z);
 
   d = d? gtofp(d,prec): real_0(prec);
@@ -521,11 +522,11 @@ nupow(GEN x, GEN n, GEN L)
   y = qfi_1_by_disc(D);
   if (!signe(n)) return y;
   if (!L) L = sqrtnint(absi_shallow(D), 4);
-  y = gen_pow(x, n, (void*)L, &mul_nudupl, &mul_nucomp);
+  y = gen_pow_i(x, n, (void*)L, &mul_nudupl, &mul_nucomp);
   if (signe(n) < 0
   && !absequalii(gel(y,1),gel(y,2))
   && !absequalii(gel(y,1),gel(y,3))) togglesign(gel(y,2));
-  return gerepileupto(av, y);
+  return gerepilecopy(av, y);
 }
 
 /* Reduction */
@@ -652,7 +653,7 @@ redimagsl2(GEN q, GEN *U)
     b = negi(b);
     z = u1; u1 = u2; u2 = negi(z);
   }
-  avma = av;
+  set_avma(av);
   a = icopy(a); gel(Q,1) = a;
   b = icopy(b); gel(Q,2) = b;
   c = icopy(c); gel(Q,3) = c;
@@ -665,7 +666,7 @@ redimagsl2(GEN q, GEN *U)
   v1 = subii(mulii(z, u1), mulii(a, u2)); v1 = diviiexact(v1, gel(q,3));
   z = subii(z, b);
   v2 = addii(mulii(z, u2), mulii(c, u1)); v2 = diviiexact(v2, gel(q,3));
-  avma = av;
+  set_avma(av);
   v1 = icopy(v1);
   v2 = icopy(v2);
   *U = mkmat2(mkcol2(u1,v1), mkcol2(u2,v2)); return Q;
@@ -698,7 +699,7 @@ redimag_1(pari_sp av, GEN a, GEN b, GEN c)
     if (uel(a,2) <= uel(c,2))
     { /* lg(b) <= 3 but may be too large for itos */
       long s = signe(b);
-      avma = av;
+      set_avma(av);
       if (!s) return redimag_1_b0(a[2], c[2]);
       if (a[2] == c[2]) s = 1;
       return setq(a[2], b[2], c[2], s);
@@ -706,7 +707,7 @@ redimag_1(pari_sp av, GEN a, GEN b, GEN c)
     swap(a,c); b = negi(b);
   }
   /* b != 0 */
-  avma = av;
+  set_avma(av);
   ua = a[2];
   ub = sb = b[2]; if (signe(b) < 0) sb = -sb;
   uc = c[2];
@@ -755,7 +756,7 @@ redimag_av(pari_sp av, GEN q)
   /* size of reduced Qfb(a,b,c) <= 3 lg(c) + 4 <= 4 lg(c) */
   (void)new_chunk(lc<<2);
   a = icopy(a); b = icopy(b); c = icopy(c);
-  avma = av;
+  set_avma(av);
   retmkqfi(icopy(a), icopy(b), icopy(c));
 }
 GEN
@@ -777,7 +778,7 @@ rhoimag(GEN x)
   x = cgetg(4, t_QFI);
   (void)new_chunk(lgefint(a) + lgefint(b) + lgefint(c) + 3);
   swap(a,c); b = negi(b);
-  REDB(a, &b, &c); avma = (pari_sp)x;
+  REDB(a, &b, &c); set_avma((pari_sp)x);
   gel(x,1) = icopy(a);
   gel(x,2) = icopy(b);
   gel(x,3) = icopy(c); return x;
@@ -994,7 +995,7 @@ qfr5_init(GEN x, struct qfr_data *S)
     pari_sp av=avma;
     long e;
     S->isqrtD = gcvtoi(S->sqrtD,&e);
-    if (e>-2) { avma = av; S->isqrtD = sqrti(S->D); }
+    if (e>-2) { set_avma(av); S->isqrtD = sqrti(S->D); }
   }
   else if (typ(S->isqrtD) != t_INT) pari_err_TYPE("qfr_init",S->isqrtD);
   return x;
@@ -1249,6 +1250,38 @@ primeform(GEN x, GEN p, long prec)
   return y;
 }
 
+static GEN
+normforms(GEN D, GEN fa)
+{
+  long i, j, k, lB, aN, sa;
+  GEN a, L, V, B, N, N2;
+  int D_odd = mpodd(D);
+  a = typ(fa) == t_INT ? fa: typ(fa) == t_VEC? gel(fa,1): factorback(fa);
+  sa = signe(a);
+  if (sa==0 || (signe(D)<0 && sa<0)) return NULL;
+  V = D_odd? Zn_quad_roots(fa, gen_1, shifti(subsi(1, D), -2))
+           : Zn_quad_roots(fa, gen_0, negi(shifti(D, -2)));
+  if (!V) return NULL;
+  N = gel(V,1); B = gel(V,2); lB = lg(B);
+  N2 = shifti(N,1);
+  aN = itou(diviiexact(a, N)); /* |a|/N */
+  L = cgetg((lB-1)*aN+1, t_VEC);
+  for (k = 1, i = 1; i < lB; i++)
+  {
+    GEN b = shifti(gel(B,i), 1), c, C;
+    if (D_odd) b = addiu(b , 1);
+    c = diviiexact(shifti(subii(sqri(b), D), -2), a);
+    for (j = 0;; b = addii(b, N2))
+    {
+      gel(L, k++) = mkvec3(a,b,c);
+      if (++j == aN) break;
+      C = addii(b, N); if (aN > 1) C = diviuexact(C, aN);
+      c = sa > 0? addii(c, C): subii(c, C);
+    }
+  }
+  return L;
+}
+
 /* Let M and N in SL2(Z), return (N*M^-1)[,1] */
 static GEN
 SL2_div_mul_e1(GEN N, GEN M)
@@ -1258,6 +1291,7 @@ SL2_div_mul_e1(GEN N, GEN M)
   GEN q = subii(mulii(gcoeff(N,2,1), d), mulii(gcoeff(N,2,2), b));
   return mkvec2(p,q);
 }
+
 /* Let M and N in SL2(Z), return (N*[1,0;0,-1]*M^-1)[,1] */
 static GEN
 SL2_swap_div_mul_e1(GEN N, GEN M)
@@ -1277,14 +1311,23 @@ GL2_qfb_equal(GEN a, GEN b)
    &&    equalii(gel(a,3),gel(b,3));
 }
 
+static GEN
+qfisolve_normform(GEN Q, GEN P)
+{
+  GEN a = gel(Q,1), N = gel(Q,2);
+  GEN M, b = redimagsl2(P, &M);
+  if (!GL2_qfb_equal(a,b) || signe(gel(a,2))!=signe(gel(b,2)))
+    return NULL;
+  return SL2_div_mul_e1(N,M);
+}
+
 static GEN
 qfbsolve_cornacchia(GEN c, GEN p, int swap)
 {
   pari_sp av = avma;
   GEN M, N;
-  if (kronecker(negi(c), p) < 0 || !cornacchia(c, p, &M,&N)) {
-    avma = av; return gen_0;
-  }
+  if (kronecker(negi(c), p) < 0 || !cornacchia(c, p, &M,&N))
+  { set_avma(av); return gen_0; }
   return gerepilecopy(av, swap? mkvec2(N,M): mkvec2(M,N));
 }
 
@@ -1308,20 +1351,20 @@ qfisolvep(GEN Q, GEN p)
     if (!signe(gel(a,2)))
     {
       a = qfbsolve_cornacchia(gel(a,3), p, 0);
-      if (a == gen_0) { avma = av; return gen_0; }
+      if (a == gen_0) { set_avma(av); return gen_0; }
       a = ZM_ZC_mul(N, a);
       a[0] = evaltyp(t_VEC) | _evallg(3); /* transpose */
       return gerepileupto(av, a);
     }
     /* x^2 + xy + ((1-d)/4)y^2 = p <==> (2x + y)^2 - d y^2 = 4p */
-    if (!cornacchia2(negi(d), p, &x, &y)) { avma = av; return gen_0; }
-    x = divis_rem(subii(x,y), 2, &r); if (r) { avma = av; return gen_0; }
+    if (!cornacchia2(negi(d), p, &x, &y)) { set_avma(av); return gen_0; }
+    x = divis_rem(subii(x,y), 2, &r); if (r) { set_avma(av); return gen_0; }
     a = ZM_ZC_mul(N, mkvec2(x,y));
     a[0] = evaltyp(t_VEC) | _evallg(3); /* transpose */
     return gerepileupto(av, a);
   }
   b = redimagsl2(primeform(d, p, 0), &M);
-  if (!GL2_qfb_equal(a,b)) { avma = av; return gen_0; }
+  if (!GL2_qfb_equal(a,b)) { set_avma(av); return gen_0; }
   if (signe(gel(a,2))==signe(gel(b,2)))
     x = SL2_div_mul_e1(N,M);
   else
@@ -1400,7 +1443,7 @@ qfbredsl2(GEN q, GEN S)
       else
       {
         D = qfb_disc(q);
-        isD = sqrtint(D);
+        isD = sqrti(D);
       }
       v = redrealsl2(q,D,isD);
       gel(v,1) = qfr3_to_qfr(gel(v,1), real_0(precision(gel(q,4))));
@@ -1408,54 +1451,129 @@ qfbredsl2(GEN q, GEN S)
 
     default:
         pari_err_TYPE("qfbredsl2",q);
-        return NULL;
+        return NULL;/*LCOV_EXCL_LINE*/
   }
 }
 
-GEN
-qfrsolvep(GEN Q, GEN p)
+static GEN
+qfrsolve_normform(GEN N, GEN P, GEN P2, GEN d, GEN rd)
 {
   pari_sp ltop = avma, btop;
-  GEN N, P, P1, P2, M, rd, d = qfb_disc(Q);
-  if (kronecker(d, p) < 0) { avma = ltop; return gen_0; }
-  rd = sqrti(d);
-  M = N = redrealsl2(Q, d,rd);
-  P = primeform(d, p, DEFAULTPREC);
-  P1 = redrealsl2(P, d,rd);
-  togglesign( gel(P,2) );
-  P2 = redrealsl2(P, d,rd);
+  GEN M = N, P1 = redrealsl2(P, d,rd);
   btop = avma;
   for(;;)
   {
     if (ZV_equal(gel(M,1), gel(P1,1))) { N = gel(P1,2); break; }
-    if (ZV_equal(gel(M,1), gel(P2,1))) { N = gel(P2,2); break; }
+    if (P2 && ZV_equal(gel(M,1), gel(P2,1))) { N = gel(P2,2); break; }
     M = redrealsl2step(M, d,rd);
-    if (ZV_equal(gel(M,1), gel(N,1))) { avma = ltop; return gen_0; }
+    if (ZV_equal(gel(M,1), gel(N,1))) { set_avma(ltop); return NULL; }
     if (gc_needed(btop, 1)) M = gerepileupto(btop, M);
   }
   return gerepilecopy(ltop, SL2_div_mul_e1(gel(M,2),N));
 }
 
 GEN
-qfbsolve(GEN Q,GEN n)
+qfrsolvep(GEN Q, GEN p)
 {
-  if (typ(n)!=t_INT) pari_err_TYPE("qfbsolve",n);
-  switch(typ(Q))
+  pari_sp ltop = avma;
+  GEN P, N, x, rd, d = qfb_disc(Q);
+  if (kronecker(d, p) < 0) { set_avma(ltop); return gen_0; }
+  P = primeform(d, p, DEFAULTPREC);
+  rd = sqrti(d);
+  N = redrealsl2(Q, d,rd);
+  x = qfrsolve_normform(N, P, invraw(P), d,rd);
+  if (!x) { set_avma(ltop); return gen_0; }
+  return gerepileupto(ltop, x);
+}
+
+static GEN
+redsl2(GEN Q, GEN d)
+{
+  GEN P, U;
+  if( signe(d)>0) return redrealsl2(Q, d, sqrti(d));
+  P = redimagsl2(Q, &U); return mkvec2(P,U);
+}
+
+static GEN
+qfsolve_normform(GEN Q, GEN f, GEN d, GEN rd)
+{ return rd? qfrsolve_normform(Q, f, NULL, d, rd): qfisolve_normform(Q, f); }
+static GEN
+qfbsolve_primitive_i(GEN Q, GEN d, GEN rd, GEN *Qr, GEN fa, long all)
+{
+  GEN x, W, F = normforms(d, fa);
+  long i, j, l;
+  if (!F) return NULL;
+  if (!*Qr) *Qr = redsl2(Q, d);
+  l = lg(F); W = all? cgetg(l, t_VEC): NULL;
+  for (j = i = 1; i < l; i++)
+    if ((x = qfsolve_normform(*Qr, gel(F,i), d, rd)))
+    {
+      if (!all) return x;
+      gel(W,j++) = x;
+    }
+  if (j == 1) return NULL;
+  setlg(W,j); return W;
+}
+
+static void
+qfb_initrd(GEN Q, GEN *pd, GEN *prd)
+{
+  *pd = qfb_disc(Q);
+  *prd = signe(*pd) > 0? sqrti(*pd): NULL;
+}
+static GEN
+qfbsolve_primitive(GEN Q, GEN fa, long all)
+{
+  pari_sp av = avma;
+  GEN Qr = NULL, dQ, rdQ, x;
+  qfb_initrd(Q, &dQ, &rdQ);
+  x = qfbsolve_primitive_i(Q, dQ, rdQ, &Qr, fa, all);
+  if (!x) { set_avma(av); return cgetg(1, t_VEC); }
+  return gerepilecopy(av, x);
+}
+
+/* f / g^2 */
+static GEN
+famat_divsqr(GEN f, GEN g)
+{ return famat_reduce(famat_div_shallow(f, famat_pows_shallow(g,2))); }
+static GEN
+qfbsolve_all(GEN Q, GEN n, long all)
+{
+  pari_sp av = avma;
+  GEN W, dQ, rdQ, Qr = NULL, fa = factorint(n, 0);
+  GEN D = divisors_factored(mkmat2(gel(fa,1), gshift(gel(fa,2),-1)));
+  long i, j, l = lg(D);
+  qfb_initrd(Q, &dQ, &rdQ);
+  W = all? cgetg(l, t_VEC): NULL;
+  for (i = j = 1; i < l; i++)
   {
-  case t_QFI: return qfisolvep(Q,n);
-  case t_QFR: return qfrsolvep(Q,n);
-  default:
-    pari_err_TYPE("qfbsolve",Q);
-    return NULL; /* LCOV_EXCL_LINE */
+    GEN w, d = gel(D,i), FA = i == 1? fa: famat_divsqr(fa, gel(d,2));
+    if ((w = qfbsolve_primitive_i(Q, dQ, rdQ, &Qr, FA, all)))
+    {
+      if (i != 1) w = RgV_Rg_mul(w, gel(d,1));
+      if (!all) return gerepilecopy(av, w);
+      gel(W,j++) = w;
+    }
   }
+  if (j == 1) { set_avma(av); return cgetg(1, t_VEC); }
+  setlg(W,j); return gerepilecopy(av, shallowconcat1(W));
+}
+
+GEN
+qfbsolve(GEN Q, GEN n, long fl)
+{
+  if (!is_qfb_t(typ(Q))) pari_err_TYPE("qfbsolve",Q);
+  if (fl < 0 || fl > 3) pari_err_FLAG("qfbsolve");
+  return (fl & 2)? qfbsolve_all(Q, n, fl & 1)
+                 : qfbsolve_primitive(Q, n, fl & 1);
 }
 
 /* 1 if there exists x,y such that x^2 + dy^2 = p [prime], 0 otherwise */
 long
 cornacchia(GEN d, GEN p, GEN *px, GEN *py)
 {
-  pari_sp av = avma, av2;
-  GEN a, b, c, L, r;
+  pari_sp av = avma;
+  GEN a, b, c, r;
 
   if (typ(d) != t_INT) pari_err_TYPE("cornacchia", d);
   if (typ(p) != t_INT) pari_err_TYPE("cornacchia", p);
@@ -1463,55 +1581,47 @@ cornacchia(GEN d, GEN p, GEN *px, GEN *py)
   *px = *py = gen_0;
   b = subii(p, d);
   if (signe(b) < 0) return 0;
-  if (signe(b) == 0) { avma = av; *py = gen_1; return 1; }
+  if (signe(b) == 0) { *py = gen_1; return gc_long(av,1); }
   b = Fp_sqrt(b, p); /* sqrt(-d) */
-  if (!b) { avma = av; return 0; }
-  if (abscmpii(shifti(b,1), p) > 0) b = subii(b,p);
-  a = p; L = sqrti(p);
-  av2 = avma;
-  while (abscmpii(b, L) > 0)
-  {
-    r = remii(a, b); a = b; b = r;
-    if (gc_needed(av2, 1)) {
-      if (DEBUGMEM>1) pari_warn(warnmem,"cornacchia");
-      gerepileall(av2, 2, &a,&b);
-    }
-  }
+  if (!b) return gc_long(av,0);
+  b = gmael(halfgcdii(p, b), 2, 2);
   a = subii(p, sqri(b));
   c = dvmdii(a, d, &r);
-  if (r != gen_0 || !Z_issquareall(c, &c)) { avma = av; return 0; }
-  avma = av;
+  if (r != gen_0 || !Z_issquareall(c, &c)) return gc_long(av,0);
+  set_avma(av);
   *px = icopy(b);
   *py = icopy(c); return 1;
 }
 
+static GEN
+lastqi(GEN Q)
+{
+  GEN s = gcoeff(Q,1,1), q = gcoeff(Q,1,2), p = absi_shallow(gcoeff(Q,2,2));
+  if (signe(q)==0) return gen_0;
+  if (signe(s)==0) return p;
+  if (is_pm1(q))   return subiu(p,1);
+  return divii(p, absi_shallow(q));
+}
+
 static long
 cornacchia2_helper(long av, GEN d, GEN p, GEN b, GEN px4, GEN *px, GEN *py)
 {
-  pari_sp av2 = avma;
-  GEN a, c, r, L;
-  long k = mod4(d);
+  GEN M, Q, V, a, c, r;
   if (!signe(b)) { /* d = p,2p,3p,4p */
-    avma = av;
+    set_avma(av);
     if (absequalii(d, px4)){ *py = gen_1; return 1; }
     if (absequalii(d, p))  { *py = gen_2; return 1; }
     return 0;
   }
-  if (mod2(b) != (k & 1)) b = subii(p,b);
-  a = shifti(p,1); L = sqrti(px4);
-  av2 = avma;
-  while (cmpii(b, L) > 0)
-  {
-    r = remii(a, b); a = b; b = r;
-    if (gc_needed(av2, 1)) {
-      if (DEBUGMEM>1) pari_warn(warnmem,"cornacchia");
-      gerepileall(av2, 2, &a,&b);
-    }
-  }
+  if (mod2(b) != mod2(d)) b = subii(p,b);
+  M = halfgcdii(shifti(p,1), b); Q = gel(M,1); V = gel(M, 2);
+  b = addii(mulii(gel(V,1), lastqi(Q)),gel(V,2));
+  if (cmpii(sqri(b),px4) > 0) b = gel(V,1);
+  if (cmpii(sqri(b),px4) > 0) b = gel(V,2);
   a = subii(px4, sqri(b));
   c = dvmdii(a, d, &r);
-  if (r != gen_0 || !Z_issquareall(c, &c)) { avma = av; return 0; }
-  avma = av;
+  if (r != gen_0 || !Z_issquareall(c, &c)) return gc_long(av,0);
+  set_avma(av);
   *px = icopy(b);
   *py = icopy(c); return 1;
 }
@@ -1531,10 +1641,10 @@ cornacchia2(GEN d, GEN p, GEN *px, GEN *py)
   k = mod4(d);
   if (k == 1 || k == 2) pari_err_DOMAIN("cornacchia2","-d mod 4", ">",gen_1,d);
   px4 = shifti(p,2);
-  if (abscmpii(px4, d) < 0) { avma = av; return 0; }
+  if (abscmpii(px4, d) < 0) return gc_long(av,0);
   if (absequaliu(p, 2))
   {
-    avma = av;
+    set_avma(av);
     switch (itou_or_0(d)) {
       case 4: *px = gen_2; break;
       case 7: *px = gen_1; break;
@@ -1543,7 +1653,7 @@ cornacchia2(GEN d, GEN p, GEN *px, GEN *py)
     *py = gen_1; return 1;
   }
   b = Fp_sqrt(negi(d), p);
-  if (!b) { avma = av; return 0; }
+  if (!b) return gc_long(av,0);
   return cornacchia2_helper(av, d, p, b, px4, px, py);
 }
 
@@ -1555,6 +1665,6 @@ cornacchia2_sqrt(GEN d, GEN p, GEN b, GEN *px, GEN *py)
   GEN px4;
   *px = *py = gen_0;
   px4 = shifti(p,2);
-  if (abscmpii(px4, d) < 0) { avma = av; return 0; }
+  if (abscmpii(px4, d) < 0) return gc_long(av,0);
   return cornacchia2_helper(av, d, p, b, px4, px, py);
 }
