@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -44,7 +45,7 @@ ZX_add(GEN x, GEN y)
   for (i=2; i<ly; i++) gel(z,i) = addii(gel(x,i),gel(y,i));
   for (   ; i<lx; i++) gel(z,i) = icopy(gel(x,i));
   if (lx == ly) z = ZX_renormalize(z, lx);
-  if (!lgpol(z)) { avma = (pari_sp)(z + lx); return pol_0(varn(x)); }
+  if (!lgpol(z)) { set_avma((pari_sp)(z + lx)); return pol_0(varn(x)); }
   return z;
 }
 
@@ -60,7 +61,7 @@ ZX_sub(GEN x,GEN y)
     if (lx == ly)
     {
       z = ZX_renormalize(z, lx);
-      if (!lgpol(z)) { avma = (pari_sp)(z + lx); z = pol_0(varn(x)); }
+      if (!lgpol(z)) { set_avma((pari_sp)(z + lx)); z = pol_0(varn(x)); }
     }
     else
       for (   ; i<lx; i++) gel(z,i) = icopy(gel(x,i));
@@ -121,7 +122,7 @@ ZX_Z_add(GEN y, GEN x)
 {
   long lz, i;
   GEN z = cgetg_copy(y, &lz);
-  if (lz == 2) { avma = (pari_sp)(z + 2); return scalar_ZX(x,varn(y)); }
+  if (lz == 2) { set_avma((pari_sp)(z + 2)); return scalar_ZX(x,varn(y)); }
   z[1] = y[1];
   gel(z,2) = addii(gel(y,2),x);
   for(i=3; i<lz; i++) gel(z,i) = icopy(gel(y,i));
@@ -133,7 +134,7 @@ ZX_Z_add_shallow(GEN y, GEN x)
 {
   long lz, i;
   GEN z = cgetg_copy(y, &lz);
-  if (lz == 2) { avma = (pari_sp)(z + 2); return scalar_ZX_shallow(x,varn(y)); }
+  if (lz == 2) { set_avma((pari_sp)(z + 2)); return scalar_ZX_shallow(x,varn(y)); }
   z[1] = y[1];
   gel(z,2) = addii(gel(y,2),x);
   for(i=3; i<lz; i++) gel(z,i) = gel(y,i);
@@ -149,7 +150,7 @@ ZX_Z_sub(GEN y, GEN x)
   if (lz == 2)
   { /* scalarpol(negi(x), v) */
     long v = varn(y);
-    avma = (pari_sp)(z + 2);
+    set_avma((pari_sp)(z + 2));
     if (!signe(x)) return pol_0(v);
     z = cgetg(3,t_POL);
     z[1] = evalvarn(v) | evalsigne(1);
@@ -167,7 +168,7 @@ Z_ZX_sub(GEN x, GEN y)
 {
   long lz, i;
   GEN z = cgetg_copy(y, &lz);
-  if (lz == 2) { avma = (pari_sp)(z + 2); return scalar_ZX(x,varn(y)); }
+  if (lz == 2) { set_avma((pari_sp)(z + 2)); return scalar_ZX(x,varn(y)); }
   z[1] = y[1];
   gel(z,2) = subii(x, gel(y,2));
   for(i=3; i<lz; i++) gel(z,i) = negi(gel(y,i));
@@ -184,6 +185,24 @@ ZX_Z_divexact(GEN y,GEN x)
   return z;
 }
 
+GEN
+ZX_divuexact(GEN y, ulong x)
+{
+  long i, l = lg(y);
+  GEN z = cgetg(l,t_POL); z[1] = y[1];
+  for(i=2; i<l; i++) gel(z,i) = diviuexact(gel(y,i),x);
+  return z;
+}
+
+GEN
+zx_z_divexact(GEN y, long x)
+{
+  long i, l = lg(y);
+  GEN z = cgetg(l,t_VECSMALL); z[1] = y[1];
+  for (i=2; i<l; i++) z[i] = y[i]/x;
+  return z;
+}
+
 GEN
 ZX_Z_mul(GEN y,GEN x)
 {
@@ -317,13 +336,124 @@ ZX_div_by_X_1(GEN a, GEN *r)
   return z;
 }
 
-/* Return 2^(n degpol(P))  P(x >> n), not memory clean if P is a ZX */
+/* return P(X + c) using destructive Horner, optimize for c = 1,-1 */
+static GEN
+ZX_translate_basecase(GEN P, GEN c)
+{
+  pari_sp av = avma;
+  GEN Q, R;
+  long i, k, n;
+
+  if (!signe(P) || !signe(c)) return ZX_copy(P);
+  Q = leafcopy(P);
+  R = Q+2; n = degpol(P);
+  if (equali1(c))
+  {
+    for (i=1; i<=n; i++)
+    {
+      for (k=n-i; k<n; k++) gel(R,k) = addii(gel(R,k), gel(R,k+1));
+      if (gc_needed(av,2))
+      {
+        if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate(1), i = %ld/%ld", i,n);
+        Q = gerepilecopy(av, Q); R = Q+2;
+      }
+    }
+  }
+  else if (equalim1(c))
+  {
+    for (i=1; i<=n; i++)
+    {
+      for (k=n-i; k<n; k++) gel(R,k) = subii(gel(R,k), gel(R,k+1));
+      if (gc_needed(av,2))
+      {
+        if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate(-1), i = %ld/%ld", i,n);
+        Q = gerepilecopy(av, Q); R = Q+2;
+      }
+    }
+  }
+  else
+  {
+    for (i=1; i<=n; i++)
+    {
+      for (k=n-i; k<n; k++) gel(R,k) = addmulii_inplace(gel(R,k), c, gel(R,k+1));
+      if (gc_needed(av,2))
+      {
+        if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate, i = %ld/%ld", i,n);
+        Q = gerepilecopy(av, Q); R = Q+2;
+      }
+    }
+  }
+  return gerepilecopy(av, Q);
+}
+
+static GEN
+Z_Xpm1_powu(long n, long s, long v)
+{
+  long d, k;
+  GEN C;
+  if (!n) return pol_1(v);
+  d = (n + 1) >> 1;
+  C = cgetg(n+3, t_POL);
+  C[1] = evalsigne(1)| evalvarn(v);
+  gel(C,2) = gen_1;
+  gel(C,3) = utoipos(n);
+  for (k=2; k <= d; k++)
+    gel(C,k+2) = diviuexact(mului(n-k+1, gel(C,k+1)), k);
+  if (s < 0)
+    for (k = odd(n)? 0: 1; k <= d; k += 2)
+      togglesign_safe(&gel(C,k+2));
+  if (s > 0 || !odd(n))
+    for (k = d+1; k <= n; k++) gel(C,k+2) = gel(C,n-k+2);
+  else
+    for (k = d+1; k <= n; k++) gel(C,k+2) = negi(gel(C,n-k+2));
+  return C;
+}
+/* return (x+u)^n */
+static GEN
+Z_XpN_powu(GEN u, long n, long v)
+{
+  pari_sp av;
+  long k;
+  GEN B, C, V;
+  if (!n) return pol_1(v);
+  if (is_pm1(u))
+    return Z_Xpm1_powu(n, signe(u), v);
+  av = avma;
+  V = gpowers(u, n);
+  B = vecbinomial(n);
+  C = cgetg(n+3, t_POL);
+  C[1] = evalsigne(1)| evalvarn(v);
+  for (k=1; k <= n+1; k++)
+    gel(C,k+1) = mulii(gel(V,n+2-k), gel(B,k));
+  return gerepileupto(av, C);
+}
+
+GEN
+ZX_translate(GEN P, GEN c)
+{
+  pari_sp av = avma;
+  long n = degpol(P);
+  if (n < 220)
+    return ZX_translate_basecase(P, c);
+  else
+  {
+    long d = n >> 1;
+    GEN Q = ZX_translate(RgX_shift_shallow(P, -d), c);
+    GEN R = ZX_translate(RgXn_red_shallow(P, d), c);
+    GEN S = Z_XpN_powu(c, d, varn(P));
+    return gerepileupto(av, ZX_add(ZX_mul(Q, S), R));
+  }
+}
+
+/* Return 2^(n degpol(P))  P(x >> n) */
 GEN
 ZX_rescale2n(GEN P, long n)
 {
   long i, l = lg(P), ni = n;
-  GEN Q = cgetg(l,t_POL);
-  Q[l-1] = P[l-1];
+  GEN Q;
+  if (l==2) return pol_0(varn(P));
+  Q = cgetg(l,t_POL);
+  gel(Q,l-1) = icopy(gel(P,l-1));
   for (i=l-2; i>=2; i--)
   {
     gel(Q,i) = shifti(gel(P,i), ni);
@@ -366,7 +496,6 @@ ZX_rescale_lt(GEN P)
   Q[1] = P[1]; return Q;
 }
 
-
 /*Eval x in 2^(k*BIL) in linear time*/
 static GEN
 ZX_eval2BILspec(GEN x, long k, long nx)
@@ -605,7 +734,7 @@ ZX_rem(GEN x, GEN y)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     if (signe(p1)) { sx = 1; break; }
     if (!i) break;
-    avma=av;
+    set_avma(av);
   }
   lr=i+3; rem -= lr;
   rem[0] = evaltyp(t_POL) | evallg(lr);
@@ -938,6 +1067,19 @@ ZXX_renormalize(GEN x, long lx)
   setlg(x, i+1); setsigne(x, i!=1); return x;
 }
 
+GEN
+ZXX_evalx0(GEN y)
+{
+  long i, l = lg(y);
+  GEN z = cgetg(l,t_POL); z[1] = y[1];
+  for(i=2; i<l; i++)
+  {
+    GEN yi = gel(y,i);
+    gel(z,i) = typ(yi)==t_INT? yi: constant_coeff(yi);
+  }
+  return ZX_renormalize(z,l);
+}
+
 long
 ZXX_max_lg(GEN x)
 {
@@ -1023,12 +1165,65 @@ ZXX_to_Kronecker_spec(GEN P, long lP, long n)
   y-=2; setlg(y, k+2); y[1] = evalsigne(1); return y;
 }
 
+/* shallow, n = deg(T) */
+GEN
+Kronecker_to_ZXX(GEN z, long n, long v)
+{
+  long i,j,lx,l, N = (n<<1)+1;
+  GEN x, t;
+  l = lg(z); lx = (l-2) / (N-2);
+  x = cgetg(lx+3,t_POL);
+  x[1] = z[1];
+  for (i=2; i<lx+2; i++)
+  {
+    t = cgetg(N,t_POL); t[1] = evalvarn(v);
+    for (j=2; j<N; j++) gel(t,j) = gel(z,j);
+    z += (N-2);
+    gel(x,i) = ZX_renormalize(t,N);
+  }
+  N = (l-2) % (N-2) + 2;
+  t = cgetg(N,t_POL); t[1] = evalvarn(v);
+  for (j=2; j<N; j++) gel(t,j) = gel(z,j);
+  gel(x,i) = ZX_renormalize(t,N);
+  return ZXX_renormalize(x, i+1);
+}
+
 GEN
 ZXX_to_Kronecker(GEN P, long n)
 {
   GEN z = ZXX_to_Kronecker_spec(P+2, lgpol(P), n);
   setvarn(z,varn(P)); return z;
 }
+GEN
+ZXX_mul_Kronecker(GEN x, GEN y, long n)
+{ return ZX_mul(ZXX_to_Kronecker(x,n), ZXX_to_Kronecker(y,n)); }
+
+GEN
+ZXX_sqr_Kronecker(GEN x, long n)
+{ return ZX_sqr(ZXX_to_Kronecker(x,n)); }
+
+/* shallow, n = deg(T) */
+GEN
+Kronecker_to_ZXQX(GEN z, GEN T)
+{
+  long i,j,lx,l, N = (degpol(T)<<1)+1;
+  GEN x, t;
+  l = lg(z); lx = (l-2) / (N-2);
+  x = cgetg(lx+3,t_POL);
+  x[1] = z[1];
+  for (i=2; i<lx+2; i++)
+  {
+    t = cgetg(N,t_POL); t[1] = T[1];
+    for (j=2; j<N; j++) gel(t,j) = gel(z,j);
+    z += (N-2);
+    gel(x,i) = ZX_rem(ZX_renormalize(t,N), T);
+  }
+  N = (l-2) % (N-2) + 2;
+  t = cgetg(N,t_POL); t[1] = T[1];
+  for (j=2; j<N; j++) gel(t,j) = gel(z,j);
+  gel(x,i) = ZX_rem(ZX_renormalize(t,N), T);
+  return ZXX_renormalize(x, i+1);
+}
 
 GEN
 ZXQX_sqr(GEN x, GEN T)
@@ -1036,8 +1231,8 @@ ZXQX_sqr(GEN x, GEN T)
   pari_sp av = avma;
   long n = degpol(T);
   GEN z = ZXX_sqr_Kronecker(x, n);
-  z = Kronecker_to_ZXX(z, n, varn(T));
-  return gerepileupto(av, z);
+  z = Kronecker_to_ZXQX(z, T);
+  return gerepilecopy(av, z);
 }
 
 GEN
@@ -1046,8 +1241,20 @@ ZXQX_mul(GEN x, GEN y, GEN T)
   pari_sp av = avma;
   long n = degpol(T);
   GEN z = ZXX_mul_Kronecker(x, y, n);
-  z = Kronecker_to_ZXX(z, n, varn(T));
-  return gerepileupto(av, z);
+  z = Kronecker_to_ZXQX(z, T);
+  return gerepilecopy(av, z);
+}
+
+GEN
+ZXQX_ZXQ_mul(GEN P, GEN U, GEN T)
+{
+  long i, lP;
+  GEN res;
+  res = cgetg_copy(P, &lP); res[1] = P[1];
+  for(i=2; i<lP; i++)
+    gel(res,i) = typ(gel(P,i))==t_POL? ZXQ_mul(U, gel(P,i), T)
+                                     : gmul(U, gel(P,i));
+  return ZXX_renormalize(res,lP);
 }
 
 GEN
@@ -1084,3 +1291,52 @@ QX_ZX_rem(GEN x, GEN y)
   if (d) r = ZX_Q_mul(r, d);
   return gerepileupto(av, r);
 }
+
+GEN
+QXQX_mul(GEN x, GEN y, GEN T)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN dy, ny = Q_primitive_part(y, &dy);
+  GEN z = ZXQX_mul(nx, ny, T);
+  if (dx || dy)
+  {
+    GEN d = dx ? dy ? gmul(dx, dy): dx : dy;
+    return ZXX_Q_mul(z, d);
+  } else
+    return z;
+}
+
+GEN
+QXQX_sqr(GEN x, GEN T)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN z = ZXQX_sqr(nx, T);
+  if (dx)
+    return ZXX_Q_mul(z, gsqr(dx));
+  else
+    return z;
+}
+
+GEN
+QXQX_powers(GEN P, long n, GEN T)
+{
+  GEN v = cgetg(n+2, t_VEC);
+  long i;
+  gel(v, 1) = pol_1(varn(T));
+  if (n==0) return v;
+  gel(v, 2) = gcopy(P);
+  for (i = 2; i <= n; i++) gel(v,i+1) = QXQX_mul(P, gel(v,i), T);
+  return v;
+}
+
+GEN
+QXQX_QXQ_mul(GEN P, GEN U, GEN T)
+{
+  long i, lP;
+  GEN res;
+  res = cgetg_copy(P, &lP); res[1] = P[1];
+  for(i=2; i<lP; i++)
+    gel(res,i) = typ(gel(P,i))==t_POL? QXQ_mul(U, gel(P,i), T)
+                                     : gmul(U, gel(P,i));
+  return ZXX_renormalize(res,lP);
+}
