@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -35,10 +36,11 @@ typedef struct Red {
 #define cache_cyc(C)  (gel((C),3))
 #define cache_E(C)  (gel((C),4))
 #define cache_eta(C)  (gel((C),5))
-#define cache_matvite(C)  (gel((C),6))
-#define cache_matinvvite(C)  (gel((C),7))
-#define cache_avite(C)  (gel((C),8))
-#define cache_pkvite(C)  (gel((C),9))
+/* the last 4 only when N = 1 mod p^k */
+#define cache_mat(C)  (gel((C),6))
+#define cache_matinv(C)  (gel((C),7))
+#define cache_a(C)  (gel((C),8)) /* a has order p^k in (Z/NZ)^* */
+#define cache_pk(C)  (gel((C),9)) /* p^k */
 
 static GEN
 makepoldeg1(GEN c, GEN d)
@@ -104,53 +106,33 @@ red_cyclo2n_ip(GEN x, long n)
 static GEN
 red_cyclo2n(GEN x, long n) { return red_cyclo2n_ip(leafcopy(x), n); }
 
-/* x a non-zero VECSMALL */
-static GEN
-smallpolrev(GEN x)
-{
-  long i,j, lx = lg(x);
-  GEN y;
-
-  while (lx-- && x[lx]==0) /* empty */;
-  i = lx+2; y = cgetg(i,t_POL);
-  y[1] = evalsigne(1);
-  for (j=2; j<i; j++) gel(y,j) = stoi(x[j-1]);
-  return y;
-}
-
-/* x polynomial in t_VECSMALL form, T t_POL return x mod T */
-static GEN
-u_red(GEN x, GEN T) {
-  return RgX_rem(smallpolrev(x), T);
-}
-
 /* special case R->C = polcyclo(2^n) */
 static GEN
-_red_cyclo2n(GEN x, Red *R) {
-  return centermod_i(red_cyclo2n(x, R->n), R->N, R->N2);
-}
+_red_cyclo2n(GEN x, Red *R)
+{ return centermod_i(red_cyclo2n(x, R->n), R->N, R->N2); }
 /* special case R->C = polcyclo(p) */
 static GEN
-_red_cyclop(GEN x, Red *R) {
-  return centermod_i(red_cyclop(x, R->n), R->N, R->N2);
-}
+_red_cyclop(GEN x, Red *R)
+{ return centermod_i(red_cyclop(x, R->n), R->N, R->N2); }
 static GEN
-_red(GEN x, Red *R) {
-  return centermod_i(grem(x, R->C), R->N, R->N2);
-}
+_red(GEN x, Red *R)
+{ return centermod_i(ZX_rem(x, R->C), R->N, R->N2); }
 static GEN
-_redsimple(GEN x, Red *R) { return centermodii(x, R->N, R->N2); }
-
+modZ(GEN x, Red *R) { return centermodii(x, R->N, R->N2); }
 static GEN
-sqrmod(GEN x, Red *R) {
-  return R->red(gsqr(x), R);
-}
+sqrmodZ(GEN x, Red *R) { return modZ(sqri(x), R); }
+static GEN
+sqrmod(GEN x, Red *R) { return R->red(ZX_sqr(x), R); }
+static GEN
+_mul(GEN x, GEN y, Red *R)
+{ return typ(x) == t_INT? modZ(mulii(x,y), R)
+                        : R->red(ZX_mul(x,y), R); }
 
 static GEN
 sqrconst(GEN pol, Red *R)
 {
   GEN z = cgetg(3,t_POL);
-  gel(z,2) = centermodii(sqri(gel(pol,2)), R->N, R->N2);
+  gel(z,2) = modZ(sqri(gel(pol,2)), R);
   z[1] = pol[1]; return z;
 }
 
@@ -165,9 +147,8 @@ sqrmod3(GEN pol, Red *R)
   if (lv==3) return sqrconst(pol, R);
   a = gel(pol,3);
   b = gel(pol,2); bma = subii(b,a);
-  A = centermodii(mulii(a,addii(b,bma)), R->N, R->N2);
-  B = centermodii(mulii(bma,addii(a,b)), R->N, R->N2);
-  return makepoldeg1(A,B);
+  A = modZ(mulii(a,addii(b,bma)), R);
+  B = modZ(mulii(bma,addii(a,b)), R); return makepoldeg1(A,B);
 }
 
 /* pol^2 mod (x^2+1,N) */
@@ -181,9 +162,8 @@ sqrmod4(GEN pol, Red *R)
   if (lv==3) return sqrconst(pol, R);
   a = gel(pol,3);
   b = gel(pol,2);
-  A = centermodii(mulii(a, shifti(b,1)), R->N, R->N2);
-  B = centermodii(mulii(subii(b,a),addii(b,a)), R->N, R->N2);
-  return makepoldeg1(A,B);
+  A = modZ(mulii(a, shifti(b,1)), R);
+  B = modZ(mulii(subii(b,a),addii(b,a)), R); return makepoldeg1(A,B);
 }
 
 /* pol^2 mod (polcyclo(5),N) */
@@ -199,12 +179,9 @@ sqrmod5(GEN pol, Red *R)
   d = gel(pol,2);
   if (lv==4)
   {
-    A = sqri(d);
-    B = mulii(c2, d);
-    C = sqri(c);
-    A = centermodii(A, R->N, R->N2);
-    B = centermodii(B, R->N, R->N2);
-    C = centermodii(C, R->N, R->N2); return mkpoln(3,A,B,C);
+    A = modZ(sqri(d), R);
+    B = modZ(mulii(c2, d), R);
+    C = modZ(sqri(c), R); return mkpoln(3,A,B,C);
   }
   b = gel(pol,4);
   if (lv==5)
@@ -226,15 +203,12 @@ sqrmod5(GEN pol, Red *R)
     /* 2a(b - c) + (d-b)(d+b) */
     D = addii(mulii(a2, subii(b,c)), mulii(subii(d,b), addii(d,b)));
   }
-  A = centermodii(A, R->N, R->N2);
-  B = centermodii(B, R->N, R->N2);
-  C = centermodii(C, R->N, R->N2);
-  D = centermodii(D, R->N, R->N2); return mkpoln(4,A,B,C,D);
+  A = modZ(A, R);
+  B = modZ(B, R);
+  C = modZ(C, R);
+  D = modZ(D, R); return mkpoln(4,A,B,C,D);
 }
 
-static GEN
-_mul(GEN x, GEN y, Red *R) { return R->red(gmul(x,y), R); }
-
 /* jac^floor(N/pk) mod (N, polcyclo(pk)), flexible window */
 static GEN
 _powpolmod(GEN C, GEN jac, Red *R, GEN (*_sqr)(GEN, Red *))
@@ -269,13 +243,12 @@ static GEN
 _powpolmodsimple(GEN C, Red *R, GEN jac)
 {
   pari_sp av = avma;
-  GEN w = ZM_ZX_mul(cache_matvite(C), jac);
+  GEN w = ZM_ZX_mul(cache_mat(C), jac);
   long j, ph = lg(w);
 
-  R->red = &_redsimple;
-  for (j=1; j<ph; j++)
-    gel(w,j) = _powpolmod(C, centermodii(gel(w,j), R->N, R->N2), R, &sqrmod);
-  w = centermod_i( gmul(cache_matinvvite(C), w), R->N, R->N2 );
+  R->red = &modZ;
+  for (j=1; j<ph; j++) gel(w,j) = _powpolmod(C, modZ(gel(w,j), R), R, &sqrmodZ);
+  w = centermod_i( ZM_ZC_mul(cache_matinv(C), w), R->N, R->N2);
   w = gerepileupto(av, w);
   return RgV_to_RgX(w, 0);
 }
@@ -285,7 +258,7 @@ powpolmod(GEN C, Red *R, long p, long k, GEN jac)
 {
   GEN (*_sqr)(GEN, Red *);
 
-  if (!isintzero(cache_matvite(C))) return _powpolmodsimple(C, R, jac);
+  if (!isintzero(cache_mat(C))) return _powpolmodsimple(C, R, jac);
   if (p == 2) /* p = 2 */
   {
     if (k == 2) _sqr = &sqrmod4;
@@ -501,7 +474,7 @@ compute_t(GEN N, GEN *e, GEN *faet)
       pari_sp av = avma;
       *e = compute_e(t, faet);
       if (cmpii(*e, B) > 0) break;
-      avma = av;
+      set_avma(av);
     }
 #else
     *e = NULL; /* LCOV_EXCL_LINE */
@@ -546,6 +519,18 @@ compute_g(ulong q)
   return T;
 }
 
+/* x a nonzero VECSMALL with >= 0 entries */
+static GEN
+zv_to_ZX(GEN x)
+{
+  long i,j, lx = lg(x);
+  GEN y;
+
+  while (lx-- && x[lx]==0) /* empty */;
+  i = lx+2; y = cgetg(i,t_POL); y[1] = evalsigne(1);
+  for (j=2; j<i; j++) gel(y,j) = utoi(x[j-1]);
+  return y;
+}
 /* p odd prime */
 static GEN
 get_jac(GEN C, ulong q, long pk, GEN tabg)
@@ -555,7 +540,7 @@ get_jac(GEN C, ulong q, long pk, GEN tabg)
 
   for (x=2; x<=qs2; x++) vpk[ tabg[x]%pk + 1 ] += 2;
   vpk[ tabg[x]%pk + 1 ]++; /* x = (q+1)/2 */
-  return u_red(vpk, cache_cyc(C));
+  return ZX_rem(zv_to_ZX(vpk), cache_cyc(C));
 }
 
 /* p = 2 */
@@ -600,9 +585,9 @@ static GEN
 finda(GEN Cp, GEN N, long pk, long p)
 {
   GEN a, pv;
-  if (Cp && !isintzero(cache_avite(Cp))) {
-    a  = cache_avite(Cp);
-    pv = cache_pkvite(Cp);
+  if (Cp && !isintzero(cache_a(Cp))) {
+    a  = cache_a(Cp);
+    pv = cache_pk(Cp);
   }
   else
   {
@@ -630,8 +615,8 @@ finda(GEN Cp, GEN N, long pk, long p)
     if (!gequal1(b)) return NULL;
 
     if (Cp) {
-      cache_avite(Cp)  = a; /* a has order p^v */
-      cache_pkvite(Cp) = pv;
+      cache_a(Cp)  = a; /* a has order p^v */
+      cache_pk(Cp) = pv;
     }
   }
   return Fp_pow(a, divis(pv, pk), N);
@@ -680,14 +665,14 @@ filltabs(GEN C, GEN Cp, Red *R, long p, long pk, long ltab)
     if (!a) return 0;
     ph = pk - pk/p;
     vpa = cgetg(ph+1,t_COL); gel(vpa,1) = a;
-    if (pk > p) a2 = centermodii(sqri(a), R->N, R->N2);
+    if (pk > p) a2 = modZ(sqri(a), R);
     jj = 1;
     for (i=2; i<pk; i++) /* vpa = { a^i, (i,p) = 1 } */
       if (i%p) {
         GEN z = mulii((i%p==1) ? a2 : a, gel(vpa,jj));
-        gel(vpa,++jj) = centermodii(z , R->N, R->N2);
+        gel(vpa,++jj) = modZ(z , R);
       }
-    if (!gequal1( centermodii( mulii(a, gel(vpa,ph)), R->N, R->N2) )) return 0;
+    if (!gequal1( modZ( mulii(a, gel(vpa,ph)), R) )) return 0;
     p1 = cgetg(ph+1,t_MAT);
     p2 = cgetg(ph+1,t_COL); gel(p1,1) = p2;
     for (i=1; i<=ph; i++) gel(p2,i) = gen_1;
@@ -696,11 +681,11 @@ filltabs(GEN C, GEN Cp, Red *R, long p, long pk, long ltab)
     {
       p2 = cgetg(ph+1,t_COL); gel(p1,j) = p2;
       for (i=1; i<=ph; i++)
-        gel(p2,i) = centermodii(mulii(gel(vpa,i),gel(p3,i)), R->N, R->N2);
+        gel(p2,i) = modZ(mulii(gel(vpa,i),gel(p3,i)), R);
       p3 = p2;
     }
-    cache_matvite(C) = p1;
-    cache_matinvvite(C) = FpM_inv(p1, R->N);
+    cache_mat(C) = p1;
+    cache_matinv(C) = FpM_inv(p1, R->N);
   }
 
   tabt = cgetg(ltab+1, t_VECSMALL);
@@ -715,7 +700,7 @@ filltabs(GEN C, GEN Cp, Red *R, long p, long pk, long ltab)
   }
   setlg(taba, e); cache_aall(C) = taba;
   setlg(tabt, e); cache_tall(C) = tabt;
-  avma = av; return 1;
+  return gc_long(av,1);
 }
 
 static GEN
@@ -761,9 +746,7 @@ aut(long pk, GEN z, long a)
   GEN v;
   long b, i, dz = degpol(z);
   if (a == 1 || dz < 0) return z;
-  v = cgetg(pk+2,t_POL);
-  v[1] = evalvarn(0);
-  b = 0;
+  v = cgetg(pk+2,t_POL); v[1] = evalvarn(0); b = 0;
   gel(v,2) = gel(z,2); /* i = 0 */
   for (i = 1; i < pk; i++)
   {
@@ -777,28 +760,30 @@ aut(long pk, GEN z, long a)
 static GEN
 autvec_TH(long pk, GEN z, GEN v, GEN C)
 {
+  pari_sp av = avma;
   long i, lv = lg(v);
   GEN s = pol_1(varn(C));
   for (i=1; i<lv; i++)
   {
     long y = v[i];
-    if (y) s = RgXQ_mul(s, RgXQ_powu(aut(pk,z, y), y, C), C);
+    if (y) s = ZXQ_mul(s, ZXQ_powu(aut(pk,z, y), y, C), C);
   }
-  return s;
+  return gerepileupto(av,s);
 }
 
 static GEN
 autvec_AL(long pk, GEN z, GEN v, Red *R)
 {
+  pari_sp av = avma;
   const long r = umodiu(R->N, pk);
   GEN s = pol_1(varn(R->C));
   long i, lv = lg(v);
   for (i=1; i<lv; i++)
   {
     long y = (r*v[i]) / pk;
-    if (y) s = RgXQ_mul(s, RgXQ_powu(aut(pk,z, v[i]), y, R->C), R->C);
+    if (y) s = RgXQ_mul(s, ZXQ_powu(aut(pk,z, v[i]), y, R->C), R->C);
   }
-  return s;
+  return gerepileupto(av,s);
 }
 
 /* 0 <= i < pk, such that x^i = z mod polcyclo(pk),  -1 if no such i exist */
@@ -905,11 +890,8 @@ step4d(Red *R, ulong q)
   return -1;
 }
 
-static GEN
-_res(long a, long b) { return b? mkvec2s(a, b): mkvecs(a); }
-
-/* return 1 [OK so far] or <= 0 [not a prime] */
-static GEN
+/* return 1 [OK so far] or 0 [not a prime] */
+static long
 step5(GEN pC, Red *R, long p, GEN et, ulong ltab, long lpC)
 {
   pari_sp av;
@@ -922,7 +904,7 @@ step5(GEN pC, Red *R, long p, GEN et, ulong ltab, long lpC)
   while( (q = u_forprime_next(&T)) )
   { /* q = 1 (mod p) */
     if (umodiu(et,q) == 0) continue;
-    if (umodiu(R->N,q) == 0) return _res(1,p);
+    if (umodiu(R->N,q) == 0) return 0;
     k = u_lval(q-1, p);
     pk = upowuu(p,k);
     if (pk <= lpC && !isintzero(gel(pC,pk))) {
@@ -931,22 +913,21 @@ step5(GEN pC, Red *R, long p, GEN et, ulong ltab, long lpC)
     } else {
       C = gel(pC,1);
       Cp = NULL;
-      cache_matvite(C) = gen_0; /* re-init */
+      cache_mat(C) = gen_0; /* re-init */
     }
-
     av = avma;
-    if (!filltabs(C, Cp, R, p, pk, ltab)) return _res(1,0);
+    if (!filltabs(C, Cp, R, p, pk, ltab)) return 0;
     R->C = cache_cyc(C);
     if (p >= 3)      fl = step4a(C,R, q,p,k, NULL);
     else if (k >= 3) fl = step4b(C,R, q,k);
     else if (k == 2) fl = step4c(C,R, q);
     else             fl = step4d(R, q);
-    if (fl == -1) return _res(q,p);
-    if (fl == 1) return NULL; /*OK*/
-    avma = av;
+    if (fl == -1) return 0;
+    if (fl == 1) return 1; /*OK*/
+    set_avma(av);
   }
   pari_err_BUG("aprcl test fails! This is highly improbable");
-  return NULL;
+  return 0;/*LCOV_EXCL_LINE*/
 }
 
 GEN
@@ -958,33 +939,33 @@ aprcl_step6_worker(GEN r, long t, GEN N, GEN N1, GEN et)
   {
     r = remii(mulii(r,N1), et);
     if (equali1(r)) break;
-    if (dvdii(N,r) && !equalii(r,N)) return mkvec2(r, gen_0);
+    if (dvdii(N,r) && !equalii(r,N)) return gen_0; /* not prime */
     if ((i & 0x1f) == 0) r = gerepileuptoint(av, r);
   }
-  return gen_0;
+  return cgetg(1,t_VECSMALL);
 }
 
-static GEN
+/* return 1 if N prime, else 0 */
+static long
 step6(GEN N, ulong t, GEN et)
 {
-  GEN r, rk, N1 = remii(N, et);
+  GEN worker, r, rk, N1 = remii(N, et);
   ulong k = 10000;
   ulong i;
-  GEN worker, res = NULL;
-  long pending = 0;
+  long pending = 0, res = 1;
   struct pari_mt pt;
   pari_sp btop;
-  worker = strtoclosure("_aprcl_step6_worker", 3, N, N1, et);
+  worker = snm_closure(is_entry("_aprcl_step6_worker"), mkvec3(N, N1, et));
   r = gen_1;
   rk = Fp_powu(N1, k, et);
   mt_queue_start_lim(&pt, worker, (t-1+k-1)/k);
   btop = avma;
-  for (i=1; (i<t && !res) || pending; i+=k)
+  for (i=1; (i<t && res) || pending; i+=k)
   {
     GEN done;
-    mt_queue_submit(&pt, i, (i<t && !res) ? mkvec2(r, utoi(minss(k,t-i))): NULL);
+    mt_queue_submit(&pt, i, (i<t && res)? mkvec2(r, utoi(minss(k,t-i))): NULL);
     done = mt_queue_get(&pt, NULL, &pending);
-    if (done && !isintzero(done)) res = done;
+    if (res && done && typ(done) != t_VECSMALL) res = 0; /* not a prime */
     r = Fp_mul(r, rk, et);
     if (gc_needed(btop, 2))
     {
@@ -993,8 +974,7 @@ step6(GEN N, ulong t, GEN et)
     }
   }
   mt_queue_end(&pt);
-  if (res) return res;
-  return gen_1;
+  return res;
 }
 
 GEN
@@ -1011,7 +991,7 @@ aprcl_step4_worker(ulong q, GEN pC, GEN N, GEN v)
   R.N2= shifti(N, -1);
   R.k = v[1]; R.lv = v[2]; R.mask = uel(v,3); R.n = v[4];
   av2 = avma;
-  for (j=1, k=1; j<lfaq; j++, avma = av2)
+  for (j=1, k=1; j<lfaq; j++, set_avma(av2))
   {
     long p = P[j], e = E[j], pe = PE[j], fl;
     GEN C = gel(pC,pe);
@@ -1020,42 +1000,40 @@ aprcl_step4_worker(ulong q, GEN pC, GEN N, GEN v)
     else if (e >= 3) fl = step4b(C,&R, q,e);
     else if (e == 2) fl = step4c(C,&R, q);
     else             fl = step4d(&R, q);
-    if (fl == -1) return _res(q,p);
-    if (fl == 1)  flags[k++] = p;
+    if (fl == -1) return gen_0; /* not prime */
+    if (fl == 1) flags[k++] = p;
   }
   setlg(flags, k);
-  return gerepileuptoleaf(av1, flags);
+  return gerepileuptoleaf(av1, flags); /* OK so far */
 }
 
-static GEN
+/* return 1 if prime, else 0 */
+static long
 aprcl(GEN N)
 {
-  GEN et, fat, flaglp, faet = NULL; /*-Wall*/
-  long i, j, l, ltab, lfat, lpC;
+  GEN pC, worker, et, fat, flaglp, faet = NULL; /*-Wall*/
+  long i, j, l, ltab, lfat, lpC, workid, pending = 0, res = 1;
   ulong t;
   Red R;
-  GEN pC;
-  GEN worker, res = NULL;
-  long pending = 0, workid;
   struct pari_mt pt;
 
   if (typ(N) != t_INT) pari_err_TYPE("aprcl",N);
   if (cmpis(N,12) <= 0)
     switch(itos(N))
     {
-      case 2: case 3: case 5: case 7: case 11: return gen_1;
-      default: return _res(0,0);
+      case 2: case 3: case 5: case 7: case 11: return 1;
+      default: return 0;
     }
-  if (Z_issquare(N)) return _res(0,0);
+  if (Z_issquare(N)) return 0;
   t = compute_t(N, &et, &faet);
   if (DEBUGLEVEL) err_printf("Starting APRCL with t = %ld\n",t);
   if (cmpii(sqri(et),N) < 0) pari_err_BUG("aprcl: e(t) too small");
-  if (!equali1(gcdii(N,mului(t,et)))) return _res(1,0);
+  if (!equali1(gcdii(N,mului(t,et)))) return 0;
 
   R.N = N;
   R.N2= shifti(N, -1);
   pC = calcglobs(&R, t, &lpC, &ltab, &fat);
-  if (!pC) return _res(1,0);
+  if (!pC) return 0;
   lfat = lg(fat);
   flaglp = cgetg(lfat, t_VECSMALL);
   flaglp[1] = 0;
@@ -1066,34 +1044,33 @@ aprcl(GEN N)
   }
   vecsmall_sort(faet);
   l = lg(faet);
-  worker = strtoclosure("_aprcl_step4_worker", 3,
-                        pC, N, mkvecsmall4(R.k, R.lv, R.mask, R.n));
+  worker = snm_closure(is_entry("_aprcl_step4_worker"),
+                       mkvec3(pC, N, mkvecsmall4(R.k, R.lv, R.mask, R.n)));
   if (DEBUGLEVEL>2) err_printf("Step4: %ld q-values\n", l-1);
   mt_queue_start_lim(&pt, worker, l-1);
-  for (i=l-1; (i>0 && !res) || pending; i--)
+  for (i=l-1; (i>0 && res) || pending; i--)
   {
     GEN done;
     ulong q = i>0 ? faet[i]: 0;
     mt_queue_submit(&pt, q, q>0? mkvec(utoi(q)): NULL);
     done = mt_queue_get(&pt, &workid, &pending);
-    if (done)
+    if (res && done)
     {
       long lf = lg(done);
-      if (typ(done) == t_VEC) res = done;
+      if (typ(done) != t_VECSMALL) { res = 0; continue; } /* not prime */
       for (j=1; j<lf; j++) flaglp[ zv_search(fat, done[j]) ] = 0;
       if (DEBUGLEVEL>2) err_printf("testing Jacobi sums for q = %ld...OK\n", workid);
     }
   }
   mt_queue_end(&pt);
-  if (res) return res;
+  if (!res) return 0;
   if (DEBUGLEVEL>2) err_printf("Step5: testing conditions lp\n");
   for (i=2; i<lfat; i++) /*skip fat[1] = 2*/
   {
     pari_sp av = avma;
     long p = fat[i];
-    GEN r;
-    if (flaglp[i] && (r = step5(pC, &R, p, et, ltab, lpC))) return r;
-    avma = av;
+    if (flaglp[i] && !step5(pC, &R, p, et, ltab, lpC)) return 0;
+    set_avma(av);
   }
   if (DEBUGLEVEL>2)
     err_printf("Step6: testing potential divisors\n");
@@ -1102,11 +1079,7 @@ aprcl(GEN N)
 
 long
 isprimeAPRCL(GEN N)
-{
-  pari_sp av = avma;
-  GEN res = aprcl(N);
-  avma = av; return (typ(res) == t_INT);
-}
+{ pari_sp av = avma; return gc_long(av, aprcl(N)); }
 
 /*******************************************************************/
 /*              DIVISORS IN RESIDUE CLASSES (LENSTRA)              */
