@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -72,7 +73,7 @@ factoru_pow(ulong n)
   F = factoru(n);
   P = gel(F,1);
   E = gel(F,2); l = lg(P);
-  avma = av;
+  set_avma(av);
   gel(f,1) = p = cgetg(l,t_VECSMALL);
   gel(f,2) = e = cgetg(l,t_VECSMALL);
   gel(f,3) = c = cgetg(l,t_VECSMALL);
@@ -158,7 +159,7 @@ factor_eulerphi(GEN n)
         B = to_mat(p, e-1);
     }
   }
-  avma = av;
+  set_avma(av);
   if (!B) return v? to_mat(gen_2, v): trivial_fact();
   A = cgetg(3, t_MAT);
   P = gel(B,1); E = gel(B,2); l = lg(P);
@@ -192,7 +193,7 @@ RgV_is_ZVpos(GEN v)
   }
   return 1;
 }
-/* check whether v is a ZV with non-0 entries */
+/* check whether v is a ZV with nonzero entries */
 int
 RgV_is_ZVnon0(GEN v)
 {
@@ -204,7 +205,7 @@ RgV_is_ZVnon0(GEN v)
   }
   return 1;
 }
-/* check whether v is a ZV with non-zero entries OR exactly [0] */
+/* check whether v is a ZV with nonzero entries OR exactly [0] */
 static int
 RgV_is_ZV0(GEN v)
 {
@@ -297,9 +298,9 @@ check_arith_pos(GEN n, const char *f) {
       return n;
   }
   pari_err_TYPE(f,n);
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
-/* n attached to a factorization of a non-0 integer */
+/* n attached to a factorization of a nonzero integer */
 GEN
 check_arith_non0(GEN n, const char *f) {
   switch(typ(n))
@@ -315,7 +316,7 @@ check_arith_non0(GEN n, const char *f) {
       return n;
   }
   pari_err_TYPE(f,n);
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 /* n attached to a factorization of an integer */
 GEN
@@ -332,7 +333,7 @@ check_arith_all(GEN n, const char *f) {
       return n;
   }
   pari_err_TYPE(f,n);
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 /***********************************************************************/
@@ -451,6 +452,48 @@ divisors_factored(GEN N)
   for (i = 1; i <= n; i++) gmael(D,i,2) = fa_clean(P, gmael(D,i,2));
   return gerepilecopy(av, D);
 }
+static int
+cmpu1(void *E, GEN va, GEN vb)
+{ long a = va[1], b = vb[1]; (void)E; return a>b? 1: (a<b? -1: 0); }
+static GEN
+fa_clean_u(GEN P, GEN E)
+{
+  long i, j, l = lg(E);
+  GEN Q = cgetg(l, t_VECSMALL);
+  for (i = j = 1; i < l; i++)
+    if (E[i]) { Q[j] = P[i]; E[j] = E[i]; j++; }
+  setlg(Q,j);
+  setlg(E,j); return mkmat2(Q,E);
+}
+GEN
+divisorsu_fact_factored(GEN fa)
+{
+  pari_sp av = avma;
+  GEN *d, *t1, *t2, *t3, vD, D, P = gel(fa,1), E = gel(fa,2);
+  long i, j, l, n = ndiv(E);
+
+  D = cgetg(n+1,t_VEC); d = (GEN*)D;
+  l = lg(E);
+  *++d = mkvec2((GEN)1, const_vecsmall(l-1,0));
+  for (i=1; i<l; i++)
+    for (t1=(GEN*)D,j=E[i]; j; j--,t1=t2)
+      for (t2=d, t3=t1; t3<t2; )
+      {
+        ulong a;
+        GEN b;
+        a = (*++t3)[1] * P[i];
+        b = leafcopy(gel(*t3,2)); b[i]++;
+        *++d = mkvec2((GEN)a,b);
+      }
+  gen_sort_inplace(D,NULL,&cmpu1,NULL);
+  vD = cgetg(n+1, t_VECSMALL);
+  for (i = 1; i <= n; i++)
+  {
+    vD[i] = umael(D,i,1);
+    gel(D,i) = fa_clean_u(P, gmael(D,i,2));
+  }
+  return gerepilecopy(av, mkvec2(vD,D));
+}
 GEN
 divisors(GEN N)
 {
@@ -475,6 +518,17 @@ divisors0(GEN N, long flag)
   return flag? divisors_factored(N): divisors(N);
 }
 
+GEN
+divisorsu_moebius(GEN P)
+{
+  GEN d, t, t2, t3;
+  long i, l = lg(P);
+  d = t = cgetg((1 << (l-1)) + 1, t_VECSMALL);
+  *++d = 1;
+  for (i=1; i<l; i++)
+    for (t2=d, t3=t; t3<t2; ) *(++d) = *(++t3) * -P[i];
+  return t;
+}
 GEN
 divisorsu_fact(GEN fa)
 {
@@ -602,10 +656,8 @@ long
 omegau(ulong n)
 {
   pari_sp av;
-  GEN F;
   if (n == 1UL) return 0;
-  av = avma; F = factoru(n);
-  avma = av; return lg(gel(F,1))-1;
+  av = avma; return gc_long(av, nbrows(factoru(n)));
 }
 long
 omega(GEN n)
@@ -621,17 +673,15 @@ omega(GEN n)
   if (lgefint(n) == 3) return omegau(n[2]);
   av = avma;
   F = absZ_factor(n);
-  P = gel(F,1); avma = av; return lg(P)-1;
+  return gc_long(av, nbrows(F));
 }
 
 long
 bigomegau(ulong n)
 {
   pari_sp av;
-  GEN F;
   if (n == 1) return 0;
-  av = avma; F = factoru(n);
-  avma = av; return zv_sum(gel(F,2));
+  av = avma; return gc_long(av, zv_sum(gel(factoru(n),2)));
 }
 long
 bigomega(GEN n)
@@ -650,7 +700,7 @@ bigomega(GEN n)
   else
     E = gel(absZ_factor(n), 2);
   E = ZV_to_zv(E);
-  avma = av; return zv_sum(E);
+  return gc_long(av, zv_sum(E));
 }
 
 /* assume f = factoru(n), possibly with 0 exponents. Return phi(n) */
@@ -676,11 +726,9 @@ eulerphiu_fact(GEN f)
 ulong
 eulerphiu(ulong n)
 {
-  pari_sp av = avma;
-  GEN F;
+  pari_sp av;
   if (!n) return 2;
-  F = factoru(n);
-  avma = av; return eulerphiu_fact(F);
+  av = avma; return gc_long(av, eulerphiu_fact(factoru(n)));
 }
 GEN
 eulerphi(GEN n)
@@ -698,7 +746,7 @@ eulerphi(GEN n)
       ulong e;
       F = mkmat2(ZV_to_nv(gel(F,1)), ZV_to_nv(gel(F,2)));
       e = eulerphiu_fact(F);
-      avma = av; return utoipos(e);
+      set_avma(av); return utoipos(e);
     }
   }
   else if (lgefint(n) == 3) return utoipos(eulerphiu(uel(n,2)));
@@ -731,10 +779,8 @@ long
 numdivu(long N)
 {
   pari_sp av;
-  GEN fa;
   if (N == 1) return 1;
-  av = avma; fa = factoru(N);
-  avma = av; return numdivu_fact(fa);
+  av = avma; return gc_long(av, numdivu_fact(factoru(N)));
 }
 static GEN
 numdiv_aux(GEN F)
@@ -920,7 +966,7 @@ issquarefree(GEN x)
     case t_POL:
       if (!signe(x)) return 0;
       av = avma; d = RgX_gcd(x, RgX_deriv(x));
-      avma = av; return (lg(d) == 3);
+      return gc_long(av, lg(d)==3);
     case t_VEC:
     case t_MAT: return fa_issquarefree(check_arith_all(x,"issquarefree"));
     default: pari_err_TYPE("issquarefree",x);
@@ -1056,7 +1102,7 @@ digits_dacsmall(GEN x, GEN vB, long l, ulong* z)
   q = dvmdii(x, gel(vB,m), &r);
   digits_dacsmall(q,vB,l-m,z);
   digits_dacsmall(r,vB,m,z+l-m);
-  avma = av;
+  set_avma(av);
 }
 
 GEN
@@ -1068,8 +1114,8 @@ digits(GEN x, GEN B)
   if (typ(x)!=t_INT) pari_err_TYPE("digits",x);
   B = check_basis(B);
   if (signe(B)<0) pari_err_DOMAIN("digits","B","<",gen_0,B);
-  if (!signe(x))       {avma = av; return cgetg(1,t_VEC); }
-  if (abscmpii(x,B)<0) {avma = av; retmkvec(absi(x)); }
+  if (!signe(x))       {set_avma(av); return cgetg(1,t_VEC); }
+  if (abscmpii(x,B)<0) {set_avma(av); retmkvec(absi(x)); }
   if (Z_ispow2(B))
   {
     long k = expi(B);
@@ -1078,11 +1124,11 @@ digits(GEN x, GEN B)
     {
       (void)new_chunk(4*(expi(x) + 2)); /* HACK */
       z = binary_2k_nv(x, k);
-      avma = av; return Flv_to_ZV(z);
+      set_avma(av); return Flv_to_ZV(z);
     }
     else
     {
-      avma = av; return binary_2k(x, k);
+      set_avma(av); return binary_2k(x, k);
     }
   }
   x = absi_shallow(x);
@@ -1098,7 +1144,7 @@ digits(GEN x, GEN B)
     (void)new_chunk(3*lz); /* HACK */
     z = zero_zv(lz);
     digits_dacsmall(x,vB,lz,(ulong*)(z+1));
-    avma = av; return Flv_to_ZV(z);
+    set_avma(av); return Flv_to_ZV(z);
   }
 }
 
@@ -1201,11 +1247,11 @@ sumdigitsu(ulong n)
   return s;
 }
 
-/* res=array of 9-digits integers, return \sum_{0 <= i < l} sumdigits(res[i]) */
+/* res=array of 9-digits integers, return sum_{0 <= i < l} sumdigits(res[i]) */
 static ulong
 sumdigits_block(ulong *res, long l)
 {
-  long s = sumdigitsu(*--res);
+  ulong s = sumdigitsu(*--res);
   while (--l > 0) s += sumdigitsu(*--res);
   return s;
 }
@@ -1213,26 +1259,25 @@ sumdigits_block(ulong *res, long l)
 GEN
 sumdigits(GEN n)
 {
+  const long L = (long)(ULONG_MAX / 81);
   pari_sp av = avma;
-  ulong s, *res;
+  ulong *res;
   long l;
 
   if (typ(n) != t_INT) pari_err_TYPE("sumdigits", n);
-  l = lgefint(n);
-  switch(l)
+  switch(lgefint(n))
   {
     case 2: return gen_0;
     case 3: return utoipos(sumdigitsu(n[2]));
   }
   res = convi(n, &l);
-  if ((ulong)l < ULONG_MAX / 81)
+  if (l < L)
   {
-    s = sumdigits_block(res, l);
-    avma = av; return utoipos(s);
+    ulong s = sumdigits_block(res, l);
+    set_avma(av); return utoipos(s);
   }
   else /* Huge. Overflows ulong */
   {
-    const long L = (long)(ULONG_MAX / 81);
     GEN S = gen_0;
     while (l > L)
     {
@@ -1258,19 +1303,19 @@ sumdigits0(GEN x, GEN B)
   if (Z_ispow2(B))
   {
     long k = expi(B);
-    if (k == 1) { avma = av; return utoi(hammingweight(x)); }
+    if (k == 1) { set_avma(av); return utoi(hammingweight(x)); }
     if (k < BITS_IN_LONG)
     {
       GEN z = binary_2k_nv(x, k);
       if (lg(z)-1 > 1L<<(BITS_IN_LONG-k)) /* may overflow */
         return gerepileuptoint(av, ZV_sum(Flv_to_ZV(z)));
-      avma = av; return utoi(zv_sum(z));
+      set_avma(av); return utoi(zv_sum(z));
     }
     return gerepileuptoint(av, ZV_sum(binary_2k(x, k)));
   }
-  if (!signe(x))       { avma = av; return gen_0; }
-  if (abscmpii(x,B)<0) { avma = av; return absi(x); }
-  if (absequaliu(B,10))   { avma = av; return sumdigits(x); }
+  if (!signe(x))       { set_avma(av); return gen_0; }
+  if (abscmpii(x,B)<0) { set_avma(av); return absi(x); }
+  if (absequaliu(B,10))   { set_avma(av); return sumdigits(x); }
   x = absi_shallow(x); lz = logint(x,B) + 1;
   z = gen_digits_i(x, B, lz, NULL, &Z_ring, _dvmdii);
   return gerepileuptoint(av, ZV_sum(z));
