@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -21,9 +22,28 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 int new_galois_format = 0;
 
+/* v a t_VEC, lg(v) = 13, sanity check for true rnf */
+static int
+v13checkrnf(GEN v)
+{ return typ(gel(v,6)) == t_VEC; }
+static int
+rawcheckbnf(GEN v) { return typ(v)==t_VEC && lg(v)==11; }
+static int
+rawchecknf(GEN v) { return typ(v)==t_VEC && lg(v)==10; }
+/* v a t_VEC, lg(v) = 11, sanity check for true bnf */
+static int
+v11checkbnf(GEN v) { return rawchecknf(bnf_get_nf(v)); }
+/* v a t_VEC, lg(v) = 10, sanity check for true nf */
+static int
+v10checknf(GEN v) { return typ(gel(v,1))==t_POL; }
+/* v a t_VEC, lg(v) = 9, sanity check for true gal */
+static int
+v9checkgal(GEN v)
+{ GEN x = gel(v,2); return typ(x) == t_VEC && lg(x) == 4; }
+
 int
 checkrnf_i(GEN rnf)
-{ return (typ(rnf)==t_VEC && lg(rnf)==13); }
+{ return (typ(rnf)==t_VEC && lg(rnf)==13 && v13checkrnf(rnf)); }
 
 void
 checkrnf(GEN rnf)
@@ -39,7 +59,7 @@ checkbnf_i(GEN X)
         if (typ(gel(X,6)) != t_INT) return NULL; /* pre-2.2.4 format */
         if (lg(gel(X,10)) != 4) return NULL; /* pre-2.8.1 format */
         return X;
-      case 7:  return checkbnf_i(bnr_get_bnf(X));
+      case 7: return checkbnf_i(bnr_get_bnf(X));
     }
   return NULL;
 }
@@ -74,20 +94,18 @@ checknf(GEN x)
   return nf;
 }
 
+GEN
+checkbnr_i(GEN bnr)
+{
+  if (typ(bnr)!=t_VEC || lg(bnr)!=7 || !checkbnf_i(bnr_get_bnf(bnr)))
+    return NULL;
+  return bnr;
+}
 void
 checkbnr(GEN bnr)
 {
-  if (typ(bnr)!=t_VEC || lg(bnr)!=7)
+  if (!checkbnr_i(bnr))
     pari_err_TYPE("checkbnr [please apply bnrinit()]",bnr);
-  (void)checkbnf(bnr_get_bnf(bnr));
-}
-
-void
-checkbnrgen(GEN bnr)
-{
-  checkbnr(bnr);
-  if (lg(bnr_get_clgp(bnr))<=3)
-    pari_err_TYPE("checkbnrgen [apply bnrinit(,,1), not bnrinit()]",bnr);
 }
 
 void
@@ -176,15 +194,21 @@ get_bnf(GEN x, long *t)
                 *t = typ_QUA; return NULL;
         case 6: *t = typv6(x); return NULL;
         case 7:  *t = typ_BNR;
-          x = bnr_get_bnf(x); if (typ(x)!=t_VEC || lg(x)!=11) break;
+          x = bnr_get_bnf(x);
+          if (!rawcheckbnf(x)) break;
           return x;
         case 9:
-          x = gel(x,2);
-          if (typ(x) == t_VEC && lg(x) == 4) *t = typ_GAL;
-          return NULL;
-        case 10: *t = typ_NF; return NULL;
-        case 11: *t = typ_BNF; return x;
-        case 13: *t = typ_RNF; return NULL;
+          if (!v9checkgal(x)) break;
+          *t = typ_GAL; return NULL;
+        case 10:
+          if (!v10checknf(x)) break;
+          *t = typ_NF; return NULL;
+        case 11:
+          if (!v11checkbnf(x)) break;
+          *t = typ_BNF; return x;
+        case 13:
+          if (!v13checkrnf(x)) break;
+          *t = typ_RNF; return NULL;
         case 17: *t = typ_ELL; return NULL;
       }
       break;
@@ -212,19 +236,22 @@ get_nf(GEN x, long *t)
           if (typ(gel(x,1)) != t_INT) break;
           *t = typ_QUA; return NULL;
         case 6: *t = typv6(x); return NULL;
-        case 7: *t = typ_BNR;
-          x = bnr_get_bnf(x); if (typ(x)!=t_VEC || lg(x)!=11) break;
-          x = bnf_get_nf(x);  if (typ(x)!=t_VEC || lg(x)!=10) break;
-          return x;
+        case 7:
+          x = bnr_get_bnf(x);
+          if (!rawcheckbnf(x) || !rawchecknf(x = bnf_get_nf(x))) break;
+          *t = typ_BNR; return x;
         case 9:
-          x = gel(x,2);
-          if (typ(x) == t_VEC && lg(x) == 4) *t = typ_GAL;
-          return NULL;
-        case 10: *t = typ_NF; return x;
-        case 11: *t = typ_BNF;
-          x = bnf_get_nf(x); if (typ(x)!=t_VEC || lg(x)!=10) break;
-          return x;
-        case 13: *t = typ_RNF; return NULL;
+          if (!v9checkgal(x)) break;
+          *t = typ_GAL; return NULL;
+        case 10:
+          if (!v10checknf(x)) break;
+          *t = typ_NF; return x;
+        case 11:
+          if (!rawchecknf(x = bnf_get_nf(x))) break;
+          *t = typ_BNF; return x;
+        case 13:
+          if (!v13checkrnf(x)) break;
+          *t = typ_RNF; return NULL;
         case 17: *t = typ_ELL; return NULL;
       }
       break;
@@ -245,22 +272,24 @@ nftyp(GEN x)
     case t_VEC:
       switch(lg(x))
       {
-        case 13: return typ_RNF;
+        case 13:
+          if (!v13checkrnf(x)) break;
+          return typ_RNF;
         case 10:
-          if (typ(gel(x,1))!=t_POL) break;
+          if (!v10checknf(x)) break;
           return typ_NF;
         case 11:
-          x = bnf_get_nf(x); if (typ(x)!=t_VEC || lg(x)!=10) break;
+          if (!v11checkbnf(x)) break;
           return typ_BNF;
         case 7:
-          x = bnr_get_bnf(x); if (typ(x)!=t_VEC || lg(x)!=11) break;
-          x = bnf_get_nf(x);  if (typ(x)!=t_VEC || lg(x)!=10) break;
+          x = bnr_get_bnf(x);
+          if (!rawcheckbnf(x) || !v11checkbnf(x)) break;
           return typ_BNR;
         case 6:
           return typv6(x);
         case 9:
-          x = gel(x,2);
-          if (typ(x) == t_VEC && lg(x) == 4) return typ_GAL;
+          if (!v9checkgal(x)) break;
+          return typ_GAL;
         case 17: return typ_ELL;
       }
   }
@@ -294,7 +323,7 @@ tschirnhaus(GEN x)
   while (degpol(RgX_gcd(u,RgX_deriv(u)))); /* while u not separable */
   if (DEBUGLEVEL>1)
     err_printf("Tschirnhaus transform. New pol: %Ps",u);
-  avma=av2; return gerepileupto(av,u);
+  set_avma(av2); return gerepileupto(av,u);
 }
 
 /* Assume pol in Z[X], monic of degree n. Find L in Z such that
@@ -364,7 +393,7 @@ ZX_primitive_to_monic(GEN pol, GEN *pL)
   else
     POL = leafcopy(pol);
   a = POL+2; lc = gel(a,n);
-  fa = Z_factor_limit(lc,0); L = gen_1;
+  fa = absZ_factor_limit(lc,0); L = gen_1;
   P = gel(fa,1);
   E = gel(fa,2);
   for (i = lg(P)-1; i > 0; i--)
@@ -458,6 +487,22 @@ ZX_to_monic(GEN pol, GEN *L)
   return ZX_primitive_to_monic(Q_primpart(pol), L);
 }
 
+GEN
+ZXX_Q_mul(GEN A, GEN z)
+{
+  long i, l;
+  GEN B;
+  if (typ(z)==t_INT) return ZXX_Z_mul(A,z);
+  B = cgetg_copy(A, &l);
+  B[1] = A[1];
+  for (i=2; i<l; i++)
+  {
+    GEN Ai = gel(A,i);
+    gel(B,i) = typ(Ai)==t_POL ? ZX_Q_mul(Ai, z): gmul(Ai, z);
+  }
+  return B;
+}
+
 /* Evaluate pol in s using nfelt arithmetic and Horner rule */
 GEN
 nfpoleval(GEN nf, GEN pol, GEN s)
@@ -539,25 +584,31 @@ zk_galoisapplymod(GEN nf, GEN z, GEN S, GEN p)
 
 /* true nf */
 static GEN
+pr_make(GEN nf, GEN p, GEN u, GEN e, GEN f)
+{
+  GEN t = FpM_deplin(zk_multable(nf, u), p);
+  t = zk_scalar_or_multable(nf, t);
+  return mkvec5(p, u, e, f, t);
+}
+static GEN
 pr_galoisapply(GEN nf, GEN pr, GEN aut)
 {
-  GEN p, t, u;
-  if (typ(pr_get_tau(pr)) == t_INT) return pr; /* inert */
-  p = pr_get_p(pr);
-  u = zk_galoisapplymod(nf, pr_get_gen(pr), aut, p);
-  t = FpM_deplin(zk_multable(nf, u), p);
-  t = zk_scalar_or_multable(nf, t);
-  return mkvec5(p, u, gel(pr,3), gel(pr,4), t);
+  GEN p = pr_get_p(pr), u = zk_galoisapplymod(nf, pr_get_gen(pr), aut, p);
+  return pr_make(nf, p, u, gel(pr,3), gel(pr,4));
+}
+static GEN
+pr_galoismatrixapply(GEN nf, GEN pr, GEN M)
+{
+  GEN p = pr_get_p(pr), u = FpC_red(ZM_ZC_mul(M, pr_get_gen(pr)), p);
+  return pr_make(nf, p, u, gel(pr,3), gel(pr,4));
 }
 
 static GEN
-vecgaloisapply(GEN nf, GEN aut, GEN v)
-{
-  long i, l;
-  GEN V = cgetg_copy(v, &l);
-  for (i = 1; i < l; i++) gel(V,i) = galoisapply(nf, aut, gel(v,i));
-  return V;
-}
+vecgaloisapply(GEN nf, GEN aut, GEN x)
+{ pari_APPLY_same(galoisapply(nf, aut, gel(x,i))); }
+static GEN
+vecgaloismatrixapply(GEN nf, GEN aut, GEN x)
+{ pari_APPLY_same(nfgaloismatrixapply(nf, aut, gel(x,i))); }
 
 /* x: famat or standard algebraic number, aut automorphism in ZC form
  * simplified from general galoisapply */
@@ -585,6 +636,17 @@ elt_galoisapply(GEN nf, GEN aut, GEN x)
   pari_err_TYPE("galoisapply",x);
   return NULL; /* LCOV_EXCL_LINE */
 }
+/* M automorphism in matrix form */
+static GEN
+elt_galoismatrixapply(GEN nf, GEN M, GEN x)
+{
+  if (typ(x) == t_MAT)
+    switch(lg(x)) {
+      case 1: return cgetg(1, t_MAT);
+      case 3: retmkmat2(vecgaloismatrixapply(nf,M,gel(x,1)), ZC_copy(gel(x,2)));
+    }
+  return nfgaloismatrixapply(nf, M, x);
+}
 
 GEN
 galoisapply(GEN nf, GEN aut, GEN x)
@@ -609,7 +671,9 @@ galoisapply(GEN nf, GEN aut, GEN x)
       aut = algtobasis(nf, aut);
       switch(lg(x))
       {
-        case 6: return gerepilecopy(av, pr_galoisapply(nf, x, aut));
+        case 6:
+          if (pr_is_inert(x)) { set_avma(av); return gcopy(x); }
+          return gerepilecopy(av, pr_galoisapply(nf, x, aut));
         case 3: y = cgetg(3,t_VEC);
           gel(y,1) = galoisapply(nf, aut, gel(x,1));
           gel(y,2) = elt_galoisapply(nf, aut, gel(x,2));
@@ -631,6 +695,49 @@ galoisapply(GEN nf, GEN aut, GEN x)
   return NULL; /* LCOV_EXCL_LINE */
 }
 
+/* M automorphism in galoismatrix form */
+GEN
+nfgaloismatrixapply(GEN nf, GEN M, GEN x)
+{
+  pari_sp av = avma;
+  long lx;
+  GEN y;
+
+  nf = checknf(nf);
+  switch(typ(x))
+  {
+    case t_INT:  return icopy(x);
+    case t_FRAC: return gcopy(x);
+
+    case t_POLMOD: x = gel(x,2); /* fall through */
+    case t_POL:
+      x = algtobasis(nf, x);
+      return gerepileupto(av, basistoalg(nf, RgM_RgC_mul(M, x)));
+
+    case t_VEC:
+      switch(lg(x))
+      {
+        case 6:
+          if (pr_is_inert(x)) { set_avma(av); return gcopy(x); }
+          return gerepilecopy(av, pr_galoismatrixapply(nf, x, M));
+        case 3: y = cgetg(3,t_VEC);
+          gel(y,1) = nfgaloismatrixapply(nf, M, gel(x,1));
+          gel(y,2) = elt_galoismatrixapply(nf, M, gel(x,2));
+          return gerepileupto(av, y);
+      }
+      break;
+
+    case t_COL: return RgM_RgC_mul(M, x);
+
+    case t_MAT: /* ideal */
+      lx = lg(x); if (lx==1) return cgetg(1,t_MAT);
+      if (nbrows(x) != nf_get_degree(nf)) break;
+      return gerepileupto(av, idealhnf_shallow(nf,RgM_mul(M, x)));
+  }
+  pari_err_TYPE("galoisapply",x);
+  return NULL; /* LCOV_EXCL_LINE */
+}
+
 /* compute action of automorphism s on nf.zk */
 GEN
 nfgaloismatrix(GEN nf, GEN s)
@@ -678,16 +785,16 @@ idealquasifrob(GEN nf, GEN gal, GEN grp, GEN pr, GEN subg, GEN *S, GEN aut)
   for (i=1; i<=n; i++)
   {
     GEN g = gel(grp,i);
-    if ((!subg && perm_order(g)==f)
+    if ((!subg && perm_orderu(g) == (ulong)f)
       || (subg && perm_relorder(g, subg)==f))
     {
       *S = get_aut(nf, gal, aut, g);
       if (ZC_prdvd(ZC_galoisapply(nf, *S, pi), pr)) return g;
-      avma = av;
+      set_avma(av);
     }
   }
   pari_err_BUG("idealquasifrob [Frobenius not found]");
-  return NULL; /*LCOV_EXCL_LINE*/
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 GEN
@@ -719,7 +826,7 @@ idealfrobenius_aut(GEN nf, GEN gal, GEN pr, GEN aut)
   GEN T, p, a, b, modpr;
   long f, n, s;
   f = pr_get_f(pr); n = nf_get_degree(nf);
-  if (f==1) { avma = av; return identity_perm(n); }
+  if (f==1) { set_avma(av); return identity_perm(n); }
   g = idealquasifrob(nf, gal, gal_get_group(gal), pr, NULL, &S, aut);
   if (f==2) return gerepileuptoleaf(av, g);
   modpr = zk_to_Fq_init(nf,&pr,&T,&p);
@@ -730,7 +837,7 @@ idealfrobenius_aut(GEN nf, GEN gal, GEN pr, GEN aut)
     a = Fq_pow(a, p, T, p);
     if (ZX_equal(a, b)) break;
   }
-  g = perm_pow(g, Fl_inv(s, f));
+  g = perm_powu(g, Fl_inv(s, f));
   return gerepileupto(av, g);
 }
 
@@ -755,7 +862,7 @@ idealramfrobenius_aut(GEN nf, GEN gal, GEN pr, GEN ram, GEN aut)
   GEN isog, deco;
   long f, n, s;
   f = pr_get_f(pr); n = nf_get_degree(nf);
-  if (f==1) { avma = av; return identity_perm(n); }
+  if (f==1) { set_avma(av); return identity_perm(n); }
   modpr = zk_to_Fq_init(nf,&pr,&T,&p);
   deco = group_elts(gel(ram,1), nf_get_degree(nf));
   isog = group_set(gel(ram,2),  nf_get_degree(nf));
@@ -764,7 +871,7 @@ idealramfrobenius_aut(GEN nf, GEN gal, GEN pr, GEN ram, GEN aut)
   b = nf_to_Fq(nf, zk_galoisapplymod(nf, modpr_genFq(modpr), S, p), modpr);
   for (s=0; !ZX_equal(a, b); s++)
     a = Fq_pow(a, p, T, p);
-  g = perm_pow(g, Fl_inv(s, f));
+  g = perm_powu(g, Fl_inv(s, f));
   return gerepileupto(av, g);
 }
 
@@ -786,17 +893,17 @@ idealinertiagroup(GEN nf, GEN gal, GEN aut, GEN pr)
   for (i=1; i<=n; i++)
   {
     GEN iso = gel(grp,i);
-    if (perm_order(iso) == e)
+    if (perm_orderu(iso) == (ulong)e)
     {
       GEN S = get_aut(nf, gal, aut, iso);
       if (ZC_prdvd(ZC_galoisapply(nf, S, pi), pr)
           && (coprime || gequalX(nf_to_Fq(nf, galoisapply(nf,S,b), modpr))))
           return iso;
-      avma = ltop;
+      set_avma(ltop);
     }
   }
   pari_err_BUG("idealinertiagroup [no isotropic element]");
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 static GEN
@@ -811,7 +918,7 @@ idealramgroupstame(GEN nf, GEN gal, GEN aut, GEN pr)
     if (f==1)
       return cgetg(1,t_VEC);
     frob = idealquasifrob(nf, gal, grp, pr, NULL, &S, aut);
-    avma = av;
+    set_avma(av);
     res = cgetg(2, t_VEC);
     gel(res, 1) = cyclicgroup(frob, f);
     return res;
@@ -819,7 +926,7 @@ idealramgroupstame(GEN nf, GEN gal, GEN aut, GEN pr)
   res = cgetg(3, t_VEC);
   av = avma;
   iso = idealinertiagroup(nf, gal, aut, pr);
-  avma = av;
+  set_avma(av);
   giso = cyclicgroup(iso, e);
   gel(res, 2) = giso;
   if (f==1)
@@ -830,7 +937,7 @@ idealramgroupstame(GEN nf, GEN gal, GEN aut, GEN pr)
   av = avma;
   isog = group_set(giso, nf_get_degree(nf));
   frob = idealquasifrob(nf, gal, grp, pr, isog, &S, aut);
-  avma = av;
+  set_avma(av);
   gel(res, 1) = dicyclicgroup(iso,frob,e,f);
   return res;
 }
@@ -881,7 +988,7 @@ idealramgroupswild(GEN nf, GEN gal, GEN aut, GEN pr)
   {
     GEN S, Spi, piso, iso = gel(grp, i);
     long j, o, ix = iso[1];
-    if (idx[ix] >= 0 || rorder % (o = perm_order(iso))) continue;
+    if (idx[ix] >= 0 || rorder % (o = (long)perm_orderu(iso))) continue;
 
     piso = iso;
     S = get_aut(nf, gal, aut, iso);
@@ -903,7 +1010,7 @@ idealramgroupswild(GEN nf, GEN gal, GEN aut, GEN pr)
       piso = perm_mul(piso,iso);
       if (ugcd(j,o)==1) idx[ piso[1] ] = idx[ix];
     }
-    avma = av2;
+    set_avma(av2);
   }
   return gerepileuptoleaf(av, idx);
 }
@@ -977,14 +1084,24 @@ incl_disc(GEN nfa, GEN a, int nolocal)
   return d;
 }
 
+static int
+badp(GEN fa, GEN db, long q)
+{
+  GEN P = gel(fa,1), E = gel(fa,2);
+  long i, l = lg(P);
+  for (i = 1; i < l; i++)
+    if (mod2(gel(E,i)) && !dvdii(db, powiu(gel(P,i),q))) return 1;
+  return 0;
+}
+
 /* is isomorphism / inclusion (a \subset b) compatible with what we know about
  * basic invariants ? (degree, signature, discriminant); test for isomorphism
  * if fliso is set and for inclusion otherwise */
 static int
 tests_OK(GEN a, GEN nfa, GEN b, GEN nfb, long fliso)
 {
-  GEN da2, da, db, fa, P, E, U;
-  long i, l, nP, q, m = degpol(a), n = degpol(b);
+  GEN da2, da, db, fa, P, U;
+  long i, l, q, m = degpol(a), n = degpol(b);
 
   if (m <= 0) pari_err_IRREDPOL("nfisincl",a);
   if (n <= 0) pari_err_IRREDPOL("nfisincl",b);
@@ -1016,8 +1133,9 @@ tests_OK(GEN a, GEN nfa, GEN b, GEN nfb, long fliso)
     P = nf_get_ramified_primes(nfb); l = lg(P);
     for (i = 1; i < l; i++)
     {
-      long va = Z_pval(da, gel(P,i));
-      if (va && Z_pval(db, gel(P,i)) < q * va) return 0;
+      GEN p = gel(P,i);
+      long va = Z_pval(nfdisc(mkvec2(a, mkvec(p))), p);
+      if (va && Z_pval(db, gel(P,i)) < va * q) return 0;
     }
     return 1;
   }
@@ -1032,28 +1150,12 @@ tests_OK(GEN a, GEN nfa, GEN b, GEN nfb, long fliso)
     da = diviiexact(da, da2);
   }
   if (is_pm1(da)) return 1;
-  fa = absZ_factor_limit(da, 0);
-  P = gel(fa,1);
-  E = gel(fa,2); nP = lg(P) - 1;
-  for (i=1; i<nP; i++) /* all but last factor (primes) */
-    if (mod2(gel(E,i)) && !dvdii(db, powiu(gel(P,i),q))) return 0;
-  U = gel(P,nP); /* unknown */
-  if (mod2(gel(E,i)) && expi(U) < 150)
-  { /* "unfactored" cofactor is small, finish */
-    if (abscmpiu(U, maxprime()) > 0)
-    {
-      fa = Z_factor(U);
-      P = gel(fa,1);
-      E = gel(fa,2);
-    }
-    else
-    {
-      P = mkvec(U);
-      E = mkvec(gen_1);
-    }
-    nP = lg(P) - 1;
-    for (i=1; i<=nP; i++)
-      if (mod2(gel(E,i)) && !dvdii(db, powiu(gel(P,i),q))) return 0;
+  fa = absZ_factor_limit_strict(da, 0, &U);
+  if (badp(fa, db, q)) return 0;
+  if (U && mod2(gel(U,2)) && expi(gel(U,1)) < 150)
+  { /* cofactor is small, finish */
+    fa = absZ_factor(gel(U,1));
+    if (badp(fa, db, q)) return 0;
   }
   return 1;
 }
@@ -1071,7 +1173,7 @@ nfisisom(GEN a, GEN b)
   if (!nfa) { a = Q_primpart(a); RgX_check_ZX(a, "nfisisom"); }
   if (!nfb) { b = Q_primpart(b); RgX_check_ZX(b, "nfisisom"); }
   if (nfa && !nfb) { swap(a,b); nfb = nfa; nfa = NULL; sw = 1; }
-  if (!tests_OK(a, nfa, b, nfb, 1)) { avma = av; return gen_0; }
+  if (!tests_OK(a, nfa, b, nfb, 1)) { set_avma(av); return gen_0; }
 
   if (nfb) lb = gen_1; else nfb = b = ZX_Q_normalize(b,&lb);
   if (nfa) la = gen_1; else nfa = a = ZX_Q_normalize(a,&la);
@@ -1079,7 +1181,7 @@ nfisisom(GEN a, GEN b)
   if (newvar) { a = leafcopy(a); setvarn(a, fetch_var_higher()); }
   y = lift_shallow(nfroots(nfb,a));
   if (newvar) (void)delete_var();
-  lx = lg(y); if (lx==1) { avma=av; return gen_0; }
+  lx = lg(y); if (lx==1) { set_avma(av); return gen_0; }
   if (sw) { vb = va; b = leafcopy(b); setvarn(b, vb); }
   for (i=1; i<lx; i++)
   {
@@ -1093,65 +1195,44 @@ nfisisom(GEN a, GEN b)
 }
 
 static GEN
-partmap_reverse(GEN a, GEN b, GEN F, long v)
+partmap_reverse(GEN a, GEN b, GEN t, GEN la, GEN lb, long v)
 {
   pari_sp av = avma;
-  long i, j, k;
-  long da = degpol(a), d = degpol(F);
-  GEN M1, M2,  W, U, V;
-  M1 = cgetg(1+da-d, t_MAT);
-  M2 = cgetg(1+da-d, t_MAT);
-  for (i=1; i<=da-d; i++)
-  {
-    gel(M1, i) = zerocol(d);
-    gel(M2, i) = zerocol(da-d);
-  }
-  W = monomial(gen_1, d-1, varn(F));
-  for (i=1; i<=da-d; i++)
-  {
-    long l;
-    GEN M1i = gel(M1, i), M2i = gel(M2,i);
-    W = RgX_shift_shallow(W,1);
-    if (degpol(W)==d)
-      W = RgX_sub(W, RgXQX_RgXQ_mul(F, gel(W, d+2), b));
-    l = lg(W);
-    for (j=2; j<l; j++)
-    {
-      GEN Wj = gel(W, j);
-      if (typ(Wj) == t_INT)
-        gel(M1i, j-1) = Wj;
-      else
-      {
-        long lj = lg(Wj), u;
-        if (lj > 2)
-          gel(M1i, j-1) = gel(Wj, 2);
-        for (k=3, u=j-1; k<lj; k++, u+=d)
-          gel(M2i, u) = gel(Wj, k);
-      }
-    }
-  }
-  V = QM_gauss(M2, col_ei(da-d, 1));
-  if (!V) { setvarn(a,v); pari_err_IRREDPOL("nfisincl", a); }
-  U = RgC_neg(QM_QC_mul(M1, V));
-  return gerepilecopy(av, RgV_to_RgX(shallowconcat(U,V), v));
+  GEN rnf = rnfequation2(a, t), z;
+  if (!RgX_equal(b, gel(rnf,1)))
+    { setvarn(b,v); pari_err_IRREDPOL("nfisincl", b); }
+  z = liftpol_shallow(gel(rnf, 2));
+  setvarn(z, v);
+  if (!isint1(lb)) z = RgX_unscale(z, lb);
+  if (!isint1(la)) z = RgX_Rg_div(z, la);
+  return gerepilecopy(av, z);
 }
 
 GEN
-nfisincl(GEN fa, GEN fb)
+nfisincl0(GEN fa, GEN fb, long flag)
 {
   pari_sp av = avma;
   long i, k, vb, lx;
   long da, db, d;
   GEN a, b, nfa, nfb, x, y, la, lb;
   int newvar;
-
+  if (flag < 0 || flag > 1) pari_err_FLAG("nfisincl");
   a = get_nfpol(fa, &nfa);
   b = get_nfpol(fb, &nfb);
   da = degpol(a); db = degpol(b);
-  if (da == db) return nfisisom(fa, fb);
   if (!nfa) { a = Q_primpart(a); RgX_check_ZX(a, "nsisincl"); }
   if (!nfb) { b = Q_primpart(b); RgX_check_ZX(b, "nsisincl"); }
-  if (!tests_OK(a, nfa, b, nfb, 0)) { avma = av; return gen_0; }
+  if (ZX_equal(a, b))
+  {
+    if (flag==1)
+    {
+      x = pol_x(varn(b));
+      return degpol(b) > 1 ? x: RgX_rem(x,b);
+    }
+    x = galoisconj(fb, NULL); settyp(x, t_VEC);
+    return gerepilecopy(av,x);
+  }
+  if (!tests_OK(a, nfa, b, nfb, 0)) { set_avma(av); return gen_0; }
 
   if (nfb) lb = gen_1; else nfb = b = ZX_Q_normalize(b,&lb);
   if (nfa) la = gen_1; else nfa = a = ZX_Q_normalize(a,&la);
@@ -1163,20 +1244,88 @@ nfisincl(GEN fa, GEN fb)
   x = cgetg(lx, t_VEC);
   for (i=1, k=1; i<lx; i++)
   {
-    GEN t = gel(y,i), z;
+    GEN t = gel(y,i);
     if (degpol(t)!=d) continue;
-    z = partmap_reverse(b, a, t, vb);
-    if (lb != gen_1) z = RgX_unscale(z, lb);
-    if (la != gen_1) z = RgX_Rg_div(z, la);
-    gel(x, k++) = z;
+    gel(x, k++) = partmap_reverse(a, b, t, la, lb, vb);
+    if (flag==1) break;
   }
   if (newvar) (void)delete_var();
-  if (k==1) { avma = av; return gen_0; }
+  if (k==1) { set_avma(av); return gen_0; }
+  if (flag==1) return gerepilecopy(av,gel(x,1));
   setlg(x, k);
   gen_sort_inplace(x, (void*)&cmp_RgX, &cmp_nodata, NULL);
   return gerepilecopy(av,x);
 }
 
+GEN
+nfisincl(GEN fa, GEN fb) { return nfisincl0(fa, fb, 0); }
+
+static GEN
+lastel(GEN x) { return gel(x, lg(x)-1); }
+
+static GEN
+nfsplitting_composite(GEN P)
+{
+  GEN F = gel(ZX_factor(P), 1), Q = NULL;
+  long i, n = lg(F)-1;
+  for (i = 1; i <= n; i++)
+  {
+    GEN Fi = gel(F, i);
+    if (degpol(Fi) == 1) continue;
+    Q = Q ? lastel(compositum(Q, Fi)): Fi;
+  }
+  return Q ? Q: pol_x(varn(P));
+}
+GEN
+nfsplitting(GEN T, GEN D)
+{
+  pari_sp av = avma;
+  long d, v;
+  GEN F, K;
+  T = get_nfpol(T,&K);
+  if (!K)
+  {
+    if (typ(T) != t_POL) pari_err_TYPE("nfsplitting",T);
+    T = Q_primpart(T);
+    RgX_check_ZX(T,"nfsplitting");
+  }
+  T = nfsplitting_composite(T);
+  d = degpol(T);
+  if (d<=1) return pol_x(varn(T));
+  if (!K) {
+    if (!isint1(leading_coeff(T))) K = T = polredbest(T,0);
+    K = T;
+  }
+  if (D)
+  {
+    if (typ(D) != t_INT || signe(D) < 1) pari_err_TYPE("nfsplitting",D);
+  }
+  else
+  {
+    char *data = stack_strcat(pari_datadir, "/galdata");
+    long dmax = pari_is_dir(data)? 11: 7;
+    D = (d <= dmax)? gel(polgalois(T,DEFAULTPREC), 1): mpfact(d);
+  }
+  d = itos_or_0(D);
+  v = varn(T);
+  T = leafcopy(T); setvarn(T, fetch_var_higher());
+  for(F = T;;)
+  {
+    GEN P = gel(nffactor(K, F), 1), Q = gel(P,lg(P)-1);
+    if (degpol(gel(P,1)) == degpol(Q)) break;
+    F = rnfequation(K,Q);
+    if (degpol(F) == d) break;
+  }
+  if (umodiu(D,degpol(F)))
+  {
+    char *sD = itostr(D);
+    pari_warn(warner,stack_strcat("ignoring incorrect degree bound ",sD));
+  }
+  (void)delete_var();
+  setvarn(F,v);
+  return gerepilecopy(av, F);
+}
+
 /*************************************************************************/
 /**                                                                     **/
 /**                               INITALG                               **/
@@ -1205,7 +1354,7 @@ get_roots(GEN x, long r1, long prec)
   else
   {
     long n = degpol(x);
-    z = (r1 == n)? realroots(x, NULL, prec): QX_complex_roots(x,prec);
+    z = (r1 == n)? ZX_realroots_irred(x, prec): QX_complex_roots(x,prec);
     ru = (n+r1)>>1;
   }
   for (i=r1+1; i<=ru; i++) gel(z,i) = gel(z, (i<<1)-r1);
@@ -1498,7 +1647,7 @@ nfmaxord_complete(nfmaxord_t *S)
       S->index = get_nfindex(S->basis);
   }
   if (!S->dK) S->dK = diviiexact(S->dT, sqri(S->index));
-  if (S->r1 < 0) S->r1 = ZX_sturm(S->T);
+  if (S->r1 < 0) S->r1 = ZX_sturm_irred(S->T);
   if (!S->basden) S->basden = get_bas_den(S->basis);
 }
 
@@ -1519,6 +1668,7 @@ nfmaxord_to_nf(nfmaxord_t *S, GEN ro, long prec)
   gel(nf,3) = S->dK;
   gel(nf,4) = S->index;
   gel(nf,5) = mat;
+  if (gprecision(gel(F.ro,1)) > prec) F.ro = gprec_wtrunc(F.ro, prec);
   gel(nf,6) = F.ro;
   w = S->basis;
   if (!is_pm1(S->index)) w = Q_remove_denom(w, NULL);
@@ -1550,7 +1700,7 @@ nfmaxord_to_nf(nfmaxord_t *S, GEN ro, long prec)
   gel(mat,3) = RM_round_maxrank(F.G);
   gel(mat,4) = Tr;
   gel(mat,5) = D;
-  gel(mat,8) = S->dKP? shallowtrans(S->dKP): cgetg(1,t_VEC);
+  gel(mat,8) = shallowtrans(S->dKP? S->dKP: gel(absZ_factor(S->dK), 1));
   return nf;
 }
 
@@ -1603,7 +1753,7 @@ get_red_G(nfmaxord_t *S, GEN *pro)
     if (DEBUGLEVEL)
       err_printf("get_red_G: starting LLL, prec = %ld (%ld + %ld)\n",
                   prec + F.extraprec, prec, F.extraprec);
-    if ((u = lllfp(G, 0.99, LLL_KEEP_FIRST|LLL_COMPATIBLE)))
+    if ((u = lllfp(G, 0.99, LLL_KEEP_FIRST)))
     {
       if (lg(u)-1 == n) break;
       /* singular ==> loss of accuracy */
@@ -1624,12 +1774,16 @@ static void
 set_LLL_basis(nfmaxord_t *S, GEN *pro, double DELTA)
 {
   GEN B = S->basis;
-  if (S->r1 < 0) S->r1 = ZX_sturm(S->T);
+  long N = degpol(S->T);
+  if (S->r1 < 0)
+  {
+    S->r1 = ZX_sturm_irred(S->T);
+    if (odd(N - S->r1)) pari_err_IRREDPOL("set_LLL_basis", S->T);
+  }
   if (!S->basden) S->basden = get_bas_den(B);
-  if (S->r1 == degpol(S->T)) {
+  if (S->r1 == N) {
     pari_sp av = avma;
-    GEN u = ZM_lll(make_Tr(S), DELTA,
-                   LLL_GRAM|LLL_KEEP_FIRST|LLL_IM|LLL_COMPATIBLE);
+    GEN u = ZM_lll(make_Tr(S), DELTA, LLL_GRAM|LLL_KEEP_FIRST|LLL_IM);
     B = gerepileupto(av, RgV_RgM_mul(B, u));
     *pro = NULL;
   }
@@ -1706,13 +1860,18 @@ nfpolred(nfmaxord_t *S, GEN *pro)
 static long
 nf_input_type(GEN x)
 {
-  GEN T, V;
+  GEN T, V, DKP = NULL;
   long i, d, v;
   switch(typ(x))
   {
     case t_POL: return t_POL;
     case t_VEC:
-      if (lg(x) != 3) return t_VEC; /* nf or incorrect */
+      switch(lg(x))
+      {
+        case 4: DKP = gel(x,3);
+        case 3: break;
+        default: return t_VEC; /* nf or incorrect */
+      }
       T = gel(x,1); V = gel(x,2);
       if (typ(T) != t_POL) return -1;
       switch(typ(V))
@@ -1736,6 +1895,7 @@ nf_input_type(GEN x)
           default: return -1;
         }
       }
+      if (DKP && (typ(DKP) != t_VEC || !RgV_is_ZV(DKP))) return -1;
       return 0;
   }
   return t_VEC; /* nf or incorrect */
@@ -1759,11 +1919,9 @@ nf_basden(GEN nf)
 void
 nfinit_basic(nfmaxord_t *S, GEN T)
 {
-  long t = nf_input_type(T);
-  if (t == t_POL) { nfmaxord(S, T, 0); return; }
-  S->dTP = S->dTE = S->dKE = S->basden = NULL;
-  switch (t)
+  switch (nf_input_type(T))
   {
+    case t_POL: nfmaxord(S, T, 0); return;
     case t_VEC:
     { /* nf, bnf, bnr */
       GEN nf = checknf(T);
@@ -1771,23 +1929,24 @@ nfinit_basic(nfmaxord_t *S, GEN T)
       S->basis = nf_get_zk(nf); /* probably useless */
       S->basden = nf_basden(nf);
       S->index = nf_get_index(nf);
-      S->dK    = nf_get_disc(nf);
+      S->dK = nf_get_disc(nf);
       S->dKP = nf_get_ramified_primes(nf);
       S->dT = mulii(S->dK, sqri(S->index));
       S->r1 = nf_get_r1(nf); break;
     }
-    case 0: /* monic integral polynomial + integer basis */
+    case 0: /* monic integral polynomial + integer basis (+ ramified primes)*/
       S->T = S->T0 = gel(T,1);
       S->basis = gel(T,2);
+      S->basden = NULL;
       S->index = NULL;
       S->dK = NULL;
-      S->dKP = NULL;
+      S->dKP = lg(T) == 4? gel(T,3): NULL;
       S->dT = NULL;
       S->r1 = -1; break;
     default: /* -1 */
-      pari_err_TYPE("nfbasic_init", T);
-      return;
+      pari_err_TYPE("nfinit_basic", T);
   }
+  S->dTP = S->dTE = S->dKE = NULL;
   S->unscale = gen_1;
 }
 
@@ -1799,7 +1958,7 @@ nfinit_complete(nfmaxord_t *S, long flag, long prec)
   if (!ZX_is_irred(S->T)) pari_err_IRREDPOL("nfinit",S->T);
   if (!(flag & nf_RED) && !ZX_is_monic(S->T0))
   {
-    pari_warn(warner,"non-monic polynomial. Result of the form [nf,c]");
+    pari_warn(warner,"nonmonic polynomial. Result of the form [nf,c]");
     flag |= nf_RED | nf_ORIG;
   }
   unscale = S->unscale;
@@ -1895,7 +2054,7 @@ nfnewprec_shallow(GEN nf, long prec)
   F.r1 = nf_get_r1(nf);
   F.basden = nf_basden(nf);
   F.extraprec = -1;
-  F.prec = prec; make_M_G(&F, 1);
+  F.prec = prec; make_M_G(&F, 0);
   gel(NF,5) = m = leafcopy(gel(NF,5));
   gel(m,1) = F.M;
   gel(m,2) = F.G;
@@ -1935,7 +2094,7 @@ embednorm_T2(GEN x, long r1)
   return avma == av? gcopy(p): gerepileupto(av, p);
 }
 
-/* simplified version of gnorm for scalar, non-complex inputs, without GC */
+/* simplified version of gnorm for scalar, noncomplex inputs, without GC */
 static GEN
 real_norm(GEN x)
 {
@@ -1946,7 +2105,7 @@ real_norm(GEN x)
     case t_FRAC: return sqrfrac(x);
   }
   pari_err_TYPE("real_norm", x);
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 /* simplified version of gnorm, without GC */
 static GEN
@@ -2052,15 +2211,15 @@ try_polmin(CG_data *d, nfmaxord_t *S, GEN v, long flag, GEN *ai)
   if (best)
   {
     ed = expo(embed_disc(v, d->r1, LOWDEFAULTPREC));
-    avma = av; if (d->expo_best_disc < ed) return NULL;
+    set_avma(av); if (d->expo_best_disc < ed) return NULL;
   }
   else
     ed = 0;
   g = get_pol(d, v);
   /* accuracy too low, compute algebraically */
-  if (!g) { avma = av; g = ZXQ_charpoly(*ai, S->T, varn(S->T)); }
+  if (!g) { set_avma(av); g = ZXQ_charpoly(*ai, S->T, varn(S->T)); }
   g = ZX_radical(g);
-  if (best && degpol(g) != degpol(S->T)) { avma = av; return NULL; }
+  if (best && degpol(g) != degpol(S->T)) return gc_NULL(av);
   g = gerepilecopy(av, g);
   d->expo_best_disc = ed;
   if (flag & nf_ORIG)
@@ -2083,9 +2242,9 @@ chk_gen(void *data, GEN x)
   if (!g) pari_err_PREC("chk_gen");
   av1 = avma;
   h = ZX_gcd(g, ZX_deriv(g));
-  if (degpol(h)) { avma = av; return NULL; }
+  if (degpol(h)) return gc_NULL(av);
   if (DEBUGLEVEL>3) err_printf("  generator: %Ps\n",g);
-  avma = av1; return gerepileupto(av, g);
+  set_avma(av1); return gerepileupto(av, g);
 }
 
 static long
@@ -2119,7 +2278,7 @@ remove_duplicates(GEN v)
     }
   l = k+1;
   gel(A,k) = a; setlg(A,l);
-  gel(P,k) = x; setlg(P,l); avma = av;
+  gel(P,k) = x; setlg(P,l); set_avma(av);
 }
 
 static void
@@ -2158,7 +2317,7 @@ findmindisc(GEN y)
   return x;
 }
 /* filter [y,b] from polred_aux: keep a single polynomial of degree n in y
- * [ the best wrt discriminant ordering ], but keep all non-primitive
+ * [ the best wrt discriminant ordering ], but keep all imprimitive
  * polynomials */
 static void
 filter(GEN y, GEN b, long n)
@@ -2173,7 +2332,7 @@ filter(GEN y, GEN b, long n)
     {
       pari_sp av = avma;
       if (!dx) dx = ZX_disc(yi);
-      else if (!ZX_is_better(yi,x,&dx)) { avma = av; continue; }
+      else if (!ZX_is_better(yi,x,&dx)) { set_avma(av); continue; }
       x = yi; a = ai; continue;
     }
     gel(y,k) = yi;
@@ -2289,7 +2448,7 @@ polredbest_aux(nfmaxord_t *S, GEN *pro, GEN *px, GEN *pdx, GEN *pb)
     {
       GEN yi = gel(y,i);
       pari_sp av = avma;
-      if (ZX_is_better(yi,x,&dx)) { x = yi; b = gel(a,i); } else avma = av;
+      if (ZX_is_better(yi,x,&dx)) { x = yi; b = gel(a,i); } else set_avma(av);
     }
     *pb = b;
   }
@@ -2300,7 +2459,7 @@ polredbest_aux(nfmaxord_t *S, GEN *pro, GEN *px, GEN *pdx, GEN *pb)
     {
       GEN yi = gel(y,i);
       pari_sp av = avma;
-      if (ZX_is_better(yi,x,&dx)) x = yi; else avma = av;
+      if (ZX_is_better(yi,x,&dx)) x = yi; else set_avma(av);
     }
   }
   if (pdx) { if (!dx) dx = ZX_disc(x); *pdx = dx; }
@@ -2565,7 +2724,7 @@ polredabs_i(GEN x, nfmaxord_t *S, GEN *u, long flag)
     return mkvec2(mkvec( pol_x(vx) ),
                   mkvec( deg1pol_shallow(gen_1, negi(gel(S->T,2)), vx) ));
   }
-  if (!(flag & nf_PARTIALFACT) && S->dKP)
+  if (!(flag & nf_PARTIALFACT) && S->dK && S->dKP)
   {
     GEN vw = primes_certify(S->dK, S->dKP);
     v = gel(vw,1); l = lg(v);
@@ -2649,15 +2808,6 @@ polredabs(GEN x) { return polredabs0(x,0); }
 GEN
 polredabs2(GEN x) { return polredabs0(x,nf_ORIG); }
 
-/* true nf, partial nf_rnfeq */
-GEN
-nf_rnfeq_partial(GEN nf, GEN R)
-{
-  long sa;
-  GEN T = nf_get_pol(nf), pol = rnfequationall(nf, R, &sa, NULL);
-  return mkvec5(pol,gen_0,stoi(sa),T,liftpol_shallow(R));
-}
-
 /* relative polredabs/best. Returns relative polynomial by default (flag = 0)
  * flag & nf_ORIG: + element (base change)
  * flag & nf_ABSOLUTE: absolute polynomial */
@@ -2680,7 +2830,7 @@ rnfpolred_i(GEN nf, GEN R, long flag, long best)
   R = RgX_nffix(f, T, R, 0);
   if (best || (flag & nf_PARTIALFACT))
   {
-    rnfeq = abs? nf_rnfeq(nf, R): nf_rnfeq_partial(nf, R);
+    rnfeq = abs? nf_rnfeq(nf, R): nf_rnfeqsimple(nf, R);
     pol = gel(rnfeq,1);
     if (listP) pol = mkvec2(pol, listP);
     red = best? polredbest_i(pol, abs? 1: 2)
@@ -2748,11 +2898,11 @@ rnfpolred_i(GEN nf, GEN R, long flag, long best)
   return gerepilecopy(av, P);
 }
 GEN
-rnfpolredabs(GEN nf, GEN relpol, long flag)
-{ return rnfpolred_i(nf,relpol,flag, 0); }
+rnfpolredabs(GEN nf, GEN R, long flag)
+{ return rnfpolred_i(nf,R,flag, 0); }
 GEN
-rnfpolredbest(GEN nf, GEN relpol, long flag)
+rnfpolredbest(GEN nf, GEN R, long flag)
 {
   if (flag < 0 || flag > 3) pari_err_FLAG("rnfpolredbest");
-  return rnfpolred_i(nf,relpol,flag, 1);
+  return rnfpolred_i(nf,R,flag, 1);
 }
