@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -19,30 +20,26 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
-/* must return a t_POL */
+/* eq is an rnfeq; must return a t_POL */
 GEN
-eltreltoabs(GEN rnfeq, GEN x)
+eltreltoabs(GEN eq, GEN x)
 {
-  long i, k, v;
+  GEN Pabs = gel(eq,1), a = gel(eq,2), k = gel(eq,3), T = gel(eq,4), b, s;
+  long i, v = varn(Pabs);
   pari_sp av = avma;
-  GEN T, pol, teta, a, s;
 
-  pol = gel(rnfeq,1);
-  a = gel(rnfeq,2);
-  k = itos(gel(rnfeq,3));
-  T = gel(rnfeq,4);
-
-  v = varn(pol);
   if (varncmp(gvar(x), v) > 0) x = scalarpol(x,v);
   x = RgX_nffix("eltreltoabs", T, x, 1);
-  /* Mod(X - k a, pol(X)), a root of the polynomial defining base */
-  teta = gadd(pol_x(v), gmulsg(-k,a));
+  /* Mod(X - k a, Pabs(X)) is a root of the relative polynomial */
+  if (signe(k))
+    x = RgXQX_translate(x, deg1pol_shallow(negi(k), gen_0, varn(T)), T);
+  b = pol_x(v);
   s = gen_0;
   for (i=lg(x)-1; i>1; i--)
   {
     GEN c = gel(x,i);
-    if (typ(c) == t_POL) c = RgX_RgXQ_eval(c, a, pol);
-    s = RgX_rem(gadd(c, gmul(teta,s)), pol);
+    if (typ(c) == t_POL) c = RgX_RgXQ_eval(c, a, Pabs);
+    s = RgX_rem(gadd(c, gmul(b,s)), Pabs);
   }
   return gerepileupto(av, s);
 }
@@ -76,27 +73,29 @@ rnfeltreltoabs(GEN rnf,GEN x)
       if (varn(x) == rnf_get_nfvarn(rnf)) return rnfeltup(rnf,x);
       retmkpolmod(eltreltoabs(rnf_get_map(rnf), x), ZX_copy(pol));
   }
-  pari_err_TYPE(f,x); return NULL;
+  pari_err_TYPE(f,x);
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 GEN
 eltabstorel_lift(GEN rnfeq, GEN P)
 {
-  GEN k, T = gel(rnfeq,4), relpol = gel(rnfeq,5);
+  GEN k, T = gel(rnfeq,4), R = gel(rnfeq,5);
   if (is_scalar_t(typ(P))) return P;
   k = gel(rnfeq,3);
   P = lift_shallow(P);
-  if (signe(k)) P = RgXQX_translate(P, deg1pol_shallow(k, gen_0, varn(T)), T);
-  P = RgXQX_rem(P, relpol, T);
+  if (signe(k))
+    P = RgXQX_translate(P, deg1pol_shallow(k, gen_0, varn(T)), T);
+  P = RgXQX_rem(P, R, T);
   return QXQX_to_mod_shallow(P, T);
 }
-/* rnfeq = [pol,a,k,T,relpol], P a t_POL or scalar
+/* rnfeq = [pol,a,k,T,R], P a t_POL or scalar
  * Return Mod(P(x + k Mod(y, T(y))), pol(x)) */
 GEN
 eltabstorel(GEN rnfeq, GEN P)
 {
-  GEN T = gel(rnfeq,4), relpol = gel(rnfeq,5);
-  return mkpolmod(eltabstorel_lift(rnfeq,P), QXQX_to_mod_shallow(relpol,T));
+  GEN T = gel(rnfeq,4), R = gel(rnfeq,5);
+  return mkpolmod(eltabstorel_lift(rnfeq,P), QXQX_to_mod_shallow(R,T));
 }
 GEN
 rnfeltabstorel(GEN rnf,GEN x)
@@ -129,7 +128,7 @@ rnfeltabstorel(GEN rnf,GEN x)
       x = nf_to_scalar_or_alg(NF,x); break;
     default:
       pari_err_TYPE(f,x);
-      return NULL;
+      return NULL;/*LCOV_EXCL_LINE*/
   }
   switch(typ(x))
   {
@@ -146,7 +145,7 @@ rnfeltabstorel(GEN rnf,GEN x)
   }
   switch(lg(x))
   {
-    case 2: avma = av; return gen_0;
+    case 2: set_avma(av); return gen_0;
     case 3: return gerepilecopy(av, gel(x,2));
   }
 END:
@@ -173,7 +172,7 @@ modulereltoabs(GEN rnf, GEN x)
     if (lg(id) == 1) continue; /* must be a t_MAT */
     id = Q_primitive_part(id, &cid);
     w = Q_primitive_part(eltreltoabs(rnfeq,w), &c0);
-    c0 = mul_content(c0, mul_content(cid,inv_content(dzknf)));
+    c0 = div_content(mul_content(c0,cid), dzknf);
     if (typ(id) == t_INT)
       for (j=1; j<=m; j++)
       {
@@ -197,7 +196,7 @@ modulereltoabs(GEN rnf, GEN x)
 GEN
 rnf_zkabs(GEN rnf)
 {
-  GEN d, M = modulereltoabs(rnf, rnf_get_zk(rnf));
+  GEN d, v, M = modulereltoabs(rnf, rnf_get_zk(rnf));
   GEN T = rnf_get_polabs(rnf);
   long n = degpol(T);
   M = Q_remove_denom(M, &d); /* t_VEC of t_POL */
@@ -209,7 +208,14 @@ rnf_zkabs(GEN rnf)
   }
   else
     M = matid(n);
-  return mkvec2(T, RgM_to_RgXV(M, varn(T)));
+  v = rnf_get_ramified_primes(rnf);
+  if (lg(v) == 1)
+  {
+    GEN D = gel(rnf_get_disc(rnf),1);
+    if (!isint1(D)) pari_err_TYPE("rnf_zkabs (old style rnf)", rnf);
+  }
+  v = shallowconcat(nf_get_ramified_primes(rnf_get_nf(rnf)), v);
+  return mkvec3(T, RgM_to_RgXV(M, varn(T)), ZV_sort_uniq(v));
 }
 
 static GEN
@@ -259,37 +265,72 @@ nf_nfzk(GEN nf, GEN rnfeq)
   return Q_primpart(QXV_QXQ_eval(nf_get_zkprimpart(nf), a, pol));
 }
 
+static GEN
+rnfdisc_get_T_i(GEN P, GEN *lim)
+{
+  *lim = NULL;
+  if (typ(P) == t_VEC && lg(P) == 3)
+  {
+    GEN L = gel(P,2);
+    long i, l;
+    *lim = L;
+    switch(typ(L))
+    {
+      case t_INT:
+        if (signe(L) <= 0) return NULL;
+        break;
+      case t_VEC: case t_COL:
+        l = lg(L);
+        for (i = 1; i < l; i++)
+        {
+          GEN p = gel(L,i);
+          if (typ(p) == t_INT)
+          { if (signe(p) <= 0) return NULL; }
+          else checkprid(p);
+        }
+        break;
+      default: return NULL;
+    }
+    P = gel(P,1);
+  }
+  return (typ(P) == t_POL)? P: NULL;
+}
 /* true nf */
 GEN
-check_polrel(GEN nf, GEN P, ulong *lim)
+rnfdisc_get_T(GEN nf, GEN P, GEN *lim)
 {
-  if (typ(P) != t_VEC || lg(P) != 3) *lim = 0;
-  else { *lim = gtou(gel(P,2)); P = gel(P,1); }
-  if (typ(P) != t_POL) pari_err_TYPE("rnfinit",P);
-  P = RgX_nffix("rnfinit", nf_get_pol(nf), P, 0);
-  if (!gequal1(leading_coeff(P)))
-    pari_err_IMPL("non-monic relative polynomials");
-  return P;
+  GEN T = rnfdisc_get_T_i(P, lim);
+  if (!T) pari_err_TYPE("rnfdisc",P);
+  return RgX_nffix("rnfdisc", nf_get_pol(nf), T, 1);
+}
+
+GEN
+rnfpseudobasis(GEN nf, GEN pol)
+{
+  pari_sp av = avma;
+  GEN D, z, lim;
+  nf = checknf(nf);
+  pol = rnfdisc_get_T(nf, pol, &lim);
+  z = rnfallbase(nf, pol, lim, NULL, &D, NULL, NULL);
+  return gerepilecopy(av, shallowconcat(z,D));
 }
 
 GEN
 rnfinit0(GEN nf, GEN T, long flag)
 {
   pari_sp av = avma;
-  GEN bas, D, f, B, T0, rnfeq, rnf = obj_init(11, 2);
-  ulong lim;
+  GEN lim, bas, D, f, B, DKP, rnfeq, rnf = obj_init(11, 2);
   nf = checknf(nf);
-  T0 = check_polrel(nf, T, &lim);
-  T = lift_shallow(T0);
+  T = rnfdisc_get_T(nf, T, &lim);
   gel(rnf,11) = rnfeq = nf_rnfeq(nf,T);
   gel(rnf,2) = nf_nfzk(nf, rnfeq);
-  bas = rnfallbase(nf, T0, lim, rnf, &D, &f);
+  bas = rnfallbase(nf, T, lim, rnf, &D, &f, &DKP);
   B = matbasistoalg(nf,gel(bas,1));
   gel(bas,1) = lift_if_rational( RgM_to_RgXV(B,varn(T)) );
   gel(rnf,1) = T;
   gel(rnf,3) = D;
   gel(rnf,4) = f;
-  gel(rnf,5) = cgetg(1, t_VEC); /* dummy */
+  gel(rnf,5) = DKP;
   gel(rnf,6) = cgetg(1, t_VEC); /* dummy */
   gel(rnf,7) = bas;
   gel(rnf,8) = lift_if_rational( RgM_inv_upper(B) );
@@ -318,13 +359,13 @@ rnfeltup0(GEN rnf, GEN x, long flag)
     if (flag) x = nf_to_scalar_or_basis(NF,x);
     return gerepilecopy(av, x);
   }
-  if (NF && tx == t_COL && lg(x)-1 == nf_get_degree(NF))
+  nf = rnf_get_nf(rnf);
+  if (NF && tx == t_COL && lg(x)-1 == degpol(POL) && nf_get_degree(rnf) > 1)
   {
     x = flag? nf_to_scalar_or_basis(NF,x)
             : mkpolmod(nf_to_scalar_or_alg(NF,x), POL);
     return gerepilecopy(av, x);
   }
-  nf = rnf_get_nf(rnf);
   if (NF)
   {
     GEN d, proj;
@@ -356,7 +397,7 @@ nfeltup(GEN nf, GEN x, GEN zknf)
   x = Q_primitive_part(x, &c);
   if (!RgV_is_ZV(x)) pari_err_TYPE("rnfeltup", x);
   if (gequal1(dzknf)) dzknf = NULL;
-  c = mul_content(c, inv_content(dzknf));
+  c = div_content(c, dzknf);
   x = RgV_RgC_mul(zknf, x); if (c) x = RgX_Rg_mul(x, c);
   return x;
 }
@@ -620,7 +661,7 @@ rnfidealdown(GEN rnf,GEN x)
         GEN z, proj = obj_check(rnf,rnf_MAPS), ZK = gel(proj,1);
         long i, lz, l;
         x = idealhnf(NF,x);
-        if (lg(x) == 1) { avma = av; return cgetg(1,t_MAT); }
+        if (lg(x) == 1) { set_avma(av); return cgetg(1,t_MAT); }
         z = ZM_lll(shallowconcat(ZK,x), 0.99, LLL_KER);
         lz = lg(z); l = lg(ZK);
         for (i = 1; i < lz; i++) setlg(gel(z,i), l);
@@ -631,7 +672,7 @@ rnfidealdown(GEN rnf,GEN x)
     }
   }
   x = rnfidealhnf(rnf,x); I = gel(x,2);
-  if (lg(I) == 1) { avma = av; return cgetg(1,t_MAT); }
+  if (lg(I) == 1) { set_avma(av); return cgetg(1,t_MAT); }
   return gerepilecopy(av, gel(I,1));
 }
 
@@ -699,7 +740,7 @@ rnfidealtwoelement(GEN rnf, GEN x)
   y = matalgtobasis(NF, y); settyp(y, t_MAT);
   y = Q_primitive_part(y, &cy);
   y = ZM_hnf(y);
-  if (lg(y) == 1) { avma = av; return mkvec2(gen_0, gen_0); }
+  if (lg(y) == 1) { set_avma(av); return mkvec2(gen_0, gen_0); }
   y = idealtwoelt(NF, y);
   if (cy) y = RgV_Rg_mul(y, cy);
   z = gel(y,2);
@@ -714,7 +755,7 @@ rnfidealmul(GEN rnf,GEN x,GEN y)
   GEN nf, z, x1, x2, p1, p2, bas;
 
   y = rnfidealtwoelement(rnf,y);
-  if (isintzero(gel(y,1))) { avma = av; return rnfideal0(); }
+  if (isintzero(gel(y,1))) { set_avma(av); return rnfideal0(); }
   nf = rnf_get_nf(rnf);
   bas = rnf_get_zk(rnf);
   x = rnfidealhnf(rnf,x);
@@ -726,12 +767,14 @@ rnfidealmul(GEN rnf,GEN x,GEN y)
   return gerepileupto(av, nfhnf(nf,z));
 }
 
+/* prK wrt NF ~ Q[x]/(polabs) */
 static GEN
 rnfidealprimedec_1(GEN rnf, GEN SL, GEN prK)
 {
-  GEN v, piL = rnfeltup0(rnf, pr_get_gen(prK), 1);
+  GEN v, piL, piK = pr_get_gen(prK);
   long i, c, l;
-  if (typ(piL) != t_COL) return SL; /* p inert in K/Q */
+  if (pr_is_inert(prK)) return SL;
+  piL = rnfeltup0(rnf, piK, 1);
   v = cgetg_copy(SL, &l);
   for (i = c = 1; i < l; i++)
   {
@@ -749,16 +792,8 @@ rnfidealprimedec(GEN rnf, GEN pr)
   rnfcomplete(rnf);
   NF = obj_check(rnf,rnf_NFABS);
   nf = rnf_get_nf(rnf);
-  if (typ(pr) == t_INT)
-  {
-    p = pr;
-    pr = NULL;
-  }
-  else
-  {
-    checkprid(pr);
-    p = pr_get_p(pr);
-  }
+  if (typ(pr) == t_INT) { p = pr; pr = NULL; }
+  else { checkprid(pr); p = pr_get_p(pr); }
   SL = idealprimedec(NF, p);
   if (pr) z = rnfidealprimedec_1(rnf, SL, pr);
   else
@@ -817,7 +852,7 @@ rnfequation0(GEN A, GEN B, long flall)
   if (flall)
   { /* a,b,c root of A,B,C = compositum, c = b + k a */
     GEN a, mH0 = RgX_neg(gel(LPRS,1)), H1 = gel(LPRS,2);
-    a = RgXQ_mul(mH0, QXQ_inv(H1, C), C);
+    a = QXQ_div(mH0, H1, C);
     C = mkvec3(C, mkpolmod(a, C), stoi(k));
   }
   return gerepilecopy(av, C);
@@ -827,25 +862,25 @@ rnfequation(GEN nf, GEN pol) { return rnfequation0(nf,pol,0); }
 GEN
 rnfequation2(GEN nf, GEN pol) { return rnfequation0(nf,pol,1); }
 GEN
-nf_rnfeq(GEN nf, GEN relpol)
+nf_rnfeq(GEN nf, GEN R)
 {
   GEN pol, a, k, junk, eq;
-  relpol = liftpol_shallow(relpol);
-  eq = rnfequation2(nf, relpol);
+  R = liftpol_shallow(R);
+  eq = rnfequation2(nf, R);
   pol = gel(eq,1);
   a = gel(eq,2); if (typ(a) == t_POLMOD) a = gel(a,2);
   k = gel(eq,3);
-  return mkvec5(pol,a,k,get_nfpol(nf, &junk),relpol);
+  return mkvec5(pol,a,k,get_nfpol(nf, &junk),R);
 }
 /* only allow abstorel */
 GEN
-nf_rnfeqsimple(GEN nf, GEN relpol)
+nf_rnfeqsimple(GEN nf, GEN R)
 {
   long sa;
   GEN junk, pol;
-  relpol = liftpol_shallow(relpol);
-  pol = rnfequationall(nf, relpol, &sa, NULL);
-  return mkvec5(pol,gen_0/*dummy*/,stoi(sa),get_nfpol(nf, &junk),relpol);
+  R = liftpol_shallow(R);
+  pol = rnfequationall(nf, R, &sa, NULL);
+  return mkvec5(pol,gen_0/*dummy*/,stoi(sa),get_nfpol(nf, &junk),R);
 }
 
 /*******************************************************************/
@@ -1342,7 +1377,7 @@ zero_nfbezout(GEN nf,GEN bB, GEN b, GEN A,GEN B,GEN *u,GEN *v,GEN *w,GEN *di)
 
 /* Given elements a,b and ideals A, B, outputs d = a.A+b.B and gives
  * di=d^-1, w=A.B.di, u, v such that au+bv=1 and u in A.di, v in B.di.
- * Assume A, B non-zero, but a or b can be zero (not both) */
+ * Assume A, B nonzero, but a or b can be zero (not both) */
 static GEN
 nfbezout(GEN nf,GEN a,GEN b, GEN A,GEN B, GEN *pu,GEN *pv,GEN *pw,GEN *pdi,
          int red)
@@ -1605,7 +1640,7 @@ nfsnf0(GEN nf, GEN x, long flag)
         /* find d in D = I[k]/I[i] not in J[i]/(A[k,l] J[l]) */
         D = idealdiv(nf,gel(I,k),gel(I,i));
         p2 = idealdiv(nf,gel(J,i), p1);
-        d = RgV_find_denom( RgM_solve(p2, D) );
+        d = RgV_find_denom(QM_gauss(p2, D));
         if (!d) pari_err_BUG("nfsnf");
         p1 = element_mulvecrow(nf,d,A,k,i);
         for (t=1; t<=i; t++) gcoeff(A,i,t) = gadd(gcoeff(A,i,t),gel(p1,t));
@@ -1709,7 +1744,7 @@ nfdetint(GEN nf, GEN x)
       gerepileall(av1,6, &det1,&piv,&pivprec,&pass,&v,&idprod);
     }
   }
-  if (!cm) { avma = av; return cgetg(1,t_MAT); }
+  if (!cm) { set_avma(av); return cgetg(1,t_MAT); }
   return gerepileupto(av, idealmul(nf,idprod,det1));
 }
 
