@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -77,7 +78,6 @@ sliding_window_powu(GEN x, ulong n, long e, void *E, GEN (*sqr)(void*,GEN),
   return z;
 }
 
-
 /* assume n != 0, t_INT. Compute x^|n| using sliding window powering */
 static GEN
 sliding_window_pow(GEN x, GEN n, long e, void *E, GEN (*sqr)(void*,GEN),
@@ -350,39 +350,32 @@ producttree_scheme(long n)
     }
     swap(v,w); l<<=1;
   }
-  fixlg(v, l+1);
-  avma = (pari_sp) v;
-  return v;
+  fixlg(v, l+1); set_avma((pari_sp)v); return v;
 }
 
 GEN
-gen_product(GEN x, void *data, GEN (*mul)(void *,GEN,GEN))
+gen_product(GEN x, void *E, GEN (*mul)(void *,GEN,GEN))
 {
-  pari_sp ltop;
-  long i,k,lx = lg(x),lv;
+  pari_sp av;
+  long i, k, l = lg(x);
   pari_timer ti;
-  GEN v;
-  if (DEBUGLEVEL>7) timer_start(&ti);
+  GEN y, v;
 
-  if (lx == 1) return gen_1;
-  if (lx == 2) return gcopy(gel(x,1));
-  x = leafcopy(x); k = lx;
-  v = producttree_scheme(lx-1); lv = lg(v);
-  ltop = avma;
-  for (k=1, i=1; k<lv; i+=v[k++])
-    gel(x,k) = v[k]==1 ? gel(x,i): mul(data,gel(x,i),gel(x,i+1));
+  if (l <= 2) return l == 1? gen_1: gcopy(gel(x,1));
+  y = cgetg(l, t_VEC); av = avma;
+  v = producttree_scheme(l-1);
+  l = lg(v);
+  if (DEBUGLEVEL>7) timer_start(&ti);
+  for (k = i = 1; k < l; i += v[k++])
+    gel(y,k) = v[k]==1? gel(x,i): mul(E, gel(x,i), gel(x,i+1));
   while (k > 2)
   {
-    if (DEBUGLEVEL>7)
-      timer_printf(&ti,"gen_product: remaining objects %ld",k-1);
-    lx = k; k = 1;
-    for (i=1; i<lx-1; i+=2)
-      gel(x,k++) = mul(data,gel(x,i),gel(x,i+1));
-    if (i < lx) gel(x,k++) = gel(x,i);
-    if (gc_needed(ltop,1))
-      gerepilecoeffs(ltop,x+1,k-1);
+    long n = k - 1;
+    if (DEBUGLEVEL>7) timer_printf(&ti,"gen_product: remaining objects %ld",n);
+    for (k = i = 1; i < n; i += 2) gel(y,k++) = mul(E, gel(y,i), gel(y,i+1));
+    if (gc_needed(av,1)) gerepilecoeffs(av, y+1, k-1);
   }
-  return gel(x,1);
+  return gel(y,1);
 }
 
 /***********************************************************************/
@@ -390,22 +383,17 @@ gen_product(GEN x, void *data, GEN (*mul)(void *,GEN,GEN))
 /**                    DISCRETE LOGARITHM                             **/
 /**                                                                   **/
 /***********************************************************************/
-
 static GEN
 iter_rho(GEN x, GEN g, GEN q, GEN A, ulong h, void *E, const struct bb_group *grp)
 {
-  GEN a = gel(A,1);
-  switch((h|grp->hash(a))%3UL)
+  GEN a = gel(A,1), b = gel(A,2), c = gel(A,3);
+  switch((h | grp->hash(a)) % 3UL)
   {
-    case 0:
-      return mkvec3(grp->pow(E,a,gen_2),Fp_mulu(gel(A,2),2,q),
-                                        Fp_mulu(gel(A,3),2,q));
-    case 1:
-      return mkvec3(grp->mul(E,a,x),addis(gel(A,2),1),gel(A,3));
-    case 2:
-      return mkvec3(grp->mul(E,a,g),gel(A,2),addiu(gel(A,3),1));
+    case 0: return mkvec3(grp->pow(E,a,gen_2), Fp_mulu(b,2,q), Fp_mulu(c,2,q));
+    case 1: return mkvec3(grp->mul(E,a,x), addiu(b,1), c);
+    case 2: return mkvec3(grp->mul(E,a,g), b, addiu(c,1));
   }
-  return NULL;
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /*Generic Pollard rho discrete log algorithm*/
@@ -518,7 +506,7 @@ gen_Shanks_log(GEN x, GEN g, GEN q, void *E, const struct bb_group *grp)
   av1 = avma;
   for (p1=x, i=1;;i++)
   {
-    if (grp->equal1(p1)) { avma = av; return stoi(i-1); }
+    if (grp->equal1(p1)) { set_avma(av); return stoi(i-1); }
     table[i] = grp->hash(p1); if (i==lbaby) break;
     p1 = grp->mul(E,p1,ginv);
     if (gc_needed(av1,2))
@@ -552,7 +540,7 @@ gen_Shanks_log(GEN x, GEN g, GEN q, void *E, const struct bb_group *grp)
       p1 = gerepileupto(av1, p1);
     }
   }
-  avma = av; return cgetg(1, t_VEC); /* no solution */
+  set_avma(av); return cgetg(1, t_VEC); /* no solution */
 }
 
 /*Generic discrete logarithme in a group of prime order p*/
@@ -612,16 +600,15 @@ get_arith_Z(GEN o)
   return NULL; /* LCOV_EXCL_LINE */
 }
 
-/* grp->easylog() is an optional trapdoor function that catch easy logarithms*/
-/* Generic Pohlig-Hellman discrete logarithm*/
-/* smallest integer n such that g^n=a. Assume g has order ord */
+/* Generic Pohlig-Hellman discrete logarithm: smallest integer n >= 0 such that
+ * g^n=a. Assume ord(g) | ord; grp->easylog() is an optional trapdoor
+ * function that catches easy logarithms */
 GEN
 gen_PH_log(GEN a, GEN g, GEN ord, void *E, const struct bb_group *grp)
 {
   pari_sp av = avma;
-  GEN v,t0,a0,b,q,g_q,n_q,ginv0,qj,ginv;
-  GEN fa, ex;
-  long e,i,j,l;
+  GEN v, ginv, fa, ex;
+  long i, j, l;
 
   if (grp->equal(g, a)) /* frequent special case */
     return grp->equal1(g)? gen_0: gen_1;
@@ -637,43 +624,39 @@ gen_PH_log(GEN a, GEN g, GEN ord, void *E, const struct bb_group *grp)
   fa = gel(fa,1); l = lg(fa);
   ginv = grp->pow(E,g,gen_m1);
   v = cgetg(l, t_VEC);
-  for (i=1; i<l; i++)
+  for (i = 1; i < l; i++)
   {
-    q = gel(fa,i);
-    e = itos(gel(ex,i));
+    GEN q = gel(fa,i), qj, gq, nq, ginv0, a0, t0;
+    long e = itos(gel(ex,i));
     if (DEBUGLEVEL>5)
       err_printf("Pohlig-Hellman: DL mod %Ps^%ld\n",q,e);
     qj = new_chunk(e+1);
     gel(qj,0) = gen_1;
     gel(qj,1) = q;
-    for (j=2; j<=e; j++) gel(qj,j) = mulii(gel(qj,j-1), q);
+    for (j = 2; j <= e; j++) gel(qj,j) = mulii(gel(qj,j-1), q);
     t0 = diviiexact(ord, gel(qj,e));
     a0 = grp->pow(E, a, t0);
-    ginv0 = grp->pow(E, ginv, t0); /* order q^e */
-    if (grp->equal1(ginv0))
-    {
-      gel(v,i) = mkintmod(gen_0, gen_1);
-      continue;
-    }
-    do { g_q = grp->pow(E,g, mulii(t0, gel(qj,--e))); /* order q */
-    } while (grp->equal1(g_q));
-    n_q = gen_0;
-    for (j=0;; j++)
-    { /* n_q = sum_{i<j} b_i q^i */
-      b = grp->pow(E,a0, gel(qj,e-j));
-      /* early abort: cheap and very effective */
+    ginv0 = grp->pow(E, ginv, t0); /* ord(ginv0) | q^e */
+    if (grp->equal1(ginv0)) { gel(v,i) = mkintmod(gen_0, gen_1); continue; }
+    do gq = grp->pow(E,g, mulii(t0, gel(qj,--e))); while (grp->equal1(gq));
+    /* ord(gq) = q */
+    nq = gen_0;
+    for (j = 0;; j++)
+    { /* nq = sum_{i<j} b_i q^i */
+      GEN b = grp->pow(E,a0, gel(qj,e-j));
+      /* cheap early abort: wrong local order */
       if (j == 0 && !grp->equal1(grp->pow(E,b,q))) {
-        avma = av; return cgetg(1, t_VEC);
+        set_avma(av); return cgetg(1, t_VEC);
       }
-      b = gen_plog(b, g_q, q, E, grp);
-      if (typ(b) != t_INT) { avma = av; return cgetg(1, t_VEC); }
-      n_q = addii(n_q, mulii(b, gel(qj,j)));
+      b = gen_plog(b, gq, q, E, grp);
+      if (typ(b) != t_INT) { set_avma(av); return cgetg(1, t_VEC); }
+      nq = addii(nq, mulii(b, gel(qj,j)));
       if (j == e) break;
 
       a0 = grp->mul(E,a0, grp->pow(E,ginv0, b));
       ginv0 = grp->pow(E,ginv0, q);
     }
-    gel(v,i) = mkintmod(n_q, gel(qj,e+1));
+    gel(v,i) = mkintmod(nq, gel(qj,e+1));
   }
   return gerepileuptoint(av, lift(chinese1_coprime_Z(v)));
 }
@@ -794,13 +777,13 @@ gen_select_order(GEN o, void *E, const struct bb_group *grp)
       lasto = newo;
       if (!grp->equal1(t))
       {
-        if (--nbo == 1) { avma=ltop; return icopy(lastgood); }
+        if (--nbo == 1) { set_avma(ltop); return icopy(lastgood); }
         vo[i] = 1;
       }
       else
         lastgood = lasto;
     }
-    avma = btop;
+    set_avma(btop);
   }
 }
 
@@ -820,7 +803,7 @@ gen_lgener(GEN l, long e, GEN r,GEN *zeta, void *E, const struct bb_group *grp)
   const pari_sp av1 = avma;
   GEN m, m1;
   long i;
-  for (;; avma = av1)
+  for (;; set_avma(av1))
   {
     m1 = m = grp->pow(E, grp->rand(E), r);
     if (grp->equal1(m)) continue;
@@ -888,9 +871,9 @@ gen_Shanks_sqrtl(GEN a, GEN l, long e, GEN r, GEN y, GEN m,void *E,
       z = p1; p1 = grp->pow(E,p1,l);
       k++;
     } while(!grp->equal1(p1));
-    if (k==e) { avma = av; return NULL; }
+    if (k==e) return gc_NULL(av);
     dl = gen_plog(z,m,l,E,grp);
-    if (typ(dl) != t_INT) { avma = av; return NULL; }
+    if (typ(dl) != t_INT) return gc_NULL(av);
     dl = negi(dl);
     p1 = grp->pow(E, grp->pow(E,y, dl), powiu(l,e-k-1));
     m = grp->pow(E,m,dl);
@@ -948,7 +931,7 @@ gen_Shanks_sqrtn(GEN a, GEN n, GEN q, GEN *zetan, void *E, const struct bb_group
         do
         {
           a = gen_Shanks_sqrtl(a,l,e,r,y,zeta,E,grp);
-          if (!a) { avma = ltop; return NULL;}
+          if (!a) return gc_NULL(ltop);
         } while (--j);
       }
       if (gc_needed(ltop,1))
@@ -1005,7 +988,7 @@ gen_ellgroup(GEN N, GEN d, GEN *pt_m, void *E, const struct bb_group *grp,
   if (pt_m) *pt_m = gen_1;
   if (is_pm1(N)) return cgetg(1,t_VEC);
   F = ellgroup_d2(N, d);
-  if (!F) {avma = av; return mkveccopy(N);}
+  if (!F) {set_avma(av); return mkveccopy(N);}
   N0 = gel(F,1); N1 = diviiexact(N, N0);
   while(1)
   {
@@ -1013,15 +996,15 @@ gen_ellgroup(GEN N, GEN d, GEN *pt_m, void *E, const struct bb_group *grp,
     GEN P, Q, d, s, t, m;
 
     P = grp->pow(E,grp->rand(E), N1);
-    s = gen_order(P, F, E, grp); if (equalii(s, N0)) {avma = av; return mkveccopy(N);}
+    s = gen_order(P, F, E, grp); if (equalii(s, N0)) {set_avma(av); return mkveccopy(N);}
 
     Q = grp->pow(E,grp->rand(E), N1);
-    t = gen_order(Q, F, E, grp); if (equalii(t, N0)) {avma = av; return mkveccopy(N);}
+    t = gen_order(Q, F, E, grp); if (equalii(t, N0)) {set_avma(av); return mkveccopy(N);}
 
     m = lcmii(s, t);
     d = pairorder(E, P, Q, m, F);
     /* structure is [N/d, d] iff m d == N0. Note that N/d = N1 m */
-    if (is_pm1(d) && equalii(m, N0)) {avma = av; return mkveccopy(N);}
+    if (is_pm1(d) && equalii(m, N0)) {set_avma(av); return mkveccopy(N);}
     if (equalii(mulii(m, d), N0))
     {
       GEN g = mkvec2(mulii(N1,m), d);
@@ -1029,7 +1012,7 @@ gen_ellgroup(GEN N, GEN d, GEN *pt_m, void *E, const struct bb_group *grp,
       gerepileall(av,pt_m?2:1,&g,pt_m);
       return g;
     }
-    avma = av2;
+    set_avma(av2);
   }
 }
 
@@ -1045,14 +1028,14 @@ gen_ellgens(GEN D1, GEN d2, GEN m, void *E, const struct bb_group *grp,
   av = avma;
   do
   {
-    avma = av;
+    set_avma(av);
     P = grp->rand(E);
     s = gen_order(P, F, E, grp);
   } while (!equalii(s, d1));
   av = avma;
   do
   {
-    avma = av;
+    set_avma(av);
     Q = grp->rand(E);
     d = pairorder(E, grp->pow(E, P, dm), grp->pow(E, Q, dm), m, F);
   } while (!equalii(d, d2));
