@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -521,7 +522,7 @@ gen_last_inv_diago(GEN A, void* data, const struct bb_hermite *R)
 static GEN
 /* remove_zerocols: 0 none, 1 until square, 2 all */
 /* early abort: if not right-invertible, abort, return NULL, and set ops to the
- * non-invertible pivot */
+ * noninvertible pivot */
 gen_howell_i(GEN A, long remove_zerocols, long permute_zerocols, long early_abort, long only_triangular, GEN* ops, void *data, const struct bb_hermite *R)
 {
   pari_sp av = avma, av1;
@@ -596,7 +597,6 @@ gen_howell_i(GEN A, long remove_zerocols, long permute_zerocols, long early_abor
       }
     }
 
-
     if (early_abort)
     {
       d = gcoeff(H,i,si);
@@ -786,15 +786,24 @@ gen_matimage(GEN A, GEN* U, void *data, const struct bb_hermite *R)
   GEN ops, H;
   if (U)
   {
-    pari_sp av = avma;
-    long m, n, i, r, n2;
+    pari_sp av = avma, av1;
+    long m, n, i, r, n2, pergc;
     RgM_dimensions(A,&m,&n);
     H = gen_howell_i(A, 2, 1, 0, 0, &ops, data, R);
+    av1 = avma;
     r = lg(H)-1;
     *U = shallowmatconcat(mkvec2(gen_zeromat(n, maxss(0,m-n+1), data, R), gen_matid_hermite(n, data, R)));
     n2 = lg(*U)-1;
+    pergc = maxss(m,n);
     for (i=1; i<lg(ops); i++)
+    {
       gen_rightapply(*U, gel(ops,i), data, R);
+      if (!(i%pergc) && gc_needed(av,1))
+      {
+        if (DEBUGMEM>1) pari_warn(warnmem,"gen_matimage. i=%ld",i);
+        gerepileall(av1,1,U);
+      }
+    }
     if (r<n2) *U = vecslice(*U, n2-r+1, n2);
     gerepileall(av, 2, &H, U);
     return H;
@@ -952,7 +961,7 @@ gen_solve_from_howell(GEN H, GEN ops, long m, long n, GEN Y, void* data, const s
   RgM_dimensions(H,&mH,&nH);
   n2 = maxss(n,m+1);
   Z = gen_reduce_mod_howell(H, Y, &X, data, R);
-  if (!gen_is_zerocol(Z,m,data,R)) { avma=av; return NULL; }
+  if (!gen_is_zerocol(Z,m,data,R)) { set_avma(av); return NULL; }
 
   X = shallowconcat(zerocol(n2-nH),X);
   for (i=lg(ops)-1; i>0; i--) gen_leftapply(X, gel(ops,i), data, R);
@@ -1198,7 +1207,7 @@ msolvemod0(GEN M, GEN D, GEN Y, long flag)
   if (!flag) return gaussmoduloall(M,D,Y,NULL);
   y = cgetg(3,t_VEC);
   x = gaussmoduloall(M,D,Y,&U);
-  if (x == gen_0) { avma = av; return gen_0; }
+  if (x == gen_0) { set_avma(av); return gen_0; }
   gel(y,1) = x;
   gel(y,2) = U; return y;
 
