@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -28,13 +29,13 @@ lemma6(GEN T, GEN p, long nu, GEN x)
   pari_sp av = avma;
   GEN gpx, gx = poleval(T, x);
 
-  if (Zp_issquare(gx, p)) { avma = av; return 1; }
+  if (Zp_issquare(gx, p)) return gc_long(av,1);
 
   la = Z_pval(gx, p);
   gpx = poleval(ZX_deriv(T), x);
   mu = signe(gpx)? Z_pval(gpx,p)
                  : la+nu+1; /* mu = +oo */
-  avma = av;
+  set_avma(av);
   if (la > mu<<1) return 1;
   if (la >= nu<<1 && mu >= nu) return 0;
   return -1;
@@ -51,7 +52,7 @@ lemma7(GEN T, long nu, GEN x)
 
   gpx = poleval(ZX_deriv(T), x);
   la = Z_lvalrem(gx, 2, &oddgx);
-  odd4 = umodiu(oddgx,4); avma = av;
+  odd4 = umodiu(oddgx,4); set_avma(av);
 
   mu = vali(gpx);
   if (mu < 0) mu = la+nu+1; /* mu = +oo */
@@ -89,7 +90,7 @@ zpsol(GEN T, GEN p, long nu, GEN pnu, GEN x0)
   for (i=0; i < itos(p); i++)
   {
     x = addii(x,pnu);
-    if (zpsol(T,p,nu+1,pnup,x)) { avma = av; return 1; }
+    if (zpsol(T,p,nu+1,pnup,x)) return gc_long(av,1);
     if (gc_needed(btop, 2))
     {
       x = gerepileupto(btop, x);
@@ -97,7 +98,7 @@ zpsol(GEN T, GEN p, long nu, GEN pnu, GEN x0)
         pari_warn(warnmem, "hyperell_locally_soluble: %ld/%Ps",i,p);
     }
   }
-  avma = av; return 0;
+  return gc_long(av,0);
 }
 
 /* return 1 if equation y^2=T(x) has a rational p-adic solution (possibly
@@ -111,7 +112,7 @@ hyperell_locally_soluble(GEN T,GEN p)
   if (typ(p)!=t_INT) pari_err_TYPE("hyperell_locally_soluble",p);
   RgX_check_ZX(T, "hyperell_locally_soluble");
   res = zpsol(T,p,0,gen_1,gen_0) || zpsol(RgX_recip_shallow(T), p, 1, p, gen_0);
-  avma = av; return res;
+  return gc_long(av, res);
 }
 
 /* is t a square in (O_K/pr) ? Assume v_pr(t) = 0 */
@@ -133,7 +134,7 @@ quad_char(GEN nf, GEN t, GEN pr)
   }
   return kronecker(t, p);
 }
-/* quad_char(x), x in Z, non-zero mod p */
+/* quad_char(x), x in Z, nonzero mod p */
 static long
 Z_quad_char(GEN x, GEN pr)
 {
@@ -162,34 +163,68 @@ psquarenf(GEN nf,GEN x,GEN pr,GEN modpr)
     if (v&1) return 0;
     v = (quad_char(nf, x, modpr) == 1);
   }
-  avma = av; return v;
+  return gc_long(av,v);
 }
 
-/* Is  x a square in (ZK / pr^(1+2e))^* ?  pr | 2 */
 static long
-check2(GEN nf, GEN x, GEN sprk)
+ZV_iseven(GEN zlog)
 {
-  GEN zlog = zlog_pr(nf, x, sprk);
   long i, l = lg(zlog);
-  for (i=1; i<l; i++) /* all elementary divisors are even (1+2e > 1) */
+  for (i = 1; i < l; i++)
     if (mpodd(gel(zlog,i))) return 0;
   return 1;
 }
 
+/* pr | 2, project to principal units (trivializes later discrete log) */
+static GEN
+to_principal_unit(GEN nf, GEN x, GEN pr, GEN sprk)
+{
+  if (pr_get_f(pr) != 1)
+  {
+    GEN prk = gel(sprk,3);
+    x = nfpowmodideal(nf, x, gmael(sprk,5,1), prk);
+  }
+  return x;
+}
 /* pr | 2. Return 1 if x in Z_K is square in Z_{K_pr}, 0 otherwise */
 static int
-psquare2nf_i(GEN nf,GEN x,GEN pr,GEN sprk)
+psquare2nf(GEN nf, GEN x, GEN pr, GEN sprk)
 {
   long v = nfvalrem(nf, x, pr, &x);
-  /* now (x,pr) = 1 */
-  return v == LONG_MAX || (!odd(v) && check2(nf,x,sprk));
+  if (v == LONG_MAX) return 1; /* x = 0 */
+  /* (x,pr) = 1 */
+  if (odd(v)) return 0;
+  x = to_principal_unit(nf, x, pr, sprk); /* = 1 mod pr */
+  return ZV_iseven(sprk_log_prk1(nf, x, sprk));
 }
-static int
-psquare2nf(GEN nf,GEN x,GEN pr,GEN sprk)
+
+/*
+For z in nf, z != 0.
+quadratic characters modulo the prime ideal pr in nf.
+pr output by nfmodprinit
+pstar output by idealstar (only for p | 2).
+For p odd, the output is a vector [v,c]*Mod(1,2), where
+v = valuation(z,pr)
+c = !issquare( z/pr^v mod pr)
+For p | 2, the output is similar, with a longer sequence of 0,1 for c.
+*/
+
+GEN
+nf_quadchar_modpr(GEN nf, GEN z, GEN modpr, GEN pstar)
 {
   pari_sp av = avma;
-  long v = psquare2nf_i(nf,x,pr,sprk);
-  avma = av; return v;
+  GEN pr = modpr_get_pr(modpr);
+  GEN v = stoi(nfvalrem(nf, z, pr, &z));
+  if( equaliu(pr_get_p(pr),2))
+  {
+    GEN c = ideallog(nf, z, pstar);
+    return gerepilecopy(av, shallowconcat(v, shallowtrans(c)));
+  }
+  else
+  {
+    GEN c = quad_char(nf, z, modpr)==1? gen_0: gen_1;
+    return gerepilecopy(av, mkvec2(v,c));
+  }
 }
 
 /* pr above an odd prime */
@@ -204,46 +239,59 @@ lemma6nf(GEN nf, GEN T, GEN pr, long nu, GEN x, GEN modpr)
 
   la = nfval(nf,gx,pr);
   gpx = nfpoleval(nf, RgX_deriv(T), x);
-  mu = gequal0(gpx)? la+nu+1: nfval(nf,gpx,pr);
-  avma = av;
+  mu = gequal0(gpx)? la+nu+1 /* +oo */: nfval(nf,gpx,pr);
+  set_avma(av);
   if (la > (mu<<1)) return 1;
-  if (la >= (nu<<1)  && mu >= nu) return 0;
+  if (la >= (nu<<1) && mu >= nu) return 0;
   return -1;
 }
 /* pr above 2 */
 static long
 lemma7nf(GEN nf, GEN T, GEN pr, long nu, GEN x, GEN sprk)
 {
-  long res, la, mu, q;
-  GEN gpx, gx = nfpoleval(nf, T, x);
-
-  if (psquare2nf(nf,gx,pr,sprk)) return 1;
+  long i, res, la, mu, q, e, v;
+  GEN M, y, gpx, loggx = NULL, gx = nfpoleval(nf, T, x);
 
+  la = nfvalrem(nf, gx, pr, &gx); /* gx /= pi^la, pi a pr-uniformizer */
+  if (la == LONG_MAX) return 1;
+  if (!odd(la))
+  {
+    gx = to_principal_unit(nf, gx, pr, sprk); /* now 1 mod pr */
+    loggx = sprk_log_prk1(nf, gx, sprk); /* cheap */
+    if (ZV_iseven(loggx)) return 1;
+  }
   gpx = nfpoleval(nf, RgX_deriv(T), x);
-  /* gx /= pi^la, pi a pr-uniformizer */
-  la = nfvalrem(nf, gx, pr, &gx);
-  mu = gequal0(gpx)? la+nu+1: nfval(nf,gpx,pr);
+  mu = gequal0(gpx)? la+nu+1 /* oo */: nfval(nf,gpx,pr);
 
-  if (la > (mu<<1)) return 1;
+  if (la > (mu << 1)) return 1;
   if (nu > mu)
   {
-    if (la&1) return -1;
+    if (odd(la)) return -1;
     q = mu+nu-la; res = 1;
   }
   else
   {
-    long nu2 = nu<<1;
-    if (la >= nu2) return 0;
+    q = (nu << 1) - la;
+    if (q <= 0) return 0;
     if (odd(la)) return -1;
-    q = nu2-la; res = 0;
+    res = 0;
   }
-  if (q > pr_get_e(pr)<<1)  return -1;
+  /* la even */
+  e = pr_get_e(pr);
+  if (q > e << 1)  return -1;
   if (q == 1) return res;
 
-  /* is gx a square mod pi^q ? FIXME : highly inefficient */
-  sprk = zlog_pr_init(nf, pr, q);
-  if (!check2(nf, gx, sprk)) res = -1;
-  return res;
+  /* gx = 1 mod pr; square mod pi^q ? */
+  v = nfvalrem(nf, nfadd(nf, gx, gen_m1), pr, &y);
+  if (v >= q) return res;
+  /* 1 + pi^v y = (1 + pi^vz z)^2 mod pr^q ? v < q <= 2e => vz < e => vz = vy/2
+   * => y = x^2 mod pr^(min(q-v, e+v/2)), (y,pr) = 1 */
+  if (odd(v)) return -1;
+  /* e > 1 */
+  M = cgetg(2*e+1 - q + 1, t_VEC);
+  for (i = q+1; i <= 2*e+1; i++) gel(M, i-q) = sprk_log_gen_pr(nf, sprk, i);
+  M = ZM_hnfmodid(shallowconcat1(M), gen_2);
+  return hnf_solve(M,loggx)? res: -1;
 }
 /* zinit either a sprk (pr | 2) or a modpr structure (pr | p odd).
    pnu = pi^nu, pi a uniformizer */
@@ -256,7 +304,7 @@ zpsolnf(GEN nf,GEN T,GEN pr,long nu,GEN pnu,GEN x0,GEN repr,GEN zinit)
 
   res = typ(zinit) == t_VEC? lemma7nf(nf,T,pr,nu,x0,zinit)
                            : lemma6nf(nf,T,pr,nu,x0,zinit);
-  avma = av;
+  set_avma(av);
   if (res== 1) return 1;
   if (res==-1) return 0;
   pnup = nfmul(nf, pnu, pr_get_gen(pr));
@@ -264,9 +312,9 @@ zpsolnf(GEN nf,GEN T,GEN pr,long nu,GEN pnu,GEN x0,GEN repr,GEN zinit)
   for (i=1; i<lg(repr); i++)
   {
     GEN x = nfadd(nf, x0, nfmul(nf,pnu,gel(repr,i)));
-    if (zpsolnf(nf,T,pr,nu,pnup,x,repr,zinit)) { avma = av; return 1; }
+    if (zpsolnf(nf,T,pr,nu,pnup,x,repr,zinit)) return gc_long(av,1);
   }
-  avma = av; return 0;
+  return gc_long(av,0);
 }
 
 /* Let y = copy(x); y[k] := j; return y */
@@ -295,18 +343,17 @@ repres(GEN nf, GEN pr)
 /* = 1 if equation y^2 = z^deg(T) * T(x/z) has a pr-adic rational solution
  * (possibly (1,y,0) = oo), 0 otherwise.
  * coeffs of T are algebraic integers in nf */
-long
-nf_hyperell_locally_soluble(GEN nf,GEN T,GEN pr)
+static long
+locally_soluble(GEN nf,GEN T,GEN pr)
 {
-  GEN repr, zinit, p1;
-  pari_sp av = avma;
+  GEN repr, zinit;
 
   if (typ(T)!=t_POL) pari_err_TYPE("nf_hyperell_locally_soluble",T);
   if (gequal0(T)) return 1;
   checkprid(pr); nf = checknf(nf);
   if (absequaliu(pr_get_p(pr), 2))
   { /* tough case */
-    zinit = zlog_pr_init(nf, pr, 1+2*pr_get_e(pr));
+    zinit = log_prk_init(nf, pr, 1+2*pr_get_e(pr), NULL);
     if (psquare2nf(nf,constant_coeff(T),pr,zinit)) return 1;
     if (psquare2nf(nf, leading_coeff(T),pr,zinit)) return 1;
   }
@@ -316,12 +363,16 @@ nf_hyperell_locally_soluble(GEN nf,GEN T,GEN pr)
     if (psquarenf(nf,constant_coeff(T),pr,zinit)) return 1;
     if (psquarenf(nf, leading_coeff(T),pr,zinit)) return 1;
   }
-  repr = repres(nf,pr);
-  if (zpsolnf(nf,T,pr,0,gen_1,gen_0,repr,zinit)) { avma=av; return 1; }
-  p1 = pr_get_gen(pr);
-  if (zpsolnf(nf,RgX_recip_shallow(T),pr,1,p1,gen_0,repr,zinit)) { avma=av; return 1; }
-
-  avma = av; return 0;
+  repr = repres(nf,pr); /* FIXME: inefficient if Npr is large */
+  return zpsolnf(nf, T, pr, 0, gen_1, gen_0, repr, zinit) ||
+         zpsolnf(nf, RgX_recip_shallow(T), pr, 1, pr_get_gen(pr),
+                 gen_0, repr, zinit);
+}
+long
+nf_hyperell_locally_soluble(GEN nf,GEN T,GEN pr)
+{
+  pari_sp av = avma;
+  return gc_long(av, locally_soluble(nf, T, pr));
 }
 
 /* return a * denom(a)^2, as an 'liftalg' */
@@ -347,25 +398,21 @@ static long
 hilb2nf(GEN nf,GEN a,GEN b,GEN p)
 {
   pari_sp av = avma;
-  long rep;
   GEN pol;
-
   a = den_remove(nf, a);
   b = den_remove(nf, b);
   pol = mkpoln(3, a, gen_0, b);
   /* varn(nf.pol) = 0, pol is not a valid GEN  [as in Pol([x,x], x)].
    * But it is only used as a placeholder, hence it is not a problem */
-
-  rep = nf_hyperell_locally_soluble(nf,pol,p)? 1: -1;
-  avma = av; return rep;
+  return gc_long(av, nf_hyperell_locally_soluble(nf,pol,p)? 1: -1);
 }
 
-/* local quadratic Hilbert symbol (a,b)_pr, for a,b (non-zero) in nf */
+/* local quadratic Hilbert symbol (a,b)_pr, for a,b (nonzero) in nf */
 static long
 nfhilbertp(GEN nf, GEN a, GEN b, GEN pr)
 {
   GEN t;
-  long va, vb, rep;
+  long va, vb;
   pari_sp av = avma;
 
   if (absequaliu(pr_get_p(pr), 2)) return hilb2nf(nf,a,b,pr);
@@ -373,33 +420,29 @@ nfhilbertp(GEN nf, GEN a, GEN b, GEN pr)
   /* pr not above 2, compute t = tame symbol */
   va = nfval(nf,a,pr);
   vb = nfval(nf,b,pr);
-  if (!odd(va) && !odd(vb)) { avma = av; return 1; }
+  if (!odd(va) && !odd(vb)) return gc_long(av,1);
   /* Trick: pretend the exponent is 2, result is OK up to squares ! */
   t = famat_makecoprime(nf, mkvec2(a,b), mkvec2s(vb, -va),
                         pr, pr_hnf(nf, pr), gen_2);
-  if (typ(t) == t_INT) {
-    if (odd(va) && odd(vb)) t = negi(t);
-    /* t = (-1)^(v(a)v(b)) a^v(b) b^(-v(a)) */
-    rep = Z_quad_char(t, pr);
+  /* quad. symbol is image of t = (-1)^(v(a)v(b)) a^v(b) b^(-v(a))
+   * by the quadratic character  */
+  switch(typ(t))
+  {
+    default: /* t_COL */
+      if (!ZV_isscalar(t)) break;
+      t = gel(t,1); /* fall through */
+    case t_INT:
+      if (odd(va) && odd(vb)) t = negi(t);
+      return gc_long(av,  Z_quad_char(t, pr));
   }
-  else if (ZV_isscalar(t)) {
-    t = gel(t,1);
-    if (odd(va) && odd(vb)) t = negi(t);
-    /* t = (-1)^(v(a)v(b)) a^v(b) b^(-v(a)) */
-    rep = Z_quad_char(t, pr);
-  } else {
-    if (odd(va) && odd(vb)) t = ZC_neg(t);
-    /* t = (-1)^(v(a)v(b)) a^v(b) b^(-v(a)) */
-    rep = quad_char(nf, t, pr);
-  }
-  /* quad. symbol is image of t by the quadratic character  */
-  avma = av; return rep;
+  if (odd(va) && odd(vb)) t = ZC_neg(t);
+  return gc_long(av, quad_char(nf, t, pr));
 }
 
 /* Global quadratic Hilbert symbol (a,b):
  *  =  1 if X^2 - aY^2 - bZ^2 has a point in projective plane
  *  = -1 otherwise
- * a, b should be non-zero */
+ * a, b should be nonzero */
 long
 nfhilbert(GEN nf, GEN a, GEN b)
 {
@@ -418,7 +461,7 @@ nfhilbert(GEN nf, GEN a, GEN b)
     {
       if (DEBUGLEVEL>3)
         err_printf("nfhilbert not soluble at real place %ld\n",i);
-      avma = av; return -1;
+      return gc_long(av,-1);
     }
 
   /* local solutions in finite completions ? (pr | 2ab)
@@ -435,9 +478,9 @@ nfhilbert(GEN nf, GEN a, GEN b)
     {
       if (DEBUGLEVEL>3)
         err_printf("nfhilbert not soluble at finite place %Ps\n",S[i]);
-      avma = av; return -1;
+      return gc_long(av,-1);
     }
-  avma = av; return 1;
+  return gc_long(av,1);
 }
 
 long
@@ -453,196 +496,6 @@ nfhilbert0(GEN nf,GEN a,GEN b,GEN p)
   return nfhilbert(nf,a,b);
 }
 
-/* S a list of prime ideal in idealprimedec format. Return res:
- * res[1] = generators of (S-units / units), as polynomials
- * res[2] = [perm, HB, den], for bnfissunit
- * res[3] = [] (was: log. embeddings of res[1])
- * res[4] = S-regulator ( = R * det(res[2]) * \prod log(Norm(S[i])))
- * res[5] = S class group
- * res[6] = S */
-GEN
-bnfsunit0(GEN bnf, GEN S, long flag, long prec)
-{
-  pari_sp av = avma;
-  long i,j,ls;
-  GEN p1,nf,gen,M,U,H;
-  GEN sunit,card,sreg,res,pow;
-
-  if (!is_vec_t(typ(S))) pari_err_TYPE("bnfsunit",S);
-  bnf = checkbnf(bnf);
-  nf = bnf_get_nf(bnf);
-  gen = bnf_get_gen(bnf);
-
-  sreg = bnf_get_reg(bnf);
-  res=cgetg(7,t_VEC);
-  gel(res,1) = gel(res,2) = gel(res,3) = cgetg(1,t_VEC);
-  gel(res,4) = sreg;
-  gel(res,5) = bnf_get_clgp(bnf);
-  gel(res,6) = S; ls=lg(S);
-
- /* M = relation matrix for the S class group (in terms of the class group
-  * generators given by gen)
-  * 1) ideals in S
-  */
-  M = cgetg(ls,t_MAT);
-  for (i=1; i<ls; i++)
-  {
-    p1 = gel(S,i); checkprid(p1);
-    gel(M,i) = isprincipal(bnf,p1);
-  }
-  /* 2) relations from bnf class group */
-  M = shallowconcat(M, diagonal_shallow(bnf_get_cyc(bnf)));
-
-  /* S class group */
-  H = ZM_hnfall(M, &U, 1);
-  card = gen_1;
-  if (lg(H) > 1)
-  { /* non trivial (rare!) */
-    GEN A, u, D = ZM_snfall_i(H, &u, NULL, 1);
-    long l;
-    ZV_snf_trunc(D); l = lg(D);
-    card = ZV_prod(D);
-    A = cgetg(l, t_VEC); pow = ZM_inv(u, NULL);
-    for(i = 1; i < l; i++) gel(A,i) = idealfactorback(nf, gen, gel(pow,i), 1);
-    gel(res,5) = mkvec3(card, D, A);
-  }
-
-  /* S-units */
-  if (ls>1)
-  {
-    GEN den, Sperm, perm, dep, B, A, U1 = U;
-    long lH, lB, FLAG = flag|nf_FORCE;
-
-   /* U1 = upper left corner of U, invertible. S * U1 = principal ideals
-    * whose generators generate the S-units */
-    setlg(U1,ls); p1 = cgetg(ls, t_MAT); /* p1 is junk for mathnfspec */
-    for (i=1; i<ls; i++) { setlg(U1[i],ls); gel(p1,i) = cgetg(1,t_COL); }
-    H = mathnfspec(U1,&perm,&dep,&B,&p1);
-    lH = lg(H);
-    lB = lg(B);
-    if (lg(dep) > 1 && lgcols(dep) > 1) pari_err_BUG("bnfsunit");
-   /*                   [ H B  ]            [ H^-1   - H^-1 B ]
-    * perm o HNF(U1) =  [ 0 Id ], inverse = [  0         Id   ]
-    * (permute the rows)
-    * S * HNF(U1) = _integral_ generators for S-units  = sunit */
-    Sperm = cgetg(ls, t_VEC); sunit = cgetg(ls, t_VEC);
-    for (i=1; i<ls; i++) Sperm[i] = S[perm[i]]; /* S o perm */
-
-    setlg(Sperm, lH);
-    for (i=1; i<lH; i++)
-    {
-      GEN v = isprincipalfact(bnf, NULL,Sperm,gel(H,i), FLAG);
-      v = gel(v,2); if (flag == nf_GEN) v = nf_to_scalar_or_alg(nf, v);
-      gel(sunit,i) = v;
-    }
-    for (j=1; j<lB; j++,i++)
-    {
-      GEN v = isprincipalfact(bnf, gel(Sperm,i),Sperm,gel(B,j),FLAG);
-      v = gel(v,2); if (flag == nf_GEN) v = nf_to_scalar_or_alg(nf, v);
-      gel(sunit,i) = v;
-   }
-    H = ZM_inv(H,&den);
-    A = shallowconcat(H, ZM_neg(ZM_mul(H,B))); /* top part of inverse * den */
-    /* HNF in split form perm + (H B) [0 Id missing] */
-    gel(res,1) = sunit;
-    gel(res,2) = mkvec3(perm,A,den);
-  }
-
-  /* S-regulator */
-  sreg = mpmul(sreg,card);
-  for (i=1; i<ls; i++)
-  {
-    GEN p = pr_get_p( gel(S,i) );
-    sreg = mpmul(sreg, logr_abs(itor(p,prec)));
-  }
-  gel(res,4) = sreg;
-  return gerepilecopy(av,res);
-}
-GEN
-bnfsunit(GEN bnf,GEN S,long prec) { return bnfsunit0(bnf,S,nf_GEN,prec); }
-
-static GEN
-make_unit(GEN nf, GEN bnfS, GEN *px)
-{
-  long lB, cH, i, ls;
-  GEN den, gen, S, v, p1, xp, xb, N, N0, HB, perm;
-
-  if (gequal0(*px)) return NULL;
-  S = gel(bnfS,6); ls = lg(S);
-  if (ls==1) return cgetg(1, t_COL);
-
-  xb = nf_to_scalar_or_basis(nf,*px);
-  switch(typ(xb))
-  {
-    case t_INT:  N = xb; break;
-    case t_FRAC: N = mulii(gel(xb,1),gel(xb,2)); break;
-    default: { GEN d = Q_denom(xb); N = mulii(idealnorm(nf,gmul(*px,d)), d); }
-  } /* relevant primes divide N */
-  if (is_pm1(N)) return zerocol(ls -1);
-
-  p1 = gel(bnfS,2);
-  perm = gel(p1,1);
-  HB   = gel(p1,2);
-  den  = gel(p1,3);
-  cH = nbrows(HB);
-  lB = lg(HB) - cH;
-  v = zero_zv(ls-1);
-  N0 = N;
-  for (i=1; i<ls; i++)
-  {
-    GEN P = gel(S,i), p = pr_get_p(P);
-    if (dvdii(N, p))
-    {
-      v[i] = nfval(nf,xb,P);
-      (void)Z_pvalrem(N0, p, &N0);
-    }
-  }
-  if (!is_pm1(N0)) return NULL;
-  /* here, x = S v */
-  p1 = vecsmallpermute(v, perm);
-  v = ZM_zc_mul(HB, p1);
-  for (i=1; i<=cH; i++)
-  {
-    GEN r, w = dvmdii(gel(v,i), den, &r);
-    if (r != gen_0) return NULL;
-    gel(v,i) = w;
-  }
-  p1 += cH; p1[0] = evaltyp(t_VECSMALL) | evallg(lB);
-  v = shallowconcat(v, zc_to_ZC(p1)); /* append bottom of p1 (= [0 Id] part) */
-
-  gen = gel(bnfS,1);
-  xp = trivial_fact();
-  for (i=1; i<ls; i++)
-  {
-    GEN e = gel(v,i);
-    if (!signe(e)) continue;
-    xp = famat_mulpow_shallow(xp, gel(gen,i), negi(e));
-  }
-  if (lgcols(xp) != 1) *px = famat_mulpow_shallow(xp, xb, gen_1);
-  return v;
-}
-
-/* Analog to bnfisunit, for S-units. Let v the result
- * If x not an S-unit, v = []~, else
- * x = \prod_{i=0}^r e_i^v[i] * prod{i=r+1}^{r+s} s_i^v[i]
- * where the e_i are the field units (cf bnfisunit), and the s_i are
- * the S-units computed by bnfsunit (in the same order) */
-GEN
-bnfissunit(GEN bnf,GEN bnfS,GEN x)
-{
-  pari_sp av = avma;
-  GEN v, w, nf;
-
-  bnf = checkbnf(bnf);
-  nf = bnf_get_nf(bnf);
-  if (typ(bnfS)!=t_VEC || lg(bnfS)!=7) pari_err_TYPE("bnfissunit",bnfS);
-  x = nf_to_scalar_or_alg(nf,x);
-  v = NULL;
-  if ( (w = make_unit(nf, bnfS, &x)) ) v = bnfisunit(bnf, x);
-  if (!v || lg(v) == 1) { avma = av; return cgetg(1,t_COL); }
-  return gerepileupto(av, gconcat(v, w));
-}
-
 static void
 p_append(GEN p, hashtable *H, hashtable *H2)
 {
@@ -700,10 +553,10 @@ hash_create_INT(ulong s)
 { return hash_create(s, (ulong(*)(void*))&hash_GEN,
                         (int(*)(void*,void*))&equalii, 1); }
 GEN
-rnfisnorminit(GEN T, GEN relpol, int galois)
+rnfisnorminit(GEN T, GEN R, int galois)
 {
   pari_sp av = avma;
-  long i, l, drel;
+  long i, l, dR;
   GEN S, gen, cyc, bnf, nf, nfabs, rnfeq, bnfabs, k, polabs;
   GEN y = cgetg(9, t_VEC);
   hashtable *H = hash_create_INT(100UL);
@@ -713,18 +566,18 @@ rnfisnorminit(GEN T, GEN relpol, int galois)
   if (!bnf) bnf = Buchall(nf? nf: T, nf_FORCE, DEFAULTPREC);
   if (!nf) nf = bnf_get_nf(bnf);
 
-  relpol = get_bnfpol(relpol, &bnfabs, &nfabs);
-  if (!gequal1(leading_coeff(relpol))) pari_err_IMPL("non monic relative equation");
-  drel = degpol(relpol);
-  if (drel <= 2) galois = 1;
+  R = get_bnfpol(R, &bnfabs, &nfabs);
+  if (!gequal1(leading_coeff(R))) pari_err_IMPL("non monic relative equation");
+  dR = degpol(R);
+  if (dR <= 2) galois = 1;
 
-  relpol = RgX_nffix("rnfisnorminit", T, relpol, 1);
+  R = RgX_nffix("rnfisnorminit", T, R, 1);
   if (nf_get_degree(nf) == 1) /* over Q */
-    rnfeq = mkvec5(relpol,gen_0,gen_0,T,relpol);
+    rnfeq = mkvec5(R,gen_0,gen_0,T,R);
   else if (galois == 2) /* needs eltup+abstorel */
-    rnfeq = nf_rnfeq(nf, relpol);
+    rnfeq = nf_rnfeq(nf, R);
   else /* needs abstorel */
-    rnfeq = nf_rnfeqsimple(nf, relpol);
+    rnfeq = nf_rnfeqsimple(nf, R);
   polabs = gel(rnfeq,1);
   k = gel(rnfeq,3);
   if (!bnfabs || !gequal0(k))
@@ -733,7 +586,7 @@ rnfisnorminit(GEN T, GEN relpol, int galois)
 
   if (galois == 2)
   {
-    GEN P = polabs==relpol? leafcopy(relpol): nfX_eltup(nf, rnfeq, relpol);
+    GEN P = polabs==R? leafcopy(R): nfX_eltup(nf, rnfeq, R);
     setvarn(P, fetch_var_higher());
     galois = !!nfroots_if_split(&nfabs, P);
     (void)delete_var();
@@ -744,18 +597,18 @@ rnfisnorminit(GEN T, GEN relpol, int galois)
   for(i=1; i<l; i++)
   {
     GEN g = gel(gen,i);
-    if (ugcdiu(gel(cyc,i), drel) == 1) break;
+    if (ugcdiu(gel(cyc,i), dR) == 1) break;
     Zfa_append(gcoeff(g,1,1), H, NULL);
   }
   if (!galois)
   {
-    GEN Ndiscrel = diviiexact(nf_get_disc(nfabs), powiu(nf_get_disc(nf), drel));
+    GEN Ndiscrel = diviiexact(nf_get_disc(nfabs), powiu(nf_get_disc(nf), dR));
     Zfa_append(Ndiscrel, H, NULL);
   }
   S = hash_keys(H); settyp(S,t_VEC);
   gel(y,1) = bnf;
   gel(y,2) = bnfabs;
-  gel(y,3) = relpol;
+  gel(y,3) = R;
   gel(y,4) = rnfeq;
   gel(y,5) = S;
   gel(y,6) = nf_pV_to_prV(nf, S);
@@ -774,9 +627,9 @@ GEN
 rnfisnorm(GEN T, GEN x, long flag)
 {
   pari_sp av = avma;
-  GEN bnf, rel, relpol, rnfeq, nfpol;
+  GEN bnf, rel, R, rnfeq, nfpol;
   GEN nf, aux, H, U, Y, M, A, bnfS, sunitrel, futu, S, S1, S2, Sx;
-  long L, i, drel, itu;
+  long L, i, itu;
   hashtable *H0, *H2;
   if (typ(T) != t_VEC || lg(T) != 9)
     pari_err_TYPE("rnfisnorm [please apply rnfisnorminit()]", T);
@@ -786,12 +639,12 @@ rnfisnorm(GEN T, GEN x, long flag)
   rel = checkbnf(rel);
   nf = bnf_get_nf(bnf);
   x = nf_to_scalar_or_alg(nf,x);
-  if (gequal0(x)) { avma = av; return mkvec2(gen_0, gen_1); }
-  if (gequal1(x)) { avma = av; return mkvec2(gen_1, gen_1); }
-  relpol = gel(T,3);
+  if (gequal0(x)) { set_avma(av); return mkvec2(gen_0, gen_1); }
+  if (gequal1(x)) { set_avma(av); return mkvec2(gen_1, gen_1); }
+  R = gel(T,3);
   rnfeq = gel(T,4);
-  drel = degpol(relpol);
-  if (gequalm1(x) && odd(drel)) { avma = av; return mkvec2(gen_m1, gen_1); }
+  if (gequalm1(x) && odd(degpol(R)))
+  { set_avma(av); return mkvec2(gen_m1, gen_1); }
 
   /* build set T of ideals involved in the solutions */
   nfpol = nf_get_pol(nf);
@@ -802,8 +655,6 @@ rnfisnorm(GEN T, GEN x, long flag)
   for (i = 1; i < L; i++) p_append(gel(S,i),H0,NULL);
   S1 = gel(T,6);
   S2 = gel(T,7);
-  if (flag && !gequal0(gel(T,8)))
-    pari_warn(warner,"useless flag in rnfisnorm: the extension is Galois");
   if (flag > 0)
   {
     forprime_t T;
