@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -341,7 +342,7 @@ ncharvecexpo(GEN G, GEN nchi)
         *++e = Fl_add(*++u3, x, ord); v[*d] = *e;
       }
   }
-  avma = av; return v;
+  set_avma(av); return v;
 }
 
 /*****************************************************************************/
@@ -375,7 +376,7 @@ coprimes_zv(ulong N)
     ulong p = P[i], j;
     for (j = p; j <= N; j += p) v[j] = 0;
   }
-  avma = av; return v;
+  set_avma(av); return v;
 }
 /* cf zv_cyc_minimal: return k such that g*k is minimal (wrt lex) */
 long
@@ -407,7 +408,7 @@ zv_cyc_minimize(GEN cyc, GEN g, GEN coprime)
     gk = vecmoduu(gk, cyc);
     if (vecsmall_lexcmp(gk, best) < 0) { best = gk; bestk = k; }
   }
-  avma = av; return bestk == 1? k0: Fl_mul(k0, bestk, o);
+  return gc_long(av, bestk == 1? k0: (long) Fl_mul(k0, bestk, o));
 }
 /* g of order o in abelian group G attached to cyc. Is g a minimal generator
  * [wrt lex order] of the cyclic subgroup it generates;
@@ -431,16 +432,16 @@ zv_cyc_minimal(GEN cyc, GEN g, GEN coprime)
   g   = vecslice(g,  k,l-1);
   e = cyc[1];
   /* find k in (Z/e)^* such that g*k mod cyc is lexicographically minimal,
-   * k = 1 mod d to fix the first non-zero entry */
+   * k = 1 mod d to fix the first nonzero entry */
   gd = Flv_Fl_mul(g, d, e); maxi = e/ugcd(d,e);
   for (gk = g, k = d+1, i = 1; i < maxi; i++, k += d)
   {
     long ko = k % o;
     gk = Flv_add(gk, gd, e); if (!coprime[ko]) continue;
     gk = vecmoduu(gk, cyc);
-    if (vecsmall_lexcmp(gk, g) < 0) { avma = av; return 0; }
+    if (vecsmall_lexcmp(gk, g) < 0) return gc_long(av,0);
   }
-  avma = av; return 1;
+  return gc_long(av,1);
 }
 
 static GEN
@@ -503,7 +504,7 @@ chargalois(GEN G, GEN ORD)
   long maxord, i, l;
   GEN v, cyc = (typ(G) == t_VEC && RgV_is_ZVpos(G))? G: member_cyc(G);
   if (lg(cyc) == 1) retmkvec(cgetg(1,t_VEC));
-  maxord = itou(gel(cyc,1));
+  maxord = itou(cyc_get_expo(cyc));
   if (ORD && gequal0(ORD)) ORD = NULL;
   if (ORD)
     switch(typ(ORD))
@@ -551,7 +552,7 @@ znstar0(GEN N, long flag)
   if (!signe(N))
   {
     if (flag) pari_err_IMPL("znstar(0,1)");
-    avma = av;
+    set_avma(av);
     retmkvec3(gen_2, mkvec(gen_2), mkvec(gen_m1));
   }
   N = absi_shallow(N);
@@ -774,7 +775,7 @@ znconreyfromchar(GEN bid, GEN chi)
 }
 
 /* discrete log on canonical "primitive root" generators
- * Allow log(x) instead of x [usual discrete log on bid's generators */
+ * Allow log(x) instead of x [usual discrete log on bid's generators] */
 GEN
 znconreylog(GEN bid, GEN x)
 {
@@ -956,7 +957,7 @@ znconreyconductor(GEN bid, GEN chi, GEN *pm)
   E = gel(F,2);
   if (l == 1)
   {
-    avma = av;
+    set_avma(av);
     if (pm) *pm = cgetg(1,t_COL);
     if (lg(P) == 1) return gen_1;
     retmkvec2(gen_1, trivial_fact());
@@ -1079,7 +1080,7 @@ zncharinduce(GEN G, GEN chi, GEN N)
   e2 = (E[1] >= 3 && absequaliu(gel(P,1),2)); /* 2 generators at 2 mod N */
   if (ZV_equal0(chi))
   {
-    avma = av;
+    set_avma(av);
     return equali1(N)? cgetg(1, t_COL): zerocol(l+e2 - 1);
   }
 
@@ -1188,7 +1189,7 @@ znchartokronecker(GEN G, GEN chi, long flag)
   s = zncharisodd(G, chi)? -1: 1;
   if (typ(chi) != t_COL) chi = znconreylog(G, chi);
   o = zncharorder(G, chi);
-  if (abscmpiu(o,2) > 0) { avma = av; return gen_0; }
+  if (abscmpiu(o,2) > 0) { set_avma(av); return gen_0; }
   F = znconreyconductor(G, chi, NULL);
   if (typ(F) == t_INT)
   {
