@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -447,7 +448,7 @@ shallowmatconcat(GEN v)
       return M;
     default:
       pari_err_TYPE("shallowmatconcat", v);
-      return NULL;
+      return NULL;/*LCOV_EXCL_LINE*/
   }
 }
 GEN
