@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -14,7 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
-#define dbg_mode() if (DEBUGLEVEL >= 2)
+#define dbg_mode()  if (DEBUGLEVEL >= 2)
+#define dbg_mode1() if (DEBUGLEVEL >= 3)
 
 #define ANSI_RED     "\x1b[31m"
 #define ANSI_GREEN   "\x1b[32m"
@@ -56,13 +58,17 @@ zv_binsearch0(void *E, long (*f)(void* E, GEN x), GEN x)
 }
 
 INLINE long
-timer_record(GEN* X0, const char* Xx, pari_timer* ti)
+time_record(GEN* X0, const char* Xx, long t)
 {
-  long t = timer_delay(ti), i = Xx[0]-'A'+1, j = Xx[1]-'1'+1;
+  long i = Xx[0]-'A'+1, j = Xx[1]-'1'+1;
   umael3(*X0, 1, i, j) += t;
   umael3(*X0, 2, i, j) ++; return t;
 }
 
+INLINE long
+timer_record(GEN* X0, const char* Xx, pari_timer* ti)
+{ return time_record(X0, Xx, timer_delay(ti)); }
+
 INLINE long
 FpJ_is_inf(GEN P) { return signe(gel(P, 3)) == 0; }
 
@@ -84,7 +90,7 @@ D_get_wD(long D)
  *   pd: the degree of polclass; equal to h except when bi is a double
  *       eta-quotient w_p,q with p|D and q|D, where pd = h/2.
  * Dfac: the prime factorization of D; we have D = q0 q1* q2* ... qn*
- *       where q0 = 1, 4, -4, 8, qi* = 1 mod 4 and |qi| is a prime.
+ *       where q0 = 1, -4, 8, -8, qi* = 1 mod 4 and |qi| is a prime.
  *       The factorization is a vecsmall listing the indices of the qi* as
  *       they appear in the primelist (q0 = 1 is omitted) */
 INLINE long
@@ -141,6 +147,8 @@ ecpp_param_get_primelist(GEN param) { return gmael(param, 1, 2); }
 INLINE GEN
 ecpp_param_get_disclist(GEN param) { return gmael(param, 1, 3); }
 INLINE GEN
+ecpp_param_get_primeord(GEN param) { return gmael(param, 1, 4); }
+INLINE GEN
 ecpp_param_get_primorial_vec(GEN param) { return gel(param, 2); }
 INLINE GEN
 ecpp_param_get_tune(GEN param) { return gel(param, 3); }
@@ -172,7 +180,7 @@ primorial_vec(ulong x)
    Dinfo, as discussed in the comment about Dinfo
    m, as in the theorem in ??ecpp
    q, as in the theorem in ??ecpp
-   g, a quadratic non-residue modulo N
+   g, a quadratic nonresidue modulo N
    sqrt, a list of squareroots
 */
 INLINE GEN
@@ -395,6 +403,25 @@ ecpp_disclist_init(ulong maxdisc, GEN primelist)
   return gerepilecopy(av, merge);
 }
 
+static GEN
+ecpp_primeord_init(GEN primelist, GEN disclist)
+{
+  long i, k=1, lgdisclist = lg(disclist), lprimelist = lg(primelist);
+  GEN primeord = zero_Flv(lprimelist-1);
+  for (i=1;i < lgdisclist; i++)
+  {
+    GEN Dinfo = gel(disclist, i), Dfac = Dinfo_get_Dfac(Dinfo);
+    long j, l = lg(Dfac);
+    for (j = 1; j < l; j++)
+    {
+      long ip = Dfac[j];
+      if (primeord[ip]==0)
+        primeord[ip]=k++;
+    }
+  }
+  return perm_inv(primeord);
+}
+
 /*  Input: a vector tune whose components are [maxsqrt,maxpcdg,tdivexp,expiN]
  * Output: vector param of precomputations
  *   let x =  be a component of tune then
@@ -412,8 +439,9 @@ ecpp_param_set(GEN tune, GEN x)
   GEN T = mkvecsmall3(maxsqrt, maxdisc, maxpcdg);
   GEN Plist = ecpp_primelist_init(maxsqrt);
   GEN Dlist = ecpp_disclist_init(maxdisc, Plist);
+  GEN Olist = ecpp_primeord_init(Plist, Dlist);
   GEN primorial = primorial_vec(tdivexp);
-  return gerepilecopy(av, mkvec3(mkvec3(T,Plist,Dlist), primorial, tune));
+  return gerepilecopy(av, mkvec3(mkvec4(T,Plist,Dlist,Olist), primorial, tune));
 }
 
 /* cert contains [N, t, s, a4, [x, y]] as documented in ??ecpp; the following
@@ -541,7 +569,7 @@ NqE_check(GEN N, GEN q, GEN a, GEN b, GEN s)
  * the ith element of the certificate; uses N, D, m, q, J from previous steps.
  * All computations are modulo N unless stated otherwise */
 
-/* g is a quadratic and cubic non-residue modulo N */
+/* g is a quadratic and cubic nonresidue modulo N */
 static GEN
 j0_find_g(GEN N)
 {
@@ -617,10 +645,8 @@ realgenusfield(GEN Dfac, GEN sq, GEN p)
 static GEN
 FpX_classtower_oneroot(GEN P, GEN Dfac, GEN sq, GEN p)
 {
-  pari_timer ti;
   pari_sp av = avma;
   GEN C;
-  dbg_mode() timer_start(&ti);
   if (degpol(P) > 1)
   {
     GEN N = NULL, V = realgenusfield(Dfac, sq, p), v = gel(V,1), R = gel(V,2);
@@ -637,18 +663,45 @@ FpX_classtower_oneroot(GEN P, GEN Dfac, GEN sq, GEN p)
       P = liftpol_shallow(gmael(nffactor(N,P),1,1));
     }
     if (N)
-    {
       P = FpXY_evalx(Q_primpart(P), R, p);
-      dbg_mode() err_printf(ANSI_BRIGHT_GREEN " %6ld[%ld]" ANSI_RESET,
-                            timer_delay(&ti), l-1);
-    } else
-      dbg_mode() err_printf("          ");
   }
   C = FpX_oneroot_split(P, p);
-  dbg_mode() err_printf(" %6ld", timer_delay(&ti));
   return gerepileupto(av, C);
 }
 
+GEN
+ecpp_step2_worker(GEN S, GEN HD, GEN primelist)
+{
+  pari_sp av = avma;
+  pari_timer ti;
+  GEN J, t, s, EP, rt, res;
+  GEN N = NDmqg_get_N(S), Dinfo = NDmqg_get_Dinfo(S);
+  GEN m = NDmqg_get_m(S), q = NDmqg_get_q(S);
+  GEN g = NDmqg_get_g(S), sq = NDmqg_get_sqrt(S);
+  long D = Dinfo_get_D(Dinfo), inv = Dinfo_get_bi(Dinfo);
+  GEN Dfacp = Dfac_to_p(Dinfo_get_Dfac(Dinfo), primelist);
+  long C2 = 0, C3 = 0, D1 = 0;
+  GEN c = getrand();
+  setrand(gen_1); /* for reproducibility */
+  /* C2: Find a root modulo N of polclass(D,inv) */
+  dbg_mode() timer_start(&ti);
+  rt = FpX_classtower_oneroot(HD, Dfacp, sq, N);
+  dbg_mode() C2 = timer_delay(&ti);
+  /* C3: Convert root from previous step into the appropriate j-invariant */
+  J = Fp_modinv_to_j(rt, inv, N); /* root of polclass(D) */
+  dbg_mode() C3 = timer_delay(&ti);
+  /* D1: Find an elliptic curve E with a point P satisfying the theorem */
+  s = diviiexact(m, q);
+  EP = find_EP(N, D, q, g, J, s);
+  dbg_mode() D1 = timer_delay(&ti);
+
+  /* D2: Compute for t and s */
+  t = subii(addiu(N, 1), m); /* t = N+1-m */
+  res = mkvec2(mkvec5(N, t, s, gel(EP,1), gel(EP,2)),mkvecsmall3(C2,C3,D1));
+  setrand(c);
+  return gerepilecopy(av, res);
+}
+
 /* This uses [N, D, m, q] from step 1 to find the appropriate j-invariants
  * to be used in step 3. Step 2 is divided into substeps 2a, 2b, 2c */
 static GEN
@@ -659,73 +712,108 @@ ecpp_step2(GEN step1, GEN *X0, GEN primelist)
   GEN perm = gen_indexsort(step1, NULL, &sort_NDmq_by_D);
   GEN step2 = cgetg(lg(step1), t_VEC);
   GEN HD = NULL, db = polmodular_db_init_allinv();
-
-  for (j = 1; j < lg(step2); j++)
+  long ls = lg(step2), pending = 0;
+  GEN worker;
+  struct pari_mt pt;
+  GEN Hi = cgetg(ls, t_VEC);
+  for (j = 1; j < ls; j++)
   {
     long i = uel(perm, j);
-    GEN J, t, s, EP, rt, S = gel(step1, i);
-    GEN N = NDmqg_get_N(S), Dinfo = NDmqg_get_Dinfo(S);
-    GEN m = NDmqg_get_m(S), q = NDmqg_get_q(S);
-    GEN g = NDmqg_get_g(S), sq = NDmqg_get_sqrt(S);
+    GEN S = gel(step1, i);
+    GEN Dinfo = NDmqg_get_Dinfo(S);
     long D = Dinfo_get_D(Dinfo), inv = Dinfo_get_bi(Dinfo);
-    GEN Dfacp = Dfac_to_p(Dinfo_get_Dfac(Dinfo), primelist);
-
     /* C1: Find the appropriate class polynomial modulo N */
     dbg_mode() timer_start(&ti);
     if (D != Dprev) HD = D_polclass(D, inv, &db);
     dbg_mode() {
+      GEN N = NDmqg_get_N(S);
       long tt = timer_record(X0, "C1", &ti);
       err_printf(ANSI_BRIGHT_GREEN "\n[ %3d | %4ld bits]" ANSI_RESET, i, expi(N));
       err_printf(ANSI_GREEN " D = %8ld poldeg = %4ld" ANSI_RESET, D, degpol(HD));
       if (D == Dprev) err_printf(" %6ld", tt);
       else err_printf(ANSI_BRIGHT_WHITE " %6ld" ANSI_RESET, tt);
     }
-    /* C2: Find a root modulo N of polclass(D,inv) */
-    dbg_mode() timer_start(&ti);
-    rt = FpX_classtower_oneroot(HD, Dfacp, sq, N);
-    dbg_mode() err_printf(" %6ld", timer_record(X0, "C2", &ti));
-
-    /* C3: Convert root from previous step into the appropriate j-invariant */
-    dbg_mode() timer_start(&ti);
-    J = Fp_modinv_to_j(rt, inv, N); /* root of polclass(D) */
-    dbg_mode() err_printf(" %6ld", timer_record(X0, "C3", &ti));
-
-    /* D1: Find an elliptic curve E with a point P satisfying the theorem */
-    dbg_mode() timer_start(&ti);
-    s = diviiexact(m, q);
-    EP = find_EP(N, D, q, g, J, s);
-    dbg_mode() err_printf(" %6ld", timer_record(X0, "D1", &ti));
-
-    /* D2: Compute for t and s */
-    t = subii(addiu(N, 1), m); /* t = N+1-m */
-    gel(step2, i) = mkvec5(N, t, s, gel(EP,1), gel(EP,2));
-    Dprev = D;
+    gel(Hi, i) = HD;
   }
-  gunclone_deep(db); return step2;
+  gunclone_deep(db);
+  worker = snm_closure(is_entry("_ecpp_step2_worker"),mkvec(primelist));
+  mt_queue_start_lim(&pt, worker, ls-1);
+  for (j = 1; j < ls || pending; j++)
+  {
+    long jj;
+    GEN S = gel(step1, j), HD = gel(Hi, j), done;
+    mt_queue_submit(&pt, j, j < ls ? mkvec2(S, HD) : NULL);
+    done = mt_queue_get(&pt, &jj, &pending);
+    if (!done) continue;
+    dbg_mode() {
+      GEN T = gel(done,2);
+      GEN N = NDmqg_get_N(gel(step1, jj));
+      err_printf(ANSI_BRIGHT_GREEN "\n[ %3d | %4ld bits]" ANSI_RESET, jj, expi(N));
+      err_printf(" %6ld", time_record(X0, "C2", T[1]));
+      err_printf(" %6ld", time_record(X0, "C3", T[2]));
+      err_printf(" %6ld", time_record(X0, "D1", T[3]));
+    }
+    gel(step2, jj) = gel(done,1);
+  }
+  mt_queue_end(&pt);
+  return step2;
 }
 /* end of functions for step 2 */
 
+GEN
+ecpp_sqrt_worker(GEN p, GEN g, GEN N)
+{
+  GEN r = Fp_sqrt_i(p, g, N);
+  return r ? r: gen_0;
+}
+
+static void
+ecpp_parsqrt(GEN N, GEN param, GEN kroP, GEN sqrtlist, GEN g, long nb, long *nbsqrt)
+{
+  GEN primelist = ecpp_param_get_primelist(param);
+  GEN ordinv = ecpp_param_get_primeord(param);
+  long i, k, l = lg(ordinv), n = *nbsqrt+1;
+  GEN worker = snm_closure(is_entry("_ecpp_sqrt_worker"), mkvec2(g, N));
+  GEN W = cgetg(nb+1,t_VEC), V;
+  for (i=n, k=1; k<=nb && i<l; i++)
+  {
+    long pi = ordinv[i];
+    long s = kroP[pi];
+    if (s > 1) kroP[pi] = s = krosi(primelist[pi], N); /* update cache */
+    if (kroP[pi] > 0)
+      gel(W,k++) = stoi(primelist[pi]);
+  }
+  *nbsqrt=i-1;
+  setlg(W,k);
+  V = gen_parapply(worker, W);
+  for (i=n, k=1; k<=nb && i<l; i++)
+  {
+    long pi = ordinv[i];
+    if (kroP[pi] > 0)
+    {
+      dbg_mode() err_printf(ANSI_MAGENTA "S" ANSI_RESET);
+      if (!signe(gel(V,k)))
+        pari_err_BUG("D_find_discsqrt"); /* possible if N composite */
+      gel(sqrtlist,pi) = gel(V,k++);
+    }
+  }
+}
+
 /* start of functions for step 1 */
 
 /* This finds the square root of D modulo N [given by Dfac]: find the square
  * root modulo N of each prime p dividing D and multiply them out */
 static GEN
-D_find_discsqrt(GEN N, GEN primelist, GEN Dfac, GEN sqrtlist, GEN g)
+D_find_discsqrt(GEN N, GEN param, GEN Dfac, GEN kroP, GEN sqrtlist, GEN g, long *nbsqrt)
 {
-  GEN s = NULL;
+  GEN s = NULL, sj;
   long i, l = lg(Dfac);
   for (i = 1; i < l; i++)
   {
     long j = Dfac[i];
-    GEN sj = gel(sqrtlist,j);
-    if (!signe(sj))
-    {
-      GEN p = stoi(primelist[j]);
-      dbg_mode() err_printf(ANSI_MAGENTA "S" ANSI_RESET);
-      /* A4: Get the square root of a prime factor of D. */
-      sj = gel(sqrtlist, j) = Fp_sqrt_i(p, g, N);
-      if (!sj) pari_err_BUG("D_find_discsqrt"); ; /* possible if N composite */
-    }
+    while(!signe(gel(sqrtlist,j)))
+      ecpp_parsqrt(N, param, kroP, sqrtlist, g, mt_nbthreads(), nbsqrt);
+    sj = gel(sqrtlist,j);
     s = s? Fp_mul(s, sj, N): sj;
   }
   return s;/* != NULL */
@@ -766,7 +854,7 @@ NUV_find_m(GEN Dinfo, GEN N, GEN U, GEN V, long wD)
    Finally, sqrtlist and g help compute the square root modulo N of D.
 */
 static long
-D_collectcards(GEN N, GEN param, GEN* X0, GEN Dinfo, GEN sqrtlist, GEN g, GEN Dmbatch, GEN kroP)
+D_collectcards(GEN N, GEN param, GEN* X0, GEN Dinfo, GEN sqrtlist, GEN g, GEN Dmbatch, GEN kroP, long *nbsqrt)
 {
   long i, l, corn_succ, wD, D = Dinfo_get_D(Dinfo);
   GEN U, V, sqrtofDmodN, Dfac = Dinfo_get_Dfac(Dinfo);
@@ -784,14 +872,14 @@ D_collectcards(GEN N, GEN param, GEN* X0, GEN Dinfo, GEN sqrtlist, GEN g, GEN Dm
   }
   /* A3: Get square root of D mod N */
   dbg_mode() timer_start(&ti);
-  sqrtofDmodN = D_find_discsqrt(N, primelist, Dfac, sqrtlist, g);
+  sqrtofDmodN = D_find_discsqrt(N, param, Dfac, kroP, sqrtlist, g, nbsqrt);
   dbg_mode() timer_record(X0, "A3", &ti);
   /* A5: Use square root with Cornacchia to solve U^2 + |D|V^2 = 4N */
   dbg_mode() timer_start(&ti);
   corn_succ = cornacchia2_sqrt( utoi(labs(D)), N, sqrtofDmodN, &U, &V);
   dbg_mode() timer_record(X0, "A5", &ti);
   if (!corn_succ) {
-    dbg_mode() err_printf(ANSI_YELLOW "c" ANSI_RESET);
+    dbg_mode1() err_printf(ANSI_YELLOW "c" ANSI_RESET);
     return 0;
   }
   /* A6: Collect the w(D) cardinalities of E/(F_N) with CM by D */
@@ -893,22 +981,14 @@ Dmbatch_factor_Dmqvec(GEN N, long expiN, GEN* X0, GEN Dmbatch, GEN param)
   /* B2: For each batch, remove cardinalities lower than (N^(1/4)+1)^2
    *     and cardinalities in which we didn't win enough bits */
   Dmqvec = Dmqvec_slice(N, Dmqvec);
-  if (!Dmqvec) { avma = av; return NULL; } /* nothing is left */
+  if (!Dmqvec) return gc_NULL(av); /* nothing is left */
   return gerepilecopy(av, Dmqvec);
 }
 
-/* Dmq (where q has no small prime factors): determine if q is pseudoprime */
-static long
-Dmq_isgoodq(GEN Dmq, GEN* X0)
-{
-  pari_timer ti;
-  long s;
-  /* B3: Check pseudoprimality of each q on the list. */
-  dbg_mode() timer_start(&ti);
-  s = BPSW_psp_nosmalldiv(gel(Dmq,3));
-  dbg_mode() timer_record(X0, "B3", &ti);
-  return s; /* did not find for this m */
-}
+GEN
+ecpp_ispsp_worker(GEN N)
+{ return mkvecsmall(BPSW_psp_nosmalldiv(N)); }
+
 static GEN
 mkNDmqg(GEN z, GEN N, GEN Dmq, GEN g, GEN sqrtlist)
 {
@@ -918,15 +998,15 @@ mkNDmqg(GEN z, GEN N, GEN Dmq, GEN g, GEN sqrtlist)
 }
 /* expi(N) > 64. Return [ NDmqg, N_downrun(q) ]; NDmqg is a vector of the form
  * [N,D,m,q,g,sqrt]. For successive D, find a square root of D mod N (g is a
- * quadratic non-residue), solve U^2 + |D|V^2 = 4N, then find candidates for m.
+ * quadratic nonresidue), solve U^2 + |D|V^2 = 4N, then find candidates for m.
  * When enough m's, batch-factor them to produce the q's. If one of the q's is
  * pseudoprime, recursive call with N = q. May return gen_0 at toplevel
  * => N has a small prime divisor */
 static GEN
-N_downrun(GEN N, GEN param, GEN *X0, long *depth, long persevere)
+N_downrun(GEN N, GEN param, GEN worker, GEN *X0, long *depth, long persevere)
 {
   pari_timer T, ti;
-  long lgdisclist, lprimelist, t, i, j, expiN = expi(N);
+  long lgdisclist, lprimelist, nbsqrt = 0, t, i, j, expiN = expi(N);
   long persevere_next = 0, FAIL = 0;
   ulong maxpcdg;
   GEN primelist, disclist, sqrtlist, g, Dmbatch, kroP;
@@ -957,6 +1037,8 @@ N_downrun(GEN N, GEN param, GEN *X0, long *depth, long persevere)
     case 5: kroP[1] = -1; kroP[2] = 1; kroP[3] =-1; break;
     case 7: kroP[1] =  1; kroP[2] =-1; kroP[3] =-1; break;
   }
+  for(i=4; i<lprimelist; i++)
+    kroP[i]= krosi(primelist[i],N);
   dbg_mode() timer_record(X0, "A2", &ti);
 
   /* Print the start of this iteration. */
@@ -989,7 +1071,7 @@ N_downrun(GEN N, GEN param, GEN *X0, long *depth, long persevere)
       GEN Dinfo = gel(disclist, i);
       long n;
       if (!persevere && Dinfo_get_pd(Dinfo) > maxpcdg) { FAIL = 1; break; }
-      n = D_collectcards(N,param, X0, Dinfo, sqrtlist, g, Dmbatch, kroP);
+      n = D_collectcards(N,param, X0, Dinfo, sqrtlist, g, Dmbatch, kroP, &nbsqrt);
       if (n < 0) return gen_0;
       last_i = i++;
       numcard += n; if (numcard >= t) break;
@@ -1012,9 +1094,31 @@ N_downrun(GEN N, GEN param, GEN *X0, long *depth, long persevere)
     lgDmqlist = lg(Dmqlist);
     for (j = 1; j < lgDmqlist; j++)
     {
-      GEN z, Dmq = gel(Dmqlist,j), q = gel(Dmq,3);
-      dbg_mode() err_printf(ANSI_WHITE "." ANSI_RESET);
-      if (!Dmq_isgoodq(Dmq, X0)) continue;
+      GEN z, Dmq, q;
+      struct pari_mt pt;
+      pari_timer ti2;
+      long running, stop = 0, pending = 0;
+      mt_queue_start_lim(&pt, worker, lgDmqlist-j);
+      dbg_mode() timer_start(&ti2);
+      while ((running = (!stop && j < lgDmqlist)) || pending)
+      {
+        long jj;
+        GEN done;
+        mt_queue_submit(&pt, j, running ? mkvec(gmael(Dmqlist,j,3)) : NULL);
+        done = mt_queue_get(&pt, &jj, &pending);
+        if (done)
+        {
+          if (done[1] == 0)
+          { dbg_mode() err_printf(ANSI_WHITE "." ANSI_RESET); }
+          else if (!stop || jj < stop)
+            stop = jj;
+        }
+        j++;
+      }
+      mt_queue_end(&pt);
+      dbg_mode() timer_record(X0, "B3", &ti2);
+      if (!stop && j >= lgDmqlist) break;
+      j = stop; Dmq = gel(Dmqlist,j); q = gel(Dmq,3);
 
       dbg_mode() {
         err_printf(ANSI_BRIGHT_RED "\n       %5ld bits " ANSI_RESET,
@@ -1026,7 +1130,7 @@ N_downrun(GEN N, GEN param, GEN *X0, long *depth, long persevere)
       if (expi(q) < 64) z = gen_1; /* q is prime; sentinel */
       else
       {
-        z = N_downrun(q, param, X0, depth, persevere_next);
+        z = N_downrun(q, param, worker, X0, depth, persevere_next);
         if (!z) {
           dbg_mode() { char o = persevere? '<': '[', c = persevere? '>': ']';
                        err_printf(ANSI_CYAN "\n%c %3d | %4ld bits%c "
@@ -1071,8 +1175,9 @@ ecpp_step1(GEN N, GEN param, GEN* X0)
 {
   pari_sp av = avma;
   long depth = 0;
-  GEN downrun = N_downrun(N, param, X0, &depth, 1);
-  if (downrun == NULL) { avma = av; return NULL; }
+  GEN worker = strtofunction("_ecpp_ispsp_worker");
+  GEN downrun = N_downrun(N, param, worker, X0, &depth, 1);
+  if (downrun == NULL) return gc_NULL(av);
   return gerepilecopy(av, ecpp_flattencert(downrun, depth));
 }
 
@@ -1098,8 +1203,8 @@ ecpp0(GEN N, GEN param)
   X0 = mkvec3(Tv, Cv, zero_zv(1));
 
   step1 = ecpp_step1(N, param, &X0);
-  if (step1 == NULL) { avma = av; return NULL; }
-  if (typ(step1) != t_VEC) { avma = av; return gen_0; }
+  if (step1 == NULL) return gc_NULL(av);
+  if (typ(step1) != t_VEC) { set_avma(av); return gen_0; }
 
   answer = ecpp_step2(step1, &X0, ecpp_param_get_primelist(param));
 
@@ -1182,10 +1287,8 @@ long
 isprimeECPP(GEN N)
 {
   pari_sp av = avma;
-  GEN res;
   if (!BPSW_psp(N)) return 0;
-  res = ecpp(N);
-  avma = av; return !isintzero(res);
+  return gc_long(av, !isintzero(ecpp(N)));
 }
 
 /* PARI ECPP Certificate -> Human-readable format */
@@ -1207,7 +1310,7 @@ cert_out(GEN x)
     str_printf(&str, " N = %Ps\n", cert_get_N(xi));
     str_printf(&str, " t = %Ps\n", cert_get_t(xi));
     str_printf(&str, " s = %Ps\n", cert_get_s(xi));
-    str_printf(&str, "a4 = %Ps\n", cert_get_a4(xi));
+    str_printf(&str, "a = %Ps\n", cert_get_a4(xi));
     str_printf(&str, "D = %Ps\n", cert_get_D(xi));
     str_printf(&str, "m = %Ps\n", cert_get_m(xi));
     str_printf(&str, "q = %Ps\n", cert_get_q(xi));
@@ -1322,13 +1425,59 @@ Nq_isvalid(GEN N, GEN q)
   return (cmpii(sqri(c), shifti(mulii(N,q), 4)) > 0);
 }
 
+GEN
+primecertisvalid_ecpp_worker(GEN certi)
+{
+  GEN N, W, mP, sP, r, m, s, a, P, q;
+  if (lg(certi)-1 != 5) return gen_0;
+
+  N = cert_get_N(certi);
+  if (typ(N) != t_INT || signe(N) <= 0) return gen_0;
+  switch(umodiu(N,6))
+  {
+    case 1: case 5: break; /* Check if N is not divisible by 2 or 3 */
+    default: return gen_0;
+  }
+
+  W = cert_get_t(certi);
+  if (typ(W) != t_INT) return gen_0;
+  /* Check if W^2 < 4N (Hasse interval) */
+  if (!(cmpii(sqri(W), shifti(N,2)) < 0)) return gen_0;
+
+  s = cert_get_s(certi);
+  if (typ(s) != t_INT || signe(s) < 0) return gen_0;
+
+  m = cert_get_m(certi);
+  q = dvmdii(m, s, &r);
+  /* Check m%s == 0 */
+  if (!isintzero(r)) return gen_0;
+  /* Check q > (N^(1/4) + 1)^2 */
+  if (!Nq_isvalid(N, q)) return gen_0;
+
+  a = cert_get_a4(certi);
+  if (typ(a) != t_INT) return gen_0;
+
+  P = cert_get_P(certi);
+  if (typ(P) != t_VEC || lg(P) != 3) return gen_0;
+  P = FpE_to_FpJ(P);
+
+  /* Check mP == 0 */
+  mP = FpJ_mul(P, m, a, N);
+  if (!FpJ_is_inf(mP)) return gen_0;
+
+  /* Check sP != 0 and third component is coprime to N */
+  sP = FpJ_mul(P, s, a, N);
+  if (!isint1(gcdii(gel(sP, 3), N))) return gen_0;
+  return q;
+}
+
 /* return 1 if 'cert' is a valid PARI ECPP certificate */
 static long
 ecppisvalid_i(GEN cert)
 {
   const long trustbits = 64;/* a pseudoprime < 2^trustbits is prime */
   long i, lgcert = lg(cert);
-  GEN ql, q = gen_0;
+  GEN ql, q = gen_0, worker, check;
 
   if (typ(cert) == t_INT)
   {
@@ -1339,61 +1488,21 @@ ecppisvalid_i(GEN cert)
   ql = gel(cert, lgcert-1); if (lg(ql)-1 != 5) return 0;
   ql = cert_get_q(ql);
   if (expi(ql) >= trustbits || !BPSW_psp(ql)) return 0;
-
+  worker = strtofunction("_primecertisvalid_ecpp_worker");
+  check = gen_parapply(worker, cert);
   for (i = 1; i < lgcert; i++)
   {
-    GEN N, W, mP, sP, r, m, s, a, P, certi = gel(cert, i);
-    if (lg(certi)-1 != 5) return 0;
-
-    N = cert_get_N(certi);
-    if (typ(N) != t_INT || signe(N) <= 0) return 0;
-    switch(umodiu(N,6))
-    {
-      case 1: case 5: break; /* Check if N is not divisible by 2 or 3 */
-      default: return 0;
-    }
-    /* Check if N matches the q from the previous entry. */
+    GEN certi = gel(cert, i);
+    GEN qq = gel(check,i), N = cert_get_N(certi);
+    if (isintzero(qq)) return 0;
     if (i > 1 && !equalii(N, q)) return 0;
-
-    W = cert_get_t(certi);
-    if (typ(W) != t_INT) return 0;
-    /* Check if W^2 < 4N (Hasse interval) */
-    if (!(cmpii(sqri(W), shifti(N,2)) < 0)) return 0;
-
-    s = cert_get_s(certi);
-    if (typ(s) != t_INT || signe(s) < 0) return 0;
-
-    m = cert_get_m(certi);
-    q = dvmdii(m, s, &r);
-    /* Check m%s == 0 */
-    if (!isintzero(r)) return 0;
-    /* Check q > (N^(1/4) + 1)^2 */
-    if (!Nq_isvalid(N, q)) return 0;
-
-    a = cert_get_a4(certi);
-    if (typ(a) != t_INT) return 0;
-
-    P = cert_get_P(certi);
-    if (typ(P) != t_VEC || lg(P) != 3) return 0;
-    P = FpE_to_FpJ(P);
-
-    /* Check mP == 0 */
-    mP = FpJ_mul(P, m, a, N);
-    if (!FpJ_is_inf(mP)) return 0;
-
-    /* Check sP != 0 and third component is coprime to N */
-    sP = FpJ_mul(P, s, a, N);
-    if (!isint1(gcdii(gel(sP, 3), N))) return 0;
+    q = qq;
   }
   return 1;
 }
 long
 ecppisvalid(GEN cert)
-{
-  pari_sp av = avma;
-  long v = ecppisvalid_i(cert);
-  avma = av; return v;
-}
+{ pari_sp av = avma; return gc_long(av, ecppisvalid_i(cert)); }
 
 GEN
 ecppexport(GEN cert, long flag)
