@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -65,7 +66,7 @@ gen_BG_add(void *E, bg_fun *fun, struct bg_data *bg, GEN n, long i, GEN a, GEN l
     nexta = mulis(a, bg->an[p]);
     if (i == j && umodiu(bg->N, p)) nexta = subii(nexta, mului(p, lasta));
     gen_BG_add(E, fun, bg, pn, j, nexta, a);
-    avma = av;
+    set_avma(av);
   }
 }
 
@@ -97,7 +98,7 @@ gen_BG_rec(void *E, bg_fun *fun, struct bg_data *bg)
     ulong pp = bg->p[i];
     long ap = bg->an[pp];
     gen_BG_add(E, fun, bg, utoipos(pp), i, stoi(ap), gen_1);
-    avma = av2;
+    set_avma(av2);
   }
   if (DEBUGLEVEL) err_printf("2nd stage, looping for p <= %Ps\n", bndov2);
   while ( (p = forprime_next(&S)) )
@@ -117,9 +118,9 @@ gen_BG_rec(void *E, bg_fun *fun, struct bg_data *bg)
       a = mulis(ap, aj);
       n = muliu(p, j);
       fun(E, n, a);
-      avma = av3;
+      set_avma(av3);
     }
-    avma = av2;
+    set_avma(av2);
     if (abscmpii(p, bndov2) >= 0) break;
   }
   if (DEBUGLEVEL) err_printf("3nd stage, looping for p <= %Ps\n", bg->bnd);
@@ -128,9 +129,9 @@ gen_BG_rec(void *E, bg_fun *fun, struct bg_data *bg)
     GEN ap = ellap(bg->E, p);
     if (!signe(ap)) continue;
     fun(E, p, ap);
-    avma = av2;
+    set_avma(av2);
   }
-  avma = av;
+  set_avma(av);
 }
 
 /******************************************************************
@@ -180,27 +181,25 @@ param_points(GEN N, double Y, double tmax, long bprec, long *cprec, long *L,
 static GEN
 vecF2_lk(GEN E, GEN K, GEN rbnd, GEN Q, GEN sleh, long prec)
 {
-  pari_sp av = avma, av2;
+  pari_sp av;
   long l, L  = lg(K)-1;
   GEN a = ellanQ_zv(E, itos(gel(K,1)));
   GEN S = cgetg(L+1, t_VEC);
+
+  for (l = 1; l <= L; l++) gel(S,l) = cgetr(prec);
+  av = avma;
   for (l = 1; l <= L; l++)
-    gel(S,l) = cgetr(prec);
-  av2 = avma;
-  for (l = 1; l <= L; ++l)
   {
-    GEN e1, Sl;
-    long aB, b, A, B;
-    GEN z, zB;
-    pari_sp av3;
-    long Kl = itou(gel(K,l));
+    GEN e1, Sl, z, zB;
+    long aB, b, A, B, Kl = itou(gel(K,l));
+    pari_sp av2;
     /* FIXME: could reduce prec here (useful for large prec) */
     e1 = gel(Q, l);
     Sl = real_0(prec);;
     /* baby-step giant step */
-    A = rbnd[l]; B = A;
+    B = A = rbnd[l];
     z = powersr(e1, B); zB = gel(z, B+1);
-    av3 = avma;
+    av2 = avma;
     for (aB = A*B; aB >= 0; aB -= B)
     {
       GEN s = real_0(prec); /* could change also prec here */
@@ -208,15 +207,14 @@ vecF2_lk(GEN E, GEN K, GEN rbnd, GEN Q, GEN sleh, long prec)
       {
         long k = aB+b;
         if (k <= Kl && a[k]) s = addrr(s, mulsr(a[k], gel(z, b+1)));
-        if (gc_needed(av3, 1))
-          gerepileall(av3, 2, &s, &Sl);
+        if (gc_needed(av2, 1)) gerepileall(av2, 2, &s, &Sl);
       }
       Sl = addrr(mulrr(Sl, zB), s);
     }
     affrr(mulrr(Sl, gel(sleh,l)), gel(S, l)); /* to avoid copying all S */
-    avma = av2;
+    set_avma(av);
   }
-  return gerepilecopy(av, S);
+  return S;
 }
 
 /* Return C, C[i][j] = Q[j]^i, i = 1..nb */
@@ -269,14 +267,13 @@ ellL1_add(void *E, GEN n, GEN a)
       ulong r, q = uabsdiviu_rem(n, bb->rbnd[j], &r);
       GEN giant = gel(bb->giant, j), baby = gel(bb->baby, j);
       affrr(addrr(gel(giant, q+1), mulri(gel(baby, r+1), a)), gel(giant, q+1));
-      avma = av;
+      set_avma(av);
     } else break;
 }
 
 static GEN
 vecF2_lk_bsgs(GEN E, GEN bnd, GEN rbnd, GEN Q, GEN sleh, GEN N, long prec)
 {
-  pari_sp av = avma;
   struct bg_data bg;
   struct baby_giant bb;
   long k, L = lg(bnd)-1;
@@ -287,75 +284,68 @@ vecF2_lk_bsgs(GEN E, GEN bnd, GEN rbnd, GEN Q, GEN sleh, GEN N, long prec)
   S = cgetg(L+1, t_VEC);
   for (k = 1; k <= L; ++k)
   {
-    pari_sp av2 = avma;
+    pari_sp av = avma;
     long j, g = rbnd[k];
-    GEN giant = gmael(bb.baby, k, g+1);
-    GEN Sl = real_0(prec);
-    for (j = g; j >=1; j--)
-      Sl = addrr(mulrr(Sl, giant), gmael(bb.giant,k,j));
-    gel(S, k) = gerepileupto(av2, mulrr(gel(sleh,k), Sl));
+    GEN giant = gmael(bb.baby, k, g+1), Sl = gmael(bb.giant, k, g);
+    for (j = g-1; j >=1; j--) Sl = addrr(mulrr(Sl, giant), gmael(bb.giant,k,j));
+    gel(S, k) = gerepileuptoleaf(av, mulrr(gel(sleh,k), Sl));
   }
-  return gerepileupto(av, S);
+  return S;
 }
 
+static long
+_sqrt(GEN x) { pari_sp av = avma; return gc_long(av, itou(sqrtint(x))); }
+
 static GEN
 vecF(struct lcritical *C, GEN E)
 {
-  pari_sp av = avma, av2;
-  long prec = C->cprec, Ks = itos_or_0(C->K), l, L = C->L;
-  GEN N = ellQ_get_N(E);
-  GEN PiN = shiftr(divrr(mppi(prec), gsqrt(N, prec)), 1);
-  GEN eh = mpexp(C->h), elh = powersr(eh, L-1), sleh = elh;
-  GEN Q, bnd, rbnd, vec;
+  pari_sp av = avma;
+  long prec = C->cprec, Ks = itos_or_0(C->K), L = C->L, l;
+  GEN N = ellQ_get_N(E), PiN;
+  GEN e = mpexp(C->h), elh = powersr(e, L-1), Q, bnd, rbnd, vec;
+
+  PiN = divrr(Pi2n(1,prec), sqrtr_abs(itor(N, prec)));
+  setsigne(PiN, -1); /* - 2Pi/sqrt(N) */
+  bnd = gpowers0(invr(e), L-1, C->K); /* bnd[i] = K exp(-(i-1)h) */
   rbnd = cgetg(L+1, t_VECSMALL);
-  av2 = avma;
-  bnd = cgetg(L+1, t_VEC);
   Q  = cgetg(L+1, t_VEC);
-  for (l = 1; l <= L; ++l)
+  for (l = 1; l <= L; l++)
   {
-    gel(bnd,l) = l==1 ? C->K: ceil_safe(divir(C->K, gel(elh, l)));
-    rbnd[l] = itou(sqrtint(gel(bnd,l)))+1;
-    gel(Q, l) = mpexp(mulrr(negr(PiN), gel(elh, l)));
+    gel(bnd,l) = ceil_safe(gel(bnd,l));
+    rbnd[l] = _sqrt(gel(bnd,l)) + 1;
+    gel(Q, l) = mpexp(mulrr(PiN, gel(elh, l)));
   }
-  gerepileall(av2, 2, &bnd, &Q);
   if (Ks && baby_size(rbnd, Ks, prec) > (Ks>>1))
-    vec = vecF2_lk(E, bnd, rbnd, Q, sleh, prec);
+    vec = vecF2_lk(E, bnd, rbnd, Q, elh, prec);
   else
-    vec = vecF2_lk_bsgs(E, bnd, rbnd, Q, sleh, N, prec);
+    vec = vecF2_lk_bsgs(E, bnd, rbnd, Q, elh, N, prec);
   return gerepileupto(av, vec);
 }
 
-/* ************************************************************************
- *
- * Compute Lambda function by Fourier inversion
- *
- */
-
+/* Lambda function by Fourier inversion. vec is a grid, t a scalar or t_SER */
 static GEN
 glambda(GEN t, GEN vec, GEN h, long real, long prec)
 {
-  GEN ehs, elhs;
-  GEN r;
-  long L = lg(vec)-1, l;
-  /* assume vec is a grid */
-  ehs = gexp(gmul(gen_I(),gmul(h, t)), prec);
-  elhs = (real == 1) ? gen_1 : mkcomplex(gen_0, gen_m1);
-  r = gmul2n(real_i(gmul(real_i(gel(vec, 1)), elhs)), -1);
+  GEN z, r, e = gexp(gmul(mkcomplex(gen_0,h), t), prec);
+  long n = lg(vec)-1, i;
+
+  r = real == 1? gmul2n(real_i(gel(vec, 1)), -1): gen_0;
+  z = real == 1? e: gmul(powIs(3), e);
   /* FIXME: summing backward may be more stable */
-  for (l = 2; l <= L; ++l)
+  for (i = 2; i <= n; i++)
   {
-    elhs = gmul(elhs, ehs);
-    r = gadd(r, real_i(gmul(gel(vec, l), elhs)));
+    r = gadd(r, real_i(gmul(gel(vec,i), z)));
+    if (i < n) z = gmul(z, e);
   }
   return gmul(mulsr(4, h), r);
 }
 
 static GEN
-Lpoints(struct lcritical *C, GEN e, GEN tmax, long bprec)
+Lpoints(struct lcritical *C, GEN e, double tmax, long bprec)
 {
   double h = 0, Y = .97;
   GEN N = ellQ_get_N(e);
-  param_points(N, Y, gtodouble(tmax), bprec, &C->cprec, &C->L, &C->K, &h);
+  param_points(N, Y, tmax, bprec, &C->cprec, &C->L, &C->K, &h);
   C->real = ellrootno_global(e);
   C->h = rtor(dbltor(h), C->cprec);
   return vecF(C, e);
@@ -386,7 +376,7 @@ ellL1_eval(GEN e, GEN vec, struct lcritical *C, GEN t, long prec)
 static GEN
 ellL1_der(GEN e, GEN vec, struct lcritical *C, GEN t, long der, long prec)
 {
-  GEN r = polcoeff0(ellL1_eval(e, vec, C, t, prec), der, 0);
+  GEN r = polcoef(ellL1_eval(e, vec, C, t, prec), der, 0);
   r = gmul(r,powIs(C->real == 1 ? -der: 1-der));
   return gmul(real_i(r), mpfact(der));
 }
@@ -401,8 +391,8 @@ ellL1_bitprec(GEN E, long r, long bitprec)
   if (r < 0)
     pari_err_DOMAIN("ellL1", "derivative order", "<", gen_0, stoi(r));
   e = ellanal_globalred(E, NULL);
-  if (r == 0 && ellrootno_global(e) < 0) { avma = av; return gen_0; }
-  vec = Lpoints(&C, e, gen_0, bitprec);
+  if (r == 0 && ellrootno_global(e) < 0) { set_avma(av); return gen_0; }
+  vec = Lpoints(&C, e, 0., bitprec);
   t = r ? scalarser(gen_1, 0, r):  zeroser(0, 0);
   setvalp(t, 1);
   return gerepileupto(av, ellL1_der(e, vec, &C, t, r, prec));
@@ -429,7 +419,7 @@ ellanalyticrank_bitprec(GEN E, GEN eps, long bitprec)
       if (typ(eps) != t_REAL) pari_err_TYPE("ellanalyticrank", eps);
     }
   e = ellanal_globalred(E, NULL);
-  vec = Lpoints(&C, e, gen_0, bitprec);
+  vec = Lpoints(&C, e, 0., bitprec);
   if (DEBUGLEVEL) timer_printf(&ti, "init L");
   av2 = avma;
   for (rk = C.real>0 ? 0: 1;  ; rk += 2)
@@ -441,7 +431,7 @@ ellanalyticrank_bitprec(GEN E, GEN eps, long bitprec)
     if (DEBUGLEVEL) timer_printf(&ti, "L^(%ld)=%Ps", rk, Lrk);
     if (abscmprr(Lrk, eps) > 0)
       return gerepilecopy(av, mkvec2(stoi(rk), Lrk));
-    avma = av2;
+    set_avma(av2);
   }
 }
 
@@ -541,13 +531,11 @@ heegner_psi(GEN E, GEN N, GEN points, long bitprec)
   pari_sp av = avma, av2;
   struct baby_giant bb;
   struct bg_data bg;
-  long k, L = lg(points)-1, prec = nbits2prec(bitprec)+EXTRAPRECWORD;
-  GEN  Q, pi2 = Pi2n(1, prec), bnd, rbnd;
-  long l;
+  long l, k, L = lg(points)-1, prec = nbits2prec(bitprec)+EXTRAPRECWORD;
+  GEN Q, pi2 = Pi2n(1, prec), bnd, rbnd, bndmax;
   GEN B = divrr(mulur(bitprec,mplog2(DEFAULTPREC)), pi2);
-  GEN bndmax;
-  rbnd = cgetg(L+1, t_VECSMALL);
-  av2 = avma;
+
+  rbnd = cgetg(L+1, t_VECSMALL); av2 = avma;
   bnd = cgetg(L+1, t_VEC);
   Q  = cgetg(L+1, t_VEC);
   for (l = 1; l <= L; ++l)
@@ -561,18 +549,15 @@ heegner_psi(GEN E, GEN N, GEN points, long bitprec)
   gen_BG_init(&bg, E, N, bndmax);
   if (bitprec >= 1900)
   {
-    GEN S;
+    GEN S = cgetg(L+1, t_VEC);
     baby_init2(&bb, Q, bnd, rbnd, prec);
     gen_BG_rec((void*)&bb, heegner_L1_bg, &bg);
-    S = cgetg(L+1, t_VEC);
     for (k = 1; k <= L; ++k)
     {
       pari_sp av2 = avma;
       long j, g = rbnd[k];
-      GEN giant = gmael(bb.baby, k, g+1);
-      GEN Sl = real_0(prec);
-      for (j = g; j >=1; j--)
-        Sl = gadd(gmul(Sl, giant), gmael(bb.giant,k,j));
+      GEN giant = gmael(bb.baby, k, g+1), Sl = real_0(prec);
+      for (j = g; j >=1; j--) Sl = gadd(gmul(Sl, giant), gmael(bb.giant,k,j));
       gel(S, k) = gerepileupto(av2, real_i(Sl));
     }
     return gerepileupto(av, S);
@@ -877,7 +862,7 @@ best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
           best = N; break;
         }
       }
-      avma = av2;
+      set_avma(av2);
     }
   }
 #ifdef DEBUG
@@ -949,7 +934,6 @@ heegner_indexmult(GEN om, long t, GEN tam, long prec)
   return gerepileupto(av, divru(divir(tam, Ovr), 4*t*t));
 }
 
-
 /* omega(gcd(D, N)), given faN = factor(N) */
 static long
 omega_N_D(GEN faN, ulong D)
@@ -1033,7 +1017,7 @@ heegner_find_point(GEN e, GEN om, GEN ht, GEN z1, long k, long prec)
       P = heegner_try_point(e, lambdas, ht, z2, prec);
       if (P) return P;
     }
-    avma = av;
+    set_avma(av);
   }
   pari_err_BUG("ellheegner, point not found");
   return NULL; /* LCOV_EXCL_LINE */
@@ -1110,9 +1094,9 @@ listDisc(GEN fa4N, GEN bad, long d)
       v[j++] = d;
       if (j > NDISC) break;
     }
-    avma = av;
+    set_avma(av);
   }
-  avma = av; return v;
+  set_avma(av); return v;
 }
 /* L = vector of [q1,q2] or [q1,q2,q2']
  * cd = (b^2 - D)/(4N) */
@@ -1250,7 +1234,7 @@ heegner_find_disc(GEN *points, GEN *coefs, long *pind, GEN E,
         }
       } while(0);
     }
-    d = listD[l-1]; avma = av;
+    d = listD[l-1]; set_avma(av);
   }
 }
 
