@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -20,7 +21,6 @@ static long
 ellisweierstrasspoint(GEN E, GEN Q)
 { return ell_is_inf(Q) || gequal0(ec_dmFdy_evalQ(E, Q)); }
 
-
 /* Given an elliptic curve E = [a1, a2, a3, a4, a6] and t,w in the ring of
  * definition of E, return the curve
  *  E' = [a1, a2, a3, a4 - 5t, a6 - (E.b2 t + 7w)] */
@@ -43,50 +43,6 @@ get_isog_vars(GEN phi, long *vx, long *vy)
   if (*vy == *vx) *vy = gvar2(gel(phi,2));
 }
 
-static GEN
-RgX_homogenous_evalpow(GEN P, GEN A, GEN B)
-{
-  pari_sp av = avma;
-  long d, i, v;
-  GEN s;
-  if (typ(P)!=t_POL)
-    return mkvec2(P, gen_1);
-  d = degpol(P); v = varn(A);
-  s = scalarpol_shallow(gel(P, d+2), v);
-  for (i = d-1; i >= 0; i--)
-  {
-    s = gadd(gmul(s, A), gmul(gel(B,d+1-i), gel(P,i+2)));
-    if (gc_needed(av,1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"RgX_homogenous_eval(%ld)",i);
-      s = gerepileupto(av, s);
-    }
-  }
-  s = gerepileupto(av, s);
-  return mkvec2(s, gel(B,d+1));
-}
-
-static GEN
-RgXQX_homogenous_evalpow(GEN P, GEN A, GEN B, GEN T)
-{
-  pari_sp av = avma;
-  long i, d = degpol(P), v = varn(A);
-  GEN s;
-  if (signe(P)==0) return mkvec2(pol_0(v), pol_1(v));
-  s = scalarpol_shallow(gel(P, d+2), v);
-  for (i = d-1; i >= 0; i--)
-  {
-    s = RgX_add(RgXQX_mul(s, A, T), RgXQX_RgXQ_mul(gel(B,d+1-i), gel(P,i+2), T));
-    if (gc_needed(av,1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"RgX_homogenous_eval(%ld)",i);
-      s = gerepileupto(av, s);
-    }
-  }
-  s = gerepileupto(av, s);
-  return mkvec2(s, gel(B,d+1));
-}
-
 /* x must be nonzero */
 INLINE long _degree(GEN x) { return typ(x)==t_POL ? degpol(x): 0; }
 
@@ -96,20 +52,14 @@ static GEN
 ellcompisog(GEN F, GEN G)
 {
   pari_sp av = avma;
-  GEN Fv, Gh, Gh2, Gh3, f, g, h, h2, h3, den, num;
+  GEN Gh, Gh2, Gh3, f, g, h, h2, h3, den, num;
   GEN K, K2, K3, F0, F1, g0, g1, Gp;
-  long v, vx, vy, d;
+  long vx, vy, d;
   checkellisog(F);
   checkellisog(G);
   get_isog_vars(F, &vx, &vy);
-  v = fetch_var_higher();
-  Fv = shallowcopy(gel(F,3)); setvarn(Fv, v);
   Gh = gel(G,3); Gh2 = gsqr(Gh); Gh3 = gmul(Gh, Gh2);
-  K = gmul(polresultant0(Fv, deg1pol(gneg(Gh2),gel(G,1), v), v, 0), Gh);
-  delete_var();
-  K = RgX_normalize(RgX_div(K, RgX_gcd(K,deriv(K,0))));
-  K2 = gsqr(K); K3 = gmul(K, K2);
-  F0 = polcoeff0(gel(F,2), 0, vy); F1 = polcoeff0(gel(F,2), 1, vy);
+  F0 = polcoef(gel(F,2), 0, vy); F1 = polcoef(gel(F,2), 1, vy);
   d = maxss(maxss(degpol(gel(F,1)),_degree(gel(F,3))),
             maxss(_degree(F0),_degree(F1)));
   Gp = gpowers(Gh2, d);
@@ -117,6 +67,9 @@ ellcompisog(GEN F, GEN G)
   g0 = RgX_homogenous_evalpow(F0, gel(G,1), Gp);
   g1 = RgX_homogenous_evalpow(F1, gel(G,1), Gp);
   h =  RgX_homogenous_evalpow(gel(F,3), gel(G,1), Gp);
+  K = gmul(gel(h,1), Gh);
+  K = RgX_normalize(RgX_div(K, RgX_gcd(K,RgX_deriv(K))));
+  K2 = gsqr(K); K3 = gmul(K, K2);
   h2 = mkvec2(gsqr(gel(h,1)), gsqr(gel(h,2)));
   h3 = mkvec2(gmul(gel(h,1),gel(h2,1)), gmul(gel(h,2),gel(h2,2)));
   f  = gdiv(gmul(gmul(K2, gel(f,1)),gel(h2,2)), gmul(gel(f,2), gel(h2,1)));
@@ -129,7 +82,7 @@ ellcompisog(GEN F, GEN G)
 static GEN
 to_RgX(GEN P, long vx)
 {
-  return typ(P) == t_POL ? lift(P): scalarpol_shallow(lift(P), vx);
+  return typ(P) == t_POL && varn(P)==vx? P: scalarpol_shallow(P, vx);
 }
 
 static GEN
@@ -151,61 +104,57 @@ static GEN
 ellnfcompisog(GEN nf, GEN F, GEN G)
 {
   pari_sp av = avma;
-  GEN Fv, Gh, Gh2, Gh3, f, g, gd, h, h21, h22, h31, h32, den;
+  GEN Gh, Gh2, Gh3, f, g, gd, h, h21, h22, h31, h32, den;
   GEN K, K2, K3, F0, F1, G0, G1, g0, g1, Gp;
   GEN num0, num1, gn0, gn1;
   GEN g0d, g01, k3h32;
-  GEN T, res;
+  GEN T;
   pari_timer ti;
-  long v, vx, vy, d;
+  long vx, vy, d;
   if (!nf) return ellcompisog(F, G);
   T = nf_get_pol(nf);
   timer_start(&ti);
   checkellisog(F);
   checkellisog(G);
+  F = liftpol(F); G = liftpol(G);
   get_isog_vars(F, &vx, &vy);
-  v = fetch_var_higher();
-  Fv = shallowcopy(gel(F,3)); setvarn(Fv, v);
-  Gh = lift(gel(G,3)); Gh2 = RgXQX_sqr(Gh, T); Gh3 = RgXQX_mul(Gh, Gh2, T);
-  res = to_RgX(polresultant0(Fv, deg1pol(gmul(gneg(Gh2),gmodulo(gen_1,T)),gel(G,1), v), v, 0),vx);
-  delete_var();
-  K = Q_remove_denom(RgXQX_mul(res, Gh, T), NULL);
-  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: resultant");
-  K = RgXQX_div(K, nfgcd(K, deriv(K,0), T, NULL), T);
+  Gh = to_RgX(gel(G,3),vx); Gh2 = QXQX_sqr(Gh, T); Gh3 = QXQX_mul(Gh, Gh2, T);
+  F0 = to_RgX(polcoef(gel(F,2), 0, vy), vx);
+  F1 = to_RgX(polcoef(gel(F,2), 1, vy), vx);
+  G0 = to_RgX(polcoef(gel(G,2), 0, vy), vx);
+  G1 = to_RgX(polcoef(gel(G,2), 1, vy), vx);
+  d = maxss(maxss(degpol(gel(F,1)),degpol(gel(F,3))),maxss(degpol(F0),degpol(F1)));
+  Gp = QXQX_powers(Gh2, d, T);
+  f  = QXQX_homogenous_evalpow(to_RgX(gel(F,1),vx), gel(G,1), Gp, T);
+  g0 = QXQX_homogenous_evalpow(F0, to_RgX(gel(G,1),vx), Gp, T);
+  g1 = QXQX_homogenous_evalpow(F1, to_RgX(gel(G,1),vx), Gp, T);
+  h  = QXQX_homogenous_evalpow(to_RgX(gel(F,3),vx), gel(G,1), Gp, T);
+  K = Q_remove_denom(QXQX_mul(to_RgX(gel(h,1),vx), Gh, T), NULL);
+  K = RgXQX_div(K, nfgcd(K, RgX_deriv(K), T, NULL), T);
   K = RgX_normalize(K);
   if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: nfgcd");
-  K2 = RgXQX_sqr(K, T); K3 = RgXQX_mul(K, K2, T);
-  F0 = to_RgX(polcoeff0(gel(F,2), 0, vy), vx);
-  F1 = to_RgX(polcoeff0(gel(F,2), 1, vy), vx);
-  G0 = to_RgX(polcoeff0(gel(G,2), 0, vy), vx);
-  G1 = to_RgX(polcoeff0(gel(G,2), 1, vy), vx);
-  d = maxss(maxss(degpol(gel(F,1)),degpol(gel(F,3))),maxss(degpol(F0),degpol(F1)));
-  Gp = RgXQX_powers(Gh2, d, T);
-  f  = RgXQX_homogenous_evalpow(to_RgX(gel(F,1),vx), gel(G,1), Gp, T);
-  g0 = RgXQX_homogenous_evalpow(F0, to_RgX(gel(G,1),vx), Gp, T);
-  g1 = RgXQX_homogenous_evalpow(F1, to_RgX(gel(G,1),vx), Gp, T);
-  h  = RgXQX_homogenous_evalpow(to_RgX(gel(F,3),vx), gel(G,1), Gp, T);
+  K2 = QXQX_sqr(K, T); K3 = QXQX_mul(K, K2, T);
   if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: evalpow");
-  h21 = RgXQX_sqr(gel(h,1),T);
-  h22 = RgXQX_sqr(gel(h,2),T);
-  h31 = RgXQX_mul(gel(h,1), h21,T);
-  h32 = RgXQX_mul(gel(h,2), h22,T);
+  h21 = QXQX_sqr(gel(h,1),T);
+  h22 = QXQX_sqr(gel(h,2),T);
+  h31 = QXQX_mul(gel(h,1), h21,T);
+  h32 = QXQX_mul(gel(h,2), h22,T);
   if (DEBUGLEVEL) timer_printf(&ti,"h");
-  f  = RgXQX_div(RgXQX_mul(RgXQX_mul(K2, gel(f,1), T), h22, T),
-                           RgXQX_mul(gel(f,2), h21, T), T);
+  f  = RgXQX_div(QXQX_mul(QXQX_mul(K2, gel(f,1), T), h22, T),
+                          QXQX_mul(gel(f,2), h21, T), T);
   if (DEBUGLEVEL) timer_printf(&ti,"f");
-  den = RgXQX_mul(Gh3, gel(g1,2), T);
+  den = QXQX_mul(Gh3, gel(g1,2), T);
   if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: den");
-  g0d = RgXQX_mul(gel(g0,1),den, T);
-  g01 = RgXQX_mul(gel(g1,1),gel(g0,2),T);
-  num0 = RgX_add(g0d, RgXQX_mul(G0,g01, T));
-  num1 = RgXQX_mul(G1,g01, T);
+  g0d = QXQX_mul(gel(g0,1),den, T);
+  g01 = QXQX_mul(gel(g1,1),gel(g0,2),T);
+  num0 = RgX_add(g0d, QXQX_mul(G0,g01, T));
+  num1 = QXQX_mul(G1,g01, T);
   if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: num");
-  k3h32 = RgXQX_mul(K3,h32,T);
-  gn0 = RgXQX_mul(num0, k3h32, T);
-  gn1 = RgXQX_mul(num1, k3h32, T);
+  k3h32 = QXQX_mul(K3,h32,T);
+  gn0 = QXQX_mul(num0, k3h32, T);
+  gn1 = QXQX_mul(num1, k3h32, T);
   if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: gn");
-  gd = RgXQX_mul(RgXQX_mul(gel(g0,2), den, T), h31, T);
+  gd = QXQX_mul(QXQX_mul(gel(g0,2), den, T), h31, T);
   if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: gd");
   g = divy(gn0, gn1, gd, T, vy);
   if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: divy");
@@ -229,7 +178,7 @@ ellisogenyapply(GEN phi, GEN P)
   h = gel(phi, 3);
   get_isog_vars(phi, &vx, &vy);
   img_h = poleval(h, gel(P, 1));
-  if (gequal0(img_h)) { avma = ltop; return ellinf(); }
+  if (gequal0(img_h)) { set_avma(ltop); return ellinf(); }
 
   img_h2 = gsqr(img_h);
   img_h3 = gmul(img_h, img_h2);
@@ -383,7 +332,6 @@ first_three_power_sums(GEN pol, GEN *p1, GEN *p2, GEN *p3)
   *p3 = gadd(gmul(s1, gsub(*p2, s2)), gmulsg(-3L, ms3));
 }
 
-
 /* Let E and a t_POL h of degree 1 or 3 whose roots are 2-torsion points on E.
  * - if only_image != 0, return [t, w] used to compute the equation of the
  *   quotient by the given 2-torsion points
@@ -1205,7 +1153,7 @@ ellQ_goodl(GEN E)
       if ((mask &_13) && kross(D,13)==-1) mask &= ~_13;
     }
   }
-  avma = av; return mask;
+  return gc_ulong(av, mask);
 }
 
 static long
@@ -1232,7 +1180,7 @@ ellQ_goodl_l(GEN E, long l)
         long D = t*t-4*p;
         if (kross(D,l)==-1) return 0;
       }
-      avma = av;
+      set_avma(av);
     }
   }
   return 1;
@@ -1276,7 +1224,7 @@ ellnf_goodl_l(GEN E, GEN v)
         }
       }
     }
-    avma = av;
+    set_avma(av);
   }
   return w^((1UL<<(lv-1))-1);
 }
@@ -1356,7 +1304,7 @@ ellnf_prime_degree(GEN E)
     B = gcdii(B, ellnf_get_degree(E, utoi(p)));
     if (Z_issquareall(B,&rB)) B=rB;
   }
-  if (signe(B)==0) pari_err_IMPL("ellisomat, CM case");
+  if (signe(B)==0) pari_err_DOMAIN("ellisomat", "E","has",strtoGENstr("CM"),E);
   P = vec_to_vecsmall(gel(Z_factor(B),1));
   return shallowextract(P, utoi(ellnf_goodl_l(E, P)));
 }
@@ -1613,7 +1561,8 @@ get_isomat(GEN v)
   /* wE a vector of [a4,a6] */
   for (i = 1; i < l; i++)
   {
-    GEN e = ellinit(gel(wE,i), gen_1, 0), E = ellminimalmodel(e, NULL);
+    GEN e = ellinit(gel(wE,i), gen_1, DEFAULTPREC);
+    GEN E = ellminimalmodel(e, NULL);
     obj_free(e); gel(wE,i) = E;
   }
   return mkvec2(wE, M);
@@ -1659,7 +1608,7 @@ ellisotree(GEN E)
   vE = gel(L,1);
   adj = gel(L,2);
   n = lg(vE)-1; L = cgetg(n+1, t_VEC);
-  for (i = 1; i <= n; i++) gel(L,i) = ellR_area(gel(vE,i), LOWDEFAULTPREC);
+  for (i = 1; i <= n; i++) gel(L,i) = ellR_area(gel(vE,i), DEFAULTPREC);
   M = zeromatcopy(n,n);
   for (i = 1; i <= n; i++)
     for (j = i+1; j <= n; j++)
