@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -158,11 +159,8 @@ gtolong(GEN x)
       return itos(x);
     case t_REAL:
       return (long)(rtodbl(x) + 0.5);
-    case t_FRAC: {
-      pari_sp av = avma;
-      long y = itos(ground(x));
-      avma = av; return y;
-    }
+    case t_FRAC:
+    { pari_sp av = avma; return gc_long(av, itos(ground(x))); }
     case t_COMPLEX:
       if (gequal0(gel(x,2))) return gtolong(gel(x,1)); break;
     case t_QUAD:
@@ -421,7 +419,6 @@ int
 gequalm1(GEN x)
 {
   pari_sp av;
-  long y;
   GEN p1;
 
   switch(typ(x))
@@ -436,7 +433,7 @@ gequalm1(GEN x)
       return s < 0 ? absrnz_equal1(x): 0;
     }
     case t_INTMOD:
-      av=avma; y=equalii(addui(1,gel(x,2)), gel(x,1)); avma=av; return y;
+      av = avma; return gc_bool(av, equalii(addui(1,gel(x,2)), gel(x,1)));
 
     case t_FRAC:
       return 0;
@@ -451,17 +448,17 @@ gequalm1(GEN x)
       return gequalm1(gel(x,2)) && gequal0(gel(x,3));
 
     case t_PADIC:
-      av=avma; y=equalii(addui(1,gel(x,4)), gel(x,3)); avma=av; return y;
+      av = avma; return gc_bool(av, equalii(addui(1,gel(x,4)), gel(x,3)));
 
     case t_POLMOD:
-      av=avma; p1 = gaddgs(gel(x,2), 1);
-      y = gequal0(p1) || gequal(p1,gel(x,1)); avma=av; return y;
+      av = avma; p1 = gaddgs(gel(x,2), 1);
+      return gc_bool(av, gequal0(p1) || gequal(p1,gel(x,1)));
 
     case t_POL: return is_monomial_test(x, 0, &gequalm1);
     case t_SER: return is_monomial_test(x, valp(x), &gequalm1);
 
     case t_RFRAC:
-      av=avma; y=gequal(gel(x,1), gneg_i(gel(x,2))); avma=av; return y;
+      av = avma; return gc_bool(av, gequal(gel(x,1), gneg_i(gel(x,2))));
     case t_COL: return col_test(x, &gequalm1);
     case t_MAT: return mat_test(x, &gequalm1);
   }
@@ -531,23 +528,20 @@ cmp_universal(GEN x, GEN y)
       {
         long tx = list_typ(x), ty = list_typ(y);
         GEN vx, vy;
+        pari_sp av;
         if (tx < ty) return -1;
         if (tx > ty) return 1;
         vx = list_data(x);
         vy = list_data(y);
         if (!vx) return vy? -1: 0;
         if (!vy) return 1;
-        switch (tx)
+        av = avma;
+        if (tx == t_LIST_MAP)
         {
-        case t_LIST_MAP:
-          {
-            pari_sp av = avma;
-            int ret = cmp_universal_rec(maptomat_shallow(x), maptomat_shallow(y),1);
-            avma = av; return ret;
-          }
-        default:
-          return cmp_universal_rec(vx, vy, 1);
+          vx = maptomat_shallow(x);
+          vy = maptomat_shallow(y);
         }
+        return gc_int(av, cmp_universal_rec(vx, vy, 1));
       }
     default:
       return cmp_universal_rec(x, y, lontyp[tx]);
@@ -560,31 +554,27 @@ cmpfrac(GEN x, GEN y)
   pari_sp av = avma;
   GEN a = gel(x,1), b = gel(x,2);
   GEN c = gel(y,1), d = gel(y,2);
-  int r = cmpii(mulii(a, d), mulii(b, c));
-  avma = av; return r;
+  return gc_bool(av, cmpii(mulii(a, d), mulii(b, c)));
 }
 static int
 cmpifrac(GEN a, GEN y)
 {
   pari_sp av = avma;
   GEN c = gel(y,1), d = gel(y,2);
-  int r = cmpii(mulii(a, d), c);
-  avma = av; return r;
+  return gc_int(av, cmpii(mulii(a, d), c));
 }
 static int
 cmprfrac(GEN a, GEN y)
 {
   pari_sp av = avma;
   GEN c = gel(y,1), d = gel(y,2);
-  int r = cmpri(mulri(a, d), c);
-  avma = av; return r;
+  return gc_int(av, cmpri(mulri(a, d), c));
 }
 static int
 cmpgen(GEN x, GEN y)
 {
   pari_sp av = avma;
-  int s = gsigne(gsub(x,y));
-  avma = av; return s;
+  return gc_int(av, gsigne(gsub(x,y)));
 }
 
 /* returns the sign of x - y when it makes sense. 0 otherwise */
@@ -653,12 +643,11 @@ gcmpsg(long s, GEN y)
     case t_REAL: return cmpsr(s,y);
     case t_FRAC: {
       pari_sp av = avma;
-      GEN n = gel(y,1), d = gel(y,2);
-      int f = cmpii(mulsi(s,d), n); avma = av; return f;
+      return gc_int(av, cmpii(mulsi(s,gel(y,2)), gel(y,1)));
     }
     case t_QUAD: {
       pari_sp av = avma;
-      int f = gsigne(gsubsg(s, y)); avma = av; return f;
+      return gc_int(av, gsigne(gsubsg(s, y)));
     }
     case t_INFINITY: return -inf_get_sign(y);
   }
@@ -790,20 +779,17 @@ lexcmpsg(long x, GEN y)
 }
 
 /* as gcmp for vector/matrices, using lexicographic ordering on components */
-int
-lexcmp(GEN x, GEN y)
+static int
+lexcmp_i(GEN x, GEN y)
 {
   const long tx = roughtype(x), ty = roughtype(y);
   if (tx == ty)
     switch(tx)
     {
       case t_MAT:
-      case t_VEC:
-        return lexcmp_similar(x,y);
-      case t_VECSMALL:
-        return vecsmall_lexcmp(x,y);
-      default:
-        return gcmp(x,y);
+      case t_VEC: return lexcmp_similar(x,y);
+      case t_VECSMALL: return vecsmall_lexcmp(x,y);
+      default: return gcmp(x,y);
     }
   if (tx == t_VECSMALL) return  lexcmp_vecsmall_other(x,y,ty);
   if (ty == t_VECSMALL) return -lexcmp_vecsmall_other(y,x,tx);
@@ -812,7 +798,24 @@ lexcmp(GEN x, GEN y)
   if (ty == t_INT) return -lexcmp_scal_matvec(y,x);
 
   if (ty==t_MAT) return  lexcmp_vec_mat(x,y);
-  /*tx==t_MAT*/  return -lexcmp_vec_mat(y,x);
+  return -lexcmp_vec_mat(y,x); /*tx==t_MAT*/
+}
+int
+lexcmp(GEN x, GEN y)
+{
+  pari_sp av = avma;
+  if (typ(x) == t_COMPLEX)
+  {
+    x = mkvec2(gel(x,1), gel(x,2));
+    if (typ(y) == t_COMPLEX) y = mkvec2(gel(y,1), gel(y,2));
+    else y = mkvec2(y, gen_0);
+  }
+  else if (typ(y) == t_COMPLEX)
+  {
+    x = mkvec2(x, gen_0);
+    y = mkvec2(gel(y,1), gel(y,2));
+  }
+  return gc_int(av, lexcmp_i(x, y));
 }
 
 /*****************************************************************/
@@ -882,7 +885,7 @@ list_cmp(GEN x, GEN y, int cmp(GEN x, GEN y))
       pari_sp av = avma;
       GEN mx  = maptomat_shallow(x), my = maptomat_shallow(y);
       int ret = gidentical(gel(mx, 1), gel(my, 1)) && cmp(gel(mx, 2), gel(my, 2));
-      avma = av; return ret;
+      return gc_bool(av, ret);
     }
   default:
     return cmp(vx, vy);
@@ -1036,7 +1039,7 @@ gequal(GEN x, GEN y)
         return gequal(gel(x,2),gel(y,2)) && gequal(gel(x,1),gel(y,1));
       case t_PADIC:
         if (!equalii(gel(x,2),gel(y,2))) return 0;
-        av = avma; i = gequal0(gsub(x,y)); avma = av;
+        av = avma; i = gequal0(gsub(x,y)); set_avma(av);
         return i;
       case t_POLMOD:
         if (varn(gel(x,1)) != varn(gel(y,1))) break;
@@ -1067,8 +1070,9 @@ gequal(GEN x, GEN y)
         GEN a = gel(x,1), b = gel(x,2), c = gel(y,1), d = gel(y,2);
         if (gequal(b,d)) return gequal(a,c); /* simple case */
         av = avma;
-        i = gequal(simplify_shallow(gmul(a,d)), simplify_shallow(gmul(b,c)));
-        avma = av; return i;
+        a = simplify_shallow(gmul(a,d));
+        b = simplify_shallow(gmul(b,c));
+        return gc_bool(av, gequal(a,b));
       }
 
       case t_STR:
@@ -1084,18 +1088,16 @@ gequal(GEN x, GEN y)
       case t_INFINITY:
         return gequal(gel(x,1),gel(y,1));
     }
-  (void)&av; /* emulate volatile */
-  av = avma; i = gequal_try(x, y);
-  avma = av; return i;
+  if (is_noncalc_t(tx) || is_noncalc_t(ty)) return 0;
+  if (tx == t_INT && !signe(x)) return gequal0(y);
+  if (ty == t_INT && !signe(y)) return gequal0(x);
+  (void)&av; av = avma; /* emulate volatile */
+  return gc_bool(av, gequal_try(x, y));
 }
 
 int
 gequalsg(long s, GEN x)
-{
-  pari_sp av = avma;
-  int f = gequal(stoi(s), x);
-  avma = av; return f;
-}
+{ pari_sp av = avma; return gc_bool(av, gequal(stoi(s), x)); }
 
 /* a and b are t_INT, t_FRAC, t_REAL or t_COMPLEX of those. Check whether
  * a-b is invertible */
@@ -1104,11 +1106,9 @@ cx_approx_equal(GEN a, GEN b)
 {
   pari_sp av = avma;
   GEN d;
-  int r;
   if (a == b) return 1;
   d = gsub(a,b);
-  r = (gequal0(d) || (typ(d) == t_COMPLEX && gequal0(cxnorm(d))));
-  avma = av; return r;
+  return gc_bool(av, gequal0(d) || (typ(d)==t_COMPLEX && gequal0(cxnorm(d))));
 }
 /*******************************************************************/
 /*                                                                 */
@@ -1137,20 +1137,33 @@ intdvd(GEN x, GEN y, GEN *z) { GEN r; *z = dvmdii(x,y,&r); return (r==gen_0); }
 
 /* x t_FRAC, p t_INT, return v_p(x) */
 static long
-ratval(GEN x, GEN p) {
+frac_val(GEN x, GEN p) {
   long v = Z_pval(gel(x,2),p);
   if (v) return -v;
   return Z_pval(gel(x,1),p);
 }
-
 long
-Q_pval(GEN x, GEN p) { return (typ(x) == t_INT)? Z_pval(x, p): ratval(x, p); }
+Q_pval(GEN x, GEN p)
+{
+  if (lgefint(p) == 3) return Q_lval(x, uel(p,2));
+  return (typ(x)==t_INT)? Z_pval(x, p): frac_val(x, p);
+}
+
+static long
+frac_lval(GEN x, ulong p) {
+  long v = Z_lval(gel(x,2),p);
+  if (v) return -v;
+  return Z_lval(gel(x,1),p);
+}
+long
+Q_lval(GEN x, ulong p){return (typ(x)==t_INT)? Z_lval(x, p): frac_lval(x, p);}
 
 long
 Q_pvalrem(GEN x, GEN p, GEN *y)
 {
   GEN a, b;
   long v;
+  if (lgefint(p) == 3) return Q_lvalrem(x, uel(p,2), y);
   if (typ(x) == t_INT) return Z_pvalrem(x, p, y);
   a = gel(x,1);
   b = gel(x,2);
@@ -1159,6 +1172,19 @@ Q_pvalrem(GEN x, GEN p, GEN *y)
   v = Z_pvalrem(a, p, &a);
   *y = mkfrac(a, b); return v;
 }
+long
+Q_lvalrem(GEN x, ulong p, GEN *y)
+{
+  GEN a, b;
+  long v;
+  if (typ(x) == t_INT) return Z_lvalrem(x, p, y);
+  a = gel(x,1);
+  b = gel(x,2);
+  v = Z_lvalrem(b, p, &b);
+  if (v) { *y = isint1(b)? a: mkfrac(a, b); return -v; }
+  v = Z_lvalrem(a, p, &a);
+  *y = mkfrac(a, b); return v;
+}
 
 long
 gvaluation(GEN x, GEN p)
@@ -1198,14 +1224,14 @@ gvaluation(GEN x, GEN p)
       if (tp == t_POL) return signe(b)? 0: LONG_MAX;
       av = avma;
       if (!intdvd(a, p, &a)) break;
-      if (!intdvd(b, p, &b)) { avma = av; return 0; }
+      if (!intdvd(b, p, &b)) return gc_long(av,0);
       val = 1; while (intdvd(a,p,&a) && intdvd(b,p,&b)) val++;
-      avma = av; return val;
+      return gc_long(av,val);
     }
 
     case t_FRAC:
       if (tp == t_POL) return 0;
-      return ratval(x, p);
+      return frac_val(x, p);
 
     case t_PADIC:
       if (tp == t_POL) return 0;
@@ -1222,11 +1248,11 @@ gvaluation(GEN x, GEN p)
       a = RgX_divrem(a, p, ONLY_DIVIDES);
       if (!a) break;
       if (typ(b) != t_POL || varn(b) != v ||
-          !(b = RgX_divrem(b, p, ONLY_DIVIDES)) ) { avma = av; return 0; }
+          !(b = RgX_divrem(b, p, ONLY_DIVIDES)) ) return gc_long(av,0);
       val = 1;
       while ((a = RgX_divrem(a, p, ONLY_DIVIDES)) &&
              (b = RgX_divrem(b, p, ONLY_DIVIDES)) ) val++;
-      avma = av; return val;
+      return gc_long(av,val);
     }
     case t_POL: {
       if (tp == t_POL) {
@@ -1239,7 +1265,7 @@ gvaluation(GEN x, GEN p)
           for (val=0; ; val++)
           {
             x = RgX_divrem(x,p,ONLY_DIVIDES);
-            if (!x) { avma = av; return val; }
+            if (!x) return gc_long(av,val);
             if (gc_needed(av,1))
             {
               if(DEBUGMEM>1) pari_warn(warnmem,"gvaluation");
@@ -1282,7 +1308,7 @@ gpvaluation(GEN x, GEN p)
   return v == LONG_MAX? mkoo(): stoi(v);
 }
 
-/* x is non-zero */
+/* x is nonzero */
 long
 u_lvalrem(ulong x, ulong p, ulong *py)
 {
@@ -1382,7 +1408,7 @@ Z_lval(GEN x, ulong p)
       break;
     }
   }
-  avma = av; return vx;
+  return gc_long(av,vx);
 }
 long
 Z_lvalrem(GEN x, ulong p, GEN *py)
@@ -1411,7 +1437,7 @@ Z_lvalrem(GEN x, ulong p, GEN *py)
       break;
     }
   }
-  avma = av; *py = icopy(x); setsigne(*py, sx); return vx;
+  set_avma(av); *py = icopy(x); setsigne(*py, sx); return vx;
 }
 
 /* Is |q| <= p ? */
@@ -1451,7 +1477,7 @@ Z_lvalrem_stop(GEN *n, ulong p, int *stop)
     return v;
   }
   av = avma; v = 0; q = absdiviu_rem(*n, p, &r);
-  if (r) avma = av;
+  if (r) set_avma(av);
   else
   {
     do {
@@ -1469,7 +1495,7 @@ Z_lvalrem_stop(GEN *n, ulong p, int *stop)
   *stop = isless_iu(q,p); return v;
 }
 
-/* x is a non-zero integer, |p| > 1 */
+/* x is a nonzero integer, |p| > 1 */
 long
 Z_pvalrem(GEN x, GEN p, GEN *py)
 {
@@ -1482,7 +1508,7 @@ Z_pvalrem(GEN x, GEN p, GEN *py)
   for(;;)
   {
     GEN r, q = dvmdii(x,p,&r);
-    if (r != gen_0) { avma = av; *py = icopy(x); return vx; }
+    if (r != gen_0) { set_avma(av); *py = icopy(x); return vx; }
     vx++; x = q;
   }
 }
@@ -1509,7 +1535,7 @@ Z_pval(GEN x, GEN p) {
   for(;;)
   {
     GEN r, q = dvmdii(x,p,&r);
-    if (r != gen_0) { avma = av; return vx; }
+    if (r != gen_0) return gc_long(av,vx);
     vx++; x = q;
   }
 }
@@ -1518,7 +1544,9 @@ Z_pval(GEN x, GEN p) {
 long
 factorial_lval(ulong n, ulong p)
 {
-  ulong q = p, v = 0;
+  ulong q, v;
+  if (p == 2) return n - hammingl(n);
+  q = p; v = 0;
   do { v += n/q; q *= p; } while (n >= q);
   return (long)v;
 }
@@ -1539,7 +1567,7 @@ gen_Z_divides(GEN x, GEN q, long imin)
     GEN r, xi = gel(x,i);
     if (!signe(xi)) { gel(y,i) = xi; continue; }
     gel(y,i) = dvmdii(xi, q, &r);
-    if (r != gen_0) { avma = (pari_sp)(y+l); return NULL; }
+    if (r != gen_0) { set_avma((pari_sp)(y+l)); return NULL; }
   }
   return y;
 }
@@ -1558,7 +1586,7 @@ gen_z_divides(GEN x, ulong q, long imin)
     GEN xi = gel(x,i);
     if (!signe(xi)) { gel(y,i) = xi; continue; }
     gel(y,i) = absdiviu_rem(xi, q, &r);
-    if (r) { avma = (pari_sp)(y+l); return NULL; }
+    if (r) { set_avma((pari_sp)(y+l)); return NULL; }
     affectsign_safe(xi, &gel(y,i));
   }
   return y;
@@ -1579,7 +1607,7 @@ gen_pvalrem_DC(GEN x, GEN q, GEN *py, long imin)
     GEN r, xi = gel(x,i);
     if (!signe(xi)) { gel(y,i) = xi; continue; }
     gel(y,i) = dvmdii(xi, q, &r);
-    if (r != gen_0) { avma = av; *py = x; return 0; }
+    if (r != gen_0) { *py = x; return gc_long(av,0); }
     lz = minss(lz, lgefint(gel(y,i)));
   }
   if (2 * lgefint(q) <= lz+3) /* avoid squaring if pointless */
@@ -1618,8 +1646,9 @@ gen_lval(GEN x, ulong p, long imin)
   for(v = 0;; v++)
     for (i = imin; i < lx; i++)
     {
-      ulong r; gel(y,i) = absdiviu_rem(gel(y,i), p, &r);
-      if (r) { avma = av; return v; }
+      ulong r;
+      gel(y,i) = absdiviu_rem(gel(y,i), p, &r);
+      if (r) return gc_long(av,v);
     }
 }
 long
@@ -1627,6 +1656,20 @@ ZX_lval(GEN x, ulong p) { return gen_lval(x, p, 2); }
 long
 ZV_lval(GEN x, ulong p) { return gen_lval(x, p, 1); }
 
+long
+zx_lval(GEN f, long p)
+{
+  long i, l = lg(f), x = LONG_MAX;
+  for(i=2; i<l; i++)
+  {
+    long y;
+    if (f[i] == 0) continue;
+    y = z_lval(f[i], p);
+    if (y < x) { x = y; if (x == 0) return x; }
+  }
+  return x;
+}
+
 static long
 gen_pval(GEN x, GEN p, long imin)
 {
@@ -1642,13 +1685,13 @@ gen_pval(GEN x, GEN p, long imin)
     {
       if (is_pm1(p)) pari_err_DOMAIN("gen_pval", "p", "=", p, p);
       v += gen_pvalrem_DC(y, p, &y, imin);
-      avma = av; return v;
+      return gc_long(av,v);
     }
 
     for (i = imin; i < lx; i++)
     {
       GEN r; gel(y,i) = dvmdii(gel(y,i), p, &r);
-      if (r != gen_0) { avma = av; return v; }
+      if (r != gen_0) return gc_long(av,v);
     }
   }
 }
@@ -1663,8 +1706,8 @@ ZV_Z_dvd(GEN v, GEN p)
   pari_sp av = avma;
   long i, l = lg(v);
   for (i=1; i<l; i++)
-    if (!dvdii(gel(v,i), p)) { avma = av; return 0; }
-  avma = av; return 1;
+    if (!dvdii(gel(v,i), p)) return gc_long(av,0);
+  return gc_long(av,1);
 }
 
 static long
@@ -2141,7 +2184,7 @@ padic_to_Fp(GEN x, GEN Y) {
   if (!signe(Y)) pari_err_INV("padic_to_Fp",Y);
   vy = Z_pvalrem(Y,p, &z);
   if (vx < 0 || !gequal1(z)) pari_err_OP("",x, mkintmod(gen_1,Y));
-  if (vx >= vy) { avma = av; return gen_0; }
+  if (vx >= vy) { set_avma(av); return gen_0; }
   z = gel(x,4);
   if (!signe(z) || vy > vx + precp(x)) pari_err_OP("",x, mkintmod(gen_1,Y));
   if (vx) z = mulii(z, powiu(p,vx));
@@ -2235,7 +2278,7 @@ gaffect(GEN x, GEN y)
           av = avma;
           setvalp(y, Z_pvalrem(x,gel(y,2),&p1));
           affii(modii(p1,gel(y,3)), gel(y,4));
-          avma = av; break;
+          set_avma(av); break;
 
         case t_QUAD: gaffect(x,gel(y,2)); gaffsg(0,gel(y,3)); break;
         default: pari_err_TYPE2("=",x,y);
@@ -2257,7 +2300,7 @@ gaffect(GEN x, GEN y)
         case t_INTMOD: av = avma;
           p1 = Fp_inv(gel(x,2),gel(y,1));
           affii(modii(mulii(gel(x,1),p1),gel(y,1)), gel(y,2));
-          avma = av; break;
+          set_avma(av); break;
         case t_COMPLEX: gaffect(x,gel(y,1)); gaffsg(0,gel(y,2)); break;
         case t_PADIC:
           if (!signe(gel(x,1))) { padicaff0(y); break; }
@@ -2267,7 +2310,7 @@ gaffect(GEN x, GEN y)
           if (!vx) vx = -Z_pvalrem(den,gel(y,2),&den);
           setvalp(y,vx);
           p1 = mulii(num,Fp_inv(den,gel(y,3)));
-          affii(modii(p1,gel(y,3)), gel(y,4)); avma = av; break;
+          affii(modii(p1,gel(y,3)), gel(y,4)); set_avma(av); break;
         case t_QUAD: gaffect(x,gel(y,2)); gaffsg(0,gel(y,3)); break;
         default: pari_err_TYPE2("=",x,y);
       }
@@ -2283,7 +2326,7 @@ gaffect(GEN x, GEN y)
       {
         case t_INTMOD:
           av = avma; affii(padic_to_Fp(x, gel(y,1)), gel(y,2));
-          avma = av; break;
+          set_avma(av); break;
         default: pari_err_TYPE2("=",x,y);
       }
       break;
@@ -2295,10 +2338,10 @@ gaffect(GEN x, GEN y)
           pari_err_TYPE2("=",x,y);
 
         case t_REAL:
-          av = avma; affgr(quadtofp(x,realprec(y)), y); avma = av; break;
+          av = avma; affgr(quadtofp(x,realprec(y)), y); set_avma(av); break;
         case t_COMPLEX:
           ly = precision(y); if (!ly) pari_err_TYPE2("=",x,y);
-          av = avma; gaffect(quadtofp(x,ly), y); avma = av; break;
+          av = avma; gaffect(quadtofp(x,ly), y); set_avma(av); break;
         default: pari_err_TYPE2("=",x,y);
       }
     default: pari_err_TYPE2("=",x,y);
@@ -2661,12 +2704,12 @@ gsigne(GEN x)
       if (signe(gel(T,3))) a = gadd(a,b);
       /* a + b sqrt(D) > 0 ? */
       sa = gsigne(a);
-      sb = gsigne(b); if (sa == sb) { avma = av; return sa; }
-      if (sa == 0) { avma = av; return sb; }
-      if (sb == 0) { avma = av; return sa; }
+      sb = gsigne(b); if (sa == sb) return gc_int(av,sa);
+      if (sa == 0) return gc_int(av,sb);
+      if (sb == 0) return gc_int(av,sa);
       /* different signs, take conjugate expression */
       sb = gsigne(gsub(gsqr(a), gmul(quad_disc(x), gsqr(b))));
-      avma = av; return sb * sa;
+      return gc_int(av, sb*sa);
     }
     case t_INFINITY: return inf_get_sign(x);
   }
@@ -2801,7 +2844,7 @@ listinsert(GEN L, GEN x, long index)
     pari_err_TYPE("listinsert",L);
   z = list_data(L); l = z? lg(z): 1;
   if (index <= 0) pari_err_COMPONENT("listinsert", "<=", gen_0, stoi(index));
-  if (index > l) pari_err_COMPONENT("listinsert", ">", stoi(l), stoi(index));
+  if (index > l) index = l;
   ensure_nb(L, l);
   BLOCK_SIGINT_START
   z = list_data(L);
@@ -2898,5 +2941,5 @@ listsort(GEN L, long flag)
     for (i=1; i<l; i++) gel(vnew,i) = gel(v,perm[i]);
   }
   for (i=1; i<l; i++) gel(v,i) = gel(vnew,i);
-  v[0] = vnew[0]; avma = av;
+  v[0] = vnew[0]; set_avma(av);
 }
