@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -136,7 +137,7 @@ hnffinal(GEN matgen,GEN perm,GEN* ptdep,GEN* ptB,GEN* ptC)
       for (   ; k<=lig;  k++) gel(z,k) = subii(gel(z,k), mulii(p1, gel(Hi,k-nlze)));
       gel(Cnew,j) = gsub(gel(Cnew,j), gmul(p1, gel(Cnew,i+zc)));
     }
-    if (gc_needed(av,3))
+    if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"hnffinal, i = %ld",i);
       gerepileall(av, 2, &Cnew, &B);
@@ -205,7 +206,7 @@ p_mat(GEN mat, GEN perm, long k)
   err_printf("Permutation: %Ps\n",perm);
   if (DEBUGLEVEL > 6)
     err_printf("matgen = %Ps\n", zm_to_ZM( rowpermute(mat, perm) ));
-  avma = av;
+  set_avma(av);
 }
 
 static GEN
@@ -225,7 +226,7 @@ ZM_rowrankprofile(GEN x, long *nlze)
 
   x = shallowtrans(x); l = lg(x);
   (void)new_chunk(l); /* HACK */
-  d = ZM_pivots(x,&r); avma = av;
+  d = ZM_pivots(x,&r); set_avma(av);
   *nlze = r;
   if (!d) return identity_perm(l-1);
   y = cgetg(l,t_VECSMALL);
@@ -283,7 +284,7 @@ hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
 
   i = lig = li-1; col = co-1; lk0 = k0;
   T = (k0 || (lg(C) > 1 && lgcols(C) > 1))? matid(col): NULL;
-  /* Look for lines with a single non-0 entry, equal to 1 in absolute value */
+  /* Look for lines with a single nonzero entry, equal to 1 in absolute value */
   while (i > lk0 && col)
     switch( count(mat,perm[i],col,&n) )
     {
@@ -301,7 +302,7 @@ hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
           if (T)
           {
             p1 = gel(T,col);
-            for (i=1; ; i++) /* T is a permuted identity: single non-0 entry */
+            for (i=1; ; i++) /* T = permuted identity: single nonzero entry */
               if (signe(gel(p1,i))) { togglesign_safe(&gel(p1,i)); break; }
           }
         }
@@ -322,7 +323,7 @@ hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
       if (count(mat,perm[i],col,&n) > 0) break;
     if (i == lk0) break;
 
-    /* only 0, +/- 1 entries, at least 2 of them non-zero */
+    /* only 0, +/- 1 entries, at least 2 of them nonzero */
     lswap(perm[i], perm[lig]);
     swap(gel(mat,n), gel(mat,col)); p = gel(mat,col);
     if (T) swap(gel(T,n), gel(T,col));
@@ -348,7 +349,7 @@ hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
     if (gc_needed(av,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"hnfspec[1]");
-      if (T) T = gerepilecopy(av, T); else avma = av;
+      if (T) T = gerepilecopy(av, T); else set_avma(av);
     }
   }
   /* As above with lines containing a +/- 1 (no other assumption).
@@ -431,13 +432,12 @@ END2: /* clean up mat: remove everything to the right of the 1s on diagonal */
         for (h=1; h<i0; h++) gel(Bj,h) = subii(gel(Bj,h), mulii(v,gel(Bk,h)));
       }
       if (T) ZC_lincomb1_inplace(gel(T,j), gel(T,k), negi(v));
-      if (gc_needed(av,3))
-      {
-        if(DEBUGMEM>1) pari_warn(warnmem,"hnfspec[3], (i,j) = %ld,%ld", i,j);
-        for (h=1; h<co; h++) setlg(matb[h], i0+1); /* bottom can be forgotten */
-        gerepileall(av, T? 2: 1, &matb, &T);
-        Bk = gel(matb,k);
-      }
+    }
+    if (gc_needed(av,2))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"hnfspec[3], i = %ld", i);
+      for (h=1; h<co; h++) setlg(matb[h], i0+1); /* bottom can be forgotten */
+      gerepileall(av, T? 2: 1, &matb, &T);
     }
   }
   for (j=1; j<co; j++) setlg(matb[j], lig-k0+1); /* bottom can be forgotten */
@@ -445,7 +445,7 @@ END2: /* clean up mat: remove everything to the right of the 1s on diagonal */
   if (DEBUGLEVEL>5) err_printf("    matb cleaned up (using Id block)\n");
 
   nlze = lk0 - k0;  /* # of 0 rows */
-  lnz = lig-nlze+1; /* 1 + # of non-0 rows (!= 0...0 1 0 ... 0) */
+  lnz = lig-nlze+1; /* 1 + # of nonzero rows (!= 0...0 1 0 ... 0) */
   if (T) matt = ZM_mul(matt,T); /* update top rows */
   extramat = cgetg(col+1,t_MAT); /* = new C minus the 0 rows */
   for (j=1; j<=col; j++)
@@ -454,7 +454,7 @@ END2: /* clean up mat: remove everything to the right of the 1s on diagonal */
     GEN t = (gel(matb,j)) + nlze - k0;
     p2=cgetg(lnz,t_COL); gel(extramat,j) = p2;
     for (i=1; i<=k0; i++) gel(p2,i) = gel(z,i); /* top k0 rows */
-    for (   ; i<lnz; i++) gel(p2,i) = gel(t,i); /* other non-0 rows */
+    for (   ; i<lnz; i++) gel(p2,i) = gel(t,i); /* other nonzero rows */
   }
   if (!col) {
     permpro = identity_perm(lnz);
@@ -516,7 +516,7 @@ END2: /* clean up mat: remove everything to the right of the 1s on diagonal */
       gel(p1,k) = (i <= k0)? gel(y,i): gel(z,i);
     }
   }
-  if (T) C = typ(C)==t_MAT? RgM_mul(C,T): RgV_RgM_mul(C,T);
+  if (T) C = typ(C)==t_MAT? RgM_ZM_mul(C,T): RgV_RgM_mul(C,T);
   gerepileall(av, 4, &matbnew, &B, &dep, &C);
   *ptdep = dep;
   *ptB = B;
@@ -1464,7 +1464,7 @@ ZM_hnfmodprime(GEN x, GEN p)
   l = lgcols(x);
   x = FpM_echelon(x, &P, p);
   lP = lg(P); /* rank = lP-1 */
-  if (lP == l) { avma = av; return matid(l-1); }
+  if (lP == l) { set_avma(av); return matid(l-1); }
   y = scalarmat_shallow(p, l-1);
   for (i = 1; i < lP; i++) gel(y,P[i]) = gel(x,i);
   return gerepilecopy(av, FpM_hnfend(av,y,p));
@@ -1546,7 +1546,7 @@ Minus(long j, GEN lambda)
   for (k=j+1; k<n; k++) togglesign_safe(&gcoeff(lambda,j,k));
 }
 
-/* index of first non-zero entry */
+/* index of first nonzero entry */
 static long
 findi(GEN M)
 {
@@ -1658,8 +1658,7 @@ must_swap(long k, GEN lambda, GEN D)
 {
   pari_sp av = avma;
   GEN z = addii(mulii(gel(D,k-2),gel(D,k)), sqri(gcoeff(lambda,k-1,k)));
-  long s = cmpii(z, sqri(gel(D,k-1)));
-  avma = av; return s < 0;
+  return gc_bool(av, cmpii(z, sqri(gel(D,k-1))) < 0);
 }
 
 GEN
@@ -1695,13 +1694,6 @@ ZM_hnflll(GEN A, GEN *ptB, int remove)
       {
         long row0, row1;
         reduce2(A,B,k,i,&row0,&row1,lambda,D);
-        if (gc_needed(av,3))
-        {
-          GEN b = D-1;
-          if (DEBUGMEM>1) pari_warn(warnmem,"hnflll (reducing), kmax = %ld",kmax);
-          gerepileall(av, B? 4: 3, &A, &lambda, &b, &B);
-          D = b+1;
-        }
       }
       if (++k > kmax) kmax = k;
     }
@@ -1710,6 +1702,7 @@ ZM_hnflll(GEN A, GEN *ptB, int remove)
       GEN b = D-1;
       if (DEBUGMEM>1) pari_warn(warnmem,"hnflll, kmax = %ld / %ld",kmax,n-1);
       gerepileall(av, B? 4: 3, &A, &lambda, &b, &B);
+      if (gc_needed(av,1)) paristack_resize(0); /* avoid desperation GC */
       D = b+1;
     }
   }
@@ -2002,12 +1995,12 @@ ZM_hnfall_i(GEN A, GEN *ptB, long remove)
         /* zero a = Aij  using  Aik */
         if (signe(a)) ZC_elem(a,gcoeff(A,i,k), A,B,j,k);
         ZM_reduce(A,B, i,k); /* ensure reduced entries even if a = 0 */
-        if (gc_needed(av,1))
-        {
-          if (DEBUGMEM>1)
-            pari_warn(warnmem,"ZM_hnfall[1], li = %ld, j = %ld", li, j);
-          gerepileall(av, B? 2: 1, &A, &B);
-        }
+      }
+      if (gc_needed(av,1) && (j & 0x7f) == 0)
+      {
+        if (DEBUGMEM>1)
+          pari_warn(warnmem,"ZM_hnfall[1], li = %ld, j = %ld", li, j);
+        gerepileall(av, B? 2: 1, &A, &B);
       }
       if (signe( gcoeff(A,li,j) )) break;
       h[j] = li-1;
@@ -2036,18 +2029,20 @@ ZM_hnfall_i(GEN A, GEN *ptB, long remove)
   if (DEBUGLEVEL>5) err_printf("\nhnfall, final phase: ");
   r--; /* first r cols are in the image the n-r (independent) last ones */
   for (j=1; j<=r; j++)
+  {
     for (i=h[j]; i; i--)
     {
       a = gcoeff(A,i,j);
       k = c[i];
       if (signe(a)) ZC_elem(a,gcoeff(A,i,k), A,B, j,k);
       ZM_reduce(A,B, i,k); /* ensure reduced entries, even if a = 0 */
-      if (gc_needed(av,1))
-      {
-        if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfall[3], j = %ld", j);
-        gerepileall(av, B? 2: 1, &A, &B);
-      }
     }
+    if (gc_needed(av,1) && (j & 0x7f) == 0)
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfall[3], j = %ld", j);
+      gerepileall(av, B? 2: 1, &A, &B);
+    }
+  }
   if (DEBUGLEVEL>5) err_printf("\n");
   if (remove) remove_0cols(r, &A, &B, remove);
   if (ptB) *ptB = B;
@@ -2136,17 +2131,17 @@ hnfdivide(GEN A, GEN B)
     b = gel(B,k);
     m = gel(b,k);
     gel(u,k) = dvmdii(m, gcoeff(A,k,k), &r);
-    if (r != gen_0) { avma = av; return 0; }
+    if (r != gen_0) return gc_long(av, 0);
     for (i=k-1; i>0; i--)
     {
       m = gel(b,i);
       for (j=i+1; j<=k; j++) m = subii(m, mulii(gcoeff(A,i,j),gel(u,j)));
       m = dvmdii(m, gcoeff(A,i,i), &r);
-      if (r != gen_0) { avma = av; return 0; }
+      if (r != gen_0) return gc_long(av, 0);
       gel(u,i) = m;
     }
   }
-  avma = av; return 1;
+  return gc_long(av, 1);
 }
 
 /* A upper HNF, b integral vector. Return A^(-1) b if integral,
@@ -2170,11 +2165,11 @@ hnf_invimage(GEN A, GEN b)
     for (j=i+1; j<=n; j++) t = subii(t, mulii(gcoeff(A,k,j),gel(u,j)));
     if (!signe(Aki))
     {
-      if (signe(t)) { avma = av;return NULL; }
-      avma = av2; gel(u,i) = gen_0; continue;
+      if (signe(t)) return gc_NULL(av);
+      set_avma(av2); gel(u,i) = gen_0; continue;
     }
     t = dvmdii(t, Aki, &r);
-    if (r != gen_0) { avma = av; return NULL; }
+    if (r != gen_0) return gc_NULL(av);
     gel(u,i) = gerepileuptoint(av2, t);
     if (--i == 0) break;
   }
@@ -2186,8 +2181,8 @@ hnf_invimage(GEN A, GEN b)
     GEN t = gel(b,k);
     if (typ(t) != t_INT) pari_err_TYPE("hnf_invimage",t);
     for (j=1; j<=n; j++) t = subii(t, mulii(gcoeff(A,k,j),gel(u,j)));
-    if (signe(t)) { avma = av;return NULL; }
-    avma = av2;
+    if (signe(t)) return gc_NULL(av);
+    set_avma(av2);
   }
   return u;
 }
@@ -2203,9 +2198,10 @@ hnf_solve(GEN A, GEN B)
 
   if (typ(B) == t_COL) return hnf_invimage(A, B);
   av = avma; C = cgetg_copy(B, &l);
-  for (i = 1; i < l; i++) {
+  for (i = 1; i < l; i++)
+  {
     GEN c = hnf_invimage(A, gel(B,i));
-    if (!c) { avma = av; return NULL; }
+    if (!c) return gc_NULL(av);
     gel(C,i) = c;
   }
   return C;
@@ -2228,6 +2224,14 @@ trivsmith(long all)
   gel(z,3) = cgetg(1,t_MAT); return z;
 }
 
+static void
+snf_pile1(pari_sp av, GEN *x, GEN *U)
+{
+  GEN *gptr[2];
+  int c = 1; gptr[0]=x;
+  if (*U) gptr[c++] = U;
+  gerepilemany(av,gptr,c);
+}
 static void
 snf_pile(pari_sp av, GEN *x, GEN *U, GEN *V)
 {
@@ -2261,21 +2265,14 @@ bezout_step(GEN *pa, GEN *pb, GEN *pu, GEN *pv)
 static int
 negcmpii(void *E, GEN x, GEN y) { (void)E; return -cmpii(x,y); }
 
-/* does b = x[i,i] divide all entries in x[1..i-1,1..i-1] ? If so, return 0;
- * else return the index of a problematic row */
+/* x square of maximal rank; does b = x[i,i] divide all entries in
+ * x[1..i-1, 1..i-1] ? If so, return 0; else the index of a problematic row */
 static long
 ZM_snf_no_divide(GEN x, long i)
 {
   GEN b = gcoeff(x,i,i);
   long j, k;
 
-  if (!signe(b))
-  { /* impossible in the current implementation : x square of maximal rank */
-    for (k = 1; k < i; k++)
-      for (j = 1; j < i; j++)
-        if (signe(gcoeff(x,k,j))) return k;
-    return 0;
-  }
   if (is_pm1(b)) return 0;
   for (k = 1; k < i; k++)
     for (j = 1; j < i; j++)
@@ -2283,14 +2280,33 @@ ZM_snf_no_divide(GEN x, long i)
   return 0;
 }
 
+static void
+ZM_redpart(GEN x, GEN p, long I)
+{
+  long l = lgefint(p), i, j;
+  for (i = 1; i <= I; i++)
+    for (j = 1; j <= I; j++)
+    {
+      GEN c = gcoeff(x,i,j);
+      if (lgefint(c) > l) gcoeff(x,i,j) = remii(c, p);
+    }
+}
+static void
+ZMrow_divexact_inplace(GEN M, long i, GEN c)
+{
+  long j, l = lg(M);
+  for (j = 1; j < l; j++) gcoeff(M,i,j) = diviiexact(gcoeff(M,i,j), c);
+}
+
 /* Return the SNF D of matrix X. If ptU/ptV non-NULL set them to U/V
  * to that D = UXV */
 GEN
-ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
+ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, long flag)
 {
   pari_sp av0 = avma, av;
+  const long return_vec = flag & 1;
   long i, j, k, m0, m, n0, n;
-  GEN p1, u, v, U, V, V0, mdet, ys, perm = NULL;
+  GEN u, v, U, V, V0, mdet, A = NULL, perm = NULL;
 
   n0 = n = lg(x)-1;
   if (!n) {
@@ -2298,101 +2314,82 @@ ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
     if (ptV) *ptV = cgetg(1,t_MAT);
     return cgetg(1, return_vec? t_VEC: t_MAT);
   }
-  av = avma;
   m0 = m = nbrows(x);
 
-  U = ptU? gen_1: NULL; /* TRANSPOSE of row transform matrix [act on columns]*/
-  V = ptV? gen_1: NULL;
-  V0 = NULL;
-  x = RgM_shallowcopy(x);
+  U = V = V0 = NULL; /* U = TRANSPOSE of row transform matrix [act on columns]*/
   if (m == n && ZM_ishnf(x))
   {
-    mdet = ZM_det_triangular(x);
-    if (V) *ptV = matid(n);
+    mdet = ZM_det_triangular(x); /* != 0 */
+    if (ptV) *ptV = matid(n);
   }
   else
   {
     mdet = ZM_detmult(x);
-    if (signe(mdet))
-    {
-      if (!V)
-        p1 = ZM_hnfmod(x,mdet);
-      else
-      {
-        if (m == n)
-        {
-          p1 = ZM_hnfmod(x,mdet);
-          *ptV = RgM_solve(x,p1);
-        }
-        else
-          p1 = ZM_hnfperm(x, ptV, ptU? &perm: NULL);
-      }
-      mdet = ZM_det_triangular(p1);
-    }
+    if (!signe(mdet))
+      x = ZM_hnfperm(x, ptV, ptU? &perm: NULL);
     else
-      p1 = ZM_hnfperm(x, ptV, ptU? &perm: NULL);
-    x = p1;
-  }
-  n = lg(x)-1;
-  if (V)
-  {
-    V = *ptV;
-    if (n != n0)
+    { /* m <= n */
+      if (!ptV)
+        x = ZM_hnfmod(x,mdet);
+      else if (m == n)
+      {
+        GEN H = ZM_hnfmod(x,mdet);
+        *ptV = ZM_gauss(x,H);
+        x = H;
+      }
+      else
+        x = ZM_hnfperm(x, ptV, ptU? &perm: NULL);
+      mdet = ZM_det_triangular(x); /* > 0 */
+    }
+    n = lg(x)-1; /* n independent columns */
+    if (ptV)
     {
-      V0 = vecslice(V, 1, n0 - n); /* kernel */
-      V  = vecslice(V, n0-n+1, n0);
-      av = avma;
+      V = *ptV;
+      if (n != n0)
+      {
+        V0 = vecslice(V, 1, n0 - n); /* kernel */
+        V  = vecslice(V, n0-n+1, n0);
+      }
+    }
+    if (!signe(mdet))
+    {
+      if (n)
+      {
+        x = ZM_snfall_i(shallowtrans(x), ptV, ptU, return_vec); /* swap V,U */
+        if (!return_vec && n != m) x = shallowtrans(x);
+        if (ptV) V = ZM_mul(V, shallowtrans(*ptV));
+        if (ptU) U = *ptU; /* TRANSPOSE */
+      }
+      else /* 0 matrix */
+      {
+        x = cgetg(1,t_MAT);
+        if (ptV) V = cgetg(1, t_MAT);
+        if (ptU) U = matid(m);
+      }
+      goto THEEND;
     }
   }
-  /* independent columns */
-  if (!signe(mdet))
-  {
-    if (n)
-    {
-      x = ZM_snfall_i(shallowtrans(x), ptV, ptU, return_vec); /* swap ptV,ptU */
-      if (typ(x) == t_MAT && n != m) x = shallowtrans(x);
-      if (V) V = ZM_mul(V, shallowtrans(*ptV));
-      if (U) U = *ptU; /* TRANSPOSE */
-    }
-    else /* 0 matrix */
-    {
-      x = cgetg(1,t_MAT);
-      if (V) V = cgetg(1, t_MAT);
-      if (U) U = matid(m);
-    }
-    goto THEEND;
-  }
-  if (U) U = matid(n);
+  if (ptV || ptU) U = matid(n); /* we will compute V in terms of U */
+  if (DEBUGLEVEL>7) err_printf("starting SNF loop");
 
   /* square, maximal rank n */
-  p1 = gen_indexsort(RgM_diagonal_shallow(x), NULL, &negcmpii);
-  ys = cgetg(n+1,t_MAT);
-  for (j=1; j<=n; j++) gel(ys,j) = vecpermute(gel(x, p1[j]), p1);
-  x = ys;
-  if (U) U = vecpermute(U, p1);
-  if (V) V = vecpermute(V, p1);
-
-  p1 = ZM_hnfmod(x, mdet);
-  if (V) V = ZM_mul(V, RgM_solve(x,p1));
-  x = p1;
-
-  if (DEBUGLEVEL>7) err_printf("starting SNF loop");
-  for (i=n; i>1; i--)
+  A = x; x = shallowcopy(x); av = avma;
+  for (i = n; i > 1; i--)
   {
     if (DEBUGLEVEL>7) err_printf("\ni = %ld: ",i);
     for(;;)
     {
       int c = 0;
       GEN a, b;
-      for (j=i-1; j>=1; j--)
+      for (j = i-1; j >= 1; j--)
       {
         b = gcoeff(x,i,j); if (!signe(b)) continue;
         a = gcoeff(x,i,i);
-        ZC_elem(b, a, x,V, j,i);
+        ZC_elem(b, a, x,NULL, j,i);
         if (gc_needed(av,1))
         {
           if (DEBUGMEM>1) pari_warn(warnmem,"[1]: ZM_snfall i = %ld", i);
-          snf_pile(av, &x,&U,&V);
+          snf_pile1(av, &x,&U);
         }
       }
       if (DEBUGLEVEL>7) err_printf("; ");
@@ -2415,7 +2412,7 @@ ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
         if (gc_needed(av,1))
         {
           if (DEBUGMEM>1) pari_warn(warnmem,"[2]: ZM_snfall, i = %ld", i);
-          snf_pile(av, &x,&U,&V);
+          snf_pile1(av, &x,&U);
         }
         c = 1;
       }
@@ -2425,25 +2422,47 @@ ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
         if (!k) break;
 
         /* x[k,j] != 0 mod b */
-        for (j=1; j<=i; j++)
+        for (j = 1; j <= i; j++)
           gcoeff(x,i,j) = addii(gcoeff(x,i,j),gcoeff(x,k,j));
         if (U) gel(U,i) = gadd(gel(U,i),gel(U,k));
       }
+      ZM_redpart(x, mdet, i);
+      if (U && (flag & 2)) ZM_redpart(U, mdet, n);
       if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"[3]: ZM_snfall");
-        snf_pile(av, &x,&U,&V);
+        snf_pile1(av, &x,&U);
       }
     }
   }
   if (DEBUGLEVEL>7) err_printf("\n");
-  for (k=1; k<=n; k++)
-    if (signe(gcoeff(x,k,k)) < 0)
-    {
-      if (V) ZV_togglesign(gel(V,k));
-      togglesign(gcoeff(x,k,k));
-    }
+  for (k = n; k; k--)
+  {
+    GEN d = gcdii(gcoeff(x,k,k), mdet);
+    gcoeff(x,k,k) = d;
+    if (!is_pm1(d)) mdet = diviiexact(mdet,d);
+  }
 THEEND:
+  if (U) U = shallowtrans(U);
+  if (ptV && A)
+  { /* U A V = D => D^(-1) U A = V^(-1) */
+    long l = lg(x);
+    GEN W = ZM_mul(U, A);
+    for (i = 1; i < l; i++)
+    {
+      GEN c = gcoeff(x,i,i);
+      if (is_pm1(c)) break; /* only 1 from now on */
+      ZMrow_divexact_inplace(W, i, c);
+    }
+    if (flag & 2)
+    {
+      W = FpM_red(W, gcoeff(x,1,1));
+      W = matinvmod(W, gcoeff(x,1,1));
+    }
+    else
+      W = ZM_inv(W, NULL);
+    V = V? ZM_mul(V, W): W;
+  }
   if (return_vec)
   {
     long l = lg(x)-1;
@@ -2452,15 +2471,11 @@ THEEND:
   }
 
   if (V0)
-  {
+  { /* add kernel */
     if (!return_vec) x = shallowconcat(zeromat(m,n0-n), x);
-    if (V) V = shallowconcat(V0, V);
-  }
-  if (U)
-  {
-    U = shallowtrans(U);
-    if (perm) U = vecpermute(U, perm_inv(perm));
+    if (ptV) V = shallowconcat(V0, V);
   }
+  if (perm && U) U = vecpermute(U, perm_inv(perm));
   snf_pile(av0, &x,&U,&V);
   if (ptU) *ptU = U;
   if (ptV) *ptV = V;
@@ -2470,19 +2485,12 @@ GEN
 ZM_snfall(GEN x, GEN *U, GEN *V) { return ZM_snfall_i(x, U, V, 0); }
 GEN
 ZM_snf(GEN x) { return ZM_snfall_i(x, NULL,NULL, 1); }
-
 GEN
-smith(GEN x) {
-  if (typ(x)!=t_MAT) pari_err_TYPE("smith",x);
-  RgM_check_ZM(x, "smith");
-  return ZM_snfall_i(x, NULL,NULL, 1);
-}
+smith(GEN x) { return ZM_snfall_i(x, NULL,NULL, 1); }
 GEN
 smithall(GEN x)
 {
   GEN z = cgetg(4, t_VEC);
-  if (typ(x)!=t_MAT) pari_err_TYPE("smithall",x);
-  RgM_check_ZM(x, "smithall");
   gel(z,3) = ZM_snfall_i(x, (GEN*)(z+1),(GEN*)(z+2), 0);
   return z;
 }
@@ -2658,16 +2666,14 @@ RgM_hnfall(GEN A, GEN *pB, long remove)
 }
 
 static GEN
-gsmithall_i(GEN x,long all)
+RgXM_snf(GEN x,long all)
 {
   pari_sp av;
   long i, j, k, n;
   GEN z, u, v, U, V;
   long vx = gvar(x);
-  if (typ(x)!=t_MAT) pari_err_TYPE("gsmithall",x);
-  if (vx==NO_VARIABLE) return all? smithall(x): smith(x);
-  n = lg(x)-1;
-  if (!n) return trivsmith(all);
+  n = lg(x)-1; if (!n) return trivsmith(all);
+  if (vx==NO_VARIABLE) pari_err_TYPE("RgXM_snf",x);
   if (lgcols(x) != n+1) pari_err_DIM("gsmithall");
   av = avma;
   x = RgM_shallowcopy(x);
@@ -2742,17 +2748,16 @@ matsnf0(GEN x,long flag)
   pari_sp av = avma;
   if (flag > 7) pari_err_FLAG("matsnf");
   if (typ(x) == t_VEC && flag & 4) return smithclean(x);
-  if (flag & 2) x = flag&1 ? gsmithall(x): gsmith(x);
-  else          x = flag&1 ?  smithall(x):  smith(x);
+  if (typ(x)!=t_MAT) pari_err_TYPE("matsnf",x);
+  if (RgM_is_ZM(x)) x = flag&1 ? smithall(x): smith(x);
+  else              x = RgXM_snf(x, flag&1);
   if (flag & 4) x = gerepileupto(av, smithclean(x));
   return x;
 }
-
 GEN
-gsmith(GEN x) { return gsmithall_i(x,0); }
-
+gsmith(GEN x) { return RgXM_snf(x,0); }
 GEN
-gsmithall(GEN x) { return gsmithall_i(x,1); }
+gsmithall(GEN x) { return RgXM_snf(x,1); }
 
 /* H is a relation matrix, either in HNF or a t_VEC (diagonal HNF) */
 static GEN
@@ -2800,7 +2805,7 @@ snf_group(GEN H, GEN D, GEN *newU, GEN *newUi)
 GEN
 ZM_snf_group(GEN H, GEN *newU, GEN *newUi)
 {
-  GEN D = ZM_snfall_i(H, newU, newUi, 1);
+  GEN D = ZM_snfall_i(H, newU, newUi, 1 + 2);
   return snf_group(H, D, newU, newUi);
 }
 
@@ -2825,7 +2830,7 @@ ZV_snfall(GEN D, GEN *pU, GEN *pV)
       if (V) ZV_togglesign(gel(V,j));
     }
   }
-  /* entries are non-negative integers */
+  /* entries are nonnegative integers */
   p = gen_indexsort(D, NULL, &negcmpii);
   D = vecpermute(D, p);
   if (U) U = vecpermute(U, p);
