@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -163,7 +164,7 @@ ZpXXQ_frob(GEN S, GEN U, GEN V, long k, GEN T, ulong p, long e)
     A = FpX_add(FpX_mul(U,M,q), FpX_mul(H, Tp, q),q);
     v = u_lvalrem(2*i+1,p,&r);
     Bc = ZX_deriv(B);
-    Bc = FpX_Fp_mul(ZX_Z_divexact(Bc,powuu(p,v)),Fp_div(gen_2, utoi(r), q), q);
+    Bc = FpX_Fp_mul(ZX_divuexact(Bc,upowuu(p,v)),Fp_divu(gen_2, r, q), q);
     M = FpX_add(to_ZX(gel(S,i+2),vT), FpX_add(A, Bc, q), q);
     if (gc_needed(av2,1))
     {
@@ -455,7 +456,7 @@ ZpXQXXQ_frob(GEN F, GEN U, GEN V, long k, GEN S, GEN T, ulong p, long e)
     A = FpXX_add(FpXQX_mul(U, M, T, q), FpXQX_mul(H, Sp, T, q),q);
     v = u_lvalrem(2*i+1,p,&r);
     Bc = RgX_deriv(B);
-    Bc = FpXX_Fp_mul(ZXX_Z_divexact(Bc,powuu(p,v)), Fp_div(gen_2, utoi(r), q), q);
+    Bc = FpXX_Fp_mul(ZXX_Z_divexact(Bc,powuu(p,v)), Fp_divu(gen_2, r, q), q);
     M = FpXX_add(gel(F,i+2), FpXX_add(A, Bc, q), q);
     if (gc_needed(av2,1))
     {
@@ -491,7 +492,6 @@ ZpXQXXQ_frob(GEN F, GEN U, GEN V, long k, GEN S, GEN T, ulong p, long e)
     return gerepilecopy(av, R);
 }
 
-
 static GEN
 Fq_diff_red(GEN s, GEN A, long m, GEN S, GEN T, GEN q, GEN p, long e)
 {
@@ -607,6 +607,16 @@ nfhyperellpadicfrobenius(GEN H, GEN T, ulong p, long n)
   return gerepileupto(av, m);
 }
 
+GEN
+hyperellpadicfrobenius0(GEN H, GEN Tp, long n)
+{
+  GEN T, p;
+  if (!ff_parse_Tp(Tp, &T,&p,0)) pari_err_TYPE("hyperellpadicfrobenius", Tp);
+  if (lgefint(p) > 3) pari_err_IMPL("large prime in hyperellpadicfrobenius");
+  return T? nfhyperellpadicfrobenius(H, T, itou(p), n)
+          : hyperellpadicfrobenius(H, itou(p), n);
+}
+
 static GEN
 F2x_genus2charpoly_naive(GEN P, GEN Q)
 {
@@ -683,14 +693,14 @@ Flx_genus2charpoly_naive(GEN H, ulong p)
     {
       GEN r2 = gel(V, n+1);
       c += uel(r2,2) ?
-        (uel(r2,1) ? k[1+Fl2_norm_pre(r2, D, p, pi)]: e)
+        (uel(r2,1) ? uel(k,1+Fl2_norm_pre(r2, D, p, pi)): e)
          : !!uel(r2,1);
       if (i == p-1) break;
       V = Fl2V_next(V, p);
     }
-    avma = av2;
+    set_avma(av2);
   }
-  avma = av;
+  set_avma(av);
   return mkvecsmalln(6, 0UL, p*p, a*p, (b+2*c+a*a)>>1, a, 1UL);
 }
 
@@ -712,8 +722,7 @@ hyperell_Weil_bound(GEN q, ulong g, GEN p)
 {
   pari_sp av = avma;
   GEN w = mulii(binomialuu(2*g,g),sqrtint(shifti(powiu(q, g),2)));
-  long e = logint(w, p) + 1;
-  avma = av; return e;
+  return gc_long(av, logint(w,p) + 1);
 }
 
 GEN
@@ -792,3 +801,38 @@ hyperellcharpoly(GEN PQ)
   R = charpoly_funceq(R, q);
   return gerepilecopy(av, R);
 }
+
+GEN
+hyperell_redsl2(GEN Q)
+{
+  pari_sp av = avma;
+  long d = degpol(Q), v = varn(Q);
+  GEN den, P = Q_primitive_part(Q, &den);
+  GEN a = gel(P,d+2), b = gel(P,d+1), c = gel(P, d);
+  GEN q1, q2, q3, D, M, A, B, R;
+  q1 = mulis(sqri(a), d);
+  q2 = shifti(mulii(a,b), 1);
+  q3 = subii(sqri(b),shifti(mulii(a,c), 1));
+  D = gcdii(gcdii(q1, q2), q3);
+  if (!equali1(D))
+  {
+    q1 = diviiexact(q1,D);
+    q2 = diviiexact(q2,D);
+    q3 = diviiexact(q3,D);
+  }
+  D = qfb_disc3(q1, q2, q3);
+  if (!signe(D))
+    M = mkmat22(gen_1,truedivii(negi(q2),shifti(q1,1)),gen_0,gen_1);
+  else if (issquare(D))
+    M = mkmat22(gen_1,divii(negi(q2),shifti(q1,1)),gen_0,gen_1);
+  else
+  {
+    GEN Q = signe(D) < 0 ? qfi(q1,q2,q3): qfr(q1,q2,q3,real_0(DEFAULTPREC));
+    M = gel(qfbredsl2(Q, NULL),2);
+  }
+  A = deg1pol(gcoeff(M,1,1),gcoeff(M,1,2),v);
+  B = gpowers(deg1pol(gcoeff(M,2,1),gcoeff(M,2,2),v), d);
+  R = gel(RgX_homogenous_evalpow(P, A, B),1);
+  if (den) R = gmul(R, den);
+  return gerepilecopy(av,mkvec2(R,M));
+}
