@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -21,40 +22,31 @@ static const double LLLDFT = 0.99;
 static GEN
 lll_trivial(GEN x, long flag)
 {
-  GEN y;
   if (lg(x) == 1)
   { /* dim x = 0 */
     if (! (flag & LLL_ALL)) return cgetg(1,t_MAT);
-    y=cgetg(3,t_VEC);
-    gel(y,1) = cgetg(1,t_MAT);
-    gel(y,2) = cgetg(1,t_MAT); return y;
+    retmkvec2(cgetg(1,t_MAT), cgetg(1,t_MAT));
   }
   /* dim x = 1 */
   if (gequal0(gel(x,1)))
   {
     if (flag & LLL_KER) return matid(1);
     if (flag & (LLL_IM|LLL_INPLACE)) return cgetg(1,t_MAT);
-    y = cgetg(3,t_VEC);
-    gel(y,1) = matid(1);
-    gel(y,2) = cgetg(1,t_MAT); return y;
+    retmkvec2(matid(1), cgetg(1,t_MAT));
   }
   if (flag & LLL_INPLACE) return gcopy(x);
   if (flag & LLL_KER) return cgetg(1,t_MAT);
   if (flag & LLL_IM)  return matid(1);
-  y=cgetg(3,t_VEC);
-  gel(y,1) = cgetg(1,t_MAT);
-  gel(y,2) = (flag & LLL_GRAM)? gcopy(x): matid(1);
-  return y;
+  retmkvec2(cgetg(1,t_MAT), (flag & LLL_GRAM)? gcopy(x): matid(1));
 }
 
-/* vecslice(h,#h-k,#h) in place. Works for t_MAT, t_VEC/t_COL */
+/* vecslice(x,#x-k,#x) in place. Works for t_MAT, t_VEC/t_COL */
 static GEN
-lll_get_im(GEN h, long k)
+vectail_inplace(GEN x, long k)
 {
-  ulong mask = h[0] & ~LGBITS;
-  long l = lg(h) - k;
-  h += k; h[0] = mask | evallg(l);
-  return h;
+  if (!k) return x;
+  x[k] = ((ulong)x[0] & ~LGBITS) | evallg(lg(x) - k);
+  return x + k;
 }
 
 /* k = dim Kernel */
@@ -62,10 +54,11 @@ static GEN
 lll_finish(GEN h, long k, long flag)
 {
   GEN g;
+  if (!(flag & (LLL_IM|LLL_KER|LLL_ALL|LLL_INPLACE))) return h;
+  if (flag & (LLL_IM|LLL_INPLACE)) return vectail_inplace(h, k);
   if (flag & LLL_KER) { setlg(h,k+1); return h; }
-  if (flag & LLL_IM) return lll_get_im(h, k);
-  g = vecslice(h,1,k);
-  return mkvec2(g, lll_get_im(h, k));
+  g = vecslice(h,1,k); /* done first: vectail_inplace kills h */
+  return mkvec2(g, vectail_inplace(h, k));
 }
 
 INLINE GEN
@@ -78,7 +71,7 @@ mulshift(GEN y, GEN z, long e)
   lz = lgefint(z);
   av = avma; (void)new_chunk(ly+lz+nbits2lg(e)); /* HACK */
   t = mulii(z, y);
-  avma = av; return shifti(t, e);
+  set_avma(av); return shifti(t, e);
 }
 
 INLINE GEN
@@ -94,7 +87,7 @@ submulshift(GEN x, GEN y, GEN z, long e)
   lz = lgefint(z);
   av = avma; (void)new_chunk(lx+ly+lz+nbits2lg(e)); /* HACK */
   t = shifti(mulii(z, y), e);
-  avma = av; return subii(x, t);
+  set_avma(av); return subii(x, t);
 }
 
 /********************************************************************/
@@ -102,6 +95,22 @@ submulshift(GEN x, GEN y, GEN z, long e)
 /**                   FPLLL (adapted from D. Stehle's code)        **/
 /**                                                                **/
 /********************************************************************/
+/* n < 10; gerepileall supporting &NULL arguments. Maybe rename and export ? */
+INLINE void
+gc_lll(pari_sp av, int n, ...)
+{
+  int i, j;
+  GEN *gptr[10];
+  va_list a; va_start(a, n);
+  for (i=j=0; i<n; i++)
+  {
+    GEN *x = va_arg(a,GEN*);
+    if (*x) { gptr[j++] = x; *x = (GEN)copy_bin(*x); }
+  }
+  set_avma(av);
+  for (--j; j>=0; j--) *gptr[j] = bin_copy((GENbin*)*gptr[j]);
+  va_end(a);
+}
 /* Babai() and fplll() are a conversion to libpari API and data types
    of the file proved.c in fplll-1.3 by Damien Stehle'.
 
@@ -118,6 +127,14 @@ submulshift(GEN x, GEN y, GEN z, long e)
   http://www.shoup.net/ntl/
 */
 
+/* x t_REAL, |x| >= 1/2. Test whether |x| <= 1.5 */
+static int
+absrsmall(GEN x)
+{
+  long e = expo(x);
+  return (e < 0) || (e == 0 && (ulong)x[2] <= (3UL << (BITS_IN_LONG-2)));
+}
+
 /***********************************************/
 /* Babai's Nearest Plane algorithm (iterative) */
 /***********************************************/
@@ -127,19 +144,20 @@ mu, r, s updated in place (affrr).
 */
 static long
 Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
-      long a, long zeros, long maxG, long n, GEN eta, GEN halfplus1, long prec)
+      long a, long zeros, long maxG, GEN eta, long prec)
 {
   pari_sp av0 = avma;
-  GEN B = *pB, G = *pG, U = *pU, tmp, rtmp, ztmp;
-  long k, aa = (a > zeros)? a : zeros+1;
+  GEN G = *pG, B = *pB, U = *pU, tmp, rtmp, ztmp;
+  long k, d, n, aa = a > zeros? a: zeros+1;
   GEN maxmu = gen_0, max2mu = gen_0;
-  /* N.B: we set d = 0 (resp. n = 0) to avoid updating U (resp. B) */
-  const long d = U ? lg(U)-1: 0;
 
+  /* N.B: we set d = 0 (resp. n = 0) to avoid updating U (resp. B) */
+  d = U? lg(U)-1: 0;
+  n = B? nbrows(B): 0;
   if (gc_needed(av,2))
   {
     if(DEBUGMEM>1) pari_warn(warnmem,"Babai[0], a=%ld", aa);
-    gerepileall(av,U?3:2,&B,&G,&U);
+    gc_lll(av,3,&G,&B,&U);
   }
   for (;;) {
     int go_on = 0;
@@ -149,7 +167,7 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
     if (gc_needed(av0,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"Babai[1], a=%ld", aa);
-      gerepileall(av,U?5:4,&B,&G,&maxmu,&max2mu,&U);
+      gc_lll(av,5,&G,&B,&U,&maxmu,&max2mu);
     }
     /* Step2: compute the GSO for stage kappa */
     max3mu = max2mu;
@@ -175,14 +193,12 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
       affrr(divrr(gmael(r,kappa,j), gmael(r,j,j)), gmael(mu,kappa,j));
       if (abscmprr(maxmu, gmael(mu,kappa,j))<0)
         maxmu = gmael(mu,kappa,j);
-      avma = btop;
+      set_avma(btop);
     }
-    maxmu = absr(maxmu);
-    if (typ(max3mu)==t_REAL && abscmprr(max3mu, shiftr(max2mu, 5))<=0)
-    {
-      *pB = B; *pG = G; *pU = U;
-      if (DEBUGLEVEL>5) err_printf("prec too low\n");
-      return kappa;
+    maxmu = absr(maxmu); /* copy needed: we 'affrr' to mu later */
+    if (typ(max3mu)==t_REAL && abscmprr(max3mu, shiftr(max2mu, 5)) <= 0)
+    { /* precision too low */
+      *pG = G; *pB = B; *pU = U; return kappa;
     }
 
     /* Step3--5: compute the X_j's  */
@@ -194,17 +210,17 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
       if (gc_needed(av0,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"Babai[2], a=%ld, j=%ld", aa,j);
-        gerepileall(av,U?5:4,&B,&G,&maxmu,&max2mu,&U);
+        gc_lll(av,5,&G,&B,&U,&maxmu,&max2mu);
       }
       go_on = 1;
       /* we consider separately the case |X| = 1 */
-      if (abscmprr(tmp, halfplus1) <= 0)
+      if (absrsmall(tmp))
       {
         if (signe(tmp) > 0) { /* in this case, X = 1 */
           pari_sp btop = avma;
           for (k=zeros+1; k<j; k++)
             affrr(subrr(gmael(mu,kappa,k), gmael(mu,j,k)), gmael(mu,kappa,k));
-          avma = btop;
+          set_avma(btop);
 
           for (i=1; i<=n; i++)
             gmael(B,kappa,i) = subii(gmael(B,kappa,i), gmael(B,j,i));
@@ -224,10 +240,10 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
           pari_sp btop = avma;
           for (k=zeros+1; k<j; k++)
             affrr(addrr(gmael(mu,kappa,k), gmael(mu,j,k)), gmael(mu,kappa,k));
-          avma = btop;
+          set_avma(btop);
 
           for (i=1; i<=n; i++)
-            gmael(B,kappa,i) = addii(gmael(B,kappa,i), gmael(B,j,i));
+            gmael(B,kappa,i) = addii(gmael(B,kappa,i),gmael(B,j,i));
           for (i=1; i<=d; i++)
             gmael(U,kappa,i) = addii(gmael(U,kappa,i),gmael(U,j,i));
           btop = avma;
@@ -256,7 +272,7 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
             rtmp = subrr(gmael(mu,kappa,k), mulur(xx, gmael(mu,j,k)));
             affrr(rtmp, gmael(mu,kappa,k));
           }
-          avma = btop;
+          set_avma(btop);
           for (i=1; i<=n; i++)
             gmael(B,kappa,i) = submuliu_inplace(gmael(B,kappa,i), gmael(B,j,i), xx);
           for (i=1; i<=d; i++)
@@ -280,7 +296,7 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
             rtmp = addrr(gmael(mu,kappa,k), mulur(xx, gmael(mu,j,k)));
             affrr(rtmp, gmael(mu,kappa,k));
           }
-          avma = btop;
+          set_avma(btop);
           for (i=1; i<=n; i++)
             gmael(B,kappa,i) = addmuliu_inplace(gmael(B,kappa,i), gmael(B,j,i), xx);
           for (i=1; i<=d; i++)
@@ -301,17 +317,17 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
       else
       {
         pari_sp btop;
-        GEN X, tmp2  = itor(ztmp,prec);
-        long e = expo(tmp2) - prec2nbits(prec);
+        long e = expi(ztmp) - prec2nbits(prec);
+        GEN X = ztmp;
 
-        X = trunc2nr(tmp2, -e); if (e < 0) { X = shifti(X,e); e = 0; }
+        if (e <= 0) e = 0; else X = shifti(X, -e);
         btop = avma;
         for (k=zeros+1; k<j; k++)
         {
           rtmp = subrr(gmael(mu,kappa,k), mulir(ztmp, gmael(mu,j,k)));
           affrr(rtmp, gmael(mu,kappa,k));
         }
-        avma = btop;
+        set_avma(btop);
         for (i=1; i<=n; i++)
           gmael(B,kappa,i) = submulshift(gmael(B,kappa,i), gmael(B,j,i), X, e);
         for (i=1; i<=d; i++)
@@ -341,60 +357,41 @@ Babai(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, GEN mu, GEN r, GEN s,
     tmp = subrr(gel(s,k), mulrr(gmael(mu,kappa,k), gmael(r,kappa,k)));
     affrr(tmp, gel(s,k+1));
   }
-  *pB = B; *pG = G; *pU = U; avma = av;
-  return 0;
+  *pG = G; *pB = B; *pU = U; return gc_long(av, 0);
 }
 
 static void
-rotate(GEN mu, long kappa2, long kappa, long d)
+rotate(GEN A, long k2, long k)
 {
-  long i, j;
-  pari_sp av = avma;
-  GEN mutmp = leafcopy(gel(mu,kappa2));
-  for (i=kappa2; i>kappa; i--)
-    for (j=1;j<=d;j++) gmael(mu,i,j) = gmael(mu,i-1,j);
-  for (j=1;j<=d;j++)   gmael(mu,kappa,j) = gel(mutmp,j);
-  avma = av;
+  long i;
+  GEN B = gel(A,k2);
+  for (i = k2; i > k; i--) gel(A,i) = gel(A,i-1);
+  gel(A,k) = B;
 }
 
 /* ****************** */
 /* The LLL Algorithm  */
 /* ****************** */
-
-/* LLL-reduces the integer matrix(ces) (G,B,U)? "in place" */
-static GEN
-fplll(GEN *ptrB, GEN *ptrU, GEN *ptrr, double DELTA, double ETA, long flag, long prec)
+/* G integral Gram matrix, LLL-reduces (G,B,U) in place [apply base change
+ * transforms to B and U]. If (keepfirst), never swap with first vector.
+ * If G = NULL, we compute the Gram matrix incrementally.
+ * Return -1 on failure, else zeros = dim Kernel (>= 0) */
+static long
+fplll(GEN *pG, GEN *pB, GEN *pU, GEN *pr, double DELTA, double ETA,
+      long keepfirst, long prec)
 {
-  const long gram = flag & LLL_GRAM; /*Gram matrix*/
-  const long keepfirst = flag & LLL_KEEP_FIRST; /*never swap with first vector*/
   pari_sp av, av2;
-  long kappa, kappa2, d, n, i, j, zeros, kappamax, maxG, bab;
-  GEN G, mu, r, s, tmp, SPtmp, alpha;
-  GEN delta = dbltor(DELTA), eta = dbltor(ETA), halfplus1 = dbltor(1.5);
-  const long triangular = 0;
-  pari_timer T;
-  GEN B = *ptrB, U;
-  long cnt = 0;
+  GEN mu, r, s, tmp, SPtmp, alpha, G = *pG, B = *pB, U = *pU;
+  GEN delta = dbltor(DELTA), eta = dbltor(ETA);
+  long d, maxG, kappa, kappa2, i, j, zeros, kappamax, incgram = !G, cnt = 0;
 
-  d = lg(B)-1;
-  if (gram)
-  {
-    G = B;
-    n = d;
-    B = cgetg(1, t_VECSMALL); /* dummy */
+  if (incgram)
+  { /* incremental Gram matrix */
+    maxG = 2; d = lg(B)-1;
+    G = zeromatcopy(d, d);
   }
   else
-  {
-    G = zeromatcopy(d,d);
-    n = nbrows(B);
-  }
-  U = *ptrU; /* NULL if inplace */
-
-  if(DEBUGLEVEL>=4)
-  {
-    timer_start(&T);
-    err_printf("Entering L^2: LLL-parameters (%P.3f,%.3Pf), working precision %d words\n",delta,eta, prec);
-  }
+    maxG = d = lg(G)-1;
 
   mu = cgetg(d+1, t_MAT);
   r  = cgetg(d+1, t_MAT);
@@ -405,47 +402,47 @@ fplll(GEN *ptrB, GEN *ptrU, GEN *ptrr, double DELTA, double ETA, long flag, long
     gel(mu,j)= M;
     gel(r,j) = R;
     gel(s,j) = cgetr(prec);
-    for (i = 1; i <= d; i++) {
+    for (i = 1; i <= d; i++)
+    {
       gel(R,i) = cgetr(prec);
       gel(M,i) = cgetr(prec);
     }
   }
-  SPtmp = zerovec(d+1);
+  SPtmp = cgetg(d+1, t_VEC);
   alpha = cgetg(d+1, t_VECSMALL);
   av = avma;
 
   /* Step2: Initializing the main loop */
   kappamax = 1;
   i = 1;
-  maxG = d; /* later updated to kappamax if (!gram) */
-
   do {
-    if (!gram) gmael(G,i,i) = ZV_dotsquare(gel(B,i));
+    if (incgram) gmael(G,i,i) = ZV_dotsquare(gel(B,i));
     affir(gmael(G,i,i), gmael(r,i,i));
-  } while (signe(gmael(G,i,i)) == 0 && (++i <=d));
-  zeros = i-1; /* all vectors B[i] with i <= zeros are zero vectors */
+  } while (!signe(gmael(G,i,i)) && ++i <= d);
+  zeros = i-1; /* all basis vectors b_i with i <= zeros are zero vectors */
   kappa = i;
   if (zeros < d) affir(gmael(G,zeros+1,zeros+1), gmael(r,zeros+1,zeros+1));
   for (i=zeros+1; i<=d; i++) alpha[i]=1;
 
   while (++kappa <= d)
   {
-    if (kappa>kappamax)
+    if (kappa > kappamax)
     {
       if (DEBUGLEVEL>=4) err_printf("K%ld ",kappa);
       kappamax = kappa;
-      if (!gram) {
+      if (incgram)
+      {
         for (i=zeros+1; i<=kappa; i++)
           gmael(G,kappa,i) = ZV_dotproduct(gel(B,kappa), gel(B,i));
         maxG = kappamax;
       }
     }
     /* Step3: Call to the Babai algorithm, mu,r,s updated in place */
-    bab = Babai(av, kappa, &G,&B,&U, mu,r,s, alpha[kappa], zeros, maxG,
-      gram? 0 : ((triangular && kappamax <= n) ? kappamax: n),
-      eta, halfplus1, prec);
-    if (bab) {*ptrB=(gram?G:B); *ptrU=U; return NULL; }
-
+    if (Babai(av, kappa, &G,&B,&U, mu,r,s, alpha[kappa], zeros, maxG, eta,prec))
+    {
+      if (incgram) G = NULL;
+      *pG = G; *pB = B; *pU = U; return -1;
+    }
     av2 = avma;
     if ((keepfirst && kappa == 2) ||
         cmprr(mulrr(gmael(r,kappa-1,kappa-1), delta), gel(s,kappa-1)) <= 0)
@@ -453,106 +450,94 @@ fplll(GEN *ptrB, GEN *ptrU, GEN *ptrr, double DELTA, double ETA, long flag, long
       alpha[kappa] = kappa;
       tmp = mulrr(gmael(mu,kappa,kappa-1), gmael(r,kappa,kappa-1));
       affrr(subrr(gel(s,kappa-1), tmp), gmael(r,kappa,kappa));
-      avma = av2;
+      set_avma(av2); continue;
     }
-    else
-    { /* Step5: Find the right insertion index kappa, kappa2 = initial kappa */
-      if (DEBUGLEVEL>=4 && kappa==kappamax && signe(gel(s,kappa-1)))
-        if (++cnt > 20) { cnt = 0; err_printf("(%ld) ", expo(gel(s,1))); }
-      kappa2 = kappa;
-      do {
-        kappa--;
-        if (kappa<zeros+2 + (keepfirst ? 1: 0)) break;
-        tmp = mulrr(gmael(r,kappa-1,kappa-1), delta);
-      } while (cmprr(gel(s,kappa-1), tmp) <=0 );
-      avma = av2;
+    /* Step5: Find the right insertion index kappa, kappa2 = initial kappa */
+    if (DEBUGLEVEL>=4 && kappa==kappamax && signe(gel(s,kappa-1)))
+      if (++cnt > 20) { cnt = 0; err_printf("(%ld) ", expo(gel(s,1))); }
+    kappa2 = kappa;
+    do {
+      kappa--;
+      if (kappa < zeros+2 + (keepfirst ? 1: 0)) break;
+      tmp = mulrr(gmael(r,kappa-1,kappa-1), delta);
+    } while (cmprr(gel(s,kappa-1), tmp) <= 0);
+    set_avma(av2);
 
-      for (i=kappa; i<kappa2; i++)
-        if (kappa <= alpha[i]) alpha[i] = kappa;
-      for (i=kappa2; i>kappa; i--) alpha[i] = alpha[i-1];
-      for (i=kappa2+1; i<=kappamax; i++)
-        if (kappa < alpha[i]) alpha[i] = kappa;
-      alpha[kappa] = kappa;
+    for (i=kappa; i<kappa2; i++)
+      if (kappa <= alpha[i]) alpha[i] = kappa;
+    for (i=kappa2; i>kappa; i--) alpha[i] = alpha[i-1];
+    for (i=kappa2+1; i<=kappamax; i++)
+      if (kappa < alpha[i]) alpha[i] = kappa;
+    alpha[kappa] = kappa;
 
-      /* Step6: Update the mu's and r's */
-      rotate(mu,kappa2,kappa,d);
-      rotate(r,kappa2,kappa,d);
-      affrr(gel(s,kappa), gmael(r,kappa,kappa));
+    /* Step6: Update the mu's and r's */
+    rotate(mu, kappa2, kappa);
+    rotate(r, kappa2, kappa);
+    affrr(gel(s,kappa), gmael(r,kappa,kappa));
 
-      /* Step7: Update B, G, U */
-      if (!gram) rotate(B,kappa2,kappa,n);
-      if (U) rotate(U,kappa2,kappa,d);
+    /* Step7: Update G, B, U */
+    if (U) rotate(U, kappa2, kappa);
+    if (B) rotate(B, kappa2, kappa);
+    for (i=1; i<=kappa2; i++) gel(SPtmp,i) = gmael(G,kappa2,i);
+    for (   ; i<=maxG;   i++) gel(SPtmp,i) = gmael(G,i,kappa2);
+    for (i=kappa2; i>kappa; i--)
+    {
+      for (j=1; j<kappa; j++) gmael(G,i,j) = gmael(G,i-1,j);
+      gmael(G,i,kappa) = gel(SPtmp,i-1);
+      for (j=kappa+1; j<=i; j++) gmael(G,i,j) = gmael(G,i-1,j-1);
+      for (j=kappa2+1; j<=maxG; j++) gmael(G,j,i) = gmael(G,j,i-1);
+    }
+    for (i=1; i<kappa; i++) gmael(G,kappa,i) = gel(SPtmp,i);
+    gmael(G,kappa,kappa) = gel(SPtmp,kappa2);
+    for (i=kappa2+1; i<=maxG; i++) gmael(G,i,kappa) = gel(SPtmp,i);
 
-      for (i=1; i<=kappa2; i++) gel(SPtmp,i) = gmael(G,kappa2,i);
-      for (i=kappa2+1; i<=maxG; i++) gel(SPtmp,i) = gmael(G,i,kappa2);
-      for (i=kappa2; i>kappa; i--)
-      {
-        for (j=1; j<kappa; j++) gmael(G,i,j) = gmael(G,i-1,j);
-        gmael(G,i,kappa) = gel(SPtmp,i-1);
-        for (j=kappa+1; j<=i; j++) gmael(G,i,j) = gmael(G,i-1,j-1);
-        for (j=kappa2+1; j<=maxG; j++) gmael(G,j,i) = gmael(G,j,i-1);
-      }
-      for (i=1; i<kappa; i++) gmael(G,kappa,i) = gel(SPtmp,i);
-      gmael(G,kappa,kappa) = gel(SPtmp,kappa2);
-      for (i=kappa2+1; i<=maxG; i++) gmael(G,i,kappa) = gel(SPtmp,i);
-
-      /* Step8: Prepare the next loop iteration */
-      if (kappa == zeros+1 && !signe(gmael(G,kappa,kappa)))
-      {
-        zeros++; kappa++;
-        affir(gmael(G,kappa,kappa), gmael(r,kappa,kappa));
-      }
+    /* Step8: Prepare the next loop iteration */
+    if (kappa == zeros+1 && !signe(gmael(G,kappa,kappa)))
+    {
+      zeros++; kappa++;
+      affir(gmael(G,kappa,kappa), gmael(r,kappa,kappa));
     }
   }
-
-  if (DEBUGLEVEL>=4) timer_printf(&T,"LLL");
-  if (ptrr) *ptrr = RgM_diagonal_shallow(r);
-  if (!U)
-  {
-    if (zeros) {
-      if (gram) {
-        G = lll_get_im(G, zeros);
-        d -= zeros;
-        for (i = 1; i <= d; i++) gel(G,i) = lll_get_im(gel(G,i), zeros);
-      }
-      else
-        B = lll_get_im(B, zeros);
-    }
-  }
-  else if (flag & (LLL_IM|LLL_KER|LLL_ALL))
-    U = lll_finish(U, zeros, flag);
-  if (gram)
-  {
-    if (U) return U;
-    for (i = 1; i <= d; i++)
-      for (j = i+1; j <= d; j++) gmael(G,i,j) = gmael(G,j,i);
-    return G;
-  }
-  return U? U: B;
+  if (pr) *pr = RgM_diagonal_shallow(r);
+  *pG = G; *pB = B; *pU = U; return zeros; /* success */
 }
 
-/* Assume x a ZM, if ptB != NULL, set it to Gram-Schmidt (squared) norms */
+/* Assume x a ZM, if pN != NULL, set it to Gram-Schmidt (squared) norms
+ * The following modes are supported:
+ * - flag & LLL_INPLACE: x a lattice basis, return x*U
+ * - flag & LLL_GRAM: x a Gram matrix / else x a lattice basis; return
+ *     LLL base change matrix U [LLL_IM]
+ *     kernel basis [LLL_KER, nonreduced]
+ *     both [LLL_ALL] */
 GEN
-ZM_lll_norms(GEN x, double DELTA, long flag, GEN *B)
+ZM_lll_norms(GEN x, double DELTA, long flag, GEN *pN)
 {
-  pari_sp ltop = avma;
-  const long compat = flag & LLL_COMPATIBLE;
+  pari_sp av = avma;
   const double ETA = 0.51;
-  long p, n = lg(x)-1;
-  GEN U;
+  long p, zeros, n = lg(x)-1;
+  GEN G, B, U;
+  pari_timer T;
+
   if (n <= 1) return lll_trivial(x, flag);
   x = RgM_shallowcopy(x);
-  U = (flag & LLL_INPLACE)? NULL: matid(n);
-  for (p = compat? DEFAULTPREC: LOWDEFAULTPREC;;)
+  if (flag & LLL_GRAM)
+  { G = x; B = NULL; U = matid(n); }
+  else
+  { G = NULL; B = x; U = (flag & LLL_INPLACE)? NULL: matid(n); }
+  for (p = DEFAULTPREC;; p += EXTRAPREC64)
   {
-    GEN m = fplll(&x, &U, B, DELTA, ETA, flag, p);
-    if (m) return m;
-    if (compat)
-      p += DEFAULTPREC-2;
-    else
-      incrprec(p);
-    gerepileall(ltop, U? 2: 1, &x, &U);
+    if (DEBUGLEVEL>=4)
+    {
+      err_printf("Entering L^2: LLL-parameters (%.3f,%.3f), prec = %d\n",
+                 DELTA,ETA, p);
+      timer_start(&T);
+    }
+    zeros = fplll(&G, &B, &U, pN, DELTA, ETA, flag & LLL_KEEP_FIRST, p);
+    if (zeros >= 0) break;
+    gc_lll(av, 3, &G, &B, &U);
   }
+  if (DEBUGLEVEL>=4) timer_printf(&T,"LLL");
+  return lll_finish(U? U: B, zeros, flag);
 }
 
 /********************************************************************/
@@ -634,7 +619,7 @@ do_SWAPgen(GEN h, GEN L, GEN B, long k, GEN fl, int *flc)
       gcoeff(L,i,k) = gcoeff(L,i,k-1);
       gcoeff(L,i,k-1) = gen_0;
     }
-    B[k] = B[k-1]; fl[k] = 1; fl[k-1] = 0;
+    gel(B,k) = gel(B,k-1); fl[k] = 1; fl[k-1] = 0;
   }
   return 1;
 }
@@ -643,12 +628,12 @@ static void
 incrementalGSgen(GEN x, GEN L, GEN B, long k, GEN fl)
 {
   GEN u = NULL; /* gcc -Wall */
-  long i, j, tu;
-  for (j=1; j<=k; j++)
+  long i, j;
+  for (j = 1; j <= k; j++)
     if (j==k || fl[j])
     {
-      u = gcoeff(x,k,j); tu = typ(u);
-      if (! is_extscalar_t(tu)) pari_err_TYPE("incrementalGSgen",u);
+      u = gcoeff(x,k,j);
+      if (!is_extscalar_t(typ(u))) pari_err_TYPE("incrementalGSgen",u);
       for (i=1; i<j; i++)
         if (fl[i])
         {
@@ -657,7 +642,7 @@ incrementalGSgen(GEN x, GEN L, GEN B, long k, GEN fl)
         }
       gcoeff(L,k,j) = u;
     }
-  if (gequal0(u)) B[k+1] = B[k];
+  if (gequal0(u)) gel(B,k+1) = gel(B,k);
   else
   {
     gel(B,k+1) = gcoeff(L,k,k); gcoeff(L,k,k) = gen_1; fl[k] = 1;
@@ -712,7 +697,7 @@ static GEN
 lllallgen(GEN x, long flag)
 {
   pari_sp av = avma;
-  if ((flag & LLL_GRAM) == 0) x = gram_matrix(x);
+  if (!(flag & LLL_GRAM)) x = gram_matrix(x);
   else if (!RgM_square(x)) pari_err_DIM("qflllgram");
   return gerepilecopy(av, lllgramallgen(x, flag));
 }
@@ -727,19 +712,19 @@ lllgramkerimgen(GEN x)  { return lllallgen(x, LLL_ALL|LLL_GRAM); }
 
 static GEN
 lllall(GEN x, long flag)
-{
-  pari_sp av = avma;
-  if ((flag & LLL_GRAM) && !RgM_square(x)) pari_err_DIM("qflllgram");
-  return gerepilecopy(av, ZM_lll(x, LLLDFT, flag));
-}
+{ pari_sp av = avma; return gerepilecopy(av, ZM_lll(x, LLLDFT, flag)); }
 GEN
 lllint(GEN x) { return lllall(x, LLL_IM); }
 GEN
 lllkerim(GEN x) { return lllall(x, LLL_ALL); }
 GEN
-lllgramint(GEN x) { return lllall(x, LLL_IM | LLL_GRAM); }
+lllgramint(GEN x)
+{ if (!RgM_square(x)) pari_err_DIM("qflllgram");
+  return lllall(x, LLL_IM | LLL_GRAM); }
 GEN
-lllgramkerim(GEN x) { return lllall(x, LLL_ALL | LLL_GRAM); }
+lllgramkerim(GEN x)
+{ if (!RgM_square(x)) pari_err_DIM("qflllgram");
+  return lllall(x, LLL_ALL | LLL_GRAM); }
 
 GEN
 lllfp(GEN x, double D, long flag)
@@ -767,6 +752,7 @@ qflll0(GEN x, long flag)
     case 0: return lll(x);
     case 1: RgM_check_ZM(x,"qflll"); return lllint(x);
     case 2: RgM_check_ZM(x,"qflll"); return lllintpartial(x);
+    case 3: RgM_check_ZM(x,"qflll"); return lllall(x, LLL_INPLACE);
     case 4: RgM_check_ZM(x,"qflll"); return lllkerim(x);
     case 5: return lllkerimgen(x);
     case 8: return lllgen(x);
