@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -58,64 +59,73 @@ typedef struct
 /*******************************************************************/
 /* NOT stack clean. a, b stay on the stack */
 static GEN
-lift_to_frac_tdenom(GEN t, GEN mod, GEN amax, GEN bmax, GEN denom, GEN tdenom)
+lift_to_frac(GEN t, GEN N, GEN amax, GEN bmax, GEN den, GEN tden)
 {
   GEN a, b;
-  if (signe(t) < 0) t = addii(t, mod); /* in case t is a centerlift */
-  if (tdenom)
+  if (signe(t) < 0) t = addii(t, N); /* in case t is a centerlift */
+  if (tden)
   {
     pari_sp av = avma;
-    a = Fp_center_i(Fp_mul(t, tdenom, mod), mod, shifti(mod,-1));
-    if (abscmpii(a, amax) < 0)
-    {
-      GEN d = gcdii(a, tdenom);
-      a = diviiexact(a, d);
-      b = diviiexact(tdenom, d);
-      if (is_pm1(b)) { return gerepileuptoint(av, a); }
-      return gerepilecopy(av, mkfrac(a, b));
-    }
-    avma = av;
+    a = Fp_center_i(Fp_mul(t, tden, N), N, shifti(N,-1));
+    if (abscmpii(a, amax) < 0) return gerepileupto(av, Qdivii(a, tden));
+    set_avma(av);
   }
-  if (!Fp_ratlift(t, mod, amax,bmax, &a,&b)
-     || (denom && !dvdii(denom,b))
+  if (!Fp_ratlift(t, N, amax,bmax, &a,&b)
+     || (den && !dvdii(den,b))
      || !is_pm1(gcdii(a,b))) return NULL;
   if (is_pm1(b)) { cgiv(b); return a; }
   return mkfrac(a, b);
 }
 
-static GEN
-lift_to_frac(GEN t, GEN mod, GEN amax, GEN bmax, GEN denom)
-{
-  return lift_to_frac_tdenom(t, mod, amax, bmax, denom, NULL);
-}
-
-/* Compute rational lifting for all the components of M modulo mod.
- * Assume all Fp_ratlift preconditions are met; we allow centerlifts but in
- * that case are no longer stack clean. If one component fails, return NULL.
- * If denom != NULL, check that the denominators divide denom.
- *
- * We suppose gcd(mod, denom) = 1, then a and b are coprime; so we can use
- * mkfrac rather than gdiv */
+/* Compute rational lifting for all the components of P modulo N. Assume
+ * Fp_ratlift preconditions are met; we allow centerlifts. If one component
+ * fails, return NULL. If den != NULL, check that the deninators divide den;
+ * assume (N, den) = 1. */
 GEN
-FpC_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
+FpC_ratlift(GEN P, GEN N, GEN amax, GEN bmax, GEN den)
 {
-  pari_sp ltop = avma;
+  pari_sp av = avma;
   long j, l;
-  GEN a, d, tdenom = NULL, Q = cgetg_copy(P, &l);
+  GEN tden = NULL, Q = cgetg_copy(P, &l);
   if (l==1) return Q;
+  if (den && cmpii(bmax, den) > 0) bmax = den;
   for (j = 1; j < l; ++j)
   {
-    a = lift_to_frac_tdenom(gel(P,j), mod, amax, bmax, denom, tdenom);
-    if (!a) { avma = ltop; return NULL; }
-    d = Q_denom(a);
-    tdenom = tdenom ? cmpii(tdenom, d)<0? d: tdenom : d;
+    GEN a = lift_to_frac(gel(P,j), N, amax, bmax, den, tden);
+    if (!a) return gc_NULL(av);
+    if (typ(a) == t_FRAC)
+    {
+      GEN d = gel(a,2);
+      tden = tden? (cmpii(tden, d) < 0? d: tden): d;
+    }
+    gel(Q,j) = a;
+  }
+  return Q;
+}
+GEN
+FpX_ratlift(GEN P, GEN N, GEN amax, GEN bmax, GEN den)
+{
+  pari_sp av = avma;
+  long j, l;
+  GEN tden = NULL, Q = cgetg_copy(P, &l);
+  Q[1] = P[1];
+  if (den && cmpii(bmax, den) > 0) bmax = den;
+  for (j = 2; j < l; ++j)
+  {
+    GEN a = lift_to_frac(gel(P,j), N, amax, bmax, den, tden);
+    if (!a) return gc_NULL(av);
+    if (typ(a) == t_FRAC)
+    {
+      GEN d = gel(a,2);
+      tden = tden? (cmpii(tden, d) < 0? d: tden): d;
+    }
     gel(Q,j) = a;
   }
   return Q;
 }
 
 GEN
-FpM_ratlift(GEN M, GEN mod, GEN amax, GEN bmax, GEN denom)
+FpM_ratlift(GEN M, GEN mod, GEN amax, GEN bmax, GEN den)
 {
   pari_sp av = avma;
   long j, l = lg(M);
@@ -123,38 +133,26 @@ FpM_ratlift(GEN M, GEN mod, GEN amax, GEN bmax, GEN denom)
   if (l == 1) return N;
   for (j = 1; j < l; ++j)
   {
-    GEN a = FpC_ratlift(gel(M, j), mod, amax, bmax, denom);
-    if (!a) { avma = av; return NULL; }
+    GEN a = FpC_ratlift(gel(M, j), mod, amax, bmax, den);
+    if (!a) return gc_NULL(av);
     gel(N,j) = a;
   }
   return N;
 }
 
-GEN
-FpX_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
-{
-  pari_sp ltop = avma;
-  long j, l;
-  GEN a, Q = cgetg_copy(P, &l);
-  Q[1] = P[1];
-  for (j = 2; j < l; ++j)
-  {
-    a = lift_to_frac(gel(P,j), mod, amax,bmax,denom);
-    if (!a) { avma = ltop; return NULL; }
-    gel(Q,j) = a;
-  }
-  return Q;
-}
-
 /*******************************************************************/
 /*              GCD in K[X], K NUMBER FIELD                        */
 /*******************************************************************/
-/* P a non-zero ZXQX */
+/* P a nonzero ZXQX */
 static GEN
 lead_simplify(GEN P)
 {
-  GEN x = gel(P, lg(P)-1); /* x a non-zero ZX or t_INT */
-  if (typ(x) == t_POL && !degpol(x)) x = gel(x,2);
+  GEN x = gel(P, lg(P)-1); /* x a nonzero ZX or t_INT */
+  if (typ(x) == t_POL)
+  {
+    if (degpol(x)) return x;
+    x = gel(x,2);
+  }
   return is_pm1(x)? NULL: x;
 }
 /* P,Q in Z[X,Y], T in Z[Y] irreducible. compute GCD in Q[Y]/(T)[X].
@@ -197,7 +195,7 @@ nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
     else
       lden = gcdii(ZX_resultant(lP, T), ZX_resultant(lQ, T));
     if (is_pm1(lden)) lden = NULL;
-    den = mul_denom(den, lden);
+    if (den && lden) den = mulii(den, lden);
   }
   init_modular_small(&S);
   btop = avma;
@@ -215,7 +213,7 @@ nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
                            ZXX_to_FlxX(Q,p,vT),
                            Tp, p)) == NULL) continue;
     dR = degpol(R);
-    if (dR == 0) { avma = ltop; if (Pnew) *Pnew = P; return pol_1(vP); }
+    if (dR == 0) { set_avma(ltop); if (Pnew) *Pnew = P; return pol_1(vP); }
     if (mod && dR > dM) continue; /* p divides Res(P/gcd, Q/gcd). Discard. */
 
     R = FlxX_to_Flm(R, dT);
@@ -249,15 +247,27 @@ nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
 }
 GEN
 nfgcd(GEN P, GEN Q, GEN T, GEN den)
-{ return nfgcd_all(P,Q,T,den,NULL); }
+{ return nfgcd_all(P, Q, T, den, NULL); }
+
+GEN
+ZXQX_gcd(GEN P, GEN Q, GEN T)
+{ return nfgcd_all(P, Q, T, NULL, NULL); }
+
+GEN
+QXQX_gcd(GEN P, GEN Q, GEN T)
+{
+  pari_sp av = avma;
+  GEN P1 = Q_remove_denom(P, NULL);
+  GEN Q1 = Q_remove_denom(Q, NULL);
+  return gerepileupto(av, ZXQX_gcd(P1, Q1, T));
+}
 
 int
 nfissquarefree(GEN nf, GEN x)
 {
   pari_sp av = avma;
   GEN g, y = RgX_deriv(x);
-  if (RgX_is_rational(x))
-    g = QX_gcd(x, y);
+  if (RgX_is_rational(x)) g = QX_gcd(x, y);
   else
   {
     GEN T = get_nfpol(nf,&nf);
@@ -265,7 +275,7 @@ nfissquarefree(GEN nf, GEN x)
     y = Q_primpart( liftpol_shallow(y) );
     g = nfgcd(x, y, T, nf? nf_get_index(nf): NULL);
   }
-  avma = av; return (degpol(g) == 0);
+  return gc_bool(av, degpol(g) == 0);
 }
 
 /*******************************************************************/
@@ -321,7 +331,7 @@ RgX_int_normalize(GEN P)
   /* cater for t_POL */
   if (typ(P0) == t_POL)
   {
-    P0 = gel(P0,2); /* non-0 constant */
+    P0 = gel(P0,2); /* nonzero constant */
     P = shallowcopy(P);
     gel(P,lg(P)-1) = P0; /* now leading term is a t_INT */
   }
@@ -331,7 +341,7 @@ RgX_int_normalize(GEN P)
 }
 
 /* discard change of variable if nf is of the form [nf,c] as return by nfinit
- * for non-monic polynomials */
+ * for nonmonic polynomials */
 static GEN
 proper_nf(GEN nf)
 { return (lg(nf) == 3)? gel(nf,1): nf; }
@@ -342,7 +352,7 @@ proper_nf(GEN nf)
 static GEN
 fix_nf(GEN *pnf, GEN *pT, GEN *pA)
 {
-  GEN nf, NF, fa, P, Q, q, D, T = *pT;
+  GEN nf, NF, P, Q, q, D, T = *pT;
   nfmaxord_t S;
   long i, l;
 
@@ -368,14 +378,12 @@ fix_nf(GEN *pnf, GEN *pT, GEN *pA)
 
   D = nf_get_disc(nf);
   if (is_pm1(D)) return gen_1;
-  fa = absZ_factor_limit(D, 0);
-  P = gel(fa,1); q = gel(P, lg(P)-1);
-  if (BPSW_psp(q)) return gen_1;
-  /* nf_get_disc(nf) may be incorrect */
-  P = nf_get_ramified_primes(nf);
-  l = lg(P);
-  Q = q; q = gen_1;
-  for (i = 1; i < l; i++)
+  (void)absZ_factor_limit_strict(D, 0, &Q);
+  if (!Q) return gen_1;
+  /* D may be incorrect */
+  Q = gel(Q,1);
+  P = nf_get_ramified_primes(nf); l = lg(P);
+  for (i = 1, q = gen_1; i < l; i++)
   {
     GEN p = gel(P,i);
     if (Z_pvalrem(Q, p, &Q) && !BPSW_psp(p)) q = mulii(q, p);
@@ -403,7 +411,7 @@ get_nfsqff_data(GEN *pnf, GEN *pT, GEN *pA, GEN *pB, GEN *ptbad)
   if (nfsqff_use_Trager(n, degpol(A)))
   {
     *pnf = T;
-    bad = den = ZX_disc(T);
+    bad = den = absi_shallow(ZX_disc(T));
     if (is_pm1(leading_coeff(T))) den = indexpartial(T, den);
   }
   else
@@ -508,7 +516,7 @@ nf_bestlift_to_pol(GEN elt, GEN bound, nflift_t *L)
   }
   else
   {
-    v = gclone(v); avma = av;
+    v = gclone(v); set_avma(av);
     u = RgV_dotproduct(L->topow, v);
     gunclone(v);
   }
@@ -569,7 +577,7 @@ fact_from_sqff(GEN rep, GEN A, GEN B, GEN y, GEN T, GEN bad)
       long j;
       forprime_t S;
       u_forprime_init(&S, degpol(T), ULONG_MAX);
-      for (; ; avma = av1)
+      for (; ; set_avma(av1))
       {
         pp = u_forprime_next(&S);
         if (! umodiu(bad,pp) || !umodiu(lb, pp)) continue;
@@ -608,22 +616,20 @@ fact_from_sqff(GEN rep, GEN A, GEN B, GEN y, GEN T, GEN bad)
   gel(rep,2) = ex;
 }
 
-/* return the factorization of x in nf */
-GEN
-nffactor(GEN nf,GEN pol)
+/* return the factorization of polynomial pol in nf */
+static GEN
+nffactor_i(GEN nf,GEN T,GEN pol)
 {
-  GEN bad, A, B, y, T, den, rep = cgetg(3, t_MAT);
+  GEN bad, A, B, y, den, rep = cgetg(3, t_MAT);
   pari_sp av = avma;
   long dA;
   pari_timer ti;
 
   if (DEBUGLEVEL>2) { timer_start(&ti); err_printf("\nEntering nffactor:\n"); }
-  T = get_nfpol(nf, &nf);
-  RgX_check_ZX(T,"nffactor");
   A = RgX_nffix("nffactor",T,pol,1);
   dA = degpol(A);
   if (dA <= 0) {
-    avma = (pari_sp)(rep + 3);
+    set_avma((pari_sp)(rep + 3));
     return (dA == 0)? trivial_fact(): zerofact(varn(pol));
   }
   if (dA == 1) {
@@ -654,7 +660,27 @@ nffactor(GEN nf,GEN pol)
   if (DEBUGLEVEL>3) err_printf("number of factor(s) found: %ld\n", lg(y)-1);
 
   fact_from_sqff(rep, A, B, y, T, bad);
-  return sort_factor_pol(rep, cmp_RgX);
+  return rep;
+}
+
+/* return the factorization of P in nf */
+GEN
+nffactor(GEN nf, GEN P)
+{
+  GEN y, T = get_nfpol(nf, &nf);
+  if (!nf) RgX_check_ZX(T,"nffactor");
+  if (typ(P) == t_RFRAC)
+  {
+    pari_sp av = avma;
+    GEN a = gel(P, 1), b = gel(P, 2);
+    y = famat_inv_shallow(nffactor_i(nf, T, b));
+    if (typ(a) == t_POL && varn(a) == varn(b))
+      y = famat_mul_shallow(nffactor_i(nf, T, a), y);
+    y = gerepilecopy(av, y);
+  }
+  else
+    y = nffactor_i(nf, T, P);
+  return sort_factor_pol(y, cmp_RgX);
 }
 
 /* assume x scalar or t_COL, G t_MAT */
@@ -759,7 +785,7 @@ PRECPB:
   }
   lt = leading_coeff(polbase);
   s = mulri(s, muliu(sqri(lt), n));
-  C = powruhalf(stor(3,DEFAULTPREC), 3 + 2*d); /* 3^{3/2 + d} */
+  C = powruhalf(utor(3,DEFAULTPREC), 3 + 2*d); /* 3^{3/2 + d} */
   return divrr(mulrr(C, s), mulur(d, mppi(DEFAULTPREC)));
 }
 
@@ -861,7 +887,7 @@ typedef struct {
   GEN S1, P1;   /* S = S0 + S1 q, idem P */
 } trace_data;
 
-/* S1 * u - P1 * round(P^-1 S u). K non-zero coords in u given by ind */
+/* S1 * u - P1 * round(P^-1 S u). K nonzero coords in u given by ind */
 static GEN
 get_trace(GEN ind, trace_data *T)
 {
@@ -922,7 +948,7 @@ init_trace(trace_data *T, GEN S, nflift_t *L, GEN q)
     pari_sp av = avma;
     T->PinvSdbl[j] = t;
     for (i=1; i < h; i++) t[i] = rtodbl(mulri(invd, gel(c,i)));
-    avma = av;
+    set_avma(av);
   }
 
   T->d  = L->pk;
@@ -1051,7 +1077,7 @@ nfcmbf(nfcmbf_t *T, long klim, long *pmaxK, int *done)
         t2 = Fq_Fp_mul(t2, lt2dn, Tpk, pk);
       }
       gel(trace1,i) = gclone( nf_bestlift(t1, NULL, T->L) );
-      gel(trace2,i) = gclone( nf_bestlift(t2, NULL, T->L) ); avma = av;
+      gel(trace2,i) = gclone( nf_bestlift(t2, NULL, T->L) ); set_avma(av);
     }
     T1 = init_trace(&_T1, trace1, T->L, q);
     T2 = init_trace(&_T2, trace2, T->L, q);
@@ -1089,7 +1115,7 @@ nextK:
         if (rtodbl(_norml2(t)) > Bhigh)
         {
           if (DEBUGLEVEL>6) err_printf(".");
-          avma = av; goto NEXT;
+          set_avma(av); goto NEXT;
         }
       }
       if (T2)
@@ -1098,10 +1124,10 @@ nextK:
         if (rtodbl(_norml2(t)) > Bhigh)
         {
           if (DEBUGLEVEL>3) err_printf("|");
-          avma = av; goto NEXT;
+          set_avma(av); goto NEXT;
         }
       }
-      avma = av;
+      set_avma(av);
       y = ltdn; /* full computation */
       for (i=1; i<=K; i++)
       {
@@ -1113,7 +1139,7 @@ nextK:
       if (!y)
       {
         if (DEBUGLEVEL>3) err_printf("@");
-        avma = av; goto NEXT;
+        set_avma(av); goto NEXT;
       }
       /* y = topowden*dn*lt*\prod_{i in ind} famod[i] is apparently in O_K[X],
        * in fact in (Z[Y]/nf.pol)[X] due to multiplication by C = topowden*dn.
@@ -1122,7 +1148,7 @@ nextK:
       if (!q)
       {
         if (DEBUGLEVEL>3) err_printf("*");
-        avma = av; goto NEXT;
+        set_avma(av); goto NEXT;
       }
       /* Original T->pol in O_K[X] with leading coeff lt in Z,
        * y = C*lt \prod famod[i] is in O_K[X] with leading coeff in Z
@@ -1358,7 +1384,7 @@ bestlift_init(long a, GEN nf, GEN C, nflift_t *L)
   }
   timer_start(&ti);
   if (!a) a = (long)bestlift_bound(C, d, alpha, p, f);
-  for (;; avma = av, a += (a==1)? 1: (a>>1)) /* roughly a *= 1.5 */
+  for (;; set_avma(av), a += (a==1)? 1: (a>>1)) /* roughly a *= 1.5 */
   {
     GEN B, q = powiu(p,a), Tq = FpXQ_powu(T, a, FpX_red(nf_get_pol(nf), q), q);
     if (DEBUGLEVEL>2) err_printf("exponent %ld\n",a);
@@ -1399,7 +1425,7 @@ get_V(GEN Tra, GEN M_L, GEN PRK, GEN PRKinv, GEN pk, long *eT2)
     av2 = avma;
     T2 = ZC_sub(T2, ZM_ZC_mul(PRK, v));
     e = gexpo(T2); if (e > *eT2) *eT2 = e;
-    avma = av2;
+    set_avma(av2);
     gel(V,i) = gerepileupto(av, v); /* small */
   }
   return V;
@@ -1525,7 +1551,7 @@ AGAIN:
     if (i == r && ZM_equal(CM_L, oldCM_L))
     {
       CM_L = oldCM_L;
-      avma = av2; continue;
+      set_avma(av2); continue;
     }
 
     CM_Lp = FpM_image(CM_L, utoipos(27449)); /* inexpensive test */
@@ -1613,7 +1639,7 @@ nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, long fl, nflift_t *L)
     av = avma;
     gel(Cltx_r,2) = gneg(r); /* check P(r) == 0 */
     dvd = ZXQX_dvd(D.C2ltpol, Cltx_r, nfpol); /* integral */
-    avma = av;
+    set_avma(av);
     /* don't go on with q, usually much larger that C2ltpol */
     if (dvd) {
       if (D.Clt) r = gdiv(r, D.Clt);
@@ -1651,7 +1677,7 @@ get_good_factor(GEN T, ulong p, long maxf)
     }
     setlg(v,j); if (j > 1) return v;
   }
-  avma = av; return NULL; /* failure */
+  return gc_NULL(av); /* failure */
 }
 
 /* n = number of modular factors, f = residue degree; nold/fold current best
@@ -1706,6 +1732,8 @@ nf_pick_prime(GEN nf, GEN pol, long fl, GEN *lt, GEN *Tp, ulong *pp)
   pari_timer ti_pr;
 
   if (DEBUGLEVEL>3) timer_start(&ti_pr);
+  /* if field degree is large, try hard to find an inert prime */
+  if (nfdeg > 30) ct = maxss(ct, 2*nfdeg);
   *lt  = leading_coeff(pol); /* t_INT */
   if (gequal1(*lt)) *lt = NULL;
   *pp = 0;
@@ -1732,14 +1760,14 @@ nf_pick_prime(GEN nf, GEN pol, long fl, GEN *lt, GEN *Tp, ulong *pp)
       { /* degree 1 */
         red = FlxX_to_Flx(red);
         if (ltp) red = Flx_normalize(red, p);
-        if (!Flx_is_squarefree(red, p)) { avma = av2; continue; }
+        if (!Flx_is_squarefree(red, p)) { set_avma(av2); continue; }
         ok = 1;
         n = (fl == FACTORS)? Flx_nbfact(red,p): Flx_nbroots(red,p);
       }
       else
       {
         if (ltp) red = FlxqX_normalize(red, T, p);
-        if (!FlxqX_is_squarefree(red, T, p)) { avma = av2; continue; }
+        if (!FlxqX_is_squarefree(red, T, p)) { set_avma(av2); continue; }
         ok = 1;
         n = (fl == FACTORS)? FlxqX_nbfact(red,T,p): FlxqX_nbroots(red,T,p);
       }
@@ -1755,7 +1783,7 @@ nf_pick_prime(GEN nf, GEN pol, long fl, GEN *lt, GEN *Tp, ulong *pp)
 
       if (fl == ROOTS && f==nfdeg) { *Tp = T; *pp = p; return n; }
       if (record(nold, n, fold, f)) { nold = n; fold = f; *Tp = T; *pp = p; }
-      else avma = av2;
+      else set_avma(av2);
     }
     if (ok && --ct <= 0) break;
   }
@@ -1812,10 +1840,10 @@ polfnf(GEN a, GEN T)
   dA = degpol(A);
   if (dA <= 0)
   {
-    avma = (pari_sp)(rep + 3);
+    set_avma((pari_sp)(rep + 3));
     return (dA == 0)? trivial_fact(): zerofact(varn(A));
   }
-  bad = dent = ZX_disc(T);
+  bad = dent = absi_shallow(ZX_disc(T));
   if (tmonic) dent = indexpartial(T, dent);
   (void)nfgcd_all(A,RgX_deriv(A), T, dent, &B);
   if (degpol(B) != dA) B = Q_primpart( QXQX_normalize(B, T) );
@@ -1951,7 +1979,7 @@ nfroots_if_split(GEN *pnf, GEN pol)
   GEN T = get_nfpol(*pnf,pnf), den = fix_nf(pnf, &T, &pol);
   pari_sp av = avma;
   GEN z = nfsqff(*pnf, pol, ROOTS_SPLIT, den);
-  if (lg(z) == 1) { avma = av; return NULL; }
+  if (lg(z) == 1) return gc_NULL(av);
   return gerepilecopy(av, z);
 }
 
@@ -1966,84 +1994,17 @@ nfroots_if_split(GEN *pnf, GEN pol)
  *
  * [Step 1]: guess roots via ramification. If trivial output this.
  * [Step 2]: select prime [p] unramified and ideal [pr] above
- * [Step 3]: evaluate the maximal exponent [k] such that the fondamental domain
- *           of a LLL-reduction of [prk] = pr^k contains a ball of radius larger
- *           than the norm of any root of unity.
- * [Step 3]: select a heuristic exponent,
- *           LLL reduce prk=pr^k and verify the exponent is sufficient,
- *           otherwise try a larger one.
  * [Step 4]: factor the cyclotomic polynomial mod [pr],
- *           Hensel lift to pr^k and find the representative in the ball
- *           If there is it is a primitive root */
-
-/* Choose prime ideal unramified with "large" inertia degree */
-static void
-nf_pick_prime_for_units(GEN nf, nflift_t *L)
-{
-  GEN nfpol = nf_get_pol(nf), bad = mulii(nf_get_disc(nf), nf_get_index(nf));
-  GEN ap = NULL, r = NULL;
-  long nfdeg = degpol(nfpol), maxf = get_maxf(nfdeg);
-  ulong pp;
-  forprime_t S;
-
-  (void)u_forprime_init(&S, 2, ULONG_MAX);
-  while ( (pp = u_forprime_next(&S)) )
-  {
-    if (! umodiu(bad,pp)) continue;
-    r = get_good_factor(nfpol, pp, maxf);
-    if (r) break;
-  }
-  if (!r) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
-  r = gel(r,lg(r)-1); /* largest inertia degree */
-  ap = utoipos(pp);
-  L->p = ap;
-  L->Tp = Flx_to_ZX(r);
-  L->tozk = nf_get_invzk(nf);
-  L->topow = nf_get_zkprimpart(nf);
-  L->topowden = nf_get_zkden(nf);
-}
-
-/* *Heuristic* exponent k such that the fundamental domain of pr^k
- * should contain the ball of radius C */
-static double
-mybestlift_bound(GEN C)
-{
-  C = gtofp(C,DEFAULTPREC);
-  return ceil(log(gtodouble(C)) / 0.2) + 3;
-}
-
-/* simplified nf_DDF_roots: polcyclo(n) monic in ZX either splits or has no
- * root in nf.
- * Return a root or NULL (no root) */
-static GEN
-nfcyclo_root(long n, GEN nfpol, nflift_t *L)
-{
-  GEN q, r, Cltx_r, pol = polcyclo(n,0), gn = utoipos(n);
-  div_data D;
-
-  init_div_data(&D, pol, L);
-  (void)Fq_sqrtn(gen_1, gn, L->Tp, L->p, &r);
-  /* r primitive n-th root of 1 in Fq */
-  r = Zq_sqrtnlift(gen_1, gn, r, L->Tpk, L->p, L->k);
-  /* lt*dn*topowden * r = Clt * r */
-  r = nf_bestlift_to_pol(r, NULL, L);
-  Cltx_r = deg1pol_shallow(D.Clt? D.Clt: gen_1, gneg(r), varn(pol));
-  /* check P(r) == 0 */
-  q = RgXQX_divrem(D.C2ltpol, Cltx_r, nfpol, ONLY_DIVIDES); /* integral */
-  if (!q) return NULL;
-  if (D.Clt) r = gdiv(r, D.Clt);
-  return r;
-}
+*/
 
 /* Guesses the number of roots of unity in number field [nf].
  * Computes gcd of N(P)-1 for some primes. The value returned is a proven
  * multiple of the correct value. */
 static long
-guess_roots(GEN nf)
+guess_roots(GEN T, GEN D, GEN index)
 {
-  long c = 0, nfdegree = nf_get_degree(nf), B = nfdegree + 20, l;
+  long c = 0, nfdegree = degpol(T), B = nfdegree + 20, l;
   ulong p = 2;
-  GEN T = nf_get_pol(nf), D = nf_get_disc(nf), index = nf_get_index(nf);
   GEN nbroots = NULL;
   forprime_t S;
   pari_sp av;
@@ -2054,13 +2015,14 @@ guess_roots(GEN nf)
   for (l=1; (p = u_forprime_next(&S)); l++)
   {
     GEN old, F, pf_1, Tp;
-    long i, nb, gcdf = 0;
+    ulong i, gcdf = 0;
+    long nb;
 
     if (!umodiu(D,p) || !umodiu(index,p)) continue;
     Tp = ZX_to_Flx(T,p); /* squarefree */
     F = Flx_nbfact_by_degree(Tp, &nb, p);
     /* the gcd of the p^f - 1 is p^(gcd of the f's) - 1 */
-    for (i = 1; i <= nfdegree; i++)
+    for (i = 1; i <= (ulong) nfdegree; i++)
       if (F[i]) {
         gcdf = gcdf? ugcd(gcdf, i): i;
         if (gcdf == 1) break;
@@ -2078,7 +2040,7 @@ guess_roots(GEN nf)
   }
   if (!nbroots) pari_err_OVERFLOW("guess_roots [ran out of primes]");
   if (DEBUGLEVEL>5) err_printf("%ld loops\n",l);
-  avma = av; return itos(nbroots);
+  return gc_long(av, itos(nbroots));
 }
 
 /* T(x) an irreducible ZX. Is it of the form Phi_n(c \pm x) ?
@@ -2126,29 +2088,29 @@ static GEN
 trivroots(void) { return mkvec2(gen_2, gen_m1); }
 /* Number of roots of unity in number field [nf]. */
 GEN
-rootsof1(GEN nf)
+nfrootsof1(GEN nf)
 {
-  nflift_t L;
-  GEN T, q, fa, LP, LE, C0, z, disc;
+  GEN T, fa, LP, LE, z, disc, index;
   pari_timer ti;
   long i, l, nbguessed, nbroots, nfdegree;
   pari_sp av;
 
-  nf = checknf(nf);
-  if (nf_get_r1(nf)) return trivroots();
-
+  T = get_nfpol(nf, &nf);
+  RgX_check_ZX(T, "nfrootsof1");
+  if (nf && nf_get_r1(nf)) return trivroots();
+  disc = nf ? nf_get_disc(nf): ZX_disc(T);
+  index = nf ? nf_get_index(nf): gen_1;
   /* Step 1 : guess number of roots and discard trivial case 2 */
   if (DEBUGLEVEL>2) timer_start(&ti);
-  nbguessed = guess_roots(nf);
+  nbguessed = guess_roots(T, disc, index);
   if (DEBUGLEVEL>2)
     timer_printf(&ti, "guessing roots of 1 [guess = %ld]", nbguessed);
   if (nbguessed == 2) return trivroots();
 
-  nfdegree = nf_get_degree(nf);
+  nfdegree = degpol(T);
   fa = factoru(nbguessed);
   LP = gel(fa,1); l = lg(LP);
   LE = gel(fa,2);
-  disc = nf_get_disc(nf);
   for (i = 1; i < l; i++)
   {
     long p = LP[i];
@@ -2186,119 +2148,45 @@ rootsof1(GEN nf)
   av = avma;
 
   /* Step 1.5 : test if nf.pol == subst(polcyclo(nbguessed), x, \pm x+c) */
-  T = nf_get_pol(nf);
   if (eulerphiu_fact(fa) == (ulong)nfdegree)
   {
     z = ZXirred_is_cyclo_translate(T, uissquarefree_fact(fa));
     if (DEBUGLEVEL>2) timer_printf(&ti, "checking for cyclotomic polynomial");
     if (z)
     {
-      z = nf_to_scalar_or_basis(nf,z);
+      if (nf) z = nf_to_scalar_or_basis(nf,z);
       return gerepilecopy(av, mkvec2(utoipos(nbguessed), z));
     }
-    avma = av;
+    set_avma(av);
   }
 
-  /* Step 2 : choose a prime ideal for local lifting */
-  nf_pick_prime_for_units(nf, &L);
-  if (DEBUGLEVEL>2)
-    timer_printf(&ti, "choosing prime %Ps, degree %ld",
-             L.p, L.Tp? degpol(L.Tp): 1);
-
-  /* Step 3 : compute a reduced pr^k allowing lifting of local solutions */
-  /* evaluate maximum L2 norm of a root of unity in nf */
-  C0 = gmulsg(nfdegree, L2_bound(nf, gen_1));
-  /* lift and reduce pr^k */
-  if (DEBUGLEVEL>2) err_printf("Lift pr^k; GSmin wanted: %Ps\n",C0);
-  bestlift_init((long)mybestlift_bound(C0), nf, C0, &L);
-  L.dn = NULL;
-  if (DEBUGLEVEL>2) timer_start(&ti);
-
-  /* Step 4 : actual computation of roots */
-  nbroots = 2; z = gen_m1;
-  q = powiu(L.p,degpol(L.Tp));
+  /* Step 2 : actual computation of roots */
+  nbroots = 2; z = scalarpol(gen_m1, varn(T));
   for (i = 1; i < l; i++)
   { /* for all prime power factors of nbguessed, find a p^k-th root of unity */
     long k, p = LP[i];
-    for (k = minss(LE[i], Z_lval(subiu(q,1UL),p)); k > 0; k--)
+    for (k = LE[i]; k > 0; k--)
     { /* find p^k-th roots */
       pari_sp av = avma;
       long pk = upowuu(p,k);
       GEN r;
       if (pk==2) continue; /* no need to test second roots ! */
-      r = nfcyclo_root(pk, T, &L);
-      if (DEBUGLEVEL>2) timer_printf(&ti, "for factoring Phi_%ld^%ld", p,k);
-      if (r) {
+      r = nfisincl0(polcyclo(pk, 0), T, 1);
+      if (!isintzero(r))
+      {
         if (DEBUGLEVEL>2) err_printf("  %s root of unity found\n",uordinal(pk));
         if (p==2) { nbroots = pk; z = r; }
-        else     { nbroots *= pk; z = nfmul(nf, z,r); }
+        else
+        {
+          nbroots *= pk;
+          z = nf ? nfmul(nf, z, r): QXQ_mul(z, r, T);
+        }
         break;
       }
-      avma = av;
-      if (DEBUGLEVEL) pari_warn(warner,"rootsof1: wrong guess");
+      set_avma(av);
+      if (DEBUGLEVEL) pari_warn(warner,"nfrootsof1: wrong guess");
     }
   }
+  if (nf) z = nf_to_scalar_or_basis(nf,z);
   return gerepilecopy(av, mkvec2(utoi(nbroots), z));
 }
-
-static long
-zk_equal1(GEN y)
-{
-  if (typ(y) == t_INT) return equali1(y);
-  return equali1(gel(y,1)) && ZV_isscalar(y);
-}
-/* x^w = 1 */
-static GEN
-is_primitive_root(GEN nf, GEN fa, GEN x, long w)
-{
-  GEN P = gel(fa,1);
-  long i, l = lg(P);
-
-  for (i = 1; i < l; i++)
-  {
-    long p = itos(gel(P,i));
-    GEN y = nfpow_u(nf,x, w/p);
-    if (zk_equal1(y) > 0) /* y = 1 */
-    {
-      if (p != 2 || !equali1(gcoeff(fa,i,2))) return NULL;
-      x = gneg_i(x);
-    }
-  }
-  return x;
-}
-GEN
-rootsof1_kannan(GEN nf)
-{
-  pari_sp av = avma;
-  long N, k, i, ws, prec;
-  GEN z, y, d, list, w;
-
-  nf = checknf(nf);
-  if ( nf_get_r1(nf) ) return trivroots();
-
-  N = nf_get_degree(nf); prec = nf_get_prec(nf);
-  for (;;)
-  {
-    GEN R = R_from_QR(nf_get_G(nf), prec);
-    if (R)
-    {
-      y = fincke_pohst(mkvec(R), utoipos(N), N * N, 0, NULL);
-      if (y) break;
-    }
-    prec = precdbl(prec);
-    if (DEBUGLEVEL) pari_warn(warnprec,"rootsof1",prec);
-    nf = nfnewprec_shallow(nf,prec);
-  }
-  if (itos(ground(gel(y,2))) != N) pari_err_BUG("rootsof1 (bug1)");
-  w = gel(y,1); ws = itos(w);
-  if (ws == 2) { avma = av; return trivroots(); }
-
-  d = Z_factor(w); list = gel(y,3); k = lg(list);
-  for (i=1; i<k; i++)
-  {
-    z = is_primitive_root(nf, d, gel(list,i), ws);
-    if (z) return gerepilecopy(av, mkvec2(w, z));
-  }
-  pari_err_BUG("rootsof1");
-  return NULL; /* LCOV_EXCL_LINE */
-}
