@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -26,11 +27,23 @@ vec_isconst(GEN v)
   GEN w;
   if (l==1) return 1;
   w = gel(v,1);
-  for(i=2;i<l;i++)
+  for(i=2; i<l; i++)
     if (!gequal(gel(v,i), w)) return 0;
   return 1;
 }
 
+int
+vecsmall_isconst(GEN v)
+{
+  long i, l = lg(v);
+  ulong w;
+  if (l==1) return 1;
+  w = uel(v,1);
+  for(i=2; i<l; i++)
+    if (uel(v,i) != w) return 0;
+  return 1;
+}
+
 /* Check if all the elements of v are different.
  * Use a quadratic algorithm. Could be done in n*log(n) by sorting. */
 int
@@ -92,7 +105,7 @@ vecsmall_sortspec(GEN v, long n, GEN w)
       w[m++]=y[iy++];
   for(;ix<nx;) w[m++]=x[ix++];
   for(;iy<ny;) w[m++]=y[iy++];
-  avma=ltop;
+  set_avma(ltop);
 }
 
 /*in place sort.*/
@@ -136,7 +149,7 @@ vecsmall_indexsortspec(GEN v, long n)
       w[m++] = y[iy++]+nx;
   for(;ix<=nx;) w[m++] = x[ix++];
   for(;iy<=ny;) w[m++] = y[iy++]+nx;
-  avma = (pari_sp)w; return w;
+  set_avma((pari_sp)w); return w;
 }
 
 /*indirect sort.*/
@@ -191,10 +204,56 @@ vecsmall_duplicate(GEN x)
   if (l==1) return 0;
   for (k=x[p[1]],i=2; i<l; k=x[p[i++]])
     if (x[p[i]] == k) { r=p[i]; break; }
-  avma=av;
+  set_avma(av);
   return r;
 }
 
+static int
+vecsmall_is1to1spec(GEN v, long n, GEN w)
+{
+  pari_sp ltop=avma;
+  long nx=n>>1, ny=n-nx;
+  long m, ix, iy;
+  GEN x, y;
+  if (n<=2)
+  {
+    if (n==1)
+      w[0]=v[0];
+    else if (n==2)
+    {
+      long v0=v[0], v1=v[1];
+      if (v0==v1) return 0;
+      else if (v0<v1) { w[0]=v0; w[1]=v1; }
+      else            { w[0]=v1; w[1]=v0; }
+    }
+    return 1;
+  }
+  x = new_chunk(nx);
+  if (!vecsmall_is1to1spec(v,nx,x))    return 0;
+  y = new_chunk(ny);
+  if (!vecsmall_is1to1spec(v+nx,ny,y)) return 0;
+  for (m=0, ix=0, iy=0; ix<nx && iy<ny; )
+    if (x[ix]==y[iy]) return 0;
+    else if (x[ix]<y[iy])
+      w[m++]=x[ix++];
+    else
+      w[m++]=y[iy++];
+  for(;ix<nx;) w[m++]=x[ix++];
+  for(;iy<ny;) w[m++]=y[iy++];
+  set_avma(ltop);
+  return 1;
+}
+
+int
+vecsmall_is1to1(GEN V)
+{
+  pari_sp av = avma;
+  long l;
+  GEN W = cgetg_copy(V, &l);
+  if (l <= 2) return 1;
+  return gc_bool(av, vecsmall_is1to1spec(V+1,l,W+1));
+}
+
 /*************************************************************************/
 /**                                                                     **/
 /**             Routines for handling vectors of VECSMALL               **/
@@ -203,27 +262,26 @@ vecsmall_duplicate(GEN x)
 
 GEN
 vecvecsmall_sort(GEN x)
-{
-  return gen_sort(x, (void*)&vecsmall_lexcmp, cmp_nodata);
-}
+{ return gen_sort(x, (void*)&vecsmall_lexcmp, cmp_nodata); }
+GEN
+vecvecsmall_sort_shallow(GEN x)
+{ return gen_sort_shallow(x, (void*)&vecsmall_lexcmp, cmp_nodata); }
+
+void
+vecvecsmall_sort_inplace(GEN x, GEN *perm)
+{ gen_sort_inplace(x, (void*)&vecsmall_lexcmp, cmp_nodata, perm); }
 
 GEN
 vecvecsmall_sort_uniq(GEN x)
-{
-  return gen_sort_uniq(x, (void*)&vecsmall_lexcmp, cmp_nodata);
-}
+{ return gen_sort_uniq(x, (void*)&vecsmall_lexcmp, cmp_nodata); }
 
 GEN
 vecvecsmall_indexsort(GEN x)
-{
-  return gen_indexsort(x, (void*)&vecsmall_lexcmp, cmp_nodata);
-}
+{ return gen_indexsort(x, (void*)&vecsmall_lexcmp, cmp_nodata); }
 
 long
 vecvecsmall_search(GEN x, GEN y, long flag)
-{
-  return gen_search(x,y,flag,(void*)vecsmall_prefixcmp, cmp_nodata);
-}
+{ return gen_search(x,y,flag,(void*)vecsmall_prefixcmp, cmp_nodata); }
 
 /* assume x non empty */
 long
@@ -256,6 +314,46 @@ perm_mul_inplace2(GEN s, GEN t)
   for (i = 1; i < l; i++) t[i] = s[t[i]];
 }
 
+GEN
+vecperm_extendschreier(GEN C, GEN v, long n)
+{
+  pari_sp av = avma;
+  long mj, lv = lg(v), m = 1, mtested = 1;
+  GEN bit = const_vecsmall(n, 0);
+  GEN cy = cgetg(n+1, t_VECSMALL);
+  GEN sh = const_vec(n, gen_0);
+  for(mj=1; mj<=n; mj++)
+  {
+    if (isintzero(gel(C,mj))) continue;
+    gel(sh,mj) = gcopy(gel(C,mj));
+    if (bit[mj]) continue;
+    cy[m++] = mj;
+    bit[mj] = 1;
+    for(;;)
+    {
+      long o, mold = m;
+      for (o = 1; o < lv; o++)
+      {
+        GEN vo = gel(v,o);
+        long p;
+        for (p = mtested; p < mold; p++) /* m increases! */
+        {
+          long j = vo[ cy[p] ];
+          if (!bit[j])
+          {
+            gel(sh,j) = perm_mul(vo, gel(sh, cy[p]));
+            cy[m++] = j;
+          }
+          bit[j] = 1;
+        }
+      }
+      mtested = mold;
+      if (m == mold) break;
+    }
+  }
+  return gerepileupto(av, sh);
+}
+
 /* Orbits of the subgroup generated by v on {1,..,n} */
 static GEN
 vecperm_orbits_i(GEN v, long n)
@@ -300,15 +398,7 @@ vecperm_orbits(GEN v, long n)
   return gerepilecopy(av, vecperm_orbits_i(v, n));
 }
 
-/* Compute the cyclic decomposition of a permutation */
-GEN
-perm_cycles(GEN v)
-{
-  pari_sp av = avma;
-  return gerepilecopy(av, vecperm_orbits_i(mkvec(v), lg(v)-1));
-}
-
-static long
+static int
 isperm(GEN v)
 {
   pari_sp av = avma;
@@ -319,24 +409,59 @@ isperm(GEN v)
   for (i=1; i<=n; i++)
   {
     long d = v[i];
-    if (d < 1 || d > n || w[d]) { avma = av; return 0; }
+    if (d < 1 || d > n || w[d]) return gc_bool(av,0);
     w[d] = 1;
   }
-  avma = av; return 1;
+  return gc_bool(av,1);
+}
+
+/* Compute the cyclic decomposition of a permutation */
+GEN
+perm_cycles(GEN v)
+{
+  pari_sp av = avma;
+  return gerepilecopy(av, vecperm_orbits_i(mkvec(v), lg(v)-1));
+}
+
+GEN
+permcycles(GEN v)
+{
+  if (!isperm(v)) pari_err_TYPE("permcycles",v);
+  return perm_cycles(v);
 }
 
 /* Output the order of p */
-long
-perm_order(GEN v)
+ulong
+perm_orderu(GEN v)
 {
-  pari_sp ltop = avma;
+  pari_sp av = avma;
   GEN c = vecperm_orbits_i(mkvec(v), lg(v)-1);
   long i, d;
   for(i=1, d=1; i<lg(c); i++) d = ulcm(d, lg(gel(c,i))-1);
-  avma = ltop; return d;
+  return gc_ulong(av,d);
 }
 
-long
+static GEN
+_domul(void *data, GEN x, GEN y)
+{
+  GEN (*mul)(GEN,GEN)=(GEN (*)(GEN,GEN)) data;
+  return mul(x,y);
+}
+
+/* Output the order of p */
+GEN
+perm_order(GEN v)
+{
+  pari_sp av = avma;
+  GEN c = vecperm_orbits_i(mkvec(v), lg(v)-1);
+  long i, l = lg(c);
+  GEN V = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+    gel(V,i) = utoi(lg(gel(c,i))-1);
+  return gerepileuptoint(av, gen_product(V, (void *)lcmii, _domul));
+}
+
+GEN
 permorder(GEN v)
 {
   if (!isperm(v)) pari_err_TYPE("permorder",v);
@@ -352,7 +477,7 @@ perm_sign(GEN v)
   long i, l = lg(c), s = 1;
   for (i = 1; i < l; i++)
     if (odd(lg(gel(c, i)))) s = -s;
-  avma = av; return s;
+  return gc_long(av,s);
 }
 
 long
@@ -379,7 +504,7 @@ Z_to_perm(long n, GEN x)
       if (uel(v,i) > a) uel(v,i)++;
     uel(v,r) = a+1;
   }
-  avma = av; return v;
+  set_avma(av); return v;
 }
 GEN
 numtoperm(long n, GEN x)
@@ -395,7 +520,7 @@ perm_to_Z_inplace(GEN v)
 {
   long l = lg(v), i, r;
   GEN x = gen_0;
-  if (!isperm(v)) pari_err_TYPE("permsign",v);
+  if (!isperm(v)) return NULL;
   for (i = 1; i < l; i++)
   {
     long vi = v[i];
@@ -424,14 +549,14 @@ permtonum(GEN p)
     case t_VECSMALL: return perm_to_Z(p);
     case t_VEC: case t_COL:
       if (RgV_is_ZV(p)) { v = ZV_to_zv(p); break; }
-    default: pari_err_TYPE("permtonum",p); return NULL;
+    default: pari_err_TYPE("permtonum",p);
+      return NULL;/*LCOV_EXCL_LINE*/
   }
   x = perm_to_Z_inplace(v);
   if (!x) pari_err_TYPE("permtonum",p);
   return gerepileuptoint(av, x);
 }
 
-
 GEN
 cyc_pow(GEN cyc, long exp)
 {
@@ -446,7 +571,7 @@ cyc_pow(GEN cyc, long exp)
   for (r = j = 1; j < lg(cyc); j++)
   {
     GEN v = gel(cyc,j);
-    long n = lg(v) - 1, e = smodss(exp,n), g = (long)ugcd(n, e), m = n / g;
+    long n = lg(v) - 1, e = umodsu(exp,n), g = (long)ugcd(n, e), m = n / g;
     for (i = 0; i < g; i++)
     {
       GEN p = cgetg(m+1, t_VECSMALL);
@@ -473,7 +598,7 @@ cyc_pow_perm(GEN cyc, long exp)
   for (j = 1; j < lg(cyc); j++)
   {
     GEN v = gel(cyc,j);
-    n = lg(v) - 1; e = smodss(exp, n);
+    n = lg(v) - 1; e = umodsu(exp, n);
     for (k = 1, l = e; k <= n; k++)
     {
       p[v[k]] = v[l+1];
@@ -484,7 +609,7 @@ cyc_pow_perm(GEN cyc, long exp)
 }
 
 GEN
-perm_pow(GEN perm, long exp)
+perm_pow(GEN perm, GEN exp)
 {
   long i, r = lg(perm)-1;
   GEN p = zero_zv(r);
@@ -495,19 +620,41 @@ perm_pow(GEN perm, long exp)
     long e, n, k, l;
     if (p[i]) continue;
     v[1] = i;
-    for (n=1, k=perm[i]; k!=i; k=perm[k], n++)
-      v[n+1] = k;
-    e = smodss(exp, n);
+    for (n=1, k=perm[i]; k!=i; k=perm[k], n++) v[n+1] = k;
+    e = umodiu(exp, n);
     for (k = 1, l = e; k <= n; k++)
     {
       p[v[k]] = v[l+1];
       if (++l == n) l = 0;
     }
   }
-  avma = av; return p;
+  set_avma(av); return p;
 }
 
-static GEN
+GEN
+perm_powu(GEN perm, ulong exp)
+{
+  ulong i, r = lg(perm)-1;
+  GEN p = zero_zv(r);
+  pari_sp av = avma;
+  GEN v = cgetg(r+1, t_VECSMALL);
+  for (i=1; i<=r; i++)
+  {
+    ulong e, n, k, l;
+    if (p[i]) continue;
+    v[1] = i;
+    for (n=1, k=perm[i]; k!=i; k=perm[k], n++) v[n+1] = k;
+    e = exp % n;
+    for (k = 1, l = e; k <= n; k++)
+    {
+      p[v[k]] = v[l+1];
+      if (++l == n) l = 0;
+    }
+  }
+  set_avma(av); return p;
+}
+
+GEN
 perm_to_GAP(GEN p)
 {
   pari_sp ltop=avma;
@@ -593,10 +740,9 @@ long
 perm_relorder(GEN p, GEN set)
 {
   pari_sp ltop = avma;
-  long n = 1;
-  long q = p[1];
+  long n = 1, q = p[1];
   while (!F2v_coeff(set,q)) { q = p[q]; n++; }
-  avma = ltop; return n;
+  return gc_long(ltop,n);
 }
 
 GEN
@@ -676,6 +822,18 @@ group_elts(GEN G, long n)
   return res;
 }
 
+GEN
+groupelts_conj_set(GEN elts, GEN p)
+{
+  long i, j, l = lg(elts), n = lg(p)-1;
+  GEN res = zero_F2v(n);
+  for(j = 1; j < n; j++)
+    if (p[j]==1) break;
+  for(i = 1; i < l; i++)
+    F2v_set(res, p[mael(elts,i,j)]);
+  return res;
+}
+
 GEN
 groupelts_set(GEN elts, long n)
 {
@@ -696,7 +854,7 @@ group_set(GEN G, long n)
   long i, l = lg(elts);
   for(i=1; i<l; i++)
     F2v_set(res,mael(elts,i,1));
-  avma = av;
+  set_avma(av);
   return res;
 }
 
@@ -740,13 +898,13 @@ dicyclicgroup(GEN g1, GEN g2, long s1, long s2)
 /* gen (vecvecsmall): coset generators
  * coset (vecsmall): vecsmall of coset number) */
 GEN
-group_quotient(GEN G, GEN H)
+groupelts_quotient(GEN elt, GEN H)
 {
   pari_sp ltop = avma;
   GEN  p2, p3;
   long i, j, a = 1;
-  long n = group_domain(G), o = group_order(H);
-  GEN  elt = group_elts(G,n), el;
+  long n = lg(gel(elt,1))-1, o = group_order(H);
+  GEN  el;
   long le = lg(elt)-1;
   GEN used = zero_F2v(le+1);
   long l = le/o;
@@ -773,6 +931,12 @@ group_quotient(GEN G, GEN H)
   return gerepilecopy(ltop,mkvec2(p2,p3));
 }
 
+GEN
+group_quotient(GEN G, GEN H)
+{
+  return groupelts_quotient(group_elts(G, group_domain(G)), H);
+}
+
 /*Compute the image of a permutation by a quotient map.*/
 GEN
 quotient_perm(GEN C, GEN p)
@@ -840,14 +1004,25 @@ quotient_group(GEN C, GEN G)
   return gerepilecopy(ltop,Q);
 }
 
+GEN
+quotient_groupelts(GEN C)
+{
+  GEN G = quo_get_gen(C);
+  long i, l = lg(G);
+  GEN Q = cgetg(l, t_VEC);
+  for (i = 1; i < l; ++i)
+    gel(Q,i) = quotient_perm(C, gel(G,i));
+  return Q;
+}
+
 /* Return 1 if g normalizes N, 0 otherwise */
 long
 group_perm_normalize(GEN N, GEN g)
 {
   pari_sp ltop = avma;
-  long r = gequal(vecvecsmall_sort(group_leftcoset(N, g)),
-                  vecvecsmall_sort(group_rightcoset(N, g)));
-  avma = ltop; return r;
+  long r = gequal(vecvecsmall_sort_shallow(group_leftcoset(N, g)),
+                  vecvecsmall_sort_shallow(group_rightcoset(N, g)));
+  return gc_long(ltop, r);
 }
 
 /* L is a list of subgroups, C is a coset and r a relative order.*/
@@ -898,7 +1073,7 @@ liftsubgroup(GEN C, GEN H, GEN S)
   return gerepilecopy(ltop,V);
 }
 
-/* 1:A4 2:S4 0: other */
+/* 1:A4, 2:S4, 3:F36, 0: other */
 long
 group_isA4S4(GEN G)
 {
@@ -906,6 +1081,14 @@ group_isA4S4(GEN G)
   GEN ord = grp_get_ord(G);
   long n = lg(ord);
   if (n != 4 && n != 5) return 0;
+  if (n==4 && ord[1]==3 && ord[2]==3 && ord[3]==4)
+  {
+    long i;
+    GEN p = gel(elt,1), q = gel(elt,2), r = gel(elt,3);
+    for(i=1; i<=36; i++)
+      if (p[r[i]]!=r[q[i]]) return 0;
+    return 3;
+  }
   if (ord[1]!=2 || ord[2]!=2 || ord[3]!=3) return 0;
   if (perm_commute(gel(elt,1),gel(elt,3))) return 0;
   if (n==4) return 1;
@@ -922,8 +1105,22 @@ group_subgroups(GEN G)
   GEN gen = grp_get_gen(G);
   GEN ord = grp_get_ord(G);
   long lM, i, j, n = lg(gen);
+  long t;
   if (n == 1) return trivialsubgroups();
-  if (group_isA4S4(G))
+  t = group_isA4S4(G);
+  if (t == 3)
+  {
+    GEN H = mkvec2(mkvec3(gel(gen,1), gel(gen,2), perm_sqr(gel(gen,3))),
+                   mkvecsmall3(3, 3, 2));
+    GEN S = group_subgroups(H);
+    GEN V = cgetg(11,t_VEC);
+    gel(V,1) = cyclicgroup(gel(gen,3),4);
+    for (i=2; i<10; i++)
+      gel(V,i) = cyclicgroup(perm_mul(gmael3(V,i-1,1,1),gel(gen,i%3==1 ? 2:1)),4);
+    gel(V,10) = G;
+    return gerepilecopy(ltop,shallowconcat(S,V));
+  }
+  else if (t)
   {
     GEN s = gel(gen,1);       /*s = (1,2)(3,4) */
     GEN t = gel(gen,2);       /*t = (1,3)(2,4) */
@@ -939,13 +1136,13 @@ group_subgroups(GEN G)
       GEN u = gel(gen,3);
       GEN v = gel(gen,4), w, u2;
       if (zv_equal(perm_conj(u,s), t)) /*u=(2,3,4)*/
-        u2 = perm_mul(u,u);
+        u2 = perm_sqr(u);
       else
       {
         u2 = u;
-        u = perm_mul(u,u);
+        u = perm_sqr(u);
       }
-      if (perm_order(v)==2)
+      if (perm_orderu(v)==2)
       {
         if (!perm_commute(s,v)) /*v=(1,2)*/
         {
@@ -957,10 +1154,10 @@ group_subgroups(GEN G)
       else
       {
         w = v;
-        if (!zv_equal(perm_mul(w,w), s)) /*w=(1,4,2,3)*/
+        if (!zv_equal(perm_sqr(w), s)) /*w=(1,4,2,3)*/
         {
           w = perm_conj(u,w);
-          if (!zv_equal(perm_mul(w,w), s)) w = perm_conj(u,w);
+          if (!zv_equal(perm_sqr(w), s)) w = perm_conj(u,w);
         }
         v = perm_mul(w,t); /*v=(1,2)*/
       }
@@ -974,8 +1171,8 @@ group_subgroups(GEN G)
   }
   else
   {
-    long osig = mael(factoru(ord[1]), 1, 1);
-    GEN sig = perm_pow(gel(gen,1), ord[1]/osig);
+    ulong osig = mael(factoru(ord[1]), 1, 1);
+    GEN sig = perm_powu(gel(gen,1), ord[1]/osig);
     H = cyclicgroup(sig,osig);
     sg3 = NULL;
   }
@@ -1035,10 +1232,10 @@ group_abelianHNF(GEN G, GEN S)
     GEN P, C = cgetg(n,t_COL);
     pari_sp av = avma;
     gel(M,i) = C;
-    P = perm_pow(gel(g,i), o[i]);
+    P = perm_inv(perm_powu(gel(g,i), o[i]));
     for(j=1; j<lg(S); j++)
       if (zv_equal(P, gel(S,j))) break;
-    avma = av;
+    set_avma(av);
     if (j==lg(S)) pari_err_BUG("galoisisabelian [inconsistent group]");
     j--;
     for(k=1; k<i; k++)
@@ -1088,25 +1285,61 @@ abelian_group(GEN v)
   return G;
 }
 
+static long
+groupelts_subgroup_isnormal(GEN G, GEN H)
+{
+  long i, n = lg(G);
+  for(i = 1; i < n; i++)
+    if (!group_perm_normalize(H, gel(G,i))) return 0;
+  return 1;
+}
+
 /*return 1 if H is a normal subgroup of G*/
 long
 group_subgroup_isnormal(GEN G, GEN H)
 {
-  GEN g = grp_get_gen(G);
-  long i, n = lg(g);
   if (lg(grp_get_gen(H)) > 1 && group_domain(G) != group_domain(H))
     pari_err_DOMAIN("group_subgroup_isnormal","domain(H)","!=",
                     strtoGENstr("domain(G)"), H);
+  return groupelts_subgroup_isnormal(grp_get_gen(G), H);
+}
+
+static GEN
+group_subgroup_kernel_set(GEN G, GEN H)
+{
+  pari_sp av;
+  GEN g = grp_get_gen(G);
+  long i, n = lg(g);
+  GEN S, elts;
+  long d = group_domain(G);
+  if (lg(grp_get_gen(H)) > 1 && group_domain(G) != group_domain(H))
+    pari_err_DOMAIN("group_subgroup_isnormal","domain(H)","!=",
+                    strtoGENstr("domain(G)"), H);
+  elts = group_elts(H,d);
+  S = groupelts_set(elts, d);
+  av = avma;
   for(i=1; i<n; i++)
-    if (!group_perm_normalize(H, gel(g,i))) return 0;
-  return 1;
+  {
+    F2v_and_inplace(S, groupelts_conj_set(elts,gel(g,i)));
+    set_avma(av);
+  }
+  return S;
+}
+
+int
+group_subgroup_is_faithful(GEN G, GEN H)
+{
+  pari_sp av = avma;
+  GEN K = group_subgroup_kernel_set(G,H);
+  F2v_clear(K,1);
+  return gc_long(av, F2v_equal0(K));
 }
 
 long
 groupelts_exponent(GEN elts)
 {
   long i, n = lg(elts)-1, expo = 1;
-  for(i=1; i<=n; i++) expo = ulcm(expo, perm_order(gel(elts,i)));
+  for(i=1; i<=n; i++) expo = ulcm(expo, perm_orderu(gel(elts,i)));
   return expo;
 }
 
@@ -1149,7 +1382,7 @@ groupelts_conjclasses(GEN elts, long *pnbcl)
         GEN h = perm_conj(gel(elts,j), g);
         long i2 = gen_search(elts,h,0,(void*)&vecsmall_lexcmp,&cmp_nodata);
         c[i2] = cl;
-        avma = av;
+        set_avma(av);
       }
   }
   if (pnbcl) *pnbcl = cl;
@@ -1189,7 +1422,7 @@ group_to_cc(GEN G)
   else
   {
     long i, l = lg(elts);
-    elts = gen_sort(elts,(void*)vecsmall_lexcmp,cmp_nodata); /* general case */
+    elts = gen_sort_shallow(elts,(void*)vecsmall_lexcmp,cmp_nodata);
     for (i = 1; i < l; i++)
       if (gel(elts,i)[1] != i) { flag = 0; break; }
   }
@@ -1271,3 +1504,73 @@ group_export(GEN G, long format)
   pari_err_FLAG("galoisexport");
   return NULL; /*-Wall*/
 }
+
+static GEN
+groupelts_cyclic_subgroups(GEN G)
+{
+  pari_sp av = avma;
+  long i, j, n = lg(G)-1;
+  GEN elts = zero_F2v(n+1), f;
+  GEN ord = cgetg(n+1, t_VECSMALL);
+  GEN V = cgetg(n+1, t_VEC);
+  for (i=1, j=1; i<=n; i++)
+  {
+    long k = 1, o, c = 0;
+    GEN p = gel(G, i);
+    if (F2v_coeff(elts, p[1])) continue;
+    o = perm_orderu(p);
+    ord[j] = o;
+    gel(V,j++) = p;
+    do
+    {
+      if (cgcd(o, ++c)==1) F2v_set(elts, p[k]);
+      k = p[k];
+    } while (k!=1);
+  }
+  setlg(ord, j);
+  setlg(V, j);
+  f = vecsmall_indexsort(ord);
+  return gerepilecopy(av, mkvec2(vecpermute(V, f), vecpermute(ord, f)));
+}
+
+GEN
+groupelts_to_group(GEN G)
+{
+  pari_sp av = avma;
+  GEN L = groupelts_cyclic_subgroups(G);
+  GEN cyc = gel(L,1), ord = gel(L,2);
+  long i, l = lg(cyc), n = lg(G)-1;
+  for (i = l-1; i >= 2; i--)
+  {
+    GEN p = gel(cyc,i);
+    long o = ord[i];
+    GEN H;
+    if (o == n) { set_avma(av); return cyclicgroup(p, o); }
+    H = cyclicgroup(p, o);
+    if (groupelts_subgroup_isnormal(G, H))
+    {
+      GEN C = groupelts_quotient(G, H);
+      GEN Q = quotient_groupelts(C);
+      GEN R = groupelts_to_group(Q);
+      if (!R) return gc_NULL(av);
+      return gerepilecopy(av, quotient_subgroup_lift(C, H, R));
+    }
+  }
+  if (n==12 && l==9 && ord[2]==2 && ord[3]==2 && ord[5]==3)
+    return gerepilecopy(av,
+      mkvec2(mkvec3(gel(cyc,2), gel(cyc,3), gel(cyc,5)), mkvecsmall3(2,2,3)));
+  if (n==24 && l==18 && ord[11]==3 && ord[15]==4 && ord[16]==4)
+  {
+    GEN t21 = perm_sqr(gel(cyc,15));
+    GEN t22 = perm_sqr(gel(cyc,16));
+    return gerepilecopy(av,
+      mkvec2(mkvec4(t21,t22, gel(cyc,11), gel(cyc,15)), mkvecsmall4(2,2,3,2)));
+  }
+  if (n==36 && l==24 && ord[11]==3 && ord[15]==4)
+  {
+    GEN t1 = gel(cyc,11), t3 = gel(cyc,15);
+    return gerepilecopy(av,
+      mkvec2(mkvec3(t1, perm_conj(t3, t1), t3), mkvecsmall3(3,3,4)));
+  }
+  return gc_NULL(av);
+}
