@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -248,6 +249,15 @@ random_F2x(long d, long vs)
   return F2x_renormalize(y,l);
 }
 
+GEN
+random_zv(long n)
+{
+  GEN y = cgetg(n+1, t_VECSMALL);
+  long i;
+  for (i=1; i<=n; i++) uel(y,i) = pari_rand();
+  return y;
+}
+
 GEN
 randomr(long prec)
 {
@@ -260,7 +270,7 @@ randomr(long prec)
   y = randomi(int2n(b));
   if (!signe(y)) return real_0_bit(b);
   affir(y, x); shiftr_inplace(x, - b);
-  avma = av; return x;
+  set_avma(av); return x;
 }
 
 static GEN
