@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -38,21 +39,20 @@ gp_ask_confirm(const char *s)
   pari_hit_return();
 }
 
-/* numerr < 0: from SIGINT */
+/* numerr < 0: after changing PARI stack size
+ * numerr > 0: normal error, including SIGINT */
 static void
 gp_err_recover(long numerr)
 {
   longjmp(env[s_env.n-1], numerr);
 }
-/* numerr < 0: from SIGINT */
+
+/* numerr >= 0 */
 static void
 gp_pre_recover(long numerr)
 {
-  if (numerr>=0)
-  {
-    out_puts(pariErr, "\n"); pariErr->flush();
-  }
-  longjmp(env[s_env.n-1], numerr);
+  out_puts(pariErr, "\n"); pariErr->flush();
+  gp_err_recover(numerr);
 }
 
 static void
@@ -111,7 +111,7 @@ gp_head(void)
 {
   pari_print_version();
   pari_putc('\n');
-  pari_center("Copyright (C) 2000-2018 The PARI Group");
+  pari_center("Copyright (C) 2000-2020 The PARI Group");
   pari_putc('\n');
   print_text("PARI/GP is free software, covered by the GNU General Public \
 License, and comes WITHOUT ANY WARRANTY WHATSOEVER.");
@@ -329,7 +329,7 @@ static GEN
 gp_main_loop(long ismain)
 {
   VOLATILE GEN z = gnil;
-  VOLATILE long t = 0;
+  VOLATILE long t = 0, r = 0;
   VOLATILE pari_sp av = avma;
   filtre_t F;
   Buffer *b = filtered_buffer(&F);
@@ -346,7 +346,7 @@ gp_main_loop(long ismain)
       tmp_restore(rec.file.file);
       gp_context_save(&rec);
     }
-    avma = av = pari_mainstack->top;
+    set_avma(av = pari_mainstack->top);
     parivstack_reset();
     kill_buffers_upto(b);
     pari_alarm(0);
@@ -364,6 +364,7 @@ gp_main_loop(long ismain)
     {
       reset_ctrlc();
       timer_start(GP_DATA->T);
+      walltimer_start(GP_DATA->Tw);
       pari_set_last_newline(1);
     }
     if (gp_meta(b->buf,ismain)) continue;
@@ -372,17 +373,29 @@ gp_main_loop(long ismain)
     if (!ismain) continue;
 
     t = timer_delay(GP_DATA->T);
+    r = walltimer_delay(GP_DATA->Tw);
     if (!pari_last_was_newline()) pari_putc('\n');
     pari_alarm(0);
     if (t && GP_DATA->chrono)
     {
-      pari_puts("time = ");
-      pari_puts(gp_format_time(t));
+      if (pari_mt_nbthreads==1)
+      {
+        pari_puts("time = ");
+        pari_puts(gp_format_time(t));
+      }
+      else
+      {
+        pari_puts("cpu time = ");
+        pari_puts(gp_format_time(t));
+        pari_puts(", real time = ");
+        pari_puts(gp_format_time(r));
+      }
+      pari_puts(".\n");
     }
     if (GP_DATA->simplify) z = simplify_shallow(z);
-    pari_add_hist(z, t);
+    pari_add_hist(z, t, r);
     if (z != gnil && ! is_silent(b->buf) ) gp_output(z);
-    avma = av;
+    set_avma(av);
     parivstack_reset();
   }
 }
@@ -403,7 +416,7 @@ read_main(const char *s)
     else z = gp_main_loop(0);
   }
   if (!z) err_printf("... skipping file '%s'\n", s);
-  avma = pari_mainstack->top;
+  set_avma(pari_mainstack->top);
 }
 
 static const char *
@@ -427,7 +440,7 @@ break_loop(int numerr)
   pari_sp av;
 
   if (numerr == e_SYNTAX) return 0;
-  if (numerr == e_STACK) { evalstate_clone(); avma = pari_mainstack->top; }
+  if (numerr == e_STACK) { evalstate_clone(); set_avma(pari_mainstack->top); }
   gp_context_save(&rec1);
 
   b = filtered_buffer(&F);
@@ -449,7 +462,7 @@ break_loop(int numerr)
   {
     GEN x;
     long er, br_status;
-    avma = av;
+    set_avma(av);
     gp_context_save(&rec2);
     if ((er=setjmp(env[nenv])))
     {
@@ -555,8 +568,10 @@ main(int argc, char **argv)
   pari_init_defaults();
   pari_library_path = DL_DFLT_NAME;
   pari_stack_init(&s_A,sizeof(*A),(void**)&A);
-  pari_init_opts(1000000 * sizeof(long), 0, INIT_SIGm | INIT_noPRIMEm | INIT_noIMTm);
+  /* must be defined here in case an error is raised in pari_init_opts, e.g.
+   * when parsing function prototypes */
   cb_pari_err_recover = gp_err_recover;
+  pari_init_opts(1000000 * sizeof(long), 0, INIT_SIGm | INIT_noPRIMEm | INIT_noIMTm);
   cb_pari_pre_recover = gp_pre_recover;
   cb_pari_break_loop = break_loop;
   cb_pari_is_interactive = is_interactive;
@@ -586,6 +601,7 @@ main(int argc, char **argv)
   if (GP_DATA->flags & gpd_TEXMACS) init_texmacs();
 
   timer_start(GP_DATA->T);
+  walltimer_start(GP_DATA->Tw);
   if (!(GP_DATA->flags & gpd_QUIET)) gp_head();
   if (GP_DATA->flags & gpd_TEST) init_test();
   if (s_A.n)
