@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -81,6 +82,7 @@ DTOL(double t) { return (long)(t + 0.5); }
 
 static const long PS_WIDTH = 1120 - 60; /* 1400 - 60 for hi-res */
 static const long PS_HEIGH = 800 - 40; /* 1120 - 60 for hi-res */
+static const long PS_SCALE = 1000; /* Allowing 64x zoom on 500ppi */
 
 static void
 _psdraw_scale(PARI_plot *T, GEN w, GEN x, GEN y)
@@ -89,7 +91,7 @@ _psdraw_scale(PARI_plot *T, GEN w, GEN x, GEN y)
   FILE *F = fopen(current_psfile, "a");
   if (!F) pari_err_FILE("postscript file",current_psfile);
   fputs(rect2ps(w,x,y,T), F);
-  fclose(F); avma = av;
+  fclose(F); set_avma(av);
 }
 static void
 _psdraw(PARI_plot *T, GEN w, GEN x, GEN y)
@@ -126,12 +128,8 @@ pari_get_svgplot(PARI_plot *T)
 /**                      RECTPLOT FUNCTIONS                        **/
 /**                                                                **/
 /********************************************************************/
-static void
-get_plot_null(PARI_plot *T)
-{ (void)T; pari_err(e_MISC,"high resolution graphics disabled"); }
-
 void
-pari_init_graphics(void) { pari_get_plot = &get_plot_null; }
+pari_init_graphics(void) { pari_get_plot = &pari_get_svgplot; }
 
 void
 pari_set_plot_engine(void (*plot)(PARI_plot *))
@@ -608,7 +606,7 @@ plotpoints(long ne, GEN X, GEN Y)
     px[i] = gtodouble(gel(X,i));
     py[i] = gtodouble(gel(Y,i));
   }
-  plotpoints0(ne,px,py,lx); avma = av;
+  plotpoints0(ne,px,py,lx); set_avma(av);
 }
 
 /* ROt_ML */
@@ -658,7 +656,7 @@ plotlines(long ne, GEN X, GEN Y, long flag)
     x[i] = gtodouble(gel(X,i));
     y[i] = gtodouble(gel(Y,i));
   }
-  rectlines0(ne,x,y,lx,flag); avma = av;
+  rectlines0(ne,x,y,lx,flag); set_avma(av);
 }
 
 /* ROt_ST */
@@ -1017,19 +1015,15 @@ set_xrange(dblPointList *f, double x)
 { if (x < f->xsml) f->xsml = x;
   if (x > f->xbig) f->xbig = x; }
 static void
-Appendxat(dblPointList *f, dblPointList *l,long n,double x)
-{ (l->d)[n] = x; l->nb = n+1; set_xrange(f,x); }
+Appendx(dblPointList *f, dblPointList *l, double x)
+{ (l->d)[l->nb++] = x; set_xrange(f,x); }
 static void
 set_yrange(dblPointList *f, double y)
 { if (y < f->ysml) f->ysml = y;
   if (y > f->ybig) f->ybig = y; }
 static void
-Appendyat(dblPointList *f, dblPointList *l,long n,double y)
-{ (l->d)[n] = y; l->nb = n+1; set_yrange(f,y); }
-static void
-Appendx(dblPointList *f, dblPointList *l,double x) { Appendxat(f,l,l->nb,x); }
-static void
-Appendy(dblPointList *f, dblPointList *l,double y) { Appendyat(f,l,l->nb,y); }
+Appendy(dblPointList *f, dblPointList *l, double y)
+{ (l->d)[l->nb++] = y; set_yrange(f,y); }
 
 static void
 get_xy(long cplx, GEN t, double *x, double *y)
@@ -1104,29 +1098,31 @@ gtodblList(GEN data, long flags)
   lx1 = lg(gel(data,1));
   if (!param && lx1 == 1) return NULL;
 
+  /* Check input first */
   if (nl == 1 && !cplx) pari_err_DIM("gtodblList");
-  /* Allocate memory, then convert coord. to double */
+  for (i = 0; i < nl; i += cplx? 1: 2)
+  {
+    GEN x = gel(data,i+1), y = cplx? NULL: gel(data,i+2);
+    long lx = lg(x);
+    if (!is_vec_t(typ(x))) pari_err_TYPE("gtodblList",x);
+    if (y)
+    {
+      if (!is_vec_t(typ(y))) pari_err_TYPE("gtodblList",y);
+      if (lg(y) != lx || (!param && lx != lx1)) pari_err_DIM("gtodblList");
+    }
+  }
+  /* Now allocate memory, then convert coord. to double */
   l = (dblPointList*)pari_malloc((cplx? 2*nl: nl)*sizeof(dblPointList));
   L = &l[0];
   for (i = 0; i < nl; i += cplx? 1: 2)
   {
-    GEN x = gel(data,i+1), y;
-    long lx = lg(x);
-    if (!is_vec_t(typ(x))) pari_err_TYPE("gtodblList",x);
-    if (cplx) y = NULL;
-    else
-    {
-      y = gel(data,i+2);
-      if (!is_vec_t(typ(y))) pari_err_TYPE("gtodblList",y);
-      if (lg(y) != lx || (!param && lx != lx1)) pari_err_DIM("gtodblList");
-    }
-    lx--;
+    GEN x = gel(data,i+1), y = cplx? NULL: gel(data,i+2);
+    long lx = lg(x)-1;
     l[i].d   = X = (double*)pari_malloc(lx*sizeof(double));
     l[i+1].d = Y = (double*)pari_malloc(lx*sizeof(double));
     for (j=1; j<=lx; j++) get_xy_from_vec2(cplx, x, y, j, X+(j-1), Y+(j-1));
     l[i].nb = l[i+1].nb = lx;
   }
-
   /* Compute extremas */
   if (param)
   {
@@ -1157,13 +1153,13 @@ gtodblList(GEN data, long flags)
   return l;
 }
 
-/* (x+y)/2 */
+/* x,y t_REAL; return (x+y)/2,  */
 static GEN
 rmiddle(GEN x, GEN y) { GEN z = addrr(x,y); shiftr_inplace(z,-1); return z; }
 
 static void
 single_recursion(void *E, GEN(*eval)(void*,GEN), dblPointList *pl,
-                 GEN xleft,double yleft, GEN xright,double yright,long depth)
+                 GEN xl,double yl, GEN xr,double yr,long depth)
 {
   GEN xx;
   pari_sp av = avma;
@@ -1171,39 +1167,38 @@ single_recursion(void *E, GEN(*eval)(void*,GEN), dblPointList *pl,
 
   if (depth==RECUR_MAXDEPTH) return;
 
-  xx = rmiddle(xleft,xright);
+  xx = rmiddle(xl,xr);
   yy = gtodouble(eval(E,xx));
 
-  if (dy && fabs(yleft+yright-2*yy) < dy*RECUR_PREC) return;
-  single_recursion(E,eval, pl,xleft,yleft, xx,yy, depth+1);
+  if (dy && fabs(yl+yr-2*yy) < dy*RECUR_PREC) return;
+  single_recursion(E,eval, pl,xl,yl, xx,yy, depth+1);
   Appendx(&pl[0],&pl[0],rtodbl(xx));
   Appendy(&pl[0],&pl[1],yy);
-  single_recursion(E,eval, pl,xx,yy, xright,yright, depth+1);
-  avma = av;
+  single_recursion(E,eval, pl,xx,yy, xr,yr, depth+1);
+  set_avma(av);
 }
 
 static void
 param_recursion(void *E,GEN(*eval)(void*,GEN), long cplx, dblPointList *pl,
-  GEN tleft,double xleft, double yleft,
-  GEN tright,double xright,double yright, long depth)
+  GEN tl,double xl, double yl, GEN tr,double xr,double yr, long depth)
 {
-  GEN tt;
+  GEN t;
   pari_sp av = avma;
   double xx, dy=pl[0].ybig - pl[0].ysml;
   double yy, dx=pl[0].xbig - pl[0].xsml;
 
   if (depth==RECUR_MAXDEPTH) return;
 
-  tt = rmiddle(tleft,tright);
-  get_xy(cplx, eval(E,tt), &xx,&yy);
+  t = rmiddle(tl,tr);
+  get_xy(cplx, eval(E,t), &xx,&yy);
 
-  if (dx && dy && fabs(xleft+xright-2*xx) < dx*RECUR_PREC
-               && fabs(yleft+yright-2*yy) < dy*RECUR_PREC) return;
-  param_recursion(E,eval, cplx, pl, tleft,xleft,yleft, tt,xx,yy, depth+1);
+  if (dx && dy && fabs(xl+xr-2*xx) < dx*RECUR_PREC
+               && fabs(yl+yr-2*yy) < dy*RECUR_PREC) return;
+  param_recursion(E,eval, cplx, pl, tl,xl,yl, t,xx,yy, depth+1);
   Appendx(&pl[0],&pl[0],xx);
   Appendy(&pl[0],&pl[1],yy);
-  param_recursion(E,eval,cplx, pl, tt,xx,yy, tright,xright,yright, depth+1);
-  avma = av;
+  param_recursion(E,eval,cplx, pl, t,xx,yy, tr,xr,yr, depth+1);
+  set_avma(av);
 }
 
 /* Graph 'code' for parameter values in [a,b], using 'N' sample points
@@ -1218,10 +1213,9 @@ plotrecthin(void *E, GEN(*eval)(void*, GEN), GEN a, GEN b, ulong flags,
   const long param = flags & (PLOT_PARAMETRIC|PLOT_COMPLEX);
   const long recur = flags & PLOT_RECURSIVE;
   const long cplx = flags & PLOT_COMPLEX;
-  GEN t,dx,x;
+  GEN t, dx, x;
   dblPointList *pl;
   long tx, i, j, sig, nc, nl, ncoords, nbpoints, non_vec = 0;
-  pari_sp av = avma;
 
   sig = gcmp(b,a); if (!sig) return NULL;
   if (sig < 0) swap(a, b);
@@ -1244,7 +1238,7 @@ plotrecthin(void *E, GEN(*eval)(void*, GEN), GEN a, GEN b, ulong flags,
     nl = lg(t);
     nc = nl-1;
   }
-  if (!nc) { avma = av; return NULL; }
+  if (!nc) return NULL;
   if (recur && nc > 1) pari_err_TYPE("ploth [multi-curves + recursive]",t);
 
   ncoords = cplx? 2*nl: nl;
@@ -1269,78 +1263,68 @@ plotrecthin(void *E, GEN(*eval)(void*, GEN), GEN a, GEN b, ulong flags,
   dx = divru(gtofp(gsub(b,a),prec), N-1);
   if (recur)
   { /* recursive plot */
-    double yleft, yright = 0;
+    double yl, yr = 0;
     if (param)
     {
-      GEN tleft = cgetr(prec), tright = cgetr(prec);
-      double xleft, xright = 0;
+      GEN tl = cgetr(prec), tr = cgetr(prec);
+      double xl, xr = 0;
       pari_sp av2 = avma;
-      affgr(a,tleft);
-      t = eval(E, tleft);
-      get_xy(cplx,t, &xleft,&yleft);
-      for (i=0; i<N-1; i++, avma = av2)
+      affgr(a, tl);
+      t = eval(E, tl);
+      get_xy(cplx,t, &xl,&yl);
+      for (i=0; i<N-1; i++, set_avma(av2))
       {
-        if (i) { affrr(tright,tleft); xleft = xright; yleft = yright; }
-        addrrz(tleft,dx,tright);
-        t = eval(E, tright);
-        get_xy(cplx,t, &xright,&yright);
-        Appendx(&pl[0],&pl[0],xleft);
-        Appendy(&pl[0],&pl[1],yleft);
-        param_recursion(E,eval, cplx, pl, tleft,xleft,yleft, tright,xright,yright, 0);
+        if (i) { affrr(tr,tl); xl = xr; yl = yr; }
+        addrrz(tl,dx,tr);
+        t = eval(E, tr);
+        get_xy(cplx,t, &xr,&yr);
+        Appendx(&pl[0],&pl[0],xl);
+        Appendy(&pl[0],&pl[1],yl);
+        param_recursion(E,eval, cplx, pl, tl,xl,yl, tr,xr,yr, 0);
       }
-      Appendx(&pl[0],&pl[0],xright);
-      Appendy(&pl[0],&pl[1],yright);
+      Appendx(&pl[0],&pl[0],xr);
+      Appendy(&pl[0],&pl[1],yr);
     }
     else /* single curve */
     {
-      GEN xleft = cgetr(prec), xright = cgetr(prec);
+      GEN xl = cgetr(prec), xr = cgetr(prec);
       pari_sp av2 = avma;
-      affgr(a,xleft);
-      yleft = gtodouble(eval(E,xleft));
-      for (i=0; i<N-1; i++, avma = av2)
+      affgr(a,xl);
+      yl = gtodouble(eval(E,xl));
+      for (i=0; i<N-1; i++, set_avma(av2))
       {
-        addrrz(xleft,dx,xright);
-        yright = gtodouble(eval(E,xright));
-
-        Appendx(&pl[0],&pl[0],rtodbl(xleft));
-        Appendy(&pl[0],&pl[1],yleft);
-
-        single_recursion(E,eval, pl,xleft,yleft,xright,yright,0);
-        affrr(xright,xleft); yleft = yright;
+        addrrz(xl,dx,xr);
+        yr = gtodouble(eval(E,xr));
+        Appendx(&pl[0],&pl[0],rtodbl(xl));
+        Appendy(&pl[0],&pl[1],yl);
+        single_recursion(E,eval, pl,xl,yl,xr,yr,0);
+        affrr(xr,xl); yl = yr;
       }
-      Appendx(&pl[0],&pl[0],rtodbl(xright));
-      Appendy(&pl[0],&pl[1],yright);
+      Appendx(&pl[0],&pl[0],rtodbl(xr));
+      Appendy(&pl[0],&pl[1],yr);
     }
   }
-  else /* non-recursive plot */
+  else /* nonrecursive plot */
   {
-    pari_sp av2;
-    GEN worker = NULL, vx = NULL;
-    long pending = 0, workid;
-    struct pari_mt pt;
+    GEN V, X = cgetg(N+1, t_VEC);
+    for (i = 1; i <= N; i++) { gel(X,i) = x; x = addrr(x,dx); }
     if (flags & PLOT_PARA && eval == gp_call)
     {
-      worker = snm_closure(is_entry("_parvector_worker"), mkvec((GEN)E));
-      mt_queue_start_lim(&pt, worker, N-1);
-      vx = mkvec(x);
+      GEN worker = snm_closure(is_entry("_parapply_slice_worker"),
+                               mkvec((GEN)E));
+      V = gen_parapply_slice(worker, X, mt_nbthreads());
+    }
+    else
+    {
+      V = cgetg(N+1, t_VEC);
+      for (i = 1; i <= N; i++) gel(V,i) = eval(E,gel(X,i));
     }
-    av2 = avma;
     if (param)
     {
-      for (i=0; i<N || pending; i++, avma = av2)
+      for (i = 1; i <= N; i++)
       {
-        long k, nt;
-        if (worker)
-        {
-          mt_queue_submit(&pt, i, i<N ? vx : NULL);
-          t = mt_queue_get(&pt, &workid, &pending);
-          if (!t) continue;
-        }
-        else
-        {
-          t = eval(E,x);
-          workid = i;
-        }
+        long nt, k, j;
+        t = gel(V,i);
         if (typ(t) != t_VEC)
         {
           if (cplx) nt = 1;
@@ -1354,60 +1338,27 @@ plotrecthin(void *E, GEN(*eval)(void*, GEN), GEN a, GEN b, ulong flags,
         {
           double xx, yy;
           get_xy_from_vec(cplx, t, &j, &xx, &yy);
-          Appendxat(&pl[0], &pl[k++], workid, xx);
-          Appendyat(&pl[0], &pl[k++], workid, yy);
+          Appendx(&pl[0], &pl[k++], xx);
+          Appendy(&pl[0], &pl[k++], yy);
         }
-        if (i<N) affrr(addrr(x,dx), x);
       }
     }
     else if (non_vec)
-      for (i=0; i<N || pending; i++, avma = av2)
+      for (i = 1; i <= N; i++)
       {
-        if (worker)
-        {
-          mt_queue_submit(&pt, i, i<N ? vx : NULL);
-          t = mt_queue_get(&pt, &workid, &pending);
-        }
-        else
-        {
-          t = eval(E,x);
-          workid = i;
-        }
-        if (t) Appendyat(&pl[0], &pl[1], workid, gtodouble(t));
-        if (i<N)
-        {
-          pl[0].d[i] = gtodouble(x);
-          affrr(addrr(x,dx), x);
-        }
+        Appendy(&pl[0], &pl[1], gtodouble(gel(V,i)));
+        pl[0].d[i-1] = gtodouble(gel(X,i));
       }
-    else /* vector of non-parametric curves */
-      for (i=0; i<N || pending; i++, avma = av2)
+    else /* vector of nonparametric curves */
+      for (i = 1; i <= N; i++)
       {
-        if (worker)
-        {
-          mt_queue_submit(&pt, i, i<N ? vx : NULL);
-          t = mt_queue_get(&pt, &workid, &pending);
-        }
-        else
-        {
-          t = eval(E,x);
-          workid = i;
-        }
-        if (t)
-        {
-          if (typ(t) != t_VEC || lg(t) != nl) pari_err_DIM("plotrecth");
-          for (j=1; j<nl; j++)
-            Appendyat(&pl[0], &pl[j], workid, gtodouble(gel(t,j)));
-        }
-        if (i<N)
-        {
-          pl[0].d[i] = gtodouble(x);
-          affrr(addrr(x,dx), x);
-        }
+        t = gel(V,i);
+        if (typ(t) != t_VEC || lg(t) != nl) pari_err_DIM("plotrecth");
+        for (j = 1; j < nl; j++) Appendy(&pl[0], &pl[j], gtodouble(gel(t,j)));
+        pl[0].d[i-1] = gtodouble(gel(X,i));
       }
-    if (worker) mt_queue_end(&pt);
   }
-  pl[0].nb = nc; avma = av; return pl;
+  pl[0].nb = nc; return pl;
 }
 
 static GEN
@@ -1443,10 +1394,10 @@ rectsplines(long ne, double *x, double *y, long lx, long flag)
 
     xa++; ya++;
     if (param) {
-      gel(pol3,1) = polint_i(tas, xa, X, 4, NULL);
-      gel(pol3,2) = polint_i(tas, ya, X, 4, NULL);
+      gel(pol3,1) = polintspec(tas, xa, X, 4, NULL);
+      gel(pol3,2) = polintspec(tas, ya, X, 4, NULL);
     } else {
-      pol3 = polint_i(xa, ya, X, 4, NULL);
+      pol3 = polintspec(xa, ya, X, 4, NULL);
       tas = xa;
     }
     /* Start with 3 points */
@@ -1454,9 +1405,9 @@ rectsplines(long ne, double *x, double *y, long lx, long flag)
                i== 0 ? gel(tas,0) : gel(tas,1),
                i==lx-4 ? gel(tas,3) : gel(tas,2),
                fl, 2, DEFAULTPREC);
-    avma = av;
+    set_avma(av);
   }
-  avma = av0;
+  set_avma(av0);
 }
 
 static void
@@ -1484,7 +1435,10 @@ fmt_convert(GEN fmt, GEN w, GEN x, GEN y, PARI_plot *T)
 
 static void
 Draw(PARI_plot *T, GEN w, GEN x, GEN y)
-{ if (T->draw) T->draw(T, w,x,y); else get_plot_null(NULL); }
+{
+  if (!T->draw) pari_err(e_MISC,"high resolution graphics disabled");
+  T->draw(T, w,x,y);
+}
 static void
 set_range(double m, double M, double *sml, double *big)
 {
@@ -1632,7 +1586,7 @@ plotrecthrawin(GEN fmt, PARI_plot *W, long ne, dblPointList *data, long flags)
     plotkill(w[2]);
     if (fmt) return s;
   }
-  avma = av;
+  set_avma(av);
   retmkvec4(dbltor(xsml), dbltor(xbig), dbltor(ysml), dbltor(ybig));
 }
 
@@ -1648,8 +1602,9 @@ static GEN
 plotrecth_i(GEN fmt, void *E, GEN(*f)(void*,GEN), PARI_plot *T, long ne,
             GEN a,GEN b, ulong flags,long n, long prec)
 {
+  pari_sp av = avma;
   dblPointList *pl = plotrecthin(E,f, a,b, flags, n, prec);
-  return plotrecthrawin(fmt, T, ne, pl, flags);
+  set_avma(av); return plotrecthrawin(fmt, T, ne, pl, flags);
 }
 GEN
 plotrecth(void *E, GEN(*f)(void*,GEN), long ne, GEN a,GEN b,
@@ -1658,21 +1613,22 @@ plotrecth(void *E, GEN(*f)(void*,GEN), long ne, GEN a,GEN b,
 GEN
 plotrecth0(long ne, GEN a,GEN b,GEN code,ulong flags,long n, long prec)
 { EXPR_WRAP(code, plotrecth(EXPR_ARG, ne, a,b, flags, n, prec)); }
+static GEN
+_ploth(void *E, GEN(*f)(void*,GEN), GEN a, GEN b,long flags, long n, long prec)
+{
+  PARI_plot T; pari_get_plot(&T);
+  return plotrecth_i(NULL, E,f, &T, NUMRECT-1, a,b, flags,n, prec);
+}
 GEN
 ploth(void *E, GEN(*f)(void*,GEN), GEN a, GEN b,long flags, long n, long prec)
-{
-  PARI_plot T; pari_get_plot(&T);
-  return plotrecth_i(NULL, E,f, &T, NUMRECT-1, a,b, flags&~PLOT_PARA,n, prec);
-}
+{ return _ploth(E, f, a, b, flags&~PLOT_PARA, n, prec); }
 GEN
 parploth(GEN a, GEN b, GEN code, long flags, long n, long prec)
-{
-  PARI_plot T; pari_get_plot(&T);
-  return plotrecth_i(NULL, code,gp_call,&T, NUMRECT-1, a,b, flags|PLOT_PARA,n, prec);
-}
+{ return _ploth(code, gp_call, a, b, flags|PLOT_PARA, n, prec); }
 GEN
 ploth0(GEN a, GEN b, GEN code, long flags,long n, long prec)
 { EXPR_WRAP(code, ploth(EXPR_ARG, a,b,flags,n, prec)); }
+
 GEN
 psploth(void *E, GEN(*f)(void*,GEN), GEN a,GEN b, long flags, long n, long prec)
 {
@@ -1682,19 +1638,27 @@ psploth(void *E, GEN(*f)(void*,GEN), GEN a,GEN b, long flags, long n, long prec)
 GEN
 psploth0(GEN a, GEN b, GEN code, long flags, long n, long prec)
 { EXPR_WRAP(code, psploth(EXPR_ARG, a, b, flags, n, prec)); }
-GEN
-plothexport(GEN fmt, void *E, GEN(*f)(void*,GEN), GEN a,GEN b, long flags,
+
+static GEN
+_plothexport(GEN fmt, void *E, GEN(*f)(void*,GEN), GEN a,GEN b, long flags,
             long n, long prec)
 {
   pari_sp av = avma;
   GEN s;
   PARI_plot T; pari_get_fmtplot(fmt, &T);
-  s = plotrecth_i(fmt, E,f, &T, NUMRECT-1, a,b, flags&~PLOT_PARA,n, prec);
+  s = plotrecth_i(fmt, E,f, &T, NUMRECT-1, a,b, flags,n, prec);
   return gerepileuptoleaf(av, s);
 }
 GEN
+plothexport(GEN fmt, void *E, GEN(*f)(void*,GEN), GEN a,GEN b, long flags,
+            long n, long prec)
+{ return _plothexport(fmt, E, f, a, b, flags&~PLOT_PARA, n, prec); }
+GEN
 plothexport0(GEN fmt, GEN a, GEN b, GEN code, long flags, long n, long prec)
 { EXPR_WRAP(code, plothexport(fmt, EXPR_ARG, a, b, flags, n, prec)); }
+GEN
+parplothexport(GEN fmt, GEN a, GEN b, GEN code, long flags, long n, long prec)
+{ return _plothexport(fmt, code, gp_call, a, b, flags|PLOT_PARA, n, prec); }
 
 /* Draw list of points */
 static GEN
@@ -1774,19 +1738,24 @@ wxy_n(GEN wxy)
   return 0;/*LCOV_EXCL_LINE*/
 }
 static void
-wxy_init(GEN wxy, GEN W, GEN X, GEN Y, PARI_plot *T)
+wxy_init(GEN wxy, GEN *pW, GEN *pX, GEN *pY, PARI_plot *T)
 {
-  long i, l = lg(X);
+  long i, n = wxy_n(wxy);
+  GEN W, X, Y;
+  *pW = W = cgetg(n+1, t_VECSMALL); /* win number */
+  *pX = X = cgetg(n+1, t_VECSMALL);
+  *pY = Y = cgetg(n+1, t_VECSMALL); /* (x,y)-offset */
   if (typ(wxy) == t_INT)
   {
     W[1] = itos(wxy); check_rect_init(W[1]);
     X[1] = 0;
     Y[1] = 0; return;
   }
-  for (i = 1; i < l; i++)
+  for (i = 1; i <= n; i++)
   {
     GEN w = gel(wxy,3*i-2), x = gel(wxy,3*i-1), y = gel(wxy,3*i);
     if (typ(w) != t_INT) pari_err_TYPE("plotdraw",w);
+    W[i] = itos(w); check_rect_init(W[i]);
     if (T) {
       X[i] = DTOL(gtodouble(x)*(T->width - 1));
       Y[i] = DTOL(gtodouble(y)*(T->height - 1));
@@ -1794,23 +1763,24 @@ wxy_init(GEN wxy, GEN W, GEN X, GEN Y, PARI_plot *T)
       X[i] = gtos(x);
       Y[i] = gtos(y);
     }
-    W[i] = itos(w); check_rect_init(W[i]);
   }
 }
 /* if flag is set, rescale wrt T */
 static void
 gendraw(PARI_plot *T, GEN wxy, long flag)
 {
-  long n = wxy_n(wxy);
-  /* malloc mandatory in case draw() forks then pari_close() */
-  GEN W = cgetalloc(t_VECSMALL, n+1); /* win number */
-  GEN X = cgetalloc(t_VECSMALL, n+1);
-  GEN Y = cgetalloc(t_VECSMALL, n+1); /* (x,y)-offset */
-  wxy_init(wxy, W,X,Y, flag? T: NULL);
+  GEN w, x, y, W, X, Y;
+  long i, l;
+  wxy_init(wxy, &w,&x,&y, flag? T: NULL);
+  l = lg(w);
+  /* malloc mandatory in case draw() forks then pari_close(). Done after
+   * wxy_init to avoid leak on error */
+  W = cgetalloc(t_VECSMALL, l);
+  X = cgetalloc(t_VECSMALL, l);
+  Y = cgetalloc(t_VECSMALL, l);
+  for (i = 1; i < l; i++) { W[i] = w[i]; X[i] = x[i]; Y[i] = y[i]; }
   Draw(T,W,X,Y);
-  pari_free(W);
-  pari_free(X);
-  pari_free(Y);
+  pari_free(W); pari_free(X); pari_free(Y);
 }
 void
 psdraw(GEN wxy, long flag)
@@ -1823,13 +1793,10 @@ GEN
 plotexport(GEN fmt, GEN wxy, long flag)
 {
   pari_sp av = avma;
-  long n = wxy_n(wxy);
-  GEN w = cgetg(n+1, t_VECSMALL);
-  GEN x = cgetg(n+1, t_VECSMALL);
-  GEN y = cgetg(n+1, t_VECSMALL);
+  GEN w, x, y;
   PARI_plot _T, *T = flag? &_T: NULL;
   if (T) pari_get_plot(T);
-  wxy_init(wxy, w, x, y, T);
+  wxy_init(wxy, &w, &x, &y, T);
   return gerepileuptoleaf(av, fmt_convert(fmt, w, x, y, T));
 }
 
@@ -2125,7 +2092,7 @@ static void
 ps_rect(void *data, long x, long y, long w, long h)
 {
   pari_str *S = (pari_str*)data;
-  str_printf(S,"%ld %ld m %ld %ld l %ld %ld l %ld %ld l closepath stroke\n",
+  str_printf(S,"%ld %ld m %ld %ld l %ld %ld l %ld %ld l closepath currentlinejoin 0 setlinejoin stroke setlinejoin\n",
              y,x, y,x+w, y+h,x+w, y+h,x);
 }
 
@@ -2133,7 +2100,7 @@ static void
 ps_fillrect(void *data, long x, long y, long w, long h)
 {
   pari_str *S = (pari_str*)data;
-  str_printf(S,"%ld %ld m %ld %ld l %ld %ld l %ld %ld l closepath fill\n",
+  str_printf(S,"%ld %ld m %ld %ld l %ld %ld l %ld %ld l closepath currentlinejoin 0 setlinejoin fill setlinejoin\n",
              y,x, y,x+w, y+h,x+w, y+h,x);
 }
 
@@ -2177,11 +2144,11 @@ rect2ps_i(GEN w, GEN x, GEN y, PARI_plot *T, int plotps)
   struct plot_eng pl;
   PARI_plot U;
   pari_str S;
-  double xs = 0.65, ys = 0.65;
+  double xs = 0.65*PS_SCALE, ys = 0.65*PS_SCALE;
   if (T) /* res wrt T dimens */
   {
     if (plotps)
-      xs = ys = 1;
+      xs = ys = PS_SCALE;
     else
     {
       xs *= ((double)PS_WIDTH) / T->width;
@@ -2196,12 +2163,15 @@ rect2ps_i(GEN w, GEN x, GEN y, PARI_plot *T, int plotps)
   /* Definitions taken from post terminal of Gnuplot. */
   str_printf(&S, "%%!\n\
 50 50 translate\n\
+1 %d div 1 %d div scale\n\
+1 setlinejoin\n\
 /p {moveto 0 2 rlineto 2 0 rlineto 0 -2 rlineto closepath fill} def\n\
 /c0 {0 0 0 setrgbcolor} def\n\
 /c {setrgbcolor} def\n\
 /l {lineto} def\n\
 /m {moveto} def\n"
-"/Times-Roman findfont %ld scalefont setfont\n", DTOL(T->fheight * xs));
+"/Times-Roman findfont %ld scalefont setfont\n",
+PS_SCALE, PS_SCALE, DTOL(T->fheight * xs));
 
   pl.sc = &ps_sc;
   pl.pt = &ps_point;
@@ -2214,7 +2184,7 @@ rect2ps_i(GEN w, GEN x, GEN y, PARI_plot *T, int plotps)
   pl.pl = T;
   pl.data = (void*)&S;
 
-  if (plotps) str_printf(&S,"0 %ld translate -90 rotate\n", T->height - 50);
+  if (plotps) str_printf(&S,"0 %ld translate -90 rotate\n", (T->height - 50)*PS_SCALE);
   gen_draw(&pl, w, x, y, xs, ys);
   str_puts(&S,"stroke showpage\n");
   *S.cur = 0; return S.string;
