@@ -5,7 +5,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -26,8 +27,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
  * pointer is NULL  (but the computations will happen nonetheless).
  */
 
-GEN
-bezout(GEN a, GEN b, GEN *pu, GEN *pv)
+static GEN
+bezout_lehmer(GEN a, GEN b, GEN *pu, GEN *pv)
 {
   GEN t,u,u1,v,v1,d,d1,q,r;
   GEN *pt;
@@ -111,7 +112,7 @@ bezout(GEN a, GEN b, GEN *pu, GEN *pv)
     q = dvmdii(absi_shallow(a), d, &d1);
     if (!signe(d1))                /* a == qb */
     {
-      avma = av;
+      set_avma(av);
       if (pu) *pu = gen_0;
       if (pv) *pv = sb < 0 ? gen_m1 : gen_1;
       return icopy(d);
@@ -191,7 +192,7 @@ bezout(GEN a, GEN b, GEN *pu, GEN *pv)
     u = subii(muliu(u,xu), muliu(u1, xv));
     v = subii(muliu(v,xu), muliu(v1, xv));
     if (s < 0) { sa = -sa; sb = -sb; }
-    avma = av;
+    set_avma(av);
     if (pu) *pu = sa < 0 ? negi(u) : icopy(u);
     if (pv) *pv = sb < 0 ? negi(v) : icopy(v);
     if (g == 1) return gen_1;
@@ -201,9 +202,52 @@ bezout(GEN a, GEN b, GEN *pu, GEN *pv)
   /* get here when the final sprint was skipped (d1 was zero already).
    * Now the matrix is final, and d contains the gcd.
    */
-  avma = av;
+  set_avma(av);
   if (pu) *pu = sa < 0 ? negi(u) : icopy(u);
   if (pv) *pv = sb < 0 ? negi(v) : icopy(v);
   return icopy(d);
 }
 
+static GEN
+addmulmul(GEN u, GEN v, GEN x, GEN y)
+{ return addii(mulii(u, x),mulii(v, y)); }
+
+static GEN
+bezout_halfgcd(GEN x, GEN y, GEN *ptu, GEN *ptv)
+{
+  pari_sp av=avma;
+  GEN u, v, R = matid(2);
+  while (lgefint(y)-2 >= EXTGCD_HALFGCD_LIMIT)
+  {
+    GEN M = HGCD0(x,y);
+    R = ZM_mul2(R, gel(M, 1));
+    x = gel(M,2); y = gel(M,3);
+    if (signe(y) && expi(y)<magic_threshold(x))
+    {
+      GEN r, q = dvmdii(x, y, &r);
+      x = y; y = r;
+      R = mulq(R, q);
+    }
+    if (gc_needed(av, 1))
+      gerepileall(av,3,&x,&y,&R);
+  }
+  y = bezout_lehmer(x,y,&u,&v);
+  R = ZM_inv2(R);
+  if (ptu) *ptu = addmulmul(u,v,gcoeff(R,1,1),gcoeff(R,2,1));
+  if (ptv) *ptv = addmulmul(u,v,gcoeff(R,1,2),gcoeff(R,2,2));
+  return y;
+}
+
+GEN
+bezout(GEN x, GEN y, GEN *ptu, GEN *ptv)
+{
+  pari_sp ltop=avma;
+  GEN d;
+  if (lgefint(y)-2 >= EXTGCD_HALFGCD_LIMIT)
+    d = bezout_halfgcd(x, y, ptu, ptv);
+  else
+    d = bezout_lehmer(x, y, ptu, ptv);
+  if (ptv) gerepileall(ltop,ptu?3:2,&d,ptv,ptu);
+  else gerepileall(ltop,ptu?2:1,&d,ptu);
+  return d;
+}
