@@ -5,7 +5,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -21,8 +22,12 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "paripriv.h"
 #include "../src/kernel/none/tune-gen.h"
 
-void pari_kernel_init(void) { }
-void pari_kernel_close(void) { }
+void
+pari_kernel_init(void) { }
+void
+pari_kernel_close(void) { }
+const char *
+pari_kernel_version(void) { return ""; }
 
 /* NOTE: arguments of "spec" routines (muliispec, addiispec, etc.) aren't
  * GENs but pairs (long *a, long na) representing a list of digits (in basis
@@ -32,7 +37,7 @@ void pari_kernel_close(void) { }
 #define LIMBS(x)  ((x)+2)
 #define NLIMBS(x) (lgefint(x)-2)
 
-/* Normalize a non-negative integer */
+/* Normalize a nonnegative integer */
 GEN
 int_normalize(GEN x, long known_zero_words)
 {
@@ -43,7 +48,7 @@ int_normalize(GEN x, long known_zero_words)
   for (i = 2+known_zero_words; i < lx; i++)
     if (x[i]) break;
   x0 = x; i -= 2; x += i;
-  if (x0 == (GEN)avma) avma = (pari_sp)x;
+  if (x0 == (GEN)avma) set_avma((pari_sp)x);
   else stackdummy((pari_sp)(x0+i), (pari_sp)x0);
   lx -= i;
   x[0] = evaltyp(t_INT) | evallg(lx);
@@ -150,7 +155,7 @@ adduispec(ulong s, GEN x, long nx)
   while (xd > x) *--zd = *--xd;
   *--zd = evalsigne(1) | evallgefint(lz);
   *--zd = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)zd; return zd;
+  return gc_const((pari_sp)zd, zd);
 }
 
 GEN
@@ -194,7 +199,7 @@ addiispec(GEN x, GEN y, long nx, long ny)
   zd += i+1;
   *--zd = evalsigne(1) | evallgefint(lz);
   *--zd = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)zd; return zd;
+  return gc_const((pari_sp)zd, zd);
 }
 
 /* assume x >= s */
@@ -222,7 +227,7 @@ subiuspec(GEN x, ulong s, long nx)
     do  *--zd = *--xd; while (xd > x);
   *--zd = evalsigne(1) | evallgefint(lz);
   *--zd = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)zd; return zd;
+  return gc_const((pari_sp)zd, zd);
 }
 
 /* assume x > y */
@@ -257,7 +262,7 @@ subiispec(GEN x, GEN y, long nx, long ny)
   zd += i+1;
   *--zd = evalsigne(1) | evallgefint(lz);
   *--zd = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)zd; return zd;
+  return gc_const((pari_sp)zd, zd);
 }
 
 static void
@@ -342,8 +347,8 @@ shiftispec(GEN x, long nx, long n)
       shift_right(yd,x, 0,ny, 0,m);
       if (yd[0] == 0)
       {
-        if (ny==1) { avma = (pari_sp)(y+3); return gen_0; }
-        ny--; avma = (pari_sp)(++y);
+        if (ny==1) return gc_const((pari_sp)(y+3), gen_0);
+        ny--; set_avma((pari_sp)(++y));
       }
     } else {
       for (i=0; i<ny; i++) yd[i]=x[i];
@@ -457,7 +462,7 @@ muluispec(ulong x, GEN y, long ny)
   if (hiremainder) *--z = hiremainder; else lz--;
   *--z = evalsigne(1) | evallgefint(lz);
   *--z = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)z; return z;
+  return gc_const((pari_sp)z, z);
 }
 
 /* a + b*|Y| */
@@ -482,7 +487,7 @@ addumului(ulong a, ulong b, GEN Y)
   if (hiremainder) *--z = hiremainder; else lz--;
   *--z = evalsigne(1) | evallgefint(lz);
   *--z = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)z; return z;
+  return gc_const((pari_sp)z, z);
 }
 
 /***********************************************************************/
@@ -755,7 +760,7 @@ divri(GEN x, GEN y)
   }
   lx = lg(x); z = cgetr(lx); av = avma;
   affrr(divrr(x, itor(y, lx+1)), z);
-  avma = av; return z;
+  return gc_const(av, z);
 }
 
 /* Integer division x / y: such that sign(r) = sign(x)
@@ -819,7 +824,7 @@ TRIVIAL:
     if (z) *z = icopy(x);
     return gen_0;
   }
-DIVIDE: /* quotient is non-zero */
+DIVIDE: /* quotient is nonzero */
   av=avma; if (sx<0) sy = -sy;
   r1 = new_chunk(lx); sh = bfffo(y[2]);
   if (sh)
@@ -882,14 +887,14 @@ DIVIDE: /* quotient is non-zero */
     while (lz--) *--qd = *--xd;
     *--qd = evalsigne(sy) | evallgefint(lq);
     *--qd = evaltyp(t_INT) | evallg(lq);
-    avma = (pari_sp)qd; return (GEN)qd;
+    return gc_const((pari_sp)qd, (GEN)qd);
   }
 
   j=lq; while (j<lx && !x[j]) j++;
   lz = lx-j;
   if (z == ONLY_REM)
   {
-    if (lz==0) { avma = av; return gen_0; }
+    if (lz==0) return gc_const(av, gen_0);
     rd = (ulong*)av; lr = lz+2;
     xd = (ulong*)(x + lx);
     if (!sh) while (lz--) *--rd = *--xd;
@@ -908,7 +913,7 @@ DIVIDE: /* quotient is non-zero */
     }
     *--rd = evalsigne(sx) | evallgefint(lr);
     *--rd = evaltyp(t_INT) | evallg(lr);
-    avma = (pari_sp)rd; return (GEN)rd;
+    return gc_const((pari_sp)rd, (GEN)rd);
   }
 
   lr = lz+2;
@@ -952,7 +957,7 @@ DIVIDE: /* quotient is non-zero */
     while (lr--) *--qd = *--rd;
     *z = (GEN)qd;
   }
-  avma = (pari_sp)qd; return q;
+  return gc_const((pari_sp)qd, q);
 }
 
 /* Montgomery reduction.
@@ -1036,7 +1041,7 @@ red_montgomery(GEN T, GEN N, ulong inv)
   Td = (GEN)av;
   while (*scratch == 0 && Te > scratch) scratch++; /* strip leading 0s */
   while (Te > scratch) *--Td = *--Te;
-  k = (GEN)av - Td; if (!k) { avma = av; return gen_0; }
+  k = (GEN)av - Td; if (!k) return gc_const(av, gen_0);
   k += 2;
   *--Td = evalsigne(1) | evallgefint(k);
   *--Td = evaltyp(t_INT) | evallg(k);
@@ -1050,7 +1055,7 @@ red_montgomery(GEN T, GEN N, ulong inv)
     || cmpii(Td, addii(shifti(T, -s), N)) >= 0) pari_err_BUG("red_montgomery");
 }
 #endif
-  avma = (pari_sp)Td; return Td;
+  return gc_const((pari_sp)Td, Td);
 }
 
 /* EXACT INTEGER DIVISION */
@@ -1103,7 +1108,7 @@ diviuexact_i(GEN x, ulong y)
   z[0] = evaltyp(t_INT)|evallg(lz);
   z[1] = evalsigne(1)|evallg(lz);
   if (lz == 2) pari_err_OP("exact division", x, utoi(y));
-  avma = (pari_sp)z; return z;
+  return gc_const((pari_sp)z, z);
 }
 
 /* assume y != 0 and the division is exact */
@@ -1125,16 +1130,16 @@ diviuexact(GEN x, ulong y)
   av = avma; (void)new_chunk(lx); vy = vals(y);
   if (vy) {
     y >>= vy;
-    if (y == 1) { avma = av; return shifti(x, -vy); }
+    if (y == 1) { set_avma(av); return shifti(x, -vy); }
     x = shifti(x, -vy);
     if (lx == 3) {
       ulong q = uel(x,2) / y;
-      avma = av;
+      set_avma(av);
       if (!q) pari_err_OP("exact division", x, utoi(y));
       return (s > 0)? utoipos(q): utoineg(q);
     }
   } else x = icopy(x);
-  avma = av;
+  set_avma(av);
   z = diviuexact_i(x, y);
   setsigne(z, s); return z;
 }
@@ -1166,7 +1171,7 @@ diviiexact(GEN x, GEN y)
     x = shifti(x,-vy); lx = lgefint(x);
   }
   else x = icopy(x); /* necessary because we destroy x */
-  avma = av; /* will erase our x,y when exiting */
+  set_avma(av); /* will erase our x,y when exiting */
   /* now y is odd */
   ly = lgefint(y);
   if (ly == 3)
@@ -1215,7 +1220,7 @@ diviiexact(GEN x, GEN y)
   z[0] = evaltyp(t_INT)|evallg(lz);
   z[1] = evalsigne((sx+sy)? 1: -1) | evallg(lz);
   if (lz == 2) pari_err_OP("exact division", x, y);
-  avma = (pari_sp)z; return z;
+  return gc_const((pari_sp)z, z);
 }
 
 /* assume yz != and yz | x */
@@ -1276,7 +1281,7 @@ muliispec_basecase(GEN x, GEN y, long nx, long ny)
   if (*zd == 0) { zd++; lz--; } /* normalize */
   *--zd = evalsigne(1) | evallgefint(lz);
   *--zd = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)zd; return zd;
+  return gc_const((pari_sp)zd, zd);
 }
 
 INLINE GEN
@@ -1339,7 +1344,7 @@ END:
   if (*zd == 0) { zd++; lz--; } /* normalize */
   *--zd = evalsigne(1) | evallgefint(lz);
   *--zd = evaltyp(t_INT) | evallg(lz);
-  avma=(pari_sp)zd; return zd;
+  return gc_const((pari_sp)zd, zd);
 }
 
 /********************************************************************/
@@ -1464,7 +1469,7 @@ muliifft_dit(ulong o, ulong ord, GEN M, GEN FFT, long d, long step)
     GEN b = Zf_mulsqrt2(Zf_sub(gel(FFT,i), gel(FFT,i+hstep), M), j, ord, M);
     affii(a,gel(FFT,i));
     affii(b,gel(FFT,i+hstep));
-    avma = av;
+    set_avma(av);
   }
   if (hstep>1)
   {
@@ -1493,7 +1498,7 @@ muliifft_dis(ulong o, ulong ord, GEN M, GEN FFT, long d, long step)
     GEN b = Zf_sub(gel(FFT,i), z, M);
     affii(a,gel(FFT,i));
     affii(b,gel(FFT,i+hstep));
-    avma = av;
+    set_avma(av);
   }
 }
 
@@ -1570,13 +1575,13 @@ sqrispec_fft(GEN a, long na)
   for(i=1; i<=n; i++)
   {
     affii(Zf_sqr(gel(FFT,i), M), gel(FFT,i));
-    avma=av;
+    set_avma(av);
   }
   muliifft_dis(ord-o, ord, M, FFT, 0, n);
   for(i=1; i<=n; i++)
   {
     affii(Zf_shift(gel(FFT,i), (ord>>1)-k, M), gel(FFT,i));
-    avma=av;
+    set_avma(av);
   }
   return gerepileuptoint(ltop, muliifft_unspliti(FFT,bs,2+len));
 }
@@ -1604,14 +1609,14 @@ muliispec_fft(GEN a, GEN b, long na, long nb)
   for(i=1; i<=n; i++)
   {
     affii(Zf_mul(gel(FFT,i), gel(FFTb,i), M), gel(FFT,i));
-    avma=av2;
+    set_avma(av2);
   }
-  avma=av;
+  set_avma(av);
   muliifft_dis(ord-o, ord, M, FFT, 0, n);
   for(i=1; i<=n; i++)
   {
     affii(Zf_shift(gel(FFT,i),(ord>>1)-k,M), gel(FFT,i));
-    avma=av;
+    set_avma(av);
   }
   return gerepileuptoint(ltop, muliifft_unspliti(FFT,bs,2+len));
 }
@@ -2048,7 +2053,7 @@ sqrtremi(GEN N, GEN *r)
   else
     S = sqrtispec(n, l2, &R);
 
-  if (!r) { avma = (pari_sp)S; return gerepileuptoint(av, S); }
+  if (!r) { set_avma((pari_sp)S); return gerepileuptoint(av, S); }
   gerepileall(av, 2, &S, &R); *r = R; return S;
 }
 
@@ -2080,7 +2085,7 @@ sqrtr_abs(GEN x)
     if ( u&HIGHBIT || (u == ~HIGHBIT && cmpii(c,b) > 0))
       roundr_up_ip(res, l+2);
   }
-  avma = (pari_sp)res; return res;
+  return gc_const((pari_sp)res, res);
 }
 
 #else /* use t_REAL: currently much slower (quadratic division) */
@@ -2143,10 +2148,10 @@ sqrtr_abs(GEN x)
     setlg(a, l1 + 2);
     setlg(t, l1 + 2);
     affrr(addrr(t, divrr(a,t)), t); shiftr_inplace(t, -1);
-    avma = av;
+    set_avma(av);
   }
   affrr(t,y); shiftr_inplace(y, (ex>>1));
-  avma = av0; return y;
+  return gc_const(av0, y);
 }
 
 #endif
@@ -2168,7 +2173,7 @@ convi_dac(GEN x, ulong l, ulong *res)
   x1=dvmdii(x,powuu(1000000000UL,m),&x2);
   convi_dac(x1,l-m,res+m);
   convi_dac(x2,m,res);
-  avma=ltop;
+  set_avma(ltop);
 }
 
 /* convert integer --> base 10^9 [not memory clean] */
