@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -118,23 +119,6 @@ FIND:
 /**                   HASH TABLE MANIPULATIONS                     **/
 /**                                                                **/
 /********************************************************************/
-/* return hashing value for identifier s */
-static ulong
-hashvalue(const char *s)
-{
-  ulong n = 0, c;
-  while ( (c = (ulong)*s++) ) n = (n<<1) ^ c;
-  return n;
-}
-
-static ulong
-hashvalue_raw(const char *s, long len)
-{
-  long n = 0, i;
-  for(i=0; i<len; i++) { n = (n<<1) ^ *s; s++; }
-  return n;
-}
-
 static void
 insertep(entree *ep, entree **table, ulong hash)
 {
@@ -166,7 +150,7 @@ initep(const char *name, long len)
 static entree *
 findentry(const char *s, long len, entree **T, int insert)
 {
-  ulong hash = hashvalue_raw(s, len);
+  ulong hash = hash_str_len(s, len);
   entree *ep;
   for (ep = T[hash % functions_tblsz]; ep; ep = ep->next)
     if (ep->hash == hash)
@@ -195,39 +179,69 @@ fetch_entry(const char *s) { return fetch_entry_raw(s, strlen(s)); }
 /*                  SYNTACTICAL ANALYZER FOR GP                    */
 /*                                                                 */
 /*******************************************************************/
+static GEN
+readseq_i(char *t)
+{
+  if (gp_meta(t,0)) return gnil;
+  return closure_evalres(pari_compile_str(t));
+}
 GEN
 readseq(char *t)
-{
-  pari_sp av = avma;
-  GEN x;
-  if (gp_meta(t,0)) return gnil;
-  x = pari_compile_str(t);
-  return gerepileupto(av, closure_evalres(x));
-}
+{ pari_sp av = avma; return gerepileupto(av, readseq_i(t)); }
 
 /* filtered readseq = remove blanks and comments */
 GEN
 gp_read_str(const char *s)
+{ pari_sp av = avma; return gerepileupto(av, readseq_i(gp_filter(s))); }
+
+GEN
+compile_str(const char *s) { return pari_compile_str(gp_filter(s)); }
+
+GEN
+gp_read_str_bitprec(const char *s, long bitprec)
 {
-  char *t = gp_filter(s);
-  GEN x = readseq(t);
-  pari_free(t); return x;
+  GEN x;
+  push_localbitprec(bitprec);
+  x = gp_read_str(s);
+  pop_localprec();
+  return x;
 }
 
 GEN
-compile_str(const char *s)
-{
-  char *t = gp_filter(s);
-  GEN x = pari_compile_str(t);
-  pari_free(t); return x;
-}
+gp_read_str_prec(const char *s, long prec)
+{ return gp_read_str_bitprec(s, prec2nbits(prec)); }
 
+/* valid return type */
+static int
+isreturn(char c)
+{ return c == 'l' || c == 'v' || c == 'i' || c == 'm' || c == 'u'; }
+
+/* if is known that 2 commas follow s; base-10 signed integer followed
+ * by comma? */
+static int
+is_long(const char *s)
+{
+  while (isspace(*s)) s++;
+  if (*s == '+' || *s == '-') s++;
+  while (isdigit(*s)) s++;
+  return *s == ',';
+}
+/* if is known that 2 commas follow s; base-10 unsigned integer followed
+ * by comma? */
+static int
+is_ulong(const char *s)
+{
+  while (isspace(*s)) s++;
+  if (*s == '+') s++;
+  while (isdigit(*s)) s++;
+  return *s == ',';
+}
 static long
 check_proto(const char *code)
 {
   long arity = 0;
-  const char *s = code, *old;
-  if (*s == 'l' || *s == 'v' || *s == 'i' || *s == 'm' || *s == 'u') s++;
+  const char *s = code;
+  if (isreturn(*s)) s++;
   while (*s && *s != '\n') switch (*s++)
   {
     case '&':
@@ -245,33 +259,51 @@ check_proto(const char *code)
     case 'p':
     case 'b':
     case 'r':
-      arity++;
-      break;
+      arity++; break;
     case 'E':
     case 's':
       if (*s == '*') s++;
-      arity++;
-      break;
+      arity++; break;
     case 'D':
-      if (*s == 'G' || *s == '&' || *s == 'n' || *s == 'I' || *s == 'E'
-                    || *s == 'V' || *s == 'P' || *s == 's' || *s == 'r')
+      switch(*s)
       {
-        if (*s != 'V') arity++;
-        s++; break;
+        case 'G': case '&': case 'n': case 'I': case 'E':
+        case 'P': case 's': case 'r': s++; arity++; break;
+        case 'V': s++; break;
+        case 0:
+          pari_err(e_SYNTAX,"function has incomplete prototype", s,code);
+          break;
+        default:
+        {
+          const char *p;
+          long i;
+          for(i = 0, p = s; *p && i < 2; p++) i += *p==','; /* skip 2 commas */
+          if (i < 2) pari_err(e_SYNTAX,"missing comma",s,code);
+          arity++;
+          switch(p[-2])
+          {
+            case 'L':
+              if (!is_long(s)) pari_err(e_SYNTAX,"not a long",s,code);
+              break;
+            case 'U':
+              if (!is_ulong(s)) pari_err(e_SYNTAX,"not an ulong",s,code);
+              break;
+            case 'G': case 'r': case 's': case 'M':
+              break;
+            default: pari_err(e_SYNTAX,"incorrect type",s-2,code);
+          }
+          s = p;
+        }
       }
-      old = s; while (*s && *s != ',') s++;
-      if (*s != ',') pari_err(e_SYNTAX, "missing comma", old, code);
       break;
     case 'V':
     case '=':
     case ',': break;
     case '\n': break; /* Before the mnemonic */
-
-    case 'm':
-    case 'l':
-    case 'i':
-    case 'v': pari_err(e_SYNTAX, "this code has to come first", s-1, code);
-    default: pari_err(e_SYNTAX, "unknown parser code", s-1, code);
+    default:
+      if (isreturn(s[-1]))
+        pari_err(e_SYNTAX, "this code has to come first", s-1, code);
+      pari_err(e_SYNTAX, "unknown parser code", s-1, code);
   }
   if (arity > 20) pari_err_IMPL("functions with more than 20 parameters");
   return arity;
@@ -337,13 +369,6 @@ addhelp(const char *e, char *s)
   if (f && !EpSTATIC(ep)) pari_free(f);
 }
 
-GEN
-type0(GEN x)
-{
-  const char *s = type_name(typ(x));
-  return strtoGENstr(s);
-}
-
 /*******************************************************************/
 /*                                                                 */
 /*                              PARSER                             */
@@ -547,7 +572,7 @@ real_read(pari_sp av, const char **s, GEN y, long prec)
       {
         if (**s == 'E' || **s == 'e') {
           n = exponent(s);
-          if (!signe(y)) { avma = av; return real_0_digits(n); }
+          if (!signe(y)) { set_avma(av); return real_0_digits(n); }
           break;
         }
         --*s; return y; /* member */
@@ -556,14 +581,14 @@ real_read(pari_sp av, const char **s, GEN y, long prec)
       n = old - *s;
       if (**s != 'E' && **s != 'e')
       {
-        if (!signe(y)) { avma = av; return real_0(prec); }
+        if (!signe(y)) { set_avma(av); return real_0(prec); }
         break;
       }
     }
     /* Fall through */
     case 'E': case 'e':
       n += exponent(s);
-      if (!signe(y)) { avma = av; return real_0_digits(n); }
+      if (!signe(y)) { set_avma(av); return real_0_digits(n); }
   }
   l = nbits2prec(bit_accuracy(lgefint(y)));
   if (l < prec) l = prec; else prec = l;
@@ -798,34 +823,6 @@ pari_lex(union token_value *yylval, struct node_loc *yylloc, char **lex)
   return (unsigned char) *(*lex)++;
 }
 
-/********************************************************************/
-/**                                                                **/
-/**                            STRINGS                             **/
-/**                                                                **/
-/********************************************************************/
-
-/* return the first n0 chars of s as a GEN [s may not be 0-terminated] */
-GEN
-strntoGENstr(const char *s, long n0)
-{
-  long n = nchar2nlong(n0+1);
-  GEN x = cgetg(n+1, t_STR);
-  char *t = GSTR(x);
-  x[n] = 0;
-  strncpy(t, s, n0); t[n0] = 0; return x;
-}
-
-GEN
-strtoGENstr(const char *s) { return strntoGENstr(s, strlen(s)); }
-
-GEN
-chartoGENstr(char c)
-{
-  GEN x = cgetg(2, t_STR);
-  char *t = GSTR(x);
-  t[0] = c; t[1] = 0; return x;
-}
-
 /********************************************************************/
 /*                                                                  */
 /*                Formal variables management                       */
@@ -835,8 +832,6 @@ static THREAD long max_priority, min_priority;
 static THREAD long max_avail; /* max variable not yet used */
 static THREAD long nvar; /* first GP free variable */
 static hashtable *h_polvar;
-static struct pari_varstate global_varstate;
-static long *global_varpriority;
 
 void
 varstate_save(struct pari_varstate *s)
@@ -911,25 +906,25 @@ varstate_restore(struct pari_varstate *s)
 }
 
 void
-pari_thread_init_varstate(void)
+pari_set_varstate(long *vp, struct pari_varstate *vs)
 {
-  long i;
-  var_restore(&global_varstate);
-  varpriority = (long*)newblock((MAXVARN+2)) + 1;
-  varpriority[-1] = 1-LONG_MAX;
-  for (i = 0; i < max_avail; i++) varpriority[i] = global_varpriority[i];
-}
-
-void
-pari_pthread_init_varstate(void)
-{
-  varstate_save(&global_varstate);
-  global_varpriority = varpriority;
+  var_restore(vs);
+  varpriority = (long*)newblock(MAXVARN+2) + 1;
+  memcpy(varpriority-1,vp-1,(MAXVARN+2)*sizeof(long));
 }
 
+/* must come before destruction of functions_hash */
 void
 pari_var_close(void)
 {
+  GEN h = hash_values(h_polvar);
+  long i, l = lg(h);
+  for (i = 1; i < l; i++)
+  {
+    long v = h[i];
+    entree *ep = varentries[v];
+    if (ep && ep != is_entry(ep->name)) pari_free(ep);
+  }
   free((void*)varentries);
   free((void*)(varpriority-1));
   hash_destroy(h_polvar);
@@ -1120,7 +1115,7 @@ static void
 fill_hashtable_single(entree **table, entree *ep)
 {
   EpSETSTATIC(ep);
-  insertep(ep, table,  hashvalue(ep->name));
+  insertep(ep, table, hash_str(ep->name));
   if (ep->code) ep->arity = check_proto(ep->code);
   ep->pvalue = NULL;
 }
@@ -1143,6 +1138,13 @@ pari_add_function(entree *ep)
 /**                                                                **/
 /********************************************************************/
 
+GEN
+arity0(GEN C)
+{
+  if (typ(C)!=t_CLOSURE) pari_err_TYPE("arity", C);
+  return utoi(closure_arity(C));
+}
+
 #define ALIAS(ep) (entree *) ((GEN)ep->value)[1]
 
 entree *
