@@ -4,7 +4,8 @@ This file is part of the PARI package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -74,8 +75,10 @@ pari_translate_string(const char *src, char *s, char *entry)
 static GEN
 strntoGENexp(const char *str, long len)
 {
-  GEN z = cgetg(1+nchar2nlong(len-1), t_STR);
+  long n = nchar2nlong(len-1);
+  GEN z = cgetg(1+n, t_STR);
   const char *t = str+1;
+  z[n] = 0;
   if (!translate(&t, GSTR(z))) compile_err("run-away string",str);
   return z;
 }
@@ -102,11 +105,12 @@ struct frame_s
 };
 
 static THREAD pari_stack s_opcode, s_operand, s_data, s_lvar;
-static THREAD pari_stack s_dbginfo, s_frame;
+static THREAD pari_stack s_dbginfo, s_frame, s_accesslex;
 static THREAD char *opcode;
 static THREAD long *operand;
+static THREAD long *accesslex;
 static THREAD GEN *data;
-static THREAD long offset;
+static THREAD long offset, nblex;
 static THREAD struct vars_s *localvars;
 static THREAD const char **dbginfo, *dbgstart;
 static THREAD struct frame_s *frames;
@@ -116,17 +120,19 @@ pari_init_compiler(void)
 {
   pari_stack_init(&s_opcode,sizeof(*opcode),(void **)&opcode);
   pari_stack_init(&s_operand,sizeof(*operand),(void **)&operand);
+  pari_stack_init(&s_accesslex,sizeof(*operand),(void **)&accesslex);
   pari_stack_init(&s_data,sizeof(*data),(void **)&data);
   pari_stack_init(&s_lvar,sizeof(*localvars),(void **)&localvars);
   pari_stack_init(&s_dbginfo,sizeof(*dbginfo),(void **)&dbginfo);
   pari_stack_init(&s_frame,sizeof(*frames),(void **)&frames);
-  offset=-1;
+  offset=-1; nblex=0;
 }
 void
 pari_close_compiler(void)
 {
   pari_stack_delete(&s_opcode);
   pari_stack_delete(&s_operand);
+  pari_stack_delete(&s_accesslex);
   pari_stack_delete(&s_data);
   pari_stack_delete(&s_lvar);
   pari_stack_delete(&s_dbginfo);
@@ -135,8 +141,8 @@ pari_close_compiler(void)
 
 struct codepos
 {
-  long opcode, data, localvars, frames;
-  long offset;
+  long opcode, data, localvars, frames, accesslex;
+  long offset, nblex;
   const char *dbgstart;
 };
 
@@ -144,8 +150,10 @@ static void
 getcodepos(struct codepos *pos)
 {
   pos->opcode=s_opcode.n;
+  pos->accesslex=s_accesslex.n;
   pos->data=s_data.n;
   pos->offset=offset;
+  pos->nblex=nblex;
   pos->localvars=s_lvar.n;
   pos->dbgstart=dbgstart;
   pos->frames=s_frame.n;
@@ -157,11 +165,13 @@ compilestate_reset(void)
 {
   s_opcode.n=0;
   s_operand.n=0;
+  s_accesslex.n=0;
   s_dbginfo.n=0;
   s_data.n=0;
   s_lvar.n=0;
   s_frame.n=0;
   offset=-1;
+  nblex=0;
   dbgstart=NULL;
 }
 
@@ -170,8 +180,10 @@ compilestate_save(struct pari_compilestate *comp)
 {
   comp->opcode=s_opcode.n;
   comp->operand=s_operand.n;
+  comp->accesslex=s_accesslex.n;
   comp->data=s_data.n;
   comp->offset=offset;
+  comp->nblex=nblex;
   comp->localvars=s_lvar.n;
   comp->dbgstart=dbgstart;
   comp->dbginfo=s_dbginfo.n;
@@ -183,8 +195,10 @@ compilestate_restore(struct pari_compilestate *comp)
 {
   s_opcode.n=comp->opcode;
   s_operand.n=comp->operand;
+  s_accesslex.n=comp->accesslex;
   s_data.n=comp->data;
   offset=comp->offset;
+  nblex=comp->nblex;
   s_lvar.n=comp->localvars;
   dbgstart=comp->dbgstart;
   s_dbginfo.n=comp->dbginfo;
@@ -192,63 +206,104 @@ compilestate_restore(struct pari_compilestate *comp)
 }
 
 static GEN
-getfunction(const struct codepos *pos, long arity, long nbmvar, GEN text, long gap)
+gcopyunclone(GEN x) { GEN y = gcopy(x); gunclone(x); return y; }
+
+static void
+access_push(long x)
 {
-  long lop =s_opcode.n+1-pos->opcode;
-  long ldat=s_data.n+1-pos->data;
-  long lfram=s_frame.n+1-pos->frames;
-  GEN cl=cgetg(nbmvar?8:(text?7:6),t_CLOSURE);
-  GEN frpc, fram, dbg;
+  long a = pari_stack_new(&s_accesslex);
+  accesslex[a] = x;
+}
+
+static GEN
+genctx(long nbmvar, long paccesslex)
+{
+  GEN acc = const_vec(nbmvar,gen_1);
+  long i, lvl = 1 + nbmvar;
+  for (i = paccesslex; i<s_accesslex.n; i++)
+  {
+    long a = accesslex[i];
+    if (a > 0) { lvl+=a; continue; }
+    a += lvl;
+    if (a <= 0) pari_err_BUG("genctx");
+    if (a <= nbmvar)
+      gel(acc, a) = gen_0;
+  }
+  s_accesslex.n = paccesslex;
+  for (i = 1; i<=nbmvar; i++)
+    if (signe(gel(acc,i))==0)
+      access_push(i-nbmvar-1);
+  return acc;
+}
+
+static GEN
+getfunction(const struct codepos *pos, long arity, long nbmvar, GEN text,
+            long gap)
+{
+  long lop  = s_opcode.n+1 - pos->opcode;
+  long ldat = s_data.n+1 - pos->data;
+  long lfram = s_frame.n+1 - pos->frames;
+  GEN cl = cgetg(nbmvar && text? 8: (text? 7: 6), t_CLOSURE);
+  GEN frpc, fram, dbg, op, dat;
   char *s;
   long i;
+
   cl[1] = arity;
   gel(cl,2) = cgetg(nchar2nlong(lop)+1, t_STR);
-  gel(cl,3) = cgetg(lop,  t_VECSMALL);
-  gel(cl,4) = cgetg(ldat, t_VEC);
+  gel(cl,3) = op = cgetg(lop, t_VECSMALL);
+  gel(cl,4) = dat = cgetg(ldat, t_VEC);
   dbg = cgetg(lop,  t_VECSMALL);
   frpc = cgetg(lfram,  t_VECSMALL);
   fram = cgetg(lfram,  t_VEC);
   gel(cl,5) = mkvec3(dbg, frpc, fram);
   if (text) gel(cl,6) = text;
-  if (nbmvar) gel(cl,7) = zerovec(nbmvar);
-  s=GSTR(gel(cl,2))-1;
-  for(i=1;i<lop;i++)
+  s = GSTR(gel(cl,2)) - 1;
+  for (i = 1; i < lop; i++)
   {
-    s[i] = opcode[i+pos->opcode-1];
-    mael(cl, 3, i) = operand[i+pos->opcode-1];
-    dbg[i] = dbginfo[i+pos->opcode-1]-dbgstart;
-    if (dbg[i]<0) dbg[i]+=gap;
+    long j = i+pos->opcode-1;
+    s[i] = opcode[j];
+    op[i] = operand[j];
+    dbg[i] = dbginfo[j] - dbgstart;
+    if (dbg[i] < 0) dbg[i] += gap;
   }
-  s[i]=0;
-  s_opcode.n=pos->opcode;
-  s_operand.n=pos->opcode;
-  s_dbginfo.n=pos->opcode;
-  for(i=1;i<ldat;i++)
-    if(data[i+pos->data-1])
-    {
-      gmael(cl, 4, i) = gcopy(data[i+pos->data-1]);
-      gunclone(data[i+pos->data-1]);
-    }
-  s_data.n=pos->data;
-  while (s_lvar.n>pos->localvars && !localvars[s_lvar.n-1].inl)
+  s[i] = 0;
+  s_opcode.n = pos->opcode;
+  s_operand.n = pos->opcode;
+  s_dbginfo.n = pos->opcode;
+  if (lg(cl)==8)
+    gel(cl,7) = genctx(nbmvar, pos->accesslex);
+  else if (nbmvar==0)
+    s_accesslex.n = pos->accesslex;
+  else
+  {
+    pari_sp av = avma;
+    (void) genctx(nbmvar, pos->accesslex);
+    set_avma(av);
+  }
+  for (i = 1; i < ldat; i++)
+    if (data[i+pos->data-1]) gel(dat,i) = gcopyunclone(data[i+pos->data-1]);
+  s_data.n = pos->data;
+  while (s_lvar.n > pos->localvars && !localvars[s_lvar.n-1].inl)
+  {
+    if (localvars[s_lvar.n-1].type==Lmy) nblex--;
     s_lvar.n--;
-  for(i=1;i<lfram;i++)
-  {
-    long j=i+pos->frames-1;
-    frpc[i] = frames[j].pc-pos->opcode+1;
-    gel(fram, i) = gcopy(frames[j].frame);
-    gunclone(frames[j].frame);
   }
-  s_frame.n=pos->frames;
-  offset=pos->offset;
-  dbgstart=pos->dbgstart;
+  for (i = 1; i < lfram; i++)
+  {
+    long j = i+pos->frames-1;
+    frpc[i] = frames[j].pc - pos->opcode+1;
+    gel(fram, i) = gcopyunclone(frames[j].frame);
+  }
+  s_frame.n = pos->frames;
+  offset = pos->offset;
+  dbgstart = pos->dbgstart;
   return cl;
 }
 
 static GEN
-getclosure(struct codepos *pos)
+getclosure(struct codepos *pos, long nbmvar)
 {
-  return getfunction(pos,0,0,NULL,0);
+  return getfunction(pos, 0, nbmvar, NULL, 0);
 }
 
 static void
@@ -301,6 +356,7 @@ var_push(entree *ep, Ltype type)
   localvars[n].ep   = ep;
   localvars[n].inl  = 0;
   localvars[n].type = type;
+  if (type == Lmy) nblex++;
 }
 
 static void
@@ -315,7 +371,7 @@ static GEN
 pack_localvars(void)
 {
   GEN pack=cgetg(3,t_VEC);
-  long i,l=s_lvar.n;
+  long i, l=s_lvar.n;
   GEN t=cgetg(1+l,t_VECSMALL);
   GEN e=cgetg(1+l,t_VECSMALL);
   gel(pack,1)=t;
@@ -325,6 +381,8 @@ pack_localvars(void)
     t[i]=localvars[i-1].type;
     e[i]=(long)localvars[i-1].ep;
   }
+  for(i=1;i<=nblex;i++)
+    access_push(-i);
   return pack;
 }
 
@@ -376,8 +434,9 @@ debug_context(void)
 GEN
 localvars_read_str(const char *x, GEN pack)
 {
+  pari_sp av = avma;
   GEN code;
-  long l=0;
+  long l=0, nbmvar=nblex;
   if (pack)
   {
     GEN t=gel(pack,1);
@@ -389,7 +448,8 @@ localvars_read_str(const char *x, GEN pack)
   }
   code = compile_str(x);
   s_lvar.n -= l;
-  return closure_evalres(code);
+  nblex = nbmvar;
+  return gerepileupto(av, closure_evalres(code));
 }
 
 long
@@ -447,8 +507,6 @@ parseproto(char const **q, char *c, const char *str)
   case 'D':
     switch(p[1])
     {
-    case 0:
-      compile_err("function has incomplete prototype",str);
     case 'G':
     case '&':
     case 'W':
@@ -460,16 +518,11 @@ parseproto(char const **q, char *c, const char *str)
     case 'P':
     case 'r':
     case 's':
-      *c=p[1];
-      *q=p+2;
-      return PPdefault;
+      *c=p[1]; *q=p+2; return PPdefault;
     default:
       for(i=0;*p && i<2;p++) i+=*p==',';
-      if (i<2)
-        compile_err("function has incomplete prototype",str);
-      *c=p[-2];
-      *q=p;
-      return PPdefaultmulti;
+      /* assert(i>=2) because check_proto validated the protototype */
+      *c=p[-2]; *q=p; return PPdefaultmulti;
     }
     break;
   case 'C':
@@ -477,38 +530,25 @@ parseproto(char const **q, char *c, const char *str)
   case 'b':
   case 'P':
   case 'f':
-    *c=*p;
-    *q=p+1;
-    return PPauto;
+    *c=*p; *q=p+1; return PPauto;
   case '&':
-    *c='*';
-    *q=p+1;
-    return PPstd;
+    *c='*'; *q=p+1; return PPstd;
   case 'V':
     if (p[1]=='=')
     {
       if (p[2]!='G')
         compile_err("function prototype is not supported",str);
-      *c='=';
-      p+=2;
+      *c='='; p+=2;
     }
     else
       *c=*p;
-    *q=p+1;
-    return PPstd;
+    *q=p+1; return PPstd;
   case 'E':
   case 's':
-    if (p[1]=='*')
-    {
-      *c=*p++;
-      *q=p+1;
-      return PPstar;
-    }
+    if (p[1]=='*') { *c=*p++; *q=p+1; return PPstar; }
     /*fall through*/
   }
-  *c=*p;
-  *q=p+1;
-  return PPstd;
+  *c=*p; *q=p+1; return PPstd;
 }
 
 static long
@@ -582,7 +622,7 @@ fetch_member_raw(const char *s, long len)
   entree *ep;
   t[0] = '_'; strncpy(t+1, s, len); t[++len] = 0; /* prepend '_' */
   ep = fetch_entry_raw(t, len);
-  avma = av; return ep;
+  set_avma(av); return ep;
 }
 static entree *
 getfunc(long n)
@@ -607,9 +647,13 @@ getentry(long n)
   return getfunc(n);
 }
 
-/* match Fentry that are not actually EpSTATIC functions called without parens*/
 static entree *
 getvar(long n)
+{ return getentry(n); }
+
+/* match Fentry that are not actually EpSTATIC functions called without parens*/
+static entree *
+getvardyn(long n)
 {
   entree *ep = getentry(n);
   if (EpSTATIC(do_alias(ep)))
@@ -632,22 +676,19 @@ getmvar(entree *ep)
   return 0;
 }
 
-static long
-ctxmvar(void)
+static void
+ctxmvar(long n)
 {
   pari_sp av=avma;
-  long i, n=0;
   GEN ctx;
-  for(i=s_lvar.n-1;i>=0;i--)
-    if(localvars[i].type==Lmy)
-      n++;
-  if (n==0) return 0;
+  long i;
+  if (n==0) return;
   ctx = cgetg(n+1,t_VECSMALL);
   for(n=0, i=0; i<s_lvar.n; i++)
     if(localvars[i].type==Lmy)
       ctx[++n]=(long)localvars[i].ep;
   frame_push(ctx);
-  avma=av; return n;
+  set_avma(av);
 }
 
 INLINE int
@@ -672,6 +713,17 @@ str_defproto(const char *p, const char *q, const char *loc)
   op_push_loc(OCpushgen,data_push(strntoGENexp(q+1,len)),loc);
 }
 
+static long
+countmatrixelts(long n)
+{
+  long x,i;
+  if (n==-1 || tree[n].f==Fnoarg) return 0;
+  for(x=n, i=0; tree[x].f==Fmatrixelts; x=tree[x].x)
+    if (tree[tree[x].y].f!=Fnoarg) i++;
+  if (tree[x].f!=Fnoarg) i++;
+  return i;
+}
+
 static long
 countlisttogen(long n, Ffunc f)
 {
@@ -748,14 +800,21 @@ compilestore(long vn, entree *ep, long n)
   if (vn)
     op_push(OCstorelex,vn,n);
   else
+  {
+    if (EpSTATIC(do_alias(ep)))
+      compile_varerr(tree[n].str);
     op_push(OCstoredyn,(long)ep,n);
+  }
 }
 
 INLINE void
 compilenewptr(long vn, entree *ep, long n)
 {
   if (vn)
+  {
+    access_push(vn);
     op_push(OCnewptrlex,vn,n);
+  }
   else
     op_push(OCnewptrdyn,(long)ep,n);
 }
@@ -881,10 +940,12 @@ compilevec(long n, long mode, op_code op)
   op_push(op,l,n);
   for (i=1;i<l;i++)
   {
+    if (tree[arg[i]].f==Fnoarg)
+      compile_err("missing vector element",tree[arg[i]].str);
     compilenode(arg[i],Ggen,FLsurvive);
     op_push(OCstackgen,i,n);
   }
-  avma=ltop;
+  set_avma(ltop);
   op_push(OCpop,1,n);
   compilecast(n,Gvec,mode);
 }
@@ -915,16 +976,17 @@ compilemat(long n, long mode)
     for(j=1;j<lgcol;j++)
     {
       k-=lglin;
+      if (tree[col[j]].f==Fnoarg)
+        compile_err("missing matrix element",tree[col[j]].str);
       compilenode(col[j], Ggen, FLsurvive);
       op_push(OCstackgen,k,n);
     }
   }
-  avma=ltop;
+  set_avma(ltop);
   op_push(OCpop,1,n);
   compilecast(n,Gvec,mode);
 }
 
-
 static GEN
 cattovec(long n, long fnum)
 {
@@ -957,15 +1019,13 @@ cattovec(long n, long fnum)
 }
 
 static GEN
-compilelambda(long n, long y, GEN vep, struct codepos *pos)
+compilelambda(long y, GEN vep, long nbmvar, struct codepos *pos)
 {
-  long nbmvar, lev = vep ? lg(vep)-1 : 0;
+  long lev = vep ? lg(vep)-1 : 0;
   GEN text=cgetg(3,t_VEC);
   gel(text,1)=strtoGENstr(lev? ((entree*) vep[1])->name: "");
   gel(text,2)=strntoGENstr(tree[y].str,tree[y].len);
   dbgstart = tree[y].str;
-  nbmvar=ctxmvar()-lev;
-  if (lev) op_push(OCgetargs,lev,n);
   compilenode(y,Ggen,FLsurvive|FLreturn);
   return getfunction(pos,lev,nbmvar,text,2);
 }
@@ -975,7 +1035,7 @@ compilecall(long n, int mode, entree *ep)
 {
   pari_sp ltop=avma;
   long j;
-  long x=tree[n].x;
+  long x=tree[n].x, tx = tree[x].x;
   long y=tree[n].y;
   GEN arg=listtogen(y,Flistarg);
   long nb=lg(arg)-1;
@@ -988,7 +1048,10 @@ compilecall(long n, int mode, entree *ep)
   {
     long vn=getmvar(ep);
     if (vn)
+    {
+      access_push(vn);
       op_push(OCpushlex,vn,n);
+    }
     else
       op_push(OCpushdyn,(long)ep,n);
   }
@@ -997,6 +1060,15 @@ compilecall(long n, int mode, entree *ep)
     long x = tree[arg[j]].x, f = tree[arg[j]].f;
     if (f==Fseq)
       compile_err("unexpected ';'", tree[x].str+tree[x].len);
+    else if (f==Findarg)
+    {
+      compilenode(tree[arg[j]].x, Ggen,FLnocopy);
+      op_push(OClock,0,n);
+    } else if (tx==CSTmember)
+    {
+      compilenode(arg[j], Ggen,FLnocopy);
+      op_push(OClock,0,n);
+    }
     else if (f!=Fnoarg)
       compilenode(arg[j], Ggen,j>=lnl?FLnocopylex:0);
     else
@@ -1004,7 +1076,7 @@ compilecall(long n, int mode, entree *ep)
   }
   op_push(OCcalluser,nb,x);
   compilecast(n,Ggen,mode);
-  avma=ltop;
+  set_avma(ltop);
 }
 
 static GEN
@@ -1013,9 +1085,11 @@ compilefuncinline(long n, long c, long a, long flag, long isif, long lev, long *
   struct codepos pos;
   int type=c=='I'?Gvoid:Ggen;
   long rflag=c=='I'?0:FLsurvive;
+  long nbmvar = nblex;
   GEN vep = NULL;
   if (isif && (flag&FLreturn)) rflag|=FLreturn;
   getcodepos(&pos);
+  if (c=='J') ctxmvar(nbmvar);
   if (lev)
   {
     long i;
@@ -1032,15 +1106,18 @@ compilefuncinline(long n, long c, long a, long flag, long isif, long lev, long *
       var_push(ve,Lmy);
     }
     checkdups(varg,vep);
+    if (c=='J')
+      op_push(OCgetargs,lev,n);
+    access_push(lg(vep)-1);
     frame_push(vep);
   }
   if (c=='J')
-    return compilelambda(n,a,vep,&pos);
-  else if (tree[a].f==Fnoarg)
+    return compilelambda(a,vep,nbmvar,&pos);
+  if (tree[a].f==Fnoarg)
     compilecast(a,Gvoid,type);
   else
     compilenode(a,type,rflag);
-  return getclosure(&pos);
+  return getclosure(&pos, nbmvar);
 }
 
 static long
@@ -1055,7 +1132,7 @@ countvar(GEN arg)
     {
       long x = detag(tree[a].x);
       if (tree[x].f==Fvec && tree[x].x>=0)
-        n += countlisttogen(tree[x].x,Fmatrixelts)-1;
+        n += countmatrixelts(tree[x].x)-1;
     }
   }
   return n;
@@ -1070,7 +1147,7 @@ compileuninline(GEN arg)
   for(j=0; j<s_lvar.n; j++)
     if(!localvars[j].inl)
       pari_err(e_MISC,"uninline is only valid at top level");
-  s_lvar.n = 0;
+  s_lvar.n = 0; nblex = 0;
 }
 
 static void
@@ -1097,10 +1174,11 @@ compilemy(GEN arg, const char *str, int inl)
         GEN vars = listtogen(tree[x].x,Fmatrixelts);
         long nv = lg(vars)-1;
         for (j=1; j<=nv; j++)
-        {
-          ver[++k] = vars[j];
-          vep[k] = (long)getvar(ver[k]);
-        }
+          if (tree[vars[j]].f!=Fnoarg)
+          {
+            ver[++k] = vars[j];
+            vep[k] = (long)getvar(ver[k]);
+          }
         continue;
       } else ver[++k] = x;
     } else ver[++k] = a;
@@ -1109,6 +1187,7 @@ compilemy(GEN arg, const char *str, int inl)
   checkdups(ver,vep);
   for(i=1; i<=n; i++) var_push(NULL,Lmy);
   op_push_loc(OCnewframe,inl?-n:n,str);
+  access_push(lg(vep)-1);
   frame_push(vep);
   for (k=0, i=1; i<l; i++)
   {
@@ -1119,19 +1198,22 @@ compilemy(GEN arg, const char *str, int inl)
       if (tree[x].f==Fvec && tree[x].x>=0)
       {
         GEN vars = listtogen(tree[x].x,Fmatrixelts);
-        long nv = lg(vars)-1;
+        long nv = lg(vars)-1, m = nv;
         compilenode(tree[a].y,Ggen,FLnocopy);
-        if (nv > 1) op_push(OCdup,nv-1,x);
         for (j=1; j<=nv; j++)
-        {
-          long v = detag(vars[j]);
-          op_push(OCpushlong,j,v);
-          op_push(OCcompo1,Ggen,v);
-          k++;
-          op_push(OCstorelex,-n+k-1,a);
-          localvars[s_lvar.n-n+k-1].ep=(entree*)vep[k];
-          localvars[s_lvar.n-n+k-1].inl=inl;
-        }
+          if (tree[vars[j]].f==Fnoarg) m--;
+        if (m > 1) op_push(OCdup,m-1,x);
+        for (j=1; j<=nv; j++)
+          if (tree[vars[j]].f!=Fnoarg)
+          {
+            long v = detag(vars[j]);
+            op_push(OCpushlong,j,v);
+            op_push(OCcompo1,Ggen,v);
+            k++;
+            op_push(OCstorelex,-n+k-1,a);
+            localvars[s_lvar.n-n+k-1].ep=(entree*)vep[k];
+            localvars[s_lvar.n-n+k-1].inl=inl;
+          }
         continue;
       }
       else if (!is_node_zero(tree[a].y))
@@ -1149,7 +1231,7 @@ compilemy(GEN arg, const char *str, int inl)
 static long
 localpush(op_code op, long a)
 {
-  entree *ep = getvar(a);
+  entree *ep = getvardyn(a);
   long vep  = (long) ep;
   op_push(op,vep,a);
   var_push(ep,Llocal);
@@ -1172,17 +1254,20 @@ compilelocal(GEN arg)
       if (tree[x].f==Fvec && tree[x].x>=0)
       {
         GEN vars = listtogen(tree[x].x,Fmatrixelts);
-        long nv = lg(vars)-1;
+        long nv = lg(vars)-1, m = nv;
         compilenode(tree[a].y,Ggen,FLnocopy);
-        if (nv > 1) op_push(OCdup,nv-1,x);
         for (j=1; j<=nv; j++)
-        {
-          long v = detag(vars[j]);
-          op_push(OCpushlong,j,v);
-          op_push(OCcompo1,Ggen,v);
-          vep[++k] = localpush(OClocalvar, v);
-          ver[k] = v;
-        }
+          if (tree[vars[j]].f==Fnoarg) m--;
+        if (m > 1) op_push(OCdup,m-1,x);
+        for (j=1; j<=nv; j++)
+          if (tree[vars[j]].f!=Fnoarg)
+          {
+            long v = detag(vars[j]);
+            op_push(OCpushlong,j,v);
+            op_push(OCcompo1,Ggen,v);
+            vep[++k] = localpush(OClocalvar, v);
+            ver[k] = v;
+          }
         continue;
       } else if (!is_node_zero(tree[a].y))
       {
@@ -1200,6 +1285,42 @@ compilelocal(GEN arg)
   checkdups(ver,vep);
 }
 
+static void
+compileexport(GEN arg)
+{
+  long i, l = lg(arg);
+  for (i=1; i<l; i++)
+  {
+    long a=arg[i];
+    if (tree[a].f==Fassign)
+    {
+      long x = detag(tree[a].x);
+      long v = (long) getvardyn(x);
+      compilenode(tree[a].y,Ggen,FLnocopy);
+      op_push(OCexportvar,v,x);
+    } else
+    {
+      long x = detag(a);
+      long v = (long) getvardyn(x);
+      op_push(OCpushdyn,v,x);
+      op_push(OCexportvar,v,x);
+    }
+  }
+}
+
+static void
+compileunexport(GEN arg)
+{
+  long i, l = lg(arg);
+  for (i=1; i<l; i++)
+  {
+    long a = arg[i];
+    long x = detag(a);
+    long v = (long) getvardyn(x);
+    op_push(OCunexportvar,v,x);
+  }
+}
+
 static void
 compilefunc(entree *ep, long n, int mode, long flag)
 {
@@ -1247,35 +1368,49 @@ compilefunc(entree *ep, long n, int mode, long flag)
   {
     if (nb==0) op_push(OCpushgnil,0,n);
     else compilenode(arg[1],Ggen,FLsurvive|FLreturn);
-    avma=ltop;
+    set_avma(ltop);
     return;
   }
   else if (is_func_named(ep,"inline"))
   {
     compilemy(arg, str, 1);
     compilecast(n,Gvoid,mode);
-    avma=ltop;
+    set_avma(ltop);
     return;
   }
   else if (is_func_named(ep,"uninline"))
   {
     compileuninline(arg);
     compilecast(n,Gvoid,mode);
-    avma=ltop;
+    set_avma(ltop);
     return;
   }
   else if (is_func_named(ep,"my"))
   {
     compilemy(arg, str, 0);
     compilecast(n,Gvoid,mode);
-    avma=ltop;
+    set_avma(ltop);
     return;
   }
   else if (is_func_named(ep,"local"))
   {
     compilelocal(arg);
     compilecast(n,Gvoid,mode);
-    avma=ltop;
+    set_avma(ltop);
+    return;
+  }
+  else if (is_func_named(ep,"export"))
+  {
+    compileexport(arg);
+    compilecast(n,Gvoid,mode);
+    set_avma(ltop);
+    return;
+  }
+  else if (is_func_named(ep,"unexport"))
+  {
+    compileunexport(arg);
+    compilecast(n,Gvoid,mode);
+    set_avma(ltop);
     return;
   }
   /*We generate dummy code for global() for compatibility with gp2c*/
@@ -1290,18 +1425,18 @@ compilefunc(entree *ep, long n, int mode, long flag)
       {
         compilenode(tree[a].y,Ggen,0);
         a=tree[a].x;
-        en=(long)getvar(a);
+        en=(long)getvardyn(a);
         op_push(OCstoredyn,en,a);
       }
       else
       {
-        en=(long)getvar(a);
+        en=(long)getvardyn(a);
         op_push(OCpushdyn,en,a);
         op_push(OCpop,1,a);
       }
     }
     compilecast(n,Gvoid,mode);
-    avma=ltop;
+    set_avma(ltop);
     return;
   }
   else if (is_func_named(ep,"O"))
@@ -1319,13 +1454,13 @@ compilefunc(entree *ep, long n, int mode, long flag)
   }
   else if (x==OPn && tree[y].f==Fsmall)
   {
-    avma=ltop;
+    set_avma(ltop);
     compilesmall(y, -tree[y].x, mode);
     return;
   }
   else if (x==OPtrans && tree[y].f==Fvec)
   {
-    avma=ltop;
+    set_avma(ltop);
     compilevec(y, mode, OCcol);
     return;
   }
@@ -1368,12 +1503,17 @@ compilefunc(entree *ep, long n, int mode, long flag)
           break;
         case 'W':
           {
-            long a = arg[j];
+            long a = tree[arg[j]].f==Findarg ? tree[arg[j]].x: arg[j];
             entree *ep = getlvalue(a);
             long vn = getmvar(ep);
-            if (vn) op_push(OCcowvarlex, vn, a);
+            if (vn)
+            {
+              access_push(vn);
+              op_push(OCcowvarlex, vn, a);
+            }
             else op_push(OCcowvardyn, (long)ep, a);
-            compilenode(arg[j++],Ggen,FLnocopy);
+            compilenode(a, Ggen,FLnocopy);
+            j++;
             break;
           }
         case 'M':
@@ -1423,7 +1563,10 @@ compilefunc(entree *ep, long n, int mode, long flag)
             if (tree[a].f==Fentry)
             {
               if (vn)
+              {
+                access_push(vn);
                 op_push(OCsimpleptrlex, vn,n);
+              }
               else
                 op_push(OCsimpleptrdyn, (long)ep,n);
             }
@@ -1636,7 +1779,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
   }
   compilecast(n,ret_typ,mode);
   if (nbpointers) op_push_loc(OCendptr,nbpointers, str);
-  avma=ltop;
+  set_avma(ltop);
 }
 
 static void
@@ -1653,9 +1796,8 @@ genclosurectx(const char *loc, long nbdata)
 }
 
 static GEN
-genclosure(entree *ep, const char *loc, long  nbdata, int check)
+genclosure(entree *ep, const char *loc, long nbdata, int check)
 {
-  pari_sp av = avma;
   struct codepos pos;
   long nb=0;
   const char *code=ep->code,*p,*q;
@@ -1781,8 +1923,10 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
       case 'E':
       case 'I':
       case 'V':
+        break;
       case 'r':
       case 's':
+        op_push_loc(OCtostr,-index,loc);
         break;
       case 'n':
         op_push_loc(OCvarn,-index,loc);
@@ -1844,17 +1988,15 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
   if (ret_flag==FLnocopy) op_push_loc(OCcopy,0,loc);
   compilecast_loc(ret_typ, Ggen, loc);
   if (dovararg) nb|=VARARGBITS;
-  return gerepilecopy(av, getfunction(&pos,nb+arity,nbdata,text,0));
+  return getfunction(&pos,nb+arity,nbdata,text,0);
 }
 
 GEN
 snm_closure(entree *ep, GEN data)
 {
-  long i;
-  long n = data ? lg(data)-1: 0;
+  long i, n = data ? lg(data)-1: 0;
   GEN C = genclosure(ep,ep->name,n,0);
-  for(i=1; i<=n; i++)
-    gmael(C,7,i) = gel(data,i);
+  for(i = 1; i <= n; i++) gmael(C,7,i) = gel(data,i);
   return C;
 }
 
@@ -1875,18 +2017,14 @@ strtoclosure(const char *s, long n,  ...)
     va_list ap;
     long i;
     va_start(ap,n);
-    for(i=1; i<=n; i++)
-      gmael(C,7,i) = va_arg(ap, GEN);
+    for(i = 1; i <= n; i++) gmael(C,7,i) = va_arg(ap, GEN);
     va_end(ap);
   }
   return gerepilecopy(av, C);
 }
 
 GEN
-strtofunction(const char *s)
-{
-  return strtoclosure(s, 0);
-}
+strtofunction(const char *s) { return strtoclosure(s, 0); }
 
 GEN
 call0(GEN fun, GEN args)
@@ -1919,7 +2057,7 @@ closurefunc(entree *ep, long n, long mode)
   }
   op_push(OCpushgen, data_push(C), n);
   compilecast(n,Gclosure,mode);
-  avma=ltop;
+  set_avma(ltop);
 }
 
 static void
@@ -1931,7 +2069,7 @@ compileseq(long n, int mode, long flag)
   for(i = 1; i < l; i++)
     compilenode(L[i],Gvoid,0);
   compilenode(L[l],mode,flag&(FLreturn|FLsurvive));
-  avma = av;
+  set_avma(av);
 }
 
 static void
@@ -1963,23 +2101,26 @@ compilenode(long n, int mode, long flag)
       GEN vars = listtogen(tree[x].x,Fmatrixelts);
       long i, l = lg(vars)-1, d = mode==Gvoid? l-1: l;
       compilenode(y,Ggen,mode==Gvoid?0:flag&FLsurvive);
+      for (i=1; i<=l; i++)
+        if (tree[vars[i]].f==Fnoarg) d--;
       if (d) op_push(OCdup, d, x);
       for(i=1; i<=l; i++)
-      {
-        long a = detag(vars[i]);
-        entree *ep=getlvalue(a);
-        long vn=getmvar(ep);
-        op_push(OCpushlong,i,a);
-        op_push(OCcompo1,Ggen,a);
-        if (tree[a].f==Fentry)
-          compilestore(vn,ep,n);
-        else
+        if (tree[vars[i]].f!=Fnoarg)
         {
-          compilenewptr(vn,ep,n);
-          compilelvalue(a);
-          op_push(OCstoreptr,0,a);
+          long a = detag(vars[i]);
+          entree *ep=getlvalue(a);
+          long vn=getmvar(ep);
+          op_push(OCpushlong,i,a);
+          op_push(OCcompo1,Ggen,a);
+          if (tree[a].f==Fentry)
+            compilestore(vn,ep,n);
+          else
+          {
+            compilenewptr(vn,ep,n);
+            compilelvalue(a);
+            op_push(OCstoreptr,0,a);
+          }
         }
-      }
       if (mode!=Gvoid)
         compilecast(n,Ggen,mode);
     }
@@ -2036,7 +2177,7 @@ compilenode(long n, int mode, long flag)
       default:
         pari_err_BUG("compilenode, unsupported constant");
       }
-      avma=ltop;
+      set_avma(ltop);
       return;
     }
   case Fsmall:
@@ -2051,12 +2192,16 @@ compilenode(long n, int mode, long flag)
   case Frefarg:
     compile_err("unexpected character '&':",tree[n].str);
     return;
+  case Findarg:
+    compile_err("unexpected character '~':",tree[n].str);
+    return;
   case Fentry:
     {
       entree *ep=getentry(n);
       long vn=getmvar(ep);
       if (vn)
       {
+        access_push(vn);
         op_push(OCpushlex,(long)vn,n);
         addcopy(n,mode,flag,FLnocopy|FLnocopylex);
         compilecast(n,Ggen,mode);
@@ -2076,7 +2221,7 @@ compilenode(long n, int mode, long flag)
   case Ffunction:
     {
       entree *ep=getfunc(n);
-      if (EpVALENCE(ep)==EpVAR || EpVALENCE(ep)==EpNEW)
+      if (getmvar(ep) || EpVALENCE(ep)==EpVAR || EpVALENCE(ep)==EpNEW)
       {
         if (tree[n].x<OPnboperator) /* should not happen */
           compile_err("operator unknown",tree[n].str);
@@ -2103,43 +2248,47 @@ compilenode(long n, int mode, long flag)
       getcodepos(&pos);
       dbgstart=tree[x].str+tree[x].len;
       gap = tree[y].str-dbgstart;
-      nbmvar=ctxmvar();
+      nbmvar = nblex;
+      ctxmvar(nbmvar);
       nb = lgarg-1;
       if (nb)
       {
         long i;
         for(i=1;i<=nb;i++)
         {
-          long a=arg[i];
-          if (i==nb && tree[a].f==Fvararg)
+          long a = arg[i], f = tree[a].f;
+          if (i==nb && f==Fvararg)
           {
             dovararg=1;
             vep[i]=(long)getvar(tree[a].x);
           }
           else
-            vep[i]=(long)getvar(tree[a].f==Fassign?tree[a].x:a);
+            vep[i]=(long)getvar(f==Fassign||f==Findarg?tree[a].x:a);
           var_push(NULL,Lmy);
         }
         checkdups(arg,vep);
         op_push(OCgetargs,nb,x);
+        access_push(lg(vep)-1);
         frame_push(vep);
         for (i=1;i<=nb;i++)
         {
-          long a=arg[i];
+          long a = arg[i], f = tree[a].f;
           long y = tree[a].y;
-          if (tree[a].f==Fassign && (strict || !is_node_zero(y)))
+          if (f==Fassign && (strict || !is_node_zero(y)))
           {
             if (tree[y].f==Fsmall)
               compilenode(y, Ggen, 0);
             else
             {
               struct codepos lpos;
+              long nbmvar = nblex;
               getcodepos(&lpos);
               compilenode(y, Ggen, 0);
-              op_push(OCpushgen, data_push(getclosure(&lpos)),a);
+              op_push(OCpushgen, data_push(getclosure(&lpos,nbmvar)),a);
             }
             op_push(OCdefaultarg,-nb+i-1,a);
-          }
+          } else if (f==Findarg)
+            op_push(OCsetref, -nb+i-1, a);
           localvars[s_lvar.n-nb+i-1].ep=(entree*)vep[i];
         }
       }
@@ -2154,7 +2303,7 @@ compilenode(long n, int mode, long flag)
       op_push(OCpushgen, data_push(getfunction(&pos,nb,nbmvar,text,gap)),n);
       if (nbmvar) op_push(OCsaveframe,!!(flag&FLsurvive),n);
       compilecast(n, Gclosure, mode);
-      avma=ltop;
+      set_avma(ltop);
       return;
     }
   case Ftag:
@@ -2183,44 +2332,41 @@ gp_closure(long n)
 }
 
 GEN
-closure_deriv(GEN G)
+closure_derivn(GEN G, long n)
 {
-  pari_sp ltop=avma;
-  long i;
+  pari_sp ltop = avma;
   struct codepos pos;
+  long arity = closure_arity(G);
   const char *code;
-  GEN text;
-  long arity=closure_arity(G);
-  if (arity==0 || closure_is_variadic(G))
-    pari_err_TYPE("derivfun",G);
-  if (typ(gel(G,6))==t_STR)
+  GEN t, text;
+
+  if (arity == 0 || closure_is_variadic(G)) pari_err_TYPE("derivfun",G);
+  t = closure_get_text(G);
+  code = GSTR((typ(t) == t_STR)? t: GENtoGENstr(G));
+  if (n > 1)
   {
-    code = GSTR(gel(G,6));
-    text = cgetg(1+nchar2nlong(2+strlen(code)),t_STR);
-    sprintf(GSTR(text),"%s'",code);
+    text = cgetg(1+nchar2nlong(9+strlen(code)+n),t_STR);
+    sprintf(GSTR(text), "derivn(%s,%ld)", code, n);
   }
   else
   {
-    code = GSTR(GENtoGENstr(G));
     text = cgetg(1+nchar2nlong(4+strlen(code)),t_STR);
-    sprintf(GSTR(text),"(%s)'",code);
+    sprintf(GSTR(text), (typ(t) == t_STR)? "%s'": "(%s)'",code);
   }
   getcodepos(&pos);
-  dbgstart=code;
-  op_push_loc(OCgetargs, arity,code);
-  op_push_loc(OCpushgen,data_push(G),code);
-  op_push_loc(OCvec,arity+1,code);
-  for (i=1;i<=arity;i++)
-  {
-    op_push_loc(OCpushlex,i-arity-1,code);
-    op_push_loc(OCstackgen,i,code);
-  }
-  op_push_loc(OCpop,1,code);
-  op_push_loc(OCprecreal,0,code);
-  op_push_loc(OCcallgen,(long)is_entry("_derivfun"),code);
-  return gerepilecopy(ltop, getfunction(&pos,arity,0,text,0));
+  dbgstart = code;
+  op_push_loc(OCpackargs, arity, code);
+  op_push_loc(OCpushgen, data_push(G), code);
+  op_push_loc(OCpushlong, n, code);
+  op_push_loc(OCprecreal, 0, code);
+  op_push_loc(OCcallgen, (long)is_entry("_derivfun"), code);
+  return gerepilecopy(ltop, getfunction(&pos, arity, 0, text, 0));
 }
 
+GEN
+closure_deriv(GEN G)
+{ return closure_derivn(G, 1); }
+
 static long
 vec_optimize(GEN arg)
 {
@@ -2241,7 +2387,7 @@ optimizevec(long n)
   long x = tree[n].x;
   GEN  arg = listtogen(x, Fmatrixelts);
   tree[n].flags = vec_optimize(arg);
-  avma = ltop;
+  set_avma(ltop);
 }
 
 static void
@@ -2257,7 +2403,7 @@ optimizemat(long n)
     GEN col=listtogen(line[i],Fmatrixelts);
     fl &= vec_optimize(col);
   }
-  avma=ltop; tree[n].flags=fl;
+  set_avma(ltop); tree[n].flags=fl;
 }
 
 static void
@@ -2279,7 +2425,6 @@ optimizematcoeff(long n)
   tree[n].flags=fl;
 }
 
-
 static void
 optimizefunc(entree *ep, long n)
 {
@@ -2348,9 +2493,12 @@ optimizefunc(entree *ep, long n)
             break;
           }
         case 'W':
-          optimizenode(arg[j++]);
-          fl=0;
+        {
+          long a = tree[arg[j]].f==Findarg ? tree[arg[j]].x: arg[j];
+          optimizenode(a);
+          fl=0; j++;
           break;
+        }
         case 'V':
         case 'r':
           tree[arg[j++]].flags=COsafelex|COsafedyn;
@@ -2416,7 +2564,7 @@ optimizefunc(entree *ep, long n)
       compile_err("too many arguments",tree[arg[j]].str);
   }
   else (void)vec_optimize(arg);
-  avma=av; tree[n].flags=fl;
+  set_avma(av); tree[n].flags=fl;
 }
 
 static void
@@ -2428,7 +2576,7 @@ optimizecall(long n)
   GEN arg=listtogen(y,Flistarg);
   optimizenode(x);
   tree[n].flags = COsafelex&tree[x].flags&vec_optimize(arg);
-  avma=av;
+  set_avma(av);
 }
 
 static void
@@ -2442,7 +2590,7 @@ optimizeseq(long n)
     optimizenode(L[i]);
     flags &= tree[L[i]].flags;
   }
-  avma = av;
+  set_avma(av);
   tree[n].flags = flags;
 }
 
@@ -2493,6 +2641,8 @@ optimizenode(long n)
   case Frefarg:
     compile_err("unexpected character '&'",tree[n].str);
     return;
+  case Findarg:
+    return;
   case Fvararg:
     compile_err("unexpected characters '..'",tree[n].str);
     return;
