@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -17,7 +18,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #  include "../systems/mingw/mingw.h"
 #endif
 
-
 /* Return all chars, up to next separator
  * [as strtok but must handle verbatim character string] */
 char*
@@ -84,11 +84,11 @@ get_int(const char *s, long dflt)
   int minus = 0;
 
   if (*p == '-') { minus = 1; p++; }
-  if (!isdigit((int)*p)) { avma = av; return dflt; }
+  if (!isdigit((int)*p)) return gc_long(av, dflt);
 
   n = (long)my_int(p);
   if (n < 0) pari_err(e_SYNTAX,"integer too large",s,s);
-  avma = av; return minus? -n: n;
+  return gc_long(av, minus? -n: n);
 }
 
 ulong
@@ -96,9 +96,8 @@ get_uint(const char *s)
 {
   pari_sp av = avma;
   char *p = get_sep(s);
-  ulong u;
   if (*p == '-') pari_err(e_SYNTAX,"arguments must be positive integers",s,s);
-  u = my_int(p); avma = av; return u;
+  return gc_ulong(av, my_int(p));
 }
 
 #if defined(__EMX__) || defined(_WIN32) || defined(__CYGWIN32__)
@@ -281,11 +280,12 @@ err_intarray(char *t, char *p, const char *s)
 static GEN
 parse_intarray(const char *v, const char *s)
 {
+  pari_sp av = avma;
   char *p, *t = gp_filter(v);
   long i, l;
   GEN w;
   if (*t != '[') err_intarray(t, t, s);
-  if (t[1] == ']') { pari_free(t); return cgetalloc(t_VECSMALL, 1); }
+  if (t[1] == ']') return gc_const(av, cgetalloc(t_VECSMALL, 1));
   for (p = t+1, l=2; *p; p++)
     if (*p == ',') l++;
     else if (*p < '0' || *p > '9') break;
@@ -297,7 +297,7 @@ parse_intarray(const char *v, const char *s)
     while (*p >= '0' && *p <= '9') n = 10*n + (*p++ -'0');
     w[++i] = n;
   }
-  pari_free(t); return w;
+  return gc_const(av, w);
 }
 GEN
 sd_intarray(const char *v, long flag, GEN *pz, const char *s)
@@ -413,7 +413,8 @@ sd_colors(const char *v, long flag)
   long c,l;
   if (v && !(GP_DATA->flags & (gpd_EMACS|gpd_TEXMACS)))
   {
-    char *v0, *s;
+    pari_sp av = avma;
+    char *s;
     disable_color=1;
     l = strlen(v);
     if (l <= 2 && strncmp(v, "no", l) == 0)
@@ -426,10 +427,9 @@ sd_colors(const char *v, long flag)
       v = "9, 13, 11, 15, 14, 10, 11";
     if (l <= 6 && strncmp(v, "boldfg", l) == 0)        /* Good for darkbg consoles */
       v = "[1,,1], [5,,1], [3,,1], [7,,1], [6,,1], , [2,,1]";
-    v0 = s = gp_filter(v);
-    for (c=c_ERR; c < c_LAST; c++)
-      gp_colors[c] = gp_get_color(&s);
-    pari_free(v0);
+    s = gp_filter(v);
+    for (c=c_ERR; c < c_LAST; c++) gp_colors[c] = gp_get_color(&s);
+    set_avma(av);
   }
   if (flag == d_ACKNOWLEDGE || flag == d_RETURN)
   {
@@ -602,25 +602,27 @@ sd_log(const char *v, long flag)
       "(on with colors)",
       "(TeX output)", NULL
   };
-  ulong s = logstyle;
+  ulong s = pari_logstyle;
   GEN res = sd_ulong(v,flag,"log", &s, 0, 3, msg);
 
-  if (!s != !logstyle) /* Compare converts to boolean */
+  if (!s != !pari_logstyle) /* Compare converts to boolean */
   { /* toggled LOG */
-    if (logstyle)
+    if (pari_logstyle)
     { /* close log */
       if (flag == d_ACKNOWLEDGE)
         pari_printf("   [logfile was \"%s\"]\n", current_logfile);
-      if (pari_logfile) /* paranoia */
-      {
-        fclose(pari_logfile);
-        pari_logfile = NULL;
-      }
+      if (pari_logfile) { fclose(pari_logfile); pari_logfile = NULL; }
     }
     else
+    {
       pari_logfile = open_logfile(current_logfile);
+      if (flag == d_ACKNOWLEDGE)
+        pari_printf("   [logfile is \"%s\"]\n", current_logfile);
+      else if (flag == d_INITRC)
+        pari_printf("Logging to %s\n", current_logfile);
+    }
   }
-  if (pari_logfile && s != logstyle && s == logstyle_TeX)
+  if (pari_logfile && s != pari_logstyle && s == logstyle_TeX)
   {
     TeX_define("PARIbreak",
                "\\hskip 0pt plus \\hsize\\relax\\discretionary{}{}{}");
@@ -631,7 +633,7 @@ sd_log(const char *v, long flag)
                 "\\vskip\\smallskipamount$\\displaystyle{\\tt\\%#1} = #2$");
   }
   /* don't record new value until we are sure everything is fine */
-  logstyle = s; return res;
+  pari_logstyle = s; return res;
 }
 
 GEN
@@ -781,6 +783,8 @@ sd_datadir(const char *v, long flag)
   const char *str;
   if (v)
   {
+    mt_broadcast(snm_closure(is_entry("default"),
+                 mkvec2(strtoGENstr("datadir"), strtoGENstr(v))));
     if (pari_datadir) pari_free(pari_datadir);
     pari_datadir = path_expand(v);
   }
@@ -796,6 +800,8 @@ sd_PATH(const char *v, long flag, const char* s, gp_path *p)
 {
   if (v)
   {
+    mt_broadcast(snm_closure(is_entry("default"),
+                 mkvec2(strtoGENstr(s), strtoGENstr(v))));
     pari_free((void*)p->PATH);
     p->PATH = pari_strdup(v);
     if (flag == d_INITRC) return gnil;
@@ -976,7 +982,7 @@ default_gp_data(void)
   static gp_hist __HIST;
   static gp_pp   __PP;
   static gp_path __PATH, __SOPATH;
-  static pari_timer __T;
+  static pari_timer __T, __Tw;
 
   D->flags       = 0;
   D->primelimit  = 500000;
@@ -995,6 +1001,7 @@ default_gp_data(void)
   D->secure      = 0;
   D->use_readline= 0;
   D->T    = &__T;
+  D->Tw   = &__Tw;
   D->hist = &__HIST;
   D->pp   = &__PP;
   D->path = &__PATH;
