@@ -4,7 +4,8 @@ This file is part of the PARI package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -52,7 +53,7 @@ return0(GEN x)
 {
   GEN y = br_res;
   br_res = (x && x != gnil)? gcloneref(x): NULL;
-  if (y) gunclone_deep(y);
+  guncloneNULL_deep(y);
   br_status = br_RETURN; return NULL;
 }
 
@@ -87,7 +88,7 @@ break0(long n)
  * functions for use in sumiter: we want a temporary ep->value, which is NOT
  * a clone (PUSH), to avoid unnecessary copies. */
 
-enum {PUSH_VAL = 0, COPY_VAL = 1, DEFAULT_VAL = 2};
+enum {PUSH_VAL = 0, COPY_VAL = 1, DEFAULT_VAL = 2, REF_VAL = 3};
 
 /* ep->args is the stack of old values (INITIAL if initial value, from
  * installep) */
@@ -157,7 +158,7 @@ pushvalue(entree *ep, GEN x) {
 }
 
 INLINE void
-zerovalue (entree *ep)
+zerovalue(entree *ep)
 {
   var_cell *v = (var_cell*) pari_malloc(sizeof(var_cell));
   v->value  = (GEN)ep->value;
@@ -169,7 +170,6 @@ zerovalue (entree *ep)
   ep->valence=EpVAR;
 }
 
-
 /* as above IF ep->value was PUSHed, or was created after block number 'loc'
    return 0 if not deleted, 1 otherwise [for recover()] */
 int
@@ -220,8 +220,8 @@ enum chk_VALUE { chk_ERROR, chk_NOCREATE, chk_CREATE };
 INLINE void
 checkvalue(entree *ep, enum chk_VALUE flag)
 {
-  if (MT_IS_THREAD)
-    pari_err(e_MISC,"mt: global variable not supported: %s",ep->name);
+  if (mt_is_thread())
+    pari_err(e_MISC,"mt: attempt to change exported variable '%s'",ep->name);
   if (ep->valence==EpNEW)
     switch(flag)
     {
@@ -237,7 +237,7 @@ checkvalue(entree *ep, enum chk_VALUE flag)
         break;
     }
   else if (ep->valence!=EpVAR)
-    err_var(strtoGENstr(ep->name));
+    pari_err(e_MISC, "attempt to change built-in %s", ep->name);
 }
 
 INLINE GEN
@@ -247,7 +247,7 @@ checkvalueptr(entree *ep)
   return ep->valence==EpNEW? gen_0: (GEN)ep->value;
 }
 
-/* make GP variables safe for avma = top */
+/* make GP variables safe for set_avma(top) */
 static void
 lvar_make_safe(void)
 {
@@ -326,7 +326,6 @@ typedef struct gp_pointer
   long sp;
 } gp_pointer;
 
-
 /* assign res at *pt in "simple array object" p and return it, or a copy.*/
 static void
 change_compo(matcomp *c, GEN res)
@@ -384,11 +383,13 @@ struct trace
 
 static THREAD long sp, rp, dbg_level;
 static THREAD long *st, *precs;
+static THREAD GEN *locks;
 static THREAD gp_pointer *ptrs;
 static THREAD entree **lvars;
 static THREAD struct var_lex *var;
 static THREAD struct trace *trace;
-static THREAD pari_stack s_st, s_ptrs, s_var, s_lvars, s_trace, s_prec;
+static THREAD pari_stack s_st, s_ptrs, s_var, s_trace, s_prec;
+static THREAD pari_stack s_lvars, s_locks;
 
 static void
 changelex(long vn, GEN x)
@@ -403,7 +404,7 @@ INLINE GEN
 copylex(long vn)
 {
   struct var_lex *v = var+s_var.n+vn;
-  if (v->flag!=COPY_VAL)
+  if (v->flag!=COPY_VAL && v->flag!=REF_VAL)
   {
     v->value = gclone(v->value);
     v->flag  = COPY_VAL;
@@ -411,6 +412,13 @@ copylex(long vn)
   return v->value;
 }
 
+INLINE void
+setreflex(long vn)
+{
+  struct var_lex *v = var+s_var.n+vn;
+  v->flag  = REF_VAL;
+}
+
 INLINE void
 pushlex(long vn, GEN x)
 {
@@ -428,13 +436,15 @@ freelex(void)
 }
 
 INLINE void
-restore_vars(long nbmvar, long nblvar)
+restore_vars(long nbmvar, long nblvar, long nblock)
 {
   long j;
   for(j=1;j<=nbmvar;j++)
     freelex();
   for(j=1;j<=nblvar;j++)
     { s_lvars.n--; pop_val(lvars[s_lvars.n]); }
+  for(j=1;j<=nblock;j++)
+    { s_locks.n--; gunclone(locks[s_locks.n]); }
 }
 
 INLINE void
@@ -444,7 +454,7 @@ restore_trace(long nbtrace)
   for(j=1;j<=nbtrace;j++)
   {
     GEN C = trace[s_trace.n-j].closure;
-    if (isclone(C)) gunclone(C);
+    clone_unlock(C);
   }
   s_trace.n-=nbtrace;
 }
@@ -456,6 +466,7 @@ trace_push(long pc, GEN C)
   BLOCK_SIGINT_START
   tr = pari_stack_new(&s_trace);
   trace[tr].pc = pc;
+  clone_lock(C);
   trace[tr].closure = C;
   BLOCK_SIGINT_END
   return tr;
@@ -511,6 +522,7 @@ pari_init_evaluator(void)
   s_ptrs.n=s_ptrs.alloc;
   pari_stack_init(&s_var,sizeof(*var),(void**)&var);
   pari_stack_init(&s_lvars,sizeof(*lvars),(void**)&lvars);
+  pari_stack_init(&s_locks,sizeof(*locks),(void**)&locks);
   pari_stack_init(&s_trace,sizeof(*trace),(void**)&trace);
   br_res = NULL;
   pari_stack_init(&s_relocs,sizeof(*relocs),(void**)&relocs);
@@ -546,56 +558,97 @@ new_ptr(void)
   return &ptrs[rp++];
 }
 
-void
-push_localprec(long p)
-{
-  long n = pari_stack_new(&s_prec);
-  precs[n] = prec2nbits(p);
-}
-
 void
 push_localbitprec(long p)
 {
   long n = pari_stack_new(&s_prec);
   precs[n] = p;
 }
+void
+push_localprec(long p) { push_localbitprec(prec2nbits(p)); }
 
 void
-pop_localprec(void)
-{
-  s_prec.n--;
-}
+pop_localprec(void) { s_prec.n--; }
 
 long
-get_localbitprec(void)
-{
-  return s_prec.n? precs[s_prec.n-1]: precreal;
-}
+get_localbitprec(void) { return s_prec.n? precs[s_prec.n-1]: precreal; }
 
 long
-get_localprec(void)
-{
-  return nbits2prec(get_localbitprec());
-}
+get_localprec(void) { return nbits2prec(get_localbitprec()); }
 
-void
-localprec(long p)
+static void
+checkprec(const char *f, long p, long M)
 {
-  long pmax = prec2ndec(LGBITS);
-  if (p < 1) pari_err_DOMAIN("localprec", "p", "<", gen_1, stoi(p));
-  if (p > pmax)
-    pari_err_DOMAIN("localprec", "p", ">", utoi(pmax), stoi(p));
-  push_localprec(ndec2prec(p));
+  if (p < 1) pari_err_DOMAIN(f, "p", "<", gen_1, stoi(p));
+  if (p > M) pari_err_DOMAIN(f, "p", ">", utoipos(M), utoi(p));
 }
-
-void
-localbitprec(long p)
+static long
+_prec(GEN p, const char *f)
 {
-  if (p < 1) pari_err_DOMAIN("localprec", "p", "<", gen_1, stoi(p));
-  if (p > (long)LGBITS)
-    pari_err_DOMAIN("localbitprec", "p", ">", utoi(LGBITS), stoi(p));
+  pari_sp av = avma;
+  if (typ(p) == t_INT) return itos(p);
+  p = gceil(p);
+  if (typ(p) != t_INT) pari_err_TYPE(f, p);
+  return gc_long(av, itos(p));
+}
+void
+localprec(GEN pp)
+{
+  long p = _prec(pp, "localprec");
+  checkprec("localprec", p, prec2ndec(LGBITS));
+  p = ndec2nbits(p); push_localbitprec(p);
+}
+void
+localbitprec(GEN pp)
+{
+  long p = _prec(pp, "localbitprec");
+  checkprec("localbitprec", p, (long)LGBITS);
   push_localbitprec(p);
 }
+long
+getlocalprec(long prec) { return prec2ndec(prec); }
+long
+getlocalbitprec(long bit) { return bit; }
+
+static GEN
+_precision0(GEN x)
+{
+  long a = gprecision(x);
+  return a? utoi(prec2ndec(a)): mkoo();
+}
+GEN
+precision0(GEN x, long n)
+{ return n? gprec(x,n): _precision0(x); }
+static GEN
+_bitprecision0(GEN x)
+{
+  long a = gprecision(x);
+  return a? utoi(prec2nbits(a)): mkoo();
+}
+GEN
+bitprecision0(GEN x, long n)
+{
+  if (n < 0)
+    pari_err_DOMAIN("bitprecision", "bitprecision", "<", gen_0, stoi(n));
+  if (n) {
+    pari_sp av = avma;
+    GEN y = gprec_w(x, nbits2prec(n));
+    return gerepilecopy(av, y);
+  }
+  return _bitprecision0(x);
+}
+GEN
+precision00(GEN x, GEN n)
+{
+  if (!n) return _precision0(x);
+  return precision0(x, _prec(n, "precision"));
+}
+GEN
+bitprecision00(GEN x, GEN n)
+{
+  if (!n) return _bitprecision0(x);
+  return bitprecision0(x, _prec(n, "bitprecision"));
+}
 
 INLINE GEN
 copyupto(GEN z, GEN t)
@@ -608,35 +661,28 @@ copyupto(GEN z, GEN t)
 
 static void closure_eval(GEN C);
 
+INLINE GEN
+get_and_reset_break(void)
+{
+  GEN z = br_res? gcopy(br_res): gnil;
+  reset_break(); return z;
+}
+
 INLINE GEN
 closure_return(GEN C)
 {
-  pari_sp ltop=avma;
+  pari_sp av = avma;
   closure_eval(C);
-  if (br_status)
-  {
-    GEN z;
-    avma=ltop;
-    z=br_res?gcopy(br_res):gnil;
-    reset_break();
-    return z;
-  }
-  return gerepileupto(ltop,gel(st,--sp));
+  if (br_status) { set_avma(av); return get_and_reset_break(); }
+  return gerepileupto(av, gel(st,--sp));
 }
 
 /* for the break_loop debugger. Not memory clean */
 GEN
 closure_evalbrk(GEN C, long *status)
 {
-  closure_eval(C);
-  *status = br_status;
-  if (br_status)
-  {
-    GEN z = br_res? gcopy(br_res): gnil;
-    reset_break();
-    return z;
-  }
-  return gel(st,--sp);
+  closure_eval(C); *status = br_status;
+  return br_status? get_and_reset_break(): gel(st,--sp);
 }
 
 INLINE long
@@ -843,14 +889,13 @@ closure_eval(GEN C)
   long loper=lg(oper);
   long saved_sp=sp-closure_arity(C);
   long saved_rp=rp, saved_prec=s_prec.n;
-  long j, nbmvar=0, nblvar=0;
+  long j, nbmvar=0, nblvar=0, nblock=0;
   long pc, t;
 #ifdef STACK_CHECK
   GEN stackelt;
   if (PARI_stack_limit && (void*) &stackelt <= PARI_stack_limit)
     pari_err(e_MISC, "deep recursion");
 #endif
-  clone_lock(C);
   t = trace_push(0, C);
   if (lg(C)==8)
   {
@@ -900,8 +945,17 @@ closure_eval(GEN C)
     case OCpushdyn:
       {
         entree *ep = (entree *)operand;
-        checkvalue(ep, chk_CREATE);
-        gel(st,sp++)=(GEN)ep->value;
+        if (!mt_is_thread())
+        {
+          checkvalue(ep, chk_CREATE);
+          gel(st,sp++)=(GEN)ep->value;
+        } else
+        {
+          GEN val = export_get(ep->name);
+          if (!val)
+            pari_err(e_MISC,"mt: please use export(%s)", ep->name);
+          gel(st,sp++)=val;
+        }
         break;
       }
     case OCpushlex:
@@ -1030,6 +1084,21 @@ closure_eval(GEN C)
     case OCcowvarlex:
       (void)copylex(operand);
       break;
+    case OCsetref:
+      setreflex(operand);
+      break;
+    case OClock:
+    {
+      GEN v = gel(st,sp-1);
+      if (isclone(v))
+      {
+        long n = pari_stack_new(&s_locks);
+        locks[n] = v;
+        nblock++;
+        ++bl_refc(v);
+      }
+      break;
+    }
     case OCstoi:
       gel(st,sp-1)=stoi(st[sp-1]);
       break;
@@ -1045,7 +1114,7 @@ closure_eval(GEN C)
     case OCtostr:
       {
         GEN z = gel(st,sp+operand);
-        st[sp+operand] = (long)GENtostr_unquoted(z);
+        st[sp+operand] = (long) (z ? GENtostr_unquoted(z): NULL);
         break;
       }
     case OCvarn:
@@ -1070,7 +1139,7 @@ closure_eval(GEN C)
             pari_warn(warnmem,"eval: recovering %ld bytes", av - av2);
           x = gerepileupto(av, x);
         }
-      } else avma = av;
+      } else set_avma(av);
       gel(st,sp-1) = x;
       break;
     }
@@ -1243,9 +1312,10 @@ closure_eval(GEN C)
       break;
     case OClocalvar:
       {
-        long n = pari_stack_new(&s_lvars);
+        long n;
         entree *ep = (entree *)operand;
         checkvalue(ep, chk_NOCREATE);
+        n = pari_stack_new(&s_lvars);
         lvars[n] = ep;
         nblvar++;
         pushvalue(ep,gel(st,--sp));
@@ -1253,14 +1323,27 @@ closure_eval(GEN C)
       }
     case OClocalvar0:
       {
-        long n = pari_stack_new(&s_lvars);
+        long n;
         entree *ep = (entree *)operand;
         checkvalue(ep, chk_NOCREATE);
+        n = pari_stack_new(&s_lvars);
         lvars[n] = ep;
         nblvar++;
         zerovalue(ep);
         break;
       }
+    case OCexportvar:
+      {
+        entree *ep = (entree *)operand;
+        mt_export_add(ep->name, gel(st,--sp));
+        break;
+      }
+    case OCunexportvar:
+      {
+        entree *ep = (entree *)operand;
+        mt_export_del(ep->name);
+        break;
+      }
 
 #define EVAL_f(f) \
   switch (ep->arity) \
@@ -1386,17 +1469,43 @@ closure_eval(GEN C)
     case OCsaveframe:
       {
         GEN cl = (operand?gcopy:shallowcopy)(gel(st,sp-1));
-        long l = lg(gel(cl,7));
-        GEN  v = cgetg(l, t_VEC);
-        for(j=1; j<l; j++)
-        {
-          GEN val = var[s_var.n-j].value;
-          gel(v,j) = operand?gcopy(val):val;
-        }
+        GEN f = gel(cl, 7);
+        long j, l = lg(f);
+        GEN v = cgetg(l, t_VEC);
+        for (j = 1; j < l; j++)
+          if (signe(gel(f,l-j))==0)
+          {
+            GEN val = var[s_var.n-j].value;
+            gel(v,j) = operand?gcopy(val):val;
+          } else
+            gel(v,j) = gnil;
         gel(cl,7) = v;
         gel(st,sp-1) = cl;
       }
       break;
+    case OCpackargs:
+    {
+      GEN def = cgetg(operand+1, t_VECSMALL);
+      GEN args = cgetg(operand+1, t_VEC);
+      pari_stack_alloc(&s_var,operand);
+      sp-=operand;
+      for (j=0;j<operand;j++)
+      {
+        if (gel(st,sp+j))
+        {
+          gel(args,j+1) = gel(st,sp+j);
+          uel(def ,j+1) = 1;
+        }
+        else
+        {
+          gel(args,j+1) = gen_0;
+          uel(def ,j+1) = 0;
+        }
+      }
+      gel(st, sp++) = args;
+      gel(st, sp++) = def;
+      break;
+    }
     case OCgetargs:
       pari_stack_alloc(&s_var,operand);
       s_var.n+=operand;
@@ -1492,7 +1601,7 @@ endeval:
   }
   s_prec.n = saved_prec;
   s_trace.n--;
-  restore_vars(nbmvar, nblvar);
+  restore_vars(nbmvar, nblvar, nblock);
   clone_unlock(C);
 }
 
@@ -1501,10 +1610,18 @@ closure_evalgen(GEN C)
 {
   pari_sp ltop=avma;
   closure_eval(C);
-  if (br_status) { avma=ltop; return NULL; }
+  if (br_status) return gc_NULL(ltop);
   return gerepileupto(ltop,gel(st,--sp));
 }
 
+long
+evalstate_get_trace(void)
+{ return s_trace.n; }
+
+void
+evalstate_set_trace(long lvl)
+{ s_trace.n = lvl; }
+
 void
 evalstate_save(struct pari_evalstate *state)
 {
@@ -1514,20 +1631,22 @@ evalstate_save(struct pari_evalstate *state)
   state->prec = s_prec.n;
   state->var  = s_var.n;
   state->lvars= s_lvars.n;
+  state->locks= s_locks.n;
   state->trace= s_trace.n;
   compilestate_save(&state->comp);
-  mtstate_save(&state->pending_threads);
+  mtstate_save(&state->mt);
 }
 
 void
 evalstate_restore(struct pari_evalstate *state)
 {
-  avma = state->avma;
-  mtstate_restore(&state->pending_threads);
+  set_avma(state->avma);
+  mtstate_restore(&state->mt);
   sp = state->sp;
   rp = state->rp;
   s_prec.n = state->prec;
-  restore_vars(s_var.n-state->var,s_lvars.n-state->lvars);
+  restore_vars(s_var.n-state->var, s_lvars.n-state->lvars,
+               s_locks.n-state->locks);
   restore_trace(s_trace.n-state->trace);
   reset_break();
   compilestate_restore(&state->comp);
@@ -1548,12 +1667,12 @@ evalstate_reset(void)
   sp = 0;
   rp = 0;
   dbg_level = 0;
-  restore_vars(s_var.n, s_lvars.n);
+  restore_vars(s_var.n, s_lvars.n, s_locks.n);
   s_trace.n = 0;
   reset_break();
   compilestate_reset();
   parsestate_reset();
-  avma = pari_mainstack->top;
+  set_avma(pari_mainstack->top);
 }
 
 void
@@ -1595,7 +1714,7 @@ closure_evalvoid(GEN C)
 {
   pari_sp ltop=avma;
   closure_eval(C);
-  avma=ltop;
+  set_avma(ltop);
 }
 
 GEN
@@ -1664,45 +1783,58 @@ parvector(long n, GEN code)
   return V;
 }
 
-GEN
-parsum(GEN a, GEN b, GEN code, GEN x)
+/* B <- {a + k * m : k = 0, ..., (b-a)/m)} */
+static void
+arithprogset(GEN B, GEN a, GEN b, long m)
 {
-  pari_sp av = avma, av2;
-  long pending = 0;
-  GEN worker = snm_closure(is_entry("_parvector_worker"), mkvec(code));
-  GEN done;
+  long k;
+  for (k = 1; cmpii(a, b) <= 0; a = addui(m,a), k++) gel(B, k) = a;
+  setlg(B, k);
+}
+static GEN
+vecsum_i(GEN v)
+{
+  long i, l = lg(v);
+  GEN s;
+  if (l == 1) return gen_0;
+  s = gel(v,1); for (i = 2; i < l; i++) s = gadd(s, gel(v,i));
+  return s;
+}
+GEN
+parsum(GEN a, GEN b, GEN code)
+{
+  pari_sp av = avma;
+  GEN worker, L, v, s, N;
+  long r, m, pending;
   struct pari_mt pt;
-  if (typ(a) != t_INT) pari_err_TYPE("parsum",a);
-  if (!x) x = gen_0;
-  if (gcmp(b,a) < 0) return gcopy(x);
+  pari_sp av2;
 
-  mt_queue_start(&pt, worker);
+  if (typ(a) != t_INT) pari_err_TYPE("parsum",a);
+  if (gcmp(b,a) < 0) return gen_0;
+  worker = snm_closure(is_entry("_parapply_slice_worker"), mkvec(code));
   b = gfloor(b);
-  a = mkvec(setloop(a));
-  av2=avma;
-  for (; cmpii(gel(a,1),b) <= 0 || pending; gel(a,1) = incloop(gel(a,1)))
+  N = addiu(subii(b, a), 1);
+  m = itou(sqrti(N)); if (cmpiu(N, m) < 0) m = itou(N);
+  mt_queue_start_lim(&pt, worker, m);
+  L = cgetg(m + 2, t_VEC); v = mkvec(L);
+  s = gen_0; a = setloop(a); pending = 0; av2 = avma;
+  for (r = 1; r <= m || pending; r++)
   {
-    mt_queue_submit(&pt, 0, cmpii(gel(a,1),b) <= 0? a: NULL);
-    done = mt_queue_get(&pt, NULL, &pending);
-    if (done)
-    {
-      x = gadd(x, done);
-      if (gc_needed(av2,1))
-      {
-        if (DEBUGMEM>1) pari_warn(warnmem,"sum");
-        x = gerepileupto(av2,x);
-      }
-    }
+    long workid;
+    GEN done;
+    if (r <= m) { arithprogset(L, icopy(a), b, m); a = incloop(a); }
+    mt_queue_submit(&pt, 0, r <= m? v: NULL);
+    done = mt_queue_get(&pt, &workid, &pending);
+    if (done) s = gerepileupto(av2, gadd(s, vecsum_i(done)));
   }
-  mt_queue_end(&pt);
-  return gerepilecopy(av, x);
+  mt_queue_end(&pt); return gerepileupto(av, s);
 }
 
 void
 parfor(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
 {
   pari_sp av = avma, av2;
-  long running, pending = 0;
+  long running, pending = 0, lim;
   long status = br_NONE;
   GEN worker = snm_closure(is_entry("_parfor_worker"), mkvec(code));
   GEN done, stop = NULL;
@@ -1719,7 +1851,8 @@ parfor(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
     else
       b = gfloor(b);
   }
-  mt_queue_start(&pt, worker);
+  lim = b ? itos_or_0(subii(addis(b,1),a)): 0;
+  mt_queue_start_lim(&pt, worker, lim);
   a = mkvec(setloop(a));
   av2 = avma;
   while ((running = (!stop && (!b || cmpii(gel(a,1),b) <= 0))) || pending)
@@ -1734,14 +1867,66 @@ parfor(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
         stop = gerepileuptoint(av2, gel(done,1));
       }
     gel(a,1) = incloop(gel(a,1));
-    if (!stop) avma = av2;
+    if (!stop) set_avma(av2);
   }
-  avma = av2;
+  set_avma(av2);
   mt_queue_end(&pt);
   br_status = status;
-  avma = av;
+  set_avma(av);
 }
 
+static void
+parforiter_init(struct parfor_iter *T, GEN code)
+{
+  T->pending = 0;
+  T->worker = snm_closure(is_entry("_parfor_worker"), mkvec(code));
+  mt_queue_start(&T->pt, T->worker);
+}
+
+static GEN
+parforiter_next(struct parfor_iter *T, GEN v)
+{
+  mt_queue_submit(&T->pt, 0, v);
+  return mt_queue_get(&T->pt, NULL, &T->pending);
+}
+
+static void
+parforiter_stop(struct parfor_iter *T)
+{
+  while (T->pending)
+  {
+    mt_queue_submit(&T->pt, 0, NULL);
+    (void) mt_queue_get(&T->pt, NULL, &T->pending);
+  }
+  mt_queue_end(&T->pt);
+}
+
+void
+parfor_init(parfor_t *T, GEN a, GEN b, GEN code)
+{
+  if (typ(a) != t_INT) pari_err_TYPE("parfor",a);
+  T->b = b ? gfloor(b): NULL;
+  T->a = mkvec(setloop(a));
+  parforiter_init(&T->iter, code);
+}
+
+GEN
+parfor_next(parfor_t *T)
+{
+  long running;
+  while ((running=((!T->b || cmpii(gel(T->a,1),T->b) <= 0))) || T->iter.pending)
+  {
+    GEN done = parforiter_next(&T->iter, running ? T->a: NULL);
+    gel(T->a,1) = incloop(gel(T->a,1));
+    if (done) return done;
+  }
+  mt_queue_end(&T->iter.pt);
+  return NULL;
+}
+
+void
+parfor_stop(parfor_t *T) { parforiter_stop(&T->iter); }
+
 static long
 gp_evalvoid2(void *E, GEN x, GEN y)
 {
@@ -1760,7 +1945,37 @@ parfor0(GEN a, GEN b, GEN code, GEN code2)
 }
 
 void
-parforprime(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
+parforprimestep_init(parforprime_t *T, GEN a, GEN b, GEN q, GEN code)
+{
+  forprimestep_init(&T->forprime, a, b, q);
+  T->v = mkvec(gen_0);
+  parforiter_init(&T->iter, code);
+}
+
+void
+parforprime_init(parforprime_t *T, GEN a, GEN b, GEN code)
+{ parforprimestep_init(T, a, b, NULL, code); }
+
+GEN
+parforprime_next(parforprime_t *T)
+{
+  long running;
+  while ((running = !!forprime_next(&T->forprime)) || T->iter.pending)
+  {
+    GEN done;
+    gel(T->v, 1) = T->forprime.pp;
+    done = parforiter_next(&T->iter, running ? T->v: NULL);
+    if (done) return done;
+  }
+  mt_queue_end(&T->iter.pt);
+  return NULL;
+}
+
+void
+parforprime_stop(parforprime_t *T) { parforiter_stop(&T->iter); }
+
+void
+parforprimestep(GEN a, GEN b, GEN q, GEN code, void *E, long call(void*, GEN, GEN))
 {
   pari_sp av = avma, av2;
   long running, pending = 0;
@@ -1770,7 +1985,7 @@ parforprime(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
   struct pari_mt pt;
   forprime_t T;
 
-  if (!forprime_init(&T, a,b)) { avma = av; return; }
+  if (!forprimestep_init(&T, a,b,q)) { set_avma(av); return; }
   mt_queue_start(&pt, worker);
   v = mkvec(gen_0);
   av2 = avma;
@@ -1786,12 +2001,18 @@ parforprime(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
         br_status = br_NONE;
         stop = gerepileuptoint(av2, gel(done,1));
       }
-    if (!stop) avma = av2;
+    if (!stop) set_avma(av2);
   }
-  avma = av2;
+  set_avma(av2);
   mt_queue_end(&pt);
   br_status = status;
-  avma = av;
+  set_avma(av);
+}
+
+void
+parforprime(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
+{
+  parforprimestep(a, b, NULL, code, E, call);
 }
 
 void
@@ -1800,6 +2021,38 @@ parforprime0(GEN a, GEN b, GEN code, GEN code2)
   parforprime(a, b, code, (void*)code2, code2? gp_evalvoid2: NULL);
 }
 
+void
+parforprimestep0(GEN a, GEN b, GEN q, GEN code, GEN code2)
+{
+  parforprimestep(a, b, q, code, (void*)code2, code2? gp_evalvoid2: NULL);
+}
+
+void
+parforvec_init(parforvec_t *T, GEN x, GEN code, long flag)
+{
+  forvec_init(&T->forvec, x, flag);
+  T->v = mkvec(gen_0);
+  parforiter_init(&T->iter, code);
+}
+
+GEN
+parforvec_next(parforvec_t *T)
+{
+  GEN v = gen_0;
+  while ((v = forvec_next(&T->forvec)) || T->iter.pending)
+  {
+    GEN done;
+    if (v) gel(T->v, 1) = v;
+    done = parforiter_next(&T->iter, v ? T->v: NULL);
+    if (done) return done;
+  }
+  mt_queue_end(&T->iter.pt);
+  return NULL;
+}
+
+void
+parforvec_stop(parforvec_t *T) { parforiter_stop(&T->iter); }
+
 void
 parforvec(GEN x, GEN code, long flag, void *E, long call(void*, GEN, GEN))
 {
@@ -1812,7 +2065,7 @@ parforvec(GEN x, GEN code, long flag, void *E, long call(void*, GEN, GEN))
   forvec_t T;
   GEN a, v = gen_0;
 
-  if (!forvec_init(&T, x, flag)) { avma = av; return; }
+  if (!forvec_init(&T, x, flag)) { set_avma(av); return; }
   mt_queue_start(&pt, worker);
   a = mkvec(gen_0);
   av2 = avma;
@@ -1828,12 +2081,12 @@ parforvec(GEN x, GEN code, long flag, void *E, long call(void*, GEN, GEN))
         br_status = br_NONE;
         stop = gerepilecopy(av2, gel(done,1));
       }
-    if (!stop) avma = av2;
+    if (!stop) set_avma(av2);
   }
-  avma = av2;
+  set_avma(av2);
   mt_queue_end(&pt);
   br_status = status;
-  avma = av;
+  set_avma(av);
 }
 
 void
@@ -1842,6 +2095,96 @@ parforvec0(GEN x, GEN code, GEN code2, long flag)
   parforvec(x, code, flag, (void*)code2, code2? gp_evalvoid2: NULL);
 }
 
+void
+parforeach_init(parforeach_t *T, GEN x, GEN code)
+{
+  switch(typ(x))
+  {
+    case t_LIST:
+      x = list_data(x); /* FALL THROUGH */
+      if (!x) return;
+    case t_MAT: case t_VEC: case t_COL:
+      break;
+    default:
+      pari_err_TYPE("foreach",x);
+      return; /*LCOV_EXCL_LINE*/
+  }
+  T->x = x; T->i = 1; T->l = lg(x);
+  T->W = mkvec(gen_0);
+  T->iter.pending = 0;
+  T->iter.worker = snm_closure(is_entry("_parvector_worker"), mkvec(code));
+  mt_queue_start(&T->iter.pt, T->iter.worker);
+}
+
+GEN
+parforeach_next(parforeach_t *T)
+{
+  while (T->i < T->l || T->iter.pending)
+  {
+    GEN done;
+    long workid;
+    if (T->i < T->l) gel(T->W,1) = gel(T->x, T->i);
+    mt_queue_submit(&T->iter.pt, T->i, T->i < T->l ? T->W: NULL);
+    T->i = minss(T->i+1, T->l);
+    done = mt_queue_get(&T->iter.pt, &workid, &T->iter.pending);
+    if (done) return mkvec2(gel(T->x,workid),done);
+  }
+  mt_queue_end(&T->iter.pt);
+  return NULL;
+}
+
+void
+parforeach_stop(parforeach_t *T) { parforiter_stop(&T->iter); }
+
+void
+parforeach(GEN x, GEN code, void *E, long call(void*, GEN, GEN))
+{
+  pari_sp av = avma, av2;
+  long pending = 0, n, i, stop = 0;
+  long status = br_NONE, workid;
+  GEN worker = snm_closure(is_entry("_parvector_worker"), mkvec(code));
+  GEN done, W;
+  struct pari_mt pt;
+  switch(typ(x))
+  {
+    case t_LIST:
+      x = list_data(x); /* FALL THROUGH */
+      if (!x) return;
+    case t_MAT: case t_VEC: case t_COL:
+      break;
+    default:
+      pari_err_TYPE("foreach",x);
+      return; /*LCOV_EXCL_LINE*/
+  }
+  clone_lock(x); n = lg(x)-1;
+  mt_queue_start_lim(&pt, worker, n);
+  W = cgetg(2, t_VEC);
+  av2 = avma;
+  for (i=1; i<=n || pending; i++)
+  {
+    if (!stop && i <= n) gel(W,1) = gel(x,i);
+    mt_queue_submit(&pt, i, !stop && i<=n? W: NULL);
+    done = mt_queue_get(&pt, &workid, &pending);
+    if (call && done && (!stop || workid < stop))
+      if (call(E, gel(x, workid), done))
+      {
+        status = br_status;
+        br_status = br_NONE;
+        stop = workid;
+      }
+  }
+  set_avma(av2);
+  mt_queue_end(&pt);
+  br_status = status;
+  set_avma(av);
+}
+
+void
+parforeach0(GEN x, GEN code, GEN code2)
+{
+  parforeach(x, code, (void*)code2, code2? gp_evalvoid2: NULL);
+}
+
 void
 closure_callvoid1(GEN C, GEN x)
 {
@@ -1851,6 +2194,18 @@ closure_callvoid1(GEN C, GEN x)
   closure_evalvoid(C);
 }
 
+GEN
+closure_callgen0prec(GEN C, long prec)
+{
+  GEN z;
+  long i, ar = closure_arity(C);
+  for(i=1; i<= ar; i++) gel(st,sp++) = NULL;
+  push_localprec(prec);
+  z = closure_returnupto(C);
+  pop_localprec();
+  return z;
+}
+
 GEN
 closure_callgen1(GEN C, GEN x)
 {
@@ -1908,6 +2263,29 @@ closure_callgenvecprec(GEN C, GEN args, long prec)
   return z;
 }
 
+GEN
+closure_callgenvecdef(GEN C, GEN args, GEN def)
+{
+  long i, l = lg(args)-1, ar = closure_arity(C);
+  st_alloc(ar);
+  if (l > ar)
+    pari_err(e_MISC,"too many parameters in user-defined function call");
+  if (closure_is_variadic(C) && l==ar && typ(gel(args,l))!=t_VEC)
+    pari_err_TYPE("call", gel(args,l));
+  for (i = 1; i <= l;  i++) gel(st,sp++) = def[i] ? gel(args,i): NULL;
+  for(      ; i <= ar; i++) gel(st,sp++) = NULL;
+  return closure_returnupto(C);
+}
+
+GEN
+closure_callgenvecdefprec(GEN C, GEN args, GEN def, long prec)
+{
+  GEN z;
+  push_localprec(prec);
+  z = closure_callgenvecdef(C, args, def);
+  pop_localprec();
+  return z;
+}
 GEN
 closure_callgenall(GEN C, long n, ...)
 {
@@ -1950,11 +2328,7 @@ gp_evalprec(void *E, GEN x, long prec)
 
 long
 gp_evalbool(void *E, GEN x)
-{
-  pari_sp av = avma;
-  long res  = !gequal0(gp_eval(E,x));
-  avma = av; return res;
-}
+{ pari_sp av = avma; return gc_long(av, !gequal0(gp_eval(E,x))); }
 
 long
 gp_evalvoid(void *E, GEN x)
@@ -1991,8 +2365,7 @@ gp_callbool(void *E, GEN x)
 {
   pari_sp av = avma;
   GEN code = (GEN)E;
-  long res  = !gequal0(closure_callgen1(code, x));
-  avma = av; return res;
+  return gc_long(av, !gequal0(closure_callgen1(code, x)));
 }
 
 long
@@ -2181,6 +2554,9 @@ closure_disassemble(GEN C)
     case OCdefaultgen:
       pari_printf("defaultgen\t%ld\n",operand);
       break;
+    case OCpackargs:
+      pari_printf("packargs\t%ld\n",operand);
+      break;
     case OCgetargs:
       pari_printf("getargs\t\t%ld\n",operand);
       break;
@@ -2199,6 +2575,18 @@ closure_disassemble(GEN C)
         pari_printf("localvar0\t%s\n",ep->name);
         break;
       }
+    case OCexportvar:
+      {
+        entree *ep = (entree *)operand;
+        pari_printf("exportvar\t%s\n",ep->name);
+        break;
+      }
+    case OCunexportvar:
+      {
+        entree *ep = (entree *)operand;
+        pari_printf("unexportvar\t%s\n",ep->name);
+        break;
+      }
     case OCcallgen:
       {
         entree *ep = (entree *)operand;
@@ -2268,6 +2656,12 @@ closure_disassemble(GEN C)
     case OCcowvarlex:
       pari_printf("cowvarlex\t%ld\n",operand);
       break;
+    case OCsetref:
+      pari_printf("setref\t\t%ld\n",operand);
+      break;
+    case OClock:
+      pari_printf("lock\t\t%ld\n",operand);
+      break;
     }
   }
 }
@@ -2312,6 +2706,7 @@ opcode_need_relink(op_code opcode)
   case OCcheckargs:
   case OCcheckargs0:
   case OCcheckuserargs:
+  case OCpackargs:
   case OCgetargs:
   case OCdefaultarg:
   case OCdefaultgen:
@@ -2328,6 +2723,8 @@ opcode_need_relink(op_code opcode)
   case OCavma:
   case OCgerepile:
   case OCcowvarlex:
+  case OCsetref:
+  case OClock:
     break;
   case OCpushvar:
   case OCpushdyn:
@@ -2336,6 +2733,8 @@ opcode_need_relink(op_code opcode)
   case OCnewptrdyn:
   case OClocalvar:
   case OClocalvar0:
+  case OCexportvar:
+  case OCunexportvar:
   case OCcallgen:
   case OCcallgen2:
   case OCcalllong:
@@ -2483,5 +2882,5 @@ bincopy_relink(GEN C, GEN V)
   pari_sp av = avma;
   hashtable *table = hash_from_link(gel(V,1),gel(V,2),1);
   gen_relink(C, table);
-  avma = av;
+  set_avma(av);
 }
