@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -83,7 +84,7 @@ static const ulong readonly_constants[] = {
   (ulong)(readonly_constants+4),
   (ulong)(readonly_constants+7)
 };
-THREAD GEN bernzone, primetab;
+THREAD GEN zetazone, bernzone, primetab;
 byteptr diffptr;
 FILE    *pari_outfile, *pari_errfile, *pari_logfile, *pari_infile;
 char    *current_logfile, *current_psfile, *pari_datadir;
@@ -93,7 +94,7 @@ ulong   DEBUGFILES, DEBUGLEVEL, DEBUGMEM;
 long    DEBUGVAR;
 ulong   pari_mt_nbthreads;
 long    precreal;
-ulong   precdl, logstyle;
+ulong   precdl, pari_logstyle;
 gp_data *GP_DATA;
 
 entree  **varentries;
@@ -138,6 +139,7 @@ static void pari_init_timer(void);
 /*#define DEBUG*/
 static THREAD long next_block;
 static THREAD GEN cur_block; /* current block in block list */
+static THREAD GEN root_block; /* current block in block list */
 #ifdef DEBUG
 static THREAD long NUM;
 #endif
@@ -145,7 +147,7 @@ static THREAD long NUM;
 static void
 pari_init_blocks(void)
 {
-  next_block = 0; cur_block = NULL;
+  next_block = 0; cur_block = NULL; root_block = NULL;
 #ifdef DEBUG
   NUM = 0;
 #endif
@@ -157,7 +159,164 @@ pari_close_blocks(void)
   while (cur_block) killblock(cur_block);
 }
 
+static long
+blockheight(GEN bl) { return bl? bl_height(bl): 0; }
+
+static long
+blockbalance(GEN bl)
+{ return bl ? blockheight(bl_left(bl)) - blockheight(bl_right(bl)): 0; }
+
+static void
+fix_height(GEN bl)
+{ bl_height(bl) = maxss(blockheight(bl_left(bl)), blockheight(bl_right(bl)))+1; }
+
+static GEN
+bl_rotright(GEN y)
+{
+  GEN x = bl_left(y), t = bl_right(x);
+  bl_right(x) = y;
+  bl_left(y)  = t;
+  fix_height(y);
+  fix_height(x);
+  return x;
+}
+
+static GEN
+bl_rotleft(GEN x)
+{
+  GEN y = bl_right(x), t = bl_left(y);
+  bl_left(y)  = x;
+  bl_right(x) = t;
+  fix_height(x);
+  fix_height(y);
+  return y;
+}
+
+static GEN
+blockinsert(GEN x, GEN bl, long *d)
+{
+  long b, c;
+  if (!bl)
+  {
+    bl_left(x)=NULL; bl_right(x)=NULL;
+    bl_height(x)=1; return x;
+  }
+  c = cmpuu((ulong)x, (ulong)bl);
+  if (c < 0)
+    bl_left(bl) = blockinsert(x, bl_left(bl), d);
+  else if (c > 0)
+    bl_right(bl) = blockinsert(x, bl_right(bl), d);
+  else return bl; /* ??? Already exist in the tree ? */
+  fix_height(bl);
+  b = blockbalance(bl);
+  if (b > 1)
+  {
+    if (*d > 0) bl_left(bl) = bl_rotleft(bl_left(bl));
+    return bl_rotright(bl);
+  }
+  if (b < -1)
+  {
+    if (*d < 0) bl_right(bl) = bl_rotright(bl_right(bl));
+    return bl_rotleft(bl);
+  }
+  *d = c; return bl;
+}
+
+static GEN
+blockdelete(GEN x, GEN bl)
+{
+  long b;
+  if (!bl) return NULL; /* ??? Do not exist in the tree */
+  if (x < bl)
+    bl_left(bl) = blockdelete(x, bl_left(bl));
+  else if (x > bl)
+    bl_right(bl) = blockdelete(x, bl_right(bl));
+  else
+  {
+    if (!bl_left(bl) && !bl_right(bl)) return NULL;
+    else if (!bl_left(bl)) return bl_right(bl);
+    else if (!bl_right(bl)) return bl_left(bl);
+    else
+    {
+      GEN r = bl_right(bl);
+      while (bl_left(r)) r = bl_left(r);
+      bl_right(r) = blockdelete(r, bl_right(bl));
+      bl_left(r) = bl_left(bl);
+      bl = r;
+    }
+  }
+  fix_height(bl);
+  b = blockbalance(bl);
+  if (b > 1)
+  {
+    if (blockbalance(bl_left(bl)) >= 0) return bl_rotright(bl);
+    else
+    { bl_left(bl) = bl_rotleft(bl_left(bl)); return bl_rotright(bl); }
+  }
+  if (b < -1)
+  {
+    if (blockbalance(bl_right(bl)) <= 0) return bl_rotleft(bl);
+    else
+    { bl_right(bl) = bl_rotright(bl_right(bl)); return bl_rotleft(bl); }
+  }
+  return bl;
+}
+
+static GEN
+blocksearch(GEN x, GEN bl)
+{
+  if (isclone(x)) return x;
+  if (isonstack(x) || is_universal_constant(x)) return NULL;
+  while (bl)
+  {
+    if (x >= bl  && x < bl + bl_size(bl))
+      return bl;
+    bl = x < bl ? bl_left(bl): bl_right(bl);
+  }
+  return NULL; /* Unknown address */
+}
+
+void
+clone_lock(GEN x)
+{
+  GEN y = blocksearch(x, root_block);
+  if (y && isclone(y))
+  {
+    if (DEBUGMEM > 2)
+      err_printf("locking block no %ld: %08lx from %08lx\n", bl_num(y), y, x);
+    ++bl_refc(y);
+  }
+}
+
+void
+clone_unlock(GEN x)
+{
+  GEN y = blocksearch(x, root_block);
+  if (y && isclone(y))
+  {
+    if (DEBUGMEM > 2)
+      err_printf("unlocking block no %ld: %08lx from %08lx\n", bl_num(y), y, x);
+    gunclone(y);
+  }
+}
+
+void
+clone_unlock_deep(GEN x)
+{
+  GEN y = blocksearch(x, root_block);
+  if (y && isclone(y))
+  {
+    if (DEBUGMEM > 2)
+      err_printf("unlocking deep block no %ld: %08lx from %08lx\n", bl_num(y), y, x);
+    gunclone_deep(y);
+  }
+}
+
 /* Return x, where:
+ * x[-8]: AVL height
+ * x[-7]: adress of left child or NULL
+ * x[-6]: adress of right child or NULL
+ * x[-5]: size
  * x[-4]: reference count
  * x[-3]: adress of next block
  * x[-2]: adress of preceding block.
@@ -166,6 +325,7 @@ pari_close_blocks(void)
 GEN
 newblock(size_t n)
 {
+  long d = 0;
   long *x = (long *) pari_malloc((n + BL_HEAD)*sizeof(long)) + BL_HEAD;
 
   bl_size(x) = n;
@@ -174,6 +334,7 @@ newblock(size_t n)
   bl_prev(x) = cur_block;
   bl_num(x)  = next_block++;
   if (cur_block) bl_next(cur_block) = x;
+  root_block = blockinsert(x, root_block, &d);
 #ifdef DEBUG
   err_printf("+ %ld\n", ++NUM);
 #endif
@@ -197,6 +358,7 @@ gunclone(GEN x)
 {
   if (--bl_refc(x) > 0) return;
   BLOCK_SIGINT_START;
+  root_block = blockdelete(x, root_block);
   if (bl_next(x)) bl_prev(bl_next(x)) = bl_prev(x);
   else
   {
@@ -248,6 +410,83 @@ pop_entree_block(entree *ep, long loc)
   gunclone_deep(x); return 1;
 }
 
+/***************************************************************************
+ **                                                                       **
+ **                           Export                                      **
+ **                                                                       **
+ ***************************************************************************/
+
+static hashtable *export_hash;
+static void
+pari_init_export(void)
+{
+  export_hash = hash_create_str(1,0);
+}
+static void
+pari_close_export(void)
+{
+  hash_destroy(export_hash);
+}
+
+/* Exported values are blocks, but do not have the clone bit set so that they
+ * are not affected by clone_lock and ensure_nb, etc. */
+
+void
+export_add(const char *str, GEN val)
+{
+  hashentry *h;
+  val = gclone(val); unsetisclone(val);
+  h = hash_search(export_hash, (void*) str);
+  if (h)
+  {
+    GEN v = (GEN)h->val;
+    h->val = val;
+    setisclone(v); gunclone(v);
+  }
+  else
+    hash_insert(export_hash,(void*)str, (void*) val);
+}
+
+void
+export_del(const char *str)
+{
+  hashentry *h = hash_remove(export_hash,(void*)str);
+  if (h)
+  {
+    GEN v = (GEN)h->val;
+    setisclone(v); gunclone(v);
+    pari_free(h);
+  }
+}
+
+GEN
+export_get(const char *str)
+{
+  return hash_haskey_GEN(export_hash,(void*)str);
+}
+
+void
+unexportall(void)
+{
+  pari_sp av = avma;
+  GEN keys = hash_keys(export_hash);
+  long i, l = lg(keys);
+  for (i = 1; i < l; i++) mt_export_del((const char *)keys[i]);
+  set_avma(av);
+}
+
+void
+exportall(void)
+{
+  long i;
+  for (i = 0; i < functions_tblsz; i++)
+  {
+    entree *ep;
+    for (ep = functions_hash[i]; ep; ep = ep->next)
+      if (EpVALENCE(ep)==EpVAR) mt_export_add(ep->name, (GEN)ep->value);
+  }
+}
+
 /*********************************************************************/
 /*                                                                   */
 /*                       C STACK SIZE CONTROL                        */
@@ -322,11 +561,19 @@ getheap(void)
   traverseheap(&f_getheap, &T); return mkvec2s(T.n, T.l);
 }
 
-void
-traverseheap( void(*f)(GEN, void *), void *data )
+static void
+traverseheap_r(GEN bl, void(*f)(GEN, void *), void *data)
 {
-  GEN x;
-  for (x = cur_block; x; x = bl_prev(x)) f(x, data);
+  if (!bl) return;
+  traverseheap_r(bl_left(bl), f, data);
+  traverseheap_r(bl_right(bl), f, data);
+  f(bl, data);
+}
+
+void
+traverseheap( void(*f)(GEN, void *), void *data)
+{
+  traverseheap_r(root_block,f, data);
 }
 
 /*********************************************************************/
@@ -350,7 +597,8 @@ pari_daemon(void)
         (void)waitpid(pid,NULL,0); /* wait for son to exit, immediate */
         return 1;
   }
-  /* grandson */
+  /* grandson. The silly '!' avoids a gcc-8 warning (unused value) */
+  (void)!freopen("/dev/null","r",stdin);
   return 0;
 }
 #else
@@ -544,7 +792,7 @@ pari_init_defaults(void)
   DEBUGFILES = DEBUGLEVEL = 0;
   DEBUGMEM = 1;
   disable_color = 1;
-  logstyle = logstyle_none;
+  pari_logstyle = logstyle_none;
 
   current_psfile = pari_strdup("pari.ps");
   current_logfile= pari_strdup("pari.log");
@@ -605,11 +853,14 @@ pari_add_defaults_module(entree *ep)
 static void *
 pari_mainstack_malloc(size_t size)
 {
+  void *b;
   /* Check that the system allows reserving "size" bytes. This is just
    * a check, we immediately free the memory. */
-  void *b = mmap(NULL, size, PROT_READ|PROT_WRITE,
-                             MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+  BLOCK_SIGINT_START;
+  b = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+  BLOCK_SIGINT_END;
   if (b == MAP_FAILED) return NULL;
+  BLOCK_SIGINT_START;
   munmap(b, size);
 
   /* Map again, this time with MAP_NORESERVE. On some operating systems
@@ -617,6 +868,7 @@ pari_mainstack_malloc(size_t size)
    * MAP_NORESERVE does not work as expected. */
   b = mmap(NULL, size, PROT_READ|PROT_WRITE,
                        MAP_PRIVATE|MAP_ANONYMOUS|MAP_NORESERVE, -1, 0);
+  BLOCK_SIGINT_END;
   if (b == MAP_FAILED) return NULL;
   return b;
 }
@@ -624,7 +876,9 @@ pari_mainstack_malloc(size_t size)
 static void
 pari_mainstack_mfree(void *s, size_t size)
 {
+  BLOCK_SIGINT_START;
   munmap(s, size);
+  BLOCK_SIGINT_END;
 }
 
 /* Completely discard the memory mapped between the addresses "from"
@@ -648,8 +902,10 @@ pari_mainstack_mreset(pari_sp from, pari_sp to)
   if (!s) return;
 
   addr = (void*)from;
+  BLOCK_SIGINT_START;
   res = mmap(addr, s, PROT_NONE,
              MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS|MAP_NORESERVE, -1, 0);
+  BLOCK_SIGINT_END;
   if (res != addr) pari_err(e_MEM);
 }
 
@@ -660,7 +916,11 @@ static int
 pari_mainstack_mextend(pari_sp from, pari_sp to)
 {
   size_t s = to - from;
-  return mprotect((void*)from, s, PROT_READ|PROT_WRITE);
+  int ret;
+  BLOCK_SIGINT_START;
+  ret = mprotect((void*)from, s, PROT_READ|PROT_WRITE);
+  BLOCK_SIGINT_END;
+  return ret;
 }
 
 /* Set actual stack size to the given size. This sets st->size and
@@ -692,11 +952,15 @@ pari_mainstack_setsize(struct pari_mainstack *st, size_t size)
 static void *
 pari_mainstack_malloc(size_t s)
 {
-  return malloc(s); /* NOT pari_malloc, e_MEM would be deadly */
+  char * tmp;
+  BLOCK_SIGINT_START;
+  tmp = malloc(s); /* NOT pari_malloc, e_MEM would be deadly */
+  BLOCK_SIGINT_END;
+  return tmp;
 }
 
 static void
-pari_mainstack_mfree(void *s, size_t size) { (void) size; free(s); }
+pari_mainstack_mfree(void *s, size_t size) { (void) size; pari_free(s); }
 
 static int
 pari_mainstack_setsize(struct pari_mainstack *st, size_t size)
@@ -763,7 +1027,7 @@ static void
 pari_mainstack_use(struct pari_mainstack *st)
 {
   pari_mainstack = st;
-  avma = st->top;
+  avma = st->top; /* don't use set_avma */
 }
 
 static void
@@ -855,6 +1119,36 @@ new_chunk_resize(size_t x)
 /*                       PARI THREAD                                 */
 /*********************************************************************/
 
+/* Initial PARI thread structure t with a stack of size s and
+ * argument arg */
+
+static void
+pari_thread_set_global(struct pari_global_state *gs)
+{
+  push_localbitprec(gs->bitprec);
+  pari_set_primetab(gs->primetab);
+  pari_set_seadata(gs->seadata);
+  pari_set_varstate(gs->varpriority, &gs->varstate);
+}
+
+static void
+pari_thread_get_global(struct pari_global_state *gs)
+{
+  gs->bitprec = get_localbitprec();
+  gs->primetab = primetab;
+  gs->seadata = pari_get_seadata();
+  varstate_save(&gs->varstate);
+  gs->varpriority = varpriority;
+}
+
+void
+pari_thread_alloc(struct pari_thread *t, size_t s, GEN arg)
+{
+  pari_mainstack_alloc(warnstackthread, &t->st,s,0);
+  pari_thread_get_global(&t->gs);
+  t->data = arg;
+}
+
 /* Initial PARI thread structure t with a stack of size s and virtual size v
  * and argument arg */
 
@@ -862,16 +1156,7 @@ void
 pari_thread_valloc(struct pari_thread *t, size_t s, size_t v, GEN arg)
 {
   pari_mainstack_alloc(warnstackthread, &t->st,s,v);
-  t->data = arg;
-}
-
-/* Initial PARI thread structure t with a stack of size s and
- * argument arg */
-
-void
-pari_thread_alloc(struct pari_thread *t, size_t s, GEN arg)
-{
-  pari_mainstack_alloc(warnstackthread, &t->st,s,0);
+  pari_thread_get_global(&t->gs);
   t->data = arg;
 }
 
@@ -884,6 +1169,8 @@ pari_thread_free(struct pari_thread *t)
 void
 pari_thread_init(void)
 {
+  long var;
+  pari_stackcheck_init((void*)&var);
   pari_init_blocks();
   pari_init_errcatch();
   pari_init_rand();
@@ -892,16 +1179,6 @@ pari_thread_init(void)
   pari_init_compiler();
   pari_init_evaluator();
   pari_init_files();
-  pari_thread_init_primetab();
-  pari_thread_init_seadata();
-}
-
-void
-pari_thread_sync(void)
-{
-  pari_pthread_init_primetab();
-  pari_pthread_init_seadata();
-  pari_pthread_init_varstate();
 }
 
 void
@@ -920,7 +1197,7 @@ pari_thread_start(struct pari_thread *t)
 {
   pari_mainstack_use(&t->st);
   pari_thread_init();
-  pari_thread_init_varstate();
+  pari_thread_set_global(&t->gs);
   return t->data;
 }
 
@@ -976,13 +1253,18 @@ pari_init_opts(size_t parisize, ulong maxprime, ulong init_opts)
   paristack_alloc(parisize, 0);
   init_universal_constants();
   diffptr = NULL;
-  if (!(init_opts&INIT_noPRIMEm))  pari_init_primes(maxprime);
+  if (!(init_opts&INIT_noPRIMEm))
+  {
+    GP_DATA->primelimit = maxprime;
+    pari_init_primes(GP_DATA->primelimit);
+  }
   if (!(init_opts&INIT_noINTGMPm)) pari_kernel_init();
   pari_init_graphics();
-  pari_init_primetab();
-  pari_init_seadata();
   pari_thread_init();
+  pari_set_primetab(NULL);
+  pari_set_seadata(NULL);
   pari_init_functions();
+  pari_init_export();
   pari_var_init();
   pari_init_timer();
   pari_init_buffers();
@@ -1005,6 +1287,7 @@ pari_close_opts(ulong init_opts)
   if ((init_opts&INIT_SIGm)) pari_sig_init(SIG_DFL);
   if (!(init_opts&INIT_noIMTm)) pari_mt_close();
 
+  pari_var_close(); /* must come before destruction of functions_hash */
   for (i = 0; i < functions_tblsz; i++)
   {
     entree *ep = functions_hash[i];
@@ -1014,9 +1297,9 @@ pari_close_opts(ulong init_opts)
       ep = EP;
     }
   }
-  pari_var_close();
   pari_close_mf();
   pari_thread_close();
+  pari_close_export();
   pari_close_files();
   pari_close_homedir();
   if (!(init_opts&INIT_noINTGMPm)) pari_kernel_close();
@@ -1453,7 +1736,7 @@ pari_err2str(GEN e)
       default:  f = op; op = ","; break;
       }
       v = pari_sprintf("%s %s %s %s %s.", what,f,type_dim(x),op,type_dim(y));
-      avma = av; return v;
+      set_avma(av); return v;
     }
   case e_COMPONENT:
     {
@@ -1461,8 +1744,8 @@ pari_err2str(GEN e)
       const char *op= GSTR(gel(e,3));
       GEN l = gel(e,4);
       if (!*f)
-        return pari_sprintf("non-existent component: index %s %Ps",op,l);
-      return pari_sprintf("non-existent component in %s: index %s %Ps",f,op,l);
+        return pari_sprintf("nonexistent component: index %s %Ps",op,l);
+      return pari_sprintf("nonexistent component in %s: index %s %Ps",f,op,l);
     }
   case e_DOMAIN:
     {
@@ -1516,7 +1799,7 @@ pari_err2str(GEN e)
       {
         sprintf(buf, "the %s stack overflows !\n"
             "  current stack size: %lu (%.3f Mbytes)\n"
-            "  [hint] set '%s' to a non-zero value in your GPRC\n",
+            "  [hint] set '%s' to a nonzero value in your GPRC\n",
             stack, (ulong)rsize, (double)rsize/1048576., var);
       }
       return buf;
@@ -1552,21 +1835,20 @@ pari_err_display(GEN err)
   }
   else
   {
-    char *s = pari_err2str(err);
+    char *s;
     closure_err(0);
     err_init_msg(numerr==e_USER);
-    pariErr->puts(s);
+    s = pari_err2str(err); pariErr->puts(s); pari_free(s);
     if (numerr==e_NOTFUNC)
     {
       GEN fun = gel(err,2);
       if (gequalX(fun))
       {
         entree *ep = varentries[varn(fun)];
-        const char *s = ep->name;
-        if (cb_pari_whatnow) cb_pari_whatnow(pariErr,s,1);
+        const char *t = ep->name;
+        if (cb_pari_whatnow) cb_pari_whatnow(pariErr,t,1);
       }
     }
-    pari_free(s);
   }
   out_term_color(pariErr, c_NONE);
   pariErr->flush(); return 0;
@@ -1728,7 +2010,7 @@ list_internal_copy(GEN z, long nmax)
   l = lg(z);
   a = newblock(nmax+1);
   for (i = 1; i < l; i++) gel(a,i) = gel(z,i)? gclone(gel(z,i)): gen_0;
-  a[0] = z[0]; return a;
+  a[0] = z[0]; setisclone(a); return a;
 }
 
 static void
@@ -2004,11 +2286,23 @@ gsizeclone(GEN x) { return (typ(x) == t_INT)? lgefint(x): gsizeclone_i(x); }
 long
 gsizeword(GEN x)
 {
-  GEN L;
-  if (typ(x) != t_LIST) return gsizeclone(x);
-  /* For t_LIST, return the actual list size, gsizeclone() is always 3 */
-  L = list_data(x);
-  return L? 3 + gsizeclone(L): 3;
+  long i, n, lx, tx = typ(x);
+  switch(tx)
+  { /* non recursive types */
+    case t_INT:
+    case t_REAL:
+    case t_STR:
+    case t_VECSMALL: return lg(x);
+
+    case t_LIST:
+      x = list_data(x);
+      return x? 3 + gsizeword(x): 3;
+
+    default:
+      n = lx = lg(x);
+      for (i=lontyp[tx]; i<lx; i++) n += gsizeword(gel(x,i));
+      return n;
+  }
 }
 long
 gsizebyte(GEN x) { return gsizeword(x) * sizeof(long); }
@@ -2117,17 +2411,9 @@ shiftaddress_canon(GEN x, long dec)
       break;
 
     /* one more special case */
-    case t_LIST: {
-      GEN Lx = list_data(x);
-      if (Lx) {
-        pari_sp av = avma;
-        GEN L = (GEN)((long)Lx+dec);
-        shiftaddress_canon(L, dec);
-        list_data(x) = list_internal_copy(L, lg(L)); avma = av;
-      }
-      break;
-    }
-    default:
+    case t_LIST:
+      if (!list_data(x)) break;
+    default: /* Fall through */
       lx = lg(x);
       for (i=lontyp[tx]; i<lx; i++) {
         if (!x[i]) gel(x,i) = gen_0;
@@ -2175,8 +2461,7 @@ obj_insert_shallow(GEN S, long K, GEN O)
   if (typ(v) != t_VEC) pari_err_TYPE("obj_insert", S);
   o = gel(v,K);
   gel(v,K) = O; /*SIGINT: before unclone(o)*/
-  if (isclone(o)) gunclone(o);
-  return gel(v,K);
+  if (isclone(o)) gunclone(o); return gel(v,K);
 }
 
 /* Does S [last position] contain data at position K ? Return it, or NULL */
@@ -2193,7 +2478,7 @@ obj_checkbuild(GEN S, long tag, GEN (*build)(GEN))
 {
   GEN O = obj_check(S, tag);
   if (!O)
-  { pari_sp av = avma; O = obj_insert(S, tag, build(S)); avma = av; }
+  { pari_sp av = avma; O = obj_insert(S, tag, build(S)); set_avma(av); }
   return O;
 }
 
@@ -2204,7 +2489,7 @@ obj_checkbuild_prec(GEN S, long tag, GEN (*build)(GEN,long),
   pari_sp av = avma;
   GEN w = obj_check(S, tag);
   if (!w || pr(w) < prec) w = obj_insert(S, tag, build(S, prec));
-  avma = av; return gcopy(w);
+  set_avma(av); return gcopy(w);
 }
 GEN
 obj_checkbuild_realprec(GEN S, long tag, GEN (*build)(GEN,long), long prec)
@@ -2364,7 +2649,7 @@ gerepile(pari_sp av, pari_sp tetpil, GEN q)
     q = (GEN) (((pari_sp)q) + dec);
 
   for (x = (GEN)av, a = (GEN)tetpil; a > (GEN)av0; ) *--x = *--a;
-  avma = (pari_sp)x;
+  set_avma((pari_sp)x);
   while (x < (GEN)av)
   {
     const long tx = typ(x), lx = lg(x);
@@ -2511,18 +2796,22 @@ timer_start(pari_timer *T)
 # endif
 #endif
 
+/* round microseconds to milliseconds */
 static long
-timer_aux(pari_timer *T, pari_timer *U)
+rndus(long x) { return (x + 500) / 1000; }
+static long
+timer_aux(pari_timer *T, pari_timer *U, void (*settime)(pari_timer *))
 {
-  long s = T->s, us = T->us; timer_start(U);
-  return 1000 * (U->s - s) + (U->us - us + 500) / 1000;
+  long s = T->s, us = T->us;
+  settime(U); return 1000 * (U->s - s) + rndus(U->us - us);
 }
-/* return delay, reset timer */
+
+/* return delay, set timer checkpoint */
 long
-timer_delay(pari_timer *T) { return timer_aux(T, T); }
-/* return delay, don't reset timer */
+timer_delay(pari_timer *T) { return timer_aux(T, T, &timer_start); }
+/* return delay, don't set checkpoint */
 long
-timer_get(pari_timer *T) { pari_timer t; return timer_aux(T, &t); }
+timer_get(pari_timer *T) {pari_timer t; return timer_aux(T, &t, &timer_start);}
 
 static void
 timer_vprintf(pari_timer *T, const char *format, va_list args)
@@ -2557,32 +2846,48 @@ msgtimer(const char *format, ...)
 }
 long
 getabstime(void)  { return timer_get(&abstimer_T);}
-#if defined(USE_CLOCK_GETTIME) || defined(USE_GETTIMEOFDAY) \
- || defined(USE_FTIMEFORWALLTIME)
-static GEN
-timetoi(ulong s, ulong m)
-{
-  pari_sp av = avma;
-  GEN r = addiu(muliu(utoi(s), 1000), m);
-  return gerepileuptoint(av, r);
-}
-#endif
-GEN
-getwalltime(void)
+
+void
+walltimer_start(pari_timer *ti)
 {
 #if defined(USE_CLOCK_GETTIME)
   struct timespec t;
   if (!clock_gettime(CLOCK_REALTIME,&t))
-    return timetoi(t.tv_sec, (t.tv_nsec + 500000)/1000000);
+  { ti->s = t.tv_sec; ti->us = rndus(t.tv_nsec); return; }
 #elif defined(USE_GETTIMEOFDAY)
   struct timeval tv;
   if (!gettimeofday(&tv, NULL))
-    return timetoi(tv.tv_sec, (tv.tv_usec + 500)/1000);
+  {  ti->s = tv.tv_sec; ti->us = tv.tv_usec; return; }
 #elif defined(USE_FTIMEFORWALLTIME)
   struct timeb tp;
-  ftime(&tp); return timetoi(tp.time, tp.millitm);
+  if (!ftime(&tp))
+  { ti->s = tp.time; ti->us = tp.millitm*1000; return; }
 #endif
-  return utoi(getabstime());
+  timer_start(ti);
+}
+/* return delay, set timer checkpoint */
+long
+walltimer_delay(pari_timer *T) { return timer_aux(T, T, &walltimer_start); }
+/* return delay, don't set checkpoint */
+long
+walltimer_get(pari_timer *T)
+{
+  pari_timer t;
+  return timer_aux(T, &t, &walltimer_start);
+}
+
+static GEN
+timetoi(ulong s, ulong m)
+{
+  pari_sp av = avma;
+  return gerepileuptoint(av, addiu(muluu(s, 1000), m));
+}
+GEN
+getwalltime(void)
+{
+  pari_timer ti;
+  walltimer_start(&ti);
+  return timetoi(ti.s, rndus(ti.us));
 }
 
 /*******************************************************************/
