@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -15,7 +16,9 @@ BEGINEXTERN
 
 typedef enum {Gvoid, Gsmall, Gvec, Gvar, Ggen, Gclosure, Gusmall} Gtype;
 
-typedef enum {OCpushlong='A',OCpushgnil,OCpushgen,OCpushreal,OCpushstoi,OCpushvar,
+typedef enum {OCexportvar='!',OCunexportvar,OCsetref,OClock,OCpackargs,
+              OCbitprecreal='@',
+              OCpushlong,OCpushgnil,OCpushgen,OCpushreal,OCpushstoi,OCpushvar,
               OCpop,
               OCstoi,OCitos,OCtostr,OCvarn,OCcopy,OCcopyifclone,
               OCprecreal,OCprecdl,
@@ -33,7 +36,6 @@ typedef enum {OCpushlong='A',OCpushgnil,OCpushgen,OCpushreal,OCpushstoi,OCpushva
               OCavma,OCgerepile,
               OCcowvardyn,OCcowvarlex,
               OCdup,OCstoreptr,OCcheckuserargs,
-              OCitou,OCutoi,OCdefaultulong,
-              OCbitprecreal='@'} op_code;
+              OCitou,OCutoi,OCdefaultulong} op_code;
 
 ENDEXTERN
