@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -80,7 +81,7 @@ forparii(GEN a, GEN b, GEN code)
       set_lex(-1,a);
     }
   }
-  pop_lex(1);  avma = av0;
+  pop_lex(1);  set_avma(av0);
 }
 
 void
@@ -102,7 +103,32 @@ forpari(GEN a, GEN b, GEN code)
     }
     set_lex(-1, a);
   }
-  pop_lex(1); avma = ltop;
+  pop_lex(1); set_avma(ltop);
+}
+
+void
+foreachpari(GEN x, GEN code)
+{
+  long i, l;
+  switch(typ(x))
+  {
+    case t_LIST:
+      x = list_data(x); /* FALL THROUGH */
+      if (!x) return;
+    case t_MAT: case t_VEC: case t_COL:
+      break;
+    default:
+      pari_err_TYPE("foreach",x);
+      return; /*LCOV_EXCL_LINE*/
+  }
+  clone_lock(x); l = lg(x);
+  push_lex(gen_0,code);
+  for (i = 1; i < l; i++)
+  {
+    set_lex(-1, gel(x,i));
+    closure_evalvoid(code); if (loop_break()) break;
+  }
+  pop_lex(1); clone_unlock_deep(x);
 }
 
 /* 0 < a <= b. Using small consecutive chunks to 1) limit memory use, 2) allow
@@ -113,7 +139,7 @@ forfactoredpos(ulong a, ulong b, GEN code)
   const ulong step = 1024;
   pari_sp av = avma;
   ulong x1;
-  for(x1 = a;; x1 += step, avma = av)
+  for(x1 = a;; x1 += step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, lv, x2 = (b >= 2*step && b - 2*step >= x1)? x1-1 + step: b;
     GEN v = vecfactoru(x1, x2);
@@ -152,7 +178,7 @@ forsquarefreepos(ulong a, ulong b, GEN code)
   const ulong step = 1024;
   pari_sp av = avma;
   ulong x1;
-  for(x1 = a;; x1 += step, avma = av)
+  for(x1 = a;; x1 += step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, lv, x2 = (b >= 2*step && b - 2*step >= x1)? x1-1 + step: b;
     GEN v = vecfactorsquarefreeu(x1, x2);
@@ -174,7 +200,7 @@ forsquarefreeneg(ulong a, ulong b, GEN code)
   const ulong step = 1024;
   pari_sp av = avma;
   ulong x2;
-  for(x2 = b;; x2 -= step, avma = av)
+  for(x2 = b;; x2 -= step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, x1 = (x2 >= 2*step && x2-2*step >= a)? x2+1 - step: a;
     GEN v = vecfactorsquarefreeu(x1, x2);
@@ -196,12 +222,20 @@ forsquarefree(GEN a, GEN b, GEN code)
   if (typ(a) != t_INT) pari_err_TYPE("forsquarefree", a);
   if (typ(b) != t_INT) pari_err_TYPE("forsquarefree", b);
   if (cmpii(a,b) > 0) return;
-  s = signe(a);
-  if (s * signe(b) < 0) pari_err_TYPE("forsquarefree [!= signs]", mkvec2(a,b));
-  push_lex(NULL,code);
-  if (s < 0) forsquarefreeneg(itou(b), itou(a), code);
-  else       forsquarefreepos(itou(a), itou(b), code);
-  pop_lex(1); avma = av;
+  s = signe(a); push_lex(NULL,code);
+  if (s < 0)
+  {
+    if (signe(b) <= 0)
+      forsquarefreeneg(itou(b), itou(a), code);
+    else
+    {
+      forsquarefreeneg(1, itou(a), code);
+      forsquarefreepos(1, itou(b), code);
+    }
+  }
+  else
+    forsquarefreepos(itou(a), itou(b), code);
+  pop_lex(1); set_avma(av);
 }
 
 /* convert factoru(n) to factor(-n); M pre-allocated factorization matrix
@@ -232,7 +266,7 @@ forfactoredneg(ulong a, ulong b, GEN code)
   E = cgetg(18, t_COL); gel(E,1) = gen_1;
   M = mkmat2(P,E);
   av = avma;
-  for(x2 = b;; x2 -= step, avma = av)
+  for(x2 = b;; x2 -= step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, x1 = (x2 >= 2*step && x2-2*step >= a)? x2+1 - step: a;
     GEN v = vecfactoru(x1, x2);
@@ -253,7 +287,7 @@ eval0(GEN code)
 {
   pari_sp av = avma;
   set_lex(-1, mkvec2(gen_0, mkmat2(mkcol(gen_0),mkcol(gen_1))));
-  closure_evalvoid(code); avma = av;
+  closure_evalvoid(code); set_avma(av);
   return loop_break();
 }
 void
@@ -269,16 +303,16 @@ forfactored(GEN a, GEN b, GEN code)
   sb = signe(b);
   if (sa < 0)
   {
-    stop = forfactoredneg((sb < 0)? b[2]: 1UL, itou(a), code);
+    stop = forfactoredneg((sb < 0)? uel(b,2): 1UL, itou(a), code);
     if (!stop && sb >= 0) stop = eval0(code);
     if (!stop && sb > 0) forfactoredpos(1UL, b[2], code);
   }
   else
   {
     if (!sa) stop = eval0(code);
-    if (!stop && sb) forfactoredpos(sa? a[2]: 1UL, itou(b), code);
+    if (!stop && sb) forfactoredpos(sa? uel(a,2): 1UL, itou(b), code);
   }
-  pop_lex(1); avma = av;
+  pop_lex(1); set_avma(av);
 }
 void
 whilepari(GEN a, GEN b)
@@ -288,10 +322,10 @@ whilepari(GEN a, GEN b)
   {
     GEN res = closure_evalnobrk(a);
     if (gequal0(res)) break;
-    avma = av;
+    set_avma(av);
     closure_evalvoid(b); if (loop_break()) break;
   }
-  avma = av;
+  set_avma(av);
 }
 
 void
@@ -304,9 +338,9 @@ untilpari(GEN a, GEN b)
     closure_evalvoid(b); if (loop_break()) break;
     res = closure_evalnobrk(a);
     if (!gequal0(res)) break;
-    avma = av;
+    set_avma(av);
   }
-  avma = av;
+  set_avma(av);
 }
 
 static int negcmp(GEN x, GEN y) { return gcmp(y,x); }
@@ -324,8 +358,10 @@ forstep(GEN a, GEN b, GEN s, GEN code)
   switch(typ(s))
   {
     case t_VEC: case t_COL: ss = gsigne(vecsum(s)); v = s; break;
-    case t_INTMOD: a = gadd(a, gmod(gsub(gel(s,2),a), gel(s,1)));
-                   s = gel(s,1);
+    case t_INTMOD:
+      if (typ(a) != t_INT) a = gceil(a);
+      a = addii(a, modii(subii(gel(s,2),a), gel(s,1)));
+      s = gel(s,1);
     default: ss = gsigne(s);
   }
   if (!ss) pari_err_DOMAIN("forstep","step","=",gen_0,s);
@@ -349,23 +385,22 @@ forstep(GEN a, GEN b, GEN s, GEN code)
     }
     set_lex(-1,a);
   }
-  pop_lex(1); avma = av0;
+  pop_lex(1); set_avma(av0);
 }
 
 static void
 _fordiv(GEN a, GEN code, GEN (*D)(GEN))
 {
+  pari_sp av = avma;
   long i, l;
-  pari_sp av2, av = avma;
   GEN t = D(a);
-  push_lex(gen_0,code); l=lg(t); av2 = avma;
+  push_lex(gen_0,code); l = lg(t);
   for (i=1; i<l; i++)
   {
     set_lex(-1,gel(t,i));
     closure_evalvoid(code); if (loop_break()) break;
-    avma = av2;
   }
-  pop_lex(1); avma=av;
+  pop_lex(1); set_avma(av);
 }
 void
 fordiv(GEN a, GEN code) { return _fordiv(a, code, &divisors); }
@@ -407,7 +442,7 @@ _next(forvec_t *d)
   }
 }
 
-/* non-decreasing order [over integers] */
+/* nondecreasing order [over integers] */
 static GEN
 _next_le_i(forvec_t *d)
 {
@@ -433,7 +468,7 @@ _next_le_i(forvec_t *d)
     if (--i <= 0) return NULL;
   }
 }
-/* non-decreasing order [generic] */
+/* nondecreasing order [generic] */
 static GEN
 _next_le(forvec_t *d)
 {
@@ -480,7 +515,7 @@ _next_lt_i(forvec_t *d)
         /* M[i] > M[i-1] >= a[i-1] */
         t = addiu(d->a[i-1],1); if (cmpii(t, d->m[i]) < 0) t = d->m[i];
         d->a[i] = resetloop(d->a[i], t);/*a[i]:=max(a[i-1]+1,m[i]) <= M[i]*/
-        avma = av;
+        set_avma(av);
       }
       return (GEN)d->a;
     }
@@ -606,14 +641,14 @@ forvec(GEN x, GEN code, long flag)
   pari_sp av = avma;
   forvec_t T;
   GEN v;
-  if (!forvec_init(&T, x, flag)) { avma = av; return; }
+  if (!forvec_init(&T, x, flag)) { set_avma(av); return; }
   push_lex((GEN)T.a, code);
   while ((v = forvec_next(&T)))
   {
     closure_evalvoid(code);
     if (loop_break()) break;
   }
-  pop_lex(1); avma = av;
+  pop_lex(1); set_avma(av);
 }
 
 /********************************************************************/
@@ -667,19 +702,18 @@ sum_init(GEN x0, GEN t)
 }
 
 GEN
-suminf(void *E, GEN (*eval)(void *, GEN), GEN a, long prec)
+suminf_bitprec(void *E, GEN (*eval)(void *, GEN), GEN a, long bit)
 {
-  long fl = 0, G = prec2nbits(prec) + 5;
+  long fl = 0, G = bit + 1;
   pari_sp av0 = avma, av;
   GEN x = NULL, _1;
 
   if (typ(a) != t_INT) pari_err_TYPE("suminf",a);
-  a = setloop(a);
-  av = avma;
+  a = setloop(a); av = avma;
   for(;;)
   {
     GEN t = eval(E, a);
-    if (!x) _1 = x = sum_init(real_1(prec), t);
+    if (!x) _1 = x = sum_init(real_1_bit(bit), t);
 
     x = gadd(x,t);
     if (!gequal0(t) && gexpo(t) > gexpo(x)-G)
@@ -696,8 +730,11 @@ suminf(void *E, GEN (*eval)(void *, GEN), GEN a, long prec)
   return gerepileupto(av0, gsub(x, _1));
 }
 GEN
-suminf0(GEN a, GEN code, long prec)
-{ EXPR_WRAP(code, suminf(EXPR_ARG, a, prec)); }
+suminf(void *E, GEN (*eval)(void *, GEN), GEN a, long prec)
+{ return suminf_bitprec(E, eval, a, prec2nbits(prec)); }
+GEN
+suminf0_bitprec(GEN a, GEN code, long bit)
+{ EXPR_WRAP(code, suminf_bitprec(EXPR_ARG, a, bit)); }
 
 GEN
 sumdivexpr(GEN num, GEN code)
@@ -714,8 +751,9 @@ sumdivexpr(GEN num, GEN code)
   }
   pop_lex(1); return gerepileupto(av,y);
 }
+
 GEN
-sumdivmultexpr(GEN num, GEN code)
+sumdivmultexpr(void *D, GEN (*fun)(void*, GEN), GEN num)
 {
   pari_sp av = avma;
   GEN y = gen_1, P,E;
@@ -723,8 +761,7 @@ sumdivmultexpr(GEN num, GEN code)
   long i, l = lg(P);
   GEN (*mul)(GEN,GEN);
 
-  if (l == 1) { avma = av; return gen_1; }
-  push_lex(gen_0, code);
+  if (l == 1) return gc_const(av, gen_1);
   mul = isint? mulii: gmul;
   for (i=1; i<l; i++)
   {
@@ -732,15 +769,18 @@ sumdivmultexpr(GEN num, GEN code)
     long j, e = E[i];
     for (j = 1; j <= e; j++, q = mul(q, p))
     {
-      set_lex(-1, q);
-      z = gadd(z, closure_evalnobrk(code));
+      z = gadd(z, fun(D, q));
       if (j == e) break;
     }
     y = gmul(y, z);
   }
-  pop_lex(1); return gerepileupto(av,y);
+  return gerepileupto(av,y);
 }
 
+GEN
+sumdivmultexpr0(GEN num, GEN code)
+{ EXPR_WRAP(code, sumdivmultexpr(EXPR_ARG, num)) }
+
 /********************************************************************/
 /**                                                                **/
 /**                           PRODUCTS                             **/
@@ -846,7 +886,7 @@ prodeuler(void *E, GEN (*eval)(void *, GEN), GEN a, GEN b, long prec)
   forprime_t T;
 
   av = avma;
-  if (!forprime_init(&T, a,b)) { avma = av; return x; }
+  if (!forprime_init(&T, a,b)) return gc_const(av, x);
 
   av = avma;
   while ( (prime = forprime_next(&T)) )
@@ -896,6 +936,9 @@ vecexpr0(GEN vec, GEN code, GEN pred)
       if (!vec) return cgetg(1, t_VEC);
       break;
     }
+    case t_VECSMALL:
+      vec = vecsmall_to_vec(vec);
+      break;
     case t_VEC: case t_COL: case t_MAT: break;
     default: pari_err_TYPE("[_|_<-_,_]",vec);
   }
@@ -949,7 +992,7 @@ vecteursmall(GEN nmax, GEN code)
   {
     c[2] = i;
     y[i] = gtos(closure_evalnobrk(code));
-    avma = av;
+    set_avma(av);
     set_lex(-1,c);
   }
   pop_lex(1); return y;
@@ -1018,12 +1061,12 @@ delt(GEN g, long n)
 static GEN
 polzag1(long n, long m)
 {
-  const long d = n - m, d2 = d<<1, r = (m+1)>>1, D = (d+1)>>1;
-  long i, k;
+  long d = n - m, i, k, d2, r, D;
   pari_sp av = avma;
   GEN g, T;
 
   if (d <= 0 || m < 0) return pol_0(0);
+  d2 = d << 1; r = (m+1) >> 1, D = (d+1) >> 1;
   g = cgetg(d+2, t_POL);
   g[1] = evalsigne(1)|evalvarn(0);
   T = cgetg(d+1,t_VEC);
@@ -1054,7 +1097,7 @@ polzag1(long n, long m)
   /* sum_{0 <= i < d} g_i x^i * (x+x^2)^r */
   g = RgX_mulXn(gmul(g, gpowgs(deg1pol(gen_1,gen_1,0),r)), r);
   if (!odd(m)) g = delt(g, n);
-  for (i=1; i<=r; i++)
+  for (i = 1; i <= r; i++)
   {
     g = delt(ZX_deriv(g), n);
     if (gc_needed(av,4))
@@ -1069,10 +1112,17 @@ GEN
 polzag(long n, long m)
 {
   pari_sp av = avma;
-  GEN g = ZX_z_unscale(polzag1(n,m), -1);
+  GEN g = polzag1(n,m);
+  if (lg(g) == 2) return g;
+  g = ZX_z_unscale(polzag1(n,m), -1);
   return gerepileupto(av, RgX_Rg_div(g,gel(g,2)));
 }
 
+/*0.39322 > 1/log_2(3+sqrt(8))*/
+static ulong
+sumalt_N(long prec)
+{ return (ulong)(0.39322*(prec2nbits(prec) + 7)); }
+
 GEN
 sumalt(void *E, GEN (*eval)(void *, GEN), GEN a, long prec)
 {
@@ -1081,8 +1131,8 @@ sumalt(void *E, GEN (*eval)(void *, GEN), GEN a, long prec)
   GEN s, az, c, d;
 
   if (typ(a) != t_INT) pari_err_TYPE("sumalt",a);
-  N = (ulong)(0.39322*(prec2nbits(prec) + 7)); /*0.39322 > 1/log_2(3+sqrt(8))*/
-  d = powru(addsr(3, sqrtr(stor(8,prec))), N);
+  N = sumalt_N(prec);
+  d = powru(addsr(3, sqrtr(utor(8,prec))), N);
   d = shiftr(addrr(d, invr(d)),-1);
   a = setloop(a);
   az = gen_m1; c = d;
@@ -1151,32 +1201,32 @@ binsum(GEN S, ulong k, void *E, GEN (*f)(void *, GEN), GEN a,
         long G, long prec)
 {
   long e, i, N = lg(S)-1, l = expu(N / k); /* k 2^l <= N < k 2^(l+1) */
-  pari_sp av;
-  GEN r, t = gen_0;
+  pari_sp av = avma;
+  GEN t = real_0(prec); /* unused unless f(a + k <<l) = 0 */
 
-  gel(S, k << l) = cgetr(prec); av = avma;
   G -= l;
-  r = utoipos(k<<l);
-  for(e=0;;e++) /* compute g(k 2^l) with absolute error ~ 2^(G-l) */
-  {
-    GEN u = gtofp(f(E, addii(a,r)), prec);
+  if (!signe(a)) a = NULL;
+  for (e = 0;; e++)
+  { /* compute g(k 2^l) with absolute error ~ 2^(G-l) */
+    GEN u, r = shifti(utoipos(k), l+e);
+    if (a) r = addii(r, a);
+    u = gtofp(f(E, r), prec);
     if (typ(u) != t_REAL) pari_err_TYPE("sumpos",u);
     if (!signe(u)) break;
     if (!e)
       t = u;
     else {
       shiftr_inplace(u, e);
-      t = addrr(t,u);
-      if (expo(u) < G) break;
+      t = addrr(t,u); if (expo(u) < G) break;
+      if ((e & 0x1ff) == 0) t = gerepileuptoleaf(av, t);
     }
-    r = shifti(r,1);
   }
   gel(S, k << l) = t = gerepileuptoleaf(av, t);
   /* g(j) = 2g(2j) + f(a+j) for all j > 0 */
   for(i = l-1; i >= 0; i--)
   { /* t ~ g(2 * k*2^i) with error ~ 2^(G-i-1) */
     GEN u;
-    av = avma; u = gtofp(f(E, addiu(a, k << i)), prec);
+    av = avma; u = gtofp(f(E, a? addiu(a, k << i): utoipos(k << i)), prec);
     if (typ(u) != t_REAL) pari_err_TYPE("sumpos",u);
     t = addrr(gtofp(u,prec), mpshift(t,1)); /* ~ g(k*2^i) */
     gel(S, k << i) = t = gerepileuptoleaf(av, t);
@@ -1202,9 +1252,9 @@ sumpos(void *E, GEN (*eval)(void *, GEN), GEN a, long prec)
 
   if (typ(a) != t_INT) pari_err_TYPE("sumpos",a);
   a = subiu(a,1);
-  N = (ulong)(0.4*(prec2nbits(prec) + 7));
+  N = sumalt_N(prec);
   if (odd(N)) N++; /* extra precision for free */
-  d = powru(addsr(3, sqrtr(stor(8,prec))), N);
+  d = powru(addsr(3, sqrtr(utor(8,prec))), N);
   d = shiftr(addrr(d, invr(d)),-1);
   az = gen_m1; c = d;
 
@@ -1338,21 +1388,14 @@ GEN
 zbrent0(GEN a, GEN b, GEN code, long prec)
 { EXPR_WRAP(code, zbrent(EXPR_ARG, a, b, prec)); }
 
-/* x = solve_start(&D, a, b, prec)
- * while (x) {
- *   y = ...(x);
- *   x = solve_next(&D, y);
- * }
- * return D.res; */
-
 /* Find zeros of a function in the real interval [a,b] by interval splitting */
 GEN
 solvestep(void *E, GEN (*f)(void *,GEN), GEN a, GEN b, GEN step, long flag, long prec)
 {
   const long ITMAX = 10;
   pari_sp av = avma;
-  GEN fa, ainit, binit;
-  long sainit, it, bit = prec2nbits(prec) / 2, ct = 0, s = gcmp(a,b);
+  GEN fa, a0, b0;
+  long sa0, it, bit = prec2nbits(prec) / 2, ct = 0, s = gcmp(a,b);
 
   if (!s) return gequal0(f(E, a)) ? gcopy(mkvec(a)): cgetg(1,t_VEC);
   if (s > 0) swap(a, b);
@@ -1363,45 +1406,46 @@ solvestep(void *E, GEN (*f)(void *,GEN), GEN a, GEN b, GEN step, long flag, long
   }
   else if (gsigne(step) <= 0)
     pari_err_DOMAIN("solvestep","step","<=",gen_0,step);
-  ainit = a = gtofp(a, prec); fa = f(E, a);
-  binit = b = gtofp(b, prec); step = gtofp(step, prec);
-  sainit = gsigne(fa);
-  if (gexpo(fa) < -bit) sainit = 0;
+  a0 = a = gtofp(a, prec); fa = f(E, a);
+  b0 = b = gtofp(b, prec); step = gtofp(step, prec);
+  sa0 = gsigne(fa);
+  if (gexpo(fa) < -bit) sa0 = 0;
   for (it = 0; it < ITMAX; it++)
   {
     pari_sp av2 = avma;
     GEN v = cgetg(1, t_VEC);
-    long sa;
-    a = ainit;
-    b = binit;
-    sa = sainit;
+    long sa = sa0;
+    a = a0; b = b0;
     while (gcmp(a,b) < 0)
     {
       GEN fc, c = (flag&4)? gmul(a, step): gadd(a, step);
       long sc;
       if (gcmp(c,b) > 0) c = b;
-      fc = f(E, c);
-      sc = gsigne(fc);
+      fc = f(E, c); sc = gsigne(fc);
       if (gexpo(fc) < -bit) sc = 0;
       if (!sc || sa*sc < 0)
       {
+        GEN z = sc? zbrent(E, f, a, c, prec): c;
         long e;
-        GEN z;
-        z = sc? zbrent(E, f, a, c, prec): c;
         (void)grndtoi(z, &e);
-        if (e  <= -bit) ct = 1;
+        if (e <= -bit) ct = 1;
         if ((flag&1) && ((!(flag&8)) || ct)) return gerepileupto(av, z);
-        v = gconcat(v, z);
+        v = shallowconcat(v, z);
       }
       a = c; fa = fc; sa = sc;
+      if (gc_needed(av2,1))
+      {
+        if (DEBUGMEM>1) pari_warn(warnmem,"solvestep");
+        gerepileall(av2, 4, &a ,&fa, &v, &step);
+      }
     }
     if ((!(flag&2) || lg(v) > 1) && (!(flag&8) || ct))
       return gerepilecopy(av, v);
-    step = (flag&4)? sqrtr(sqrtr(step)): gmul2n(step, -2);
+    step = (flag&4)? sqrtnr(step,4): gmul2n(step, -2);
     gerepileall(av2, 2, &fa, &step);
   }
-  if (it == ITMAX) pari_err_IMPL("solvestep recovery [too many iterations]");
-  return NULL;
+  pari_err_IMPL("solvestep recovery [too many iterations]");
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 GEN
@@ -1416,13 +1460,15 @@ struct deriv_data
 {
   GEN code;
   GEN args;
+  GEN def;
 };
 
 static GEN deriv_eval(void *E, GEN x, long prec)
 {
  struct deriv_data *data=(struct deriv_data *)E;
  gel(data->args,1)=x;
- return closure_callgenvecprec(data->code, data->args, prec);
+ uel(data->def,1)=1;
+ return closure_callgenvecdefprec(data->code, data->args, data->def, prec);
 }
 
 /* Rationale: (f(2^-e) - f(-2^-e) + O(2^-b)) / (2 * 2^-e) = f'(0) + O(2^-2e)
@@ -1456,57 +1502,50 @@ derivnum(void *E, GEN (*eval)(void *, GEN, long), GEN x, long prec)
 }
 
 /* Fornberg interpolation algorithm for finite differences coefficients
-* using N+1 equidistant grid points around 0 [ assume N even >= M ].
-* Compute \delta[m]_{N,nu} for all derivation orders m = 0..M such that
-*   h^m * f^{(m)}(0) = \sum_{nu = 0}^n delta[m]_{N,nu}  f(a_nu) + O(h^{N-m+1}),
+* using 2N+1 equidistant grid points around 0 [ assume 2N even >= M ].
+* Compute \delta[m]_{N,i} for all derivation orders m = 0..M such that
+*   h^m * f^{(m)}(0) = \sum_{i = 0}^n delta[m]_{N,i}  f(a_i) + O(h^{N-m+1}),
 * for step size h.
-* Return a = [0,-1,1...,-N2,N2] and vector of vectors d: d[m+1][nu+1]
-* = (M!/m!) * delta[m]_{N,nu}, nu = 0..N */
+* Return a = [0,-1,1...,-N,N] and vector of vectors d: d[m+1][i+1]
+* = w'(a_i) delta[m]_{2N,i}, i = 0..2N */
 static void
-FD(long M, long N, GEN *pd, GEN *pa)
+FD(long M, long N2, GEN *pd, GEN *pa)
 {
-  GEN d, a, b, W, Wp, t, F, Mfact;
-  long N2, m, nu, i;
+  GEN d, a, b, W, F;
+  long N = N2>>1, m, i;
 
-  if (odd(N)) N++; /* make it even */
-  N2 = N>>1;
-  F = cgetg(N+2, t_VEC);
-  a = cgetg(N+2, t_VEC);
-  b = cgetg(N2+1, t_VEC);
+  F = cgetg(N2+2, t_VEC);
+  a = cgetg(N2+2, t_VEC);
+  b = cgetg(N+1, t_VEC);
   gel(a,1) = gen_0;
-  for (i = 1; i <= N2; i++)
+  for (i = 1; i <= N; i++)
   {
     gel(a,2*i)   = utoineg(i);
     gel(a,2*i+1) = utoipos(i);
     gel(b,i) = sqru(i);
   }
   /* w = \prod (X - a[i]) = x W(x^2) */
-  Mfact = mpfact(M);
   W = roots_to_pol(b, 0);
-  Wp = ZX_deriv(W);
-  t = gel(W,2); /* w'(0) */
-  t = diviiexact(t, Mfact);
-  gel(F,1) = RgX_Rg_div(RgX_inflate(W,2), t);
-  for (i = 1; i <= N2; i++)
-  { /* t = w'(a_{2i}) = w'(a_{2i+1}) */
-    GEN r, t = mulii(shifti(gel(b,i),1), poleval(Wp, gel(b,i)));
-    GEN U, S, T;
+  gel(F,1) = RgX_inflate(W,2);
+  for (i = 1; i <= N; i++)
+  {
+    pari_sp av = avma;
+    GEN r, U, S;
     U = RgX_inflate(RgX_div_by_X_x(W, gel(b,i), &r), 2);
-    U = RgX_shift_shallow(U, 1);
-    U = RgXn_red_shallow(U, M+1); /* higher terms not needed */
-    t = diviiexact(t, Mfact);
-    U = RgX_Rg_div(U, t);
-    S = RgX_shift_shallow(U,1);
-    T = RgX_Rg_mul(U, gel(a,2*i+1));
-    gel(F,2*i)   = RgX_sub(S, T);
-    gel(F,2*i+1) = RgX_add(S, T);
+    U = RgXn_red_shallow(U, M); /* higher terms not needed */
+    U = RgX_shift_shallow(U,1); /* w(X) / (X^2-a[i]^2) mod X^(M+1) */
+    S = ZX_sub(RgX_shift_shallow(U,1),
+               ZX_Z_mul(U, gel(a,2*i+1)));
+    S = gerepileupto(av, S);
+    gel(F,2*i)   = S;
+    gel(F,2*i+1) = ZX_z_unscale(S, -1);
   }
-  /* F[i] = M! w(X) / ((X-a[i])w'(a[i])) + O(X^(M+1)) */
+  /* F[i] = w(X) / (X-a[i]) + O(X^(M+1)) in Z[X] */
   d = cgetg(M+2, t_VEC);
   for (m = 0; m <= M; m++)
   {
-    GEN v = cgetg(N+2, t_VEC); /* coeff(F[nu],X^m) */
-    for (nu = 0; nu <= N; nu++) gel(v, nu+1) = gmael(F, nu+1, m+2);
+    GEN v = cgetg(N2+2, t_VEC); /* coeff(F[i],X^m) */
+    for (i = 0; i <= N2; i++) gel(v, i+1) = gmael(F, i+1, m+2);
     gel(d,m+1) = v;
   }
   *pd = d;
@@ -1519,20 +1558,50 @@ chk_ord(long m)
   if (m < 0)
     pari_err_DOMAIN("derivnumk", "derivation order", "<", gen_0, stoi(m));
 }
+/* m! / N! for m in ind; vecmax(ind) <= N */
+static GEN
+vfact(GEN ind, long N, long prec)
+{
+  GEN v, iN;
+  long i, l;
+  ind = vecsmall_uniq(ind); chk_ord(ind[1]); l = lg(ind);
+  iN = invr(itor(mulu_interval(ind[1] + 1, N), prec));
+  v = const_vec(ind[l-1], NULL); gel(v, ind[1]) = iN;
+  for (i = 2; i < l; i++)
+    gel(v, ind[i]) = iN = mulri(iN, mulu_interval(ind[i-1] + 1, ind[i]));
+  return v;
+}
 
+static GEN
+chk_ind(GEN ind, long *M)
+{
+  *M = 0;
+  switch(typ(ind))
+  {
+    case t_INT: ind = mkvecsmall(itos(ind)); break;
+    case t_VECSMALL:
+      if (lg(ind) == 1) return NULL;
+      break;
+    case t_VEC: case t_COL:
+      if (lg(ind) == 1) return NULL;
+      if (RgV_is_ZV(ind)) { ind = ZV_to_zv(ind); break; }
+      /* fall through */
+    default:
+      pari_err_TYPE("derivnum", ind);
+      return NULL; /*LCOV_EXCL_LINE*/
+  }
+  *M = vecsmall_max(ind); chk_ord(*M); return ind;
+}
 GEN
 derivnumk(void *E, GEN (*eval)(void *, GEN, long), GEN x, GEN ind0, long prec)
 {
-  GEN A, D, X, F, ind;
-  long M, fpr, p, i, pr, l, lA, e, ex, eD, newprec;
+  GEN A, C, D, DM, T, X, F, v, ind, t;
+  long M, N, N2, fpr, p, i, pr, l, lA, e, ex, emin, emax, newprec;
   pari_sp av = avma;
   int allodd = 1;
 
-  ind = gtovecsmall(ind0);
-  l = lg(ind);
-  F = cgetg(l, t_VEC);
-  M = vecsmall_max(ind);
-  chk_ord(M);
+  ind = chk_ind(ind0, &M); if (!ind) return cgetg(1, t_VEC);
+  l = lg(ind); F = cgetg(l, t_VEC);
   if (!M) /* silly degenerate case */
   {
     X = eval(E, x, prec);
@@ -1540,16 +1609,35 @@ derivnumk(void *E, GEN (*eval)(void *, GEN, long), GEN x, GEN ind0, long prec)
     if (typ(ind0) == t_INT) F = gel(F,1);
     return gerepilecopy(av, F);
   }
-  FD(M, 3*M-1, &D,&A); /* optimal if 'eval' uses quadratic time */
+  N2 = 3*M - 1; if (odd(N2)) N2++;
+  N = N2 >> 1;
+  FD(M, N2, &D,&A); /* optimal if 'eval' uses quadratic time */
+  C = vecbinomial(N2); DM = gel(D,M);
+  T = cgetg(N2+2, t_VEC);
+  /* (2N)! / w'(i) = (2N)! / w'(-i) = (-1)^(N-i) binom(2*N, N-i) */
+  t = gel(C, N+1);
+  gel(T,1) = odd(N)? negi(t): t;
+  for (i = 1; i <= N; i++)
+  {
+    t = gel(C, N-i+1);
+    gel(T,2*i) = gel(T,2*i+1) = odd(N-i)? negi(t): t;
+  }
+  N = N2 >> 1; emin = LONG_MAX; emax = 0;
+  for (i = 1; i <= N; i++)
+  {
+    e = expi(gel(DM,i)) + expi(gel(T,i));
+    if (e < 0) continue; /* 0 */
+    if (e < emin) emin = e;
+    else if (e > emax) emax = e;
+  }
 
   p = precision(x);
   fpr = p ? prec2nbits(p): prec2nbits(prec);
-  eD = gexpo(gel(D,M));
   e = (fpr + 3*M*log2((double)M)) / (2*M);
   ex = gexpo(x);
   if (ex < 0) ex = 0; /* near 0 */
   pr = (long)ceil(fpr + e * M); /* ~ 3fpr/2 */
-  newprec = nbits2prec(pr + eD + ex + BITS_IN_LONG);
+  newprec = nbits2prec(pr + (emax - emin) + ex + BITS_IN_LONG);
   switch(typ(x))
   {
     case t_REAL:
@@ -1564,21 +1652,21 @@ derivnumk(void *E, GEN (*eval)(void *, GEN, long), GEN x, GEN ind0, long prec)
   for (i = allodd? 2: 1; i < lA; i++)
   {
     GEN t = eval(E, gadd(x, gmul2n(gel(A,i), -e)), newprec);
+    t = gmul(t, gel(T,i));
     if (!gprecision(t))
       t = is_scalar_t(typ(t))? gtofp(t, newprec): gmul(t, real_1(newprec));
-    gel(X, i) = t;
+    gel(X,i) = t;
   }
 
+  v = vfact(ind, N2, nbits2prec(fpr + 32));
   for (i = 1; i < l; i++)
   {
-    GEN t;
-    long m = ind[i]; chk_ord(m);
-    t = gmul2n(RgV_dotproduct(gel(D,m+1), X), e*m);
-    if (m < M) t = gdiv(t, mulu_interval(m+1,M));
-    gel(F,i) = t;
+    long m = ind[i];
+    GEN t = RgV_dotproduct(gel(D,m+1), X);
+    gel(F,i) = gmul(t, gmul2n(gel(v, m), e*m));
   }
   if (typ(ind0) == t_INT) F = gel(F,1);
-  return gerepilecopy(av, gprec_w(F, nbits2prec(fpr)));
+  return gerepilecopy(av, F);
 }
 /* v(t') */
 static long
@@ -1600,27 +1688,23 @@ derivfunk(void *E, GEN (*eval)(void *, GEN, long), GEN x, GEN ind0, long prec)
   case t_REAL: case t_INT: case t_FRAC: case t_COMPLEX:
     return derivnumk(E,eval, x, ind0, prec);
   case t_POL:
-    ind = gtovecsmall(ind0);
-    M = vecsmall_max(ind);
+    ind = chk_ind(ind0,&M); if (!ind) return cgetg(1,t_VEC);
     xp = RgX_deriv(x);
     x = RgX_to_ser(x, precdl+2 + M * (1+RgX_val(xp)));
     break;
   case t_RFRAC:
-    ind = gtovecsmall(ind0);
-    M = vecsmall_max(ind);
+    ind = chk_ind(ind0,&M); if (!ind) return cgetg(1,t_VEC);
     x = rfrac_to_ser(x, precdl+2 + M * (1+rfrac_val_deriv(x)));
     xp = derivser(x);
     break;
   case t_SER:
-    ind = gtovecsmall(ind0);
-    M = vecsmall_max(ind);
+    ind = chk_ind(ind0,&M); if (!ind) return cgetg(1,t_VEC);
     xp = derivser(x);
     break;
   default: pari_err_TYPE("numerical derivation",x);
     return NULL; /*LCOV_EXCL_LINE*/
   }
-  av = avma; chk_ord(M);
-  vx = varn(x);
+  av = avma; vx = varn(x);
   ixp = M? ginv(xp): NULL;
   F = cgetg(M+2, t_VEC);
   gel(F,1) = eval(E, x, prec);
@@ -1679,10 +1763,10 @@ laurentseries(void *E, GEN (*f)(void*,GEN x, long), long M, long v, long prec)
     s = f(E, s, prec);
     if (typ(s) != t_SER || varn(s) != v) pari_err_TYPE("laurentseries", s);
     vr = valp(s);
-    if (M < vr) { avma = av; return zeroser(v, M); }
+    if (M < vr) { set_avma(av); return zeroser(v, M); }
     dr = lg(s) + vr - 3 - M;
     if (dr >= 0) return gerepileupto(av, s);
-    avma = av; d -= dr;
+    set_avma(av); d -= dr;
   }
 }
 static GEN
@@ -1706,59 +1790,45 @@ derivnum0(GEN a, GEN code, GEN ind, long prec)
 { EXPR_WRAP(code, derivfunk(EXPR_ARGPREC,a,ind,prec)); }
 
 GEN
-derivfun0(GEN code, GEN args, long prec)
+derivfun0(GEN args, GEN def, GEN code, long k, long prec)
 {
+  pari_sp av = avma;
   struct deriv_data E;
-  E.code=code; E.args=args;
-  return derivfun((void*)&E, deriv_eval, gel(args,1), prec);
+  GEN z;
+  E.code=code; E.args=args; E.def=def;
+  z = gel(derivfunk((void*)&E, deriv_eval, gel(args,1), mkvecs(k), prec),1);
+  return gerepilecopy(av, z);
 }
 
 /********************************************************************/
 /**                   Numerical extrapolation                      **/
 /********************************************************************/
-/* for t_CLOSURE */
-static double
-fun_getmf(long mul)
-{
-  const double A[] = {0,
-    0.331,0.260,0.228,0.210,0.198,
-    0.188,0.181,0.175,0.170,0.166 };
-  const double B[] = {0,
-    0.166,0.142,0.132,0.125,0.120,
-    0.116,0.113,0.111,0.109,0.107 };
-  if (mul <=  10) return A[mul];
-  if (mul <= 109) return B[mul/10]; else return 0.105;
-}
-/* for t_VEC */
-static double
-vec_getmf(long mul)
-{
-  const double A[] = {0,
-    0.2062, 0.1760, 0.1613, 0.1519, 0.1459,
-    0.1401, 0.1360, 0.1327, 0.1299, 0.1280 };
-  const double B[] = {0,
-    0.1280, 0.1133, 0.1064, 0.1019, 0.0987,
-    0.0962, 0.0942, 0.0925, 0.0911, 0.0899 };
-  if (mul <=  10) return A[mul];
-  if (mul <= 109) return B[mul/10]; else return 0.0899;
-}
-
-/* [u(n*muli), u <= N], muli = 1 unless f!=NULL */
+/* [u(n), u <= N] */
 static GEN
-get_u(void *E, GEN (*f)(void *, GEN, long), long N, long muli, long prec)
+get_u(void *E, GEN (*f)(void *, GEN, long), long N, long prec)
 {
   long n;
-  GEN u = cgetg(N+1, t_VEC);
+  GEN u;
   if (f)
   {
-    for (n = 1; n <= N; n++) gel(u,n) = f(E, stoi(muli*n), prec);
+    GEN v = f(E, utoipos(N), prec);
+    u = cgetg(N+1, t_VEC);
+    if (typ(v) != t_VEC || lg(v) != N+1) { gel(u,N) = v; v = NULL; }
+    else
+    {
+      GEN w = f(E, gen_1, LOWDEFAULTPREC);
+      if (typ(w) != t_VEC || lg(w) != 2) { gel(u,N) = v; v = NULL; }
+    }
+    if (v) u = v;
+    else
+      for (n = 1; n < N; n++) gel(u,n) = f(E, utoipos(n), prec);
   }
   else
   {
     GEN v = (GEN)E;
     long t = lg(v)-1;
-    if (t < N*muli) pari_err_COMPONENT("limitnum","<",stoi(N), stoi(t));
-    for (n = 1; n <= N; n++) gel(u,n) = gel(v, muli*n);
+    if (t < N) pari_err_COMPONENT("limitnum","<",stoi(N), stoi(t));
+    u = vecslice(v, 1, N);
   }
   for (n = 1; n <= N; n++)
   {
@@ -1770,132 +1840,250 @@ get_u(void *E, GEN (*f)(void *, GEN, long), long N, long muli, long prec)
 
 struct limit
 {
-  long prec0; /* target accuracy */
   long prec; /* working accuracy */
   long N; /* number of terms */
-  GEN u; /* sequence to extrapolate */
   GEN na; /* [n^alpha, n <= N] */
-  GEN nma; /* [n^-alpha, n <= N] or NULL (alpha = 1) */
   GEN coef; /* or NULL (alpha != 1) */
 };
 
-static void
-limit_init(struct limit *L, void *E, GEN (*f)(void*,GEN,long),
-           long muli, GEN alpha, long prec)
+static GEN
+_gi(void *E, GEN x)
 {
-  long bitprec = prec2nbits(prec), n, N;
-  GEN na;
+  GEN A = (GEN)E, y = gsubgs(x, 1);
+  if (gequal0(y)) return A;
+  return gdiv(gsubgs(gpow(x, A, LOWDEFAULTPREC), 1), y);
+}
+static GEN
+_g(void *E, GEN x)
+{
+  GEN D = (GEN)E, A = gel(D,1), T = gel(D,2);
+  const long prec = LOWDEFAULTPREC;
+  return gadd(glog(x,prec), intnum((void*)A, _gi, gen_0, gaddgs(x,1), T, prec));
+}
 
-  if (muli <= 0) muli = 20;
-  L->N = N = (long)ceil((f? fun_getmf(muli): vec_getmf(muli)) * bitprec);
-  L->prec = nbits2prec(bitprec + (long)ceil(1.844*N));
-  L->prec0 = prec;
-  L->u = get_u(E, f, N, muli, L->prec);
-  if (alpha && !gequal1(alpha))
+/* solve log(b) + int_0^{b+1} (x^(1/a)-1) / (x-1) dx = 0, b in [0,1]
+ * return -log_2(b), rounded up */
+static double
+get_accu(GEN a)
+{
+  pari_sp av = avma;
+  const long prec = LOWDEFAULTPREC;
+  const double We2 = 1.844434455794; /* (W(1/e) + 1) / log(2) */
+  GEN b, T;
+  if (!a) return We2;
+  if (typ(a) == t_INT) switch(itos_or_0(a))
   {
-    long prec2 = gprecision(alpha);
-    GEN nma;
-    if (!prec2) prec2 = L->prec;
-    na = vecpowug(N, alpha, prec2);
-    L->coef = NULL;
-    L->nma = nma = cgetg(N+1, t_VEC);
-    for (n = 1; n <= N; n++) gel(nma, n) = ginv(gel(na, n));
-    if (muli != 1) na = gmul(na, gpow(utor(muli,prec2), alpha, prec2));
+    case 1: return We2;
+    case 2: return 1.186955309668;
+    case 3: return 0.883182331990;
+  }
+  else if (typ(a) == t_FRAC && equali1(gel(a,1))) switch(itos_or_0(gel(a,2)))
+  {
+    case 2: return 2.644090500290;
+    case 3: return 3.157759214459;
+    case 4: return 3.536383237500;
+  }
+  T = intnuminit(gen_0, gen_1, 0, prec);
+  b = zbrent((void*)mkvec2(ginv(a), T), &_g, dbltor(1E-5), gen_1, prec);
+  return gc_double(av, -dbllog2r(b));
+}
+
+static double
+get_c(GEN a)
+{
+  double A = a? gtodouble(a): 1.0;
+  if (A <= 0) pari_err_DOMAIN("limitnum","alpha","<=",gen_0, a);
+  if (A >= 2) return 0.2270;
+  if (A >= 1) return 0.3318;
+  if (A >= 0.5) return 0.6212;
+  if (A >= 0.3333) return 1.2;
+  return 3; /* only tested for A >= 0.25 */
+}
+static void
+limit_Nprec(struct limit *L, GEN alpha, long prec)
+{
+  long bit = prec2nbits(prec);
+  L->N = ceil(get_c(alpha) * bit);
+  L->prec = nbits2prec(bit + (long)ceil(get_accu(alpha) * L->N));
+}
+/* solve x - a log(x) = b; a, b >= 3 */
+static double
+solvedivlog(double a, double b) { return dbllemma526(a,1,1,b); }
+
+/* #u > 1, prod_{j != i} u[i] - u[j] */
+static GEN
+proddiff(GEN u, long i)
+{
+  pari_sp av = avma;
+  long l = lg(u), j;
+  GEN p = NULL;
+  if (i == 1)
+  {
+    p = gsub(gel(u,1), gel(u,2));
+    for (j = 3; j < l; j++)
+      p = gmul(p, gsub(gel(u,i), gel(u,j)));
   }
   else
   {
-    GEN coef, C = vecbinomial(N), T = vecpowuu(N, N);
-    na = cgetg(N+1, t_VEC);
-    L->coef = coef = cgetg(N+1, t_VEC);
-    L->nma = NULL;
-    for (n = 1; n <= N; n++)
-    {
-      GEN c = mulii(gel(C,n+1), gel(T,n));
-      if (odd(N-n)) togglesign_safe(&c);
-      gel(coef, n) = c;
-      gel(na, n) = utoipos(n*muli);
+    p = gsub(gel(u,i), gel(u,1));
+    for (j = 2; j < l; j++)
+      if (j != i) p = gmul(p, gsub(gel(u,i), gel(u,j)));
+  }
+  return gerepileupto(av, p);
+}
+static GEN
+vecpows(GEN u, long N)
+{
+  long i, l;
+  GEN v = cgetg_copy(u, &l);
+  for (i = 1; i < l; i++) gel(v,i) = gpowgs(gel(u,i), N);
+  return v;
+}
+
+static void
+limit_init(struct limit *L, GEN alpha, int asymp)
+{
+  long n, N = L->N, a = 0;
+  GEN c, v, T = NULL;
+
+  if (!alpha) a = 1;
+  else if (typ(alpha) == t_INT)
+  {
+    a = itos_or_0(alpha);
+    if (a > 2) a = 0;
+  }
+  else if (typ(alpha) == t_FRAC)
+  {
+    long na = itos_or_0(gel(alpha,1)), da = itos_or_0(gel(alpha,2));
+    if (da && na && da <= 4 && na <= 4)
+    { /* don't bother with other cases */
+      long e = (N-1) % da, k = (N-1) / da;
+      if (e) { N += da - e; k++; } /* N = 1 (mod d) => simpler ^ (n/d)(N-1) */
+      L->N = N;
+      T = vecpowuu(N, na * k);
     }
   }
-  L->na = na;
+  L->coef = v = cgetg(N+1, t_VEC);
+  if (!a)
+  {
+    long prec2 = gprecision(alpha);
+    GEN u;
+    if (prec2 && prec2 < L->prec) alpha = gprec_w(alpha, L->prec);
+    L->na = u = vecpowug(N, alpha, L->prec);
+    if (!T) T = vecpows(u, N-1);
+    for (n = 1; n <= N; n++) gel(v,n) = gdiv(gel(T,n), proddiff(u,n));
+    return;
+  }
+  L->na = asymp? vecpowuu(N, a): NULL;
+  c = mpfactr(N-1, L->prec);
+  if (a == 1)
+  {
+    c = invr(c);
+    gel(v,1) = c; if (!odd(N)) togglesign(c);
+    for (n = 2; n <= N; n++) gel(v,n) = divru(mulrs(gel(v,n-1), n-1-N), n);
+  }
+  else
+  { /* a = 2 */
+    c = invr(mulru(sqrr(c), (N*(N+1)) >> 1));
+    gel(v,1) = c; if (!odd(N)) togglesign(c);
+    for (n = 2; n <= N; n++) gel(v,n) = divru(mulrs(gel(v,n-1), n-1-N), N+n);
+  }
+  T = vecpowuu(N, a*N);
+  for (n = 2; n <= N; n++) gel(v,n) = mulri(gel(v,n), gel(T,n));
 }
 
 /* Zagier/Lagrange extrapolation */
 static GEN
-limitnum_i(struct limit *L)
+limitnum_i(struct limit *L, GEN u, long prec)
+{ return gprec_w(RgV_dotproduct(u,L->coef), prec); }
+GEN
+limitnum(void *E, GEN (*f)(void *, GEN, long), GEN alpha, long prec)
 {
   pari_sp av = avma;
-  GEN S;
-  if (L->nma)
-    S = polint(L->nma, L->u,gen_0,NULL);
-  else
-    S = gdiv(RgV_dotproduct(L->u,L->coef), mpfact(L->N));
-  return gerepilecopy(av, gprec_w(S, L->prec0));
-}
-GEN
-limitnum(void *E, GEN (*f)(void *, GEN, long), long muli, GEN alpha, long prec)
-{
   struct limit L;
-  limit_init(&L, E,f, muli, alpha, prec);
-  return limitnum_i(&L);
+  GEN u;
+  limit_Nprec(&L, alpha, prec);
+  limit_init(&L, alpha, 0);
+  u = get_u(E, f, L.N, L.prec);
+  return gerepilecopy(av, limitnum_i(&L, u, prec));
 }
-GEN
-limitnum0(GEN u, long muli, GEN alpha, long prec)
+typedef GEN (*LIMIT_FUN)(void*,GEN,long);
+static LIMIT_FUN get_fun(GEN u, const char *s)
 {
-  void *E = (void*)u;
-  GEN (*f)(void*,GEN,long) = NULL;
   switch(typ(u))
   {
-    case t_COL:
-    case t_VEC: break;
-    case t_CLOSURE: f = gp_callprec; break;
-    default: pari_err_TYPE("limitnum", u);
+    case t_COL: case t_VEC: break;
+    case t_CLOSURE: return gp_callprec;
+    default: pari_err_TYPE(s, u);
   }
-  return limitnum(E,f, muli,alpha, prec);
+  return NULL;
 }
+GEN
+limitnum0(GEN u, GEN alpha, long prec)
+{ return limitnum((void*)u,get_fun(u, "limitnum"), alpha, prec); }
 
 GEN
-asympnum(void *E, GEN (*f)(void *, GEN, long), long muli, GEN alpha, long prec)
+asympnum(void *E, GEN (*f)(void *, GEN, long), GEN alpha, long prec)
 {
   const long MAX = 100;
   pari_sp av = avma;
-  GEN u, vres = vectrunc_init(MAX);
+  GEN u, A = cgetg(MAX+1, t_VEC);
   long i, B = prec2nbits(prec);
   double LB = 0.9*expu(B); /* 0.9 and 0.95 below are heuristic */
   struct limit L;
-  limit_init(&L, E,f, muli, alpha, prec);
+  limit_Nprec(&L, alpha, prec);
   if (alpha) LB *= gtodouble(alpha);
-  u = L.u;
+  limit_init(&L, alpha, 1);
+  u = get_u(E, f, L.N, L.prec);
   for(i = 1; i <= MAX; i++)
   {
-    GEN a, s, v, p, q;
+    GEN a, v, q, s = limitnum_i(&L, u, prec);
     long n;
-    s = limitnum_i(&L);
     /* NOT bestappr: lindep properly ignores the lower bits */
     v = lindep_bit(mkvec2(gen_1, s), maxss((long)(0.95*floor(B - i*LB)), 32));
     if (lg(v) == 1) break;
-    p = negi(gel(v,1));
-    q = gel(v,2);
-    if (!signe(q)) break;
-    a = gdiv(p,q);
+    q = gel(v,2); if (!signe(q)) break;
+    a = gdiv(negi(gel(v,1)), q);
     s = gsub(s, a);
     /* |s|q^2 > eps */
     if (!gequal0(s) && gexpo(s) + 2*expi(q) > -17) break;
-    vectrunc_append(vres, a);
+    gel(A,i) = a;
     for (n = 1; n <= L.N; n++) gel(u,n) = gmul(gsub(gel(u,n), a), gel(L.na,n));
   }
-  return gerepilecopy(av, vres);
+  setlg(A,i); return gerepilecopy(av, A);
 }
 GEN
-asympnum0(GEN u, long muli, GEN alpha, long prec)
+asympnumraw(void *E, GEN (*f)(void *,GEN,long), long LIM, GEN alpha, long prec)
 {
-  void *E = (void*)u;
-  GEN (*f)(void*,GEN,long) = NULL;
-  switch(typ(u))
+  pari_sp av = avma;
+  double c, d, al;
+  long i, B;
+  GEN u, A;
+  struct limit L;
+
+  if (LIM < 0) return cgetg(1, t_VEC);
+  c = get_c(alpha);
+  d = get_accu(alpha);
+  al = alpha? gtodouble(alpha): 1.0;
+  B = prec2nbits(prec);
+  L.N = ceil(solvedivlog(c * al * LIM / M_LN2, c * B));
+  L.prec = nbits2prec(ceil(B + L.N / c + d * L.N));
+  limit_init(&L, alpha, 1);
+  u = get_u(E, f, L.N, L.prec);
+  A = cgetg(LIM+2, t_VEC);
+  for(i = 0; i <= LIM; i++)
   {
-    case t_COL:
-    case t_VEC: break;
-    case t_CLOSURE: f = gp_callprec; break;
-    default: pari_err_TYPE("asympnum", u);
+    GEN a = RgV_dotproduct(u,L.coef);
+    long n;
+    for (n = 1; n <= L.N; n++)
+      gel(u,n) = gprec_wensure(gmul(gsub(gel(u,n), a), gel(L.na,n)), L.prec);
+    gel(A,i+1) = gprec_wtrunc(a, prec);
   }
-  return asympnum(E,f, muli,alpha, prec);
+  return gerepilecopy(av, A);
 }
+GEN
+asympnum0(GEN u, GEN alpha, long prec)
+{ return asympnum((void*)u,get_fun(u, "asympnum"), alpha, prec); }
+GEN
+asympnumraw0(GEN u, long LIM, GEN alpha, long prec)
+{ return asympnumraw((void*)u,get_fun(u, "asympnumraw"), LIM, alpha, prec); }
