@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -111,7 +112,7 @@ ellcondfile(long f)
   sprintf(s, "%s/elldata/ell%ld", pari_datadir, n);
   F = pari_fopengz(s);
   if (!F) pari_err_FILE("elldata file",s);
-  avma = av;
+  set_avma(av);
   V = gp_read_stream(F->file);
   if (!V || typ(V)!=t_VEC ) pari_err_FILE("elldata file [read]",s);
   pari_fclose(F); return V;
@@ -130,7 +131,7 @@ ellcondlist(long f)
     GEN v = gel(V,i);
     return vecslice(v,2, lg(v)-1);
   }
-  avma = av; return cgetg(1,t_VEC);
+  set_avma(av); return cgetg(1,t_VEC);
 }
 
 static GEN
@@ -145,7 +146,7 @@ ellsearchbyname(GEN V, char *name)
   }
   x = strtoGENstr(name);
   pari_err_DOMAIN("ellsearchbyname", "name", "=", x,x);
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 static GEN
@@ -183,7 +184,7 @@ ellsearch(GEN A)
       pari_err_TYPE("ellsearch",A);
   } else {
     pari_err_TYPE("ellsearch",A);
-    return NULL;
+    return NULL;/*LCOV_EXCL_LINE*/
   }
   if (f <= 0) pari_err_DOMAIN("ellsearch", "conductor", "<=", gen_0,stoi(f));
   V = ellcondlist(f);
@@ -216,7 +217,7 @@ ellidentify(GEN E)
     if (ZV_equal(gmael(V,j,2), M))
       return gerepilecopy(ltop, mkvec2(gel(V,j),gel(G,2)));
   pari_err_BUG("ellidentify [missing curve]");
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 GEN
@@ -260,14 +261,10 @@ forell(void *E, long call(void*, GEN), long a, long b, long flag)
         if (call(E, e)) return;
       }
     }
-    avma = ltop;
+    set_avma(ltop);
   }
 }
 
 void
 forell0(long a, long b, GEN code, long flag)
-{
-  push_lex(gen_0, code);
-  forell((void*)code, &gp_evalvoid, a, b, flag);
-  pop_lex(1);
-}
+{ EXPRVOID_WRAP(code, forell(EXPR_ARGVOID, a, b, flag)) }
