@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -250,7 +251,7 @@ discpart(GEN H, GEN p, long prec)
   for(i = 1; i < lg(list); i++)
   {
     GEN t = gel(list,i);
-    for(j = 3; j < lg(t); j++) /* include if non-constant mod p */
+    for(j = 3; j < lg(t); j++) /* include if nonconstant mod p */
       if (!valp(gel(t,j))) { prod = RgX_mul(prod,t); break; }
   }
   if (degpol(prod) != 3) pari_err_BUG("discpart [prod degree]");
@@ -505,7 +506,6 @@ polymini_zi2(GEN pol)
   return mkvecsmall2(t6, beta);
 }
 
-
 struct igusa {
   GEN j2, i4, j4, j6, j8, j10, i12;
   GEN a0, A2, A3, A5, B2;
@@ -739,7 +739,7 @@ get_nr(long d, long a, long b, long *n, long *r)
   ssQ_red(a, b, &A,&B);
   c = d / ugcd(d, B);
   *n = B * c;
-  *r = smodss(A * c, *n);
+  *r = umodsu(A * c, *n);
 }
 /* n = lcm(denom(a/b), denom(c/d)); r = (a/b * n mod n); q = (c/d * n mod n);
  * assume b > 0 and d > 0 */
@@ -751,8 +751,8 @@ get_nrq(long a, long b, long c, long d, long *n, long *r, long *q)
   ssQ_red(c, d, &C,&D);
   g = ugcd(B,D);
   *n = B * (D/g);
-  *r = smodss(A * (D/g), *n);
-  *q = smodss(C * (B/g), *n);
+  *r = umodsu(A * (D/g), *n);
+  *q = umodsu(C * (B/g), *n);
 }
 
 /* Ip->tt = 1 */
@@ -818,7 +818,7 @@ tame_234_init(struct igusa *I, struct igusa_p *Ip, long *n, long *q, long *r)
   else if (120*va0+5*v12 > 36*va5 && 60*vb2 >= 12*va5+5*v12)
   {
     ssQ_red(36*va5-25*v12,240, q,n);
-    *r = smodss(-2* *q, *n);
+    *r = umodsu(-2* *q, *n);
   }
   else /* 6*va0+v12 > 9*vb2 && 12*va5+5*v12 > 60*vb2 */
   {
@@ -1080,7 +1080,7 @@ tame_567_init(struct igusa *I, struct igusa_p *Ip, long dk,
     *pn = 2*ddk;
     *pd = 2*ndk;
   }
-  *pdm = smodss(*pd, *pn);
+  *pdm = umodsu(*pd, *pn);
 }
 
 static long
