@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -18,7 +19,7 @@ static long
 groupelts_sumorders(GEN S)
 {
   long i, s = 0;
-  for(i=1; i < lg(S); i++) s += perm_order(gel(S,i));
+  for(i=1; i < lg(S); i++) s += perm_orderu(gel(S,i));
   return s;
 }
 
@@ -56,7 +57,7 @@ indexgroupcentre(GEN G, GEN Z, const long *good, const long *bad)
   for(i=1;i<lg(Z);i++)
   {
     GEN z=gel(Z,i);
-    if (perm_order(z)==2)
+    if (perm_orderu(z)==2)
     {
       pari_sp btop=avma;
       GEN H = cyclicgroup(z,2);
@@ -64,7 +65,7 @@ indexgroupcentre(GEN G, GEN Z, const long *good, const long *bad)
       GEN Q = quotient_group(C,G);
       const long *p;
       long idx=group_ident(Q,NULL);
-      avma=btop;
+      set_avma(btop);
       for(p=good;*p;p++)
         if (*p==idx) return 1;
       for(p=bad;*p;p++)
@@ -513,7 +514,7 @@ group_ident(GEN G, GEN S)
   long idx = group_ident_i(G, S);
   if (idx < 0) pari_err_TYPE("group_ident [not a group]", G);
   if (!idx) pari_err_IMPL("galoisidentify for groups of order > 127");
-  avma = av; return idx;
+  return gc_long(av, idx);
 }
 
 long
