@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -420,8 +421,7 @@ LLL_1stPass(GEN *pB0, GEN kappa, baker_s *BS, GEN *pBx)
 
 /* add solution (x,y) if not already known */
 static void
-add_sol(GEN *pS, GEN x, GEN y)
-{ *pS = shallowconcat(*pS, mkvec(mkvec2(x,y))); }
+add_sol(GEN *pS, GEN x, GEN y) { *pS = vec_append(*pS, mkvec2(x,y)); }
 /* z = P(p,q), d = deg P, |z| = |rhs|. Check signs and (possibly)
  * add solutions (p,q), (-p,-q) */
 static void
@@ -635,7 +635,7 @@ MiddleSols(GEN *pS, GEN bound, GEN roo, GEN P, GEN rhs, long s, GEN c1)
       av = avma;
       z = poleval(ZX_rescale(P,q), p); /* = P(p/q) q^dep(P) */
       Q = dvmdii(rhs, z, &R);
-      if (R != gen_0) { avma = av; continue; }
+      if (R != gen_0) { set_avma(av); continue; }
       setabssign(Q);
       if (Z_ispowerall(Q, d, &Q))
       {
@@ -650,7 +650,7 @@ MiddleSols(GEN *pS, GEN bound, GEN roo, GEN P, GEN rhs, long s, GEN c1)
       /* the theoretical value is bit_prec = gexpo(ro)+1+log2(bound) */
       prec = precdbl(precision(ro));
       if (DEBUGLEVEL>1) pari_warn(warnprec,"thue",prec);
-      roo = realroots(P, NULL, prec);
+      roo = ZX_realroots_irred(P, prec);
       if (lg(roo)-1 != s) pari_err_BUG("thue [realroots]");
       k--;
     }
@@ -722,7 +722,7 @@ SmallSols(GEN S, GEN x3, GEN poly, GEN rhs)
     for (j = l-2; j >= 2; j -= 2) togglesign( gel(P,j) );
     if (j == 0) gel(P,2) = subii(gel(P,2), rhs2);
     check_y_root(&S, P, utoineg(y));
-    if (lS == lg(S)) { avma = av2; continue; } /* no solution found */
+    if (lS == lg(S)) { set_avma(av2); continue; } /* no solution found */
 
     if (gc_needed(av,1))
     {
@@ -798,12 +798,12 @@ thueinit(GEN pol, long flag, long prec)
     delete_var(); delete_var(); delete_var();
     return gerepilecopy(av, tnf);
   }
-  /* POL monic: POL(x) = C pol(x/L), L integer */
+  /* POL monic irreducible: POL(x) = C pol(x/L), L integer */
   POL = ZX_primitive_to_monic(POL, &L);
   C = gdiv(powiu(L, dpol), gel(pol, dpol+2));
   pol = POL;
 
-  s = ZX_sturm(pol);
+  s = ZX_sturm_irred(pol);
   if (s)
   {
     long PREC, n = degpol(pol);
@@ -828,7 +828,7 @@ thueinit(GEN pol, long flag, long prec)
       if (( tnf = inithue(pol, bnf, flag, PREC) )) break;
       PREC = precdbl(PREC);
       if (DEBUGLEVEL>1) pari_warn(warnprec,"thueinit",PREC);
-      bnf = NULL; avma = av;
+      bnf = NULL; set_avma(av);
     }
   }
   else
@@ -1149,7 +1149,7 @@ START:
             !TrySol(&S, B0, i1, Delta2, Lambda, ro, Lmu, NE,MatFU,prinfo,
                     P,rhs))
           goto START;
-        if (lg(S) == lS) avma = av2;
+        if (lg(S) == lS) set_avma(av2);
       }
       else
       {
@@ -1203,12 +1203,13 @@ maybe_warn(GEN bnf, GEN a, GEN Ind)
   if (!is_pm1(Ind) && !is_pm1(bnf_get_no(bnf)) && !is_pm1(a))
     pari_warn(warner, "The result returned by 'thue' is conditional on the GRH");
 }
+static GEN bnfisintnorm_fa(GEN bnf, GEN a, GEN fa, long sa);
 /* return solutions of Norm(x) = a mod U(K)^+ */
 static GEN
-get_ne(GEN bnf, GEN a, GEN Ind)
+get_ne(GEN bnf, GEN a, GEN fa, GEN Ind)
 {
   if (DEBUGLEVEL) maybe_warn(bnf,a,Ind);
-  return bnfisintnorm(bnf, a);
+  return bnfisintnorm_fa(bnf, a, fa, signe(a));
 }
 /* return solutions of |Norm(x)| = |a| mod U(K) */
 static GEN
@@ -1253,14 +1254,12 @@ sols_from_R(GEN Rab, GEN *pS, GEN P, GEN POL, GEN rhs)
     if (typ(gel(ry,k)) == t_INT) check_y(pS, P, POL, gel(ry,k), rhs);
 }
 static GEN
-Z_factor_if_easy(GEN rhs)
+absZ_factor_if_easy(GEN rhs, GEN x3)
 {
-  GEN F, P;
-  long l;
-  if (expi(rhs) < 150) return Z_factor(rhs);
-  F = Z_factor_limit(rhs, 500000);
-  P = gel(F,1); l = lg(P);
-  return (l == 1 || BPSW_psp(gel(P,l-1)))? F: NULL;
+  GEN F, U;
+  if (expi(rhs) < 150 || expo(x3) >= BITS_IN_LONG) return absZ_factor(rhs);
+  F = absZ_factor_limit_strict(rhs, 500000, &U);
+  return U? NULL: F;
 }
 /* Given a tnf structure as returned by thueinit, a RHS and
  * optionally the solutions to the norm equation, returns the solutions to
@@ -1279,11 +1278,11 @@ thue(GEN tnf, GEN rhs, GEN ne)
   /* solve P(x,y) = rhs <=> POL(L x, y) = C rhs, with POL monic in Z[X] */
   POL = gel(tnf,1);
   C = gel(POL,2); rhs = gmul(C, rhs);
-  if (typ(rhs) != t_INT) { avma = av; return cgetg(1, t_VEC); }
+  if (typ(rhs) != t_INT) { set_avma(av); return cgetg(1, t_VEC); }
   if (!signe(rhs))
   {
     GEN v = gel(tnf,2);
-    avma = av;
+    set_avma(av);
     /* at least 2 irreducible factors, one of which has degree 1 */
     if (typ(v) == t_VECSMALL && v[1] ==1)
       pari_err_DOMAIN("thue","#sols","=",strtoGENstr("oo"),rhs);
@@ -1293,8 +1292,12 @@ thue(GEN tnf, GEN rhs, GEN ne)
   POL = gel(POL,1);
   if (lg(tnf) == 8)
   {
-    if (!ne) ne = get_ne(tnf_get_bnf(tnf), rhs, tnf_get_Ind(tnf));
-    if (lg(ne) == 1) { avma = av; return cgetg(1, t_VEC); }
+    if (!ne)
+    {
+      GEN F = absZ_factor(rhs);
+      ne = get_ne(tnf_get_bnf(tnf), rhs, F, tnf_get_Ind(tnf));
+    }
+    if (lg(ne) == 1) { set_avma(av); return cgetg(1, t_VEC); }
     S = LargeSols(POL, tnf, rhs, ne);
   }
   else if (typ(gel(tnf,3)) == t_REAL)
@@ -1306,12 +1309,12 @@ thue(GEN tnf, GEN rhs, GEN ne)
     S = cgetg(1,t_VEC);
     if (!ne && typ(bnf) == t_VEC && expo(x3) > 10)
     {
-      F = Z_factor_if_easy(rhs);
-      if (F) ne = get_ne(bnf, F, gen_1);
+      F = absZ_factor_if_easy(rhs, x3);
+      if (F) ne = get_ne(bnf, rhs, F, gen_1);
     }
     if (ne)
     {
-      if (lg(ne) == 1) { avma = av; return cgetg(1,t_VEC); }
+      if (lg(ne) == 1) { set_avma(av); return cgetg(1,t_VEC); }
       if (degpol(POL) == 2) /* quadratic imaginary */
       {
         GEN u = NULL;
@@ -1319,7 +1322,7 @@ thue(GEN tnf, GEN rhs, GEN ne)
         if (typ(bnf) == t_VEC)
         {
           u = bnf_get_tuU(bnf);
-          w =  bnf_get_tuN(bnf);
+          w = bnf_get_tuN(bnf);
         }
         else
         {
@@ -1352,7 +1355,7 @@ thue(GEN tnf, GEN rhs, GEN ne)
   {
     GEN bnf, ne1 = NULL, ne2 = NULL;
     long e = itos( gel(tnf,3) );
-    if (!Z_ispowerall(rhs, e, &rhs)) { avma = av; return cgetg(1, t_VEC); }
+    if (!Z_ispowerall(rhs, e, &rhs)) { set_avma(av); return cgetg(1, t_VEC); }
     tnf = gel(tnf,2);
     bnf = tnf_get_bnf(tnf);
     ne = get_neabs(bnf, rhs, lg(tnf)==8?tnf_get_Ind(tnf): gen_1);
@@ -1435,7 +1438,6 @@ test_sol(struct sol_abs *T, long i)
   {
     err_printf("sol = %Ps\n",s);
     if (T->partrel) err_printf("T->partrel = %Ps\n",T->partrel);
-    err_flush();
   }
 }
 /* partrel[i] <-- partrel[i-1] + u[i] * rel[i] */
@@ -1450,7 +1452,7 @@ fix_partrel(struct sol_abs *T, long i)
   long k, l = lg(part1);
   for (k=1; k < l; k++)
     affii(addii(gel(part0,k), muliu(gel(rel,k), u)), gel(part1,k));
-  avma = av;
+  set_avma(av);
 }
 
 /* Recursive loop. Suppose u[1..i] has been filled
@@ -1502,7 +1504,7 @@ isintnorm_loop(struct sol_abs *T, long i)
     if (T->partrel) {
       pari_sp av = avma;
       gaffect(ZC_add(gel(T->partrel,i), gel(T->rel,i)), gel(T->partrel,i));
-      avma = av;
+      set_avma(av);
     }
   }
 }
@@ -1532,8 +1534,7 @@ get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
     for (j=2; gcd > 1 && j < lL; j++) gcd = ugcd(gcd, pr_get_f(gel(L,j)));
     if (gcd > 1 && vn % gcd)
     {
-      if (DEBUGLEVEL>2)
-      { err_printf("gcd f_P  does not divide n_p\n"); err_flush(); }
+      if (DEBUGLEVEL>2) err_printf("gcd f_P  does not divide n_p\n");
       return 0;
     }
     v = (i==nP)? 0: nPR + lL;
@@ -1599,7 +1600,7 @@ get_unit_1(GEN bnf, GEN *unit)
 
   if (DEBUGLEVEL > 2) err_printf("looking for a fundamental unit of norm -1\n");
   if (odd(n)) { *unit = gen_m1; return 1; }
-  v = nfsign_units(bnf, NULL, 0);
+  v = nfsign_fu(bnf, NULL);
   for (i = 1; i < lg(v); i++)
     if ( Flv_sum( gel(v,i), 2) ) { *unit = gel(bnf_get_fu(bnf), i); return 1; }
   return 0;
@@ -1682,6 +1683,11 @@ bnfisintnorm_i(GEN bnf, GEN a, long sa, GEN z)
   }
   setlg(z, j); return z;
 }
+/* bnfisintnorm sa * |a|, fa = factor(|a|) */
+static GEN
+bnfisintnorm_fa(GEN bnf, GEN a, GEN fa, long sa)
+{ return bnfisintnorm_i(bnf, a, sa, bnfisintnormabs(bnf, mkvec2(a, fa)));
+}
 GEN
 bnfisintnorm(GEN bnf, GEN a)
 {
@@ -1692,5 +1698,5 @@ bnfisintnorm(GEN bnf, GEN a)
     case t_VEC: a = gel(a,1); break;
     case t_MAT: a = factorback(a); break;
   }
-  return gerepilecopy(av, bnfisintnorm_i(bnf,a,signe(a), ne));
+  return gerepilecopy(av, bnfisintnorm_i(bnf, a, signe(a), ne));
 }
