@@ -6,7 +6,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -25,16 +26,35 @@ mtsingle_queue_get(struct mt_state *mt, long *workid, long *pending)
   return done;
 }
 
+static int single_is_thread = 0;
+static long single_trace_level = 0;
+
 static void
 mtsingle_queue_submit(struct mt_state *mt, long workid, GEN work)
 {
+  single_is_thread = 1;
   mt->pending = work? closure_callgenvec(mt->worker, work): NULL;
+  single_is_thread = 0;
   mt->workid = workid;
 }
 
 static void
 mtsingle_queue_end(void) {  }
 
+int
+mtsingle_is_thread(void) { return single_is_thread; }
+
+void
+mtsingle_err_recover(long er)
+{
+  (void) er;
+  if (single_is_thread)
+  {
+    evalstate_set_trace(single_trace_level);
+    single_is_thread = 0;
+  }
+}
+
 void
 mtsingle_queue_start(struct pari_mt *pt, GEN worker)
 {
@@ -43,6 +63,7 @@ mtsingle_queue_start(struct pari_mt *pt, GEN worker)
   pt->end = mtsingle_queue_end;
   pt->mt.worker = worker;
   pt->mt.pending = NULL;
+  single_trace_level = evalstate_get_trace();
 }
 
 void
@@ -61,21 +82,27 @@ mt_queue_start(struct pari_mt *pt, GEN worker)
 { return mt_queue_start_lim(pt, worker, 0); }
 
 void
-mtstate_save(long *pending)
+mtstate_save(struct pari_mtstate *mt)
 {
-  *pending = mt_is_parallel();
+  mt->is_thread = single_is_thread;
+  mt->trace_level = single_trace_level;
+  mt->pending_threads = mt_is_parallel();
 }
 
 void
-mtstate_restore(long *pending)
+mtstate_restore(struct pari_mtstate *mt)
 {
-  if (!*pending && mt_is_parallel())
+  single_is_thread = mt->is_thread;
+  single_trace_level = mt->trace_level;
+  if (!mt->pending_threads && mt_is_parallel())
     mt_queue_reset();
 }
 
 void
 mtstate_reset(void)
 {
+  single_is_thread = 0;
+  single_trace_level = 0;
   if (mt_is_parallel())
     mt_queue_reset();
 }
