@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -23,8 +24,7 @@ static const char * pariwin32_basedir = NULL;
 const char*
 win32_basedir(void)
 {
-  if (pariwin32_basedir) return pariwin32_basedir;
-  else
+  if (!pariwin32_basedir)
   {
     char basedir[1024];
     char* slash;
@@ -32,8 +32,8 @@ win32_basedir(void)
     slash = strrchr(basedir, '\\');
     if (slash) slash[1] = 0;
     pariwin32_basedir = strdup(basedir);
-    return pariwin32_basedir;
   }
+  return pariwin32_basedir;
 }
 
 char*
@@ -41,7 +41,7 @@ win32_datadir(void)
 {
   char datadir[1029];
   const char * basedir = win32_basedir();
-  sprintf(datadir, "%sdata",basedir);
+  sprintf(datadir, "%sdata", basedir);
   return strdup(datadir);
 }
 
